/*
 ftrig.h
  for fractional integer routines (alternative to floating-point)

  Created: 2008.07.15
  Revised: 2024.06.07

23.08.13 - f64to80; moved scal10 to ftrig
23.08.08 - added _umul128 for 32-bit systems
23.08.07 - cntlz64 available on 32-bit systems
20.04.25 - fractZConvolve
19.03.09 - NDFractionReduce
19.02.15 - umuldivrem
19.02.08 - vfract30mul
17.11.26 - added macro constants
15.07.02 - u32AngleToDeg/degToU32Angle
14.03.24 - an optimized rfix16div...
14.03.17 - ufix32muldiv, ufix16div
14.03.15 - added ufixmuldiv
14.03.13 - added fractsin, public csrhlut
14.02.22-24 - POCC fixes
14.01.26 - added rfix16div
	fixed M68K ifix16div - now mimics fp overflow, ifixed not uint32.
13.11.02 - added ifract32mul
13.05.15 - added fractasin
13.05.10 - fixed POCC __inline _cntlz32
13.03.30
13.01.08 - ifix16div (handles overflow; no rounding)
11.06.24 - moved types to enviro64.h, added ufix16mul
*/

#ifndef __FTRIG_H
 #if defined(__POCC__)
  #pragma warn(push)
  #pragma warn(disable:2007 2096)
 #endif

#include "enviro64.h"

#if defined(__GNUC__) && !defined(__forceinline)
 #define __forceinline __attribute__((always_inline))
#endif

/* ---- typedefs */

#include <pshpack2.h>

/* support for this is limited in compilers though x64 supports basic x87 */
/* and x64 has support for operations that deal with 128 bit integers */

typedef struct s_float80
{
#if BIG_ENDIAN
	int16		sgnexp;
	uint64		man;
#else
	uint64		man;
	int16		sgnexp;
#endif
} s_float80;

typedef union u_fhalf16
{
#ifdef _MSC_EXTENSIONS
	struct
	{
	#if BIG_ENDIAN
		signed		sgn:1;
		unsigned	exp:5, man:10;
	#else
		unsigned	man:10, exp:5;
		signed		sgn:1;
	#endif
	};
#endif
	uint16		raw;
} u_fhalf16;

#include <poppack.h>

/* ---- macros/inlines */

#ifndef HUGE_VAL
 #ifdef env68k
 typedef float float_t;
 typedef long double double_t;
  #define sizeof_double_t 10
 #elif __POCC_FLTEVAL__ == 1
 typedef double float_t;
 typedef double double_t;
 #elif __POCC_FLTEVAL__ == 2
 typedef long double float_t;
 typedef long double double_t;
 #else /* __POCC_FLTEVAL__ <= 0 */
 typedef float float_t;
 typedef double double_t;
 #endif
#endif

#ifndef sizeof_double_t
  #define sizeof_double_t 8
#endif

#if TARGET_OS_MAC
 #define INFINITY		__inf()
 double_t scalb(double_t,
	long);
#elif __POCC__
 #define scalb			scalbn
#endif

#ifndef M_PI
 #define M_PI		(3+.14159265358979323846264338327950288)
#endif
#ifndef M_LN10
 #define M_LN10		(2+.30258509299404568401799145468436421)
#endif
#ifndef M_LN2
 #define M_LN2		(.5+0.19314718055994530941723212145817657)
#endif
#ifndef M_E
 #define M_E		(2.5+.2182818284590452353602874713526625)
#endif
#ifndef M_PHI
 #define M_PHI		(1.5+.1180339887498948482045868343656381)
#endif
#ifndef M_LOG10_2
 #define M_LOG10_2	(.25+0.05102999566398119521373889472449)
#endif

/* puts together a y coordinate and a y-fraction for a starting line */
/* inverts the fraction for a fuller line w. less of a fraction, uint32-clipped */
 #define rLineFrac(d,k,f) (void)(d=(uint16)~(f)|(k)<<16,d-=!++d)

/* returns IMIN(ifixed) on overflow */
 #define ifix16div(x,y) (ifixed)((ifixed)(x)*65536./(ifixed)(y))
/* rounds; returns 0 on overflow */
 #define ufix16div(x,y) (ufixed)((ufixed)(x)*65536./(ufixed)(y)+.5)
/* fraction compensated rate division (unsigned, unrounded) */
 #define rfix16div(x,y,f) (ufixed)(((ufixed)(x)*65536.-(ufixed)(f))/(ufixed)(y))

 #define ufract32div(x,y) (uint32)(((uint64)(uint32)(x)<<32)/(uint32)(y))

 #define u32AngleToDeg(theta) ((uint32)theta*(45/536870912.))
 uint32 degToU32Angle(double theta);

 #ifdef env68k
  #pragma parameter __D0 ufract32mul(__D0, __D1)
  extern uint32 ufract32mul(uint32, uint32) = {
/*	MULU.L	D0,D0:D1 */	0x4C00, 0x1400
  };
  /* overflow will yield zero, as 0x100000000 would yield the same */
  #undef ufract32div
  #pragma parameter __D0 uhl32div(__D1,__D0,__D2)
  extern uint32 uhl32div(uint32, uint32, uint32) = {
/*	DIVU.L	D2,D1:D0 */	0x4C42, 0x401
  };
  #define ufract32div(a,b) uhl32div(a,0,b)
  #pragma parameter __D0 ifract32mul(__D0, __D1)
  extern int32 ifract32mul(int32, int32) = {
/*	MULS.L	D0,D0:D1 */	0x4C00, 0x1C00
  };

  #pragma parameter __D0 ifract30mul(__D0, __D1)
  extern int32 ifract30mul(int32, int32) = {
/*	MULS.L	D0,D0:D1 */	0x4C00, 0x1C00,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180
  };
  #pragma parameter __D0 ufract30mul(__D0, __D1)
  extern uint32 ufract30mul(uint32, uint32) = {
/*	MULU.L	D0,D0:D1 */	0x4C00, 0x1400,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180
  };

  #pragma parameter __D1 ifract30div(__D0,__D2)
  extern int32 ifract30div(int32, int32) = {
/*	MOVEQ	#3,D1	*/	0x7203,
/*	AND.L	D0,D1	*/	0xC280,
/*	ROR.L	#2,D1	*/	0xE499,
/*	ASR.L	#2,D0	*/	0xE480,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	BPL.S	*+8	;@1 */	0x6A06,
/*	NEG.L	(A7)	*/	0x4497,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080,
/*@1	TDIVS.L	(A7),D2:D0 */	0x4C57, 0x802,
/*	DIVS.L	(A7)+,D2:D1 */	0x4C5F, 0x1C02
  };
  #pragma parameter __D1 ufract30div(__D0,__D2)
  extern uint32 ufract30div(uint32, uint32) = {
/*	MOVEQ	#3,D1	*/	0x7203,
/*	AND.L	D0,D1	*/	0xC280,
/*	ROR.L	#2,D1	*/	0xE499,
/*	LSR.L	#2,D0	*/	0xE488,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	TDIVU.L	D2,D2:D0 */	0x4C42, 2,
/*	DIVU.L	(A7)+,D2:D1 */	0x4C5F, 0x1402
  };

  #undef ifix16div
  #pragma parameter __D0 ifix16div(__D1,__D2)
  extern ifixed ifix16div(ifixed, ifixed) = {
/*	SWAP	D1	*/	0x4841,
/*	MOVE.L	D1,D0	*/	0x2001,
/*	EXT.L	D1	*/	0x48C1,
/*	CLR	D0	*/	0x4240,
/*	DIVS.L	D2,D1:D0 */	0x4C42, 0xC01,
/*	BVC.S	end	*/	0x6806,
/*	MOVE.L	#$80000000,D0*/	0x203C, 0x8000, 0
  };
  #undef ufix16div
  #pragma parameter __D0 ufix16div(__D1,__D2)
  extern ufixed ufix16div(ufixed, ufixed) = {
/*	SWAP	D1	*/	0x4841,
/*	MOVE.L	D1,D0	*/	0x2001,
/*	EXT.L	D1	*/	0x48C1,
/*	CLR	D0	*/	0x4240,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	LSR.L	#1,D2	*/	0xE28A,
/*	ADD.L	D2,D0	*/	0xD082,
/*	MOVEQ	#0,D2	*/	0x7400,
/*	ADDX.L	D2,D1	*/	0xD382,
/*	DIVU.L	(A7)+,D1:D0 */	0x4C5F, 0x401,
/*	BVC.S	end	*/	0x6802,
/*	MOVEQ	#0,D0	*/	0x7000
  };
  #undef rfix16div
  #pragma parameter __D0 rfix16div(__D0,__D2,__D1)
  extern ufixed rfix16div(ufixed, ufixed, ufixed) = {
/*	MOVEA.L	D3,A0	*/	0x2043,
/*	SWAP	D0	*/	0x4840,
/*	MOVEQ	#0,D3	*/	0x7600,
/*	MOVE	D0,D3	*/	0x3600,
/*	CLR	D0	*/	0x4240,
/*	SUB.L	D1,D0	*/	0x9081,
/*	MOVEQ	#0,D1	*/	0x7200,
/*	SUBX.L	D1,D3	*/	0x9781,
/*	DIVU.L	D2,D3:D0 */	0x4C42, 0x403,
/*	BVC.S	@1	*/	0x6802,
/*	MOVEQ	#0,D0	*/	0x7000,
/*@1	MOVE.L	A0,D3	*/	0x2608
  };
  #pragma parameter __D0 ufix16mul(__D0, __D2)
  extern ufixed ufix16mul(ufixed, ufixed) = {
/*	MULU.L	D2,D2:D0 */	0x4C02, 0x402,
/*	ADD	D0,D0	*/	0xD040,
/*	MOVE	D2,D0	*/	0x3002,
/*	MOVEQ	#0,D2	*/	0x7400,
/*	SWAP	D0	*/	0x4840,
/*	ADDX.L	D2,D0	*/	0xD182
  };
 #elif LONG_LONG
  #if __MRC__
   #define ufract32mul __mulhwu
  #elif defined(__POCC__) && envx86
   static __forceinline uint32 ufract32mul(uint32 a,
	uint32 b)
   {
	__asm
	{
		mov	eax,a
		mul	b
		mov	a,edx
	}
	return a;
   }
   static __forceinline uint32 ufract30mul(uint32 a,
	uint32 b)
   {
	__asm
	{
		mov	eax,a
		mul	b
		shld	edx,eax,2
		mov	a,edx
	}
	return a;
   }
   #undef ufract32div
   static __forceinline uint32 ufract32div(uint32 n,
	uint32 d)
   {
	__asm
	{
		mov	edx,n
		xor	eax,eax
		div	d
		mov	d,eax
	}
	return d;
   }
   #if __POCC__<1100 /* can't inline this one, don't bother */
   #undef rfix16div
   static __forceinline ufixed rfix16div(ufixed x,
	ufixed y,
	ufixed f)
   {
	__asm
	{
		mov	eax,x
		mov	edx,eax
		shr	edx,16
		shl	eax,16
		sub	eax,f
		sbb	edx,0
		cmp	y,edx
		ja	ao

		and	eax,0
		jmp	eo

	ao:	div	y
	eo:	mov	x,eax
	}
	return x;
   }
   #endif
  #else
   #define ufract32mul(x,y) ((uint32)((uint64)(uint32)(x)*(uint32)(y)>>32))
  #endif

  #define ifract32mul(x,y) ((int32)((int64)(int32)(x)*(int32)(y)>>32))

  #define ifract30mul(x,y) ((int32)((int64)(int32)(x)*(int32)(y)>>30))
  #define ufract30mul(x,y) ((uint32)((uint64)(uint32)(x)*(uint32)(y)>>30))

  #define ifract30div(x,y) (int32)(((int64)(int32)(x)<<30)/(int32)(y))
  #define ufract30div(x,y) (uint32)(((uint64)(uint32)(x)<<30)/(uint32)(y))

  #define ufix16mul(x,y) (ufixed)((uint64)(ufixed)(x)*(ufixed)(y)+32768>>16)
 #else
  #define ufract32mul(x,y) (uint32)((uint32)(x)*(1/4294967296.)*(uint32)(y))

  #define ifract32mul(x,y) (int32)((int32)(x)*(1./4294967296.)*(int32)(y))

  #define ifract30mul(x,y) (int32)((int32)(x)*(1/1073741824.)*(int32)(y))
  #define ufract30mul(x,y) (uint32)((uint32)(x)*(1/1073741824.)*(uint32)(y))

  #define ifract30div(x,y) (int32)((int32)(x)*1073741824./(int32)(y))
  #define ufract30div(x,y) (uint32)((uint32)(x)*1073741824./(uint32)(y))

  #define ufix16mul(x,y) (ufixed)((ufixed)(x)/65536.*(ufixed)(y)+.5)
 #endif

/* ---- globals */

extern const uint16	csrhlut[256];
extern const s_float80	f80pow10_xxx4[31];
#define f80_one f80pow10_xxx4[15]

/* ---- prototypes */

/*
 vfix16mul/vfix16div/vfract30mul/vfract30div
	For computing a high-precision, rounded multiply or divide.
	Overflow is handled.  Divide by zero is counted as overflow.
	The result is clipped to the full binary range.
*/
ifixed vfix16mul(ifixed a,
	ifixed b);
ifixed vfix16div(ifixed num,
	ifixed den);
ifract vfract30mul(ifract num,
	ifract den);
ifract vfract30div(ifract num,
	ifract den);

/*
 ufixmuldiv/ufix32muldiv
 	For computing a high-precision, rounded multiply and divide in one.
	The result is clipped to the full binary range.
*/
ufixed ufixmuldiv(ufixed,
	ufixed numer,
	ufixed denom);
ufixed ufix32muldiv(uint32 integral,
	uint32 fraction,
	ufixed numer,
	uint32 denom);
uint32 umuldivrem(uint32 a,
	uint32 b,
	uint32 denom,
	uint32 *remainder);

/*
 ihypot2/uhypot2/ihypot3/uhypot3
	These compute an RMS (root-mean-squareroot), to the formula,
		sqrt(a**2+b**2).
	They use usqrt64.
	Common use: distance calculation, passing the coordinate differences.
  NOTE:
	Large enough parameters to uhypot3 can produce overflow.
*/
/* `sqrt(a**2+b**2)` or `sqrt((|a|/|b|)**2+1)*|b|` */
uint32 ihypot2(int32,
	int32);
uint32 uhypot2(uint32,
	uint32);
/* 2011.06.28 */
uint32 ihypot3(int32,
	int32,
	int32);
uint32 uhypot3(uint32,
	uint32,
	uint32);

/*
 usqrt/usqrt64
	These functions compute and integer square root, rounding.
	All modern implementations use a fast integer divide method.
	On x64, usqrt is SSE-optimized to be faster than sqrt().
*/
uint32 usqrt(uint32 n);
uint32 usqrt64(uint64 n);

/*
 rslopdiv
	Computes a fixed-point divide, rounding up for a div.-like ratio.
	Optionally, the initial value can be calculated for py if !=NULL.
  Result:
	The ratio, clipped to +/-0xFFFF.FFFF.
*/
int64 rslopdiv(int64 *px,
	int32 p0x,
	int32 p0y,
	int32 p1x,
	int32 p1y, /* must be >=p0y (sort the points) */
	uint16 ky); /* skip y (clipping) */
/*
 lslopdiv
	Like rslopdiv, but simulates the behavior of QuickDraw slope calculation.
	Used for compatible results.
*/
int64 lslopdiv(int64 *px, /* must be valid */
	int32 p0x,
	int32 p0y,
	int32 p1x,
	int32 p1y,
	uint16 ky);

/*
 reduceNDFraction
	Eliminates prime factors from fraction numer/denom.
	Both pointers must be valid.
*/
void reduceNDFraction(int32 *numer, /* pointers must be valid */
	int32 *denom);

/*
 fractZFTransform
	Performs a Fourier transform of two sets of ifract "zeroes".
	Both pointers must be valid.
*/
ifract fractZFTransform(const ifract *a,
	const ifract *b,
	int count);

/*
 rincdiv
	Divides a range into the largest power of two for a uint64.
	It rounds up for a div.-like ratio.
	Optionally, the initial value can be calculated for py if !=NULL.
  Result:
	The incremental rate.  It may overflow in narrow y ranges.
*/
uint64 rincdiv(int64 *py,
	ifixed y0,
	ifixed y1, /* must be >y0 */
	uint16 ky); /* skip y (clipping) */

/*
 rarcmul
	Computes the formula w*sqrt(1-y*y), with y in +/-0x40000000 (+/-1.0).
	Values outside the valid range will result in clipped overflow.
	62-bit calculation ensures high accuracy.
  Result:
	The semicircle arc length, in [0,w/2].
*/
ufract rarcmul(ifract y,
	ufract w);

/*
 fractsin
	Computes a sine, given a binary angle.
	The binary theta values of [0,4.0) represent [0,2*pi).
	Unlike 'fastsin', a polynomial is used to render the ratio.
	The result is rounded, and accurate within a certain rounding error.
*/
ifract fractsin(ufract theta);
 #define fractcos(t) fractsin((t)+0x40000000u) /* phase shift */

/*
 fastsin
	Computes a fast sine, given a binary angle.
	The binary theta values of [0,4.0) represent [0,2*pi).
  Accuracy note:
	This function uses a table with linear interpolation.
	The larger theta is in it's lower 30 bits, the less accurate the
       result, falling in between table entries of a sub-sine curve.
	It is still accurate enough for 16 bit samples.
*/
ifract fastsin(ufract theta);
 #define fastcos(t) fastsin((t)+0x40000000u) /* phase shift */

/*
 ifractatan2/ufractatan2
	Computes an arctangent of (x/y), returning the binary angle.
	32-bit computation ensures fairly high accuracy.
*/
ufract ifractatan2(int32 x,
	int32 y);
ufract ufractatan2(uint32 x,
	uint32 y);
/*
 fractasin
	Computes a binary angle arcsine, using ifractatan2/usqrt64.
*/
/* if out of range (-1,1) returns min/max values */
ifract fractasin(ifract x);

/*
 cntlz32
	Computes a count of the leading zeroes in 32 bits, unsigned.
	This is a standard routine, utilizing native CPU instruction(s).
	cntlz32 and cntlz64 are guaranteed to handle zero as an input.
*/
int cntlz32(uint32 x);

/* built-in instructions */
 #ifdef env68k
  #pragma parameter __D0 _cntlz32(__D0)
  extern int _cntlz32(uint32) = {
/*	BFFFO	D0{0:0},D0 */	0xEDC0, 0
  };
  #define cntlz32 _cntlz32
 #elif defined(__MRC__)
  #define cntlz32(x) (int)__cntlzw((uint32)(x))
 #elif defined(__POCC__)
  #if defined(envx86) || defined(envx64)
   static __forceinline int _safe__bsr(uint32 v)
   {
   #if envx86
	__asm
	{ /* default -1 for 32 result */
		or	eax,-1
		bsr	eax,v
		mov	v,eax
	}
	return v;
   #else
   unsigned long __cdecl _bit_scan_reverse(unsigned long mask);
	return _bit_scan_reverse(v)|(-!v);
   #endif
   }
   #define cntlz32(x) (31-_safe__bsr(x))
   #ifdef envx64
   static __forceinline int _safe__bsr64(uint64 v)
   {
   unsigned long long __cdecl _bit_scan_reverse64(unsigned long long mask);
	return _bit_scan_reverse64(v)|(-!v);
   }
    #define cntlz64(x) (63-_safe__bsr64(x))
   #endif
  #endif
 #endif
 #ifndef envx64
 static __forceinline int cntlz64(uint64 v)
 {
	uint32	r = (uint32)int64hi(v);

	return r ? cntlz32(r):32+cntlz32(int64lo(v));
 }
 #endif

 #ifndef _umul128
 uint64 __cdecl _umul128(uint64 m,
	uint64 n,
	uint64 *h);
 #endif

 /* converts a fixed-point number to a decimal-adjusted double_t */
 double_t fix2double(uint32 v,
	int m, /* binary scale - negate for 'signed' */
	int c); /* maximum fraction digit count */
/* Lua-style 8-bit ufloat conversion */
 uint8 dbl2uflt(double v);
 double uflt2dbl(uint8 v);

s_float80 f64to80(double v);
s_float80 i64to80(int64 v);
double f80to64(s_float80 v);
s_float80 f80_mul(s_float80 a,
	s_float80 b);
s_float80 f80_div(s_float80 a,
	s_float80 b);
s_float80 f80_add(s_float80 a,
	s_float80 b);
s_float80 f80_sub(s_float80 a,
	s_float80 b);
s_float80 f80_scalb(s_float80 v,
	int e);
s_float80 f80_scal10(s_float80 v,
	int e);
s_float80 f80_scal36(s_float80 v,
	int e);
 #if sizeof_double_t>8
  #define scal10 f80_scal10
  #define scal36 f80_scal36
 #else
  #define scal10(v,e) f80to64(f80_scal10(f64to80(v),e))
  #define scal36(v,e) f80to64(f80_scal36(f64to80(v),e))
 #endif

u_fhalf16 float2half(float v);
float half2float(u_fhalf16 v);

 #if __POCC__
  #pragma warn(pop)
 #endif
 #define __FTRIG_H
#endif
