/*
 interpreter.h
  v1.9.9f2

  Created: 2010.03.21

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __INTERPRETER_H
#include "enviro64.h"

#include <stdlib.h>

#include "ftrig.h"

/* ---- constants */

#ifdef USE_INTERP_VERS
 #define INTERP_VERS		USE_INTERP_VERS
#else
 #define INTERP_VERS		0x199
#endif

#ifdef NO_TABLES
 #undef HAVE_IUC_TABLES
#endif

#include <pshpack2.h>

enum
{ /* error codes */
	_e_param_separation	= 1,	/* white space */
	_e_return		= 0x7FFFFFFF,
	_e_repeat		= 0x7FFFFFFE, /* loop current line */
	_e_repeat_last		= 0x7FFFFFFD, /* loop last line(s) */
	_e_repeat_end		= 0x7FFFFFFC, /* loop ended (not an expr line) */

	_e_general_error	=-1,	/* syntax/param error */
	_e_memory_error		=-2,	/* alloc failed */
	_e_read_error		=-3,	/* I/O read error */
	_e_write_error		=-4,	/* I/O write error */
	_e_incomplete_error	=-5,	/* expression missing symbols */
	_e_open_interrogator	=-6,	/* incomplete error for '?' */
	_e_no_identifier	=-7,	/* bad/missing ID */
	_e_handler_error	=-8,	/* handler not supplied for op */
	_e_stack_error		=-9,	/* too many or missing parentheses */
	_e_range_error		=-10,	/* displacement too large for size */
	_e_open_comment		=-11,	/* comment still open at end */
	_e_open_string		=-12,	/* unterminated string */
	_e_identifier_conflict	=-13,	/* ID already added */
	_e_op_error		=-14,	/* can't encode that! */
	_e_label_error		=-15,	/* can't add this as a label here */
	_e_language_error	=-16,	/* not equipped to process this */
	_e_macro_error		=-17,	/* no such macro */
	_e_cannot_optimize	=-18,	/* can't remove/reduce this code */
	_e_value_set		=-19,	/* already set, different value */
	_e_lvalue		=-20,	/* destination is a calc'd number */
	_e_overlong_data	=-21,	/* data (e.g., string) too large */
	_e_no_field		=-22,	/* no such struct/list field */
	_e_no_expr		=-23,	/* no expression for loop/conditional */
	_e_closing_symbol	=-24,	/* incorrect symbol/ID in closing */
	_e_unsupported_value	=-25,	/* unsupported value or version number */
	_e_unknown_symbol	=-26,
	_e_missing_keyword	=-27,
	_e_superfluous_keyword	=-28,
	_e_expected_boolean	=-29,	/* must be 'true'/'false' to language */
	_e_open_block		=-30	/* no end to a begin word/symbol */
};

/* ---- types */

/* multi-parameter element types */
/* IMPORTANT: these do not fit into 16 bits (use LONG_TREC for 32 bits) */
#define TYP_PAIR(a,b)		(((a)&511)|(b)<<9)
#define TYP_TRIPLET(a,b,c)	(((a)&511)|((b)&31)<<9|(c)<<14)
#define TYP_QUADRUPLET(a,b,c,d)	(((a)&511)|((b)&31)<<9|((c)&31)<<14|(d)<<19)

enum
{ /* types - values -1 to -255 are pascal string sizes */
	_int8		=-256,
	_int16		=-257,
	_int32		=-258,
	_int64		=-259,

	_uint8		=-260,
	_uint16		=-261,
	_uint32		=-262,
	_uint64		=-263,
	/* printed in hexadecimal, unclipped */
	_hint8		=-264,
	_hint16		=-265,
	_hint32		=-266,
	_hint64		=-267,

	_bool		=-268,	/* non-zero byte */
	_ifrc16		=-269,	/* i6:10 - all 1 bits = -1u */
	_float		=-270,	/* IEEE 32-bit (s1.e8.m23) */
	_double		=-271,	/* IEEE 64-bit (s1.e11.m52) */

	_flt16		=-272,	/* f16:0 (BFloat16) */
	_hpf16		=-273,	/* IEEE half-precision (s1.e5.s10) */
	_dbl32		=-274,	/* d32:0 */
	_ext80		=-275,	/* IEEE 80-bit s1.e15.o1.m79 */

	_ifixed		=-276,	/* i16:16 */
	_ufixed		=-277,	/* u16:16 */
	_ifract		=-278,	/* i2:30 */
	_ufract		=-279,	/* u2:30 */

	_alpha8		=-280,	/* 255=1.0 */
	_ufrc8		=-281,	/* u0:8 */
	_ufrc16		=-282,	/* u0:16 */
	_ufrc32		=-283,	/* u0:32 */

	_styl8		=-284,	/* [style.8: 1=bold,2=italic] */
	_uflt8		=-285,	/* Lua 8-bit (e5.m3) */
	_ifix16		=-286,	/* i8:8 */
	_itck16		=-287,	/* time in 1/30 sec. precision */
	/* script-only types */
	_tilist		=-288,	/* can be an empty list */
	_tifunc		=-289,	/* function */
	_tiref		=-4096,/* low 11 bits are permutation */
	/* >_itref is evalvalue (may be complex) */
 /* some multi-element types */
	_rgb24		= TYP_TRIPLET(_uint8,_uint8,_uint8),
	_argb32		= TYP_QUADRUPLET(_alpha8,_uint8,_uint8,_uint8),
	_rgb48		= TYP_TRIPLET(_uint16,_uint16,_uint16),
	_pt2d16		= TYP_PAIR(_int16,_int16),
	_pt3d16		= TYP_TRIPLET(_int16,_int16,_int16),
	_pt2dq		= TYP_PAIR(_ifixed,_ifixed),
	_pt3dq		= TYP_TRIPLET(_ifixed,_ifixed,_ifixed),
	_rect16		= TYP_QUADRUPLET(_int16,_int16,_int16,_int16),
	_fnts80		= TYP_QUADRUPLET(-65,_int16,_styl8,_uint8)
};

/* unsigned long type */
#if SIZEOF_LONGINT>4
 #define _ulong			_uint64
#else
 #define _ulong			_uint32
#endif

#ifdef mac
 #define CHR_NE			(-83) /* inequal */
 #define CHR_LE			(-78) /* lesser or equal */
 #define CHR_GE			(-77) /* greater or equal */
 #define CHR_PI			(-71)
 #define CHR_RA			(-61) /* radical (square root) */
 #define CHR_AE			(-59) /* almost equal to */
 #define CHR_HE			(-55) /* horizontal elipsis */
 #define CHR_NB			(-54) /* non-breaking space */
 #define CHR_MN			(-48) /* minus "en-dash" */
 #define CHR_DV			(-42) /* divide */
 #define CHR_NL			13 /* new line: CR */
#else /* ISO-8895-1-compatible */
/* Windows-1252 */
 //no NE (U+2260)
 //no official LE (U+2264)
 //no official GE (U+2265)
 //no RA (U+221A)
 #define CHR_NB			(-96)
 #define CHR_ML			(-41)
 #define CHR_DV			(-9)
 #define CHR_NL			10 /* new line: LF */
 //not in ISO-8859-1
 #define CHR_HE			(-123)
 #define CHR_MN			(-106) /* minus "en-dash" */
#endif

#ifndef CHR_MN
 #define CHR_MN			'-'
#endif

/* ---- macros */

#define CASE_NUMERIC		case'0':case'1':case'2':case'3':case'4': \
	case'5':case'6':case'7':case'8':case'9':
#define CASE_ALPHABETIC		case'A':case'B':case'C':case'D':case'E': \
	case'F':case'G':case'H':case'I':case'J':case'K':case'L':case'M':case'N': \
	case'O':case'P':case'Q':case'R':case'S':case'T':case'U':case'V':case'W': \
	case'X':case'Y':case'Z': case'a':case'b':case'c':case'd':case'e': \
	case'f':case'g':case'h':case'i':case'j':case'k':case'l':case'm':case'n': \
	case'o':case'p':case'q':case'r':case's':case't':case'u':case'v':case'w': \
	case'x':case'y':case'z':
#define CASE_BINARY_OPERATOR	case'%':case'&':case'*':case'/':case'<': \
	case'=':case'>':case'^':case'|':

#define IS_EOL(c)		((c)-10u<4u)
#define IS_DIG(c)		((c)-'0'<10u)
#define IS_ALA(c)		(((c)|32)-'a'<26u)

/* efficient, in-function copy routine that's fast for small cases */
/* this advances two int8* streams */
#if env68k || envppc /* M68020+ */
 #define _amcpy(d,s,z)		do{ \
	if (z&1) *d++=*s++; \
	if (z&2) *((int16*)d)++=*((int16*)s)++; \
	if (z>>=2) do *((int32*)d)++=*((int32*)s)++; while(--z); \
 }while(0)
#elif (envx64 || envx86) && defined(__POCC__)
 extern void __cdecl __movsb(uint8*, const uint8*, size_t); /*v6.5*/
 #define _amcpy(d,s,z)			do{ \
	__movsb((uint8*)d,(const uint8*)s,z), d+=z, s+=z; \
 }while(0)
#else
 #define _amcpy(d,s,z)		do{ \
	if (z&1) *d++=*s++; \
	if (z&2) *d++=*s++, *d++=*s++; \
	if (z>>=2) do *d++=*s++,*d++=*s++,*d++=*s++,*d++=*s++; while(--z); \
 }while(0)
#endif

/* ---- typedefs, evalvalue macros, structures */

#ifndef HUGE_VAL
 #ifdef env68k
 typedef long double double_t;
 #elif __POCC_FLTEVAL__ == 1
 typedef double float_t;
 typedef double double_t;
 #elif __POCC_FLTEVAL__ == 2
 typedef long double float_t;
 typedef long double double_t;
 #else /* __POCC_FLTEVAL__ <= 0 */
 typedef float float_t;
 typedef double double_t;
 #endif
#endif

#if TARGET_OS_MAC
 #define INFINITY		__inf()
 double_t scalb(double_t,
	long);
#elif __POCC__
 #define scalb			scalbn
#endif

#if INTERP_VERS>0x198
 #define evinttype		int64
 #define evuinttype		int64
 #define evintmin		(~0x7FFFFFFFFFFFFFFFll)
 #if !defined(HAVE_COMPLEX_EVAL) && EVAL_FLOATING
  #define HAVE_COMPLEX_EVAL
 #endif
#else
 #define evinttype		long
 #define evuinttype		unsigned long
 #define evintmin		IMIN(long)
#endif
#define evreal
#define evintt(v)		(evinttype)(v)
#define evvadd(a,b)		(a+=b)
#define evvsub(a,b)		(a-=b)
#define evvmul(a,b)		do a*=b;while(0)
#define evvdiv(a,b)		(a/=b)
#define evveq(a,b)		(a==b)
#define evvne(a,b)		!evveq(a,b)
#define evrset(v,r)		do{(v)=r;}while(0)
#define evnset(v,r)		do{(v)=-(r);}while(0)
#define evradd(v,r)		do{(v)+=r;}while(0)
#define evrmul(v,r)		do{(v)*=r;}while(0)
#define evrdiv(v,r)		do{(v)/=r;}while(0)
#define evisnil(v)		!(v)
#define evronly(a,b)
#define evronly2(a,b)
#define emod(a,b)		fmod(a,b)

#if EVAL_FLOATING
 #ifdef HAVE_COMPLEX_EVAL
  #if __POCC__
   #if __POCC__<1200 /* Geez, they nuked complex support; tsk tsk */
    #include <complex.h>
   #endif
  #else
   #include <complex.h>
  #endif
 #endif

 #define _op_dve		_op_div
/* single-char op. (e.g., `/`) is unsigned when <=4 bytes */
 #define evvunsign		evalvalue
 #ifdef HAVE_COMPLEX_EVAL
  #undef evreal
  #undef evintt
  #undef evronly
  #undef evronly2
  #undef emod
  #define evronly(a)		if(cimag(a)) {evrset(a,NAN); break;}
  #define evronly2(a,b)		if(cimag(a)||cimag(b)) {evrset(a,NAN); break;}
  #define evreal		creal
  #define sizeof_evalvalue	16
  #ifndef I
   typedef struct tcomplex
   {
	double		_r, _i;
   } evalvalue, tcomplex;
   #undef evvadd
   #undef evvsub
   #undef evvmul
   #undef evvdiv
   #undef evveq
   #undef evvne
   #undef evisnil
   #define evvadd		evvadd
   #define evvsub		evvsub
   #define creal(z)		((z)._r+0)
   #define cimag(z)		((z)._i+0)
   #define csetri(z,r,i)	do{(z)._i=i,(z)._r=r;}while(0)
   #define cretri(r,i)		do{tcomplex _z; _z._r=r,_z._i=i; return _z;}while(0)
   #define caddri(a,r,i)	do{(a)._i+=i,(a)._r+=r;}while(0)
   /* these are basic and do not do advanced checks like emult */
   #define evvmul(a,b)		do{double _u=(a)._r*(b)._r-(a)._i*(b)._i; \
   					csetri(a,_u,(a)._r*(b)._i+(a)._i*(b)._r);}while(0)
   #define evvdiv(a,b)		((a)=_cdivz(a,b))
   #define evveq(a,b)		((a)._r==(b)._r&&(a)._i==(b)._i)
   #define cabs(a)		hypot(creal(a),cimag(b))

   tcomplex _cdivz(tcomplex a,
	tcomplex b);

   #define evvadd(a,b)		((a)._i+=(b)._i,(a)._r+=(b)._r)
   #define evvsub(a,b)		((a)._i-=(b)._i,(a)._r-=(b)._r)
   #define evrset(z,r)		csetri(z,r,0)
   #define evnset(a,b)		csetri(a,-creal(a),-cimag(b))
   #define evradd(z,r)		do{(z)._r+=r;}while(0)
   #define evrmul(z,r)		do{(z)._i*=r,(z)._r*=r;}while(0)
   #define evrdiv(z,r)		do{(z)._i/=r,(z)._r/=r;}while(0)
   #define evisnil(z)		(!creal(z)&&!cimag(z))
  #else
   typedef double _Complex evalvalue, tcomplex;
   #define csetri(z,r,i)	do{(z)=(i)*I+(r);}while(0)
   #define cretri(r,i)		return (i)*I+(r)
   #define caddri(a,r,i)	do{(a)+=(i)*I+(r);}while(0)
   #define evrset(v,r)		do{(v)=r;}while(0)
   #define evnset(v,r)		do{(v)=-(r);}while(0)
   #define evradd(v,r)		do{(v)+=r;}while(0)
   #define evrmul(v,r)		do{(v)*=r;}while(0)
   #define evrdiv(v,r)		do{(v)/=r;}while(0)
  #endif
  #define native_eval		evalvalue
  evinttype evintt(evalvalue);
  evalvalue emod(evalvalue,evalvalue);
 #else
  typedef struct
  {
	double		_r, _i;
  } tcomplex;
 typedef double		evalvalue;
  #define sizeof_evalvalue	8
 #endif
typedef double		evalparse;
 #define EVAL_HUGEVAL		(evalvalue)((1L<<22)*(evalvalue)(1L<<30)-1)
 #define parse_eval(s,e,i)	parse_double(s,e,i,1)
#else
 #define _op_dve		_op_dvs
 #define evvunsign		unsigned
typedef int32		evalvalue;
typedef long int	evalparse;
 #define sizeof_evalvalue	4
 #define EVAL_HUGEVAL		((evalvalue)1<<(sizeof(evalvalue)<<3)-1)
 #define parse_eval(s,e,i)	readanyinteger(s,e-s,i)
  typedef struct
  {
	double		_r, _i;
  } tcomplex;
#endif

enum
{ /* user flags */
	_allow_qrepeat		= 0x1,	/* ?> */
	_dquote_literals	= 0x2,	/* ` "abcd" ` taken as a literal, not text */
	_whitespace_breaks	= 0x4,	/* changes behavior of `2 (3)` for example */
	_disallow_breaks	= 0x8,	/* ; */
	_CPEMDAS_ordering	= 0x10,
	_allow_approxeq		= 0x20,
	_implied_block_breaks	= 0x40,	/* end-of-line always breaks a statement */
	_take_brace_blocks	= 0x80, /* {} */
  /* current state flags */
	_within_longcomment	= 0x100000
};

#if INTERP_VERS>0x199
 #ifndef EVALLANG_STACKSIZE
  #define EVALLANG_STACKSIZE	20
 #endif

 #include "../utility/eval_op.h"

typedef struct evallangpoint
{
	evalop		key;
	uint32		offset;
} evallangpoint;
#endif

/* NOTE: this structure is subject to grow in size */
/* be sure to init via memset(&state,0,sizeof(state)) prior to any use */
typedef struct evalstate
{
	evalvalue	result, trueVal, falseVal;
	uint32		flags;
	uint32		lineOffs;

	const char *(__cdecl *getv)( /* can be NULL - return NULL on error */
		const char *s,
		const char *e,
		evalvalue*, /* may be NULL */
		uint8 **u);
	const char *(__cdecl *setv)( /* can be NULL - return NULL on error */
		const char *s,
		const char *e,
		evalvalue,
		int *err,
		uint8 **u);
	evalvalue (__cdecl *geti)(evalvalue); /* can be NULL (i.e., don't allow `@`) */

	uint8		**vbuf; /* variables, constants - content defined by s/getv */
	const uint64	*lKeys; /* hashed keywords */

	uint8		conQ;
	int8		qlv;
#if INTERP_VERS>0x199
	int16		lCount;
	uint64		conShift;
	evallangpoint	lCur;
	evallangpoint	lStack[EVALLANG_STACKSIZE];
#endif
} evalstate;

/* text/data object */
typedef struct tobj/*(8+)*/
{
	int32		z; /* <=0 means not allocated */
	char		*s;
} tobj;

/* still a work in progress */

#if INTERP_VERS>0x199
/* non-eval number types use NaN code 0xFFF5.. */
/* versions of GNUC define NAN with a random quiet NaN (>=0xFFF8..) */
#define TI_ITEM_SIG		0xFFF5
#define TI_IS_ITEM(l)		((l)->sig==TI_ITEM_SIG)
/* non-pointer short strings use a Pascal length */
#define TI_IS_PSTR(l)		(((l)->typ|0xE000u)>_int8+0u)
#define TI_IS_STR(l)		(((l)->typ|0xE000u)<0xE00D)
#define TI_IS_PTR(l)		!!((l)->typ&0x8000u)
#define TI_IS_ALLOC(l)		!!((l)->typ&0x4000u)
#define TI_IS_ANYPTR(l)		!!((l)->typ&0xC000u)
#define TI_IS_CONST(l)		!!((l)->typ&0x2000u)
#define TI_GETTYPE(l)		(int16)((l)->typ|0xE000)
#define TI_GETPTR(l)		(TI_IS_ANYPTR(l)?(l)->ptr:(l)->str.buf)
#define TI_SHORT_STR		0x1FF4
#define TI_UTYPE(t)		(0x1FFF&(t))

typedef union titem/*(16)*/
{
	evalvalue	num;
	struct
	{ /* for lists, pos is the count */
	#if BIG_ENDIAN
		uint16		sig;
		int16		typ;
		int32		pos;
		void		*ptr;
	#else
		void		*ptr;
	 #if 1//ndef envx64
		uint8 pad[8-sizeof(size_t)];
	 #endif
		int32		pos;
		int16		typ;
		uint16		sig;
	#endif
	};
	/* for cases where size is <=12 */
	struct
	{
	#if BIG_ENDIAN
		uint32		styp;
		uint8		buf[12];
	#else
		uint8		buf[12];
		uint32		styp;
	#endif
	} str;
} titem;
#endif

#ifndef _SIOF_DEFINED
 #include "csstream.h"
#endif

/* ---- prototypes */

/* system standard */

#ifndef NO_TABLES
/* these compliant C routines both take nul as termination */
int strcasecmp(const char *a,
	const char *b);
int strncasecmp(const char *a,
	const char *b,
	size_t z);
/* this one does not take nul as termination */
int memcasecmp(const char *a,
	const char *b,
	size_t z);
#endif

/* extended standard */

/* parses ASCII-compatible chars to return length of an entire line; 2011.08.13 */
/* escape char can be anything, such as '\', or 0 (null) for no escape char */
long sreadline(const char *p,
	long length,
	uint32 *lecount,
	char escape);
/* same as sreadline, but uses fgetc, allocates string w. reallocfunc; */
/* output is capped to 0x7FFFFFF0 bytes */
char *freadline(struct FILE*,
	uint32 *length, /* in/out; can be NULL (C-string result) */
	char *store, /* in-place realloc'd data; can be NULL */
	void *(__cdecl reallocfunc)(void*, /* can be NULL (default: realloc) */
		size_t),
	uint32 *lecount, /* line-ending count, in/out; can be NULL */
	int escape); /* can be 0 (no escape char) */

/* reads and buffers the string for an entire file */
char *freadfile(struct FILE *f,
	int32 *z, /* out length; can be NULL */
	int (__cdecl delim)(const char *s,
		const char *e));

int makebasestring(char *outstring,
	const char *basechars, /* can be NULL (use default chars) */
	long value,
	unsigned base,
	int minlength,
	char sign); /* pass zero to interpret unsigned */

/* returns ptr to next character */
/* current version of both returns largest negative if out of range */
const char *readanyinteger(const char *str,
	long length,
	long int *outvalue);
/* added 2013.04.01 */
const char *readanyint64(const char *s,
	const char *e,
	int64 *v);

const char *parse_float(const char *s,
	const char *e,
	s_float80 *_v,
	unsigned char r); /* real (take the fractional point) */
const char *parse_double(const char *s,
	const char *e,
	double *v,
	uint8 r); /* real (take the fractional point) */
 #ifndef parse_number
  #if sizeof_double_t>8
   #define parse_number parse_float
  #else
   #define parse_number parse_double
  #endif
 #endif

/* this requires that the characters have already been preprocessed */
int parse_literal(const char *s,
	int64 *r,
	long z);

const char *parse_eval_literal(const char *str,
	const char *e,
	char endquote,
	evalvalue *v);

/* returns next position (after closing quote) */
const char *parse_textloop(const char *s,
	const char *e,
	char endquote,
	int32 *z, /* in max./out length; negative on memory error */
	uint8 *isPascal, /* can be NULL */
	char **r, /* will use realloc to buffer this; can be NULL */
	void *u, /* ptr passed to b func */
	void (__cdecl b)(void *u,
		char v));

/* prints an escaped string to file output */
/* the bounding quote chars are not printed by this function */
/* returns total printed char count */
uint32 print_tchars(struct FILE *f,
	const char *s,
	long z,
	char q,
	uint8 bounded);
uint32 tprint_tchars(sio_func_t,
	void *out,
	const char *s,
	long z,
	char q,
	uint8 bounded, /* end-quote follows */
	uint8 strict);

/* turns a 32-bit wide char into a byte sequence */
/* does not null terminate */
/* returns length */
int smake_utf8(uint8 *d,
	uint32 v);

#ifndef NO_TABLES
const char *parse_wordbreak(const char *s, /* does nothing if either are NULL */
	const char *e,
	uint8 *h, /* non-zero first char hash (can be NULL); 2011.09.15 */
	int16 min, /* normally pass UCFDIG_0 */
	int16 max); /* normally pass UCFUNDSC */
const char *parse_rwordbreak(const char *s,
	const char *f,
	int16 limit); /* normally pass UCFUNDSC */
const char *parse_wordhash(const char *s, /* must be valid */
	const char *e,
	uint64 *h); /* must be valid */
#endif

const char *parse_skipwsp(const char *s,
	const char *e);
const char *parse_rskipwsp(const char *s,
	const char *f);

const char *parse_parameter(const char *s,
	long length,
	char *outstring, /* can be NULL; otherwise, expects at least 24 chars allocated */
	int *outerr,
	const char *(__cdecl getv)(const char *s,
		const char *e,
		evalvalue *i, /* may be NULL */
		uint8 **u),
	const char *(__cdecl setv)(const char *s,
		const char *e,
		evalvalue i,
		int *err,
		uint8 **u),
	uint8 **userdata, /* for getv */
	evalvalue (__cdecl evalindir)(evalvalue), /* indirection; (`@`); can be NULL */
	uint8 allowSQ); /* allows quoted text in '' not just "" */

/* as state struct grows, init state to 0 before setting fields and making calls */
/* returns parsed length */
long evaluate_line(const char *s,
	const char *e,
	int *err,
	evalstate *state); /* can be NULL */

/* ---- prototypes/var.c */

const char *elem_tname(int16 type);

int32 elem_tsize(int32 typ);

/* basic string storage */

/* tobj delete (free and zero) */
/* returns 1 if deallocation occured, or 0 */
int deltobj(tobj *o);

/* tobj unreference (allocate if a ref ptr) */
/* returns 1 if allocation occured, _e_memory_error on failure, or 0 */
int unrtobj(tobj *o);

/* conversion */

double cvt_tdata(int32 z,
	const void *p);
int set_tdata(int32 *type,
	const tobj *s,
	int8 *p,
	int8 lit,
	int8 first);

const int8 *read_typ_be(void *p, /* must be valid */
	const int8 *s, /* must be valid */
	int32 typ);

/* string/file printing */

int32 tprint_string(sio_func_t f,
	void *out,
	const char *s,
	int32 z,
	int t,
	int idbreak); /* allow case of compact printing without quotes (.ini format) */
//#define print_string(f,s,z,t,w) _print_string(swr_fwrite,f,s,z,t,w)
int32 tprint_tdata(sio_func_t f,
	void *out,
	int32 z,
	const void *p,
	int idbreak,
	int maxlen);
//#define print_tdata(f,z,p,w) _print_tdata(swr_fwrite,f,z,p,w,INT32_MAX)

#if INTERP_VERS>0x199
int32 tprint_titem(sio_func_t f,
	void *out,
	const titem *ti,
	int idbreak,
	int32 maxlen);
#endif

/* prints the number for an evalvalue according to macro settings */
#ifdef _CRTNOP
_CRTNOP
#endif
int sprint_evalue(char *b, /* needs to buffer >= 62 chars on max prec */
	evalvalue v,
	int prec); /* 0=default/6; <0 for hex number; >17 for auto/max */
#ifdef _CRTNOP
_CRTNOP
#endif
int sprint_fphex(char *b,
	double_t v);

#ifdef _CRTNOP
_CRTNOP
#endif
int sprint_float(char *out, /* needs at least 30 chars on max precision */
	s_float80 v,
	int prec, /* <0 to always print notation or constant length */
	int res, /* min/max resolution */
	char ptchar,
	char sepchar,
	char signchar);
#define sprint_double(b,v,p,ptc,sgc) sprint_float(b,f64to80(v),p,17,ptc,0,sgc)

/* math eval routines for accuracy */

evalvalue esqrt(evalvalue v);
evalvalue epow(evalvalue a,
	evalvalue b);
evalvalue emult(evalvalue a,
	evalvalue b);
evalvalue edivi(evalvalue a,
	evalvalue b);

/* converts a polar coordinate to a complex coordinate */
/* this function is available regardless of whether complex is on */
tcomplex epol2com(double_t r,
	double_t th);

/* reduces precision for approximation compares */
evalvalue evdbl32(evalvalue v);

/* scanning */

/* single-allocation data case */
/* returns offset to data, or ~length if not in list */
/* should never return zero */
int32 get_hbuf_entry(const char *name, /* must be valid */
	int nameLength,
	uint8 hash, /* as returned by parse_wordbreak */
	const uint8 *u, /* zero-hash terminated list */
	uint32 dz); /* data size - must be valid */
/* no in-data case (no null termination needed) */
/* returns index or ~length */
int32 get_hname_index(const char *name, /* must be valid */
	int nameLength,
	uint8 hash,
	const uint8 *u,
	uint32 tz); /* total size */

const char *parse_entrybreak(const char *str,
	const char *e,
	int32 context);

/* basic variable storage */

#ifdef _CRTNOP
_CRTNOP
#endif
const char __cdecl *eval_getv(const char *s,
	const char *e,
	evalvalue *i,
	uint8 **u);

#ifdef _CRTNOP
_CRTNOP
#endif
const char __cdecl *eval_setv(const char *s,
	const char *e,
	evalvalue i,
	int *err,
	uint8 **u);

#if INTERP_VERS>0x199
/* call this when done with calls to eval_getv/eval_setv */
int eval_free(titem*);
#endif

#ifdef INTERP_BASIC_SCRIPT_FILE
 #if INTERP_VERS>0x199
  extern const uint64 eval_moskey[];
 #endif
 #define INTERP_BASIC_SCRIPT_VARS
 long evaluate_script(char *s,
	const char *e,
	long li,
	evalstate *state,
	void (__cdecl reportError)(int err,
		char *s,
		long z,
		int32 r,
		int li,
		void *userData),
	int (__cdecl reportTrace)(char *s,
		long z,
		int li,
		const evalstate *state,
		void *userData),
	void *userData);
#endif
#ifdef INTERP_BASIC_SCRIPT_VARS
extern const char *_interp_err_strs[30];
extern const char *_interp_defbasechars;

void interp_init_basic_vars(uint8 **vbuf,
	int *err);
evalvalue __cdecl eval_rindir(evalvalue v);
#endif

#include <poppack.h>

#define __INTERPRETER_H
#endif
