/*
 * cjpegp.c
 *
 * cjpeg.c Copyright (C) 1991-1998, Thomas G. Lane.
 * Modified 2003-2013 by Guido Vollbeding.
 * Plus version 2021-22 for JPEG input and other features by Adam Pickett
 *
 * cjpeg.c is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains a command-line user interface for the JPEG compressor.
 * It should work on any system with Unix- or MS-DOS-style command lines.
 *
 * Two different command line styles are permitted, depending on the
 * compile-time switch TWO_FILE_COMMANDLINE:
 *	cjpegp [options]  inputfile outputfile
 *	cjpegp [options]  [inputfile]
 * In the second style, output is always to standard output, which you'd
 * normally redirect to a file or pipe to some other program.  Input is
 * either from a named file or from standard input (typically redirected).
 * The second style is convenient on Unix but is unhelpful on systems that
 * don't support pipes.  Also, you MUST use the first style if your system
 * doesn't do binary I/O to stdin/stdout.
 * To simplify script writing, the "-outfile" switch is provided.  The syntax
 *	cjpeg [options]  -outfile outputfile  inputfile
 * works regardless of which command line style is used.
 */

#include "cdjpeg.h"		/* Common decls for cjpeg/djpeg applications */
#include "jversion.h"		/* for version message */

#ifdef USE_CCOMMAND		/* command-line reader for Macintosh */
#ifdef __MWERKS__
#include <SIOUX.h>              /* Metrowerks needs this */
#include <console.h>		/* ... and this */
#endif
#ifdef THINK_C
#include <console.h>		/* Think declares it here */
#endif
#endif

#include "transupp.h"		/* Support routines for copying markers */

#ifdef HAVE_OUTPUT_CFILTER
char * cfilterarg;
#define MAXIMUM_RFILTER_ARGS	10 /* 2022 */
char * rfilterarg[MAXIMUM_RFILTER_ARGS];
int rfilterargs = 0;

#include "cfilter.h"
#endif

/*
 * This list defines the known input image formats.
 * Most are not handled.
 */
typedef enum {
	FMT_BMP,		/* BMP format (Windows flavor) */
	FMT_GIF,
	FMT_OS2,		/* BMP format (OS/2 flavor) */
	FMT_PPM,		/* PPM/PGM (PBMPLUS formats) */
	FMT_RLE,
	FMT_TARGA,
	FMT_TIFF,
	FMT_JPEG,
	FMT_ANY
} IMAGE_FORMATS;


/* Create the add-on message string table. */

#define JMESSAGE(code,string)	string ,

static const char * const cdjpeg_message_table[] = {
#include "cderror.h"
  NULL
};

struct jpeg_error_mgr jerr;


static int image_type;	/* user -targa switch, any, or found type */

/*
 * This routine was added to handle JPEG as a pixel image.
 * It allows for more modifications, though lossy, than jpegtran.
 */

#ifdef JPEGIN_SUPPORTED
char * gammaarg, * contrastarg, * blevelarg;

struct jpeg_decompress_struct jsrcinfo;


typedef struct _jpeg_source_struct * jpeg_source_ptr;

typedef struct _jpeg_source_struct {
  struct cjpeg_source_struct pub; /* public fields */

  j_compress_ptr cinfo;		/* back link saves passing separate parm */
  int row_size;
} jpeg_source_struct;


METHODDEF(void)
input_jpeg (j_compress_ptr cinfo, cjpeg_source_ptr sinfo)
{
  /* no work */
}

METHODDEF(JDIMENSION)
readline_jpeg (j_compress_ptr cinfo, cjpeg_source_ptr sinfo)
{
  exit(EXIT_FAILURE);
  return 1;
}

LOCAL(cjpeg_source_ptr)
jinit_read_jpeg (j_compress_ptr cinfo, FILE * infile)
{
  jpeg_source_ptr source;

  source = (jpeg_source_ptr)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
				  SIZEOF(jpeg_source_struct));
  source->cinfo = cinfo;	/* make back link for subroutines */
  source->pub.start_input = input_jpeg;
  source->pub.finish_input = input_jpeg;
  source->pub.get_pixel_rows = readline_jpeg;

  jsrcinfo.err = jpeg_std_error(&jerr);
  if (jpeg_create_decompress(&jsrcinfo) < 0)
  {
    fprintf(stderr," could not open JPEG for decompression.\n");
    exit(EXIT_FAILURE);
  }

  /* Specify data source for decompression */
  jpeg_stdio_src(&jsrcinfo, infile);

  jcopy_markers_setup(&jsrcinfo, JCOPYOPT_ALL);

  /* Read file header, set default decompression parameters */
  if (jpeg_read_header(&jsrcinfo, TRUE) < 0)
  {
    fprintf(stderr,"Input is not a working JPEG file.\n");
    exit(EXIT_FAILURE);
  }
  if (gammaarg)
    jsrcinfo.output_gamma = atof(gammaarg);
  if (contrastarg) {
    //jsrcinfo.output_contrast[0] = atof(contrastarg);
    //jsrcinfo.output_contrast[RGB_RED] = 1;
    if (sscanf(contrastarg,"%f %f %f",jsrcinfo.output_contrast+RGB_RED,
         jsrcinfo.output_contrast+RGB_GREEN,jsrcinfo.output_contrast+RGB_BLUE)<2)
      jsrcinfo.output_contrast[RGB_BLUE] = jsrcinfo.output_contrast[RGB_GREEN] =
        jsrcinfo.output_contrast[RGB_RED];
  }
  if (blevelarg) {
    //jsrcinfo.output_blacklevel = atof(blevelarg);
    //jsrcinfo.output_blevel[RGB_RED] = 0;
    if (sscanf(blevelarg,"%f %f %f",jsrcinfo.output_blevel+RGB_RED,
         jsrcinfo.output_blevel+RGB_GREEN,jsrcinfo.output_blevel+RGB_BLUE)<2)
      jsrcinfo.output_blevel[RGB_BLUE] = jsrcinfo.output_blevel[RGB_GREEN] =
        jsrcinfo.output_blevel[RGB_RED];
  }

  (void) jpeg_start_decompress(&jsrcinfo);

  cinfo->image_width = jsrcinfo.output_width;
  cinfo->image_height = jsrcinfo.output_height;
  cinfo->input_components = jsrcinfo.output_components;
  cinfo->data_precision = BITS_IN_JSAMPLE;
  //cinfo->in_color_space = jsrcinfo.out_color_space;

  source->pub.buffer_height = jsrcinfo.output_height;
  source->pub.buffer = (*jsrcinfo.mem->alloc_sarray)
		((j_common_ptr) &jsrcinfo, JPOOL_IMAGE,
		jsrcinfo.output_components * jsrcinfo.output_width,
		jsrcinfo.output_height + 1);

  return (cjpeg_source_ptr) source;
}
#endif


/*
 * This routine determines what format the input file is,
 * and selects the appropriate input-reading module.
 *
 * To determine which family of input formats the file belongs to,
 * we may look only at the first byte of the file, since C does not
 * guarantee that more than one character can be pushed back with ungetc.
 * Looking at additional bytes would require one of these approaches:
 *     1) assume we can fseek() the input file (fails for piped input);
 *     2) assume we can push back more than one character (works in
 *        some C implementations, but unportable);
 *     3) provide our own buffering (breaks input readers that want to use
 *        stdio directly, such as the RLE library);
 * or  4) don't put back the data, and modify the input_init methods to assume
 *        they start reading after the start of file (also breaks RLE library).
 * #1 is attractive for MS-DOS but is untenable on Unix.
 *
 * The most portable solution for file types that can't be identified by their
 * first byte is to make the user tell us what they are.  This is also the
 * only approach for "raw" file types that contain only arbitrary values.
 * We presently apply this method for Targa files.  Most of the time Targa
 * files start with 0x00, so we recognize that case.  Potentially, however,
 * a Targa file could start with any byte value (byte 0 is the length of the
 * seldom-used ID field), so we provide a switch to force Targa input mode.
 */

LOCAL(cjpeg_source_ptr)
select_file_type (j_compress_ptr cinfo, FILE * infile)
{
  int c;

  if (image_type == FMT_TARGA) {
#ifdef TARGA_SUPPORTED
    return jinit_read_targa(cinfo);
#else
    ERREXIT(cinfo, JERR_TGA_NOTCOMP);
#endif
  }

  if ((c = getc(infile)) == EOF)
    ERREXIT(cinfo, JERR_INPUT_EMPTY);
  if (ungetc(c, infile) == EOF)
    ERREXIT(cinfo, JERR_UNGETC_FAILED);

  switch (c) {
#ifdef BMP_SUPPORTED
  case 'B':
    image_type = FMT_BMP;
    return jinit_read_bmp(cinfo);
#endif
#ifdef GIF_SUPPORTED
  case 'G':
    image_type = FMT_GIF;
    return jinit_read_gif(cinfo);
#endif
#ifdef PPM_SUPPORTED
  case 'P':
    image_type = FMT_PPM;
    return jinit_read_ppm(cinfo);
#endif
#ifdef RLE_SUPPORTED
  case 'R':
    image_type = FMT_RLE;
    return jinit_read_rle(cinfo);
#endif
#ifdef TARGA_SUPPORTED
  case 0x00:
    image_type = FMT_TARGA;
    return jinit_read_targa(cinfo);
#endif
#ifdef JPEGIN_SUPPORTED
  case 0xFF:
    image_type = FMT_JPEG;
    return jinit_read_jpeg(cinfo, infile); /* 2021 */
#endif
  default:
    ERREXIT(cinfo, JERR_UNKNOWN_FORMAT);
    break;
  }

  return NULL;			/* suppress compiler warnings */
}


/*
 * Argument-parsing code.
 * The switch parser is designed to be useful with DOS-style command line
 * syntax, ie, intermixed switches and file names, where only the switches
 * to the left of a given file name affect processing of that file.
 * The main program in this file doesn't actually use this capability...
 */


static const char * progname;	/* program name for error messages */
static char * outfilename;	/* for -outfile switch */


LOCAL(void)
usage (void)
/* complain about bad command line */
{
  fprintf(stderr,
"\n"
"usage: %s [switches] "
#ifdef TWO_FILE_COMMANDLINE
"inputfile outputfile\n"
#else
"[inputfile]\n"
#endif
    , progname);
  fputs(
"Switches (names may be abbreviated):\n"
"  -quality N[,...]   Compression quality (0..100; 5-95 is useful range)\n"
"  -grayscale     Create monochrome JPEG file\n"
"  -rgb           Create RGB JPEG file\n"
#ifdef ENTROPY_OPT_SUPPORTED
"  -optimize      Optimize Huffman table (smaller file, but slow compression)\n"
#endif
#ifdef C_PROGRESSIVE_SUPPORTED
"  -progressive   Create progressive JPEG file\n"
#endif
#ifdef DCT_SCALING_SUPPORTED
"  -scale M/N     Scale image by fraction M/N, eg, 1/2\n"
#endif
#ifdef HAVE_OUTPUT_GAMMA
"  -gamma N       Set in-JPEG gamma (1.0=normal)\n"
"  -contrast N    Set in-JPEG contrast (1.0=normal) - can be 'R G B'\n"
"  -blevel N      Set in-JPEG black level (0.0=normal) - can be 'R G B'\n"
#endif
#ifdef HAVE_OUTPUT_CFILTER
"  -filter N      Filter in-JPEG, gaussian NxN (3, 5 or 7)\n"
"  -rfilter       Filter in-JPEG, by top,left,bottom,right,insetCount\n"
#endif
#ifdef TARGA_SUPPORTED
"  -targa         Input file is Targa format (usually not needed)\n"
#endif
    , stderr);
  fprintf(stderr, "Switches for advanced users:\n");
#ifdef C_ARITH_CODING_SUPPORTED
  fprintf(stderr, "  -arithmetic    Use arithmetic coding\n");
#endif
#ifdef DCT_SCALING_SUPPORTED
  fprintf(stderr, "  -block N       DCT block size (1..16; default is 8)\n");
#endif
#if JPEG_LIB_VERSION_MAJOR >= 9
  fprintf(stderr, "  -rgb1          Create RGB JPEG file with reversible color transform\n");
#endif
#ifdef DCT_ISLOW_SUPPORTED
  fprintf(stderr, "  -dct int       Use integer DCT method%s\n",
	  (JDCT_DEFAULT == JDCT_ISLOW ? " (default)" : ""));
#endif
#ifdef DCT_IFAST_SUPPORTED
  fprintf(stderr, "  -dct fast      Use fast integer DCT (less accurate)%s\n",
	  (JDCT_DEFAULT == JDCT_IFAST ? " (default)" : ""));
#endif
#ifdef DCT_FLOAT_SUPPORTED
  fprintf(stderr, "  -dct float     Use floating-point DCT method%s\n",
	  (JDCT_DEFAULT == JDCT_FLOAT ? " (default)" : ""));
#endif
  fprintf(stderr, "  -nosmooth      Don't use high-quality downsampling\n");
  fprintf(stderr, "  -restart N     Set restart interval in rows, or in blocks with B\n");
#ifdef INPUT_SMOOTHING_SUPPORTED
  fprintf(stderr, "  -smooth N      Smooth dithered input (N=1..100 is strength)\n");
#endif
  fprintf(stderr, "  -maxmemory N   Maximum memory to use (in kbytes)\n");
  fprintf(stderr, "  -outfile name  Specify name for output file\n");
  fprintf(stderr, "  -verbose  or  -debug   Emit debug output\n");
  fprintf(stderr, "Switches for wizards:\n");
  fprintf(stderr, "  -baseline      Force baseline quantization tables\n");
  fprintf(stderr, "  -qtables file  Use quantization tables given in file\n");
  fprintf(stderr, "  -qslots N[,...]    Set component quantization tables\n");
  fprintf(stderr, "  -sample HxV[,...]  Set component sampling factors\n");
#ifdef C_MULTISCAN_FILES_SUPPORTED
  fprintf(stderr, "  -scans file    Create multi-scan JPEG per script file\n");
#endif
  exit(EXIT_FAILURE);
}


LOCAL(int)
parse_switches (j_compress_ptr cinfo, int argc, char **argv,
		int last_file_arg_seen, boolean for_real)
/* Parse optional switches.
 * Returns argv[] index of first file-name argument (== argc if none).
 * Any file names with indexes <= last_file_arg_seen are ignored;
 * they have presumably been processed in a previous iteration.
 * (Pass 0 for last_file_arg_seen on the first or only iteration.)
 * for_real is FALSE on the first (dummy) pass; we may skip any expensive
 * processing.
 */
{
  int argn;
  char * arg;
  boolean force_baseline;
  boolean simple_progressive;
  char * qualityarg = NULL;	/* saves -quality parm if any */
  char * qtablefile = NULL;	/* saves -qtables filename if any */
  char * qslotsarg = NULL;	/* saves -qslots parm if any */
  char * samplearg = NULL;	/* saves -sample parm if any */
  char * scansarg = NULL;	/* saves -scans parm if any */

  /* Set up default JPEG parameters. */

  force_baseline = FALSE;	/* by default, allow 16-bit quantizers */
  simple_progressive = FALSE;
  if (!for_real)
    image_type = FMT_ANY;
  outfilename = NULL;
  cinfo->err->trace_level = 0;

  /* Scan command line options, adjust parameters */

  for (argn = 1; argn < argc; argn++) {
    arg = argv[argn];
    if (*arg != '-') {
      /* Not a switch, must be a file name argument */
      if (argn <= last_file_arg_seen) {
	outfilename = NULL;	/* -outfile applies to just one input file */
	continue;		/* ignore this name if previously processed */
      }
      break;			/* else done parsing switches */
    }
    arg++;			/* advance past switch marker character */

    if (keymatch(arg, "arithmetic", 1)) {
      /* Use arithmetic coding. */
#ifdef C_ARITH_CODING_SUPPORTED
      cinfo->arith_code = TRUE;
#else
      fprintf(stderr, "%s: sorry, arithmetic coding not supported\n",
	      progname);
      exit(EXIT_FAILURE);
#endif

    } else if (keymatch(arg, "baseline", 2)) {
      /* Force baseline-compatible output (8-bit quantizer values). */
      force_baseline = TRUE;

    } else if (keymatch(arg, "block", 2)) {
      /* Set DCT block size. */
#if defined DCT_SCALING_SUPPORTED && JPEG_LIB_VERSION_MAJOR >= 8 && \
      (JPEG_LIB_VERSION_MAJOR > 8 || JPEG_LIB_VERSION_MINOR >= 3)
      int val;

      if (++argn >= argc)	/* advance to next argument */
	usage();
      if (sscanf(argv[argn], "%d", &val) != 1) {
        fprintf(stderr, "%s: bad block parameter '%s'\n",
	      progname, argv[argn]);
	usage();
      }
      if (val < 1 || val > 16)
	usage();
      cinfo->block_size = val;
#else
      fprintf(stderr, "%s: sorry, block size setting not supported\n",
	      progname);
      exit(EXIT_FAILURE);
#endif

    } else if (keymatch(arg, "dct", 2)) {
      /* Select DCT algorithm. */
      if (++argn >= argc)	/* advance to next argument */
	usage();
      if (keymatch(argv[argn], "int", 1)) {
	cinfo->dct_method = JDCT_ISLOW;
      } else if (keymatch(argv[argn], "fast", 2)) {
	cinfo->dct_method = JDCT_IFAST;
      } else if (keymatch(argv[argn], "float", 2)) {
	cinfo->dct_method = JDCT_FLOAT;
      } else {
        fprintf(stderr, "%s: unknown dct parameter '%s'\n",
	      progname,argv[argn]);
	usage();
      }

    } else if (keymatch(arg, "debug", 1) || keymatch(arg, "verbose", 1)) {
      /* Enable debug printouts. */
      /* On first -d, print version identification */
      static boolean printed_version = FALSE;

      if (! printed_version) {
	fprintf(stderr, "Independent JPEG Group's CJPEG, version %s\n%s\n",
		JVERSION, JCOPYRIGHT);
	printed_version = TRUE;
      }
      cinfo->err->trace_level++;

    } else if (keymatch(arg, "grayscale", 2) || keymatch(arg, "greyscale",2)) {
      /* Force a monochrome JPEG file to be generated. */
      jpeg_set_colorspace(cinfo, JCS_GRAYSCALE);

    } else if (keymatch(arg, "rgb", 3) || keymatch(arg, "rgb1", 4)) {
      /* Force an RGB JPEG file to be generated. */
#if JPEG_LIB_VERSION_MAJOR >= 9
      /* Note: Entropy table assignment in jpeg_set_colorspace depends
       * on color_transform.
       */
      cinfo->color_transform = arg[3] ? JCT_SUBTRACT_GREEN : JCT_NONE;
#endif
      jpeg_set_colorspace(cinfo, JCS_RGB);

    } else if (keymatch(arg, "maxmemory", 3)) {
      /* Maximum memory in KB (or MB with 'm'). */
      long lval;
      char ch = 'x';

      if (++argn >= argc)	/* advance to next argument */
	usage();
      if (sscanf(argv[argn], "%ld%c", &lval, &ch) < 1)
	usage();
      if (ch == 'm' || ch == 'M')
	lval *= 1000L;
      cinfo->mem->max_memory_to_use = lval * 1000L;

    } else if (keymatch(arg, "nosmooth", 3)) {
      /* Suppress fancy downsampling. */
      cinfo->do_fancy_downsampling = FALSE;

    } else if (keymatch(arg, "optimize", 1) || keymatch(arg, "optimise", 1)) {
      /* Enable entropy parm optimization. */
#ifdef ENTROPY_OPT_SUPPORTED
      cinfo->optimize_coding = TRUE;
#else
      fprintf(stderr, "%s: sorry, entropy optimization was not compiled\n",
	      progname);
      exit(EXIT_FAILURE);
#endif

    } else if (keymatch(arg, "outfile", 4)) {
      /* Set output file name. */
      if (++argn >= argc)	/* advance to next argument */
	usage();
      outfilename = argv[argn];	/* save it away for later use */

    } else if (keymatch(arg, "progressive", 1)) {
      /* Select simple progressive mode. */
#ifdef C_PROGRESSIVE_SUPPORTED
      simple_progressive = TRUE;
      /* We must postpone execution until num_components is known. */
#else
      fprintf(stderr, "%s: sorry, progressive output was not compiled\n",
	      progname);
      exit(EXIT_FAILURE);
#endif

    } else if (keymatch(arg, "quality", 1)) {
      /* Quality ratings (quantization table scaling factors). */
      if (++argn >= argc)	/* advance to next argument */
	usage();
      qualityarg = argv[argn];

#ifdef HAVE_OUTPUT_GAMMA
    } else if (keymatch(arg, "gamma", 1)) {
      if (++argn >= argc)	/* advance to next argument */
	usage();
      gammaarg = argv[argn];
    } else if (keymatch(arg, "contrast", 1)) {
      if (++argn >= argc)	/* advance to next argument */
	usage();
      contrastarg = argv[argn];
    } else if (keymatch(arg, "blevel", 1)) {
      if (++argn >= argc)	/* advance to next argument */
	usage();
      blevelarg = argv[argn];
#endif
#ifdef HAVE_OUTPUT_CFILTER
    } else if (keymatch(arg, "filter", 1)) {
      if (++argn >= argc)	/* advance to next argument */
	usage();
      cfilterarg = argv[argn];
    } else if (keymatch(arg, "rfilter", 1)) {
      if (++argn >= argc || rfilterargs>=MAXIMUM_RFILTER_ARGS)
	usage();
      rfilterarg[rfilterargs++] = argv[argn];
#endif

    } else if (keymatch(arg, "qslots", 2)) {
      /* Quantization table slot numbers. */
      if (++argn >= argc)	/* advance to next argument */
	usage();
      qslotsarg = argv[argn];
      /* Must delay setting qslots until after we have processed any
       * colorspace-determining switches, since jpeg_set_colorspace sets
       * default quant table numbers.
       */

    } else if (keymatch(arg, "qtables", 2)) {
      /* Quantization tables fetched from file. */
      if (++argn >= argc)	/* advance to next argument */
	usage();
      qtablefile = argv[argn];
      /* We postpone actually reading the file in case -quality comes later. */

    } else if (keymatch(arg, "restart", 1)) {
      /* Restart interval in MCU rows (or in MCUs with 'b'). */
      long lval;
      char ch = 'x';

      if (++argn >= argc)	/* advance to next argument */
	usage();
      if (sscanf(argv[argn], "%ld%c", &lval, &ch) < 1) {
        fprintf(stderr, "%s: unknown restart parameter '%s'\n",
	      progname,argv[argn]);
	usage();
      }
      if (lval < 0 || lval > 65535L)
	usage();
      if (ch == 'b' || ch == 'B') {
	cinfo->restart_interval = (unsigned int) lval;
	cinfo->restart_in_rows = 0; /* else prior '-restart n' overrides me */
      } else {
	cinfo->restart_in_rows = (int) lval;
	/* restart_interval will be computed during startup */
      }

    } else if (keymatch(arg, "sample", 2)) {
      /* Set sampling factors. */
      if (++argn >= argc)	/* advance to next argument */
	usage();
      samplearg = argv[argn];
      /* Must delay setting sample factors until after we have processed any
       * colorspace-determining switches, since jpeg_set_colorspace sets
       * default sampling factors.
       */

    } else if (keymatch(arg, "scale", 4)) {
      /* Scale the image by a fraction M/N. */
      if (++argn >= argc)	/* advance to next argument */
	usage();
      if (sscanf(argv[argn], "%u/%u",
		 &cinfo->scale_num, &cinfo->scale_denom) != 2) {
        fprintf(stderr, "%s: unknown scale parameter '%s'\n",
	      progname,argv[argn]);
	usage();
      }
    } else if (keymatch(arg, "scans", 4)) {
      /* Set scan script. */
#ifdef C_MULTISCAN_FILES_SUPPORTED
      if (++argn >= argc)	/* advance to next argument */
	usage();
      scansarg = argv[argn];
      /* We must postpone reading the file in case -progressive appears. */
#else
      fprintf(stderr, "%s: sorry, multi-scan output was not compiled\n",
	      progname);
      exit(EXIT_FAILURE);
#endif

    } else if (keymatch(arg, "smooth", 2)) {
      /* Set input smoothing factor. */
      int val;

      if (++argn >= argc)	/* advance to next argument */
	usage();
      if (sscanf(argv[argn], "%d", &val) != 1)
	usage();
      if (val < 0 || val > 100)
	usage();
      cinfo->smoothing_factor = val;

    } else if (keymatch(arg, "targa", 1)) {
      /* Input file is Targa format. */
      image_type = FMT_TARGA;

    } else {
      usage();			/* bogus switch */
    }
  }

  /* Post-switch-scanning cleanup */

  if (for_real) {

    /* Set quantization tables for selected quality. */
    /* Some or all may be overridden if -qtables is present. */
    if (qualityarg != NULL)	/* process -quality if it was present */
      if (! set_quality_ratings(cinfo, qualityarg, force_baseline))
	usage();

    if (qtablefile != NULL)	/* process -qtables if it was present */
      if (! read_quant_tables(cinfo, qtablefile, force_baseline))
	usage();

    if (qslotsarg != NULL)	/* process -qslots if it was present */
      if (! set_quant_slots(cinfo, qslotsarg))
	usage();

    if (samplearg != NULL)	/* process -sample if it was present */
      if (! set_sample_factors(cinfo, samplearg))
	usage();

#ifdef C_PROGRESSIVE_SUPPORTED
    if (simple_progressive)	/* process -progressive; -scans can override */
      jpeg_simple_progression(cinfo);
#endif

#ifdef C_MULTISCAN_FILES_SUPPORTED
    if (scansarg != NULL)	/* process -scans if it was present */
      if (! read_scan_script(cinfo, scansarg))
	usage();
#endif
  }

  return argn;			/* return index of next arg (file name) */
}


/*
 * The main program.
 */

int
main (int argc, char **argv)
{
  struct jpeg_compress_struct cinfo;
#ifdef PROGRESS_REPORT
  struct cdjpeg_progress_mgr progress;
#endif
  int file_index;
  cjpeg_source_ptr src_mgr;
  FILE * input_file;
  FILE * output_file;
  JDIMENSION num_scanlines;

  /* On Mac, fetch a command line. */
#ifdef USE_CCOMMAND
  argc = ccommand(&argv);
#endif

  progname = argv[0];
  if (progname == NULL || progname[0] == 0)
    progname = "cjpegp";	/* in case C library doesn't provide it */
  else {
    const char *s = strrchr(progname,'/');
    if (s)
      progname = s+1;
    s = strrchr(progname,'\\');
    if (s && s>progname)
      progname = s+1;
  }

  /* Initialize the JPEG compression object with default error handling. */
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  /* Add some application-specific error messages (from cderror.h) */
  jerr.addon_message_table = cdjpeg_message_table;
  jerr.first_addon_message = JMSG_FIRSTADDONCODE;
  jerr.last_addon_message = JMSG_LASTADDONCODE;

  /* Now safe to enable signal catcher. */
#ifdef NEED_SIGNAL_CATCHER
  enable_signal_catcher((j_common_ptr) &cinfo);
#endif

  /* Initialize JPEG parameters.
   * Much of this may be overridden later.
   * In particular, we don't yet know the input file's color space,
   * but we need to provide some value for jpeg_set_defaults() to work.
   */

  cinfo.in_color_space = JCS_RGB; /* arbitrary guess */
  jpeg_set_defaults(&cinfo);

  /* Scan command line to find file names.
   * It is convenient to use just one switch-parsing routine, but the switch
   * values read here are ignored; we will rescan the switches after opening
   * the input file.
   */

  file_index = parse_switches(&cinfo, argc, argv, 0, FALSE);

#ifdef TWO_FILE_COMMANDLINE
  /* Must have either -outfile switch or explicit output file name */
  if (outfilename == NULL) {
    if (file_index != argc-2) {
      fprintf(stderr, "%s: must name one input and one output file\n",
	      progname);
      usage();
    }
    outfilename = argv[file_index+1];
  } else {
    if (file_index != argc-1) {
      fprintf(stderr, "%s: must name one input and one output file\n",
	      progname);
      usage();
    }
  }
#else
  /* Unix style: expect zero or one file name */
  if (file_index < argc-1) {
    fprintf(stderr, "%s: only one input file\n", progname);
    usage();
  }
#endif /* TWO_FILE_COMMANDLINE */

  /* Open the input file. */
  if (file_index < argc) {
    if ((input_file = fopen(argv[file_index], READ_BINARY)) == NULL) {
      fprintf(stderr, "%s: can't open %s\n", progname, argv[file_index]);
      exit(EXIT_FAILURE);
    }
  } else {
    /* default input file is stdin */
    input_file = read_stdin();
  }

  /* Open the output file. */
  if (outfilename != NULL) {
    if ((output_file = fopen(outfilename, WRITE_BINARY)) == NULL) {
      fprintf(stderr, "%s: can't open %s\n", progname, outfilename);
      exit(EXIT_FAILURE);
    }
  } else {
    /* default output file is stdout */
    output_file = write_stdout();
  }

#ifdef PROGRESS_REPORT
  start_progress_monitor((j_common_ptr) &cinfo, &progress);
#endif

  /* Figure out the input file format, and set up to read it. */
  src_mgr = select_file_type(&cinfo, input_file);
  src_mgr->input_file = input_file;

  /* Read the input file header to obtain file size & colorspace. */
  (*src_mgr->start_input) (&cinfo, src_mgr);

  /* Now that we know input colorspace, fix colorspace-dependent defaults */
  jpeg_default_colorspace(&cinfo);

  /* Adjust default compression parameters by re-parsing the options */
  file_index = parse_switches(&cinfo, argc, argv, 0, TRUE);

  /* Specify data destination for compression */
  jpeg_stdio_dest(&cinfo, output_file);

  /* Start compressor */
  jpeg_start_compress(&cinfo, TRUE);

#ifdef JPEGIN_SUPPORTED
  if (image_type == FMT_JPEG)
  {
    if (cinfo.err->trace_level)
      fprintf(stderr," reading...\n");
    while (jsrcinfo.output_scanline < jsrcinfo.output_height)
      (void) jpeg_read_scanlines(&jsrcinfo, src_mgr->buffer + jsrcinfo.output_scanline,
	cinfo.image_height);

    if (cinfo.err->trace_level)
      fprintf(stderr," copying markers...\n");
    jcopy_markers_execute(&jsrcinfo, &cinfo, JCOPYOPT_ALL);

#ifdef HAVE_OUTPUT_CFILTER
    if (cfilterarg)
      cfilter(src_mgr->buffer,jsrcinfo.output_width,jsrcinfo.output_height,
        0,0,1<<16,1<<16,atof(cfilterarg));
    while(rfilterargs>0)
    {
      int left, top, right, bottom, count = 1;

      left = top = right = bottom = 0;
      sscanf(rfilterarg[--rfilterargs], "%d,%d,%d,%d,%d",
        &left,&top,&right,&bottom,&count);
      if (count > 0)
        do
          cfilter(src_mgr->buffer,jsrcinfo.output_width,jsrcinfo.output_height,
            left,top,right,bottom,-1);
        while(++left,++top,--right,--bottom,--count);
    }
#endif
    if (cinfo.err->trace_level)
      fprintf(stderr," writing output...\n");
    while (cinfo.next_scanline < cinfo.image_height) {
      //fprintf(stderr," %d...\n", cinfo.next_scanline);
      (void) jpeg_write_scanlines(&cinfo, src_mgr->buffer + cinfo.next_scanline,
	cinfo.image_height);
    }
    /* must do this here because decompress trashes the buffer */
    (void) jpeg_finish_decompress(&jsrcinfo);
  } else /* Process data */
#endif
    while (cinfo.next_scanline < cinfo.image_height) {
      num_scanlines = (*src_mgr->get_pixel_rows) (&cinfo, src_mgr);
      (void) jpeg_write_scanlines(&cinfo, src_mgr->buffer, num_scanlines);
    }

  /* Finish compression and release memory */
  (*src_mgr->finish_input) (&cinfo, src_mgr);
  jpeg_finish_compress(&cinfo);
  jpeg_destroy_compress(&cinfo);

  /* Close files, if we opened them */
  if (input_file != stdin)
    fclose(input_file);
  if (output_file != stdout)
    fclose(output_file);

#ifdef PROGRESS_REPORT
  end_progress_monitor((j_common_ptr) &cinfo);
#endif

  /* All done. */
  exit(jerr.num_warnings ? EXIT_WARNING : EXIT_SUCCESS);
  return 0;			/* suppress no-return-value warnings */
}
