/*
 colorPicker.c

  Created: 2015.06.21
  Revised: 2023.08.23
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2019.01.16
*/

#include <stdio.h>
#include <string.h>
#include "mevent.h"
#include "mwind.h"

#include "colorPicker.h"
#include "ftrig.h"

/* ---- macros */

#define DEFAULT_WHEEL_SIZE	201
#define MIN_WHEEL_SIZE		55
#define WHEEL_CENTER		((wheelSize+1)>>1)
#define WHEEL_MAX		((wheelSize-1)>>1)

/* ---- globals */

#if 0
static rgb30	requiredColors[] =
{
	_PIXEL30(1023,1023,1023)|1<<31, /* 1-bit, 2-bit, 4-bit priority */
	_PIXEL30(0,0,0)|1<<31,
	_PIXEL30(682,682,682)|1<<31,
	_PIXEL30(341,341,341)|1<<31,
	_PIXEL30(1023,0,0)|1<<31,
	_PIXEL30(0,1023,0)|1<<31,
	_PIXEL30(0,0,1023)|1<<31,
	_PIXEL30(1023,1023,0)|1<<31,

	_PIXEL30(0,1023,1023)|1<<31,
	_PIXEL30(1023,0,1023)|1<<31,
	_PIXEL30(478,0,0)|1<<31,
	_PIXEL30(478,478,0)|1<<31,
	_PIXEL30(0,478,0)|1<<31,
	_PIXEL30(0,478,478)|1<<31,
	_PIXEL30(0,0,478)|1<<31,
	_PIXEL30(478,0,478)|1<<31,
};
#endif

/* ---- code */

static void hsvMakeWheelCTab(cTab64 *ct,
	const rgb48 *bkColor,
	uint16 value)
{
	cSpec		*a;
	const rgb48	*rgb = bkColor ? bkColor:&rgb_white;
	hsv64		hsv;
	int		c, sat;

	if (!ct)
		return;
	ct->count = 256;
	ct->flags = 0x8000;
	a = ct->colors;
	a->rgb = *rgb; /* transparent */
	a[255].rgb = rgb_black; /* frame */
	a[254].rgb.red = rgb->red>>1; /* frame pseudo-antialiasing color */
	a[254].rgb.green = rgb->green>>1;
	a[254].rgb.blue = rgb->blue>>1;
	++a;
	a->rgb.red = a->rgb.green = a->rgb.blue = value; /* zero saturation */
	c = 36, hsv.hue = 0, hsv.val = value;
	do
	{
		sat = 0;
		do
		{
			++a, sat+= 0xFFFF6/7;
			hsv.sat = (uint16)(sat>>4);
			HSV2RGB(&a->rgb,&hsv);
		} while(sat<0xFFFF0);
	} while(hsv.hue+=(ufract)((4294967296.+35)/36),--c);
}

static void hsvWheelPt2HS(hsv64 *r,
	int16 left,
	int16 top,
	int16 wheelSize)
{
	int32		x, y;
	uint32		d, q;

	if (!r)
		return;
	if (wheelSize<MIN_WHEEL_SIZE)
		wheelSize = wheelSize<=0 ? DEFAULT_WHEEL_SIZE:MIN_WHEEL_SIZE;
	x = left-WHEEL_CENTER+1, y = WHEEL_CENTER-1-top;
	--wheelSize;
	d = WHEEL_MAX, q = d*d;
	if ((d=x*x+y*y)!=0)
		r->hue = ifractatan2(x,y);
	r->sat = (uint16)usqrt(MIN(d,q)*((65535*65535u)/q));
}

static void hsvWheelHS2Pt(uint32 hue,
	uint16 sat,
	pt2d16 *pt,
	int16 wheelSize)
{
	int32		x, y, z;

	if (wheelSize<MIN_WHEEL_SIZE)
		wheelSize = wheelSize<=0 ? DEFAULT_WHEEL_SIZE:MIN_WHEEL_SIZE;
	z = (wheelSize-2)*(sat+(sat>>15))>>16;

	x = ifract30mul(fastcos(hue),z)+1;
	x = (MIN(x,z)>>1)+pt->x+WHEEL_CENTER-1, 
	pt->x = PIN(x,-32768,32767);

	y = ifract30mul(fastsin(hue),z)+1;
	y = WHEEL_CENTER-1-(MIN(y,z)>>1)+pt->y, 
	pt->y = PIN(y,-32768,32767);
}

pixMap *hsvMakeWheelPixMap(const rgb48 *bkColor,
	uint16 value,
	int16 wheelSize,
	uint8 edgeThickness,
	boolean dither)
{
	pixMap		*m;
	int8		*a, *w;
	int32		x, y, s, e, d, rf, wheelEdge, wheelHEdge;

	if (wheelSize<MIN_WHEEL_SIZE)
		wheelSize = wheelSize<=0 ? DEFAULT_WHEEL_SIZE:MIN_WHEEL_SIZE;
	wheelEdge = edgeThickness ? wheelSize-edgeThickness:wheelSize+1;
	s = 2-WHEEL_CENTER*4, e = (wheelSize-WHEEL_CENTER)*4+4;
	m = newPixMap(8,wheelSize+1,wheelSize+1);
	if (pmAllocate(m)<=0 || (a=lockPixels(m))==NULL)
		{disposePixMap(m); return NULL;}
	d = dither ? 2:0, wheelEdge+= wheelEdge, wheelHEdge = wheelSize*2+1;
	rf = (uint32)((4294967295.+72)/72)-(-d&(uint32)((4294967295.+144)/144));
	for(y=s; y<e; y+=4,a+=m->pitch)
		for(x=s,w=a; x<e; x+=4,++w)
		{
			uint32		sat;
			int32		dm;

			sat = ihypot2(x,y);
			if (sat>=wheelEdge+0u)
			{
				*w = (int8)(sat>wheelHEdge+0u ? 0:255-(sat>=wheelHEdge+0u));
				continue;
			}
			dm =-((x^y)>>1&d);
			if ((sat=((sat*16-(sat&-d))/(wheelSize-1)-dm)>>2)==0)
				{*w = 1; continue;}
			*w = (int8)(ufract32mul(ifractatan2(x,-y)
				+(dm&(uint32)((4294967295.+72)/72))+rf,36)*7+sat+1);
		}
	m->pTable = allocCTable(NULL,256);
	hsvMakeWheelCTab(m->pTable,bkColor,value);
	return m;
}

/* color picker dialog */

enum
{ /* items */
	iPROMPT		= iCANCEL+1,
	iSCROLL,
	iWHEEL,
	iNEW,
	iOLD,
	iHUE,
	iSATURATION,
	iVALUE,
	iRED,
	iGREEN,
	iBLUE
};

wcItem		cPickDialog[] =
{
	{{  0,  0,282,420}, NULL, winTypeModal, 0, MK4CC('c','R','G','B')},
	{{250,302,270,362}, (const uint8*)"\2OK", ctlTypePushBtn, 0, 1},
	{{250,223,270,285}, (const uint8*)"\6Cancel", ctlTypePushBtn, 0, 1},
	{{ 13, 15, 33,240}, NULL, ctlTypeStaticText, -1},
	{{ 30,390,230,406}, NULL, ctlTypeScrollBar, 0, 65535},
	{{ 30,176,231,377}, NULL, ctlTypePixMap},
	{{ 40, 50, 75,130}, NULL, -1},
	{{ 75, 50,109,130}, NULL, -1},
	{{129, 92,145,145}, NULL, ctlTypeEditText},
	{{150, 92,166,145}, NULL, ctlTypeEditText},
	{{171, 92,187,145}, NULL, ctlTypeEditText},
	{{202, 92,218,145}, NULL, ctlTypeEditText},
	{{223, 92,239,145}, NULL, ctlTypeEditText},
	{{244, 92,260,145}, NULL, ctlTypeEditText},
	{{130, 55,144, 89}, (const uint8*)"\3Hue", ctlTypeStaticText, -1},
	{{151,  9,165, 89}, (const uint8*)"\12Saturation", ctlTypeStaticText, -1},
	{{172, 42,186, 89}, (const uint8*)"\5Value", ctlTypeStaticText, -1},
	{{203, 57,217, 89}, (const uint8*)"\3Red", ctlTypeStaticText, -1},
	{{224, 43,238, 89}, (const uint8*)"\5Green", ctlTypeStaticText, -1},
	{{245, 53,259, 89}, (const uint8*)"\5Blue", ctlTypeStaticText, -1},
	{{0}}
};

/*
 todo: temporarily replace closest colors in 4/8 bit screen modes
*/

typedef struct cPickStat
{
	hsv64		hsv;
	rgb48		rgbNew, rgbOld;
	int16		size;
	pt2d16		oldPt;
	int32		trkTime;
	windowRef	wind;
	int32		oldVal;
} cPickStat;

static void xorCircle(windowRef win,
	pt2d16 pt)
{
	rect2dq		oval;
	int		x = pt.x-3, y = pt.y-3;

	win->port.pnSize.x = win->port.pnSize.y = 1<<16;
	win->port.pnMode = mSrcXor;
	FixRect(&oval,x,y,x+7,y+7);
	rOval(&win->port,&oval);
	win->port.pnMode = mSrcCopy;
	update_screen_window_rect(win,x,y,x+7,y+7);
}

static void paintSolid(windowRef win,
	const rect16 *b,
	const rgb48 *rgb)
{
	RGBFore(&win->port,rgb);
	rFillRect(&win->port,b->left,b->top,b->right,b->bottom);
	RGBFore(&win->port,&rgb_black);
}

static int printClippedReal(char *b,
	double n,
	double res)
{
	return sprintf(b,"%g",
		(int64)(n*res+.5)/res);
}

static void updateFields(cPickStat *stat,
	int16 item)
{
	char		b[32];
	windowRef	win = stat->wind;
	ctlObj		*ctl;

	if (item>iVALUE || item==iOLD)
		RGB2HSV(&stat->hsv,&stat->rgbNew);
	else if (item)
		HSV2RGB(&stat->rgbNew,&stat->hsv);
	if (item!=iHUE)
		wSetControlText(wGetIndexedControl(win,iHUE-1),b,
			printClippedReal(b,u32AngleToDeg(stat->hsv.hue),10000));
	if (item!=iSATURATION)
		wSetControlText(wGetIndexedControl(win,iSATURATION-1),b,
			printClippedReal(b,stat->hsv.sat/655.35,500));
	if (item!=iVALUE)
		wSetControlText(wGetIndexedControl(win,iVALUE-1),b,
			printClippedReal(b,stat->hsv.val/655.35,500));
	if (item!=iRED)
		wSetControlText(wGetIndexedControl(win,iRED-1),b,
			printClippedReal(b,stat->rgbNew.red/257.,500));
	if (item!=iGREEN)
		wSetControlText(wGetIndexedControl(win,iGREEN-1),b,
			printClippedReal(b,stat->rgbNew.green/257.,500));
	if (item!=iBLUE)
		wSetControlText(wGetIndexedControl(win,iBLUE-1),b,
			printClippedReal(b,stat->rgbNew.blue/257.,500));
	if ((ctl=wGetIndexedControl(win,iWHEEL-1))!=NULL && ctl->data)
	{
		pixMap		*m = ctl->data;

		if (stat->hsv.val!=stat->oldVal)
		{
			hsvMakeWheelCTab(m->pTable,&win->port.bkColor,stat->hsv.val);
			stat->oldVal = stat->hsv.val;
		}
		if (stat->oldPt.x!=-0x8000)
			xorCircle(win,stat->oldPt);
		wDraw1Control(ctl);

		stat->oldPt.x = ctl->bounds.left, stat->oldPt.y = ctl->bounds.top;
		hsvWheelHS2Pt(stat->hsv.hue,stat->hsv.sat,&stat->oldPt,0);
		xorCircle(win,stat->oldPt);
	}
	if ((ctl=wGetIndexedControl(win,iNEW-1))!=NULL)
	{
		paintSolid(win,&ctl->bounds,&stat->rgbNew);
		update_screen_window_rect(win,ctl->bounds.left,ctl->bounds.top,
			ctl->bounds.right,ctl->bounds.bottom);
	}
}

static boolean adjustField(cPickStat *stat,
	int16 item,
	double delta)
{
	double		v = 0;
	ctlObj		*ctl;
	windowRef	win = stat->wind;
	uint32		t;

	ctl = wGetIndexedControl(win,item-1);
	if (ctl==NULL || !ctl->rtn || ctl->rtn(ctl,wcMsgReadValue,&v)<=0)
		return 0;
	v+= delta;
	switch(item)
	{
	case iHUE:
		t = degToU32Angle(v)+.5;
		if (t==stat->hsv.hue)
			return 0;
		stat->hsv.hue = t;
		break;
	case iSATURATION:
		t = PIN(v,0,100)*655.35+.5;
		if (t==stat->hsv.sat)
			return 0;
		stat->hsv.sat = t;
		break;
	case iVALUE:
		t = PIN(v,0,100)*655.35+.5;
		if (t==stat->hsv.val)
			return 0;
		stat->hsv.val = t;
		break;
	case iRED:
		t = PIN(v,0,255)*257+.5;
		if (t==stat->rgbNew.red)
			return 0;
		stat->rgbNew.red = t;
		goto updclr;
	case iGREEN:
		t = PIN(v,0,255)*257+.5;
		if (t==stat->rgbNew.green)
			return 0;
		stat->rgbNew.green = t;
		goto updclr;
	case iBLUE:
		t = PIN(v,0,255)*257+.5;
		if (t==stat->rgbNew.blue)
			return 0;
		stat->rgbNew.blue = t;
	updclr:	RGB2HSV(&stat->hsv,&stat->rgbNew);
	}
	if (item>iSATURATION)
		wSetControlValue(wGetIndexedControl(win,iSCROLL-1),65535-stat->hsv.val);
	if (delta)
		updateFields(stat,-1);
	return 1;
}

#define readField(s,i) adjustField(s,i,0)

DEF_TRK_RTN(hsvValueTrackScroll)
{
	cPickStat	*stat = (cPickStat*)ctl->refCon;
	int32		time, diff;

	if (part==kControlLabelPart || !part)
		goto set;
	time = TickCount(), diff = time-stat->trkTime;
	if (ctl->visible<2)
		time+= 17; /* ~1/3 sec total */
	else if (diff<3) /* ~1/20 sec */
		return;
	stat->trkTime = time, ctl->visible = 2;
	value = ctl->value;
	if (part==kControlUpButtonPart)
		value-= 257;
	else if (part==kControlDownButtonPart)
		value+= 257;
	else if (part==kControlPageUpButtonPart)
		value-= 4369;
	else if (part==kControlPageDownButtonPart)
		value+= 4369;
set:	if (value+0u>65535u)
		value = (~value)>>31&65535;
	if (ctl->value==value)
		return;
	wSetControlValue(ctl,(uint16)value);
	stat->hsv.val = 65535-value;
	updateFields(stat,-1);
}

boolean cPickFilter(windowRef win,
	EventRecord *event,
	int16 *itemHit)
{
	pt2d16		loc, oldloc;
	ctlObj		*ctl;
	cPickStat	*stat = NULL;
	int		r;

	global_idle_proc();
	if ((ctl=wGetIndexedControl(win,iSCROLL-1))!=NULL)
		stat = (void*)ctl->refCon;
	switch(event->what)
	{
	case updateEvt:
		if ((win=(WindowRef)event->message)==NULL)
			return 0;
		wBeginUpdate(win);
		wDrawControls(win);
		if (stat)
		{
			rect2dq		qframe;
			rect16		frame;

			frame = ctl->bounds;
			if ((ctl=wGetIndexedControl(win,iOLD-1))!=NULL)
			{
				paintSolid(win,&ctl->bounds,&stat->rgbOld);
				frame = ctl->bounds;
			}
			if ((ctl=wGetIndexedControl(win,iNEW-1))!=NULL)
			{
				paintSolid(win,&ctl->bounds,&stat->rgbNew);
				UnionRect16(&frame,&ctl->bounds);
			}
			FixRect(&qframe,frame.left-1,frame.top-1,frame.right+1,frame.bottom+1);
			rRect(&win->port,&qframe);
			//update_screen_window_rect(win,frame.left-1,frame.top-1,
			//	frame.right+2,frame.bottom+2);
			if (stat->oldPt.x!=-0x8000)
				xorCircle(win,stat->oldPt);
		}
		wEndUpdate(win);
		return 1;
	case mouseUp:
	{ /* modern style: mouse wheel -> pageUp/Down on hover item */
		if (!((int)event->message&wheelFlag))
			break;
		ctl = NULL;
		get_mouse_loc(&loc.x,&loc.y);
		wFindControl(win,loc,&ctl);
		if (wGetControlIndex(win,ctl)==iSCROLL-1)
			ctl = wGetIndexedControl(win,iVALUE-1);
		event->message = (void*)(uint8)((int)event->message&wheelDownFlag
			? kPageDownCharCode:kPageUpCharCode);
		goto adjctl;
	}
	case keyDown:
	case autoKey:
		if (!(event->modifiers&menuKey)
				&& !memchr("0123456789.e-",(uint8)event->message,13))
		{
			ctl = win->ctlFocus;
		adjctl:	if (ctl && win->ctlFocus->rtn(win->ctlFocus,
					wcMsgGetType,NULL)==ctlTypeEditText)
			{
				r = wGetControlIndex(win,ctl)+1;
				switch((uint8)event->message)
				{
				case kPageUpCharCode:
					adjustField(stat,r,10);
					break;
				case kPageDownCharCode:
					adjustField(stat,r,-10);
					break;
				case kUpArrowCharCode:
					adjustField(stat,r,1);
					break;
				case kDownArrowCharCode:
					adjustField(stat,r,-1);
					break;

				default: /* non-numeric keys ignored */
					if ((uint8)event->message>=' ')
						event->what = nullEvent;
				}
			}
		}
		break;
	case nullEvent:
	case mouseDown:
		if (stat)
		{
			get_mouse_loc(&loc.x,&loc.y);
			if ((ctl=wGetIndexedControl(win,iOLD-1))!=NULL
					&& ptInRect16(&ctl->bounds,loc))
			{
				if (event->what==nullEvent)
				{
					set_mouse_cursor(_curs_colorPick,0);
					wIdleFocus(win);
				} else
				{
					stat->rgbNew = stat->rgbOld;
					updateFields(stat,iOLD);
					wSetControlValue(wGetIndexedControl(win,iSCROLL-1),
						65535-stat->hsv.val);
				}
				return 1;
			}
			if ((ctl=wGetIndexedControl(win,iWHEEL-1))==NULL)
				return 0;
			r = (stat->size-1)>>1;
			if (!ptInRect16(&ctl->bounds,loc)
					|| (int32)ihypot2(loc.x-r-ctl->bounds.left,
					loc.y-r-ctl->bounds.top)>r)
				break;
			if (event->what==nullEvent)
			{
				set_mouse_cursor(_curs_colorPick,0);
				wIdleFocus(win);
				return 1;
			}
			while(mouse_still_down())
			{
				get_mouse_loc(&loc.x,&loc.y);
				if (oldloc.x==loc.x && oldloc.y==loc.y)
					continue;
				oldloc = loc;
				hsvWheelPt2HS(&stat->hsv,
					loc.x-ctl->bounds.left,loc.y-ctl->bounds.top,stat->size);
				updateFields(stat,iVALUE);
			}
			*itemHit = iWHEEL;
			return 1;
		}
	}
	return 0;
}

boolean rgbPickColor(int16 left,
	int16 top,
	const uint8 *prompt,
	const rgb48 *inColor,
	rgb48 *outColor)
{
	cPickStat	stat;
	windowRef	win;
	ctlObj		*ctl;
	int		i, j;
	int16		item, oldFocus;

	if (!prompt)
		prompt = (const uint8*)"\0";
	cPickDialog[iPROMPT].title = prompt;
	if ((win=wNewWindowFromTemplate(cPickDialog,(windowRef)-1))==NULL)
		return 0;
	cPickDialog[iPROMPT].title = NULL;
	memset(&stat,0,sizeof(stat));
	stat.wind = win;
	stat.size = DEFAULT_WHEEL_SIZE;
	stat.oldPt.x =-0x8000;
	stat.oldVal =-1;
	if (inColor)
		stat.rgbNew = stat.rgbOld = *inColor;
	win->ctlDefault = wGetIndexedControl(win,iOK-1);
	win->ctlCancel = wGetIndexedControl(win,iCANCEL-1);
	oldFocus = iHUE;
	win->ctlFocus = wGetIndexedControl(win,iHUE-1);
	if ((left&top)!=-1)
		OffsetRect16(&win->bounds,left-win->bounds.left,top-win->bounds.top);
	left = top = i = j = 0;
	if ((ctl=wGetIndexedControl(win,iWHEEL-1))!=NULL)
	{
		left = ctl->bounds.left, top = ctl->bounds.top;
		i = ctl->bounds.right-left;
		j = ctl->bounds.bottom-top;
		ctl->max = mDither;
		/* dithering needs this */
		PenToRGB(win->port.map,win->port.bkSolid,&win->port.bkColor);

		ctl->data = hsvMakeWheelPixMap(&win->port.bkColor,stat.hsv.val,
			stat.size=MAX(i,j),2,1);
	}
	updateFields(&stat,32767);
	if ((ctl=wGetIndexedControl(win,iSCROLL-1))!=NULL)
	{
		ctl->trk = hsvValueTrackScroll;
		ctl->refCon = (void*)&stat;
		ctl->value = 65535-stat.hsv.val;
	}
	//win->palette = allocCTable(requiredColors,GET_ARRAY_COUNT(requiredColors));
	//wActivatePalette(win);
	wShowWindow(win,1);
	item = 0;
	do
	{
		wModalDialog(cPickFilter,&item);
		if (item-iHUE+0u<iBLUE-iHUE+0u)
		{
			if (item!=oldFocus)
				if (readField(&stat,oldFocus))
					updateFields(&stat,oldFocus);
			oldFocus = item;
		}
	} while(item>iCANCEL || item<=0);
	wDisposeWindow(win);
	set_mouse_cursor(_curs_arrow,0);
	if (item==iOK && outColor)
		{*outColor = stat.rgbNew; return 1;}
	return 0;
}
