/*
 demo.c
  portable graphics and input demo

  Created: 2014.02.15 20:58 EST as "spanner"
  Revised: 2024.06.09
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#define __STDC_WANT_LIB_EXT1__ 1
#define __STDC_WANT_LIB_EXT2__ 1

#include <enviro64.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>

#ifdef WIN32
 #include <windows.h>
#endif

#define MTEXT_LEGACY_NAMES
#define MEVENT_DRAGDROP

#include "my_cseries.h"
#include "mwind.h"
#include "mevent.h"
#include "mtext.h"
#include "colorPicker.h"
#include "interpreter.h"
#include "resex.h"
#include "fencoding.h"

#define devWindowCName		".SI:mOS demo"
#define DEFAULT_BPP		24
#define DISPLAY_STATS		1

#define CW_TOPLEFT		CW_USEDEFAULT // 0
#define HSV_WHEEL_TEST		1

static void update_window(boolean updateNow);

static boolean	decorActive, doDebug;
static uint8	evtRActive;
extern uint8	mEventInWMLoop;
static boolean	appReceivedParam;
static boolean	pbufrecog;
static char	pbuf[288], imgError, imgDepth, imgBPP, imgHasAlpha;

//static CRITICAL_SECTION	drawMutex;

static timerThread	*global_timer_thread, *graphic_update_thread;

/* ---- required suite */

char		temporary[512];

void global_idle_proc(void)
{
	if (!mEventInWMLoop)
		GetNextEvent(0,NULL); /* Windows will assume we're hanging without Get/PeekMessage */
	if (evtRActive)
		return;
	evtRActive = 2;
	if (decorActive && !FrontWindow())
		update_window(0);
	//else if (!evtTActive) -- this causing too many problems
	//	wIdleControls(&screen_desktop_window);
	evtRActive = 0;
}

void graphic_update_timer(timerEntry *this)
{
	//evtTActive = 1; /* wIdleControls easily messes up Windows event polling */
	(void)int64addu32(this->mcsremaining,16626);
	if (int64n(this->mcsremaining))
		int64xset(this->mcsremaining,0);
	if (mEventInWMLoop&2)
		global_idle_proc();
	//evtTActive = 0;
}

void stop_interface_fade(void)
{
}

uint8 *getpstr(void *d,
	short ID,
	int x)
{
	const char	*s = "";

	if (!d || ID<0)
		return NULL;
	if (ID==1)
		switch(x)
		{
		case 1: s = "\77File is unsupported or unrecognized by path, extension or data."; break;
		case 2: s = "\16File is empty."; /* likely won't get here...? */
		}
	memcpy(d,s,*(const uint8*)s+1);
	return d;
}

static uint8	_dbgTemp[2048];

#define ATTASKLEVEL0()		1

int dprintf(const char *form,
	...)
{
	va_list		args;
	char		*s;
	int		len;

	va_start(args,form);

	if ((len=vsprintf_s(1+(char*)_dbgTemp,255,form,args))>255)
		len = 255;

	va_end(args);

	s = memchr(1+_dbgTemp,';',*_dbgTemp=len);
	//_dbgTemp[len+1] = 0; -- already set by vsprintf

	/* if first ';' follows "g;" */
	if (!ATTASKLEVEL0() || !s || s[1]!='g') // || s[2]!=';')
	{
#if TARGET_OS_MAC
		DebugStr(_dbgTemp);
#else
		__try{__debugbreak();} __except(EXCEPTION_EXECUTE_HANDLER) {}
#endif
	}
	if (s)
		*s = 0;
	if (ATTASKLEVEL0())
		sprintf((char*)_dbgTemp+512+strlen((char*)_dbgTemp+512),"## %s\n",
			1+_dbgTemp);
	return len;
}

void _debugMsg(const char *msg,
	const char *where,
	int n)
{
	static char	*type[] = {"warning","assertion failure","pause","halt"};
	int		z, i = n&0x3FFFFFFF;

	n = (uint32)n>>30;
	z = sprintf((char*)_dbgTemp+1,msg ? "%s in '%s', line %d: %s\n":"%s in '%s', line %d\n",
		type[n],where,i,msg);
	*_dbgTemp = MIN(z,255);
	n&= 1;
	if (n)
		(void)freopen("stderr.txt","w",stderr);
	fprintf(n ? stderr:stdout,n ? "### %s\n":"## %s\n",
		_dbgTemp+1);
	if (n)
		fflush(stderr);
	__try{__debugbreak();} __except(EXCEPTION_EXECUTE_HANDLER) {}
	if (n)
		exit(-1);
}

/* ---- code */

static char *hexToStr(char *d,
	const uint8 *s,
	int z)
{
	if (!d || z<=0)
		return d;
	do
		*d++ = _interp_defbasechars[*s>>4], *d++ = _interp_defbasechars[*s++&15];
	while(--z);
	return d;
}

static boolean ptInPtRange(pt2d16 loc,
	pt2d16 obtl,
	pt2d16 obdim)
{
	uint32		r;

	r = loc.x-obtl.x+0u;
	if (r>=(obdim.x|(obdim.x>>31))+0u)
		return 0;
	r = loc.y-obtl.y+0u;
	if (r>=(obdim.y|(obdim.y>>31))+0u)
		return 0;
	return 1;
}

typedef enum
{
#define TOK(l,s) _ft_##s,
	#include "ftypelist.h"
#undef TOK
} dFileType;

const uint8	ft_extensions[] =
#define QUOTE(name) #name
#define TOK(l,s) QUOTE(l)QUOTE(s)
	#include "ftypelist.h"
#undef TOK
;

static dFileType findFileType(const char *s,
	int z,
	boolean isExt)
{
	const uint8	*l;
	int		i, x;
	char		t;

	if (z<=0)
		return -1;
	x = 0, i = t = 0;
	if (z&1)
	{
		++i;
		if (s[0]=='.' || s[0]=='\\')
			x = i+1, t = s[0];
	}
	for(; i<z; ++i)
	{
		if (s[i]=='.' || s[i]=='\\')
			x = i+1, t = s[i];
		++i;
		if (s[i]=='.' || s[i]=='\\')
			x = i+1, t = s[i];
	}
	//if (x<0)
	//	return -1;
	s+= x, z-= x;
	//if (z>0 && s[z-1]=='!' && x<=0)
	//	--z;
	for(i=0,l=ft_extensions; (x=*l++)!=0; ++i,l+=x)
		if (x==z && !memcmp(s,l,z))
		{
			if (t=='.' || isExt)
				return i;
			return -1;
		}
	return -1;
}

static const uint8 *getFileTypeStr(dFileType ft)
{
	const uint8	*l;
	int		i = ft;

	if (i+0u>LAST_FTYPE+0u)
		return NULL;
	l = ft_extensions;
	for(++i; --i;)
		l+= *l+1;
	return l;
}

/* ---- sound (optional) */

#ifdef SNDPLAY
#include "../snddec/snddec.h"

static void playSoundListFile(const char *name)
{
	ResFile_Memory	*m = NULL;
	FILE		*f;
	void		*h;

	if (!name)
		return;
	if ((f=fopen(name,"rb"))==NULL)
		return;
	if (ResFile_LoadMap(f,0,&m,0)>=0
			&& (h=ResFile_GetIndResource(m,MK4CC('s','n','d',' '),0,
			NULL,NULL))!=NULL)
	{
		sndManager	*mgr;
		sndChannel	*ch;

		if ((mgr=sndMgrOpen(48000,1,NULL))!=NULL)
		{
			if ((ch=sndNewChannel(mgr,sndInitStereo))!=NULL)
			{
				SetCursor(LoadCursor(NULL,IDC_WAIT));
				sndPlaySndList(ch,(const uint8*)hLock(h),hSize(h));
				while(sndChannelPlaying(ch))
					sndMgrWait(mgr);
				sndMgrWait(mgr);
				SetCursor(LoadCursor(NULL,IDC_ARROW));
			}
			sndMgrClose(mgr);
		}
		hFree(h);
	}
	if (m)
		free(m);
	fclose(f);
}
#endif

/* ---- active window */

#define MAXTIMERS		4

#define MINSCRW			640
#define MINSCRH			400
#define DEFSCRW			728
#define DEFSCRH			544
#define MAXSCRW			1600
#define MAXSCRH			900

enum
{
	_it_about,
	_it_dither,
	_it_smoothing,
	_it_real_tabs,
	_it_wrap,
	_it_match_case,
	_it_whole_word,
	_it_fbox
};

static uint64	timer_start;

int		staty;
int16		ffOptFont =-1;
uint16		fontsSuccess;
uint8		*evalglobals;

#define maxOpenFiles	1

typedef struct
{
	int		prefEncoding, encoding;
	dFileType	prefFType, fType;
	char		path[260];
	rect16		bounds;
	ctlObj		*tbctl;
	ufixed		maxWidth;
	ifixed		tabSize;
	int		curLine;
} tOpenFile;

uint8		tbCount, tbFront;
tOpenFile	tbOpenFile[maxOpenFiles] = {0,0,-1,-1};
dFileType	savedftype =-1;
uint8		fbMode = 15;

char		*searchH;

const rgb48	rgb_orange = {65535,32767,0};
const rgb48	rgb_tan = {37008,29041,14906};
const rgb48	rgb_ltyellow = {65535,65535,.65*65535};
const rgb48	rgb_ltblue = {65535*.7,65535*.7,65535};
const rgb48	rgb_offwhite = {65535*.97,65535*.97,65535*.97};
uint8		grayPat[8];

static ufract	startAngle = 20.5*(1u<<30)/90;
static ufract	arcAngSpan = 351.01*(1u<<30)/90;
static rect2dq	ov, rr;

 #include "mimg_file.h"

static uint8	picMode;
static pixMap	*pic, *picInitFrame;
static int	picGIFFrame =-1;
static FILE	*picGIFFile;
static double	picGIFTime;
static GIFState	*picGIFState;

int		savedWScreenW, savedWScreenH;
static rect16	*nrctp;
static int	nrctc, nrctcs, nrctox, nrctoy;
static boolean	nrcti;

static int getTBFile(const ctlObj *ctl)
{
	uint8		i;

	for(i=0; i<tbCount; ++i)
		if (tbOpenFile[i].tbctl==ctl)
			return i;
	return -1;
}

static void ovalArc(grafPort *p,
	const rect2dq *ov,
	boolean chord)
{
#if 1
	rArc(p,ov,startAngle,arcAngSpan,!chord);
#else
 /* test ovalRgn */
	SLRgn		*argn, *brgn;
	int		r;

	argn = brgn = NULL;
	ovalRgn(&argn,ov,IMIN(int16),IMAX(int16));
	//if (p->pnSize.x<0)
	{
		copyRgn(&brgn,argn);
		offsetRgn(&argn,8,8);
		//rectRgn(&brgn,brgn->bBox.left,brgn->bBox.top,brgn->bBox.right,brgn->bBox.bottom,1);
		xorRgn(&argn,brgn);
		disposeRgn(&brgn);
	}
	if ((r=rRegion(p,argn))<0)
	{
		dprintf("ovalArc: malformed region passed to rRegion (%d)",
			r);
		exit(1);
	}
	disposeRgn(&argn);
#endif
}

#ifdef _CRTCHK
_CRTCHK(printf,2,3)
#endif
static int tbprintf(textBoxObj *tBox,
	const char *form,
	...)
{
	va_list		args;
	size_t		len;

	va_start(args,form);
	if ((len=vsprintf_s((char*)_dbgTemp+257,500,form,args))>500)
		len = 500;
	va_end(args);
	tBoxInsert(tBox,tBox->length,(char*)_dbgTemp+257,len,NULL,0,1);
	return len;
}

#if 0
const s_float80	pow10correct[] =
{
#include "..\pow10correct.h"
};
#endif

static void doDebugTest(textBoxObj *tBox,
	boolean doDebug)
{
	uint64		pmcs, nmcs, amcs;
	DWORD		prev, next, sum = 0, dif, mint = IMIN(DWORD);
	TIMECAPS	tc;
	int		c = 0, j;
#if 0
	const s_float80	twothirds = {.sgnexp=0x3FFE,.man=0xAAAAAAAAAAAAAAABull};
	const s_float80	maxf80 = {.sgnexp=0x7FFE,.man=~0ull};
	s_float80 n;

	sprint_evalue(temporary,0,0); tbprintf(tBox,"%s\n",temporary);
	sprint_evalue(temporary,900,0); tbprintf(tBox,"%s\n",temporary);
	sprint_float(temporary,twothirds,21,20,'.',',',0); tbprintf(tBox,"%s\n",temporary);
	for(j=1; j<31; ++j)
	{
		n = f80pow10_xxx4[j];
		sprint_float(temporary,n,21,20,'.',',',0);
		tbprintf(tBox,"%s / ",temporary);
		n.man--;
		sprint_float(temporary,n,21,20,'.',',',0);
		tbprintf(tBox,"%s\n",temporary);
	}
	sprint_float(temporary,maxf80,21,20,'.',',',0); tbprintf(tBox,"%s\n",temporary);
#endif
	if (_dbgTemp[512])
		tbprintf(tBox,"Debug Log:\n%s\n\n\n",
			(char*)_dbgTemp+512);
	if (!doDebug)
		return;
	tbprintf(tBox,"%d fonts loaded.\n",
		fontsSuccess);
	if (nrctc)
		tbprintf(tBox,"%d rects loaded.\n",
			nrctc);
	tbprintf(tBox,
"\nAnalyzing video...\n"
"\n"
"Window: %dx%d\n"
"Depth:  %dbpp, pSize=%d, minCmp=%d, pitch=%d\n",
		screen_desktop_window.port.map->w,screen_desktop_window.port.map->h,
		screen_desktop_window.port.map->bpp,screen_desktop_window.port.map->pSize,
		screen_desktop_window.port.map->minCmp,screen_desktop_window.port.map->pitch);
	if (screen_desktop_window.port.map->pType<=pTypeFixed)
		tbprintf(tBox,"Pixel Format: %s\n",
			screen_desktop_window.port.map->pType==pTypeIndexed
			? screen_desktop_window.port.map->pTable
			? "indexed":"indexed/implied":"fixed");
	else
		tbprintf(tBox,"Pixel Format: %s, %d:%d:%d\n",
			(screen_desktop_window.port.map->pType&15)==pTypeRawRGB
			? "RGB direct":"unspecified",
			screen_desktop_window.port.map->minCmp,
			screen_desktop_window.port.map->minCmp
			+!!(screen_desktop_window.port.map->pType&pTypeLGreen),
			screen_desktop_window.port.map->minCmp);

	tbprintf(tBox,"\n"
"Some animation variables:\n"
" oval dimensions: tlbr %.5f,%.5f,%.5f,%.5f; angles: %g+%g\n"
" img W %d, H %d; depth: %d; frame #%d\n\n",
		ov.top*(1/65536.),ov.left*(1/65536.),ov.bottom*(1/65536.),ov.right*(1/65536.),
		u32AngleToDeg(startAngle),u32AngleToDeg(arcAngSpan),
		pic->w,pic->h,pic->bpp,picGIFFrame);
	j = timeGetDevCaps(&tc,sizeof(TIMECAPS));
	if (j==TIMERR_NOERROR)
		tbprintf(tBox,"Min time resolution: %d; max: %d\n",
			tc.wPeriodMin,tc.wPeriodMax);
	else
		tbprintf(tBox,"Got error %d calling timeGetDevCaps\n",
			j);
	tbprintf(tBox,"\nPerforming time tests...\n\n");
	for(prev=GetTickCount(); (next=GetTickCount())==prev; )
		;
	Microseconds(&pmcs);
	c = 40;
	do
	{
		for(prev=next; (next=GetTickCount())==prev;)
			;
		if ((dif=next-prev)<mint)
			mint = dif;
	} while(sum+=dif,--c);
	Microseconds(&nmcs);
	tbprintf(tBox,"Current native Tick resolution: ~%gms (%d lowest)\n",
		sum/40.,mint);

	c = 3, mint = IMIN(DWORD); Microseconds(&pmcs);
	do
		Microseconds(&nmcs);
	while(pmcs==nmcs);
	do
	{
		amcs = nmcs;
		j = 80;
		do
		{
			pmcs = nmcs;
			do
				Microseconds(&nmcs);
			while(pmcs==nmcs);
		} while(--j);
		dif = int64lo(nmcs)-int64lo(amcs);
		if (dif<mint)
			mint = dif;
	} while(--c);
	tbprintf(tBox,"Microseconds resolution: ~%gmcs\n",
		((mint+4)>>3)/10.);
}

static int setTBoxFont(textBoxObj *tbox,
	int32 start,
	int16 font)
{
	const tStyleRun	*stylePeek;
	tStyleRun	style;

	memset(&style,0,sizeof(style));
	style.font = font;
	style.size = 12;
	if (start<0)
	{
		tStyleRun	*a = tbox->styles;
		int		c = tbox->styleCount;

		if (start==-2 && c>1 && a)
		{
			do
				a->font = font;
			while(++a,--c);
		} else if ((tbox->styles=realloc(tbox->styles,sizeof(tStyleRun)))!=NULL)
		{
			tbox->styleCount = 1;
			*tbox->styles = style;
			return 1;
		}
		return -2;
	}
	stylePeek = tBoxCharStyle(tbox,start,NULL);
	if (stylePeek->font==font)
		return 0;
	tBoxInsert(tbox,start,NULL,0,&style,1,1);
	return 1;
}

static void drawLineInfo(textBoxObj *tbox,
	int left)
{
	const tStyleRun	*style = tBoxCharStyle(tbox,tbox->selStart,NULL);
	int32		ss = tbox->selStart, se = tbox->selEnd, loffs;
	int		r, l, li;

	l = se-ss;
	if (l<0)
		ss = tbox->selEnd, se = tbox->selStart;
	l = se-ss;
	li = 1+tBoxCharLineIndex(tbox,ss,&loffs,NULL,0);
	r = 1+tBoxCharLineIndex(tbox,se,NULL,NULL,0);
	if (!l || (!doDebug && li==r))
	{
		ufixed		x = 0;

		tBoxCharLine(tbox,ss,&x,NULL,NULL);
		r = sprintf(temporary,l ? "%d:%d%+d":"%d:%d",
			li,ss-loffs+1,l);
		if (doDebug)
		{
			//screen_desktop_window.port.chExtra =-1<<16;
			r+= sprintf(temporary+r,"/@%d:(%d) fn%d,sz%d,st%02X,sp%Xh,x=%.*f,y=%d",
				ss,tbox->styles ? (int)(style-tbox->styles):0,
				style->font,style->size,style->style,style->special,
				x&32767 ? 4:1,x/65536.,li>0 ? tbox->lines[li-1].vpos:0);
		}
	} else
	{
		r = sprintf(temporary,"%d-%d:%d%+d",
			li,r,ss-loffs+1,l);
		if (doDebug)
		{
			const char	*s = tBoxGetText(tbox);

			r+= sprintf(temporary+r,"/@%d-%d:(%d) fn%d,sz%d,st%02X,sp%Xh,h\"%02X",
				tbox->selStart,tbox->selEnd,tbox->styles ? (int)(style-tbox->styles):0,
				style->font,style->size,style->style,style->special,
				(uint8)s[ss]);
			r+= sprintf(temporary+r,l>3 ? " %02X %02X..\"":l>2 ? " %02X %02X\"":l>1
				 ? " %02X\"":"\"",
				(uint8)s[ss+1],(uint8)s[ss+2]);
		}
	}
	l = left;
	screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,
		system_colors+_system_gray73);
	rFillRect(&screen_desktop_window.port,
		l,screen_desktop_window.ctlList->bounds.top,
		MAXSCRW,screen_desktop_window.ctlList->bounds.bottom);
	screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,&rgb_black);
	screen_desktop_window.port.txMode = mTransparent;
	li = (l+2)<<16;
	if (rTextWidth(&screen_desktop_window.port,temporary,r,0)>screen_desktop_window.port.map->w)
		screen_desktop_window.port.chExtra =-1<<15; /* condense */
	rDrawText(&screen_desktop_window.port,temporary,r,
		screen_desktop_window.ctlList->bounds.bottom-7,li);
	screen_desktop_window.port.chExtra = 0<<15;
	update_screen_rect(l,screen_desktop_window.ctlList->bounds.top-4,
		screen_desktop_window.port.map->w-l+2,
		screen_desktop_window.ctlList->bounds.bottom
		-screen_desktop_window.ctlList->bounds.top+4);
}

static void dumpResInfo(textBoxObj *tbox,
	ResFile_Memory *m,
	const char *path,
	int r)
{
	ResFile_TypeEntry	*typeEntry;

	const char	*rftypes[] = {"unknown","Mac","IIGS","Win32"};
	ResFile_Map	*map;
	int		i;

	if (!m)
	{
		tbprintf(tbox,r==-3 ? "\n### '%s' is empty\n"
			:"\n### '%s' is not a recognized resource file or corrupted (%d)\n",
			path,r);
		return;
	}
	map = &m->map;
	tbprintf(tbox,"# resource file: %s\n\n",
		path);
	tbprintf(tbox,
"# rsrc file type: %s(%d); flags 0x%X; map size %d/total %d\n"
"# resource count: %d; type count %d\n",
		m->type+0u>TYPE_RSWIN32 ? "?":rftypes[m->type],
		m->type,m->flag,m->mapLen,m->mapPos+m->mapLen,
		map->totalCount,map->typeCount);
	for(typeEntry=map->typeEntries,i=0; i<map->typeCount; ++i,++typeEntry)
	{
		uint32		j, c;

		j = 4, c = typeEntry->typeCode;
		do
			if ((c&255)-32u>126-32u)
				break;
		while(c>>=8,--j);
		c = be32_native(typeEntry->typeCode);
		if (j)
			tbprintf(tbox,"\n  %08X: ",
				c);
		else
			tbprintf(tbox,"\n  '%-4.4s': ",
				(const char*)&typeEntry->typeCode);
		c = typeEntry->count;
		r = sprintf(temporary,"%d %s\n",
			c,c==1 ? "entry":"entries");
		if (!i)
		{
			memcpy(r+temporary,"      ID    size offset flag name\n",34);
			r+= 34;
		}
		tBoxInsert(tbox,tbox->length,temporary,r,NULL,0,0);
		for(j=0; j<c; ++j)
		{
			RsrcInfo	info;

			ResFile_GetIndResource(m,typeEntry->typeCode,~j,&info,&r);
			if (r<0)
			{
				tbprintf(tbox,"\n## error %d getting rsrc info for #%d\n",
					r,j);
				continue;
			}
			r = sprintf(temporary,"% 8d % 7d %06X %04X",
				info.ID,info.sizeOnDisk,info.offsetOnDisk,info.flag);
			if (*info.name)
			{
				swr_ptr		wp;

				temporary[r++] = ' ';
				temporary[r++] = '"';
				wp.dst = temporary+r;
				wp.remaining = 512;
				r+= tprint_tchars(swr_sputc,&wp,(const char*)info.name+1,*info.name,'"',1,1);
				temporary[r++] = '"';
			}
			temporary[r++] = '\n';
			tBoxInsert(tbox,tbox->length,temporary,r,NULL,0,1);
		}
	}
}

DEF_TBDRAWLN_RTN(drawLineHook)
{
	//static int32	lastoffs;
	ctlObj		*rtctl;

	if (lr || tbox->flags&tBoxDragState)
		return 0;
	//lastoffs = i;
	rtctl = wGetIndexedControl(&screen_desktop_window,_it_real_tabs);
	if (rtctl)
	{
		SLRgn		*clip = screen_desktop_window.port.clipRgn;
		rect16		oldclip;

		oldclip = rtctl->bounds;
		if (clip)
		{ /* shouldn't get here */
			oldclip = clip->bBox;
			if (clip->size==OFFSET(SLRgn,x))
				clip->bBox.right = clip->bBox.bottom = 16000;
		}
		drawLineInfo(tbox,rtctl->bounds.right+14);
		if (clip)
			clip->bBox = oldclip;
	}
	return i;
}

DEF_TBRECLLN_RTN(recalLineHook)
{
	if (!(lc&1023))
		global_idle_proc();
	return brkcode;
}

void drawBoxerBorderText(const char *s,
	int z);
static int drawBoxerBorder(rect16 *fbounds, /* returns requested rect */
	int which);
static int boxerUpdateEvt(eventRecord event);
static void boxerClickCtl(ctlObj *ctl,
	rect16 *fbounds,
	uint16 modifiers);
static int boxerGenKey(uint16 key,
	uint16 modifiers,
	boolean autoKey);
static void drawBoxerTBFrame(void);

static void adjustScreen(boolean bBorder)
{
	rect16		mbounds;
	ctlObj		*ctl, *fbctl, *wwctl;
	textBoxObj	*tbox = NULL;
	int		w, h, i, oldl;

	fbctl = wGetIndexedControl(&screen_desktop_window,_it_fbox);
	wwctl = wGetIndexedControl(&screen_desktop_window,_it_wrap);
	w = screen_desktop_window.port.map->w-savedWScreenW;
	h = screen_desktop_window.port.map->h-savedWScreenH;
	for(ctl=screen_desktop_window.ctlList;
			ctl; ctl=ctl->next)
	{
		i = getTBFile(ctl);
		//tbox = NULL;
		if (i>=0)
			tbox = (textBoxObj*)(tbOpenFile[i].tbctl->data);
		mbounds = ctl->bounds;
		if (i>=0 || (tbox && ctl==tbox->ctlScrollV))
		{
			mbounds.right+= w;
			if (tbox && ctl==tbox->ctlScrollV)
				mbounds.left+= w;
		} else
			mbounds.top+= h;
		if (tbox && ctl==tbox->ctlScrollH)
			mbounds.right+= w;
		mbounds.bottom+= h;
		ctl->bounds = mbounds; /* dirty */
		if (i<0 && ctl!=fbctl)
			continue;
		if (i<0)
			OutsetRect16(&mbounds,2,2);
		((textBoxObj*)ctl->data)->bounds
			= mbounds;
		oldl = mbounds.left;
		mbounds.left = mbounds.top = mbounds.right
			= mbounds.bottom = 2;
		if (tbox) /* preserve left coord for custom margin */
			oldl = tbox->txbounds.left;
		tBoxSetMargins(ctl->data,i>=0 ? NULL:&mbounds,i<0);
		if (tbox)
			tbox->txbounds.left = oldl;
		if (i>=0 && wwctl && wwctl->value)
		{
			tbox->maxWidth = RECTANGLE_WIDTH(&tbox->txbounds)
				-2<<16;
			tBoxRecalc(ctl->data,0);
		}
		tBoxUpdateScrollLimits(ctl->data);
	}
	for(i=0; i<tbCount; ++i)
	{
		tbOpenFile[i].bounds.bottom+= h;
		tbOpenFile[i].bounds.right+= w;
	}

	rFillRect(&screen_desktop_window.port,0,0,
		savedWScreenW=screen_desktop_window.port.map->w,
		savedWScreenH=screen_desktop_window.port.map->h);
	mbounds = fbctl->bounds;
	if (bBorder)
	{
		drawBoxerTBFrame();
		drawBoxerBorder(&mbounds,0);
		RGBBack(&screen_desktop_window.port,system_colors+_system_gray87);
	} else
		RGBBack(&screen_desktop_window.port,&rgb_white);
	wDrawControls(&screen_desktop_window);
	update_screen_rect(0,0,
		screen_desktop_window.port.map->w,
		screen_desktop_window.port.map->h);
	staty+= h;
}

static void adjustTBFforNew(void)
{
	rect16		mbounds;
	ctlObj		*ctl = tbOpenFile[0].tbctl, *wwctl;
	textBoxObj	*tbox;
	int		i, v, oldl;

	if (!ctl || tbCount<2 || !tbOpenFile[1].tbctl)
		return;
	wwctl = wGetIndexedControl(&screen_desktop_window,_it_wrap);
	tbox = (textBoxObj*)ctl->data;
	v =-400;
	if ((ctl=tbox->ctlScrollV)!=NULL)
		v =-RECTANGLE_HEIGHT(&ctl->bounds);
	v>>= 1;
	for(ctl=screen_desktop_window.ctlList;
			ctl; ctl=ctl->next)
	{
		i = getTBFile(ctl);
		if (i>=0)
			tbox = (textBoxObj*)(tbOpenFile[i].tbctl->data);
		mbounds = ctl->bounds;
		if (i>=0 || (tbox && ctl==tbox->ctlScrollV))
		{
		} else if (tbox && ctl==tbox->ctlScrollH)
			mbounds.top+= v;
		mbounds.bottom+= v;
		ctl->bounds = mbounds; /* dirty */
		if (i!=0)
			continue;
		((textBoxObj*)ctl->data)->bounds = mbounds;
		oldl = mbounds.left;
		mbounds.left = mbounds.top = mbounds.right
			= mbounds.bottom = 2;
		if (tbox) /* preserve left coord for custom margin */
			oldl = tbox->txbounds.left;
		tBoxSetMargins(ctl->data,NULL,0);
		if (tbox)
			tbox->txbounds.left = oldl;
		if (wwctl && wwctl->value)
		{
			tbox->maxWidth = RECTANGLE_WIDTH(&tbox->txbounds)
				-2<<16;
			tBoxRecalc(ctl->data,0);
		}
		tBoxUpdateScrollLimits(ctl->data);
	}
}

static void insertCmdLine(textBoxObj *tbox,
	boolean logo,
	boolean move)
{
	const char	*s = tBoxGetText(tbox);
	int		r = setTBoxFont(tbox,tbox->length,ffFWDefault);

	if (logo)
		tbprintf(tbox,"\n>> mOS interpreter %u.%u.%u",
			INTERP_VERS>>8,INTERP_VERS>>4&15,INTERP_VERS&15);
	r = 1+tBoxCharLineIndex(tbox,tbox->selStart,NULL,NULL,1);
	if (tbox->length<2 || s[tbox->length-1]!='>' || s[tbox->length-2]!='\n')
		tBoxInsert(tbox,tbox->length,r>0
			&& tbox->lines[r-1].offs==tbox->selStart ? ">":"\n>",
			-1,NULL,0,1);
	else if (move)
		tBoxSelect(tbox,tbox->length,tbox->length);
	r = 1+tBoxCharLineIndex(tbox,tbox->selStart,NULL,NULL,1);
	tbox->selCLMinPos = tbox->lines[r-1].offs+1;
}

/* converts tabs to spaces so the pointer (^) places right */
static void untab(char *s,
	const char *e)
{
	if (!s)
		return;
	while(s<e)
		if (*s++=='\t')
			s[-1] = ' ';
}

static void __cdecl tbReportError(int err,
	char *s,
	long z,
	int32 r,
	int li,
	void *userData)
{
	if (err==_e_param_separation)
		err = _e_general_error;
	if (err<0 && (err=~err)<GET_ARRAY_COUNT(_interp_err_strs))
	{
		if (s)
		{
			untab(s,s+z);
			tbprintf(userData,"\n"
"### %.*s\n"
"### % *^\n"
"### %s",
				z,s,r+1,_interp_err_strs[err]);
		} else
			tbprintf(userData,err==_e_open_comment ? "\n>>> %s"
					:"\n>>> %s @ %d:%d",
				_interp_err_strs[err],li,r);
	}
}

static int __cdecl traceExitHook(char *s,
	long z,
	int l,
	const evalstate *state,
	void *u)
{
(void)s; (void)z; (void)l; (void*)state; (void)u;
	MGetKeys(mKeyMap);
	if ((_MKDOWN(mKeyMap,MKV_LMENU)|_MKDOWN(mKeyMap,MKV_RMENU)
			 && (_MKDOWN(mKeyMap,MKV_PERIOD)
			 || _MKDOWN(mKeyMap,MKV_Q) || _MKDOWN(mKeyMap,MKV_W)
			 || _MKDOWN(mKeyMap,MKV_F4)))
			|| _MKDOWN(mKeyMap,MKV_ESCAPE))
	{
		GetNextEvent(keyDownMask|keyUpMask,NULL);
		return -128;
	}
	return 0;
}

static boolean	tbRTRun;

static int __cdecl tbReportTrace(char *s,
	long z,
	int l,
	const evalstate *state,
	void *userData)
{
	rect16		fbounds;
	ctlObj		*tbctl, *fbctl;
	textBoxObj	*tbox;
	boolean		fbsel = 0;

	if (tbRTRun)
		return traceExitHook(s,z,l,state,userData);
	if (!s || (tbctl=tbOpenFile[tbFront].tbctl)==NULL)
		return 0;
	memcpy(temporary,s,MIN(z,255));
	untab(temporary,temporary+z);
	if (z>255)
		temporary[255] = CHR_HE, ++z;
	tbprintf(userData,"\n##% 4ld: %.*s",
		l,z,temporary);
	strcpy(temporary,"Result: ");
	z = 8+sprint_evalue(temporary+8,state->result,18);
	drawBoxerBorderText(temporary,z);
	fbctl = wGetIndexedControl(&screen_desktop_window,_it_fbox);
	fbounds = fbctl->bounds;
	tbox = tbctl->data;
	for(;;)
	{
		eventRecord	event;
		ctlObj		*ctl;

		event.what = nullEvent;
		if (WaitNextEvent(everyEvent,&event,0,NULL))
		{
			l = boxerUpdateEvt(event);
			if (l>0)
			{
				if (event.what==closeEvent)
					PostMEvent(&event); /* pass it down */
				return -128;
			}
			if (l)
				continue;
			if (fbsel)
				tBoxSelect(fbctl->data,0,99);
			fbsel = 0;
		}
		if (event.what==keyDown || event.what==autoKey)
		{
			switch((uint8)event.message)
			{
			case 'r': tbRTRun = 1;
			case kReturnCharCode:
				if (screen_desktop_window.ctlFocus==fbctl)
					break;
				return 1;
			case kFunctionKeyCharCode:
				if (!(event.modifiers&menuKey)
						|| ((uint16)event.message>>8)!=MKV_F4)
					break;
			case 'w':
			case 'q': PostMEvent(&event); /* pass it down */
			case 'a':
			case '.':
				if (event.modifiers&menuKey)
					return -128;
				break;
			case kEscapeCharCode:
				if (screen_desktop_window.ctlFocus!=fbctl)
					return -128;
				break;
			case kUpArrowCharCode:
			case kDownArrowCharCode:
		/*		tBoxScroll(tbox,tBoxLineScrollV(tbox,
					1|-((uint8)event.message==kUpArrowCharCode))
					-tbox->scrollV,0);
				break;*/
			case kHomeCharCode:
			case kEndCharCode:
			case kPageUpCharCode:
			case kPageDownCharCode:
				goto pass;
			}
			if (event.modifiers&menuKey)
				event.modifiers|= controlKey;
			l = boxerGenKey((uint16)event.message,event.modifiers,
				event.what==autoKey);
			if (l<0)
			{
				if (l==-2)
					fbsel = 1;
				continue;
			}
			if (l-1u<99)
			{
				ctl = !l ? tbctl:wGetIndexedControl(&screen_desktop_window,l);
				goto hctl;
			}
			/* pass through */
		}
		if (event.what!=autoKey || screen_desktop_window.ctlFocus==fbctl)
		{
		pass:	while(evtRActive)
				;
			evtRActive = 1;
			ctl = wControlEvent(&screen_desktop_window,&event);
			evtRActive = 0;
			if (ctl==screen_desktop_window.ctlList)
				return -128;
		hctl:	boxerClickCtl(ctl,&fbounds,event.modifiers);
		}
	}
	return 0;
}

static boolean ispath(const char *text,
	int32 z)
{
	const char	*s = text;

	while(z>0 && *s==' ')
		++s, --z;
	if (z<2)
		return 0;
	if (*s!='.')
		return z>7 && (!memcasecmp(s,"\\users\\",7) || !memcasecmp(s,"/users/",7))
			|| (z>3 && s[1]==':' && s[2]=='\\' && (*s&~0x20)-'A'+0u<=25);
	return (s[1]=='.' && s[2]=='/') || s[1]=='/' ? s-text+1:0;
}

static void executeLine(textBoxObj *tbox,
	int li,
	boolean trace)
{
	char		b[80], *s = tBoxGetText(tbox);
	const char	*n;
	evalstate	state;
	double		v;
	int32		ss = tbox->selStart;
	int32		se = tbox->selEnd;
	int		r, err = 0;
	char		c;
	boolean		move;

	if (!s)
		return;
	move = li==tbox->lineCount;
	if (ss==se)
	{
		ss = li ? tbox->lines[li-1].offs:0;
		for(r=li; r-0u<tbox->lineCount-0u; ++r)
		{
			se = tbox->lines[r].offs-1;
			if (se>0 && (c=s[se])-10u<4 && c!=12)
				goto havp;
		}
		se = tbox->length;
	havp:	while(li && ss>1 && (c=s[ss-1])-10u<4 && c!=12 && s[ss-2]=='\\')
		{
			--li;
			ss = li ? tbox->lines[li-1].offs:0;
		}
		for(r=0; ss<se && s[ss]=='>'; ++ss)
			if (++r>3)
				break;
		if (r>2)
		{ /* error code with line:char */
			while(ss<se && s[ss]!='@')
				++ss;
			if (ss==se)
				return;
			++ss;
			n = parse_number(s+ss,s+se,&v,0);
			if (!n || *n!=':' || (li=(int)v)+0u>tbox->lineCount+0u)
				return;
			r = tbox->lines[li-1].offs;
			n = parse_number(n+1,s+se,&v,0);
			if (!n)
				return;
			r+= (int)v;
			if (li+0u<tbox->lineCount+0u)
			{
				se = tbox->lines[li].offs-1;
				r = MIN(r,se);
			}
			tBoxSelect(tbox,r,r);
			return;
		}
		if (ss==se && ss==tbox->length)
		{
			li = 1+tBoxCharLineIndex(tbox,ss,&r,NULL,1);
			if (ss==r)
				goto ins;
			return;
		}
		//if (tbox->selStart!=tbox->length)
		//	tBoxInsert(tbox,tbox->length,s+ss,se-ss,NULL,0,1);
	}
	memset(&state,0,sizeof(state));
	state.flags = _allow_qrepeat|_CPEMDAS_ordering|_allow_approxeq|_implied_block_breaks;
	state.getv = eval_getv;
	state.setv = eval_setv;
	state.geti = eval_rindir;
	state.vbuf = &evalglobals;
 #if INTERP_VERS>0x199
	state.lKeys = eval_moskey;
 #endif
	if ((r=ispath(s+ss,se-ss))!=0)
	{
		FILE		*f;
		char		*buf;
		int32		z = 0;

		ss+= r-1;
		err = _e_return;
		if (s[se-1]=='/')
		{
			tbprintf(tbox,"\n>>> invalid file path @ %d:1",
				1+tBoxCharLineIndex(tbox,ss,NULL,NULL,1));
		} else
		{
			memcpy(temporary,s+ss,se-ss);
			temporary[se-ss] = 0;
			f = fopen(temporary,"r");
			if (f)
			{
				buf = freadfile(f,&z,NULL);
				tbRTRun = 0;
				if (trace)
				{
					wSetControlTitle(screen_desktop_window.ctlList,(uint8*)"\5Abort");
					tbox->flags^= tBoxModifiable;
				}
				li = evaluate_script(buf,buf+z,1,&state,
					tbReportError,trace ? tbReportTrace:traceExitHook,tbox);
				free(buf);
				if (trace)
				{
					drawBoxerBorder(NULL,2);
					wSetControlTitle(screen_desktop_window.ctlList,(uint8*)"\5Close");
					tbox->flags^= tBoxModifiable;
					state.flags&=~_within_longcomment;
					tbprintf(tbox,"\n");
				}
				err = 0;
			} else
				tbprintf(tbox,"\n>>> couldn't open file @ %d:1",
					1+tBoxCharLineIndex(tbox,ss,NULL,NULL,1));
			fclose(f);
		}
	} else
		do
			r = ss+evaluate_line(s+ss,s+se,&err,&state);
		while(r>=_e_repeat_end && traceExitHook(s+ss,se-ss,li,&state,tbox));
	if (!err)
		if ((state.flags&_within_longcomment)
		#if INTERP_VERS>0x199
			|state.lCount
		#endif
			)
			err =
		 	#if INTERP_VERS>0x199
				!(state.flags&_within_longcomment) ? _e_open_block:
			#endif
				_e_open_comment;
	if (err)
	{
		li = 1+tBoxCharLineIndex(tbox,r,NULL,NULL,1);
		r-= li>0 ? tbox->lines[li-1].offs:0;
		tbReportError(err,NULL,0,r,li,tbox);
	} else
	{
		strcpy(b,"\n>> ");
		sprint_evalue(b+4,state.result,18);
		--se;
		while(++se<tbox->length && s[se-1]-10u<4)
			;
		tBoxInsert(tbox,trace ? tbox->length:se,b,-1,NULL,0,1);
	}
ins:	insertCmdLine(tbox,0,move);
}

static int getCXEncoding(const char *s,
	int32 z)
{
	int		r = csGetXTag(ml_MIME_encodings,s,z);

	if (r<0)
	{
		if (z==4 && !memcmp(s,"utf8",4))
			return _csUTF8;
		return 0;
	}
	return ml_MIME_encodings[r].ID;
}

enum
{ /* fbox modes */
	_goto_line,
	_set_tabsize,
	_set_maxwidth,
	_reload_ftype,
	_reload_encoding,
	_find_text,

	_line_interpreter	= 16
};

#define defaultFBoxWidth	94

void drawBoxerBorderText(const char *s,
	int z)
{
	screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,
		system_colors+_system_gray15);
	rFillRectMode(&screen_desktop_window.port,0,screen_desktop_window.port.map->h-15,
		MAXSCRW,MAXSCRH+8,mSrcCopy);
	screen_desktop_window.port.txMode = mTransparent;
	RGBFore(&screen_desktop_window.port,system_colors+_system_gray73);
	rDrawText(&screen_desktop_window.port,
		s,z,screen_desktop_window.port.map->h-2,8<<16);
	update_screen_rect(0,screen_desktop_window.port.map->h-16,
		MAXSCRW,15);
}

static int drawBoxerBorder(rect16 *fbounds,
	int which)
{
	rect16		b;
	textBoxObj	*tbox;
	const char	*text = NULL;
	ufixed		x = 0;
	int		r;

	if (!fbounds)
	{
		ctlObj		*fbctl = wGetIndexedControl(&screen_desktop_window,_it_fbox);

		fbounds = &b;
		if (fbctl)
			b = fbctl->bounds;
	}
	if (which<=0)
	{
		//screen_desktop_window.port.opColor = system_colors[_system_gray33];
		//RGBFore(&screen_desktop_window.port,&rgb_black);
		screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,
			system_colors+_system_gray15);
		rFillRectMode(&screen_desktop_window.port,0,screen_desktop_window.port.map->h-46,
			MAXSCRW,screen_desktop_window.port.map->h-45,mSrcCopy);
		screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,
			system_colors+_system_gray73);
		rFillRectMode(&screen_desktop_window.port,0,screen_desktop_window.port.map->h-45,
			MAXSCRW,screen_desktop_window.port.map->h-44,mSrcCopy);
		screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,
			system_colors+_system_gray67);
		rFillRectMode(&screen_desktop_window.port,0,screen_desktop_window.port.map->h-44,
			MAXSCRW,screen_desktop_window.port.map->h-16,mSrcCopy);
		screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,
			system_colors+_system_gray47);
		rFillRectMode(&screen_desktop_window.port,0,screen_desktop_window.port.map->h-16,
			MAXSCRW,screen_desktop_window.port.map->h-15,mSrcCopy);
		screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,&rgb_black);
		rFillRectMode(&screen_desktop_window.port,0,screen_desktop_window.port.map->h-15,
			MAXSCRW,screen_desktop_window.port.map->h-14,mSrcCopy);
	}
	screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,
		system_colors+_system_gray15);
	rFillRectMode(&screen_desktop_window.port,0,screen_desktop_window.port.map->h-14,
		MAXSCRW,MAXSCRH+8,mSrcCopy);
	screen_desktop_window.port.txMode = mTransparent;
	RGBFore(&screen_desktop_window.port,system_colors+_system_gray73);
	r = fbMode&15;
	if (r<15)
		text = r==_goto_line ? "Go to line:"
			:r==_set_tabsize ? " Tab size:"
			:r==_set_maxwidth ? " Max width:"
			:r==_reload_ftype ? "Reload in file type:"
			:r==_reload_encoding ? "Reload in encoding:"
			:r==_find_text ? " Find text:"
			:fbMode&_line_interpreter ? ">>"
			:"Something ain't right...";
	else if (tbOpenFile[tbFront].fType>=0)
		text = tbOpenFile[tbFront].path;
	x = rDrawText(&screen_desktop_window.port,
		text,-1,fbounds->bottom-3,8<<16);
	if (r<15)
		r = 1, OffsetRect16(fbounds,8+(x>>16)-fbounds->left,0);
	else if (text)
	{
		r = 0;
		if (tbOpenFile[tbFront].tbctl && (tbox=tbOpenFile[tbFront].tbctl->data)!=NULL
				&& tbox->flags&tBoxWasModified)
			rDrawText(&screen_desktop_window.port,
				" *",2,fbounds->bottom-3,x);
	}
	RGBFore(&screen_desktop_window.port,&rgb_black);
	screen_desktop_window.port.opColor = system_colors[_system_gray87];
	if (which==2)
		update_screen_rect(0,screen_desktop_window.port.map->h-16,
			MAXSCRW,15);
	return r;
}

static void beginFBox(ctlObj *fbctl,
	rect16 *fbounds,
	const char *s,
	int len,
	int new_mode)
{
	boolean		show_find;

	if (len==-255 && s)
		len = *(const uint8*)s, ++s;
	wSetControlText(fbctl,s,len);
	if ((fbMode&15)!=new_mode)
	{
		fbMode&=~15;
		fbMode|= new_mode;
		if (drawBoxerBorder(fbounds,0)==1)
			wSetControlBounds(fbctl,fbounds);
		wShowControl(fbctl);
		wDrawControls(&screen_desktop_window);
		update_screen_rect(0,fbounds->bottom-50,
			screen_desktop_window.port.map->w,
			50);
	}
	wCtlNewFocus(fbctl,0);
	tBoxSelect(fbctl->data,0,IMAX(int));
	show_find = new_mode==_find_text;
	wSetControlVisible(wGetIndexedControl(&screen_desktop_window,_it_match_case),show_find);
	wSetControlVisible(wGetIndexedControl(&screen_desktop_window,_it_whole_word),show_find);
}

static boolean fbModeCanKey(int mode,
	uint16 key,
	int curlen)
{
	switch(mode&=15)
	{
	case _reload_encoding:
	case _reload_ftype:
		return (uint8)key+0u<' ' || (uint8)key==127 || curlen<16;
	case _goto_line:
		if ((uint8)key=='.')
			return FALSE;
	case _set_tabsize:
	case _set_maxwidth:
		if ((uint8)key+0u<' ' || (uint8)key==127)
			return TRUE;
		if ((uint8)key-'0'+0u>9u)
			return FALSE;
		return curlen<11;

	default:
		return TRUE;
	}
}

static void drawBoxerTBFrame(void)
{
	rect2dq		frame;

	FixRect16(&frame,&tbOpenFile[tbFront].bounds);
	RGBFore(&screen_desktop_window.port,&rgb_tan);
	screen_desktop_window.port.pnMode = mBlend;
	screen_desktop_window.port.opColor = system_colors[_system_gray73];
	screen_desktop_window.port.pnSize.x = screen_desktop_window.port.pnSize.y = 2<<16;
	rRect(&screen_desktop_window.port,&frame);
	screen_desktop_window.port.pnSize.x = screen_desktop_window.port.pnSize.y = 4<<16;
	rRect(&screen_desktop_window.port,&frame);
	RGBFore(&screen_desktop_window.port,&rgb_white);
	screen_desktop_window.port.pnSize.x = screen_desktop_window.port.pnSize.y = 6<<16;
	screen_desktop_window.port.opColor = system_colors[_system_gray67];
	rRect(&screen_desktop_window.port,&frame);
	screen_desktop_window.port.pnSize.x = screen_desktop_window.port.pnSize.y = 8<<16;
	rRect(&screen_desktop_window.port,&frame);
	screen_desktop_window.port.pnSize.x = screen_desktop_window.port.pnSize.y = 10<<16;
	rRect(&screen_desktop_window.port,&frame);
}

static int boxerUpdateEvt(eventRecord event)
{
	rect16		fbounds;
	ctlObj		*fbctl;
	int		h;

	if (event.what==closeEvent)
		return 2;
	fbctl = wGetIndexedControl(&screen_desktop_window,_it_fbox);
	memset(&fbounds,0,sizeof(fbounds));
	if (fbctl)
		fbounds = fbctl->bounds;
	if (event.what==updateEvt)
	{
		wBeginUpdate(&screen_desktop_window);
		RGBBack(&screen_desktop_window.port,system_colors+_system_gray87);
		drawBoxerBorder(&fbounds,0);
		wDrawControls(&screen_desktop_window);
		wEndUpdate(&screen_desktop_window);
		return -1;
	}
	if (event.what==osEvt)
	{
		h = (uint32)event.message>>24;
		if (h==appScreenChangeMessage)
			adjustScreen(1);
		else if (h==appRequestMessage && (int)event.message&changeFlag)
			return 1;
		return -1;
	}
	return 0;
}

static void boxerClickCtl(ctlObj *ctl,
	rect16 *fbounds,
	uint16 modifiers)
{
	ctlObj		*tbctl = tbOpenFile[tbFront].tbctl, *rtctl, *fbctl;
	textBoxObj	*tbox;
	int		r = 0, i;

	if (!tbctl)
		return;
	tbox = tbctl->data;
	fbctl = wGetIndexedControl(&screen_desktop_window,_it_fbox);
	rtctl = wGetIndexedControl(&screen_desktop_window,_it_real_tabs);
	if (modifiers==1)
		goto tgctl;
	if (modifiers==2)
		goto ltabz;
	if ((i=wGetControlIndex(&screen_desktop_window,ctl))==_it_wrap)
	{
		if (modifiers&controlKey)
		{
			r = sprintf(temporary,"%.9g",
				fix2double(tbOpenFile[tbFront].maxWidth,16,5));
			fbounds->right = fbounds->left+defaultFBoxWidth;
			beginFBox(fbctl,fbounds,temporary,r,
				_set_maxwidth);
			r = 1;
		} else
		{
			wSetControlValue(ctl,!ctl->value);
			wShowControl(tbox->ctlScrollH);
		}
		tbox->maxWidth = tbOpenFile[tbFront].maxWidth;
		if (ctl->value)
		{
			tbox->maxWidth = RECTANGLE_WIDTH(&tbox->txbounds)-2<<16;
			tBoxUpdateScrollLimits(tbox);
		}
		if (r)
			goto ltabz;
		goto cmupd;
	} else if (ctl==rtctl)
	{
		if (modifiers&controlKey)
		{
			r = sprintf(temporary,"%.9g",
				fix2double(tbOpenFile[tbFront].tabSize,-16,5));
			fbounds->right = fbounds->left+defaultFBoxWidth;
			beginFBox(fbctl,fbounds,temporary,r,
				_set_tabsize);
		} else
			wSetControlValue(ctl,!ctl->value);
	ltabz:	r = rtctl->value ? tbOpenFile[tbFront].tabSize:0;
		if (tbox->tabAlign!=r)
		{
			tbox->tabAlign = r;
		cmupd:	tBoxUpdate(tbox,NULL,0);
		}
		ctl = tbctl;
		r =-1;
	} else if (i==_it_match_case || i==_it_whole_word)
	tgctl:	wSetControlValue(ctl,!ctl->value);
	if (ctl==tbctl || r<0)
		drawLineInfo(tbox,rtctl->bounds.right+14);
}

static void fbSetNoMode(void)
{
	ctlObj		*ctl;
	int		c;

	for(c=3,ctl=wGetIndexedControl(&screen_desktop_window,_it_match_case);
			ctl && c; --c,ctl=ctl->next)
		wHideControl(ctl);
	fbMode|= 15;
}

static int boxerGenKey(uint16 key,
	uint16 modifiers,
	boolean autoKey)
{
	rect16		fbounds;
	ctlObj		*tbctl = tbOpenFile[tbFront].tbctl, *fbctl, *ctl, *whctl;
	textBoxObj	*tbox, *fbox;
	int		r, li = tbOpenFile[tbFront].curLine;

	fbctl = wGetIndexedControl(&screen_desktop_window,_it_fbox);
	if (!tbctl || !fbctl)
		return -1;
	tbox = tbctl->data, fbox = fbctl->data;
	fbounds = fbctl->bounds;
	if (modifiers&menuKey)
	{
		if (autoKey)
			return -1;
		if ((uint8)key=='d')
		{
			doDebug^= 1;
			update_screen_rect(0,0,
				screen_desktop_window.port.map->w,
				screen_desktop_window.port.map->h);
			return 0;
		} else if ((key>>8)==MKV_T)
		{
			if (!(modifiers&optionKey) || modifiers&controlKey)
				return _it_real_tabs;
			free(tbox->styles); tbox->styles = NULL;
			tbox->styleCount = 0;
			tBoxUpdate(tbox,NULL,0);
		} else if ((key>>8)==MKV_R)
		{
			if (modifiers&optionKey)
				return _it_wrap;
			if (!fbctl || modifiers&controlKey)
				return -1;
			fbounds.right = fbounds.left+defaultFBoxWidth;
			if (modifiers&shiftKey)
			{
				r = tbOpenFile[tbFront].prefEncoding ? tbOpenFile[tbFront].prefEncoding
					:tbOpenFile[tbFront].encoding;
				temporary[makebasestring(temporary,NULL,r,10,0,'-')] = 0;
				beginFBox(fbctl,&fbounds,
					r && r!=_csMacRoman
					? temporary:"mac",-1,
					_reload_encoding);
			} else
				beginFBox(fbctl,&fbounds,
					(const char*)getFileTypeStr(tbOpenFile[tbFront].fType),-255,
					_reload_ftype);
			return -1;
		} else if ((uint8)key=='l')
		{
			r = 1+tBoxCharLineIndex(tbox,modifiers&shiftKey ? tbox->length
				:(tbox->selStart+tbox->selEnd)>>1,NULL,NULL,1);
			r = makebasestring(temporary,NULL,r,10,1,0);
			fbounds.right = fbounds.left+defaultFBoxWidth;
			beginFBox(fbctl,&fbounds,temporary,r,
				_goto_line);
		} else if ((uint8)key=='f')
		{
			fbounds.right = fbounds.left+352;
			r = searchH ? hSize(searchH):0;
			beginFBox(fbctl,&fbounds,hLock(searchH),r,
				_find_text);
			hUnlock(searchH);
		} else if ((key>>8)==MKV_SPACE)
		{
			tBoxClick(tbox,mEventState.where,tMenuOnly,0);
			return -1;
		} else if ((key>>8)==MKV_M)
		{
			if (screen_desktop_window.ctlFocus==fbctl)
			{
				if (!(modifiers&optionKey)
						&& (fbMode&15)==_find_text)
					return _it_match_case;
			} else
			{
				if (modifiers&optionKey)
				{
					setTBoxFont(tbox,-1,ffFWDefault);
					tBoxUpdate(tbox,NULL,0);
				}
				return 0;
			}
		} else if ((key>>8)==MKV_O)
		{
			if (screen_desktop_window.ctlFocus!=fbctl
					&& modifiers&optionKey)
			{
				setTBoxFont(tbox,-2,ffOptFont);
				tBoxUpdate(tbox,NULL,0);
			}
			return 0;
		} else if ((uint8)key=='h')
		{
			if (screen_desktop_window.ctlFocus==fbctl
					&& (fbMode&15)==_find_text)
				return _it_whole_word;
		} else if ((uint8)key=='e')
		{
			const char	*s;

			li = MIN(tbox->selStart,tbox->selEnd);
			r = MAX(tbox->selStart,tbox->selEnd);
			if (li==r)
				return -1;
			hFree(searchH); searchH = NULL;
			s = tBoxGetText(tbox);
			if (!s)
				return -1;
			searchH = tBoxCopyText(tbox,li,r);
			if (fbctl->visible && (fbMode&15)==_find_text && fbctl)
			{
				r-= li;
				tBoxSetTextRaw(fbox,s+li,r);
				tBoxUpdate(fbox,0,0);
				tBoxSelect(fbox,-1,0);
			}
			return -1;
		} else if ((uint8)key=='g')
			goto fagain;
		return 0;
	}
	if ((uint8)key=='\t')
	{ /* must process tabs before general control event handler */
		if (screen_desktop_window.ctlFocus==fbctl)
		{
			tBoxSelect(fbox,0,99);
			return -1;
		}
		tBoxKey(screen_desktop_window.ctlFocus
			? screen_desktop_window.ctlFocus->data:tbox,
			'\t',modifiers);
		return -1;
	}  else if ((key>>8)==MKV_F3)
	{
	fagain:	if (fbctl->visible && (fbMode&15)==_find_text)
			goto ffvis;
		goto tfind;
	}
	if (screen_desktop_window.ctlFocus!=fbctl)
		return -1;
	r = fbMode&15;
	if (r<15 && ((uint8)key==kReturnCharCode
			|| (uint8)key==kEnterCharCode
			|| (uint8)key==kEscapeCharCode))
	{
		const char	*s, *e;
		double		v;

		if (autoKey)
			return -1;
		if (r==_find_text)
		{
			if ((uint8)key==kEscapeCharCode)
				goto nomode;
		ffvis:	hFree(searchH);
			searchH = tBoxCopyText(fbox,0,IMAX(int));
			if (searchH==NULL)
				goto failb;
			//tBoxSetTextRaw(fbox,NULL,0);
		tfind:	ctl = wGetIndexedControl(&screen_desktop_window,_it_match_case);
			whctl = wGetIndexedControl(&screen_desktop_window,_it_whole_word);
			r = hSize(searchH);
			li = MIN(tbox->selStart,tbox->selEnd);
			li = tBoxSearch(tbox,modifiers&shiftKey ? li-1
				:li+(r==MAX(tbox->selStart,tbox->selEnd)-li),
				hLock(searchH),r,(modifiers&shiftKey ? tSearchBackwards:0)
				 |(ctl && ctl->value ? tSearchCaseSensitive:0)
				 |(whctl && whctl->value ? tSearchWholeWords:0));
			hUnlock(searchH);
			if (li<0)
			{
				if (fbctl->visible && (fbMode&15)==_find_text)
					goto failb;
			} else
				tBoxSelect(tbox,li,li+r);
			return -1;
		}
		v = 0;
		if (r==_reload_encoding)
		{
			if ((uint8)key==kEscapeCharCode
					|| !fbox->length)
				goto nomode;
			s = tBoxGetText(fbox), e = s+fbox->length;
			v =-1;
			s = parse_number(s,e,&v,0);
			if (s==e && v>=0)
				tbOpenFile[tbFront].prefEncoding = v;
			else
				tbOpenFile[tbFront].prefEncoding = getCXEncoding(tBoxGetText(fbox),
					fbox->length);
			if (tbOpenFile[tbFront].prefEncoding || v>=0)
				return 128;
			goto nomode;
		}
		if (r==_reload_ftype)
		{
			if ((uint8)key==kEscapeCharCode
					|| !fbox->length)
				goto nomode;
			tbOpenFile[tbFront].prefFType = findFileType(tBoxGetText(fbox),
				fbox->length,1);
			if (tbOpenFile[tbFront].prefFType>=0)
			{
				*pbuf = 0; /* prevent erroneous alert_user */
				return 128;
			}
			goto nomode;
		}
		if ((uint8)key==kEscapeCharCode
				|| !fbctl->rtn(fbctl,wcMsgReadValue,&v))
			goto nomode;
		if (r==_set_tabsize)
		{
			tbOpenFile[tbFront].tabSize = v*65536+.5;
			if (tbOpenFile[tbFront].tabSize-16384u>(1000<<16)-16384u)
				goto failb;
			goto nomode;
		}
		if (r==_set_maxwidth)
		{
			tbOpenFile[tbFront].maxWidth = v*65536+.5;
			if (tbOpenFile[tbFront].maxWidth-1048576u>(60000<<16)-1048576u)
			{
			failb:	wShowControl(fbctl);
				//screen_desktop_window.ctlFocus = tbctl;
				//tbox->flags|= tBoxFocusState;
				wCtlNewFocus(fbctl,0);
				tBoxSelect(fbox,-1,0); /* flash our box */
				//wMcsDelay(30000);
				return -2;
			}
			goto nomode;
		}
		if ((r=v)>0 && tbox->lines)
		{
			r = MIN(r,tbox->lineCount);
			if (r!=li)
			{
				r = tbox->lines[r-1].offs;
				tBoxSelect(tbox,r,r);
				1+tBoxCharLineIndex(tbox,r,NULL,&r,1);
				if (r+6>tbox->txbounds.bottom)
					tBoxScroll(tbox,12,0);
			}
		} else
			tBoxSelect(tbox,tbox->selStart,tbox->selEnd);
	nomode:	fbSetNoMode();
		drawBoxerBorder(&fbounds,0);
		wDrawControls(&screen_desktop_window);
		update_screen_rect(0,fbounds.top-50,
			screen_desktop_window.port.map->w,
			fbounds.bottom-fbounds.top+50);
		wCtlNewFocus(tbctl,0);
		return -1;
	} else if (!fbModeCanKey(fbMode,key,fbox->length))
		return -1;
	return 0;
}

static int isFileEmpty(const char *path)
{
	int		f, r;

	if ((f=_open(path,_O_RDONLY|_O_BINARY))<0)
		return f;
	r = _filelength64(f)<=0;
	_close(f);
	return r;
}

static int boxer(const char *path,
	rect16 *bounds,
	const char *stylesPath,
	dFileType ftype,
	boolean doPassword)
{
	static int16	msglogskip;
	rect16		fbounds;
	pt2dq		oldpens;
	rgb48		oldfore, oldback;
	timerEntry	*busytimer = NULL;
	FILE		*f = NULL;
	textBoxObj	*tbox = NULL, *fbox = NULL;
	ctlObj		*tbctl = NULL, *fbctl = NULL, *rtctl = NULL, *wwctl = NULL;
	ctlObj		*csctl = NULL, *whctl = NULL;
	void		*bufH = NULL;
	long long	z = 0;
	int		lastlineCount =-1;
	int32		r, i, li =-1;
	boolean		fbsel = 0, logo = 1;

	++tbCount;
	if (tbOpenFile[tbFront].prefFType>=0)
		ftype = tbOpenFile[tbFront].prefFType;
	tbOpenFile[tbFront].bounds = *bounds;
	tbOpenFile[tbFront].maxWidth = 7000<<16;
	tbOpenFile[tbFront].tabSize = 56<<16;
	tbOpenFile[tbFront].curLine =-1;
	fbMode&=~_line_interpreter;
	if (path && ftype+0u>=_ft_txt)
	{
		const int8	*n;
		int8		*bufA = NULL;
		long		lastnl = 0, longestln =-1;

		strcpy_s(tbOpenFile[tbFront].path,sizeof(tbOpenFile[tbFront].path)-1,path);
		if ((f=fopen(path,"rb"))==0)
			{r =-43; goto err;}
		if (global_timer_thread)
			PrimeTimer(global_timer_thread,busytimer
				=InstallTimer(global_timer_thread,mOSBusyCursor_timer),0.5e6);
		z = _filelength64(_fileno(f));
		if (z-0u>(1ull<<(sizeof(ULONG_PTR)*8-1))-2u)
			{r =-1309; goto end;}
		r = tbOpenFile[tbFront].prefEncoding;
		if (r<_csUTF16BE || r>_csUTF32LE)
			r = csTestBOM(f);
		if ((bufH=hNew(z,0))==NULL)
		{
			r =-2;
		err:	if (r!=msglogskip)
				dprintf("error %d loading file;g;",
					r);
			path = stylesPath = NULL;
			msglogskip = r;
			goto skipld;
		}
		bufA = hLock(bufH);
		z = fread(bufA,1,z,f);
		if (!r)
			if (tbOpenFile[tbFront].prefEncoding)
				r = tbOpenFile[tbFront].prefEncoding;
			else
			{
				n = memchr(bufA,0,MIN(64,z));
				if (n)
					if (n-bufA<z+1 && !n[1])
						ftype = _ft_bin;
					else if (n==bufA+1 && n-bufA<z+5 && n[1] && !n[2] && n[3])
						r = (n-bufA)&1 ? _csUTF16LE:_csUTF16BE; /* likely */
			}
		tbOpenFile[tbFront].encoding = r;
		if (r>0 && r!=_csMacRoman && r!=_cpMacRoman)
		{
			csUTF_reader	reader;
			const uint8	*rpos, *epos;
			uint8		*wpos;

			if (!msglogskip)
				dprintf("text is translated from %s(%d)!;g;",
					r==_csUTF8 ? "UTF-8"
					:r==_csUTF16BE ? "UTF-16BE"
					:r==_csUTF16LE ? "UTF-16LE":"",r);
			msglogskip = 1;
			reader = csGetUTFReader(r);
			rpos = wpos = (uint8*)bufA, epos = z+rpos;
			if ((r=reader(&rpos))>=0 && r!=0xFEFF)
				writeUTF_MacRoman(r,&wpos);
			while(rpos<epos && (r=reader(&rpos))>=0)
			{
				writeUTF_MacRoman(r,&wpos);
				/*pch = r;*/
				i = wpos-(uint8*)bufA;
				if (!(i&65535))
					global_idle_proc();
				if (r-10u>3u || r==12)
					continue;
				r = i-lastnl;
				if (r>longestln)
					longestln = r;
				lastnl = i+1;
			}
			z = wpos-(uint8*)bufA;
		} else
		{
			for(i=0; i<z; ++i)
			{
				r = bufA[i];
				if (!(i&65535))
					global_idle_proc();
				if (r-10u>3u || r==12)
					continue;
				r = i-lastnl;
				if (r>longestln)
					longestln = r;
				lastnl = i+1;
			}
		}
		if (longestln<0)
			longestln = z;
		r = MIN(longestln,9000);
		if (r>1000)
			tbOpenFile[tbFront].maxWidth = (r*7)<<16;
		fclose(f); f = NULL;
		bufA = NULL;
		hUnlock(bufH);
		hResize(bufH,z);
	}

skipld:	r = 0;
	tbctl = wNewControl(&screen_desktop_window,bounds,0,1,
			(tBoxAutoVScroll|tBoxAutoHScroll|tBoxInitHideScrolls|tBoxDragDrop
			|tBoxUseCustBgColor)<<16,
			0,0,ctlTypeStaticText,0);
	//screen_desktop_window.port.clipRgn->bBox = saveClip;
	if (tbctl==NULL || (tbox=tbctl->data)==NULL)
	{
		if (bufH)
			{hFree(bufH); bufH = NULL;}
		goto end;
	}
	tbOpenFile[tbFront].tbctl = tbctl;
	tbox->custBgColor = rgb_offwhite;
	assert(tbox->textH==NULL);
	if (bufH)
		tbox->textH = bufH, tbox->length = z;

	tbox->recalLine = recalLineHook;
	/* ctlTypeStaticText always resets its margins */
	//mbounds.left = mbounds.top = mbounds.right = mbounds.bottom = 2;
	//tBoxSetMargins(tbox,&mbounds);
	tbctl->bounds.left-= 3;
	tbox->bounds.left-= 3;

	oldfore = screen_desktop_window.port.fgColor;
	oldback = screen_desktop_window.port.bkColor;
	oldpens = screen_desktop_window.port.pnSize;
	RGBBack(&screen_desktop_window.port,system_colors+_system_gray87);
	rTextFace(&screen_desktop_window.port,helvetica,12,0);

	wHideControl(wGetIndexedControl(&screen_desktop_window,_it_dither));
	wHideControl(wGetIndexedControl(&screen_desktop_window,_it_smoothing));

	rtctl = wGetIndexedControl(&screen_desktop_window,_it_real_tabs);
	if (rtctl)
		rtctl->visible = 1;
	wwctl = wGetIndexedControl(&screen_desktop_window,_it_wrap);
	if (wwctl)
	{
		wwctl->visible = 1;
		wwctl->value = 1;
	}
	csctl = wGetIndexedControl(&screen_desktop_window,_it_match_case);
	whctl = wGetIndexedControl(&screen_desktop_window,_it_whole_word);
	fbctl = wGetIndexedControl(&screen_desktop_window,_it_fbox);

	if (stylesPath)
	{
		ResFile_Memory	*m = NULL;
		boolean		haveFailMsg = strcmp(pbuf,path) && strcmp(pbuf,stylesPath);

		if (!*stylesPath)
			stylesPath = NULL, haveFailMsg&= ftype>_ft_ppm;
		else if (ftype-_ft_ppm+0u<_ft_sfil-_ft_ppm)
			goto unsup; /* if we got here... */
		else if (ftype-_ft_sfil+0u<=_ft_bin-_ft_sfil)
		{
			if (strcmp(pbuf,stylesPath) && !pbufrecog)
			{
			unsup:	tbprintf(tbox,"### '%s' is %s\n",
					pbuf,isFileEmpty(pbuf)>0 ? "empty"
					:"unsupported or could not be recognized by path or extension");
				haveFailMsg = 0;
			} else
			{
				int32		offs = 0, rsrclen;

				tBoxSetTextRaw(tbox,0,0);
				setTBoxFont(tbox,-1,ffFWDefault);
				if (path==stylesPath || ftype==_ft_bin || (f=fopen(stylesPath,"rb"))==NULL)
				{
					f = fopen(path,"rb");
					if ((r=MacBinaryDetect(_fileno(f),&offs,&rsrclen))>0)
					{
						offs = (offs+r+127)&~127;
						tbprintf(tbox,"# MacBinary wrapper\n");
					}
				}
				if (f)
				{
					r = ResFile_LoadMap(f,0,&m,offs);
					if (r==-39 && !_filelength64(_fileno(f)))
						r =-3;
					dumpResInfo(tbox,m,stylesPath,r);
				} else
					tbprintf(tbox,"### '%s' could not be opened\n",
						stylesPath);
				fclose(f); f = NULL;
			}
		} else if ((f=fopen(stylesPath,"rb"))!=0)
		{
			void		*h;

			if (ResFile_LoadMap(f,0,&m,0)>=0
					&& (h=ResFile_GetIndResource(m,MK4CC('s','t','y','l'),0,
					NULL,NULL))!=NULL)
			{
				uint8		*p = (uint8*)hLock(h);
				int16		c = (int16)(*p<<8|p[1]);
				int		rc = (hSize(h)-2)/sizeof(tOldStyleRun);

				if (c<rc)
					c = rc;
				tbox->textH = NULL;
				tbox->length = 0;
				tBoxImport(tbox,0,(char*)hLock(bufH),z,(const tOldStyleRun*)(p+2),c,1);
				hFree(bufH); bufH = NULL;
				hFree(h);
				tBoxSelect(tbox,0,0);
			}
			fclose(f); f = NULL;
		} else if (ftype>=_ft_csv)
			if (setTBoxFont(tbox,-1,ffFWDefault)>0)
			{
				tbox->styles->size = 12;
				if (rtctl)
					rtctl->value = 1;
				if (wwctl)
					wwctl->value = 0;
			}
		if (m)
			free(m);
		if (*pbuf && (unsigned int)ftype>LAST_IMG_FTYPE)
		{
			if (haveFailMsg)
			{
				alert_user(infoError,1,(isFileEmpty(pbuf)>0)+1,0);
				*pbuf = 0;
			}
		} else
		{
			tbprintf(tbox,"# '%s':\n",
				pbuf);
			if (pic && !imgError)
			{
				i = pic->pTable ? pic->pTable->count:1<<31;
				r = 1<<imgDepth; //(imgDepth>31 ? 24:imgDepth);
				tbprintf(tbox,imgDepth==31 ? " %d x %d, %d bpp (bitfield)\n"
					:" %d x %d, %d bpp/%u max colors%s\n",
					pic->w,pic->h,imgBPP,MIN(i+0u,r+0u),
					imgHasAlpha ? " - has alpha":"");
			} else
				tbprintf(tbox,"## not loaded\n");
		}
	}
	if (!stylesPath)
		if (!path)
		{
			setTBoxFont(tbox,-1,ffFWDefault);
			doDebugTest(tbox,doDebug);
		} else if (ftype<0 && (tbox->styles=realloc(tbox->styles,
				sizeof(tStyleRun)*5))!=NULL)
		{ /* 0 size for style #2 defaults to 12 */
			tStyleRun	*styles = tbox->styles;

			memset(styles,0,sizeof(tStyleRun)*5);
			tbox->styleCount = 5;
			styles[2].font = newYork;
			styles[2].size = 14;
			styles[1].font = styles[3].font = helvetica;
			styles[3].style = styles[2].style = _style_bold;
			styles[1].special = _special_align_center; //_special_align_left_center
			styles[3].special = styles[2].special = _special_align_justify;
			styles[2].offs = 14;
			styles[3].offs = 26;
			styles[3].size = 12;
			*styles = styles[1];
			styles[4] = styles[3];
			styles[4].offs = 28;
			//styles[4].firstIndent = 28<<6;
			styles[3].special|= _special_subscript|_special_superscript;
			styles->size = 12;
			styles->style = _style_bold;
			//styles->style = 1;
			styles[1].offs = 2;
		} else if (ftype>=_ft_csv)
		{
			if (rtctl)
				rtctl->value = 1;
			if (wwctl)
				wwctl->value = 0;
		}
	tbOpenFile[tbFront].fType = ftype;
	tbox->flags&=~tBoxWasModified;

	/* add a softened frame (uses blend transfer mode) */
	drawBoxerTBFrame();
	/* have the text box own the buffer (tBoxUpdate is necessary) */
	if (doPassword)
		tbox->flags|= tBoxHideCharacters;
	if (rtctl)
		tbox->tabAlign = rtctl->value ? tbOpenFile[tbFront].tabSize:0;
	tbox->maxWidth = tbOpenFile[tbFront].maxWidth; //33500u
	tbox->ctlScrollV->visible = 1;
	if (wwctl && wwctl->value)
		tbox->maxWidth = RECTANGLE_WIDTH(&tbox->txbounds)-2<<16;
	else
		tbox->ctlScrollH->visible = 1;

	//tbox->custBgColor = system_colors[_system_gray87];
	//tbox->drawLine = _custTBStripeBG;
	tbox->drawLine = drawLineHook;
	tBoxUpdate(tbox,NULL,1);

	fbounds.left = fbounds.top = 0;
	fbounds.right = fbounds.bottom = 16000;
	wValidRect(&screen_desktop_window,&fbounds);

	fbounds.right = fbounds.bottom = 0;
	if (fbctl)
		fbox = fbctl->data, fbounds = fbctl->bounds;
	drawBoxerBorder(&fbounds,-1);
	wDrawControls(&screen_desktop_window);

	if (busytimer)
		{RemoveTimer(global_timer_thread,busytimer); busytimer = NULL;}
	set_mouse_cursor(_curs_arrow,0);
	update_screen_rect(0,0,screen_desktop_window.port.map->w,
		screen_desktop_window.port.map->h);

	screen_desktop_window.ctlFocus = tbctl;
//	screen_desktop_window.ctlDefault = 
//	screen_desktop_window.ctlCancel = screen_desktop_window.ctlList;
	r =-1;
	do
	{
		eventRecord	event;
		ctlObj		*ctl = NULL;
		uint32		oldflags = tbox->flags;
		uint16		key = 0;

		event.modifiers = 0;
		if (WaitNextEvent(everyEvent,&event,0,NULL))
		{
			r = boxerUpdateEvt(event);
			if (fbctl)
				fbounds = fbctl->bounds;
			if (r!=0)
				continue;
			key = (uint16)event.message;
			if (fbsel)
				tBoxSelect(fbox,0,99);
			fbsel = 0;
			if (event.what!=keyDown && event.what!=autoKey)
				goto genctl;
			if (event.modifiers&menuKey)
			{
				if ((uint8)key=='w')
				{
					//if (!(event.modifiers&shiftKey))
					//	continue;
					wCtlAnimHilite(wGetIndexedControl(&screen_desktop_window,_it_about),
						defaultBtnAnimDelay);
					break;
				}
				if (((uint8)key=='q'
						|| (uint32)key==(MKV_F4<<8|kFunctionKeyCharCode))
						&& event.modifiers&menuKey)
					{r = 2; break;}
				if ((key>>8)==MKV_RETURN)
					goto inret;
				/*else if ((uint8)key=='-' && tbox->txbounds.left>-32)
				{
					tbox->txbounds.left-= 3;
					if (tbox->txbounds.left<tbox->bounds.left)
						tbox->bounds.left = tbox->txbounds.left;
					tbox->txbounds.left+= 2;
					tBoxUpdate(tbox,NULL,0);
				} else if ((uint8)key=='=' && tbox->txbounds.left<32)
				{
					tbox->txbounds.left-= 1;
					if (tbox->txbounds.left<tbox->bounds.left)
						tbox->bounds.left = tbox->txbounds.left;
					tbox->txbounds.left+= 2;
					tBoxUpdate(tbox,NULL,0);
				}*/
				else if ((uint8)key=='i' && event.modifiers&shiftKey)
				{
					if (tbOpenFile[tbFront].fType>=0)
					{
						adjustTBFforNew();
					}
					fbMode|= _line_interpreter;
					drawBoxerBorder(&fbounds,0);
					tBoxSelect(tbox,IMAX(int),IMAX(int));
					//setTBoxFont(tbox,tbox->selStart,monaco);
					insertCmdLine(tbox,logo,1);
					logo = 0;
					wDrawControls(&screen_desktop_window);
					update_screen_rect(0,fbounds.bottom-50,
						screen_desktop_window.port.map->w,
						50);
					lastlineCount = tbox->lineCount;
				} else
					goto ckey;
			} else if ((uint8)key=='\t' || (key>>8)==MKV_F3
					|| screen_desktop_window.ctlFocus==fbctl)
			{
			ckey:	r = boxerGenKey(key,event.modifiers,event.what==autoKey);
				if (r<0)
				{
					if (r==-2)
						fbsel = 1;
					continue;
				}
				if (r-1u<99)
				{
					ctl = !r ? tbctl:wGetIndexedControl(&screen_desktop_window,r);
					r =-1;
					goto hctl;
				}
				if (r==128)
					break;
				r =-1;
			} else if (fbMode&_line_interpreter)
			{
				if ((uint8)key==kBackspaceCharCode)
				{
					if (tbox->selStart==tbox->selEnd)
					{ /* don't allow delete of last line */
						r = 1+tBoxCharLineIndex(tbox,tbox->length,NULL,NULL,1);
						if (r>0)
						{
							r = tbox->lines[r-1].offs;
							if (tBoxGetChar(tbox,r)=='>')
							{
								tbox->selCLMinPos = r+1;
								if (tbox->selStart-r+0u<2u)
									{r =-1; continue;}
							}
						}
					} /*else if ((r=MIN(tbox->selStart,tbox->selEnd))
							<tbox->selCLMinPos
							&& (li=MAX(tbox->selStart,tbox->selEnd))
							>=tbox->selCLMinPos)
					{
						tBoxDelete(tbox,r,li,1);
						insertCmdLine(tbox);
						r =-1;
						continue;
					}*/
					r =-1;
				} else if (((uint8)key==kReturnCharCode
						|| (uint8)key==kEnterCharCode
						|| (uint8)key==kEscapeCharCode)
						&& !(event.modifiers&controlKey))
				{
					const char	*s;

				inret:	s = tBoxGetText(tbox);
					li = 1+tBoxCharLineIndex(tbox,tbox->selStart,NULL,NULL,1);
					tbOpenFile[tbFront].curLine = li;
					if ((uint8)key==kEscapeCharCode)
					{
						r = li>0 ? tbox->lines[li-1].offs:0;
						tbox->selCLMinPos = 0;
						if (r<tbox->length && s && s[r]=='>')
							tBoxDelete(tbox,tbox->selStart-1,tbox->selStart,0);
						fbMode&= 15;
						fbSetNoMode();
						r =-1;
						continue;
					}
					/* return/enter */
					if (tbox->selStart==tbox->selEnd)
					{
						r = li-0u<tbox->lineCount-0u
							? tbox->lines[li].offs-2
							:tbox->length-1;
						if (s[r]=='\\' || (tbox->selStart>0
								&& s[tbox->selStart-1]=='\\'))
						{
							lastlineCount = tbox->lineCount+1;
							r =-1;
							goto genctl;
						}
					} else if (!tbox->selCLMinPos)
						tbox->selCLMinPos = r;
					executeLine(tbox,li,!!(event.modifiers&menuKey));
					if (fbctl)
						fbounds = fbctl->bounds;
					lastlineCount = tbox->lineCount;
					r =-1;
					goto redraw;
				}
			}
		}
	genctl:	while(evtRActive)
			;
		evtRActive = 1;
		ctl = wControlEvent(&screen_desktop_window,&event);
		evtRActive = 0;
		if (ctl==screen_desktop_window.ctlList)
			break;
		if (fbMode&_line_interpreter && tbox->lineCount!=lastlineCount)
		{ /* screw it */
			tbox->selCLMinPos = 0;
			lastlineCount = tbox->lineCount;
			if (tbox->selStart==tbox->selEnd && key<' ')
			{
				li =-1;
				1+tBoxCharLineIndex(tbox,tbox->selStart,&li,NULL,1);
				if (tBoxGetChar(tbox,li)!='>')
					insertCmdLine(tbox,0,1);
				else
					tbox->selCLMinPos = li+1;
			}
		}
	hctl:	boxerClickCtl(ctl,&fbounds,event.modifiers);
		if ((oldflags^tbox->flags)&tBoxWasModified && !(oldflags&tBoxWasModified))
		{
		redraw:	drawBoxerBorder(&fbounds,1);
			wDrawControls(&screen_desktop_window);
			update_screen_rect(0,fbounds.top-50,
				screen_desktop_window.port.map->w,
				fbounds.bottom-fbounds.top+50);
		}
		if (fbMode&_line_interpreter && key)
		{
			li = 1+tBoxCharLineIndex(tbox,tbox->length,NULL,NULL,1);
			if (li>0 && tBoxGetChar(tbox,tbox->lines[li-1].offs)=='>')
				tbox->selCLMinPos = tbox->lines[li-1].offs+1;
		}
	} while(r<=0);
	screen_desktop_window.port.pnSize = oldpens;
	RGBFore(&screen_desktop_window.port,&oldfore);
	RGBBack(&screen_desktop_window.port,&oldback);
	rFillRect(&screen_desktop_window.port,0,tbox->bounds.top-5,
		MAXSCRW,MAXSCRH);
	update_screen_rect(0,0,screen_desktop_window.port.map->w,
		screen_desktop_window.port.map->h);
	set_mouse_cursor(_curs_arrow,0);

end:	if (tbctl)
		wDisposeControl(tbctl);
	*bounds = tbOpenFile[tbFront].bounds;
	tbOpenFile[tbFront].tbctl = NULL;
	memset(tbOpenFile[tbFront].path,0,sizeof(tbOpenFile[tbFront].path));
	--tbCount;
	if (rtctl)
		rtctl->visible = 0;
	if (f)
		fclose(f);
	fbSetNoMode();
	wShowControl(wGetIndexedControl(&screen_desktop_window,_it_dither));
	wShowControl(wGetIndexedControl(&screen_desktop_window,_it_smoothing));
	if (busytimer)
		{RemoveTimer(global_timer_thread,busytimer); busytimer = NULL;}
	set_mouse_cursor(_curs_arrow,0);
	return r;
}

#if DISPLAY_STATS
static void displayStats(int x,
	int y)
{
	static boolean	init;

	static char	*modenames[] =
	{
		"cpy ","or   ","xor  ","bic  ",
		"ncpy","nor  ","nxor","nbic",
		"tran","blnd","admx","admn",
		"adpn","adov","sbpn","sbov"
	};
	char		b[sizeof(mKeyMap)*2+80], colorstr[80];
	pt2d16		where;
	uint64		mcs;
	int32		z, vblc = TickCount();

	get_mouse_loc(&where.x,&where.y);
	hexToStr(b,mKeyMap,sizeof(mKeyMap));

	colorstr[0] = 0;
	if (nrcti)
	{
		const rect16	*b = nrctp;
		int		c = nrctc, i, inst;

		for(i=inst=0; i<c; ++i,++b)
			if (where.y>=b->top && where.x>=b->left && where.y<b->bottom && where.x<b->right)
			{
				sprintf(colorstr+strlen(colorstr),inst ? ",#%d":"#%d",
					i);
				if (++inst>4)
					goto elip;
			}
		if (!inst)
			elip: strcat(colorstr,"..");
	} else
	{
		rgb48		color;

		GetPixelRGB(screen_desktop_window.port.map,
			where.x-screen_desktop_window.port.origin.x,
			where.y-screen_desktop_window.port.origin.y,&color);
		sprintf(colorstr,"%03X,%03X,%03X",
			color.red>>4,color.green>>4,color.blue>>4);
	}

	if (init)
	{
		init = 0;
		return;
	}
	if (global_timer_thread)
	{
		Microseconds(&mcs);
		z = sprintf(b+sizeof(mKeyMap)*2," %.2fs - %.3f ms; %s; %d,%d=%s",
			vblc*.016625,(mcs-timer_start)/(double)vblc/1000.,
			modenames[screen_desktop_window.port.pnMode&15],
			where.x,where.y,colorstr);
	} else
		z = sprintf(b+sizeof(mKeyMap)*2," %s; %d,%d=%s",
			modenames[screen_desktop_window.port.pnMode&15],
			where.x,where.y,colorstr);
	screen_desktop_window.port.txMode = mSrcCopy;
	rTextFace(&screen_desktop_window.port,0,0,0);
	x = rDrawText(&screen_desktop_window.port,b,z+sizeof(mKeyMap)*2,y,x<<16|0x8000);
	screen_desktop_window.port.chExtra = screen_desktop_window.port.map->w<<16;
	rDrawText(&screen_desktop_window.port," ",1,y,x);
	screen_desktop_window.port.chExtra = 0;
}
#endif

static void testGraphics(void)
{
	SLRgn		*brgn = NULL;

	/* broken mrgn code will trip one or more of these errors, or crash */
	if (rectRgn(&brgn,IMIN(int16),IMIN(int16),IMAX(int16),IMAX(int16),1)<0)
		printf("testGraphics: failed to allocate max. filled rect region\n");
	else if (!brgn)
		printf("testGraphics: generated max. rect region is NULL!\n");
	else
	{
		if (brgn->size!=0x40008)
			printf("testGraphics: generated max. rect region has incorrect size\n");
		brgn = rgnCompact(brgn,0);
		if (!brgn)
			printf("testGraphics: rgnCompact(max. rect region) now empty!\n");
		else if (brgn->size!=12)
			printf("testGraphics: rgnCompact(max. rect region) not implied rect\n");
	}
	if (rectRgn(&brgn,IMAX(int16),IMIN(int16),IMAX(int16),IMAX(int16),1)<0)
		printf("testGraphics: failed to allocate max. vert region\n");
	else if (!brgn)
		printf("testGraphics: generated max. vert region is NULL!\n");
	else
	{
		if (brgn->size!=0x2000A)
			printf("testGraphics: generated max. v-rect region has incorrect size\n");
		brgn = rgnCompact(brgn,0);
		if (brgn)
			printf("testGraphics: rgnCompact(max. vert region) not empty\n");
	}
	disposeRgn(&brgn);
	if (brgn)
		printf("testGraphics: disposeRgn failed!\n");
}

static char	workingDir[260];

static void GetWorkingDirectory(void)
{
	const char	*sep, *sepb;
	int		i;

	GetModuleFileName(NULL,workingDir,sizeof(workingDir)-1);

	sep = strrchr(workingDir,'\\');
	sepb = strrchr(workingDir,'/');
	if (!sep || (sepb && sepb<sep))
		sep = sepb;
	if (sep && (i=sep-workingDir+1)>=0)
		workingDir[i] = 0;
}

/* main demo */

uint8		weavePat[8] = {0xF8,0x74,0x22,0x47,0x8F,0x17,0x22,0x71};
static int8	txsizes[5] = {9, 12, 14, 24, 48};
static char	txs[] = "Run, rabbit, run! \xD9";
#ifdef TEST_FILTER
cTab64		*filterTable;
#endif
static pt2dq	tr[4], ln[2];
static pt2d16	txPos, pxz;
static pt2d16	pxt, pxd, pxmdv;
static boolean	pxmdrag;
static rgb48	shapeColor, lineColor;
static SLRgn	*pictClipRgn;
static int	txhw;
static boolean	wpat, wpatt;
static boolean	pmode = 0, pmodet = 0;
static boolean	animate = 1, animatet = 0, arcChord;
static boolean	incrt = 0, decrt = 0;
static int8	txstyl = 0, txsize = 1, txfont = 3, txgray = 0, txspex = 0;
static uint32	ltick;

static char	txtPath[512], stylesPath[512];
const char	*picPath = NULL;

static int __cdecl iterAltFW(const uint8 *name,
	int ID)
{
	if (!memcasecmp((const char*)name,"\7ProFont",8))
		ffFWDefault = ID;
	return 0;
}

static int __cdecl iterOptFont(const uint8 *name,
	int ID)
{
	if (ffOptFont<0 && ID>0)
		ffOptFont = ID;
	return 0;
}

static void paintRgnOp(boolean sect,
	boolean erase)
{
	SLRgn	br, *rgn = newRgn(0), *rrgn = NULL;

	rgn->bBox.top = ov.top>>16;
	rgn->bBox.left = ov.left>>16;
	rgn->bBox.bottom = ov.bottom>>16;
	rgn->bBox.right = ov.right>>16;
	br.size = RECT_RGN_SIZE;
	br.bBox.top = (ov.top+0x88000+(ln->y>>7))>>16;
	br.bBox.left = (ov.left+0x88000+(ln->x>>7))>>16;
	br.bBox.bottom = (ov.bottom+0x48000+(ln->y>>7))>>16;
	br.bBox.right = (ov.right+0x48000+(ln->x>>7))>>16;
	xorRgn(&rgn,&br);
	br.bBox.top = (ov.top+0x88000)>>16;
	br.bBox.left = (ov.left+0x88000)>>16; //+(startAngle&0x1FFFFF))>>16
	br.bBox.bottom = (ov.bottom+0x48000)>>16;
	br.bBox.right = (ov.right+0x48000)>>16;
	xorRgn(&rgn,&br);
	roundRectRgn(&rrgn,&rr,16<<16,16<<16);
	if (sect)
		sectRgn(&rgn,rrgn);
	else
		unionRgn(&rgn,rrgn);
	disposeRgn(&rrgn);
	screen_desktop_window.port.fgSolid = RGBToPen(screen_desktop_window.port.map,
		erase ? &rgb_black:&rgb_tan);
	rRegion(&screen_desktop_window.port,rgn);
	RGBFore(&screen_desktop_window.port,&screen_desktop_window.port.fgColor);
	disposeRgn(&rgn);
}

static void frameRects(const rect16 *b,
	int c,
	boolean erase)
{
	grafPort	*p = &screen_desktop_window.port;
	rect2dq		rc;
	penState	oldstate;
	uint32		oldfg;

	if (c<=0)
		return;
	oldfg = p->fgSolid;
	rPenNormal(p,&oldstate);
	p->fgSolid = p->map->bpp>8 ? 0:~0;
	if (!erase)
		p->pnMode = mSrcXor;
	do
	{
		FixRect(&rc,b->left,b->top,b->right,b->bottom);
		rRect(p,&rc);
	} while(++b,--c);
	p->fgSolid = oldfg;
	rRestorePen(p,&oldstate,1);
}

static void offsetRects(rect16 *b,
	int c,
	int x,
	int y)
{
	for(; c>0; --c,++b)
	{
		b->top+= y;
		b->left+= x;
		b->bottom+= y;
		b->right+= x;
	}
}

static void update_window(boolean updateNow)
{
	static boolean	hasDrawn, lastrgnop;
	ctlObj		*dctl, *sctl;
	uint32		tick;
	int32		c;
	uint16		mode;
	int16		mousex, mousey;
	boolean		shift, menudown = 0;

	tick = TickCount(); //GetTickCount()>>4;
	if (tick==ltick && !updateNow)
		return;
	//EnterCriticalSection(&drawMutex);
	/* update for both left and right modifier keys */
	MGetKeys(mKeyMap);
	get_mouse_loc(&mousex,&mousey);
	shift = _MKDOWN(mKeyMap,MKV_LSHIFT)|_MKDOWN(mKeyMap,MKV_RSHIFT);
	if (hasDrawn)
	{
		screen_desktop_window.port.pnMode
			= screen_desktop_window.port.txMode = mSrcCopy;
		//if (_MKDOWN(mKeyMap,MKV_COMMA))
		paintRgnOp(lastrgnop,1);
		RGBFore(&screen_desktop_window.port,&rgb_black);
		RGBBack(&screen_desktop_window.port,&rgb_black);
		ovalArc(&screen_desktop_window.port,&ov,arcChord);
		rPolygon(&screen_desktop_window.port,tr,4);
		rLine(&screen_desktop_window.port,*ln,ln[1]);
		rTriangle(&screen_desktop_window.port,tr[2],tr[3],*tr);
		rRoundRect(&screen_desktop_window.port,&rr,16<<16,16<<16);

		if (pic)
			rFillRect(&screen_desktop_window.port,pxt.x,pxt.y,
				pxt.x+pxd.x,pxt.y+pxd.y);
		frameRects(nrctp,nrctc,1);
		rTextFace(&screen_desktop_window.port,txfont&31,txsizes[txsize&7],txstyl);
		rDrawText(&screen_desktop_window.port,txs,-1,txPos.y,(txPos.x-txhw)<<16|0x8000);
		RGBBack(&screen_desktop_window.port,&rgb_white);
		//hasDrawn = 0; - will render anyway
	}
	if (animate&1)
	{
		tr->x+= (int16)rand()>>1;
		tr->y+= (int16)rand()>>1;
		tr[1].x+= (int16)rand()>>1;
		tr[1].y+= (int16)rand()>>1;
		tr[2].x+= (int16)rand()>>1;
		tr[2].y+= (int16)rand()>>1;
		ln->x+= (int16)rand()>>1;
		ln->y+= (int16)rand()>>1;
		ln[1].x+= (int16)rand()>>1;
		ln[1].y+= (int16)rand()>>1;
		ov.left+= (int16)rand()>>1;
		ov.top+= (int16)rand()>>1;
		ov.right+= (int16)rand()>>1;
		ov.bottom+= (int16)rand()>>1;
		rr.left+= (int16)rand()>>1;
		rr.top+= (int16)rand()>>1;
		rr.right+= (int16)rand()>>1;
		rr.bottom+= (int16)rand()>>1;
	}

	dctl = wGetIndexedControl(&screen_desktop_window,_it_dither);
	sctl = wGetIndexedControl(&screen_desktop_window,_it_smoothing);

	//if (!(event.modifiers&menuKey))
	if (!(_MKDOWN(mKeyMap,MKV_LMENU)|_MKDOWN(mKeyMap,MKV_RMENU)))
	{
		if (pmodet!=_MKDOWN(mKeyMap,MKV_I) && !pmodet)
			pmode = (pmode+1)&15;
		pmodet = _MKDOWN(mKeyMap,MKV_I);
		if (wpatt!=_MKDOWN(mKeyMap,MKV_H) && !wpatt)
			wpat^= 1;
		wpatt = _MKDOWN(mKeyMap,MKV_H);
		if (animatet!=_MKDOWN(mKeyMap,MKV_SPACE) && !animatet)
			animate^= 1;
		animatet = _MKDOWN(mKeyMap,MKV_SPACE);
		if (decrt!=_MKDOWN(mKeyMap,MKV_HYPHEN_USC) && !decrt) /* - key to decrease pensize */
			screen_desktop_window.port.pnSize.x
				= screen_desktop_window.port.pnSize.y
				-= screen_desktop_window.port.pnSize.y<0 ? 0:0x4000;
		decrt = _MKDOWN(mKeyMap,MKV_HYPHEN_USC);
		if (incrt!=_MKDOWN(mKeyMap,MKV_EQUALPLUS) && !incrt) /* = key to decrease pensize */
			screen_desktop_window.port.pnSize.x
				= screen_desktop_window.port.pnSize.y+= 0x4000;
		incrt = _MKDOWN(mKeyMap,MKV_EQUALPLUS);
		shapeColor.red = 65535;
		if (_MKDOWN(mKeyMap,MKV_L)) /* L to set the intensity (by v) */
			shapeColor.red = mousey*140;
		shapeColor.blue = shapeColor.red;
		shapeColor.green = MIN(mousex,510)*257>>1;

		/* A, S, D, and F for the quadrilateral */
		if (_MKDOWN(mKeyMap,MKV_A))
			tr[0].x = mousex<<16|0x8000, tr[0].y = mousey<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_S))
			tr[1].x = mousex<<16|0x8000, tr[1].y = mousey<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_D))
			tr[2].x = mousex<<16|0x8000, tr[2].y = mousey<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_F))
			tr[3].x = mousex<<16|0x8000, tr[3].y = mousey<<16|0x8000;
		/* Z and X for the line coords */
		if (_MKDOWN(mKeyMap,MKV_Z))
			ln->x = mousex<<16|0x8000, ln->y = mousey<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_X))
			ln[1].x = mousex<<16|0x8000, ln[1].y = mousey<<16|0x8000;
		/* C and V for the oval coords */
		if (_MKDOWN(mKeyMap,MKV_C))
			rr.left = mousex<<16|0x8000, rr.top = mousey<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_V))
			rr.right = mousex<<16|0x8000, rr.bottom = mousey<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_P))
		{
			pxmdv = pxt;
			mousex-= pxd.x>>1, mousey-= pxd.y>>1;
			//pxt.x = mousex-(pxd.x>>1), pxt.y = mousey-(pxd.y>>1);
			goto drag;
		}
		if (_MKDOWN(mKeyMap,MKV_SLBRACKET))
		{
			pt2d16		pxu;

			pxd.x+= pxt.x, pxd.y+= pxt.y;
			pxu = pxd;
			pxt.x = mousex, pxt.y = mousey;
			pxd.x-= pxt.x, pxd.y-= pxt.y;
			/* proportional scaling; 2016.10.31 */
			if (pic && shift
					&& pxd.x>0 && pxd.y>0)
				if (pxd.x<pxd.y)
				{
					pxd.y = (pxd.x*pic->h+(pic->w>>1))/pic->w;
					pxt.y = pxu.y-pxd.y;
				} else
				{
					pxd.x = (pxd.y*pic->w+(pic->h>>1))/pic->h;
					pxt.x = pxu.x-pxd.x;
				}
		}
		if (_MKDOWN(mKeyMap,MKV_SRBRACKET))
		{
			pxd.x = mousex-pxt.x, pxd.y = mousey-pxt.y;
			/* proportional scaling; 2016.10.31, 2022 */
			if (pic && shift
					&& pxd.x>0 && pxd.y>0)
			{
				double		aspect = sqrt((double)pic->w/pic->h);
				double		diag = hypot(pxd.x,pxd.y)*.707;

				pxd.x = (int)(diag*aspect+.5);
				pxd.y = (int)(diag/aspect+.5);
			}
		} else if (_MKDOWN(mKeyMap,MKV_BKSLASH) && pic)
		{
			pxd.x = pic->w, pxd.y = pic->h;
			if (_MKDOWN(mKeyMap,MKV_LCONTROL)|_MKDOWN(mKeyMap,MKV_RCONTROL))
				if (shift)
					pxd.x>>= 2, pxd.y>>= 2;
				else
					pxd.x>>= 1, pxd.y>>= 1;
			else if (shift)
				pxd.x+= pxd.x, pxd.y+= pxd.y;
		} else if (pxmdrag)
		{
			int		x, y;

		drag:	x = mousex-pxmdv.x, y = mousey-pxmdv.y;
			pxmdv.x = mousex, pxmdv.y = mousey;
			pxt.x+= x, pxt.y+= y;
			offsetRects(nrctp,nrctc,x,y);
			if (!_MKDOWN(mKeyMap,MKV_P))
				nrcti = 1;
		}// else 
		//	nrcti = 0;
		if (_MKDOWN(mKeyMap,MKV_W))
			lineColor = rgb_white;
		if (_MKDOWN(mKeyMap,MKV_Y))
			lineColor = chromo_colors[_rgb_yellow];
		if (_MKDOWN(mKeyMap,MKV_M))
			lineColor = chromo_colors[_rgb_magenta];
		if (_MKDOWN(mKeyMap,MKV_R))
			lineColor = chromo_colors[_rgb_red];
		if (_MKDOWN(mKeyMap,MKV_N))
			lineColor = chromo_colors[_rgb_cyan];
		if (_MKDOWN(mKeyMap,MKV_G))
			lineColor = chromo_colors[_rgb_green];
		if (_MKDOWN(mKeyMap,MKV_B))
			lineColor = chromo_colors[_rgb_blue];
		if (_MKDOWN(mKeyMap,MKV_K))
			lineColor = rgb_black;
		if (_MKDOWN(mKeyMap,MKV_O))
			lineColor = rgb_orange;
		if (_MKDOWN(mKeyMap,MKV_T))
			lineColor = rgb_tan;
		if (_MKDOWN(mKeyMap,MKV_1))
			txPos.x = mousex, txPos.y = mousey;
		if (_MKDOWN(mKeyMap,MKV_2))
		{ /* cycle classic styles */
			if (!(txstyl&0x80))
				txstyl = ((txstyl+1)&15)|(txstyl
					&(_style_strikethrough|_style_overline));
			txstyl|= 0x80;
		} else
			txstyl&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_3))
		{ /* cycle sizes */
			if (!(txsize&0x80))
				txsize = (txsize+1)%GET_ARRAY_COUNT(txsizes);
			txsize|= 0x80;
		} else
			txsize&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_4))
		{ /* cycle fonts */
			if (!(txfont&0x80))
			{
				c = 31;
				do
					++txfont, txfont&= 31;
				while(!fmFonts[txfont] && --c);
			}
			txfont|= 0x80;
		} else
			txfont&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_5))
		{ /* toggle graying */
			if (!(txgray&0x80))
				txgray^= mTextGray;
			txgray|= 0x80;
		} else
			txgray&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_6))
		{ /* cycle spacing */
			if (!(txspex&0x80) && (screen_desktop_window.port.spExtra
					+=65536/1.5+.5)>19*65536)
				screen_desktop_window.port.spExtra = 0;
			txspex|= 0x80;
		} else
			txspex&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_Q) && shift)
		{ /* move oval */
			c = ov.right-ov.left;
			ov.left = (mousex<<16)-(c>>1), ov.right = (mousex<<16)+((c+1)>>1);
			c = ov.bottom-ov.top;
			ov.top = (mousey<<16)-(c>>1), ov.bottom = (mousey<<16)+((c+1)>>1);
		}
		startAngle+= 1<<16;
		if (_MKDOWN(mKeyMap,MKV_J))
			arcAngSpan-= degToU32Angle(_MKDOWN(mKeyMap,MKV_LSHIFT) ? 3:0.2);
		arcChord = _MKDOWN(mKeyMap,MKV_U); /* TODO: fix small angle bug */
		if (!(decorActive&2))
		{
			if (dctl)
				if (_MKDOWN(mKeyMap,MKV_0)|_MKDOWN(mKeyMap,MKV_SEMICOLON))
				{
					if (!dctl->hilite)
						dctl->value = !dctl->value
							*(!_MKDOWN(mKeyMap,MKV_SEMICOLON)+1);
					dctl->hilite = 255;
				} else
					dctl->hilite = 0;
			if (sctl)
				if (_MKDOWN(mKeyMap,MKV_SQUOTE))
				{
					if (!sctl->hilite)
						sctl->value = !sctl->value;
					sctl->hilite = 255;
				} else
					sctl->hilite = 0;
		}
		if (_MKDOWN(mKeyMap,MKV_COMMA))
			paintRgnOp(lastrgnop=1,0);
		else if (_MKDOWN(mKeyMap,MKV_PERIOD))
			paintRgnOp(lastrgnop=0,0);
	} else
	{
		menudown = 1;
		if (shift && _MKDOWN(mKeyMap,MKV_OPT_APPS) && _MKDOWN(mKeyMap,MKV_U))
		{
			__try{__debugbreak();}
			__except(EXCEPTION_EXECUTE_HANDLER)
			{ /* we got here if we did not enter debugger */
			}
		} else if (_MKDOWN(mKeyMap,MKV_C))
		{
			rgbPickColor(-1,-1,(const uint8*)"\012Line Color",&lineColor,&lineColor);
			RGBFore(&screen_desktop_window.port,&rgb_black);
			rFillRect(&screen_desktop_window.port,
				screen_desktop_window.updateRect.left+screen_desktop_window.port.origin.x,
				screen_desktop_window.updateRect.top+screen_desktop_window.port.origin.y,
				screen_desktop_window.updateRect.right+screen_desktop_window.port.origin.x,
				screen_desktop_window.updateRect.bottom+screen_desktop_window.port.origin.y);
			screen_desktop_window.updateRect.bottom = screen_desktop_window.updateRect.right
				= 32767;
		} else if (_MKDOWN(mKeyMap,MKV_I))
			pmode = 0;
	}
	if (screen_desktop_window.port.map->bpp==1
			&& formGray(shapeColor.red,shapeColor.green,shapeColor.blue,c)
			-65536u<24576u)
	{
		RGBFore(&screen_desktop_window.port,&rgb_black);
		rPenPat(&screen_desktop_window.port,grayPat);
	} else
		RGBFore(&screen_desktop_window.port,&shapeColor);
	screen_desktop_window.port.pnMode = screen_desktop_window.port.txMode = pmode|txgray;
	ovalArc(&screen_desktop_window.port,&ov,arcChord);
	rPenPat(&screen_desktop_window.port,NULL);

	RGBFore(&screen_desktop_window.port,&shapeColor);

	c = rPolygon(&screen_desktop_window.port,tr,4);
#if 1
	if (c<0)
		vhalt((sprintf(temporary,"### rPolygon returned %d",
			c),temporary));
#endif
	RGBFore(&screen_desktop_window.port,&lineColor);
	rLine(&screen_desktop_window.port,*ln,ln[1]);

	screen_desktop_window.port.opColor.red = 65535*3/9;
	screen_desktop_window.port.opColor.green = 65535*4/9;
	screen_desktop_window.port.opColor.blue = 65535*5/9;
	RGBFore(&screen_desktop_window.port,system_colors+_system_gray87);
	rRoundRect(&screen_desktop_window.port,&rr,16<<16,16<<16);
	rPenPat(&screen_desktop_window.port,wpat ? weavePat:grayPat);
	RGBFore(&screen_desktop_window.port,screen_desktop_window.port.map->bpp==1
		? &rgb_black:system_colors+_system_gray73);
	rTriangle(&screen_desktop_window.port,tr[2],tr[3],*tr);
	rPenPat(&screen_desktop_window.port,NULL);

	hasDrawn = 1;
	RGBFore(&screen_desktop_window.port,pmode==(mSrcCopy|mTransparent)
		? system_colors+_system_gray15:&rgb_black);

	mode = pmode|mMergeDown;
	if (dctl && dctl->value)
		mode|= mDither;
	if (sctl && sctl->value)
	{
		mode|= mMergeUp|mMergeDown;
		if (pic && pic->pSize<32 && screen_desktop_window.port.map->bpp>16)
				//&& menudown)
			pmConvert(pic,32,pic->w,pic->h,NULL,NULL);
	}
	rClippedBltPix(&screen_desktop_window.port,pic,0,0,pxz.x,pxz.y,
		pxt.x,pxt.y,pxd.x,pxd.y,mode,pictClipRgn);
	frameRects(nrctp,nrctc,0);
	rTextFace(&screen_desktop_window.port,txfont&31,txsizes[txsize&7],txstyl);
	//sprintf(txs,"%08X",
	//	(pxz.y*65535+pxd.y)/pxd.y);
	rDrawText(&screen_desktop_window.port,txs,-1,txPos.y,(txPos.x-txhw)<<16|0x8000);
#if DISPLAY_STATS
	RGBFore(&screen_desktop_window.port,&rgb_black);
	screen_desktop_window.port.spExtra>>= 1;
	displayStats(0,staty);
	screen_desktop_window.port.spExtra<<= 1;
#endif
#ifdef TEST_FILTER
	pmFilter(screen_desktop_window.port.map,filterTable);
#endif
#ifdef TEST_REVERSEBLT
	rBltPix(&screen_desktop_window.port,screen_desktop_window.port.map,
		0,0,128,128,1,4,128,128,mSrcCopy);
#endif
	//if (!(decorActive&2) || tick&1)
	wDrawControls(&screen_desktop_window);
	update_screen_rect(0,0,screen_desktop_window.port.map->w,
		screen_desktop_window.port.map->h);

	if (animate&1 && picGIFState && ltick)
	{ /* animate GIF */
		picGIFTime+= (tick-ltick)*(SMALL_TICK_TIME/1e6*100);
		c = picGIFState->info.delayTime ? picGIFState->info.delayTime:10;
		if (c>0 && picGIFTime>=c)
		{
			int32	loop = picGIFState->info.loopCount;

			if (picGIFState->info.disposal==_gifdisp_erase)
				pmErase(pic,pic->pSize<=8 ? picGIFState->screen.Background*0x1010101u
					:picGIFState->screen.ColorMap[picGIFState->screen.Background]);
			//else if (picGIFState->info.disposal==_gifdisp_erase)
			//	pmCopy(picPrevFrame,&pic);
			picGIFTime-= c;
			decompressGIF(&pic,srd_fread,picGIFFile,-(++picGIFFrame),&picGIFState);
			while(picGIFState && picGIFState->info.delayTime<0
					&& (c=decompressGIF(&pic,srd_fread,picGIFFile,
					-(++picGIFFrame),&picGIFState))>=0)
				;
			if (!picGIFState && loop!=1)
			{ /* loop */
				pmCopy(picInitFrame,&pic);
				picGIFFrame = 1;
				/* state refresh doesn't work! */
				//if ((picGIFState=malloc(sizeof(picGIFInitState)))!=NULL)
				//	*picGIFState = picGIFInitState;
				//fseek(picGIFFile,picGIFInitOffs,SEEK_SET);
				rewind(picGIFFile);
				decompressGIF(NULL,srd_fread,picGIFFile,1,&picGIFState);
				if (picGIFState && loop>0)
					picGIFState->info.loopCount = loop-1;
			}
		}
	}
	//LeaveCriticalSection(&drawMutex);
	ltick = tick;
}

static void createMainButtons(const uint8 *textTitle)
{
	rect16		b, rtb;
	int		h = screen_desktop_window.port.map->h;

	b.top = h-41, b.left = 12;
	b.bottom = h-19, b.right = 100;
	wNewControl(&screen_desktop_window,&b,textTitle,1,0,0,1,ctlTypePushBtn,0);
	b.top = h-38, b.left = 112;
	b.bottom = h-22, b.right = 175;
	wNewControl(&screen_desktop_window,&b,(const uint8*)"\6Dither",1,0,0,2,
		ctlTypeCheckBox,(void*)'c');
	b.left = 184, b.right = 276;
	wNewControl(&screen_desktop_window,&b,(const uint8*)"\11Smoothing",1,0,0,2,
		ctlTypeCheckBox,(void*)'s');
	//b.left = 205;
	b.right = 270;
	rtb = b;
	wNewControl(&screen_desktop_window,&b,(const uint8*)"\11Real Tabs",0,0,0,1,
		ctlTypeCheckBox,(void*)'t');
	b.left = 115, b.right = 175;
	wNewControl(&screen_desktop_window,&b,(const uint8*)"\4Wrap",0,0,0,1,
		ctlTypeCheckBox,(void*)'w');

	b.top = rtb.bottom+7+1;
	b.left = 431;
	b.bottom = b.top+16;
	b.right = b.left+99;
	wNewControl(&screen_desktop_window,&b,
		(const uint8*)"\12Match Case",0,0,0,1,
		ctlTypeCheckBox,(void*)'m');
	b.left = b.right+7;
	b.right = b.left+101;
	wNewControl(&screen_desktop_window,&b,
		(const uint8*)"\12Whole Word",0,0,0,1,
		ctlTypeCheckBox,(void*)'h');

	b.top = rtb.bottom+7;
	b.left = rtb.left-98;
	b.right = b.left+defaultFBoxWidth;
	b.bottom = b.top+16; //+15 to test v-scroll oscillation
	wNewControl(&screen_desktop_window,&b,0,0,
		(tBoxDragDrop|tBoxInhibitStyles|tBoxSetOnDrag)<<16,0,0,
		ctlTypeEditText,0);
}

static dFileType checkText(const char *path,
	dFileType r)
{
	char		b[100], *n;
	FILE		*f = fopen(path,"r");
	long		z;

	if (!f)
		return -43;
	z = fread(b,1,sizeof(b),f);
	if (z && !memchr(b,0,z) && !memchr(b,1,z) && !memchr(b,127,z)
			&& ((n=memchrbit(b,0x80,z))==NULL
			 || (*n==0xA9 || *n==0xC2 || memchr(b,'\n',n-b) || memchr(b,' ',n-b))))
		r = memchr(b,'\n',z) ? _ft_text:_ft_ascii;
	fclose(f);
	return r;
}

static int loadInputFile(const char *picPath)
{
	int		r = 0;

	if (pic)
		{pmFree(pic); pic = NULL;}
	if (picInitFrame)
		{pmFree(picInitFrame); picInitFrame = NULL;}
	if (picGIFFile)
		{fclose(picGIFFile); picGIFFile = NULL;}
	if (picGIFState)
		{free(picGIFState); picGIFState = NULL;}
	imgError = imgDepth = 0;
	if (picGIFFrame>=0)
	{
		picGIFFile = fopen(picPath,"rb");
		r = decompressGIF(&pic,srd_fread,picGIFFile,++picGIFFrame,&picGIFState);
		if (r<0)
			goto picerr;
		if (picGIFState)
		{
			/* preroll non-animated pixels */
			while(picGIFState && picGIFState->info.delayTime<0
					&& (r=decompressGIF(&pic,srd_fread,picGIFFile,
					-(++picGIFFrame),&picGIFState))>=0)
				;
			if (picGIFState)
			{
				if (picGIFState->info.delayTime<0)
					{free(picGIFState); picGIFState = NULL;}
				else if (pic->pSize!=32)
				{ /* some local transitions don't do well in 8-bit */
					pmConvert(pic,32,pic->w,pic->h,NULL,NULL);
				}
			}
			//picGIFInitState = *picGIFState;
			pmCopy(pic,&picInitFrame);
			//picGIFInitOffs = ftell(picGIFFile);
		}
	} else if ((pic=decompressImageFile(picPath && *picPath ? picPath:temporary,&r,
			NULL,NULL,&picMode))==NULL)
	{
	picerr:	r = checkText(picPath,r);
		if (r<=0 && picPath)
		{
			if (isFileEmpty(picPath)>0)
				r =-39;
			sprintf((char*)_dbgTemp+512,"## couldn't load file \"%s\" (%d)%s\n",
				picPath,r,!r ? " - data too short or empty bounds"
				:r==-2 ? " - out of memory"
				:r==-3 ? " - read error"
				:r==-39 ? " - empty file"
				:r==-8 ? " - unsupported format":"");
			imgError = (char)r;
			if (savedftype+0u>=_ft_txt)
				tbOpenFile[tbFront].fType = savedftype;
			pbufrecog = 0;
		}
	#if HSV_WHEEL_TEST
		pic = hsvMakeWheelPixMap(&screen_desktop_window.port.bkColor,65535,0,1,1);
	#endif
	} else if (picPath==pbuf)
	{
		*stylesPath = 0;
		strcpy_s(txtPath,sizeof(txtPath)-1,workingDir);
		strcat_s(txtPath,sizeof(txtPath)-1,"README.txt");
	}
	if (pic)
	{
		int		t = pic->minCmp*3+!!(pic->pType&pTypeLGreen);

		imgBPP = pic->bpp;
		imgDepth = MIN(t,imgBPP);
		imgHasAlpha = !!(pic->pType&pTypeAlpha);
		pxz.x = pxd.x = pic->w, pxz.y = pxd.y = pic->h;
	}
	return r;
}

static char *processFile(const char *s,
	int r,
	dFileType *ftype)
{
	uint8		buf[16];
	const char	*n = s+r;
	char		*path = NULL;
	dFileType	prtype = *ftype;

	memcpy(pbuf,s,r);
	pbuf[r] = 0;
	pbufrecog = 0;
	for(; r && pbuf[r]!='\\' && pbuf[r]!='/'; --r)
		if ((pbuf[r]=tolower(pbuf[r]))=='.')
			break;
	if ((*ftype=findFileType(pbuf,n-s,r && pbuf[r]=='.'))>_ft_srf)
	{ /* got a text file...? */
		const char	*u;

		strcpy_s(txtPath,sizeof(txtPath)-1,pbuf);
		n = strrchr(pbuf,'\\');
		u = strrchr(pbuf,'/');
		if (u>n)
			n = u;
		n = n ? n+1:pbuf;
		memset(stylesPath,0,sizeof(stylesPath));
		memcpy_s(stylesPath,sizeof(stylesPath)-1,pbuf,n-pbuf);
		strcat_s(stylesPath,sizeof(stylesPath)-1,".rsrc/");
		strcat_s(stylesPath,sizeof(stylesPath)-1,n);
		r =-1;
		if ((r=_open(txtPath,_O_RDONLY|_O_BINARY))<0
				|| _filelength64(r)<=0)
			*ftype =-1;
		else
		{
			if (_read(r,buf,16)>15 && *(int32*)buf==_be32_native(256)
					&& _be32_native(((int32*)buf)[1])
					==_be32_native(((int32*)buf)[2])+256)
				*ftype = _ft_rsrc, pbufrecog = 1;
			if (*ftype<_ft_txt)
				strcpy(stylesPath,txtPath);
		}
		if (r>=0)
			{_close(r); r =-1;}
		if (*ftype==-1 && (r=_open(stylesPath,_O_RDONLY|_O_BINARY))>=0
				&& _filelength64(r)>0)
		{
			if (_read(r,buf,16)>15 && *(int32*)buf==_be32_native(256)
					&& _be32_native(((int32*)buf)[1])
					==_be32_native(((int32*)buf)[2])+256)
				*ftype = _ft_rsrc, pbufrecog = 1;
			//if (*ftype<_ft_txt)
			//	strcpy(stylesPath,txtPath);
		}
		if (r>=0)
			_close(r);
	} else if (strstr(pbuf,"\\.rsrc\\") || strstr(pbuf,"/.rsrc/"))
	{
		*ftype = _ft_rsrc;
		strcpy_s(txtPath,sizeof(txtPath)-1,pbuf);
		strcpy_s(stylesPath,sizeof(stylesPath)-1,pbuf);
		pbufrecog = 1;
	} else if (*ftype<_ft_srf)
		if (*ftype==_ft_nrct)
		{ /* rect array - temporary only */
			int64		z = 0;
			int		h, c;

			*ftype = prtype;
			if ((h=_open(pbuf,_O_RDONLY|_O_BINARY))>0
					&& (z=_filelength64(h))>1) /* under 1 byte invalid */
			{
				free(nrctp); nrctp = NULL;
				nrctc = 0;
				z = (z-2)>>1;
				if (_read(h,buf,2)==2)
				{
					rect16		*b;

					r = (*(const int8*)buf<<8)|buf[1];
					if (r>z)
						r = z;
					if (r>0 && (b=calloc(8,r))!=NULL)
					{
						nrctp = b;
						c = 0;
						do
						{
							if (_read(h,buf,8)!=8)
								break;
							if (c<nrctcs)
								continue;
							b->top = (*(const int8*)buf<<8)|buf[1];
							b->left = (((const int8*)buf)[2]<<8)|buf[3];
							b->bottom = (((const int8*)buf)[4]<<8)|buf[5];
							b->right = (((const int8*)buf)[6]<<8)|buf[7];
						} while(++b,++c<r);
						nrctc = c;
						offsetRects(nrctp,nrctc,nrctox+pxt.x,nrctoy+pxt.y);
					}
				}
				_close(h);
			}
			path = (char*)-1;
		} else
		{
			picGIFFrame = (*ftype==_ft_gif)-1;
			path = pbuf;
		}
#ifdef SNDPLAY
	if (*ftype-_ft_srf+0u<=_ft_rsrc-_ft_srf)
		playSoundListFile(pbuf);
#endif
	return path;
}

static char *processCmdLine(const char *s,
	long int *depth,
	dFileType *ftype)
{
	static const char	*strs[3] = {"depth","originx","originy"};

	char		*picPath = NULL;
	const char	*n;
	long int	v;
	size_t		i;
	int		r;

	if (!s)
		return NULL;
	picGIFFrame =-1;
	for(;;)
	{
		while(*s==' ')
			++s;
		if (!*s || *s!='-')
			break;
		n = "DXYS";
		if ((i=(const char*)memchr(n,s[1],4)-n)<4u)
		{
			s+= 2, n = s, r = strlen(s);
			if ((s=readanyinteger(s,r,&v))==NULL || v==IMIN(v))
			{
				dprintf("# bad number for %s at %.15s%s;g",
					strs[i],n-2,r>13 ? "...":"");
				v = 8;
				s = n;
				while(*s && *s!=' ')
					++s;
			} else
				while(*s==' ')
					++s;
			switch(i)
			{
			case 0: *depth = v; break;
			case 1: nrctox = v; break;
			case 2: nrctoy = v; break;
			case 3: nrctcs = v; break;
			}
		} else if (s[1]=='F')
		{
			n = s+=2;
			r = *s=='"' ? '"':' ';
			while(*s && *s!=(char)r)
				++s;
			i = s-n;
			if (i<255)
			{
				memcpy(temporary+256,n,i);
				temporary[256+i] = 0;
				fontsSuccess+= fmLoadMaster(temporary+256,iterOptFont);
			}
		} else
		{
			dprintf("# unknown switch '-%c';g",
				s[1]);
			while(*s!=' ')
				++s;
		}
	}
	if (*s)
	{
		if (*s=='"')
		{
			++s;
			if ((n=strchr(s,'"'))==NULL)
				goto noname; /* invalid */
		} else if ((n=strchr(s,' '))==NULL)
			n = s+strlen(s);
		picPath = processFile(s,n-s,ftype);
	} else
	noname:	picPath = NULL;

	return picPath;
}

static void postAppRequest(int flag)
{
	eventRecord	event;

	event = mEventState;
	event.what = osEvt;
	event.message = (void*)(flag|appRequestMessage<<24);
	PostMEvent(&event);
}

static int __cdecl paramRequest(TCHAR *param)
{
	long int	depth = 0;
	char		*path;
	/* needs to be processed immediately, before ptr becomes invalid */
	savedftype = tbOpenFile[tbFront].fType;
	path = processCmdLine(param,&depth,&tbOpenFile[tbFront].fType);
	if ((size_t)path!=~0u)
	{
		appReceivedParam = !!param;
		picPath = path;
	}
	postAppRequest(changeFlag);
	//printf(" got param '%s'\n",
	//	param);
	return 0;
}

static void __cdecl myDragAcceptFunc(HDROP hDrop)
{
	TCHAR		a[260];
	char		*path;
	UINT		n, i;

	n = DragQueryFile(hDrop,~0u,NULL,0);
	for(i=0; i<n; ++i)
		if (DragQueryFile(hDrop,i,NULL,0)<sizeof(a)-1)
		{
			(void)DragQueryFile(hDrop,i,a,sizeof(a));
			savedftype = tbOpenFile[tbFront].fType;
			path = processFile(a,strlen(a),&tbOpenFile[tbFront].fType);
			if ((size_t)path!=~0u)
			{
				appReceivedParam = 1;
				picPath = path;
				postAppRequest(changeFlag);
			}
		}
	DragFinish(hDrop);
}

int WINAPI WinMain(HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPSTR lpCmdLine,
	int iCmdShow)
{
	eventRecord	event;

	rect16		b;
	const uint8	*textTitle = NULL;
	long int	depth = 0;
	int		r, i;
	boolean		update = 1, enteri = 0;
(void)hPrevInstance; (void)iCmdShow;
	mEventCmdRequest = paramRequest;
	if (wRegisterDev(hInstance,
			LoadIcon(NULL,IDI_APPLICATION),
			NULL,
			0,
			devWindowCName)<=0)
	{
		wRequestSwitchDev(devWindowCName,lpCmdLine);
		return 0;
	}
	set_mouse_cursor(_curs_wait,0);

	GetWorkingDirectory();

	//srand(GetTickCount());

	r = strlen(workingDir);
	memcpy(temporary,workingDir,r);
#if 0
	strcpy(temporary+r,"stdout.txt");
	(void)freopen(temporary,"w",stdout);
#endif
	strcpy(temporary+r,"Fonts.rsrc");
	ffDefault = 3;
	ffFWDefault = monaco;
	fontsSuccess = fmLoadMaster(temporary,NULL);
	strcpy(temporary+r,"ProFont.dfont");
	i = fmLoadMaster(temporary,iterAltFW);
/*	if (i<=0)
	{ // the renderer does not handle openType well
	  // Classic Mac OS doesn't handle it at all
		strcpy(temporary+r,"Code_New_Roman.dfont");
		i = fmLoadMaster(temporary,iterAltFW);
	}*/
	fontsSuccess+= i;

	strcpy(temporary+r,"stderr.txt");
	remove(temporary);

	strcpy_s(txtPath,sizeof(txtPath)-1,workingDir);
	strcat_s(txtPath,sizeof(txtPath)-1,"README.txt");

	mEventWMin.x = MINSCRW;
	mEventWMin.y = MINSCRH;
	mEventWMax.x = MAXSCRW;
	mEventWMax.y = MAXSCRH;
	mEventDropFilesFunc = myDragAcceptFunc;

	picPath = processCmdLine(lpCmdLine,&depth,&tbOpenFile[tbFront].fType);
	if (tbOpenFile[tbFront].fType<_ft_srf)
		tbOpenFile[tbFront].fType =-1;

	if ((r=wSetVideoMode(depth ? depth:DEFAULT_BPP,DEFSCRW,DEFSCRH,CW_TOPLEFT,0))<=0)
	{
		if (r)
			vhalt((sprintf(temporary,"error setting %dx%d %dbpp mode (%d)",
				DEFSCRW,DEFSCRH,((depth-1)&31)+1,r),temporary));
		else if (depth>0)
			vhalt("could not create video window");
		goto end;
	}

	/* initialize the world port */
	lockPixels(screen_desktop_window.port.map);
	pmErase(screen_desktop_window.port.map,screen_desktop_window.port.fgSolid);
	/* Windows 8 takes brush background second */
	if (screen_desktop_window.port.map)
	{
		update_screen_rect(0,0,
			screen_desktop_window.bounds.right=screen_desktop_window.port.map->w,
			screen_desktop_window.bounds.bottom=screen_desktop_window.port.map->h);
	}
	savedWScreenW = screen_desktop_window.bounds.right;
	savedWScreenH = screen_desktop_window.bounds.bottom;

	testGraphics();

	/* start the timer */
	global_timer_thread = CreateTimerThread(MAXTIMERS);
	graphic_update_thread = CreateTimerThread(1);
	PrimeTimer(global_timer_thread,InstallTimer(global_timer_thread,mOSVBL_timer),1);
	//InitializeCriticalSection(&drawMutex);
	PrimeTimer(graphic_update_thread,InstallTimer(graphic_update_thread,graphic_update_timer),1);

	Microseconds(&timer_start);

	/* as of 2014.08.09, origin works in all areas */
//	screen_desktop_window.port.origin.x =-8;
	screen_desktop_window.port.origin.y = 4;

	/* initialize the clip test */
	pictClipRgn = NULL;
	ov.top = ov.left =-48<<16|0x8000;
	ov.bottom = DEFSCRH+48<<16|0x8000, ov.right = DEFSCRW+48<<16|0x8000;
	ovalRgn(&pictClipRgn,&ov,IMIN(int16),IMAX(int16));
/*	ln[1].x = ln->x = ln->y = 20<<16;
	ln[1].y = 320+48<<16;
	rgnXorLine(&pictClipRgn,IMIN(int16),IMAX(int16),*ln,ln[1]);
	ln[1].x = ln->x = 330<<16;
	rgnXorLine(&pictClipRgn,IMIN(int16),IMAX(int16),*ln,ln[1]);*/

	screen_desktop_window.port.clipRgn->bBox.top = 8;
	//screen_desktop_window.port.origin.x+= 16;

	/* initialize the graphics */
	pxt.x = pxt.y = 0, pxd = pxt;
	ov.left = 0x0B48000;
	tr->x = 0x08A8000;
	tr->y = 0x0008000;
	ov.right = 0x1548000;
	tr[1].x = 0x1248000;
	tr[1].y = ov.top = 0x01B8000;
	tr[2].x = 0x1208000;
	tr[2].y = ov.bottom = 0x0C08000;
	tr[3].x = 0x0718000;// rand()%(DEFSCRW*0x4000)+DEFSCRW*0x8000;
	tr[3].y = 0x0C88000;// rand()%(DEFSCRH*0x4000)+DEFSCRH*0x8000;
	ln->x = 0x1BE8000;
	ln->y = 0x1248000;
	ln[1].x = 0x24B8000;
	ln[1].y = 0x18A8000;
	rr.left = 0x10E8000;
	rr.top = 0x1448000;
	rr.right = 0x19B8000;
	rr.bottom = 0x1AA8000;
	/* initially white */
	lineColor = rgb_white;

	ov.left+= 32<<16;
	ov.right+= 32<<16;

	_get_gray_pattern(grayPat);

	rTextFace(&screen_desktop_window.port,txfont&31,txsizes[txsize&7],txstyl);
	txPos.x = 96, txPos.y =-400;
	txhw = rTextWidth(&screen_desktop_window.port,txs,-1,0)/2;

	/* load the image */
	strcpy(temporary,workingDir);
	strcat(temporary,"blot.pct");

	pxt.x = pxz.x = pxd.x = 96, pxt.y = pxz.y = pxd.y = 200;
	if ((r=loadInputFile(picPath))>1)
	{
		strcpy(txtPath,picPath);
		strcpy(stylesPath,picPath);
		tbOpenFile[tbFront].fType = r;
	}

#ifdef TEST_FILTER
	/* will come out half-intensity */
	RGBToPen(screen_desktop_window.port.map,&rgb_white); /* make sure an iTable is built */
	filterTable = calloc(8,2049);
	filterTable->flags = 0x8000;
	ctMakeGrays(filterTable->colors,-(filterTable->count=(screen_desktop_window.port.map->minCmp
		? 1u<<screen_desktop_window.port.map->minCmp:1024))*2);
#endif
	textTitle = (const uint8*)"\5About";
	if (*stylesPath)
	{
		textTitle = tbOpenFile[tbFront].fType>=_ft_txt ? (const uint8*)"\4Text":(const uint8*)"\4Info";
		event = mEventState;
		event.what = keyDown;
		event.modifiers = menuKey;
		event.message = (void*)'t';
		PostMEvent(&event);
	}

	createMainButtons(textTitle);
	staty = screen_desktop_window.port.map->h+2;

	interp_init_basic_vars(&evalglobals,&r);

	set_mouse_cursor(_curs_arrow,0);

	/* window-based updates (e.g., controls) require visibility */
	screen_desktop_window.visible = 1;

	mEventMouseDetail = 1;

	decorActive = 1;

	for(;;)
	{
		if (!WaitNextEvent(-1u,&event,0,NULL))
			goto upd;
		enteri = 0;
		if (event.what==mouseDown)
		{
			pt2d16		loc = event.where;
			ctlObj		*ctl;

			global_to_local_pt(&loc);
			decorActive|= 2;
			if ((r=wFindControl(&screen_desktop_window,loc,&ctl))>0)
			{
				if (ctl->rtn(ctl,wcMsgClick,&loc))
				{
					decorActive&=~2;
					if ((int)ctl->refCon=='c' || (int)ctl->refCon=='s')
					{
						wSetControlValue(ctl,!ctl->value);
						update = 1;
						goto upd;
					} else
						goto about;
				}
			} else if (ptInPtRange(loc,pxt,pxd))
				pxmdv = loc, pxmdrag = 1;
			else
				nrcti = nrctc>0;
			decorActive&=~2;
		} else if (event.what==mouseUp)
			pxmdrag = 0, nrcti = 0;
		else if (event.what==updateEvt)
		{
			//wBeginUpdate(&screen_desktop_window);
			rFillRect(&screen_desktop_window.port,0,0,
				screen_desktop_window.port.map->w,
				screen_desktop_window.port.map->h);
			//wEndUpdate(&screen_desktop_window);
			update = 1;
		} else if (event.what==keyDown)
			if (event.modifiers&menuKey)
			{
				enteri = (uint8)event.message=='i'
						&& event.modifiers&shiftKey;
				if ((uint8)event.message=='q'
						|| (uint32)event.message
						==(MKV_F4<<8|kFunctionKeyCharCode))
					event.what = closeEvent;
				else if (enteri || (uint8)event.message=='a'
						|| (uint8)event.message=='t'
						|| (uint8)event.message=='/'
						|| (uint8)event.message=='d')
				{ //"\10Close \x11W"
				about:	wSetControlTitle(screen_desktop_window.ctlList,(uint8*)"\5Close");
					wHiliteControl(wGetIndexedControl(&screen_desktop_window,
						_it_dither),255);
					wHiliteControl(wGetIndexedControl(&screen_desktop_window,
						_it_smoothing),255);
					b.top = 20, b.left = 30;
					b.bottom = screen_desktop_window.port.map->h-52-8;
					b.right = screen_desktop_window.port.map->w-32;
					decorActive = 0;
					doDebug = event.modifiers&shiftKey || (uint8)event.message=='d';
					if (enteri)
					{
						event.what = keyDown;
						event.modifiers = menuKey|shiftKey;
						event.message = (void*)'i';
						PostMEvent(&event);
					}
					for(r=128; r==128; )
						r = boxer((uint8)event.message=='d' ? NULL:txtPath,&b,
								(uint8)event.message=='d' ? NULL:stylesPath,
								tbOpenFile[tbFront].fType,
								!!(event.modifiers&optionKey));
					if (r>1)
						break;
					ltick = TickCount();
					decorActive = 1;
					wSetControlTitle(screen_desktop_window.ctlList,textTitle);
					wHiliteControl(wGetIndexedControl(&screen_desktop_window,_it_dither),0);
					wHiliteControl(wGetIndexedControl(&screen_desktop_window,_it_smoothing),0);
					if (!global_timer_thread)
					{
						update = 1;
						goto upd;
					}
				} else if ((uint8)event.message=='='
						&& screen_desktop_window.bounds.right<MAXSCRW)
				{
					wSetVideoMode(depth ? depth:DEFAULT_BPP,MAXSCRW,MAXSCRH,
						CW_TOPLEFT,0);
					screen_desktop_window.port.origin.y = 4;
					goto rsiz;
				}
			} else if ((uint8)event.message==kEscapeCharCode)
			{
				EndTimerThread(global_timer_thread); global_timer_thread = NULL;
				EndTimerThread(graphic_update_thread); graphic_update_thread = NULL;
			}
			else if ((uint8)event.message=='7') /* width not affected */
				txstyl^= _style_strikethrough;
			else if ((uint8)event.message=='8')
				txstyl^= _style_overline;
		if (event.what==closeEvent)
			break;
		if (event.what==osEvt)
		{
			r = (uint32)event.message>>24;
			if (r==appScreenChangeMessage)
			{
			rsiz:	adjustScreen(0);
				disposeRgn(&pictClipRgn);
			}
		}
		if (appReceivedParam)
		{
			if (pic)
				rFillRect(&screen_desktop_window.port,pxt.x,pxt.y,
					pxt.x+pxd.x,pxt.y+pxd.y);
			if ((r=loadInputFile(picPath))>1)
			{
				strcpy(txtPath,picPath);
				strcpy(stylesPath,picPath);
				tbOpenFile[tbFront].fType = r;
			}
			appReceivedParam = 0;
			if (*stylesPath)
			{
				textTitle = tbOpenFile[tbFront].fType>=_ft_txt ? (const uint8*)"\4Text"
					:(const uint8*)"\4Info";
				event = mEventState;
				event.what = keyDown;
				event.modifiers = menuKey;
				event.message = (void*)'t';
				PostMEvent(&event);
			} else
			{
				textTitle = (const uint8*)"\5About";
				wSetControlTitle(screen_desktop_window.ctlList,textTitle);
			}
		}
	upd:	if (!evtRActive)
		{
			evtRActive = 1;
			update_window(update);
			evtRActive = 0;
		}
		update = 0;
	}
end:	EndTimerThread(graphic_update_thread);
	Sleep(1);
	wKillControls(&screen_desktop_window);
	wSetVideoMode(-1,0,0,0,0);
	EndTimerThread(global_timer_thread);
	//Sleep(10);
	fmUnloadAll();
	//if (!ftell(stdout))
	//	fclose(stdout), remove("stdout.txt");
	wUnregisterDev(hInstance,devWindowCName);
	return 0;
}
