/*
 lzw.c
  LempelZivWelch algorithm

    All LZW relevant patents on the original algorithm expired
  in 2004, so there are no legal issues in distributing this.

  web.archive.org/web/20090602212118/http://www.unisys.com/about__unisys/lzw

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#include "lzw.h"

/* Code from here to end of file has been adapted from XPaint:           */
/* +-------------------------------------------------------------------+ */
/* | Copyright 1990, 1991, 1993 David Koblas.                          | */
/* | Copyright 1996 Torsten Martinsen.                                 | */
/* |   Permission to use, copy, modify, and distribute this software   | */
/* |   and its documentation for any purpose and without fee is hereby | */
/* |   granted, provided that the above copyright notice appear in all | */
/* |   copies and that both that copyright notice and this permission  | */
/* |   notice appear in supporting documentation.  This software is    | */
/* |   provided "as is" without express or implied warranty.           | */
/* +-------------------------------------------------------------------+ */

#define	ReadOK(s,b,z)		(rf(s,z,b)==z)

static int GetCode(sio_func_t rf,
	const void *s,
	int code_size,
	int flag,
	LZWstate *lzw)
{
	int		i, j, ret;
	uint8		count;

	if (flag)
	{
		lzw->curbit = lzw->lastbit = 0;
		lzw->flag = 0;
		return 0;
	}
	if ((lzw->curbit+code_size)>=lzw->lastbit)
	{
		if (lzw->flag&1)
		{
			if (lzw->curbit>=lzw->lastbit)
				; /* ran off the end of my bits */
			return -1;
		}
		lzw->buf[0] = lzw->buf[lzw->last_byte-2];
		lzw->buf[1] = lzw->buf[lzw->last_byte-1];

		if (!ReadOK(s,&count,1) || !count || !ReadOK(s,lzw->buf+2,count))
			lzw->flag|= 1;

		lzw->curbit = (lzw->curbit-lzw->lastbit)+16;
		lzw->lastbit = (lzw->last_byte=2+count)*8;
	}
	for(i=lzw->curbit,j=0,ret=0; j<code_size; ++i,++j)
		ret|= (lzw->buf[i>>3]>>(i&7)&1)<<j;

	lzw->curbit = i;

	return ret;
}

int LZWReadByte(sio_func_t rf,
	const void *s,
	int flag,
	int input_code_size,
	LZWstate *lzw)
{
	int		code, incode;
	register int	i;

	if (input_code_size-1u>=MAX_LZW_BITS-1u) /* within (0,MAX_LZW_BITS) */
		return -1; /* buffer underflow or overflow */
	if (flag)
	{
		lzw->set_code_size = input_code_size;
		lzw->code_size = lzw->set_code_size+1;
		lzw->clear_code = 1<<lzw->set_code_size;
		lzw->end_code = lzw->clear_code+1;
		lzw->max_code_size = 2*lzw->clear_code;
		lzw->max_code = lzw->clear_code+2;

		GetCode(rf,s,0,1,lzw);

		lzw->fresh = 1;

		for(i=0; i<lzw->clear_code; ++i)
		{
			lzw->table[0][i] = 0;
			lzw->table[1][i] = i;
		}
		for(; i<(1<<MAX_LZW_BITS); ++i)
			lzw->table[0][i] = lzw->table[1][0] = 0;

		lzw->sp = lzw->stack;

		return 0;
	} else if (lzw->fresh)
	{
		lzw->fresh = 0;
		do
			lzw->firstcode = lzw->oldcode
				= GetCode(rf,s,lzw->code_size,0,lzw);
		while(lzw->firstcode==lzw->clear_code);
		return lzw->firstcode;
	}
	if (lzw->sp>lzw->stack)
		return *--lzw->sp;

	while ((code=GetCode(rf,s,lzw->code_size,0,lzw))>=0)
	{
		if (code==lzw->clear_code)
		{
			for (i=0; i<lzw->clear_code; ++i)
				lzw->table[0][i] = 0, lzw->table[1][i] = i;
			for (; i<(1<<MAX_LZW_BITS); ++i)
				lzw->table[0][i] = lzw->table[1][i] = 0;
			lzw->code_size = lzw->set_code_size+1;
			lzw->max_code_size = 2*lzw->clear_code;
			lzw->max_code = lzw->clear_code+2;
			lzw->sp = lzw->stack;
			lzw->firstcode = lzw->oldcode = GetCode(rf,s,lzw->code_size,0,lzw);
			return lzw->firstcode;
		} else if (code==lzw->end_code)
		{
			uint8		buf[260];
			int		count;

			if (lzw->flag&1) //ZeroDataBlock)
				return -5;
			while(ReadOK(s,&count,1) && count && ReadOK(s,buf,count))
				;

			if (count)
			{ /* missing EOD in data stream (common occurence) */
			}
			return -5;
		}
		incode = code;

		if (code>=lzw->max_code)
			*lzw->sp++ = lzw->firstcode, code = lzw->oldcode;

		while(code>=lzw->clear_code)
		{
			*lzw->sp++ = lzw->table[1][code];
			if (code==lzw->table[0][code])
			{ /* circular table entry BIG ERROR */
				return -14;
			}
			code = lzw->table[0][code];
		}

		*lzw->sp++ = lzw->firstcode = lzw->table[1][code];

		if ((code=lzw->max_code)<(1<<MAX_LZW_BITS))
		{
			lzw->table[0][code] = lzw->oldcode;
			lzw->table[1][code] = lzw->firstcode;
			if (++lzw->max_code>=lzw->max_code_size
					&& lzw->max_code_size<(1<<MAX_LZW_BITS))
			{
				lzw->max_code_size*= 2;
				++lzw->code_size;
			}
		}
		lzw->oldcode = incode;

		if (lzw->sp>lzw->stack)
			return *--lzw->sp;
	}
	return code;
}
