/*
 mblt.c
  for low-level span/slab drawing, blitting

   Revised: 2023.08.19
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.07 (mgraf.c)
  With rgbToPenColor, colors are no longer expanded here (slightly faster).
  Translucency...

2014.03.10
  Split from "mgraf.c", into mblt and mgraphic.
  Then all the assembly into _mblt.
2014.03.30
  Separate-channel [linear]-blending, just like QD.
2014.04.05
  Rewrote most of the monochrome (8-bit) blitters; 1/2bpp work (includes text).
2014.04.11
  rCopyBits, rOrBits and rXorBits - aligned or unaligned bits, working.
2014.04.18
  Depth conversion, all depths.  Finally.  Tested source bpp: 1, 4, 8, 15.
2014.04.21
  The 'a' param for bltRoutines is gone - use the f->a field.
  Patterns work.  Not sure if origin config works as it should.
2014.04.22
  Rewriting for inclusion of QD 'not' modes...
2014.04.23
  Split 16/32-bit interpolation macros so 9-bit blending can work.
  All implemented transfer modes work.
2014.04.27
  Implemented addPin; it took too many hours.
Separate channels for grays produced inconsistent results...what took all that time.
  Added addOver, subPin.
2014.04.28
  Added adMax, adMin, subOver.  This file's gotten large with the arithmetic modes.
2014.09.21
  Indexed conv.-transparent uses bkColor like direct modes (originally 0 for t).
2019.03.05
  Replaced not-and-and-or mask method with xor-and-xor.
*/

#define __MGRAF_INTERNAL

#include <string.h>
#include "mgraf.h"
#include "mblt.h"
#include "ftrig.h"	/* for cntlz32 */

/* ---- constants */

#define blndPrec	10 /* any higher promotes overflow in 10:11:10 */

/* ---- macros */

/* highest pre-shifted to eliminate overflow */
/* signed computation still didn't fix blue overflow */
#define rgblerp32(a,v,r,g,mr,mg,mb,kr,kg,kb) \
	(r=a&kr,r=(((((v&kr)-r)>>blndPrec)*mr)+r)&kr, \
	g=a&kg,r+=((((v&kg)-g)*mg>>blndPrec)+g)&kg, \
	a&=kb,r+=((((v&kb)-a)*mb>>blndPrec)+a)&kb)
#define rgblerp16(a,v,r,g,mr,mg,mb,kr,kg,kb) \
	(r=a&kr,r=((((v&kr)-r)*mr>>blndPrec)+r)&kr, \
	g=a&kg,r+=((((v&kg)-g)*mg>>blndPrec)+g)&kg, \
	a&=kb,r+=((((v&kb)-a)*mb>>blndPrec)+a)&kb)

#define rgbaddpin(a,v,b,u,mr,mg,mb,kr,kg,kb) \
	(u=(a&kr)+(v&kr), b=MIN(u,mr), \
	u=(a&kg)+(v&kg), b|=MIN(u,mg), \
	u=(a&kb)+(v&kb), b|=MIN(u,mb))
#define rgbaddover(a,v,b,kr,kg,kb) \
	(b=((a&kr)+(v&kr))&kr, \
	b|=((a&kg)+(v&kg))&kg, \
	b|=((a&kb)+(v&kb))&kb)
#define rgbsubpin(a,v,b,u,mr,mg,mb,kr,kg,kb) \
	(u=(a&kr)-(v&kr), b=MAX(u,mr), \
	u=(a&kg)-(v&kg), b|=MAX(u,mg), \
	u=(a&kb)-(v&kb), b|=MAX(u,mb))
#define rgbsubover(a,v,b,kr,kg,kb) \
	(b=((a&kr)-(v&kr))&kr, \
	b|=((a&kg)-(v&kg))&kg, \
	b|=((a&kb)-(v&kb))&kb)
#define rgbadmax(a,v,b,u,kr,kg,kb) \
	(u=a&kr, b=MAX(u,v&kr), \
	u=a&kg, b|=MAX(u,v&kg), \
	u=a&kb, b|=MAX(u,v&kb))
#define rgbadmin(a,v,b,u,kr,kg,kb) \
	(u=a&kr, b=MIN(u,v&kr), \
	u=a&kg, b|=MIN(u,v&kg), \
	u=a&kb, b|=MIN(u,v&kb))

/* ---- code */

/* ---- transfer routines - see DEF_BLT_RTN in mgraf.h */

#if envx86 || envx64
 extern DEF_BLT_RTN(sldx1);
 extern DEF_BLT_RTN(sldx2);
 extern DEF_BLT_RTN(sldx4);
 extern DEF_BLT_RTN(sldx8);
 extern DEF_BLT_RTN(sldx16);
 extern DEF_BLT_RTN(sldx32);
#else
 #define sld_x(h)	do{ uint32 *d, k, v; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=~((uint32)((int)0x80000000>>(MIN(c,32)-1))>>o),k=be32_native(k); \
	v=f->sc, c+=o; \
	*d=((*d^v)&k)^v; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c&32) *d++=v; \
	if(c>>=6) do *d=d[1]=v; while(d+=2,--c); \
	if(o&=31) k=~0u>>o, k=be32_native(k), *d=((*d^v)&k)^v; \
 }while(0)

static DEF_BLT_RTN(sldx1)
	{sld_x(0);}

static DEF_BLT_RTN(sldx2)
	{sld_x(1);}

static DEF_BLT_RTN(sldx4)
	{sld_x(2);}

static DEF_BLT_RTN(sldx8)
{
	int8		*d = f->a+o;
	int32		v = f->sc;

	if ((int)d&1)
		*d++ = (int8)v, --c;
	if (o&c&2)
		*(int16*)d = v, d+= 2, c-= 2;
	if ((o=c)==0)
		return;
	if (c&4)
		*((int32*)d)++ = v;
	if (c>>=3)
		do
			((int32*)d)[1] = *(int32*)d = v;
		while(d+=8,--c);
	if (o&2)
		*(int16*)d = (int16)v, d+= 2;
	if (o&1)
		*d = (int8)v;
}

static DEF_BLT_RTN(sldx16)
{
	int16		*d = (int16*)f->a+o;
	int32		v = f->sc;

	if ((int)d&2)
		*d++ = (int16)v, --c;
	o = c;
	if (c&2)
		*(int32*)d = v, d+= 2;
	if (c>>=2)
		do
			((int32*)d)[1] = *(int32*)d = v;
		while(d+=4,--c);
	if (o&1)
		*d = (int16)v;
}

static DEF_BLT_RTN(sldx32)
{
	int32		*d = (int32*)f->a+o;
#if envx64
	uint64		v = (uint32)f->sc|(uint64)(uint32)f->sc<<32;

	if (c&1)
		*d++ = (uint32)v;
	if (c&2)
		*(uint64*)d = v, d+= 2;
	if (c>>=2)
		do
			((uint64*)d)[1] = *(uint64*)d = v;
		while(d+=4,--c);
#else
	int32		v = f->sc;

	if (c&1)
		*d++ = v;
	if (c&2)
		*d = v, d[1] = v, d+= 2;
	if (c>>=2)
		do
			d[3] = d[2] = d[1] = *d = v;
		while(d+=4,--c);
#endif
}
#endif

static DEF_BLT_RTN(imgx1)
	{rCopyBits(f->a,f->cnv.s,o,o-f->vx-f->cnv.sx,c);}

static DEF_BLT_RTN(imgx2)
	{rCopyBits(f->a,f->cnv.s,o<<1,(o-f->vx)<<1,c<<1);}

static DEF_BLT_RTN(imgx4)
	{rCopyBits(f->a,f->cnv.s,o<<2,(o-f->vx)<<2,c<<2);}

static DEF_BLT_RTN(imgx8)
	{memmove(f->a+o,(const int8*)f->cnv.s+o-f->vx,c);}

static DEF_BLT_RTN(imgx16)
	{memmove((int16*)f->a+o,(const int16*)f->cnv.s+o-f->vx,c*2);}

static DEF_BLT_RTN(imgx32)
	{memmove((int32*)f->a+o,(const int32*)f->cnv.s+o-f->vx,c*4);}

/* image - not */

static DEF_BLT_RTN(imgn1)
	{rNotCopyBits(f->a,f->cnv.s,o,o-f->vx,c,f->opr);}

static DEF_BLT_RTN(imgn2)
	{rNotCopyBits(f->a,f->cnv.s,o<<1,(o-f->vx)<<1,c<<1,f->opr);}

static DEF_BLT_RTN(imgn4)
	{rNotCopyBits(f->a,f->cnv.s,o<<2,(o-f->vx)<<2,c<<2,f->opr);}

static DEF_BLT_RTN(imgn8)
	{rNotCopyBits(f->a,f->cnv.s,o<<3,(o-f->vx)<<3,c<<3,f->opr);}

static DEF_BLT_RTN(imgn16)
{
	const int16	*s = (const int16*)f->cnv.s+o-f->vx;
	int16		*d = (int16*)f->a+o;
	int32		cm = f->opr;

	if (o&1)
		*d = *s^(int16)cm, ++s, ++d, --c;
	o = c;
	if (c&2)
		*(uint32*)d = urw32(s)^cm, s+= 2, d+= 2;
	if (c&4)
		*(uint32*)d = urw32(s)^cm, ((uint32*)d)[1] = urw32(s+2)^cm, s+= 4, d+= 4;
	if (c>>=3)
		do
			*(uint32*)d = urw32(s)^cm,
				((uint32*)d)[1] = urw32(s+2)^cm,
				((uint32*)d)[2] = urw32(s+4)^cm,
				((uint32*)d)[3] = urw32(s+6)^cm;
		while(s+=8,d+=8,--c);
	if (o&=1)
		*d = *s^(int16)cm;
}

static DEF_BLT_RTN(imgn32)
{
	const int32	*s = (const int32*)f->cnv.s+o-f->vx;
	int32		*d = (int32*)f->a+o;
	int32		cm = f->opr;

	if (c&1)
		*d++ = *s^cm, ++s;
	if (c&2)
		*d = *s^cm, d[1] = s[1]^cm, d+=2, s+= 2;
	if (c>>=2)
		do
			*d = *s^cm, d[1] = s[1]^cm, d[2] = s[2]^cm, d[3] = s[3]^cm;
		while(d+=4,s+=4,--c);
}

/* image - remap */

/* in the old days, a single rotate-left instruction would cover an 8-bit pattern... */

/* no rnx, as source-not will invert the table */

#define imgrx(h)	do{ const uint8 *s=f->cnv.s; const uint16 *t=f->cnv.t; \
	uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=~((uint32)((int)0x80000000>>(MIN(c,32)-1))>>o),k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,j&=f->mx,o|=t[s[j]]<<v; while(++j,--i); \
	o=be32_native(o),*d=((*d^o)&k)^o; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h,k=32,v=0; do k-=1<<h,j&=f->mx,v|=t[s[j]]<<k; while(++j,--i); \
			*d++=be32_native(v); \
		} while(--c); \
	if(o&=31) { k=~0u>>o, k=be32_native(k), \
		i=o>>h,o=32,v=0; do o-=1<<h,j&=f->mx,v|=t[s[j]]<<o; while(++j,--i); \
		v=be32_native(v),*d=((*d^v)&k)^v; \
	} \
}while(0)

static DEF_BLT_RTN(imgr1)
	{imgrx(0);}

static DEF_BLT_RTN(imgr2)
	{imgrx(1);}

static DEF_BLT_RTN(imgr4)
	{imgrx(2);}

static DEF_BLT_RTN(imgr8)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	int8		*d = f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d = t[s[j]], ++j, ++d, --c;
	if (o&c&2)
	{
		j&= f->mx, v = t[s[j]]<<BIG_ENDIAN*8, ++j;
		j&= f->mx, v|= t[s[j]]<<!BIG_ENDIAN*8, ++j;
		*(int16*)d = v, d+= 2, c-= 2;
	}
	o = c;
	if (c>>=2)
	do
	{
		j&= f->mx, v = t[s[j]]<<BIG_ENDIAN*24, ++j;
		j&= f->mx, v|= t[s[j]]<<8+BIG_ENDIAN*8, ++j;
		j&= f->mx, v|= t[s[j]]<<8+!BIG_ENDIAN*8, ++j;
		j&= f->mx, v|= t[s[j]]<<!BIG_ENDIAN*24, ++j;
		*(int32*)d = v;
	} while(d+=4,--c);
	if (o&2)
	{
		j&= f->mx, v = t[s[j]]<<BIG_ENDIAN*8, ++j;
		j&= f->mx, v|= t[s[j]]<<!BIG_ENDIAN*8, ++j;
		*(int16*)d = v, d+= 2;
	}
	if (o&1)
		j&= f->mx, *d = t[s[j]];
}

static DEF_BLT_RTN(imgr16)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d = t[s[j]], ++j, ++d, --c;
	o = c;
	if (c>>=1)
		do
		{
			j&= f->mx, v = t[s[j]]<<BIG_ENDIAN*16, ++j;
			j&= f->mx, v|= t[s[j]]<<!BIG_ENDIAN*16, ++j;
			*(int32*)d = v;
		} while(d+=2,--c);
	if (o&1)
		j&= f->mx, *d = t[s[j]];
}

static DEF_BLT_RTN(imgr32)
{
	const int32	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, j = o-f->vx;

	if (c&1)
		j&= f->mx, *d = t[s[j]], ++j, ++d;
	if (c>>=1)
		do
		{
			j&= f->mx, *d = t[s[j]], ++j;
			j&= f->mx, d[1] = t[s[j]], ++j;
		} while(d+=2,--c);
}

/* image - pattern */

#define imgpx(h)	do{ const uint8 *s=f->cnv.s; \
	uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=~((uint32)((int)0x80000000>>(MIN(c,32)-1))>>o),k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,j&=f->mx,o|=s[j]<<v; while(++j,--i); \
	o=be32_native(o),*d=((*d^o)&k)^o; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h,k=32,v=0; do k-=1<<h,j&=f->mx,v|=s[j]<<k; while(++j,--i); \
			*d++=be32_native(v); \
		} while(--c); \
	if(o&=31) { k=~0u>>o, k=be32_native(k), \
		i=o>>h,o=32,v=0; do o-=1<<h,j&=f->mx,v|=s[j]<<o; while(++j,--i); \
		v=be32_native(v),*d=((*d^v)&k)^v; \
	} \
}while(0)

static DEF_BLT_RTN(imgp1)
	{imgpx(0);}

static DEF_BLT_RTN(imgp2)
	{imgpx(1);}

static DEF_BLT_RTN(imgp4)
	{imgpx(2);}

static DEF_BLT_RTN(imgp8)
{
	const uint8	*s = f->cnv.s;
	int8		*d = f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d = s[j], ++j, ++d, --c;
	if (o&c&2)
	{
		j&= f->mx, v = s[j]<<BIG_ENDIAN*8, ++j;
		j&= f->mx, v|= s[j]<<!BIG_ENDIAN*8, ++j;
		*(int16*)d = v, d+= 2, c-= 2;
	}
	o = c;
	if (c>>=2)
	do
	{
		j&= f->mx, v = s[j]<<BIG_ENDIAN*24, ++j;
		j&= f->mx, v|= s[j]<<8+BIG_ENDIAN*8, ++j;
		j&= f->mx, v|= s[j]<<8+!BIG_ENDIAN*8, ++j;
		j&= f->mx, v|= s[j]<<!BIG_ENDIAN*24, ++j;
		*(int32*)d = v;
	} while(d+=4,--c);
	if (o&2)
	{
		j&= f->mx, v = s[j]<<BIG_ENDIAN*8, ++j;
		j&= f->mx, v|= s[j]<<!BIG_ENDIAN*8, ++j;
		*(int16*)d = v, d+= 2;
	}
	if (o&1)
		j&= f->mx, *d = s[j];
}

static DEF_BLT_RTN(imgp16)
{
	const uint16	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d = s[j], ++j, ++d, --c;
	o = c;
	if (c>>=1)
		do
		{
			j&= f->mx, v = s[j]<<BIG_ENDIAN*16, ++j;
			j&= f->mx, v|= s[j]<<!BIG_ENDIAN*16, ++j;
			*(int32*)d = v;
		} while(d+=2,--c);
	if (o&1)
		j&= f->mx, *d = s[j];
}

static DEF_BLT_RTN(imgp32)
{
	const uint32	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, j = o-f->vx;

	if (c&1)
		j&= f->mx, *d = s[j], ++j, ++d;
	if (c>>=1)
		do
		{
			j&= f->mx, *d = s[j], ++j;
			j&= f->mx, d[1] = s[j], ++j;
		} while(d+=2,--c);
}

/* image - converted */

#define imgcx(h)		do{ uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=~((uint32)((int)0x80000000>>(MIN(c,32)-1))>>o),k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,o|=f->cnv.rtn(&f->cnv,j)<<v; while(++j,--i); \
	o=be32_native(o),*d=((*d^o)&k)^o; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h,k=32,v=0; do k-=1<<h,v|=f->cnv.rtn(&f->cnv,j)<<k; while(++j,--i); \
			*d++=be32_native(v); \
		} while(--c); \
	if(o&=31) { k=~0u>>o, k=be32_native(k), \
		i=o>>h,o=32,v=0; do o-=1<<h,v|=f->cnv.rtn(&f->cnv,j)<<o; while(++j,--i); \
		v=be32_native(v),*d=((*d^v)&k)^v; \
	} \
}while(0)

static DEF_BLT_RTN(imgc1)
	{imgcx(0);}

static DEF_BLT_RTN(imgc2)
	{imgcx(1);}

static DEF_BLT_RTN(imgc4)
	{imgcx(2);}

static DEF_BLT_RTN(imgc8)
	{imgcx(3);}

static DEF_BLT_RTN(imgc16)
{
	int16		*d = (int16*)f->a+o;

	o-= f->vx;
	if (c&1)
		*d = f->cnv.rtn(&f->cnv,o), ++o, ++d;
	if (c>>=1)
		do
		{
			*d = f->cnv.rtn(&f->cnv,o), ++o;
			d[1] = f->cnv.rtn(&f->cnv,o), ++o;
		} while(d+=2,--c);
}

static DEF_BLT_RTN(imgc32)
{
	int32		*d = (int32*)f->a+o;

	o-= f->vx;
	if (c&1)
		*d = f->cnv.rtn(&f->cnv,o), ++o, ++d;
	if (c>>=1)
		do
		{
			*d = f->cnv.rtn(&f->cnv,o), ++o;
			d[1] = f->cnv.rtn(&f->cnv,o), ++o;
		} while(d+=2,--c);
}

/* image transparent */

#define imgrtx(h)	do{ const uint8 *s=f->cnv.s; const uint16 *t=f->cnv.t; \
	uint32 *d, k, v, u, i, x, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	x=32-o, i=MIN((uint32)c,x)>>h, c+=o; \
	v=k=0; do if(x-=1u<<h,(u=s[j&=f->mx])!=0) k|=((1<<(1<<h))-1)<<x,v|=t[u]<<x; while(++j,--i); \
	v=be32_native(v), *d=((*d^v)&~be32_native(k))^v; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { x=32, i=32>>h, v=k=0; \
			do if(x-=1<<h,(u=s[j&=f->mx])!=0) k|=((1<<(1<<h))-1)<<x,v|=t[u]<<x; while(++j,--i); \
			v=be32_native(v), *d=((*d^v)&~be32_native(k))^v; \
		} while(++d,--c); \
	if(o&=31) { x=32, i=o>>h, v=k=0; \
		do if(x-=1<<h,j&=f->mx,(u=s[j&=f->mx])!=0) k|=((1<<(1<<h))-1)<<x,v|=t[u]<<x; while(++j,--i); \
		v=be32_native(v), *d=((*d^v)&~be32_native(k))^v; \
	} \
}while(0)

static DEF_BLT_RTN(imrt1)
	{imgrtx(0);}

static DEF_BLT_RTN(imrt2)
	{imgrtx(1);}

static DEF_BLT_RTN(imrt4)
	{imgrtx(2);}

static DEF_BLT_RTN(imrt8)
	{imgrtx(3);}

static DEF_BLT_RTN(imrt16)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int		v, j = o-f->vx;

	if (c&1)
	{
		j&= f->mx;
		if ((v=s[j])!=0)
			*d = t[v];
		++j, ++d;
	}
	if (c>>=1)
		do
		{
			j&= f->mx;
			if ((v=s[j])!=0)
				*d = t[v];
			++j, j&= f->mx;
			if ((v=s[j])!=0)
				d[1] = t[v];
		} while(++j,d+=2,--c);
}

static DEF_BLT_RTN(imrt32)
{
	const int32	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, v, j = o-f->vx;

	if (c&1)
	{
		j&= f->mx;
		if ((v=s[j])!=0)
			*d = t[v];
		++j, ++d;
	}
	if (c>>=1)
		do
		{
			j&= f->mx;
			if ((v=s[j])!=0)
				*d = t[v];
			++j, j&= f->mx;
			if ((v=s[j])!=0)
				d[1] = t[v];
		} while(++j,d+=2,--c);
}

static DEF_BLT_RTN(imgt16)
{
	const int16	*s = (const int16*)f->cnv.s;
	int16		*d = (int16*)f->a+o, v = f->opr, u;
	int		j = o-f->vx;

	if (c&1)
	{
		j&= f->mx;
		if ((u=s[j])!=v)
			*d = u;
		++j, ++d;
	}
	if (c>>=1)
		do
		{
			j&= f->mx;
			if ((u=s[j])!=v)
				*d = u;
			++j, j&= f->mx;
			if ((u=s[j])!=v)
				d[1] = u;
		} while(++j,d+=2,--c);
}

static DEF_BLT_RTN(imgt32)
{
	const int32	*s = (const int32*)f->cnv.s;
	int32		*d = (int32*)f->a+o, v = f->opr, u, j = o-f->vx;

	if (c&1)
	{
		j&= f->mx;
		if ((u=s[j])!=v)
			*d = u;
		++j, ++d;
	}
	if (c>>=1)
		do
		{
			j&= f->mx;
			if ((u=s[j])!=v)
				*d = u;
			++j, j&= f->mx;
			if ((u=s[j])!=v)
				d[1] = u;
		} while(++j,d+=2,--c);
}

/* image transparent - converted */

#define imgctx(h)	do{ uint32 *d, k, v, u, i, x, j=o-f->vx, t=f->opr; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	x=32-o, i=MIN((uint32)c,x)>>h, c+=o; \
	v=k=0; do if(x-=1u<<h,(u=f->cnv.rtn(&f->cnv,j))!=t) k|=((1<<(1<<h))-1)<<x,v|=u<<x; while(++j,--i); \
	k=~be32_native(k),v=be32_native(v), *d=((*d^v)&k)^v; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { x=32, i=32>>h, v=k=0; \
			do if(x-=1<<h,(u=f->cnv.rtn(&f->cnv,j))!=t) k|=((1<<(1<<h))-1)<<x,v|=u<<x; while(++j,--i); \
			k=~be32_native(k),v=be32_native(v), *d=((*d^v)&k)^v; \
		} while(++d,--c); \
	if(o&=31) { x=32, i=o>>h, v=k=0; \
		do if(x-=1<<h,j&=f->mx,(u=f->cnv.rtn(&f->cnv,j))!=t) k|=((1<<(1<<h))-1)<<x,v|=u<<x; while(++j,--i); \
		k=~be32_native(k),v=be32_native(v), *d=((*d^v)&k)^v; \
	} \
}while(0)

static DEF_BLT_RTN(imct1)
	{imgctx(0);}

static DEF_BLT_RTN(imct2)
	{imgctx(1);}

static DEF_BLT_RTN(imct4)
	{imgctx(2);}

static DEF_BLT_RTN(imct8)
	{imgctx(3);}

static DEF_BLT_RTN(imct16)
{
	int16		*d = (int16*)f->a+o, u, v = f->opr;

	o-= f->vx;
	if (c&1)
	{
		if ((u=f->cnv.rtn(&f->cnv,o))!=v)
			*d = u;
		++o, ++d;
	}
	if (c>>=1)
		do
		{
			if ((u=f->cnv.rtn(&f->cnv,o))!=v)
				*d = u;
			++o;
			if ((u=f->cnv.rtn(&f->cnv,o))!=v)
				d[1] = u;
		} while(++o,d+=2,--c);
}

static DEF_BLT_RTN(imct32)
{
	int32		*d = (int32*)f->a+o, u, v = f->opr;

	o-= f->vx;
	if (c&1)
	{
		if ((u=f->cnv.rtn(&f->cnv,o))!=v)
			*d = u;
		++o, ++d;
	}
	if (c>>=1)
		do
		{
			if ((u=f->cnv.rtn(&f->cnv,o))!=v)
				*d = u;
			++o;
			if ((u=f->cnv.rtn(&f->cnv,o))!=v)
				d[1] = u;
		} while(++o,d+=2,--c);
}

/* image pattern transparent */

#define imgptx(h)	do{ const uint8 *s=f->cnv.s; \
	uint32 *d, k, v, u, i, x, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	x=32-o, i=MIN((uint32)c,x)>>h, c+=o; \
	v=k=0; do if(x-=1u<<h,(u=s[j&=f->mx])!=0) k|=((1<<(1<<h))-1)<<x,v|=u<<x; while(++j,--i); \
	v=be32_native(v), *d=((*d^v)&~be32_native(k))^v; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { x=32, i=32>>h, v=k=0; \
			do if(x-=1<<h,(u=s[j&=f->mx])!=0) k|=((1<<(1<<h))-1)<<x,v|=u<<x; while(++j,--i); \
			v=be32_native(v), *d=((*d^v)&~be32_native(k))^v; \
		} while(++d,--c); \
	if(o&=31) { x=32, i=o>>h, v=k=0; \
		do if(x-=1<<h,j&=f->mx,(u=s[j&=f->mx])!=0) k|=((1<<(1<<h))-1)<<x,v|=u<<x; while(++j,--i); \
		v=be32_native(v), *d=((*d^v)&~be32_native(k))^v; \
	} \
}while(0)

static DEF_BLT_RTN(impt1)
	{imgptx(0);}

static DEF_BLT_RTN(impt2)
	{imgptx(1);}

static DEF_BLT_RTN(impt4)
	{imgptx(2);}

static DEF_BLT_RTN(impt8)
	{imgptx(3);}

/* OR-merge (AND on direct depths) */

#define sldorx(h)	do{ uint32 *d, k, v; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=be32_native(k); \
	v=f->sc; \
	*d|=v&k, c+=o; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) do *d++|=v; while(--c); \
	if(o&=31) k=~(~0u>>o), *d|=be32_native(k)&v; \
}while(0)

static DEF_BLT_RTN(slor1)
	{sldorx(0);}

static DEF_BLT_RTN(slor2)
	{sldorx(1);}

static DEF_BLT_RTN(slor4)
	{sldorx(2);}

static DEF_BLT_RTN(slor8)
{
	int32		v = f->sc;
	int8		*d = f->a+o;

	if (o&1)
		*d|= (int8)v, ++d, --c;
	if (o&c&2)
		*(int16*)d|= (int16)v, d+= 2, c-= 2;
	o = c;
	if (c>>=2)
		do
			*(int32*)d|= v;
		while(d+=4,--c);
	if (o&2)
		*(int16*)d|= (int16)v, d+= 2;
	if (o&=1)
		*d|= (int8)v;
}

static DEF_BLT_RTN(slor16)
{
	int32		v = f->sc;
	int16		*d = (int16*)f->a+o;

	if (o&1)
		*d&= (int16)v, ++d, --c;
	o = c;
	if (c>1)
	{
		if (c&2)
			*(int32*)d&= v, d+= 2;
		if (c>>=2)
			do
				*(int32*)d&= v, ((int32*)d)[1]&= v;
			while(d+=4,--c);
	}
	if (o&=1)
		*d&= (int16)v;
}

static DEF_BLT_RTN(slor32)
{
	int32		*d = (int32*)f->a+o, v = f->sc;

	if (c&1)
		*d++&= v;
	if (c>>=1)
		do
			*d&= v, d[1]&= v;
		while(d+=2,--c);
}

/* img OR */

#define imrorx(h)	do{ const uint8 *s=f->cnv.s; const uint16 *t=f->cnv.t; \
	uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,j&=f->mx,o|=t[s[j]]<<v; while(++j,--i); \
	*d|=be32_native(o)&k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,j&=f->mx,v|=t[s[j]]<<k; while(++j,--i); \
			*d|=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,j&=f->mx,v|=t[s[j]]<<o; while(++j,--i); \
		v=be32_native(v), *d|=v&k; \
	} \
}while(0)

static DEF_BLT_RTN(mror1)
	{imrorx(0);}

static DEF_BLT_RTN(mror2)
	{imrorx(1);}

static DEF_BLT_RTN(mror4)
	{imrorx(2);}

static DEF_BLT_RTN(mror8)
	{imrorx(3);}

static DEF_BLT_RTN(mror16)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d&= t[s[j]], ++j, ++d, --c;
	o = c;
	if (c>>=1)
		do
		{
			j&= f->mx, v = t[s[j]]<<BIG_ENDIAN*16, ++j;
			j&= f->mx, v|= t[s[j]]<<!BIG_ENDIAN*16, ++j;
			*(int32*)d&= v;
		} while(d+=2,--c);
	if (o&1)
		j&= f->mx, *d&= t[s[j]];
}

static DEF_BLT_RTN(mror32)
{
	const int32	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, j = o-f->vx;

	if (c&1)
		j&= f->mx, *d&= t[s[j]], ++j, ++d;
	if (c>>=1)
		do
		{
			j&= f->mx, *d&= t[s[j]], ++j;
			j&= f->mx, d[1]&= t[s[j]], ++j;
		} while(d+=2,--c);
}

#define imcorx(h)	do{ uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,o|=f->cnv.rtn(&f->cnv,j)<<v; while(++j,--i); \
	*d|=be32_native(o)&k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,v|=f->cnv.rtn(&f->cnv,j)<<k; while(++j,--i); \
			*d|=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,v|=f->cnv.rtn(&f->cnv,j)<<o; while(++j,--i); \
		v=be32_native(v), *d|=v&k; \
	} \
}while(0)

static DEF_BLT_RTN(mcor1)
	{imcorx(0);}

static DEF_BLT_RTN(mcor2)
	{imcorx(1);}

static DEF_BLT_RTN(mcor4)
	{imcorx(2);}

static DEF_BLT_RTN(mcor8)
	{imcorx(3);}

static DEF_BLT_RTN(mcor16)
{
	int16		*d = (int16*)f->a+o;

	o-= f->vx;
	if (c&1)
		*d&= f->cnv.rtn(&f->cnv,o), ++o, ++d;
	if (c>>=1)
		do
		{
			*d&= f->cnv.rtn(&f->cnv,o), ++o;
			d[1]&= f->cnv.rtn(&f->cnv,o), ++o;
		} while(d+=2,--c);
}

static DEF_BLT_RTN(mcor32)
{
	int32		*d = (int32*)f->a+o;

	o-= f->vx;
	if (c&1)
		*d&= f->cnv.rtn(&f->cnv,o), ++o, ++d;
	if (c>>=1)
		do
		{
			*d&= f->cnv.rtn(&f->cnv,o), ++o;
			d[1]&= f->cnv.rtn(&f->cnv,o), ++o;
		} while(d+=2,--c);
}

static DEF_BLT_RTN(imor1)
	{rOrBits(f->a,f->cnv.s,o,o-f->vx,c,f->opr);}

static DEF_BLT_RTN(imor2)
	{rOrBits(f->a,f->cnv.s,o<<1,(o-f->vx)<<1,c<<1,f->opr);}

static DEF_BLT_RTN(imor4)
	{rOrBits(f->a,f->cnv.s,o<<2,(o-f->vx)<<2,c<<2,f->opr);}

static DEF_BLT_RTN(imor8)
	{rOrBits(f->a,f->cnv.s,o<<3,(o-f->vx)<<3,c<<3,f->opr);}

static DEF_BLT_RTN(imor16)
{
	const int16	*s = (const int16*)f->cnv.s+o-f->vx;
	int16		*d = (int16*)f->a+o;
	uint32		cm = f->opr;

	if (o&1)
		*d&= *s^(int16)cm, ++s, ++d, --c;
	o = c;
	if (c>1)
	{
		if (c&2)
			*(uint32*)d&= urw32(s)^cm, s+= 2, d+= 2;
		if (c&4)
			*(uint32*)d&= urw32(s)^cm, ((uint32*)d)[1]&= urw32(s+2)^cm, s+= 4, d+= 4;
		if (c>>=3)
			do
			{
				*(uint32*)d&= urw32(s)^cm, ((uint32*)d)[1]&= urw32(s+2)^cm;
				((uint32*)d)[2]&= urw32(s+4)^cm, ((uint32*)d)[3]&= urw32(s+6)^cm;
			} while(s+=8,d+=8,--c);
	}
	if (o&=1)
		*d&= *s^(int16)cm;
}

static DEF_BLT_RTN(imor32)
{
	const int32	*s = (const int32*)f->cnv.s+o-f->vx;
	int32		*d = (int32*)f->a+o, cm = f->opr;

	if (c&1)
		*d++&= *s^cm, ++s;
	if (c&2)
		*d&= *s^cm, d[1]&= s[1]^cm, d+= 2, s+= 2;
	if (c>>=2)
		do
			*d&= *s^cm, d[1]&= s[1]^cm, d[2]&= s[2]^cm, d[3]&= s[3]^cm;
		while(d+=4,s+=4,--c);
}

#define imporx(h)	do{ const uint8 *s=f->cnv.s; \
	uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,j&=f->mx,o|=s[j]<<v; while(++j,--i); \
	*d|=be32_native(o)&k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,j&=f->mx,v|=s[j]<<k; while(++j,--i); \
			*d|=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,j&=f->mx,v|=s[j]<<o; while(++j,--i); \
		v=be32_native(v), *d|=v&k; \
	} \
}while(0)

static DEF_BLT_RTN(mpor1)
	{imporx(0);}

static DEF_BLT_RTN(mpor2)
	{imporx(1);}

static DEF_BLT_RTN(mpor4)
	{imporx(2);}

static DEF_BLT_RTN(mpor8)
	{imporx(3);}

static DEF_BLT_RTN(mpor16)
{
	const uint16	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d&= s[j], ++j, ++d, --c;
	o = c;
	if (c>>=1)
		do
		{
			j&= f->mx, v = s[j]<<BIG_ENDIAN*16, ++j;
			j&= f->mx, v|= s[j]<<!BIG_ENDIAN*16, ++j;
			*(int32*)d&= v;
		} while(d+=2,--c);
	if (o&1)
		j&= f->mx, *d&= s[j];
}

static DEF_BLT_RTN(mpor32)
{
	const uint32	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, j = o-f->vx;

	if (c&1)
		j&= f->mx, *d&= s[j], ++j, ++d;
	if (c>>=1)
		do
		{
			j&= f->mx, *d&= s[j], ++j;
			j&= f->mx, d[1]&= s[j], ++j;
		} while(d+=2,--c);
}

/* Ex-OR */

#define sldxox(h)	do{ uint32 *d, k, v; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=be32_native(k); \
	v=f->sc; \
	*d^=v&k, c+=o; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) do *d++^=v; while(--c); \
	if(o&=31) k=~(~0u>>o), *d^=be32_native(k)&v; \
}while(0)

static DEF_BLT_RTN(slxo1)
	{sldxox(0);}

static DEF_BLT_RTN(slxo2)
	{sldxox(1);}

static DEF_BLT_RTN(slxo4)
	{sldxox(2);}

static DEF_BLT_RTN(slxo8)
{
	int8		*d = f->a+o;
	int32		v = f->sc;

	if (o&1)
		*d^= (int8)v, ++d, --c;
	if (o&c&2)
		*(int16*)d^= (int16)v, d+= 2, c-= 2;
	o = c;
	if (c>>=2)
		do
			*(int32*)d^= v;
		while(d+=4,--c);
	if (o&2)
		*(int16*)d^= (int16)v, d+= 2;
	if (o&=1)
		*d^= (int8)v;
}

static DEF_BLT_RTN(slxo16)
{
	int16		*d = (int16*)f->a+o;
	int32		v = f->sc;

	if (o&1)
		*d^= (int16)v, ++d, --c;
	o = c;
	if (c&2)
		*(int32*)d^= v, d+= 2;
	if (c>>=2)
		do
			*(int32*)d^= v, ((int32*)d)[1]^= v;
		while(d+=4,--c);
	if (o&=1)
		*d^= (int16)v;
}

static DEF_BLT_RTN(slxo32)
{
	int32		*d = (int32*)f->a+o, v = f->sc;

	if (c&1)
		*d++^= v;
	if (c&2)
		*d^= v, d[1]^= v, d+= 2;
	if (c>>=2)
		do
			*d^= v, d[1]^= v, d[2]^= v, d[3]^= v;
		while(d+=4,--c);
}

/* image Ex-OR */

#define imrxox(h)	do{ const uint8 *s=f->cnv.s; const uint16 *t=f->cnv.t; \
	uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,j&=f->mx,o|=t[s[j]]<<v; while(++j,--i); \
	*d^=be32_native(o)&k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,j&=f->mx,v|=t[s[j]]<<k; while(++j,--i); \
			*d^=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,j&=f->mx,v|=t[s[j]]<<o; while(++j,--i); \
		v=be32_native(v), *d^=v&k; \
	} \
}while(0)

static DEF_BLT_RTN(mrxo1)
	{imrxox(0);}

static DEF_BLT_RTN(mrxo2)
	{imrxox(1);}

static DEF_BLT_RTN(mrxo4)
	{imrxox(2);}

static DEF_BLT_RTN(mrxo8)
	{imrxox(3);}

static DEF_BLT_RTN(mrxo16)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d^= t[s[j]], ++j, ++d, --c;
	o = c;
	if (c>>=1)
		do
		{
			j&= f->mx, v = t[s[j]]<<BIG_ENDIAN*16, ++j;
			j&= f->mx, v|= t[s[j]]<<!BIG_ENDIAN*16, ++j;
			*(int32*)d^= v;
		} while(d+=2,--c);
	if (o&1)
		j&= f->mx, *d^= t[s[j]];
}

static DEF_BLT_RTN(mrxo32)
{
	const int32	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, j = o-f->vx;

	if (c&1)
		j&= f->mx, *d^= t[s[j]], ++j, ++d;
	if (c>>=1)
		do
		{
			j&= f->mx, *d^= t[s[j]], ++j;
			j&= f->mx, d[1]^= t[s[j]], ++j;
		} while(d+=2,--c);
}

#define imcxox(h)	do{ uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,o|=f->cnv.rtn(&f->cnv,j)<<v; while(++j,--i); \
	*d^=be32_native(o)&k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,v|=f->cnv.rtn(&f->cnv,j)<<k; while(++j,--i); \
			*d^=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,v|=f->cnv.rtn(&f->cnv,j)<<o; while(++j,--i); \
		v=be32_native(v), *d^=v&k; \
	} \
}while(0)

static DEF_BLT_RTN(mcxo1)
	{imcxox(0);}

static DEF_BLT_RTN(mcxo2)
	{imcxox(1);}

static DEF_BLT_RTN(mcxo4)
	{imcxox(2);}

static DEF_BLT_RTN(mcxo8)
	{imcxox(3);}

static DEF_BLT_RTN(mcxo16)
{
	int16		*d = (int16*)f->a+o;

	o-= f->vx;
	if (c&1)
		*d^= f->cnv.rtn(&f->cnv,o), ++o, ++d;
	if (c>>=1)
		do
		{
			*d^= f->cnv.rtn(&f->cnv,o), ++o;
			d[1]^= f->cnv.rtn(&f->cnv,o), ++o;
		} while(d+=2,--c);
}

static DEF_BLT_RTN(mcxo32)
{
	int32		*d = (int32*)f->a+o;

	o-= f->vx;
	if (c&1)
		*d^= f->cnv.rtn(&f->cnv,o), ++o, ++d;
	if (c>>=1)
		do
		{
			*d^= f->cnv.rtn(&f->cnv,o), ++o;
			d[1]^= f->cnv.rtn(&f->cnv,o), ++o;
		} while(d+=2,--c);
}

#define impxox(h)	do{ const uint8 *s=f->cnv.s; \
	uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,j&=f->mx,o|=s[j]<<v; while(++j,--i); \
	*d^=be32_native(o)&k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,j&=f->mx,v|=s[j]<<k; while(++j,--i); \
			*d^=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,j&=f->mx,v|=s[j]<<o; while(++j,--i); \
		v=be32_native(v), *d^=v&k; \
	} \
}while(0)

static DEF_BLT_RTN(mpxo1)
	{impxox(0);}

static DEF_BLT_RTN(mpxo2)
	{impxox(1);}

static DEF_BLT_RTN(mpxo4)
	{impxox(2);}

static DEF_BLT_RTN(mpxo8)
	{impxox(3);}

static DEF_BLT_RTN(mpxo16)
{
	const uint16	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d^= s[j], ++j, ++d, --c;
	o = c;
	if (c>>=1)
		do
		{
			j&= f->mx, v = s[j]<<BIG_ENDIAN*16, ++j;
			j&= f->mx, v|= s[j]<<!BIG_ENDIAN*16, ++j;
			*(int32*)d^= v;
		} while(d+=2,--c);
	if (o&1)
		j&= f->mx, *d^= s[j];
}

static DEF_BLT_RTN(mpxo32)
{
	const uint32	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, j = o-f->vx;

	if (c&1)
		j&= f->mx, *d^= s[j], ++j, ++d;
	if (c>>=1)
		do
		{
			j&= f->mx, *d^= s[j], ++j;
			j&= f->mx, d[1]^= s[j], ++j;
		} while(d+=2,--c);
}

/* image Ex-Or - direct */

static DEF_BLT_RTN(imxo1)
	{rXorBits(f->a,f->cnv.s,o,o-f->vx,c,f->opr);}

static DEF_BLT_RTN(imxo2)
	{rXorBits(f->a,f->cnv.s,o<<1,(o-f->vx)<<1,c<<1,f->opr);}

static DEF_BLT_RTN(imxo4)
	{rXorBits(f->a,f->cnv.s,o<<2,(o-f->vx)<<2,c<<2,f->opr);}

static DEF_BLT_RTN(imxo8)
	{rXorBits(f->a,f->cnv.s,o<<3,(o-f->vx)<<3,c<<3,f->opr);}

static DEF_BLT_RTN(imxo16)
{
	const int16	*s = (const int16*)f->cnv.s+o-f->vx;
	int16		*d = (int16*)f->a+o;
	uint32		cm = f->opr;

	if (o&1)
		*d^= *s^(int16)cm, ++s, ++d, --c;
	o = c;
	if (c>1)
	{
		if (c&2)
			*(uint32*)d^= urw32(s)^cm, s+= 2, d+= 2;
		if (c&4)
			*(uint32*)d^= urw32(s)^cm, ((uint32*)d)[1]^= urw32(s+2)^cm, s+= 4, d+= 4;
		if (c>>=3)
			do
			{
				*(uint32*)d^= urw32(s)^cm, ((uint32*)d)[1]^= urw32(s+2)^cm;
				((uint32*)d)[2]^= urw32(s+4)^cm, ((uint32*)d)[3]^= urw32(s+6)^cm;
			} while(s+=8,d+=8,--c);
	}
	if (o&=1)
		*d^= *s^(int16)cm;
}

static DEF_BLT_RTN(imxo32)
{
	const int32	*s = (const int32*)f->cnv.s+o-f->vx;
	int32		*d = (int32*)f->a+o, cm = f->opr;

	if (c&1)
		*d++^= *s^cm, ++s;
	if (c&2)
		*d^= *s^cm, d[1]^= s[1]^cm, d+= 2, s+= 2;
	if (c>>=2)
		do
			*d^= *s^cm, d[1]^= s[1]^cm, d[2]^= s[2]^cm, d[3]^= s[3]^cm;
		while(d+=4,s+=4,--c);
}

/* Bic */

#define sldbcx(h)	do{ uint32 *d, k, v; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=~be32_native(k); \
	v=f->sc; \
	*d&=v|k, c+=o; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) do *d++&=v; while(--c); \
	if(o&=31) k=~0u>>o, *d&=be32_native(k)|v; \
}while(0)

static DEF_BLT_RTN(slbc1)
	{sldbcx(0);}

static DEF_BLT_RTN(slbc2)
	{sldbcx(1);}

static DEF_BLT_RTN(slbc4)
	{sldbcx(2);}

static DEF_BLT_RTN(slbc8)
{
	int32		v = f->sc;
	int8		*d = f->a+o;

	if (o&1)
		*d&= (int8)v, ++d, --c;
	if (o&c&2)
		*(int16*)d&= (int16)v, d+= 2, c-= 2;
	o = c;
	if (c>>=2)
		do
			*(int32*)d&= v;
		while(d+=4,--c);
	if (o&2)
		*(int16*)d&= (int16)v, d+= 2;
	if (o&=1)
		*d&= (int8)v;
}

static DEF_BLT_RTN(slbc16)
{
	int32		v = f->sc;
	int16		*d = (int16*)f->a+o;

	if (o&1)
		*d|= (int16)v, ++d, --c;
	o = c;
	if (c>1)
	{
		if (c&2)
			*(int32*)d|= v, d+= 2;
		if (c>>=2)
			do
				*(int32*)d|= v, ((int32*)d)[1]|= v;
			while(d+=4,--c);
	}
	if (o&=1)
		*d|= (int16)v;
}

static DEF_BLT_RTN(slbc32)
{
	int32		*d = (int32*)f->a+o, v = f->sc;

	if (c&1)
		*d++|= v;
	if (c&2)
		*d|= v, d[1]|= v, d+= 2;
	if (c>>=2)
		do
			*d|= v, d[1]|= v, d[2]|= v, d[3]|= v;
		while(d+=4,--c);
}

/* image Bic */

#define imrbcx(h)	do{ const uint8 *s=f->cnv.s; const uint16 *t=f->cnv.t; \
	uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=~be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,j&=f->mx,o|=t[s[j]]<<v; while(++j,--i); \
	*d&=be32_native(o)|k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,j&=f->mx,v|=t[s[j]]<<k; while(++j,--i); \
			*d&=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=~be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,j&=f->mx,v|=t[s[j]]<<o; while(++j,--i); \
		v=be32_native(v), *d&=v|k; \
	} \
}while(0)

static DEF_BLT_RTN(mrbc1)
	{imrbcx(0);}

static DEF_BLT_RTN(mrbc2)
	{imrbcx(1);}

static DEF_BLT_RTN(mrbc4)
	{imrbcx(2);}

static DEF_BLT_RTN(mrbc8)
	{imrbcx(3);}

static DEF_BLT_RTN(mrbc16)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d|= t[s[j]], ++j, ++d, --c;
	o = c;
	if (c>>=1)
		do
		{
			j&= f->mx, v = t[s[j]]<<BIG_ENDIAN*16, ++j;
			j&= f->mx, v|= t[s[j]]<<!BIG_ENDIAN*16, ++j;
			*(int32*)d|= v;
		} while(d+=2,--c);
	if (o&1)
		j&= f->mx, *d|= t[s[j]];
}

static DEF_BLT_RTN(mrbc32)
{
	const int32	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, j = o-f->vx;

	if (c&1)
		j&= f->mx, *d|= t[s[j]], ++j, ++d;
	if (c>>=1)
		do
		{
			j&= f->mx, *d|= t[s[j]], ++j;
			j&= f->mx, d[1]|= t[s[j]], ++j;
		} while(d+=2,--c);
}

#define imcbcx(h)	do{ uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=~be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,o|=f->cnv.rtn(&f->cnv,j)<<v; while(++j,--i); \
	*d&=be32_native(o)|k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,v|=f->cnv.rtn(&f->cnv,j)<<k; while(++j,--i); \
			*d&=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=~be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,v|=f->cnv.rtn(&f->cnv,j)<<o; while(++j,--i); \
		v=be32_native(v), *d&=v|k; \
	} \
}while(0)

static DEF_BLT_RTN(mcbc1)
	{imcbcx(0);}

static DEF_BLT_RTN(mcbc2)
	{imcbcx(1);}

static DEF_BLT_RTN(mcbc4)
	{imcbcx(2);}

static DEF_BLT_RTN(mcbc8)
	{imcbcx(3);}

static DEF_BLT_RTN(mcbc16)
{
	int16		*d = (int16*)f->a+o;

	o-= f->vx;
	if (c&1)
		*d|= f->cnv.rtn(&f->cnv,o), ++o, ++d;
	if (c>>=1)
		do
		{
			*d|= f->cnv.rtn(&f->cnv,o), ++o;
			d[1]|= f->cnv.rtn(&f->cnv,o), ++o;
		} while(d+=2,--c);
}

static DEF_BLT_RTN(mcbc32)
{
	int32		*d = (int32*)f->a+o;

	o-= f->vx;
	if (c&1)
		*d|= f->cnv.rtn(&f->cnv,o), ++o, ++d;
	if (c>>=1)
		do
		{
			*d|= f->cnv.rtn(&f->cnv,o), ++o;
			d[1]|= f->cnv.rtn(&f->cnv,o), ++o;
		} while(d+=2,--c);
}

#define impbcx(h)	do{ const uint8 *s=f->cnv.s; \
	uint32 *d, k, v, i, j=o-f->vx; \
	o<<=h, d=(uint32*)f->a+(o>>5), o&=31, c<<=h; \
	k=(uint32)((int)0x80000000>>(MIN(c,32)-1))>>o,k=~be32_native(k); \
	v=32-o, i=MIN((uint32)c,v)>>h, c+=o; \
	o=0; do v-=1u<<h,j&=f->mx,o|=s[j]<<v; while(++j,--i); \
	*d&=be32_native(o)|k; \
	if((c-=32)<=0) return; \
	++d, o=c; \
	if(c>>=5) \
		do { i=32>>h, k=32, v=0; do k-=1<<h,j&=f->mx,v|=s[j]<<k; while(++j,--i); \
			*d&=be32_native(v); \
		} while(++d,--c); \
	if(o&=31) { k=~(~0u>>o), k=~be32_native(k), \
		i=o>>h, o=32, v=0; do o-=1<<h,j&=f->mx,v|=s[j]<<o; while(++j,--i); \
		v=be32_native(v), *d&=v|k; \
	} \
}while(0)

static DEF_BLT_RTN(mpbc1)
	{impbcx(0);}

static DEF_BLT_RTN(mpbc2)
	{impbcx(1);}

static DEF_BLT_RTN(mpbc4)
	{impbcx(2);}

static DEF_BLT_RTN(mpbc8)
	{impbcx(3);}

static DEF_BLT_RTN(mpbc16)
{
	const uint16	*s = f->cnv.s;
	uint16		*d = (uint16*)f->a+o;
	int32		v, j = o-f->vx;

	if (o&1)
		j&= f->mx, *d|= s[j], ++j, ++d, --c;
	o = c;
	if (c>>=1)
		do
		{
			j&= f->mx, v = s[j]<<BIG_ENDIAN*16, ++j;
			j&= f->mx, v|= s[j]<<!BIG_ENDIAN*16, ++j;
			*(int32*)d|= v;
		} while(d+=2,--c);
	if (o&1)
		j&= f->mx, *d|= s[j];
}

static DEF_BLT_RTN(mpbc32)
{
	const uint32	*s = f->cnv.s;
	int32		*d = (int32*)f->a+o, j = o-f->vx;

	if (c&1)
		j&= f->mx, *d|= s[j], ++j, ++d;
	if (c>>=1)
		do
		{
			j&= f->mx, *d|= s[j], ++j;
			j&= f->mx, d[1]|= s[j], ++j;
		} while(d+=2,--c);
}

/* image Bic - direct */

static DEF_BLT_RTN(imbc1)
	{rAndBits(f->a,f->cnv.s,o,o-f->vx,c,f->opr);}

static DEF_BLT_RTN(imbc2)
	{rAndBits(f->a,f->cnv.s,o<<1,(o-f->vx)<<1,c<<1,f->opr);}

static DEF_BLT_RTN(imbc4)
	{rAndBits(f->a,f->cnv.s,o<<2,(o-f->vx)<<2,c<<2,f->opr);}

static DEF_BLT_RTN(imbc8)
	{rAndBits(f->a,f->cnv.s,o<<3,(o-f->vx)<<3,c<<3,f->opr);}

static DEF_BLT_RTN(imbc16)
{
	const int16	*s = (const int16*)f->cnv.s+o-f->vx;
	int16		*d = (int16*)f->a+o;
	uint32		cm = f->opr;

	if (o&1)
		*d|= *s^(int16)cm, ++s, ++d, --c;
	o = c;
	if (c>1)
	{
		if (c&2)
			*(uint32*)d|= urw32(s)^cm, s+= 2, d+= 2;
		if (c&4)
			*(uint32*)d|= urw32(s)^cm, ((uint32*)d)[1]|= urw32(s+2)^cm, s+= 4, d+= 4;
		if (c>>=3)
			do
			{
				*(uint32*)d|= urw32(s)^cm, ((uint32*)d)[1]|= urw32(s+2)^cm;
				((uint32*)d)[2]|= urw32(s+4)^cm, ((uint32*)d)[3]|= urw32(s+6)^cm;
			} while(s+=8,d+=8,--c);
	}
	if (o&=1)
		*d|= *s^(int16)cm;
}

static DEF_BLT_RTN(imbc32)
{
	const int32	*s = (const int32*)f->cnv.s+o-f->vx;
	int32		*d = (int32*)f->a+o, cm = f->opr;

	if (c&1)
		*d++|= *s^cm, ++s;
	if (c&2)
		*d|= *s^cm, d[1]|= s[1]^cm, d+= 2, s+= 2;
	if (c>>=2)
		do
			*d|= *s^cm, d[1]|= s[1]^cm, d[2]|= s[2]^cm, d[3]|= s[3]^cm;
		while(d+=4,s+=4,--c);
}

/* solid blend */

static DEF_BLT_RTN(slb8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	uint32		v = f->sc, a, b, u, mr = f->bmr, mg = f->bmg, mb = f->bmb;
	int32		kr = f->opr, kg = f->opg, kb = f->opb;

	do
		a = dt[*d], *d = it[rgblerp16(a,v,b,u,mr,mg,mb,kr,kg,kb)];
	while(++d,--c);
}

static DEF_BLT_RTN(slb8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		m = f->bmg, v = f->sc, a;

	do
		a = dt[*d], *d = it[(a+((v-a)*m>>blndPrec))>>6];
	while(++d,--c);
}

#define slbx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		v=f->sc, a, b, u, mr=f->bmr, mg=f->bmg, mb=f->bmb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, *d=

slbx(slb16,uint16) rgblerp16(a,v,b,u,mr,mg,mb,kr,kg,kb); while(++d,--c);}
slbx(slb32,int32) rgblerp32(a,v,b,u,mr,mg,mb,kr,kg,kb); while(++d,--c);}

/* image blend - remapped */

static DEF_BLT_RTN(xrb8c)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, b, u, v, mr = f->bmr, mg = f->bmg, mb = f->bmb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = t[s[o]], *d = it[rgblerp16(a,v,b,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(xrb8g)
{
	const uint16	*t = f->cnv.t;
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint8	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;
	int		m = f->bmg, a;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, *d = it[(a+((t[s[o]]-a)*m>>blndPrec))>>6];
	while(++o,++d,--c);
}

#define xrbx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*t=f->cnv.t; \
	const uint8	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, b, u, v, mr=f->bmr, mg=f->bmg, mb=f->bmb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=t[s[o]], *d=

xrbx(xrb16,uint16) rgblerp16(a,v,b,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c);}
xrbx(xrb32,int32) rgblerp32(a,v,b,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c);}

/* image blend - converted */

static DEF_BLT_RTN(xcb8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, b, u, v, mr = f->bmr, mg = f->bmg, mb = f->bmb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], v = f->cnv.rtn(&f->cnv,o), *d = it[rgblerp16(a,v,b,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(xcb8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		m = f->bmg, a;

	o-= f->vx;
	do
		a = dt[*d], *d = it[(a+(((signed)f->cnv.rtn(&f->cnv,o)-a)*m>>blndPrec))>>6];
	while(++o,++d,--c);
}

#define xcbx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, b, u, v, mr=f->bmr, mg=f->bmg, mb=f->bmb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, v=f->cnv.rtn(&f->cnv,o), *d=

xcbx(xcb16,uint16) rgblerp16(a,v,b,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c);}
xcbx(xcb32,int32) rgblerp32(a,v,b,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c);}

/* image blend - pattern */

static DEF_BLT_RTN(xpb8c)
{
	const uint16	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, b, u, v, mr = f->bmr, mg = f->bmg, mb = f->bmb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = s[o], *d = it[rgblerp16(a,v,b,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(xpb8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint16	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;
	int		m = f->bmg, a;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, *d = it[(a+((s[o]-a)*m>>blndPrec))>>6];
	while(++o,++d,--c);
}

#define xpbx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, b, u, v, mr=f->bmr, mg=f->bmg, mb=f->bmb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=s[o], *d=

xpbx(xpb16,uint16) rgblerp16(a,v,b,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c);}
xpbx(xpb32,int32) rgblerp32(a,v,b,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c);}

/* image blend - direct */

#define imbx(n,t) \
static DEF_BLT_RTN(n) { \
	const t		*s=(const t*)f->cnv.s+o-f->vx; \
	t		*d=(t*)f->a+o, a, r, u; \
	int32		mr=f->bmr, mg=f->bmg, mb=f->bmb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, *d=

imbx(imb16,uint16) rgblerp16(a,*s,r,u,mr,mg,mb,kr,kg,kb); while(++s,++d,--c);}
imbx(imb32,int32) rgblerp32(a,*s,r,u,mr,mg,mb,kr,kg,kb); while(++s,++d,--c);}

/* solid adMax */

static DEF_BLT_RTN(smx8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v = f->sc, u, kr = f->opr, kg = f->opg, kb = f->opb;

	do
		a = dt[*d], *d = it[rgbadmax(a,v,r,u,kr,kg,kb)];
	while(++d,--c);
}

static DEF_BLT_RTN(smx8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		v = f->sc;

	do
		*d = it[MAX(dt[*d],v)>>6&1023];
	while(++d,--c);
}

#define smxx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v=f->sc, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, *d=rgbadmax(a,v,r,u,kr,kg,kb); while(++d,--c); \
}

smxx(smx16,uint16)
smxx(smx32,int32)

/* image adMax - remapped */

static DEF_BLT_RTN(rmx8c)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = t[s[o]], *d = it[rgbadmax(a,v,r,u,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(rmx8g)
{
	const uint16	*t = f->cnv.t;
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint8	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;

	o-= f->vx;
	do
		o&= f->mx, *d = it[MAX(dt[*d],t[s[o]])>>6&1023];
	while(++o,++d,--c);
}

#define rmxx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*t=f->cnv.t; \
	const uint8	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=t[s[o]], *d=rgbadmax(a,v,r,u,kr,kg,kb); while(++o,++d,--c); \
}

rmxx(rmx16,uint16)
rmxx(rmx32,int32)

/* image adMax - converted */

static DEF_BLT_RTN(cmx8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], v = f->cnv.rtn(&f->cnv,o), *d = it[rgbadmax(a,v,r,u,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(cmx8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	uint16		v;

	o-= f->vx;
	do
		v = f->cnv.rtn(&f->cnv,o), *d = it[MAX(dt[*d],v)>>6&1023];
	while(++o,++d,--c);
}

#define cmxx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, v=f->cnv.rtn(&f->cnv,o), *d=rgbadmax(a,v,r,u,kr,kg,kb); while(++o,++d,--c); \
}

cmxx(cmx16,uint16)
cmxx(cmx32,int32)

/* image adMax - direct */

#define xmxx(n,t) \
static DEF_BLT_RTN(n) { \
	const t		*s=(const t*)f->cnv.s+o-f->vx; \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, v=*s, *d=rgbadmax(a,v,r,u,kr,kg,kb); while(++s,++d,--c); \
}

xmxx(xmx16,uint16)
xmxx(xmx32,int32)

/* image adMax - pattern */

static DEF_BLT_RTN(pmx8c)
{
	const uint16	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = s[o], *d = it[rgbadmax(a,v,r,u,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(pmx8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint16	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;

	o-= f->vx;
	do
		o&= f->mx, *d = it[MAX(dt[*d],s[o])>>6&1023];
	while(++o,++d,--c);
}

#define pmxx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=s[o], *d=rgbadmax(a,v,r,u,kr,kg,kb); while(++o,++d,--c); \
}

pmxx(pmx16,uint16)
pmxx(pmx32,int32)

/* solid adMin */

static DEF_BLT_RTN(smn8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v = f->sc, u, kr = f->opr, kg = f->opg, kb = f->opb;

	do
		a = dt[*d], *d = it[rgbadmin(a,v,r,u,kr,kg,kb)];
	while(++d,--c);
}

static DEF_BLT_RTN(smn8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		v = f->sc;

	do
		*d = it[MIN(dt[*d],v)>>6&1023];
	while(++d,--c);
}

#define smnx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v=f->sc, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, *d=rgbadmin(a,v,r,u,kr,kg,kb); while(++d,--c); \
}

smnx(smn16,uint16)
smnx(smn32,int32)

/* image adMin - remapped */

static DEF_BLT_RTN(rmn8c)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = t[s[o]], *d = it[rgbadmin(a,v,r,u,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(rmn8g)
{
	const uint16	*t = f->cnv.t;
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint8	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;
	uint16		v;

	o-= f->vx;
	do
		o&= f->mx, v = t[s[o]], *d = it[MIN(dt[*d],v)>>6&1023];
	while(++o,++d,--c);
}

#define rmnx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*t = f->cnv.t; \
	const uint8	*s = f->cnv.s; \
	typ		*d = (typ*)f->a+o; \
	int32		a, r, v, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=t[s[o]], *d=rgbadmin(a,v,r,u,kr,kg,kb); while(++o,++d,--c); \
}

rmnx(rmn16,uint16)
rmnx(rmn32,int32)

/* image adMin - converted */

static DEF_BLT_RTN(cmn8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], v = f->cnv.rtn(&f->cnv,o), *d = it[rgbadmin(a,v,r,u,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(cmn8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	uint16		v;

	o-= f->vx;
	do
		v = f->cnv.rtn(&f->cnv,o), *d = it[MIN(dt[*d],v)>>6&1023];
	while(++o,++d,--c);
}

#define cmnx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, v=f->cnv.rtn(&f->cnv,o), *d=rgbadmin(a,v,r,u,kr,kg,kb); while(++o,++d,--c); \
}

cmnx(cmn16,uint16)
cmnx(cmn32,int32)

/* image adMin - direct */

#define xmnx(n,t) \
static DEF_BLT_RTN(n) { \
	const t		*s=(const t*)f->cnv.s+o-f->vx; \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, v=*s, *d=rgbadmin(a,v,r,u,kr,kg,kb); while(++s,++d,--c); \
}

xmnx(xmn16,uint16)
xmnx(xmn32,int32)

/* image adMin - pattern */

static DEF_BLT_RTN(pmn8c)
{
	const uint16	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = s[o], *d = it[rgbadmin(a,v,r,u,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(pmn8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint16	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;
	uint16		v;

	o-= f->vx;
	do
		o&= f->mx, v = s[o], *d = it[MIN(dt[*d],v)>>6&1023];
	while(++o,++d,--c);
}

#define pmnx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*s = f->cnv.s; \
	typ		*d = (typ*)f->a+o; \
	int32		a, r, v, u, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=s[o], *d=rgbadmin(a,v,r,u,kr,kg,kb); while(++o,++d,--c); \
}

pmnx(pmn16,uint16)
pmnx(pmn32,int32)

/* solid addPin */

static DEF_BLT_RTN(sap8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v = f->sc, u = *(int32*)&f->bmg;
	int		mr = u&f->opr, mg = u&f->opg, mb = u&f->opb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	do
		a = dt[*d], *d = it[rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb)];
	while(++d,--c);
}

static DEF_BLT_RTN(sap8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		v = f->sc, a, m = f->bmg;

	do
		a = dt[*d]+v, *d = it[MIN(a,m)>>6];
	while(++d,--c);
}

#define sapx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v=f->sc, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, *d=rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++d,--c); \
}

sapx(sap16,uint16)
sapx(sap32,int32)

/* image addPin - remapped */

static DEF_BLT_RTN(rap8c)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u = *(int32*)&f->bmg;
	int		mr = u&f->opr, mg = u&f->opg, mb = u&f->opb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = t[s[o]], *d = it[rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(rap8g)
{
	const uint16	*t = f->cnv.t;
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint8	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;
	int		a, m = f->bmg;

	o-= f->vx;
	do
		o&= f->mx, a = dt[*d]+t[s[o]], *d = it[MIN(a,m)>>6];
	while(++o,++d,--c);
}

#define rapx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*t=f->cnv.t; \
	const uint8	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=t[s[o]], *d=rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c); \
}

rapx(rap16,uint16)
rapx(rap32,int32)

/* image addPin - converted */

static DEF_BLT_RTN(cap8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u = *(int32*)&f->bmg;
	int		mr = u&f->opr, mg = u&f->opg, mb = u&f->opb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], v = f->cnv.rtn(&f->cnv,o), *d = it[rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(cap8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		a, m = f->bmg;

	o-= f->vx;
	do
		a = dt[*d]+f->cnv.rtn(&f->cnv,o), *d = it[MIN(a,m)>>6];
	while(++o,++d,--c);
}

#define capx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, v=f->cnv.rtn(&f->cnv,o), *d=rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c); \
}

capx(cap16,uint16)
capx(cap32,int32)

/* image addPin - direct */

#define xapx(n,t) \
static DEF_BLT_RTN(n) { \
	const t		*s=(const t*)f->cnv.s+o-f->vx; \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, v=*s, *d=rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++s,++d,--c); \
}

xapx(xap16,uint16)
xapx(xap32,int32)

/* image addPin - pattern */

static DEF_BLT_RTN(pap8c)
{
	const uint16	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u = *(int32*)&f->bmg;
	int		mr = u&f->opr, mg = u&f->opg, mb = u&f->opb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = s[o], *d = it[rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(pap8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint16	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;
	int		a, m = f->bmg;

	o-= f->vx;
	do
		o&= f->mx, a = dt[*d]+s[o], *d = it[MIN(a,m)>>6];
	while(++o,++d,--c);
}

#define papx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=s[o], *d=rgbaddpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c); \
}

papx(pap16,uint16)
papx(pap32,int32)

/* solid addOver */

static DEF_BLT_RTN(sao8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v = f->sc, kr = f->opr, kg = f->opg, kb = f->opb;

	do
		a = dt[*d], *d = it[rgbaddover(a,v,r,kr,kg,kb)];
	while(++d,--c);
}

static DEF_BLT_RTN(sao8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		v = f->sc;

	do
		*d = it[(dt[*d]+v)>>6&1023];
	while(++d,--c);
}

#define saox(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v=f->sc, kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, *d=rgbaddover(a,v,r,kr,kg,kb); while(++d,--c); \
}

saox(sao16,uint16)
saox(sao32,int32)

/* image addOver - remapped */

static DEF_BLT_RTN(rao8c)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = t[s[o]], *d = it[rgbaddover(a,v,r,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(rao8g)
{
	const uint16	*t = f->cnv.t;
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint8	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;

	o-= f->vx;
	do
		o&= f->mx, *d = it[(dt[*d]+t[s[o]])>>6&1023];
	while(++o,++d,--c);
}

#define raox(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*t=f->cnv.t; \
	const uint8	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=t[s[o]], *d=rgbaddover(a,v,r,kr,kg,kb); while(++o,++d,--c); \
}

raox(rao16,uint16)
raox(rao32,int32)

/* image addOver - converted */

static DEF_BLT_RTN(cao8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], v = f->cnv.rtn(&f->cnv,o), *d = it[rgbaddover(a,v,r,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(cao8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;

	o-= f->vx;
	do
		*d = it[(dt[*d]+f->cnv.rtn(&f->cnv,o))>>6&1023];
	while(++o,++d,--c);
}

#define caox(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, v=f->cnv.rtn(&f->cnv,o), *d=rgbaddover(a,v,r,kr,kg,kb); while(++o,++d,--c); \
}

caox(cao16,uint16)
caox(cao32,int32)

/* image addOver - direct */

#define xaox(n,t) \
static DEF_BLT_RTN(n) { \
	const t		*s=(const t*)f->cnv.s+o-f->vx; \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, v=*s, *d=rgbaddover(a,v,r,kr,kg,kb); while(++s,++d,--c); \
}

xaox(xao16,uint16)
xaox(xao32,int32)

/* image addOver - pattern */

static DEF_BLT_RTN(pao8c)
{
	const uint16	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = s[o], *d = it[rgbaddover(a,v,r,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(pao8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint16	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;

	o-= f->vx;
	do
		o&= f->mx, *d = it[(dt[*d]+s[o])>>6&1023];
	while(++o,++d,--c);
}

#define paox(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=s[o], *d=rgbaddover(a,v,r,kr,kg,kb); while(++o,++d,--c); \
}

paox(pao16,uint16)
paox(pao32,int32)

/* solid subPin */

static DEF_BLT_RTN(ssp8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v = f->sc, u = *(int32*)&f->bmg;
	int		mr = u&f->opr, mg = u&f->opg, mb = u&f->opb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	do
		a = dt[*d], *d = it[rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb)];
	while(++d,--c);
}

static DEF_BLT_RTN(ssp8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		v = f->sc, a, m = f->bmg;

	do
		a = dt[*d]-v, *d = it[MAX(a,m)>>6];
	while(++d,--c);
}

#define sspx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v=f->sc, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, *d=rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++d,--c); \
}

sspx(ssp16,uint16)
sspx(ssp32,int32)

/* image subPin - remapped */

static DEF_BLT_RTN(rsp8c)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u = *(int32*)&f->bmg;
	int		mr = u&f->opr, mg = u&f->opg, mb = u&f->opb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = t[s[o]], *d = it[rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(rsp8g)
{
	const uint16	*t = f->cnv.t;
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint8	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;
	int		a, m = f->bmg;

	o-= f->vx;
	do
		o&= f->mx, a = dt[*d]-t[s[o]], *d = it[MAX(a,m)>>6];
	while(++o,++d,--c);
}

#define rspx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*t=f->cnv.t; \
	const uint8	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=t[s[o]], *d=rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c); \
}

rspx(rsp16,uint16)
rspx(rsp32,int32)

/* image subPin - converted */

static DEF_BLT_RTN(csp8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u = *(int32*)&f->bmg;
	int		mr = u&f->opr, mg = u&f->opg, mb = u&f->opb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], v = f->cnv.rtn(&f->cnv,o), *d = it[rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(csp8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		a, m = f->bmg;

	o-= f->vx;
	do
		a = dt[*d]-f->cnv.rtn(&f->cnv,o), *d = it[MAX(a,m)>>6];
	while(++o,++d,--c);
}

#define cspx(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, v=f->cnv.rtn(&f->cnv,o), *d=rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c); \
}

cspx(csp16,uint16)
cspx(csp32,int32)

/* image subPin - direct */

#define xspx(n,t) \
static DEF_BLT_RTN(n) { \
	const t		*s=(const t*)f->cnv.s+o-f->vx; \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, v=*s, *d=rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++s,++d,--c); \
}

xspx(xsp16,uint16)
xspx(xsp32,int32)

/* image subPin - pattern */

static DEF_BLT_RTN(psp8c)
{
	const uint16	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, u = *(int32*)&f->bmg;
	int		mr = u&f->opr, mg = u&f->opg, mb = u&f->opb;
	int		kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = s[o], *d = it[rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(psp8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint16	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;
	int		a, m = f->bmg;

	o-= f->vx;
	do
		o&= f->mx, a = dt[*d]-s[o], *d = it[MAX(a,m)>>6];
	while(++o,++d,--c);
}

#define pspx(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, u=*(int32*)&f->bmg; \
	int32		mr=u&f->opr, mg=u&f->opg, mb=u&f->opb; \
	int32		kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=s[o], *d=rgbsubpin(a,v,r,u,mr,mg,mb,kr,kg,kb); while(++o,++d,--c); \
}

pspx(psp16,uint16)
pspx(psp32,int32)

/* solid subOver */

static DEF_BLT_RTN(sso8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v = f->sc, kr = f->opr, kg = f->opg, kb = f->opb;

	do
		a = dt[*d], *d = it[rgbsubover(a,v,r,kr,kg,kb)];
	while(++d,--c);
}

static DEF_BLT_RTN(sso8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;
	int		v = f->sc;

	do
		*d = it[(dt[*d]-v)>>6&1023];
	while(++d,--c);
}

#define ssox(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v=f->sc, kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, *d=rgbsubover(a,v,r,kr,kg,kb); while(++d,--c); \
}

ssox(sso16,uint16)
ssox(sso32,int32)

/* image subOver - remapped */

static DEF_BLT_RTN(rso8c)
{
	const uint16	*t = f->cnv.t;
	const uint8	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = t[s[o]], *d = it[rgbsubover(a,v,r,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(rso8g)
{
	const uint16	*t = f->cnv.t;
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint8	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;

	o-= f->vx;
	do
		o&= f->mx, *d = it[(dt[*d]-t[s[o]])>>6&1023];
	while(++o,++d,--c);
}

#define rsox(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*t=f->cnv.t; \
	const uint8	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=t[s[o]], *d=rgbsubover(a,v,r,kr,kg,kb); while(++o,++d,--c); \
}

rsox(rso16,uint16)
rsox(rso32,int32)

/* image subOver - converted */

static DEF_BLT_RTN(cso8c)
{
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], v = f->cnv.rtn(&f->cnv,o), *d = it[rgbsubover(a,v,r,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(cso8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	uint8		*d = (uint8*)f->a+o;

	o-= f->vx;
	do
		*d = it[(dt[*d]-f->cnv.rtn(&f->cnv,o))>>6&1023];
	while(++o,++d,--c);
}

#define csox(n,t) \
static DEF_BLT_RTN(n) { \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, v=f->cnv.rtn(&f->cnv,o), *d=rgbsubover(a,v,r,kr,kg,kb); while(++o,++d,--c); \
}

csox(cso16,uint16)
csox(cso32,int32)

/* image subOver - direct */

#define xsox(n,t) \
static DEF_BLT_RTN(n) { \
	const t		*s=(const t*)f->cnv.s+o-f->vx; \
	t		*d=(t*)f->a+o; \
	int32		a, r, v, kr=f->opr, kg=f->opg, kb=f->opb; \
	do a=*d, v=*s, *d=rgbsubover(a,v,r,kr,kg,kb); while(++s,++d,--c); \
}

xsox(xso16,uint16)
xsox(xso32,int32)

/* image subOver - pattern */

static DEF_BLT_RTN(pso8c)
{
	const uint16	*s = f->cnv.s;
	const uint16	*dt = (const uint16*)f->dtb;
	const uint8	*it = f->cnv.itb;
	uint8		*d = (uint8*)f->a+o;
	int		a, r, v, kr = f->opr, kg = f->opg, kb = f->opb;

	o-= f->vx;
	do
		a = dt[*d], o&= f->mx, v = s[o], *d = it[rgbsubover(a,v,r,kr,kg,kb)];
	while(++o,++d,--c);
}

static DEF_BLT_RTN(pso8g)
{
	const uint8	*it = f->cnv.itb;
	const uint16	*dt = (const uint16*)it+512;
	const uint16	*s = f->cnv.s;
	uint8		*d = (uint8*)f->a+o;

	o-= f->vx;
	do
		o&= f->mx, *d = it[(dt[*d]-s[o])>>6&1023];
	while(++o,++d,--c);
}

#define psox(n,typ) \
static DEF_BLT_RTN(n) { \
	const typ	*s=f->cnv.s; \
	typ		*d=(typ*)f->a+o; \
	int32		a, r, v, kr=f->opr, kg=f->opg, kb=f->opb; \
	o-=f->vx; \
	do a=*d, o&=f->mx, v=s[o], *d=rgbsubover(a,v,r,kr,kg,kb); while(++o,++d,--c); \
}

psox(pso16,uint16)
psox(pso32,int32)

/* ---- mid-level processing */

static DEF_BLT_RTN(mlvpr)
{
	cnvParam	cnvsave = f->cnv;
	mlvParam	*mlv = f->mlc;
	void		*asave = f->a;
	int		xsave = f->vx;

	*mlv->m2d.cache = 0;
	mlv->m2d.cache[1] = 0;
	mlv->m2d.cache[2] = 0;
	f->a = (void*)mlv->m2d.s;
	if (mlv->rtn!=imgc32)
	{ /* non-copy xfer mode; must convert dest. to mlv line first */
		f->cnv = mlv->d2m;
		f->cnv.s = asave;
		f->vx = 0;
		imgc32(o,c,f);
		f->cnv = cnvsave;
		f->vx = xsave;
	}
	/* now, paint to mlv */
	mlv->rtn(o,c,f);
	/* transfer to dest., with whatever cnv used */
	f->a = asave;
	f->cnv = mlv->m2d;
	f->cnv.cc = mlv->d2m.cc;
	f->cnv.t = mlv->d2m.t;
	f->vx = 0;
	mlv->w2d(o,c,f);
	/* finally, save the variables and restore the source cnv */
	memcpy(mlv->m2d.cache,f->cnv.cache,sizeof(mlv->m2d.cache));
	f->cnv = cnvsave;
	f->vx = xsave;
}

/* ---- setup */

#define depLCnt		(rtnLine*5)

const bltRoutine	_blt_f_rtn[depLCnt*6] =
{
 sldx1,slor1,slxo1,slbc1,sldx1,sldx1,sldx1,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgx1,imor1,imxo1,imbc1,imgn1,imor1,imgx1,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgr1,mror1,mrxo1,mrbc1,imgr1,imrt1,imgr1,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgc1,mcor1,mcxo1,mcbc1,imgc1,imct1,imgc1,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgp1,mpor1,mpxo1,mpbc1,imgp1,impt1,imgp1,0,0,0,0,0,0,0,0,0,0,0,0,0,

 sldx2,slor2,slxo2,slbc2,sldx2,sldx2,sldx2,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgx2,imor2,imxo2,imbc2,imgn2,imor2,imgx2,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgr2,mror2,mrxo2,mrbc2,imgr2,imrt2,imgr2,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgc2,mcor2,mcxo2,mcbc2,imgc2,imct2,imgc2,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgp2,mpor2,mpxo2,mpbc2,imgp2,impt2,imgp2,0,0,0,0,0,0,0,0,0,0,0,0,0,

 sldx4,slor4,slxo4,slbc4,sldx4,sldx4,sldx4,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgx4,imor4,imxo4,imbc4,imgn4,imor4,imgx4,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgr4,mror4,mrxo4,mrbc4,imgr4,imrt4,imgr4,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgc4,mcor4,mcxo4,mcbc4,imgc4,imct4,imgc4,0,0,0,0,0,0,0,0,0,0,0,0,0,
 imgp4,mpor4,mpxo4,mpbc4,imgp4,impt4,imgp4,0,0,0,0,0,0,0,0,0,0,0,0,0,

 sldx8,slor8,slxo8,slbc8,sldx8,sldx8,slb8c,slb8g,smx8c,smx8g,smn8c,smn8g,sap8c,sap8g,sao8c,sao8g,ssp8c,ssp8g,sso8c,sso8g,
 imgx8,imor8,imxo8,imbc8,imgn8,imor8,imgx8,0,0,0,0,0,0,0,0,0,0,0,0,0, //!
 imgr8,mror8,mrxo8,mrbc8,imgr8,imrt8,xrb8c,xrb8g,rmx8c,rmx8g,rmn8c,rmn8g,rap8c,rap8g,rao8c,rao8g,rsp8c,rsp8g,rso8c,rso8g,
 imgc8,mcor8,mcxo8,mcbc8,imgc8,imct8,xcb8c,xcb8g,cmx8c,cmx8g,cmn8c,cmn8g,cap8c,cap8g,cao8c,cao8g,csp8c,csp8g,cso8c,cso8g,
 imgp8,mpor8,mpxo8,mpbc8,imgp8,impt8,xpb8c,xpb8g,pmx8c,pmx8g,pmn8c,pmn8g,pap8c,pap8g,pao8c,pao8g,psp8c,psp8g,pso8c,pso8g,

 sldx16,slor16,slxo16,slbc16,sldx16,sldx16,slb16,0,smx16,0,smn16,0,sap16,0,sao16,0,ssp16,0,sso16,0,
 imgx16,imor16,imxo16,imbc16,imgn16,imgt16,imb16,0,xmx16,0,xmn16,0,xap16,0,xao16,0,xsp16,0,xso16,0,
 imgr16,mror16,mrxo16,mrbc16,imgr16,imrt16,xrb16,0,rmx16,0,rmn16,0,rap16,0,rao16,0,rsp16,0,rso16,0,
 imgc16,mcor16,mcxo16,mcbc16,imgc16,imct16,xcb16,0,cmx16,0,cmn16,0,cap16,0,cao16,0,csp16,0,cso16,0,
 imgp16,mpor16,mpxo16,mpbc16,imgp16,imgt16,xpb16,0,pmx16,0,pmn16,0,pap16,0,pao16,0,psp16,0,pso16,0,

 sldx32,slor32,slxo32,slbc32,sldx32,sldx32,slb32,0,smx32,0,smn32,0,sap32,0,sao32,0,ssp32,0,sso32,0,
 imgx32,imor32,imxo32,imbc32,imgn32,imgt32,imb32,0,xmx32,0,xmn32,0,xap32,0,xao32,0,xsp32,0,xso32,0,
 imgr32,mror32,mrxo32,mrbc32,imgr32,imrt32,xrb32,0,rmx32,0,rmn32,0,rap32,0,rao32,0,rsp32,0,rso32,0,
 imgc32,mcor32,mcxo32,mcbc32,imgc32,imct32,xcb32,0,cmx32,0,cmn32,0,cap32,0,cao32,0,csp32,0,cso32,0,
 imgp32,mpor32,mpxo32,mpbc32,imgp32,imgt32,xpb32,0,pmx32,0,pmn32,0,pap32,0,pao32,0,psp32,0,pso32
};

/* offset tables */

#define r		rtnLine
#define c		(rtnLine*2)

static const uint8	f_m2x[] =
{
	  0,  1,  2,  3,  4,  1,  2,  3,   5,  6,  8,  10,  12,  14,  16,  18,
	r+0,r+1,r+2,r+3,r+4,r+1,r+2,r+3, r+5,r+6,r+8,r+10,r+12,r+14,r+16,r+18,
	c+0,c+1,c+2,c+3,c+4,c+1,c+2,c+3, c+5,c+6,c+8,c+10,c+12,c+14,c+16,c+18
};

#undef r
#undef c

/* avoids multiple shift or multiply operations */

static const int	pDepthR[32] =
{
	5*depLCnt,0*depLCnt,1*depLCnt,2*depLCnt,
	2*depLCnt,3*depLCnt,3*depLCnt,3*depLCnt,
	3*depLCnt,4*depLCnt,4*depLCnt,4*depLCnt,
	4*depLCnt,4*depLCnt,4*depLCnt,4*depLCnt,
	4*depLCnt,5*depLCnt,5*depLCnt,5*depLCnt,
	5*depLCnt,5*depLCnt,5*depLCnt,5*depLCnt,
	5*depLCnt,5*depLCnt,5*depLCnt,5*depLCnt,
	5*depLCnt,5*depLCnt,5*depLCnt,5*depLCnt
};

/*
   Instead of using region ops (e.g., rRegion->sectRgn), which would use memory,
 or rewriting the mgraphic ops to work with clipRgn, which would be exhausting,
 an additional defer layer was added.  This is simpler and requires less code.
   In order for this to work, however, blt rtn calls must be made in sorted x
 order, left-to-right, and rEndLine must be called at the end of every line.
*/
DEF_BLT_RTN(_bltRgnLayer)
{
	int32		rx0 = *f->rr, rx1, c1;

	o+= f->ox, c+= o;
	if (c<=rx0)
		return;
	rx1 = f->rr[1];
	do
	{
		if (o>=rx1)
		{ /* input range is ahead of current rgn range; advance */
			const int16	*a = f->ra, *e = f->re;
			/* should never get here if o>=IMAX */
			/* rInitBlt advanced to the next position, so post-increment */
			do
			{
				if (a>=e || *a==RGN_STOP)
				{
					*f->rr = f->rr[1] = IMAX(*f->rr);
					return;
				}
				rx0 = *a, ++a;
				if (a>=e || *a==RGN_STOP)
				{
					rx1 = IMAX(*f->rr);
					break;
				}
				rx1 = *a, ++a;
			} while(o>=rx1);
			*f->rr = rx0;
			f->rr[1] = rx1;
			f->ra = a;
		}
		if (o<rx0)
			o = rx0;
		c1 = MIN(c,rx1);
		if (c1>o)
			f->rrtn(o-f->ox,c1-o,f);
	} while(o=rx1,c>rx1); /* end of input persists; split */
}

int rInitBlt(bltParam *f,
	grafPort *p, /* where the virtual origin/clip reside */
	const void *s,
	const cTab64 *ct,
	int o,
	int c,
	int mode)
{
	rect16		cl;
	pixMap		*m = p->map; /* where the actual origin/bounds reside */
	uint32		pmask;
	int		rtni, fmt, rb, g, pz; /* src format, red/blue bit-size, green */
	int		irtnOffs;
	uint8		dflag;
	SLRgn		*clip;

	f->cnv.ac = f->cnv.cc = 0, f->mlc = NULL;
	/* clip */
	f->ra = NULL;
	if ((clip=p->clipRgn)==NULL)
		return -1;
	cl = clip->bBox;
	f->cl.top = MAX(p->origin.y,cl.top);
	f->cl.bottom = m->h+p->origin.y;
	if (f->cl.bottom>cl.bottom)
		f->cl.bottom = cl.bottom;
	if (c>f->cl.bottom)
		c = f->cl.bottom;
	if (c<=f->cl.top)
		return -1; /* clipped out-top */
	if (o>=f->cl.bottom)
		return -1; /* clipped out-bottom */
	/* set up the parameters */
	if ((f->p=lockPixels(m))==NULL)
		return -1; /* bad parameter */
	f->ox = p->map->pOrigin.x+p->origin.x;
	f->oy = p->map->pOrigin.y+p->origin.y;
	f->oz = m->pitch;
	/* caller should tell if object l/r-clips out of existence */
	f->cl.left = MAX(p->origin.x,cl.left);
	f->cl.right = m->w+p->origin.x;
	if (f->cl.right>cl.right)
		f->cl.right = cl.right;
	pmask = (2u<<(m->bpp-1))-1;
	dflag = 0;
	f->sc = p->fgSolid;
	rb = m->minCmp, g = rb+(m->pType>>pTypeLGreenBit&1);
	pz = m->pSize;
	if ((fmt=mode>>8)!=0)
	{
		int		ifm = formPxFmt(m);

		if (!s || !((fmt&0xC0FFFF)!=ifm
				|| ((uint16)fmt<0x900 && ct && (!m->pTable
				|| ct->seed!=m->pTable->seed))))
			fmt = 0;
		else if (mode&mDither)
			if (m->bpp>29 || (uint16)fmt==ifm)
				mode&=~mDither; /* dithering prec. is 30-bit */
			else
				pz = 32, rb = g = 10;
	}
	switch(mode&15)
	{
	case mSrcXor: /* todo: handle hilite (alt mode bit 7) */
		if (pz>8)
			{f->opr = pmask*pSizeMasks[pz&31]; goto kxo;}
		goto norm;
	case mSrcNotXor:
		if (pz>8)
			{f->opr = 0; goto knxo;}
	case mSrcNotCopy:
	case mSrcNotOr:
	case mSrcNotBic:
	nmod:	f->opr = pmask*pSizeMasks[pz&31];
	knxo:	f->vz = p->fgSolid, f->sc = p->bkSolid;
		dflag|= remapReverse; /* not mode only reverses colors for patterns */
		break;
	case mTransparent:
		/* the fact that 'transparent' is outside the 'arithmetic' mode list
		  in QuickDraw.h indicates that bkColor is used for the transparent color
		*/
		f->opr = f->vz = p->bkSolid&pmask;
		break;
	case mSubPin:
	case mAdMin:
		if (pz<8)
			{mode&=~15, mode|= mSrcNotOr; goto nmod;}
		goto arith;
	case mSubOver:
	case mAddOver:
		if (pz<8)
			{mode&=~15, mode|= mSrcXor; goto norm;}
		goto arith;
	case mAdMax:
	case mAddPin:
		if (pz<8)
			{mode&=~15, mode|= mSrcBic; goto norm;}
	arith:	if (rb)
			*(int32*)&f->bmg = formDirect(&p->opColor,g,rb);
		dflag|= ccIndexed;
		goto kxc;
	case mBlend:
		if (pz<8)
			goto opaque;
		/* note that QD interpolation truncates *against* destination */
		/* (even 0,0,0 opacity may result in a slight change) */
		/* QuickTime, however, may treat the values as 0.0-to-1.0 */
		f->bmr = (p->opColor.red+(1<<15-blndPrec))>>(16-blndPrec);
		f->bmg = (p->opColor.green+(1<<15-blndPrec))>>(16-blndPrec);
		f->bmb = (p->opColor.blue+(1<<15-blndPrec))>>(16-blndPrec);
		if (!(f->bmr|f->bmg|f->bmb))
			return -1; /* 0% opacity */
		if ((f->bmr&f->bmg&f->bmb)==(1<<blndPrec))
		{ /* 100% opacity */
		opaque: mode&=~15;
			goto norm;
		}
		/* force direct/chroma values with rGenRemappingTable */
		dflag|= ccIndexed;
	kxc:	f->vz = p->bkSolid&pmask;
		break;
	default:
	norm:	f->opr = 0;
	kxo:	f->vz = p->bkSolid;
	}
	irtnOffs = rtnLine;
	if ((f->cnv.s=s)==NULL && p->pPat)
	{ /* set up pattern */
		pixPat		*pat = p->pPat;

		f->mx = (1u<<(31-cntlz32(pat->w)))-1, f->mh = pat->h;
		mode|= mSrcPat;
		if ((int16)pat->ct.flags<(int16)0x8100)
		{
			fmt = pDepthK[pz];
			pat->ct.colors[0].value = p->fgSolid&fmt;
			pat->ct.colors[1].value = p->bkSolid&fmt;
		}
		if ((fmt=rConvertPat(pat,m,dflag))<0)
			return fmt;
		f->cnv.mp = pat->pitch;
		f->cnv.s = pat->px;
		if ((f->cnv.t=pat->nx)==NULL)
			goto xpat;
		if (pat->xpitch>0)
		{
			f->cnv.mp = pat->xpitch;
			f->cnv.s = pat->nx;
		xpat:	irtnOffs = rtnLine*3;
		}
		fmt = 0;
	} else /* fg/bk default table */
		f->cnv.t = &f->vz, f->mh = f->mx =~0u;
	if (pz<9)
	{ /* processed here first so minCmp is correct */
		if (!m->iTable && pmMakeITables(m,m->bpp,0)<=0)
			return -2;
		f->cnv.itb = m->iTable;
		f->dtb = (const uint16*)(f->cnv.itb+(m->minCmp ? 1u<<m->minCmp*3:1024));
	}
	dflag&=~remapReverse;
	if ((f->sb=f->cnv.s)==NULL)
		rtni = f_m2x[mode&=15];
	else
	{
		dflag|= mode&mDither; /* should have same bit as remapForce30 */
		mode&= 31|mMergeDown;
		f->sc&= pmask, f->vz&= pmask;
		if (fmt)
			if ((mode=rInitCnv(&f->cnv,m,ct,fmt,fmt<0 ? f->vx:0x10000,
					(mode|mSrcPat)+mSrcPat,dflag
					|((mode&15)-mSrcNotCopy<4u^(pz>8 /* inversion flag */
					&& (mode&(8|3))==mSrcXor))))<0)
				return mode;
			else if (dflag&mDither && (f->mlc=rInitDithering(m))==NULL)
				return -2;
		rtni = f_m2x[mode&63]+irtnOffs, mode&= 15;
	}
	rtni+= pDepthR[pz&31]; //pDepthX[pz&31]*(rtnLine*4);
	f->opb = (1u<<rb)-1;
	if (pz<9)
	{ /* indexed destination */
		if (dflag)
		{
			if (!m->pTable || !m->minCmp)
			{ /* 8-bit gray, opColor */
				const rgb48	*cl = &p->opColor;
				/* gray routines call for a single intensity opColor */
				++rtni;
				if (mode>mBlend)
					f->bmg = formGray(cl->red,cl->green,cl->blue,g)>>1;
				else
					f->bmg = (formGray(cl->red,cl->green,cl->blue,g)
						+(1<<16-blndPrec))>>(17-blndPrec);
			}
			f->vz = ((uint16*)f->dtb)[f->vz];
			f->sc = ((uint16*)f->dtb)[f->sc&pmask];
			goto opset;
		}
	} else if (mode==mSrcXor) /* direct */
	{
		if (!s)
			pmask*= pSizeMasks[pz&31];
		f->sc^= pmask, f->vz^= pmask;
	} else if (dflag)//&~remapFullDirect)
	{
	opset:	f->opr = f->opb<<rb<<g;
		f->opg = ((1u<<g)-1)<<rb;
	}
	if (pz<17)// && !(dflag&remapFullDirect))
	{
		f->vz = f->vz<<16*BIG_ENDIAN|f->sc<<16*!BIG_ENDIAN;
		if (s)
			f->sc = f->vz;
	}
	f->rtn = _blt_f_rtn[rtni];
	if (f->mlc)
	{
		f->mlc->w2d = _blt_f_rtn[rtnLine*3+pDepthR[m->pSize&31]];
		f->mlc->rtn = f->rtn;
		f->rtn = mlvpr;
	}
	c = rInitClip2(f,o,c,&f->a,&f->e);
	if ((pz=clip->size)>RECT_RGN_SIZE && !(pz&1))
	{ /* set up region parameters */
		f->ra = _rgnGotoLine(clip->x,f->re=(const int16*)((int8*)clip+pz),
			o+c-clip->bBox.top,RECTANGLE_HEIGHT(&clip->bBox));
		f->rrtn = f->rtn;
		f->rtn = _bltRgnLayer;
		/* load first line 'slab', filling in IMAX for stops */
		/* advance as much as two, but no further than end or a stop */
		if (f->ra>=f->re || *f->ra==RGN_STOP)
			f->rr[1] = *f->rr = IMAX(*f->rr);
		else
		{
			*f->rr = *f->ra++;
			if (f->ra>=f->re || *f->ra==RGN_STOP)
				f->rr[1] = IMAX(*f->rr);
			else
				f->rr[1] = *f->ra++;
		}
	}
	return c;
}

int rEndLine(bltParam *f,
	const int8 *e)
{
	if (f->cnv.s)
	{
		f->cnv.hx =~0;
		if (f->mh)
		{
			f->vy+= f->oz>>31|1;
			f->cnv.s = (const int8*)f->cnv.s+f->cnv.mp;
			if (f->vy+0u>=f->mh)
				f->vy = 0, f->cnv.s = f->sb;
		}
	}
	if (f->ra)
	{
		while(f->ra<f->re && *f->ra++!=RGN_STOP)
			;
		if (f->ra>=f->re || *f->ra==RGN_STOP)
			f->rr[1] = *f->rr = IMAX(*f->rr);
		else
		{
			*f->rr = *f->ra++;
			if (f->ra>=f->re || *f->ra==RGN_STOP)
				f->rr[1] = IMAX(*f->rr);
			else
				f->rr[1] = *f->ra++;
		}
	}
	return (f->a+=f->oz)<e;
}

void rEndBlt(bltParam *f)
{
	rEndCnv(&f->cnv);
	if (f->mlc)
	{
		rEndCnv(&f->mlc->m2d);
		rEndCnv(&f->mlc->d2m);
		free(f->mlc); f->mlc = NULL;
	}
}

void rBltOrigin(bltParam *f,
	int x,
	int y)
{
	if (y<f->cl.top)
		y = f->cl.top;
	if (f->sb && f->mh)
	{
		y-= f->oy;
		if (y+0u>=f->mh && (y%=(int32)f->mh)<0)
			y+= f->mh;
		f->cnv.s = (const int8*)f->sb+y*f->cnv.mp;
	}
	f->vx = x, f->vy = y;
}
