/*
 mcnv.c
  pixel depth, endian and scaled conversion

   Created: 2014.04.22
   Revised: 2024.06.07

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.04.23
  Split from mblt.
2014.04.26
  Finished mcnv scaling.
2014.05.06
  Fixed dithering (mostly), split into RGB and Gray.
2014.08.15
  calloc instead of malloc appears to fix direct dithering.
2022.02.11
  Multiple-format filtered downscaling by storing masks.
2024.05.25
  Added full horizontal anti-aliasing for direct sources (if MergeDown).
2024.06.01
  Finished indexed-mode mergedown.
*/

#include <string.h>
#include "mgraf.h"
#include "mblt.h"

#include "ftrig.h"

/* ---- constants */

#define directSrc		8
#define directBSwap		1
#define direct32b		2
#define directDst		4
#define directScaled		20
#define indexedScaled		4

#define directSmoothV32		(48+14)
#define directSmooth32		(48+19)
#define directSmooth24		(48+24)
/* non-direct dest. */
#define dstGray			4
#define dstInverse		8

/* ---- typedefs */

/* replaces 'cache'/'hx'/'rx'; this is also 40 bytes at most */
typedef struct ditherParam/*(24+)*/
{
	int32		r, g, b;
	int		i;
	void		*gt;
	rgbRoutine	cc;
	rgb48		*a;
} ditherParam;

#define DITHERVAR(f)	((ditherParam*)f)

/* ---- code */

void *rGenRemappingTable(pixMap *m,
	const cTab64 *srcct,
	uint8 bpp,
	uint8 flag)
{
	rgbRoutine	crtn;
	uint32		*t, *d, i, cm, rb;
	int		c, is32b = (flag&remap32bit)>>6;
	void		*gt;

	if (!bpp) /* idiot-proofing */
		return NULL;
	if ((t=calloc(is32b+2,srcct && srcct->flags&0x8000 ? MAX(srcct->count,256):256))==NULL)
		return NULL;
	d = t;//+c;
	if (flag&remapForce30)
		gt = (void*)(rb=10), i = 31, c = ccDirect; //(c|ccDirect)&~ccIndexed;
	else
	{
		if ((c=pmGetRGBCC(m,&gt))<0)
			return NULL;
		rb = m->minCmp ? m->minCmp:10;
		if (flag&c&ccIndexed) /* indirect table supression (arithmetic modes) */
			c&=~ccNative, i = rb;
		else
			i = m->bpp;
	}
	cm =-(flag&remapDInvert)&((2u<<(i-1))-1);
	crtn = rgbCCRoutine[c], i = rb;
	if (srcct && (c=srcct->count)!=0)
	{
		const cSpec	*cs = srcct->colors;

		if (m->bpp<9) /* 2015.07.09 */
                {
			if (is32b)
				((int32*)d)[255] = 0xFF;
			else
				((int16*)d)[255] = 0xFF;
                }
		if (srcct->flags&0x8000)
			if (flag&remapReverse)
				if (is32b)
				{
					d+= c;
					do
						*--d = crtn(i,gt,&cs->rgb)^cm;
					while(++cs,--c);
				} else
				{
					d = (uint32*)((uint16*)d+c);
					do
						d = (uint32*)((uint16*)d-1), *(uint16*)d = (uint16)(crtn(i,gt,&cs->rgb)^cm);
					while(++cs,--c);
				}
			else if (is32b)
				do
					*d = crtn(i,gt,&cs->rgb)^cm;
				while(++d,++cs,--c);
			else
				do
					*(uint16*)d = (uint16)(crtn(i,gt,&cs->rgb)^cm), d = (uint32*)((uint16*)d+1);
				while(++cs,--c);
		else if (is32b)
			do
				d[cs->value] = crtn(i,gt,&cs->rgb)^cm;
			while(++cs,--c);
		else
			do
				((uint16*)d)[cs->value] = (uint16)(crtn(i,gt,&cs->rgb)^cm);
			while(++cs,--c);
	} else
	{
		rgb48		cl;
		uint32		x;
		int		dx;

		c = 1u<<bpp, dx = (65535u<<16)/(c-1), x =~0u;
		if (is32b)
			do
			{
				cl.red = cl.green = cl.blue = (uint16)(x>>16);
				*d = crtn(i,gt,&cl)^cm;
			} while(++d,x-=dx,--c);
		else
			do
			{
				cl.red = cl.green = cl.blue = (uint16)(x>>16);
				*(uint16*)d = (uint16)(crtn(i,gt,&cl)^cm), d = (uint32*)((uint16*)d+1);
			} while(x-=dx,--c);
	}
	return t;
}

/* ---- direct conversion */

/* maybe these optimized converters could be used...
 nah - they would slow down setup further, and'd require more code than this
*/

/*uint32 i15to24(uint16 v)
{
	return (uint8)(v<<3|v>>2)|(v<<6&0xF800)|(v<<1&0x700)
		|(v<<9&0xF80000)|(v<<4&0x70000);
}
uint16 i24to15(uint32 v)
{
	v>>= 3;
	return (v&0x1F)|(v>>3&0x3E0)|(v>>6&0x7C00);
}
*/

static DEF_CNV_RTN(pxroDirect32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	return cnvDirect32(v,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxroBDirect32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	v = _reverse32(v);
	return cnvDirect32(v,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxroDirect16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	return cnvDirect(v,u,t,r)^f->cm;
}

static DEF_CNV_RTN(pxroBDirect16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	v = _reverse16(v);
	return cnvDirect(v,u,t,r)^f->cm;
}

/* ---- direct through gray */

static DEF_CNV_RTN(pxroGray32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	return cnvGray32(v,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxroBGray32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	v = _reverse32(v);
	return cnvGray32(v,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxroGray16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	return cnvGray(v,u,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxroBGray16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	v = _reverse16(v);
	return cnvGray(v,u,t,r)>>1^f->cm;
}

/* ---- direct through itable */

static DEF_CNV_RTN(pxroIDirect32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	return f->itb[cnvDirect32(v,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxroIBDirect32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	v = _reverse32(v);
	return f->itb[cnvDirect32(v,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxroIDirect16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	return f->itb[cnvDirect(v,u,t,r)^f->cm];
}

static DEF_CNV_RTN(pxroIBDirect16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	v = _reverse16(v);
	return f->itb[cnvDirect(v,u,t,r)^f->cm];
}

/* ---- direct through gray itable */

static DEF_CNV_RTN(pxroIGray32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	return f->itb[cnvGray32(v,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxroIBGray32)
{
	uint32		*t = f->t, r, u, v = ((const uint32*)f->s)[o];

	v = _reverse32(v);
	return f->itb[cnvGray32(v,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxroIGray16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	return f->itb[cnvGray(v,u,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxroIBGray16)
{
	uint32		*t = f->t, r, u, v = ((const uint16*)f->s)[o];

	v = _reverse16(v);
	return f->itb[cnvGray(v,u,t,r)>>7^f->cm];
}

/* ---- same-depth direct */

static DEF_CNV_RTN(pxro32)
	{return ((const uint32*)f->s)[o]^f->cm;}

static DEF_CNV_RTN(pxroB32)
{
	uint32		v = ((const uint32*)f->s)[o];

	return _reverse32(v)^f->cm;
}

static DEF_CNV_RTN(pxro16)
	{return ((const uint16*)f->s)[o]^f->cm;}

static DEF_CNV_RTN(pxroB16)
{
	uint16		v = ((const uint16*)f->s)[o];

	return _reverse16(v)^f->cm;
}

/* ---- indexed through table */

/* this one's usually overridden with the pattern method */
static DEF_CNV_RTN(pxro8)
	{return ((uint32*)f->t)[((const uint8*)f->s)[o]];}

static DEF_CNV_RTN(pxroCache4)
{
	int		u = o>>2, v;

	if (u!=f->hx)
	{ /* 2-byte alignment */
		const uint8	*s = (const uint8*)f->s+u*2;

		f->hx = u;
		v = *s;
		*f->cache = v>>4;
		f->cache[1] = v&15;
		v = s[1];
		f->cache[2] = v>>4;
		f->cache[3] = v&15;
	}
	return ((uint32*)f->t)[f->cache[o&3]];
}

static DEF_CNV_RTN(pxroCache2)
{
	int		u = o>>2;

	if (u!=f->hx)
	{
		f->hx = u, u = ((const uint8*)f->s)[u];
		*f->cache = u>>6;
		f->cache[1] = u>>4&3;
		f->cache[2] = u>>2&3;
		f->cache[3] = u&3;
	}
	return ((uint32*)f->t)[f->cache[o&3]];
}

static DEF_CNV_RTN(pxroCache1)
{
	int		u = o>>3;

	if (u!=f->hx)
	{
		f->hx = u, u = ((const uint8*)f->s)[u];
		*f->cache = u>>7;
		f->cache[1] = u>>6&1;
		f->cache[2] = u>>5&1;
		f->cache[3] = u>>4&1;
		f->cache[4] = u>>3&1;
		f->cache[5] = u>>2&1;
		f->cache[6] = u>>1&1;
		f->cache[7] = u&1;
	}
	return ((uint32*)f->t)[f->cache[o&7]];
}

/* ---- scaled */

#define rfdi()		(void)(*d=s[q>>16],q+=f->rx)
#define rxdi(x)		{i=q>>16; if(i>=w+0u) goto end; d[x]=s[i],q+=f->rx;}

#define hscali(typ)	\
	uint32 q=o>>3, i; \
	o&=7; \
	if(q!=f->hx) { \
		const typ *s=f->s; int32 *d=f->cache, w=f->sw; \
		f->hx=q, q=q*8*f->rx-f->sx; \
		rfdi(); \
		rxdi(1) \
		rxdi(2) \
		rxdi(3) \
		rxdi(4) \
		rxdi(5) \
		rxdi(6) \
		rxdi(7) \
	}

#if 1
/*#define lerp(a,b,q,z)	do{uint32 _b=b,m2=((uint16)q)>>16-z,m1=(1<<z)-m2; \
	a=((((a&k1)*m1+(_b&k1)*m2+r1)>>z)&k1)+((((a&k2)*m1+(_b&k2)*m2+r2)>>z)&k2); \
}while(0)

#define rxdl(x,z)	{i=qx>>16; if(i>=w) {d[x]=s[w-1]; return;} --i; \
		a=s[MAX(i,0)]; lerp(a,s[i+1],qx,z); \
	d[x]=a;}*/

#include "filter.h"

#define fir3(a,a2,a3,q)	do{uint32 v2=a2,v3=a3,m1,m2,m3; \
	const int8 *fx=f->fx+(((uint16)q)>>16-7)*3; \
	m1=*fx,m2=fx[1],m3=fx[2]; \
	a=((((a&k1)*m1+(v2&k1)*m2+(v3&k1)*m3+r1)>>6)&k1) \
	+((((a&k2)*m1+(v2&k2)*m2+(v3&k2)*m3+r2)>>6)&k2); \
}while(0)

#define fir3q(a,a2,a3,q) do{uint32 v2=a2,v3=a3,m1,m2,m3,b; \
	const int8 *fx=f->fx+(((uint16)q)>>16-7)*3; \
	m1=*fx,m2=fx[1],m3=fx[2]; \
	b=(((a&k1)*m1+(v2&k1)*m2+(v3&k1)*m3)>>6)&k1; \
	a=b+((((uint64)(a&k2)*m1+(uint64)(v2&k2)*m2+(uint64)(v3&k2)*m3)>>6)&k2); \
}while(0)

/*#define fir5(a,a2,a3,a4,a5,q) do{uint32 v2=a2,v3=a3,v4=a4,v5=a5,m1,m2,m3,m4,m5; \
	const int8 *fx=f->fx+(((uint16)q)>>16-7)*5; \
	m1=*fx,m2=fx[1],m3=fx[2],m4=fx[3],m5=fx[4]; \
	a=((((a&k1)*m1+(v2&k1)*m2+(v3&k1)*m3+(v4&k1)*m4+(v5&k1)*m5+r1)>>7)&k1) \
	+((((a&k2)*m1+(v2&k2)*m2+(v3&k2)*m3+(v4&k2)*m4+(v5&k2)*m5+r2)>>7)&k2); \
}while(0)*/

/* vertical scaling subroutines */

static DEF_CNV_RTN(hscvu32)
{
	return ((const uint32*)f->s)[o];
}

static DEF_CNV_RTN(hscvu16)
{
	return ((const uint16*)f->s)[o];
}

static DEF_CNV_RTN(hscvfir32)
{
	const uint32	*s1, *s2, *s3;
	uint32		*d = (uint32*)f->t+_cvdSize;
	uint32		v1, v2, v3, m1, m2, m3, r1, r2, k1, k2;
	int		i, c;

	if (o<=f->hy)
		return d[o];
	i = f->hy+3; /* need at least 3 */
	if (i<o)
		i = o;
	c = i+3; /* need to render ahead of indexed reads, even if producing 6 in 2 pixels */
	if (c>f->sw)
		c = f->sw;
	c-= i;
	if (c>0)
	{
		f->hy = i;
		m1 = f->m[0], m2 = f->m[1], m3 = f->m[2];
		k1 = ((const int32*)f->t)[_cvsg];
		k2 = ((const int32*)f->t)[_cvrb];
		r1 = (k1&~(k1<<1))<<5, r2 = (k2&~(k2<<1))<<5;
		s1 = i+(const uint32*)f->fa, s2 = i+(const uint32*)f->s, s3 = i+(const uint32*)f->fe;
		d+= i;
		do
		{
			v1 = *s1++, v2 = *s2++, v3 = *s3++;
			*d++ = (((v1&k1)*m1+(v2&k1)*m2+(v3&k1)*m3+r1)>>6&k1)
				+(((v1&k2)*(uint64)m1+(v2&k2)*(uint64)m2+(v3&k2)*(uint64)m3+r2)>>6&k2);
		} while(--c);
	}
	return ((const uint32*)f->fb)[o];
}

static DEF_CNV_RTN(hscvfir24)
{
	const uint32	*s1, *s2, *s3;
	uint32		*d = (uint32*)f->t+_cvdSize;
	uint32		v1, v2, v3, m1, m2, m3, r1, r2, k1, k2;
	int		i, c;

	if (o<=f->hy)
		return d[o];
	i = f->hy+3; /* need at least 3 */
	if (i<o)
		i = o;
	c = i+3; /* need to render ahead of indexed reads, even if producing 6 in 2 pixels */
	if (c>f->sw)
		c = f->sw;
	c-= i;
	if (c>0)
	{
		f->hy = i;
		m1 = f->m[0], m2 = f->m[1], m3 = f->m[2];
		k1 = ((const int32*)f->t)[_cvsg];
		k2 = ((const int32*)f->t)[_cvrb];
		r1 = (k1&~(k1<<1))<<5, r2 = (k2&~(k2<<1))<<5;
		s1 = i+(const uint32*)f->fa, s2 = i+(const uint32*)f->s, s3 = i+(const uint32*)f->fe;
		d+= i;
		do
		{
			v1 = *s1++, v2 = *s2++, v3 = *s3++;
			*d++ = (((v1&k1)*m1+(v2&k1)*m2+(v3&k1)*m3+r1)>>6&k1)
				+(((v1&k2)*m1+(v2&k2)*m2+(v3&k2)*m3+r2)>>6&k2);
		} while(--c);
	}
	return ((const uint32*)f->fb)[o];
}

static DEF_CNV_RTN(hscvbc32)
{
	const uint32	*s, *e;
	uint32		*d = (uint32*)f->t+_cvdSize;
	int		i, c;

	if (o<=f->hy)
		return d[o];
	i = f->hy+3; /* need at least 3 */
	if (i<o)
		i = o;
	c = i+3; /* need to render ahead of indexed reads, even if producing 6 in 2 pixels */
	if (c>f->sw)
		c = f->sw;
	c-= i;
	if (c>0)
	{
		f->hy = i;
		s = i+(const uint32*)f->s;
		e = i+(const uint32*)f->fe;
		d+= i;
		/*if (s>=e)
			do
				*d++ = *s++;
			while(--c);
		else*/
		{
			const uint32 gk=((const int32*)f->t)[_cvsg];
			const uint32 rb=((const int32*)f->t)[_cvrb];
			const uint32 rk=rb&(-gk), bk=rb^rk;
			const int mp=f->mp;
			uint32 minr=-(f->ry>>16);
			do
			{
				const uint32 *sa=s;
				uint32 np=minr, a=*sa, t, ar=0, ag=0;
				ar=(a&rk)>>12; ag=(a&gk)>>6; a&=bk;
				do
				{
					sa=(const uint32*)((const int8*)sa+mp);
					t=*sa; a+=t&bk; ag+=(t&gk)>>6; ar+=(t&rk)>>12;
				} while(++np,sa<e);
				np = f->m[np+1];/*4294967296./np*/
				a =ufract32mul(a,np); a|= ufract32mul(ag,np)<<6&gk; a|= ufract32mul(ar,np)<<12&rk;
				*d++ = a;
			} while(++s,++e,--c);
		}
	}
	return ((const uint32*)f->fb)[o];
}

static DEF_CNV_RTN(hscvbc32_2)
{
	const uint32	*s, *e;
	uint32		*d = (uint32*)f->t+_cvdSize;
	int		i, c;

	if (o<=f->hy)
		return d[o];
	i = f->hy+3; /* need at least 3 */
	if (i<o)
		i = o;
	c = i+3; /* need to render ahead of indexed reads, even if producing 6 in 2 pixels */
	if (c>f->sw)
		c = f->sw;
	c-= i;
	if (c>0)
	{
		f->hy = i;
		s = i+(const uint32*)f->s;
		e = i+(const uint32*)f->fe;
		d+= i;
		if (s>=e)
			do
				*d++ = *s++;
			while(--c);
		else
		{
			uint32 r[8], a;
			const uint32 gk=((const int32*)f->t)[_cvsg];
			const uint32 rb=((const int32*)f->t)[_cvrb];

			a = *s; *r=a&rb, r[1]=a&gk;
			a = s[1]; r[2]=a&rb, r[3]=a&gk;
			a = s[2]; r[4]=a&rb, r[5]=a&gk;
			a = s[3]; r[6]=a&rb, r[7]=a&gk;
			s = (const uint32*)((const int8*)s+f->mp);
			a = *s; *r+=a&rb, r[1]+=a&gk;
			a = s[1]; r[2]+=a&rb, r[3]+=a&gk;
			a = s[2]; r[4]+=a&rb, r[5]+=a&gk;
			switch(c)
			{
			case 4: a = s[3]; r[6]+=a&rb, r[7]+=a&gk;
				d[3] = (r[6]>>1&rb)|(r[7]>>1)&gk;
			case 3:	d[2] = (r[4]>>1&rb)|(r[5]>>1)&gk;
			case 2:	d[1] = (r[2]>>1&rb)|(r[3]>>1)&gk;
			default: *d = (*r>>1&rb)|(r[1]>>1)&gk;
			}
		}
	}
	return ((const uint32*)f->fb)[o];
}

static DEF_CNV_RTN(hscvbc16)
{
	const uint16	*s, *e;
	uint16		*d = (uint16*)((uint32*)f->t+_cvdSize);
	int		i, c;

	if (o<=f->hy)
		return d[o];
	i = f->hy+3;
	if (i<o)
		i = o;
	c = i+3;
	if (c>f->sw)
		c = f->sw;
	c-= i;
	if (c>0)
	{
		f->hy = i;
		s = i+(const uint16*)f->s;
		e = i+(const uint16*)f->fe;
		d+= i;
		/*if (s>=e)
			do
				*d++ = *s++;
			while(--c);
		else*/
		{
			const uint32 gk=((const int32*)f->t)[_cvsg];
			const uint32 rb=((const int32*)f->t)[_cvrb];
			const uint32 rk=rb&(-gk), bk=rb^rk;
			const int mp=f->mp;
			uint32 minr=-(f->ry>>16);
			do
			{
				const uint16 *sa=s;
				uint32 np=minr, a=*sa, t, ar=0, ag=0;
				ar=a&rk; ag=a&gk; a&=bk;
				do
				{
					sa=(const uint16*)((const int8*)sa+mp);
					t=*sa; a+=t&bk; ag+=t&gk; ar+=t&rk;
				} while(++np,sa<e);
				np = f->m[np+1];
				a =ufract32mul(a,np); a|= ufract32mul(ag,np)&gk; a|= ufract32mul(ar,np)&rk;
				*d++ = a;
			} while(++s,++e,--c);
		}
	}
	return ((const uint16*)f->fb)[o];
}

static DEF_CNV_RTN(hscvbc16_2)
{
	const uint16	*s, *e;
	uint16		*d = (uint16*)((uint32*)f->t+_cvdSize);
	int		i, c;

	if (o<=f->hy)
		return d[o];
	i = f->hy+3; /* need at least 3 */
	if (i<o)
		i = o;
	c = i+3; /* need to render ahead of indexed reads, even if producing 6 in 2 pixels */
	if (c>f->sw)
		c = f->sw;
	c-= i;
	if (c>0)
	{
		f->hy = i;
		s = i+(const uint16*)f->s;
		e = i+(const uint16*)f->fe;
		d+= i;
		if (s>=e)
			do
				*d++ = *s++;
			while(--c);
		else
		{
			uint32 r[8], a;
			const uint32 gk=((const int32*)f->t)[_cvsg];
			const uint32 rb=((const int32*)f->t)[_cvrb];

			a = *s; *r=a&rb, r[1]=a&gk;
			a = s[1]; r[2]=a&rb, r[3]=a&gk;
			a = s[2]; r[4]=a&rb, r[5]=a&gk;
			a = s[3]; r[6]=a&rb, r[7]=a&gk;
			s = (const uint16*)((const int8*)s+f->mp);
			a = *s; *r+=a&rb, r[1]+=a&gk;
			a = s[1]; r[2]+=a&rb, r[3]+=a&gk;
			a = s[2]; r[4]+=a&rb, r[5]+=a&gk;
			switch(c)
			{
			case 4: a = s[3]; r[6]+=a&rb, r[7]+=a&gk;
				d[3] = (r[6]>>1&rb)|(r[7]>>1)&gk;
			case 3:	d[2] = (r[4]>>1&rb)|(r[5]>>1)&gk;
			case 2:	d[1] = (r[2]>>1&rb)|(r[3]>>1)&gk;
			default: *d = (*r>>1&rb)|(r[1]>>1)&gk;
			}
		}
	}
	return ((const uint16*)f->fb)[o];
}

/* vertical interpolator setup functions */

DEF_FLS_RTN(hssvnone)
{
	f->fb = f->s;
	f->fe = f->s;
}

DEF_FLS_RTN(hssvfir_3)
{
	const int8	*s;
	/* set up -1 and +1 with clipping */
	f->fa = (const int8*)f->s-f->mp;
	f->fa = MAX(f->fa,base);
	s = (const int8*)f->s+f->mp;
	f->fe = s>=(const int8*)f->e ? f->s:s;
	s = mg_filterQ6lut[PIN((int)(f->ry>>16-4)-0xF,0,0x1B-0xF)];
	s+= 3*((uint16)sy>>16-7), f->m[0] = *s, f->m[1] = s[1], f->m[2] = s[2];
	f->hy =-8; /* always need reset */
}

DEF_FLS_RTN(hssvbc)
{
	const int8	*e = (const int8*)f->e-f->mp;
	/* set up boxcar range; this will be [f->s,last] for easy while(s<=e) check */
	f->fe = (const int8*)base+(((sy+f->ry)>>16)-1)*f->mp;
	if ((const int8*)f->fe>e)// || f->fe<base)
		f->fe = e;
	f->hy =-8; /* always need reset */
}

/* horizontal & vertical */

#define rxd3(x)		{i=qx>>16; if(i>=w) {d[x]=f->scv(f,w-1); return;} \
	--i; a=f->scv(f,MAX(i,0)); i+=3; fir3(a,s[i-2],s[MIN(i,w)-1],qx); \
	d[x]=a;}

#define rxd3q(x)	{i=qx>>16; if(i>=w) {d[x]=f->scv(f,w-1); return;} \
	--i; a=f->scv(f,MAX(i,0)); i+=3; fir3q(a,s[i-2],s[MIN(i,w)-1],qx); \
	d[x]=a;}

#define rxd1(x)		{i=qx>>16; if(i>=w) {d[x]=f->scv(f,w-1); return;} \
	d[x]=f->scv(f,i);}

void hsc3u32(cnvParam *f,
	uint32 qx)
{
	const int32 *s=f->fb;
	const uint32 k1=((const int32*)f->t)[_cvsg];
	const uint32 k2=((const int32*)f->t)[_cvrb];
	const uint32 r1=(~(k1*2)&k1)<<5;
	const uint32 r2=(~(k2*2)&k2)<<5;
	int32 *d=f->cache, w=f->sw, a, i;
	f->hx=qx, qx=qx*4*f->rx-f->sx;
	rxd3(0) qx+=f->rx;
	rxd3(1) qx+=f->rx;
	rxd3(2) qx+=f->rx;
	rxd3(3)
}

void hsc3qu32(cnvParam *f,
	uint32 qx)
{
	const int32 *s=f->fb;
	const uint32 k1=((const int32*)f->t)[_cvsg];
	const uint32 k2=((const int32*)f->t)[_cvrb];
	int32 *d=f->cache, w=f->sw, a, i;
	f->hx=qx, qx=qx*4*f->rx-f->sx;
	rxd3q(0) qx+=f->rx;
	rxd3q(1) qx+=f->rx;
	rxd3q(2) qx+=f->rx;
	rxd3q(3)
}

void hsc1u32(cnvParam *f,
	uint32 qx)
{
	int32 *d=f->cache, w=f->sw, i;
	f->hx=qx, qx=qx*4*f->rx-f->sx;
	i=qx>>16;
	if(i>=w)
		{*d=f->scv(f,w-1); return;}
	*d=f->scv(f,MAX(i,0)); qx+=f->rx;
	rxd1(1) qx+=f->rx;
	rxd1(2) qx+=f->rx;
	rxd1(3)
}

/*void hscluint16(cnvParam *f,
	uint32 qx)
{
	const uint16 *s=f->s;
	const uint32 k1=((const int32*)f->t)[_cvsg];
	const uint32 k2=((const int32*)f->t)[_cvrb];
	const uint32 r1=(~(k1*2)&k1)<<3-1;
	const uint32 r2=(~(k2*2)&k2)<<3-1;
	int32 *d=f->cache, w=f->sw, a, i;
	f->hx=qx, qx=qx*4*f->rx-f->sx+0x8000;
	rxdl(0,3) qx+=f->rx;
	rxdl(1,3) qx+=f->rx;
	rxdl(2,3) qx+=f->rx;
	rxdl(3,3)
}*/

#define hscalh(h)	\
	uint32 qx=o>>2; o&=3; \
	if(qx!=f->hx) hsc##h(f,qx);
#endif

/* box-car (QD direct) */
/* QuickDraw used divide instructions when they could have used recipricol multiply */

void hscbcuint32(cnvParam *f,
	uint32 qx)
{
	const uint32 *s=(const uint32*)f->fb;
	const uint32 gk=((const int32*)f->t)[_cvsg];
	const uint32 rb=((const int32*)f->t)[_cvrb];
	const uint32 rk=rb&(-gk), bk=rb^rk;
	int32 *d=f->cache, w=f->sw, i, j, c=4;
	uint32 minr=f->rx>>16, a, t, ar, ag;
	f->hx=qx; qx=qx*4*f->rx-f->sx;
	do {
		i=qx>>16; if(i>=w) {*d=f->scv(f,w-1); break;}
		qx+=f->rx; j=(qx>>16)-i; a=f->scv(f,i);
		if(j>1) { uint32 np=f->m[--j-minr+4];
			ar=(a&rk)>>12,ag=(a&gk)>>6,a&=bk;
			if (j&1)
			{
				i+=2; t=s[i=MIN(i,w)-1]; a+=t&bk; ag+=(t&gk)>>6; ar+=(t&rk)>>12;
				if (!--j)
					goto skip;
			}
			do
			{
				i+=2; t=s[i=MIN(i,w)-1]; a+=t&bk; ag+=(t&gk)>>6; ar+=(t&rk)>>12;
				i+=2; t=f->scv(f,i=MIN(i,w)-1); a+=t&bk; ag+=(t&gk)>>6; ar+=(t&rk)>>12;
			} while(j-=2);
		skip:	a=ufract32mul(a,np); a|= ufract32mul(ag,np)<<6&gk; a|= ufract32mul(ar,np)<<12&rk;
		}
		*d++=a;
	} while(--c);
}

void hscbcuint16(cnvParam *f,
	uint32 qx)
{
	const uint16 *s=f->fb;
	const int gh=((const int32*)f->t)[_cvsi];
	const int rh=((const int32*)f->t)[_cvsj]+gh;
	const uint32 gk=((const int32*)f->t)[_cvsg];
	const uint32 bk=(1u<<gh)-1;
	int32 *d=f->cache, w=f->sw, i, j, c=4;
	uint32 minr=f->rx>>16, a, t, ar, ag;
	f->hx=qx; qx=qx*4*f->rx-f->sx;
	do {
		i=qx>>16; if(i>=w) {*d=f->scv(f,w-1); break;}
		qx+=f->rx; j=(qx>>16)-i; a=f->scv(f,i);
		if(j>1) { uint32 np=f->m[--j-minr+4];
			ar=(a>>rh)&bk,ag=(a&gk)>>gh,a&=bk;
			if (j&1)
			{
				i+=2; i=MIN(i,w)-1;
				t=s[i]; a+=t&bk; ag+=(t&gk)>>gh; ar+=t>>rh&bk;
				if (!--j)
					goto skip;
			}
			do
			{
				i+=2; i=MIN(i,w)-1;
				t=s[i]; a+=t&bk; ag+=(t&gk)>>gh; ar+=t>>rh&bk;
				i+=2; i=MIN(i,w)-1;
				t=f->scv(f,i); a+=t&bk; ag+=(t&gk)>>gh; ar+=t>>rh&bk;
			} while(j-=2);
		skip:	a=ufract32mul(a,np); a|= ufract32mul(ag,np)<<gh; a|= ufract32mul(ar,np)<<rh;
		}
		*d++=a;
	} while(--c);
}

#define hscbc(h)	\
	uint32 qx=o>>2; o&=3; \
	if(qx!=f->hx) hscbc##h(f,qx);

/* indexed max-ix boxcar */
#define hscbcx(h)	\
	uint32 i, j, qx=o>>2; o&=3; \
	if(qx!=f->hx) { \
		const uint8 *s=f->s; \
		uint32 *d=(uint32*)f->cache,a,t,c,sqx, w=f->sw; \
		f->hx=qx; sqx=qx=qx*4*f->rx-f->sx; \
		d[3]=d[2]=d[1]=*d=0; \
		goto frt; \
		do { qx=sqx,d=(uint32*)f->cache; frt: c=4; \
		 do { \
			i=qx>>16; if(i>=w) break; \
			qx+=f->rx; j=(qx>>16)-i; a=(s[i>>h]>>(h ? ~i<<3-h&7:0))&((1<<(1<<3-h))-1); \
			if(j>1) { \
				if(j&1) { i+=2; i=MIN(i,w)-1; \
					t=(s[i>>h]>>(h ? ~i<<3-h&7:0))&((1<<(1<<3-h))-1), a=MAX(a,t); \
					if(!--j) goto skiph; \
				} \
				do { i+=2; i=MIN(i,w)-1; \
					t=(s[i>>h]>>(h ? ~i<<3-h&7:0))&((1<<(1<<3-h))-1), a=MAX(a,t); \
					i+=2; i=MIN(i,w)-1; \
					t=(s[i>>h]>>(h ? ~i<<3-h&7:0))&((1<<(1<<3-h))-1), a=MAX(a,t); \
				} while(j-=2); \
			} \
		 skiph:	*d=MAX(a,*d); \
		 } while(++d,--c); \
	 	} while(s+=f->mp,s<(const uint8*)f->fe); \
	}

// i+=3; i=MIN(i,w)-2;
//	t=*(const uint16*)(s+(i>>h)), t=be16_native(t)<<(h ? i<<3-h&7:0);
//	u=t>>16-(1<<3-h)&((1<<(1<<3-h))-1),a=MAX(a,u);
//	t=t>>16-(2<<3-h)&((1<<(1<<3-h))-1), a=MAX(a,t);

#define hscbc1()	\
	uint32 i, j, qx=o>>2; o&=3; \
	if(qx!=f->hx) { \
		const uint8 *s=f->s; \
		int32 *d=f->cache; \
		uint32 a,c,sqx, w=f->sw; \
		f->hx=qx; sqx=qx=qx*4*f->rx-f->sx; \
		d[3]=d[2]=d[1]=*d=0; \
		goto frt; \
		do { qx=sqx,d=f->cache; frt: c=4; \
		 do { \
			i=qx>>16; if(i>=w) break; \
			qx+=f->rx; j=(qx>>16)-i; a=(s[i>>3]>>(~i&7))&1; \
			if(j>1) { \
				if(j&1) { i+=2; i=MIN(i,w)-1; \
					a|=(s[i>>3]>>(~i&7))&1; \
					if(!--j) goto skiph; \
				} \
				do { i+=2; i=MIN(i,w)-1; \
					a|=(s[i>>3]>>(~i&7))&1; \
					i+=2; i=MIN(i,w)-1; \
					a|=(s[i>>3]>>(~i&7))&1; \
				} while(j-=2); \
			} \
		 skiph:	*d|=a; \
		 } while(++d,--c); \
	 	} while(s+=f->mp,s<(const uint8*)f->fe); \
	}

/* basic */

#define rfdx(h)		{j=q>>16,*d=s[j>>h]>>(~j<<3-h&7),q+=f->rx;}
#define rxdx(h,x)	{j=q>>16,i=j>>h; if(i>=w+0u) goto end; d[x]=s[i]>>(~j<<3-h&7),q+=f->rx;}

#define hscalx(h)	\
	uint32 q=o>>3, i, j; \
	o&=7; \
	if(q!=f->hx) { \
		const uint8 *s=f->s; int32 *d=f->cache; uint32 w=f->sw; \
		f->hx=q, q=q*8*f->rx-f->sx; \
		rfdx(h); \
		rxdx(h,1) \
		rxdx(h,2) \
		rxdx(h,3) \
		rxdx(h,4) \
		rxdx(h,5) \
		rxdx(h,6) \
		rxdx(h,7) \
	}

#if 1
/* FIR interpolate */
static DEF_CNV_RTN(pxrsCIntrhv32)
{
	hscalh(3qu32)
	return f->cache[o]^f->cm;
}

static DEF_CNV_RTN(pxrsCIntrhv24)
{
	hscalh(3u32)
	return f->cache[o]^f->cm;
}

static DEF_CNV_RTN(pxrsCIntrv32)
{
	hscalh(1u32)
	return f->cache[o]^f->cm;
}

/*static DEF_CNV_RTN(pxrsCIntrhv16)
{
	hscalh(luint16)
	return f->cache[o]^f->cm;
}*/
#endif

static DEF_CNV_RTN(pxrsCMerge32)
{
	hscbc(uint32)
	return f->cache[o]^f->cm;
}

static DEF_CNV_RTN(pxrsCMerge16)
{
	hscbc(uint16)
	return f->cache[o]^f->cm;
}

static DEF_CNV_RTN(pxrsCache32)
{
	hscali(uint32);
end:	return f->cache[o]^f->cm;
}

static DEF_CNV_RTN(pxrsBCache32)
{
	hscali(uint32);
end:	return _reverse32(f->cache[o])^f->cm;
}

static DEF_CNV_RTN(pxrsCache16)
{
	hscali(uint16);
end:	return f->cache[o]^f->cm;
}

static DEF_CNV_RTN(pxrsBCache16)
{
	hscali(uint16);
end:	return _reverse16(f->cache[o])^f->cm;
}

/* ---- scaled direct conversion */

#if 1
static DEF_CNV_RTN(pxrsDIntrhv32)
{
	uint32		*t = f->t, r, i;

	hscalh(3qu32);
	i = f->cache[o];
	return cnvDirect32(i,qx,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsDIntrhv24)
{
	uint32		*t = f->t, r, i;

	hscalh(3u32);
	i = f->cache[o];
	return cnvDirect32(i,qx,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsDIntrv32)
{
	uint32		*t = f->t, r, i;

	hscalh(1u32);
	i = f->cache[o];
	return cnvDirect32(i,qx,t,r)^f->cm;
}

/*static DEF_CNV_RTN(pxrsDIntrhv16)
{
	uint32		*t = f->t, r, i;

	hscalh(luint16);
	i = f->cache[o];
	return cnvDirect32(i,qx,t,r)^f->cm;
}*/
#endif

static DEF_CNV_RTN(pxrsDMerge32)
{
	uint32		*t, r, i;

	hscbc(uint32);
	i = f->cache[o], t = f->t;
	return cnvDirect32(i,qx,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsDMerge16)
{
	uint32		*t, r, i;

	hscbc(uint16);
	i = f->cache[o], t = f->t;
	return cnvDirect32(i,qx,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsDirect32)
{
	uint32		*t = f->t, r;

	hscali(uint32);
end:	i = f->cache[o];
	return cnvDirect32(i,q,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsBDirect32)
{
	uint32		*t = f->t, r;

	hscali(uint32);
end:	i = _reverse32(f->cache[o]);
	return cnvDirect32(i,q,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsDirect16)
{
	uint32		*t = f->t, r;

	hscali(uint16);
end:	i = f->cache[o];
	return cnvDirect(i,q,t,r)^f->cm;
}

static DEF_CNV_RTN(pxrsBDirect16)
{
	uint32		*t = f->t, r;

	hscali(uint16);
end:	i = _reverse16(f->cache[o]);
	return cnvDirect(i,q,t,r)^f->cm;
}

/* ---- scaled direct through gray */

#if 1
static DEF_CNV_RTN(pxrsGIntrhv32)
{
	uint32		*t = f->t, r, i;

	hscalh(3qu32);
	i = f->cache[o];
	return cnvGray32(i,qx,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsGIntrhv24)
{
	uint32		*t = f->t, r, i;

	hscalh(3u32);
	i = f->cache[o];
	return cnvGray32(i,qx,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsGIntrv32)
{
	uint32		*t = f->t, r, i;

	hscalh(1u32);
	i = f->cache[o];
	return cnvGray32(i,qx,t,r)>>1^f->cm;
}

/*static DEF_CNV_RTN(pxrsGIntrhv16)
{
	uint32		*t = f->t, r, i;

	hscalh(luint16);
	i = f->cache[o];
	return cnvGray32(i,qx,t,r)>>1^f->cm;
}*/
#endif

static DEF_CNV_RTN(pxrsGMerge32)
{
	uint32		*t, r, i;

	hscbc(uint32);
	i = f->cache[o], t = f->t;
	return cnvGray32(i,qx,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsGMerge16)
{
	uint32		*t, r, i;

	hscbc(uint16);
	i = f->cache[o], t = f->t;
	return cnvGray32(i,qx,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsGray32)
{
	uint32		*t = f->t, r;

	hscali(uint32);
end:	i = f->cache[o];
	return cnvGray32(i,q,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsBGray32)
{
	uint32		*t = f->t, r;

	hscali(uint32);
end:	i = _reverse32(f->cache[o]);
	return cnvGray32(i,q,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsGray16)
{
	uint32		*t = f->t, r;

	hscali(uint16);
end:	i = f->cache[o];
	return cnvGray(i,q,t,r)>>1^f->cm;
}

static DEF_CNV_RTN(pxrsBGray16)
{
	uint32		*t = f->t, r;

	hscali(uint16);
end:	i = _reverse16(f->cache[o]);
	return cnvGray(i,q,t,r)>>1^f->cm;
}

/* ---- direct through itable */

#if 1
static DEF_CNV_RTN(pxrsIDIntrhv32)
{
	uint32		*t = f->t, r, i;

	hscalh(3qu32);
	i = f->cache[o];
	return f->itb[cnvDirect32(i,qx,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIDIntrhv24)
{
	uint32		*t = f->t, r, i;

	hscalh(3u32);
	i = f->cache[o];
	return f->itb[cnvDirect32(i,qx,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIDIntrv32)
{
	uint32		*t = f->t, r, i;

	hscalh(1u32);
	i = f->cache[o];
	return f->itb[cnvDirect32(i,qx,t,r)^f->cm];
}

/*static DEF_CNV_RTN(pxrsIDIntrhv16)
{
	uint32		*t = f->t, r, i;

	hscalh(luint16);
	i = f->cache[o];
	return f->itb[cnvDirect32(i,qx,t,r)^f->cm];
}*/
#endif

static DEF_CNV_RTN(pxrsIDMerge32)
{
	uint32		*t, r, i;

	hscbc(uint32);
	i = f->cache[o], t = f->t;
	return f->itb[cnvDirect32(i,qx,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIDMerge16)
{
	uint32		*t, r, i;

	hscbc(uint16);
	i = f->cache[o], t = f->t;
	return f->itb[cnvDirect32(i,qx,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIDirect32)
{
	uint32		*t = f->t, r;

	hscali(uint32);
end:	i = f->cache[o];
	return f->itb[cnvDirect32(i,q,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIBDirect32)
{
	uint32		*t = f->t, r;

	hscali(uint32);
end:	i = _reverse32(f->cache[o]);
	return f->itb[cnvDirect32(i,q,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIDirect16)
{
	uint32		*t = f->t, r;

	hscali(uint16);
end:	i = f->cache[o];
	return f->itb[cnvDirect(i,q,t,r)^f->cm];
}

static DEF_CNV_RTN(pxrsIBDirect16)
{
	uint32		*t = f->t, r;

	hscali(uint16);
end:	i = _reverse16(f->cache[o]);
	return f->itb[cnvDirect(i,q,t,r)^f->cm];
}

/* ---- scaled direct through gray itable */

#if 1
static DEF_CNV_RTN(pxrsIGIntrhv32)
{
	uint32		*t = f->t, r, i;

	hscalh(3qu32);
	i = f->cache[o];
	return f->itb[cnvGray32(i,qx,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIGIntrhv24)
{
	uint32		*t = f->t, r, i;

	hscalh(3u32);
	i = f->cache[o];
	return f->itb[cnvGray32(i,qx,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIGIntrv32)
{
	uint32		*t = f->t, r, i;

	hscalh(1u32);
	i = f->cache[o];
	return f->itb[cnvGray32(i,qx,t,r)>>7^f->cm];
}

/*static DEF_CNV_RTN(pxrsIGIntrhv16)
{
	uint32		*t = f->t, r, i;

	hscalh(luint16);
	i = f->cache[o];
	return f->itb[cnvGray32(i,qx,t,r)>>7^f->cm];
}*/
#endif

static DEF_CNV_RTN(pxrsIGMerge32)
{
	uint32		*t, r, i;

	hscbc(uint32);
	i = f->cache[o], t = f->t;
	return f->itb[cnvGray32(i,qx,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIGMerge16)
{
	uint32		*t, r, i;

	hscbc(uint16);
	i = f->cache[o], t = f->t;
	return f->itb[cnvGray32(i,qx,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIGray32)
{
	uint32		*t = f->t, r;

	hscali(uint32);
end:	i = f->cache[o];
	return f->itb[cnvGray32(i,q,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIBGray32)
{
	uint32		*t = f->t, r;

	hscali(uint32);
end:	i = _reverse32(f->cache[o]);
	return f->itb[cnvGray32(i,q,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIGray16)
{
	uint32		*t = f->t, r;

	hscali(uint16);
end:	i = f->cache[o];
	return f->itb[cnvGray(i,q,t,r)>>7^f->cm];
}

static DEF_CNV_RTN(pxrsIBGray16)
{
	uint32		*t = f->t, r;

	hscali(uint16);
end:	i = _reverse16(f->cache[o]);
	return f->itb[cnvGray(i,q,t,r)>>7^f->cm];
}

/* ---- scaled indexed */

static DEF_CNV_RTN(pxrsCache8)
{
	hscali(uint8);
end:	return ((uint32*)f->t)[f->cache[o]];
}

static DEF_CNV_RTN(pxrsCache4)
{
	hscalx(1);
end:	return ((uint32*)f->t)[f->cache[o]&15];
}

static DEF_CNV_RTN(pxrsCache2)
{
	hscalx(2);
end:	return ((uint32*)f->t)[f->cache[o]&3];
}

static DEF_CNV_RTN(pxrsCache1)
{
	hscalx(3);
end:	return ((uint32*)f->t)[f->cache[o]&1];
}

static DEF_CNV_RTN(pxrsMerge8)
{
	hscbcx(0);
	return ((uint32*)f->t)[f->cache[o]];
}

static DEF_CNV_RTN(pxrsMerge4)
{
	hscbcx(1);
	return ((uint32*)f->t)[f->cache[o]&15];
}

static DEF_CNV_RTN(pxrsMerge2)
{
	hscbcx(2);
	return ((uint32*)f->t)[f->cache[o]&3];
}

static DEF_CNV_RTN(pxrsMerge1)
{
	hscbc1();
	return ((uint32*)f->t)[f->cache[o]&1];
}

/* ---- dithering */

/*
 Robert W. Floyd/Louis Steinberg dithering:
	actual = p[0]
	closest = closest_match(actual)
	p[0] = closest
	error = actual-closest
	p[ 1][0]+= error*7/16
	p[-1][1]+= error*3/16
	p[ 0][1]+= error*5/16
	p[ 1][1]+= error*1/16

 Crafting it here:
   Take the src/dst lines, convert to high precision (30-bit).
   All graphics ops are done on that buffer line.
   Dither with the above method.
   However, calculating differences for the next line requires extra bits;
  the colors are converted into signed pseudo-rgb48 colors.

   The modern approach alternates in direction (serpentine scanning),
  which is a bit too complex the way the implementation here is driven, expected.
*/

static DEF_CNV_RTN(pxroRGBDither)
{ /* convert to dst. pixel, then back to 30-bit for subtraction */
  /* already 10 per-channel; expanding won't be necessary for rgbRoutine;
    shouldn't be dithering to 10:11:10 dest. anyway
  */
	rgb48		r, *a;
	rgb30		v = ((const rgb30*)f->s)[o];
	uint32		u, x;
	int32		dr = v>>20&1023, dg = v>>10&1023, db = v&1023;

	u = (dg*16+DITHERVAR(f)->g)<<2;
#ifdef MDITHER_POSGREEN_COMP
	/* human perception of color is not linear;
	  ease up on positive green error when green's the dimmest;
	  otherwise, the pattern may appear brighter
	*/
	if (dr>dg*2 || db>dg*2)
		u-= (int)u>>5;
#endif
	if (u>0xFFFF)
		u = (int)~u>>31&0xFFFF;
	/* convert to destination pixel, pinning the components */
	dg = (int)u>>6;
	r.green = (uint16)u;
	if ((u=(dr*16+DITHERVAR(f)->r)<<2)>0xFFFF)
		u = (int)~u>>31&0xFFFF;
	dr = (int)u>>6;
	r.red = (uint16)u;
	if ((u=(db*16+DITHERVAR(f)->b)<<2)>0xFFFF)
		u = (int)~u>>31&0xFFFF;
	db = (int)u>>6;
	r.blue = (uint16)u;
	x = DITHERVAR(f)->cc(DITHERVAR(f)->i,DITHERVAR(f)->gt,&r);

	{ /* convert back */
		uint32		*t = f->t, b;

		if (f->cc-directSrc+0u<dstInverse)
			u = cnvDirect(x,u,t,b); /* rgbRoutines don't bswap */
		else
			u = t[x];
	}

	/* get the 'next line' of differences */
	a = DITHERVAR(f)->a+o;

	/* can't do any merging tricks (red/blue in one) due to the multiplication, high prec. */
	/* error = actual-closest */
	dr-= u>>20&1023;
	dg-= u>>10&1023;
	db-= u&1023;
	/* p[ 1][0]+= error*7/16+p[1][1] */
	DITHERVAR(f)->r = dr*7+(int16)a[2].red;
	DITHERVAR(f)->g = dg*7+(int16)a[2].green;
	DITHERVAR(f)->b = db*7+(int16)a[2].blue;

	a->red+= dr*3, a->green+= dg*3, a->blue+= db*3; 	/* p[-1][1]+= error*3/16 */
	a[1].red+= dr*5, a[1].green+= dg*5, a[1].blue+= db*5;	/* p[ 0][1]+= error*5/16 */
	a[2].red = dr, a[2].green = dg, a[2].blue = db;		/* p[ 1][1]+= error*1/16 */
	return x;
}

static DEF_CNV_RTN(pxroGrayDither)
{
	rgb30		v = ((const rgb30*)f->s)[o];
	uint32		u, x;
	int32		*a, dr = v>>20&1023, dg = v>>10&1023, db = v&1023;
	//(((dr*306+dg*601+db*117)>>6)
	if ((dg=((int)formGray(dr,dg,db,u)*8+DITHERVAR(f)->g)/16)>0x3FFu)
		dg = (int)~dg>>31&0x3FF;
	x = ((const uint8*)DITHERVAR(f)->gt)[dg];
	u = ((const uint32*)f->t)[x];

	/* get the 'next line' of differences */
	a = (int32*)DITHERVAR(f)->a+o;

	dg-= u&1023;	/* error = actual-closest */
	DITHERVAR(f)->g = dg*7+a[2]; /* p[ 1][0]+= error*7/16+p[1][1] */
	*a+= dg*3; 	/* p[-1][1]+= error*3/16 */
	a[1]+= dg*5;	/* p[ 0][1]+= error*5/16 */
	a[2] = dg;	/* p[ 1][1]+= error*1/16 */
	return x;
}

mlvParam *rInitDithering(pixMap *dm)
{
	static pixMap	m30 =
		{1, 1, {0, 0}, 4, 32, pTypeNativeEndian|pTypeRawRGB, 30, 10};
	mlvParam	*f;
	void		*gt;
	int		cc, u;

	if (!dm)
		return NULL;
	/* what seems to be the adequate allocation size isn't quite right */
	u = sizeof(mlvParam)+(dm->w)*(sizeof(rgb30)+sizeof(rgb48)*3);
	if ((cc=pmGetRGBCC(dm,&gt))<0 || (f=calloc(u,1))==NULL)
		return NULL;
	/* the easy way to get the dest. pixels */
	u = rInitCnv(&f->d2m,&m30,dm->pTable,dm->pSize<<8|(dm->minCmp ? dm->minCmp:dm->bpp)
		|(dm->pType&(pTypeLSB|pTypeLGreen)),0x10000,mSrcPat*2,0);
	if (u<0)
		goto fail;
	if (rInitCnv(&f->m2d,dm,NULL,0x200A|pTypeNativeEndian,0x10000,mSrcPat*2,0)<0)
	{
		rEndCnv(&f->d2m);
	fail:	free(f);
		return NULL;
	}
	DITHERVAR(&f->m2d)->i = dm->minCmp;
	DITHERVAR(&f->m2d)->gt = gt;
	DITHERVAR(&f->m2d)->cc = rgbCCRoutine[cc];
	f->m2d.rtn = dm->minCmp ? pxroRGBDither:pxroGrayDither;
	f->m2d.s = (void*)(f+1);
	DITHERVAR(&f->m2d)->a = (rgb48*)((rgb30*)(f+1)+dm->w);
	return f;
}

/* ---- setup */

static cnvRoutine	f_cnvx[] =
{
	pxroCache1, pxroCache2, pxroCache4, pxro8,
	pxrsCache1, pxrsCache2, pxrsCache4, pxrsCache8,

	pxro16, pxroB16, pxro32, pxroB32,
	pxroDirect16, pxroBDirect16, pxroDirect32, pxroBDirect32,
	pxroGray16, pxroBGray16, pxroGray32, pxroBGray32,
	pxroIDirect16, pxroIBDirect16, pxroIDirect32, pxroIBDirect32,
	pxroIGray16, pxroIBGray16, pxroIGray32, pxroIBGray32,

	pxrsCache16, pxrsBCache16, pxrsCache32, pxrsBCache32,
	pxrsDirect16, pxrsBDirect16, pxrsDirect32, pxrsBDirect32,
	pxrsGray16, pxrsBGray16, pxrsGray32, pxrsBGray32,
	pxrsIDirect16, pxrsIBDirect16, pxrsIDirect32, pxrsIBDirect32,
	pxrsIGray16, pxrsIBGray16, pxrsIGray32, pxrsIBGray32,

	pxrsMerge1, pxrsMerge2, pxrsMerge4, pxrsMerge8,
	pxrsCMerge32, pxrsDMerge32, pxrsGMerge32, pxrsIDMerge32, pxrsIGMerge32,
	pxrsCMerge16, pxrsDMerge16, pxrsGMerge16, pxrsIDMerge16, pxrsIGMerge16,
#if 1
	pxrsCIntrv32, pxrsDIntrv32, pxrsGIntrv32, pxrsIDIntrv32, pxrsIGIntrv32,
	pxrsCIntrhv32, pxrsDIntrhv32, pxrsGIntrhv32, pxrsIDIntrhv32, pxrsIGIntrhv32,
	pxrsCIntrhv24, pxrsDIntrhv24, pxrsGIntrhv24, pxrsIDIntrhv24, pxrsIGIntrhv24
#endif
};

/* this only handles implied-index tables */
static boolean equalColors(const cTab64 *cta,
	const cTab64 *ctb)
{
	const cSpec	*a, *b;
	int		c;

	if ((c=MIN(cta->count,256))!=0 && cta->flags&ctb->flags&0x8000
			&& c==MIN(ctb->count,256))
	{
		a = cta->colors, b = ctb->colors;
		do
			if (a->rgb.red!=b->rgb.red || *(const int32*)&a->rgb.green
					!=*(const int32*)&b->rgb.green)
				return 0;
		while(++a,++b,--c);
		return 1;
	}
	return 0;
}

int rInitCnv(cnvParam *f,
	pixMap *m,
	const cTab64 *srcct,
	int srcfmt,
	ufixed iscale, /* inverse scale */
	uint8 mode,
	uint8 flag)
{ /* depth conversion or remap needed */
	cnvParam	*cf;
	uint32		*t;
	ifixed		ry;
	int		cc, i, j;
	uint8		mcc = 0;

	ry = f->ry;
	f->ssv = hssvnone;
	if ((uint16)srcfmt<0x900)
	{ /* make table */
		int		c, pmode;
		uint16		sizefmt;
		uint8		down = mode&mMergeDown;

		sizefmt = 0x2000|remap32bit;
		if ((uint8)srcfmt>4 && iscale==0x10000)
		{ /* arithmetic modes only supported in 8-bit and up */
			if (m->pSize&31 && !(flag&(remapForce30|remap32bit)))
				sizefmt = 0x1000;
			if (!down)
				mode-= mSrcPat; /* why? */
		}
		cc = pDepthX[srcfmt>>8&31];
		/* equal set? */
		if (m->bpp==(srcfmt>>8&0xFF) && m->pTable && (pmode=mode&15)<mTransparent
				&& equalColors(m->pTable,srcct))
			goto nix;
		if ((t=rGenRemappingTable(m,srcct,srcfmt&31,flag|(uint8)sizefmt))==NULL)
			return srcfmt&31 ? -2:-1;
		f->t = t;
		f->ac = sizefmt>>12;
		if (ry>0x10000 && down)
		{
			f->ssv = hssvbc;
			cc+= indexedMerge;
		} else if (iscale!=0x10000)
			cc+= iscale>0x10000 && down ? indexedMerge:indexedScaled;
		else if ((uint8)srcfmt==m->bpp && ((pmode=mode&15)<mTransparent || m->bpp==1))
		{ /* if src/dst depths are the same & the table comes out an identity matrix
		    (a no-change table), then dispose of the table and pipe it normally */
			i = 0, c = 1u<<(uint8)srcfmt;
			if (sizefmt&remap32bit)
				do
					if (*t++!=i)
						goto sett;
				while(++i<c);
			else
				do
					if (va_ne16(t)!=(uint16)i)
						goto sett;
				while(++i<c);
			free(f->t);
			f->t = NULL, f->ac = 0;
		nix:	if (pmode==mTransparent)
				mode = mSrcOr;
			mode&=~0x30;
		}
	sett:	f->cm = 0, cf = f->t;
		goto spos;
	}
	cc = ((srcfmt>>12&2)|((srcfmt&pTypeLSB)>>pTypeLSBBit^!BIG_ENDIAN))+(directSrc+4);
	if (m->pSize<9)
	{
		if (!(flag&~1))
			cc+= dstInverse;
		if (!m->pTable || !m->minCmp)
		{
			if (!(flag&remapForce30))
				cc+= dstGray;
			if (!m->iTable && pmMakeITables(m,0,0)<=0) /* default accuracy for performance */
				return -2;
			f->itb = m->iTable;
		}
	} else if (!(flag&remapForce30) && !(uint16)(srcfmt^formPxFmt(m)))
		cc-= 4;
	if (iscale!=0x10000)
		cc+= directScaled;
	if (flag&remapForce30)
		i = j = 10;
	else
		i = m->minCmp, j = i+(m->pType>>pTypeLGreenBit&1);
	mcc = ((mode&mMergeDown)>>1)|(srcfmt>>15&mMergeUp>>1);
	if (cc-directSrc<4u && !(srcfmt&0xC00000))
		{t = NULL; goto sinv;}
	/* direct mode source */
	if ((t=calloc(_cvdSize+(mcc && ry!=0x10000 ? f->sw:0),4))==NULL)
		return -2;
	f->t = t, f->ac = 4;
	{
		int		si, di;

		t[_cvsi] = si = srcfmt&0x7F;
		t[_cvsj] = di = si+(srcfmt>>pTypeLGreenBit&1);
		t[_cvdi] = si*2;
		t[_cvdj] = di*2;
		t[_cvki] =~(~0u>>si);
		t[_cvkj] =~(~0u>>di);
		t[_cvub] = 32-si;
		t[_cvug] = t[_cvub]-di;
		t[_cvur] = t[_cvug]-si;
		t[_cvsg] =~(~0u<<di)<<si;
		t[_cvrb] = t[_cvsg]^((1<<(t[_cvdi]+di))-1);
	}
	t[_cvhb] = 32-i;
	t[_cvhg] = t[_cvhb]-j;
	t[_cvhr] = t[_cvhg]-i;
	t[_cvmg] =~(~0u<<j)<<i;
	t[_cvmr] =~(~0u<<i)<<i<<j;
sinv:	f->cm = flag&1 ? i ? (1u<<(i*3+(m->pType>>pTypeLGreenBit&1)))-1:1023:0;
	/* handle scaling interpolation */
	if ((uint16)srcfmt>=0x1000 && !(cc&directBSwap))
	{
		ifixed		ry = f->ry;
		boolean		down;

		if (!mcc)
			goto spos; /* caller requested no interpolation */
		down = mcc&mMergeDown>>1;
		if ((uint16)srcfmt<0x2000)
			mcc = down;
		f->scv = (uint16)srcfmt>=0x2000 ? hscvu32:hscvu16;
		if (ry!=0x10000)
		{
			f->fb = (int32*)t+_cvdSize;
			if (mcc>=(mMergeUp>>1) && ry<0x28000)
			{
				f->ssv = hssvfir_3;
				f->scv = (srcfmt&15)>8 || (srcfmt&(15|pTypeLGreen))==(8|pTypeLGreen) ?
					hscvfir32:hscvfir24;
			} else if (ry>0x10000)
			{
				f->ssv = hssvbc;
				if (ry>0x20000)
				{
					f->scv = (uint16)srcfmt>=0x2000 ? hscvbc32:hscvbc16;
					i = ry>>16;
					f->m[0]=(0x80000000u-1+i-1)/(i+0u)*2; ++i;
					f->m[1]=(0x80000000u-1+i-1)/(i+0u)*2;
					//f->m[2]=++i<2 ? ~0u:uhl32div(1,i-1,i);
				} else
					f->scv = (uint16)srcfmt>=0x2000 ? hscvbc32_2:hscvbc16_2;
			}
		}
		if ((mcc==(mMergeDown>>1) && iscale>0x10000) || iscale>=0x28000)
		{
			i = iscale>>16;
			f->m[3]=i<2 ? ~0u:(0x80000000u-1+i-1)/(i+0u)*2;
			f->m[4]=++i<2 ? ~0u:(0x80000000u-1+i-1)/(i+0u)*2;
			cc = ((uint16)srcfmt>=0x2000 ? directMerge32:directMerge16)+((cc-28)>>2&7);
			goto spos;
		}
	#if 1
		if (cc>=directScaled
				&& iscale!=0x10000 /* indexed dst bugged without this */
				&& ((iscale>0x10000 && down)
				|| mcc>=(mMergeUp>>1)))
		{
			cc = (/*(uint16)srcfmt>=0x2000*/ /* expanded multiply only necessary for >8:8:8 */
				(srcfmt&15)>8 || (srcfmt&(15|pTypeLGreen))==(8|pTypeLGreen)
				? directSmooth32:directSmooth24/*:directSmooth16*/)
				+((cc-(directScaled+directSrc))>>2&7);
			i = (iscale>>16-4)-0xF;
			f->fx = mg_filterQ6lut[PIN(i,0,0x18-0xF)];
		} else if (ry!=0x10000 && (uint16)srcfmt>=0x2000)
			cc = (/*(uint16)srcfmt>=0x2000 ?*/ directSmoothV32/*:directSmoothV16*/)
				+((cc-(cc>=directScaled ? directScaled+directSrc:directSrc))>>2&7);
	#endif
	}
spos:	f->cc = (uint8)cc|mcc;
	f->rtn = f_cnvx[cc];
	f->hx =~0, f->rx = iscale;
	return mode;
}

void rEndCnv(cnvParam *f)
{
	if (!f->ac)
		return;
	free(f->t); f->t = NULL;
	f->ac = 0;
}

/* mask */
/* binary (1-bit mask esp.): ((b^a)&k)^a */

/* ---- basic conversion */

const int32	_b1to8t[16] =
{
	_be32_native(0x00000000),_be32_native(0x00000001),
	_be32_native(0x00000100),_be32_native(0x00000101),
	_be32_native(0x00010000),_be32_native(0x00010001),
	_be32_native(0x00010100),_be32_native(0x00010101),
	_be32_native(0x01000000),_be32_native(0x01000001),
	_be32_native(0x01000100),_be32_native(0x01000101),
	_be32_native(0x01010000),_be32_native(0x01010001),
	_be32_native(0x01010100),_be32_native(0x01010101)
};

#if !(envx86 || envx64)
void __fastcall b1to8(uint8 *d,
	const uint8 *s,
	int c)
{
	if (c<=0)
		return;
	do
	{
		uint8		v = (++s)[-1];

		*(int32*)d = _b1to8t[v>>4];
		((int32*)d)[1] = _b1to8t[v&15];
	} while(d+=8,--c);
}
#endif

void __fastcall rb1to8(uint8 *d,
	const uint8 *s,
	int c)
{
	if (c<=0)
		return;
	s+= c, d+= c*8;
	do
	{
		uint8		v = *--s;

		*(int32*)(d-=8) = _b1to8t[v>>4];
		((int32*)d)[1] = _b1to8t[v&15];
	} while(--c);
}

void __fastcall b2to8(uint8 *d,
	const uint8 *s,
	int c)
{
	if (c<=0)
		return;
	if (s==d)
	{
		s+= c, d+= c*4;
		do
		{
			uint8		v = *--s;

			*(d-=4) = v>>6;
			d[1] = v>>4&3;
			d[2] = v>>2&3;
			d[3] = v&=3;
		} while(--c);
		return;
	}
	do
	{
		uint8		v = *s++;

		*d = v>>6;
		d[1] = v>>4&3;
		d[2] = v>>2&3;
		d[3] = v&=3;
	} while(d+=4,--c);
}

void __fastcall b4to8(uint8 *d,
	const uint8 *s,
	int c)
{
	uint8		v;

	if (c<=0)
		return;
	if (s==d)
	{
		s+= c, d+= c*4;
		if (c&1)
			v = *--s, *(d-=2) = v>>4, d[1] = v&=15, d+= 2;
		if (c>>=1)
			do
			{
				v = *(s-=2), *(d-=4) = v>>4, d[1] = v&=15;
				v = s[1], d[2] = v>>4, d[3] = v&=15;
			} while(--c);
		return;
	}
	if (c&1)
		v = *s, *d = v>>4, d[1] = v&=15, d+= 2, ++s;
	if (c>>=1)
		do
		{
			v = *s, *d = v>>4, d[1] = v&=15;
			v = s[1], d[2] = v>>4, d[3] = v&=15;
		} while(d+=4,s+=2,--c);
}
