/*
 mctl.c
  control objects (buttons, sliders, text boxes, etc.)

  Revised: 2024.02.13
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2016.08.19
  Added proportionally-sized scroll bar thumbs (yay).
2019.03.14-16
  Updates for full region clipping.
*/
/*
TODO
  add slider (base on scroll bar)
  option to tab through all modifiable items (space to change checkbox)
*/

#if defined(WIN32) || defined(WIN64)
 #include <windows.h>
 #include <wchar.h>
#else
 #include <string.h>
#endif
#define __MWIND_PRIVATE
#define __MGRAF_INTERNAL
#include "mtext.h"
#include "mevent.h"
#include "mwind.h"

#include "ftrig.h"
#define NO_TABLES
#include "interpreter.h"

/* ---- globals */

const rgb48	_defMCtlColors[3] =
{
	{0,0,0}, /* frame color */
	{0,0,0}, /* title color */
	{52428,52428,52428} /* content */
};

uint16		ctlWheelScrollCount = 64;

/* ---- code */

static inline void rgb48lerp(rgb48 *out,
	const rgb48 *a,
	const rgb48 *b,
	ifix16 m)
{
	int		v;

	v = ((b->red-a->red)*m>>8)+a->red, out->red = v>65535u ? ~v>>31:v;
	v = ((b->green-a->green)*m>>8)+a->green, out->green = v>65535u ? ~v>>31:v;
	v = ((b->blue-a->blue)*m>>8)+a->blue, out->blue = v>65535u ? ~v>>31:v;
}

static inline void rgb48amp(rgb48 *out,
	const rgb48 *a,
	ifix16 m)
{
	int		v;

	v = a->red*m>>8, out->red = MIN(v,65535);
	v = a->green*m>>8, out->green = MIN(v,65535);
	v = a->blue*m>>8, out->blue = MIN(v,65535);
}

static inline void rgb48add(rgb48 *out,
	const rgb48 *a,
	uint16 m)
{
	int		v;

	v = a->red+m, out->red = MIN(v,65535);
	v = a->green+m, out->green = MIN(v,65535);
	v = a->blue+m, out->blue = MIN(v,65535);
}

static int ctlGeneral(ctlObj *ctl,
	int msg,
	const void *data)
{
	pt2d16		pt, adja;
	int32		oldvalue;
	int		type;

	if ((uint8)ctl->hilite>=kControlDisabledPart || !ctl->visible) /* disabled or invisible */
		return 0;
	if ((msg&~1)==wcMsgTest)
		return ptInRect16(&ctl->bounds,*(const pt2d16*)data);
	if (data)
		pt = *(const pt2d16*)data;
	else
		get_mouse_loc(&pt.x,&pt.y);
	*(int32*)&adja = 0, oldvalue = ctl->value;
	if (ctl->trk)
	{
		*(int32*)&adja = ctl->rtn(ctl,wcMsgMarkMeas,(void*)oldvalue);
		adja.x = pt.x-adja.x, adja.y = pt.y-adja.y;
	}
	type = ctl->rtn(ctl,wcMsgGetType,0);
	if ((msg&~1)==wcMsgClick)
	{
		pt2d16		oldpt, pt;
		int		dn, part;

		if (data)
			pt = *(const pt2d16*)data;
		else
			get_mouse_loc(&pt.x,&pt.y);
		dn = part = ctl->rtn(ctl,wcMsgTest,&pt);
		if (part+0u>=kControlDisabledPart+0u)
			return 0;
		oldpt.x = oldpt.y = IMIN(int16);
		while(mouse_still_down())
		{
			if (dn!=ctl->hilite)
				wHiliteControl(ctl,dn);
			if (ctl->trk && (dn || (type==ctlTypeScrollBar && part==kControlLabelPart)))
			{
				pt.x-= adja.x, pt.y-= adja.y;
				if (part!=kControlLabelPart || *(int32*)&pt!=*(int32*)&oldpt)
				{
					int32		value = dn-1u<kControlDisabledPart-1u
						? ctl->rtn(ctl,wcMsgDragMeas,&pt):oldvalue;

					ctl->trk(ctl,dn,value);
				}
				oldpt = pt;
			}
			get_mouse_loc(&pt.x,&pt.y);
			dn = ctl->rtn(ctl,part==kControlLabelPart ? wcMsgWTest:wcMsgTest,&pt);
			if (dn!=part)
				dn = dn-1<254u && type==ctlTypeScrollBar && part==kControlLabelPart;
		}
		ctl->visible = 1;
		if (ctl->hilite)
		{ /* if within, reset hilite and return part */
			wHiliteControl(ctl,0);
			switch(type)
			{
			case ctlTypeCheckBox: return kControlCheckBoxPart;
			case ctlTypeRadioBtn: return kControlRadioButtonPart;
			case ctlTypePicture:
			case ctlTypePixMap: return kControlPicturePart;
			case ctlTypeIcon: return kControlIconPart;
			case ctlTypePushBtn: part = kControlButtonPart;
			}
			wMcsDelay(1000000/24);
			return part;
		} else if (type==ctlTypeScrollBar && part==1 && ctl->value!=oldvalue)
			wSetControlValue(ctl,oldvalue);
		return 0;
	}
	if (msg==wcMsgDispose)
	{
		free(ctl->data);
		ctl->data = NULL;
	}
	return -8;
}

static void ctlBoxText(ctlObj *ctl,
	const char *str,
	int z,
	const rgb48 *colors,
	int o)
{
	grafPort	*p = &ctl->wind->port;

	if (z>0 && p)
	{
		ffFace		oldtf = p->txFace;
		int16		ascent, descent;

		p->fgSolid = RGBToPen(p->map,colors+wCTitle);
		p->bkSolid = RGBToPen(p->map,colors+wCContent);
		rTextFace(p,ctl->txFont,ctl->txSize,ctl->txStyle);
		rGetFontInfo(&p->txFace,NULL,&ascent,&descent,NULL,NULL);
		rDrawText(p,str,z,ctl->bounds.top
			+((ctl->bounds.bottom-ctl->bounds.top+(ascent>>1)
			+descent)>>1),(ctl->bounds.left+o)<<16);
		p->txFace = oldtf;
	}
}

#define btnEmb1			(btnRnd-2+(btnRnd<5)+(btnRnd==6))

static void pushBtnCont(grafPort *p,
	const rect16 *frame,
	uint16 hilite,
	int dn,
	const rgb48 *color,
	const rgb48 *cont,
	rgb48 *embrcol,
	int btnRnd,
	uint8 embossing)
{
	const rgb48	*emb;
	rect2dq		b;
	SLRgn		clipr, *oldclip;
	int		t, w, hBtnRnd = btnRnd>>1, ihbrl, ihbrt, ihbrr, ihbrb;
	boolean		enabled = (uint8)hilite<kControlDisabledPart;
	/* fill and frame */
	p->pnSize.x =-1;

	FixRect(&b,frame->left,frame->top,frame->right,frame->bottom);

	p->fgSolid = RGBToPen(p->map,p->map->bpp==1 && dn ? &rgb_black:cont);
	if (rRoundRect(p,&b,btnRnd<<16,btnRnd<<16)<=0)
		return;
	/* we won't get here if there's no clipRgn */
	oldclip = NULL;
	copyRgn(&oldclip,p->clipRgn);
	clipr.size = RECT_RGN_SIZE;

	p->pnSize.x = p->pnSize.y = 0x10000;
	rgb48lerp(&p->fgColor,&p->fgColor,color+wCContent,.5*256);
	ihbrl = frame->left+hBtnRnd;
	ihbrt = frame->top+hBtnRnd;
	ihbrr = frame->right-hBtnRnd;
	ihbrb = frame->bottom-hBtnRnd;
	if (enabled)
	{
		if (dn)
		{
			clipr.bBox.left = ihbrl;
			clipr.bBox.top = ihbrt;
			*(int32*)&clipr.bBox.bottom = 0x7FFF7FFF;
		} else
		{
			*(uint32*)&clipr.bBox = 0x80008000u;
			clipr.bBox.right = ihbrr;
			clipr.bBox.bottom = ihbrb;
		}
		sectRgn(&p->clipRgn,&clipr);
	}
	rgb48lerp(&p->fgColor,&p->fgColor,color,.5*256);
	RGBFore(p,&p->fgColor);
	rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
	if (enabled)
	{
		copyRgn(&p->clipRgn,oldclip);
		if (dn)
		{
			*(uint32*)&clipr.bBox = 0x80008000u;
			clipr.bBox.right = ihbrr;
			clipr.bBox.bottom = ihbrb;
		} else
		{
			clipr.bBox.left = ihbrl;
			clipr.bBox.top = ihbrt;
			*(int32*)&clipr.bBox.bottom = 0x7FFF7FFF;
		}
		sectRgn(&p->clipRgn,&clipr);
		rgb48lerp(&p->fgColor,&p->fgColor,color,.5*256);
		RGBFore(p,&p->fgColor);
		rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
	}
	p->pnSize.x = p->pnSize.y = 0xC000;
	RGBFore(p,color);

	copyRgn(&p->clipRgn,oldclip);

	rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
	p->txMode = mTransparent|mTextGray;
	if (enabled)
	{
		rgb48		lt, oldoc = p->opColor;
	  	uint8		oldpm = p->pnMode;

		if (!embossing)
			goto yh;
		p->pnSize.x = p->pnSize.y = 1<<16;
		lt.red = MIN(color[2].red,52428)+13107;
		lt.green = MIN(color[2].green,52428)+13107;
		lt.blue = MIN(color[2].blue,52428)+13107;
		emb = dn ? &lt:color;
	  /* br emboss - 1 */
		rgb48lerp(&p->fgColor,emb,cont,.5*256);
		RGBFore(p,&p->fgColor);
		FixRect(&b,frame->left-1,frame->top-1,
			frame->right-1,frame->bottom-1);

		clipr.bBox.left = frame->left+btnEmb1;
		clipr.bBox.top = frame->top+btnEmb1;
		*(int32*)&clipr.bBox.bottom = 0x7FFF7FFF;
		sectRgn(&p->clipRgn,&clipr);

		rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
	  /* br emboss - 2 */
		rgb48lerp(&p->fgColor,&p->fgColor,cont,embossing>1 ? .5*256:.75*256);
		RGBFore(p,&p->fgColor);
		if (embrcol)
			*embrcol = p->fgColor;
		FixRect(&b,frame->left-2,frame->top-2,
			frame->right-2,frame->bottom-2);

		copyRgn(&p->clipRgn,oldclip);
		clipr.bBox.left = ihbrl;
		clipr.bBox.top = ihbrt;
		sectRgn(&p->clipRgn,&clipr);

		rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
	  /* tl emboss - 2 */
		rgb48lerp(&p->fgColor,&p->fgColor,cont,1.5*256);
		RGBFore(p,&p->fgColor);
		if (!p->fgSolid)
			RGBFore(p,cont);
		if (embossing>1)
		{
			FixRect(&b,frame->left+2,frame->top+2,frame->right+2,frame->bottom+2);
			w = hBtnRnd+(dn && btnRnd<5);

			copyRgn(&p->clipRgn,oldclip);
			*(uint32*)&clipr.bBox = 0x80008000u;
			clipr.bBox.right = frame->right-4; //-w-2
			clipr.bBox.bottom = frame->bottom-4;
			sectRgn(&p->clipRgn,&clipr);

			rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		}
	  /* tl emboss - 1 */
		if (dn)
			emb = color;
		else
		{
			if (embossing>1)
				p->pnMode = mBlend, p->opColor = system_colors[_system_gray53];
			emb = &lt;
		}
		rgb48lerp(&p->fgColor,emb,cont,.5*256);
		RGBFore(p,&p->fgColor);
		FixRect(&b,frame->left+1,frame->top+1,
			frame->right+1,frame->bottom+1);

		copyRgn(&p->clipRgn,oldclip);
		*(uint32*)&clipr.bBox = 0x80008000u;
		clipr.bBox.right = frame->right-btnEmb1-1;
		clipr.bBox.bottom = frame->bottom-btnEmb1-1;
		sectRgn(&p->clipRgn,&clipr);

		rRoundRect(p,&b,btnRnd<<16,btnRnd<<16);
		if (embossing>1 && btnRnd==3 && !dn && p->map->bpp>7)
		{ /* some subtle softening of the bottom-left and top-right corners */
			pt2dq		p0, p1;

			copyRgn(&p->clipRgn,oldclip);
			clipr.bBox.right+= 2;
			clipr.bBox.bottom+= 2;
			sectRgn(&p->clipRgn,&clipr);

			p->pnSize.x =-1;
			p->pnMode = mBlend;
			p->opColor.red = p->opColor.green = p->opColor.blue = 0x808;
			RGBFore(p,color+wCFrame);
			t = p->clipRgn->bBox.bottom;
			FixPt(&p0,frame->left+1,t-2);
			FixPt(&p1,frame->left+2,t-3);
			rLine(p,p0,p1);
			//RGBFore(p,chromo_colors+_rgb_red);
			FixPt(&p0,frame->right-4,frame->top+1);
			FixPt(&p1,frame->right-5,frame->top+2);
			rLine(p,p0,p1);
		}
		p->pnMode = oldpm, p->opColor = oldoc;
	yh:	p->txMode = mTransparent;
	}
	copyRgn(&p->clipRgn,oldclip);
	free(oldclip);
}

static DEF_CTL_RTN(pushBtnProc)
{
	penState	pen;
	grafPort	*p;
	int		w, dn = 0;

	if (!ctl)
		return -1;
	p = &ctl->wind->port;
	if (msg==wcMsgDraw)
	{
		rgb48		oldfg, oldbk, cont;
		const rgb48	*color;
		SLRgn		clipr, *oldclip;
		ffFace		oldtf = p->txFace;
		int16		ascent, descent;
		uint8		oldtm;

		if (!ctl->visible || !p->map)
			return 0;
		oldtm = p->txMode;
		oldfg = p->fgColor, oldbk = p->bkColor;
		color = ctl->data ? ctl->data:_defMCtlColors;
		rPenNormal(p,&pen);

		dn = (uint8)ctl->hilite-1u<254u;
		cont = p->bkColor = color[wCContent];
		if (dn)
			cont.red>>= 1, cont.green>>= 1, cont.blue>>= 1;
		pushBtnCont(p,&ctl->bounds,ctl->hilite,dn,color,&cont,NULL,4,2);

		p->fgSolid = RGBToPen(p->map,p->map->bpp==1 && dn ? &rgb_white:color+1);
		oldclip = NULL;
		copyRgn(&oldclip,p->clipRgn);
		clipr.size = RECT_RGN_SIZE;
		clipr.bBox = ctl->bounds;
		sectRgn(&p->clipRgn,&clipr);

		rTextFace(p,ctl->txFont,ctl->txSize,ctl->txStyle);
		rGetFontInfo(&p->txFace,NULL,&ascent,&descent,NULL,NULL);
		w = rTextWidth(p,(const char*)ctl->title+1,*ctl->title,0);
		rDrawText(p,(const char*)ctl->title+1,*ctl->title,
			((ctl->bounds.bottom-ctl->bounds.top+(ascent>>1)+descent)
			>>1)+ctl->bounds.top,(ctl->bounds.right+ctl->bounds.left-w)<<15);
		p->txFace = oldtf;

		copyRgn(&p->clipRgn,oldclip);
		free(oldclip);

		if (ctl==ctl->wind->ctlDefault)
		{ /* default round frame */
			rect2dq		outset;
			rect16		bounds = ctl->bounds;
			int		m;

			p->pnSize.x = p->pnSize.y = 0x30000;
			outset.top = subu(bounds.top<<16,0x40000-0x8000);
			outset.left = subu(bounds.left<<16,0x40000-0x8000);
			outset.bottom = addu(bounds.bottom<<16,0x40000+0x8000);
			outset.right = addu(bounds.right<<16,0x40000+0x8000);

			/* unfortunately, the mBlend mode method doesn't work in redraw or <=4bpp */
			/* so some interpolation with the back color is used */
			//p->opColor = system_colors[_system_gray53];

			m = (int)(.57*256);
			if ((uint8)ctl->hilite>=kControlDisabledPart)
			{
				RGBFore(p,&rgb_black);
				if (!PenGetGray(p,&p->fgSolid))
				{
					uint8		grayPat[8];

					_get_gray_pattern(grayPat);
					rPenPat(p,grayPat);
					m = 0;
				} else
					m = (int)(.87*256);
			}
			rgb48lerp(&cont,&rgb_black,&p->bkColor,m);
			RGBFore(p,&cont);

			rRoundRect(p,&outset,0x60000,0x60000);
			outset.top = subu(bounds.top<<16,0x30000-0x8000);
			outset.left = subu(bounds.left<<16,0x30000-0x8000);
			outset.bottom = addu(bounds.bottom<<16,0x30000+0x8000);
			outset.right = addu(bounds.right<<16,0x30000+0x8000);
			rgb48lerp(&cont,&rgb_black,&p->bkColor,(m*m>>8)-(int)(.25*256));
			RGBFore(p,&cont);
			rRoundRect(p,&outset,0x40000,0x40000);

			update_screen_window_rect(ctl->wind,ctl->bounds.left-8,ctl->bounds.top-8,
				ctl->bounds.right+8,ctl->bounds.bottom+8);
		}

		rRestorePen(p,&pen,1);
		p->txMode = oldtm;
		RGBFore(p,&oldfg);
		RGBBack(p,&oldbk);
		return 1;
	} else if ((msg&~1)==wcMsgTest)
	{ /* rough simulation of a "within round-rect" test */
		pt2d16		loc = *(pt2d16*)data;
		int		h, i;

		if ((w=loc.x-ctl->bounds.left)<0)
			return 0;
		if ((i=ctl->bounds.right-1-loc.x)<0)
			return 0;
		if (i<w)
			w = i;
		if ((h=loc.y-ctl->bounds.top)<0)
			return 0;
		if ((i=ctl->bounds.bottom-1-loc.y)<0)
			return 0;
		if (i<h)
			h = i;
		return w!=h || w>1 || h>1;
	} else if (msg==wcMsgGetType)
		return ctlTypePushBtn;
	else
		return ctlGeneral(ctl,msg,data);
}

/*
 FF00FF FFFFFF EEEEEE DDDDDD CCCCCC BBBBBB AAAAAA 999999
 888888 777777 666666 555555 333333 9F9F9F E6E6E6 000000
*/
static ifix16	amplist[10] =
{
	0xCCCC00/0xCCCC+.999,
	0xBBBB00/0xCCCC+.999,
	0xAAAA00/0xCCCC+.999,
	0x999900/0xCCCC+.999,
	0x888800/0xCCCC+.999,
	0x777700/0xCCCC+.999,
	0x666600/0xCCCC+.999,
	0x555500/0xCCCC+.999,
	0x333300/0xCCCC+.999
};

/* 16-color palette for radioBtn/checkBox image base */
int _imgContColors(cTab64 *ct,
	const rgb48 *cont,
	const rgb48 *back,
	boolean disabled)
{
	cSpec		*cs = ct->colors;
	int		i;

	ct->seed = 0;
	ct->count = 16;
	ct->flags = 0x8000;
	if (back)
	{
		cs->rgb = *back;
		rgb48amp(&cs[13].rgb,back,0x9F9F00/0xFFFF+.999);
		rgb48amp(&cs[14].rgb,back,0xDFDF00/0xFFFF+.999);
	}
	cs[15].rgb.red = cs[15].rgb.green = cs[15].rgb.blue = 0;
	if (!cont)
		cont = system_colors+_system_gray73;
	cs[3].rgb = *cont;
	rgb48add(&cs[1].rgb,cont,0x3333);
	rgb48add(&cs[2].rgb,cont,0x1919);
	i = 0;
	do
		rgb48amp(&cs[i+4].rgb,cont,amplist[i]);
	while(++i<9);
	if (disabled)
	{
		cs[7].rgb = cs[6].rgb = cs[5].rgb =
			cs[4].rgb = cs[2].rgb = cs[1].rgb = *cont;
		rgb48lerp(&cs[8].rgb,&cs[8].rgb,cont,.8*256);
		if (cont!=system_colors+_system_gray93)
		{
			rgb48amp(&cs[15].rgb,cont,.5*256);
			rgb48lerp(&cs[14].rgb,&cs[14].rgb,cont,.5*256);
			rgb48lerp(&cs[13].rgb,&cs[13].rgb,cont,.5*256);
			rgb48lerp(&cs[12].rgb,&cs[12].rgb,cont,.5*256);
			rgb48lerp(&cs[11].rgb,&cs[11].rgb,cont,.5*256);
			rgb48lerp(&cs[10].rgb,&cs[10].rgb,cont,.5*256);
			rgb48lerp(&cs[9].rgb,&cs[9].rgb,cont,.6*256);
		}
		return mTransparent|mTextGray;
	}
	return mTransparent;
};

static uint8	checkBoxMerge[360] =
{ /* based on (but isn't) the Appearance 1.0 checkBox control */
	0x11,0x11,0x11,0x11,0x15,0x00,0x99,0x99,0x99,0x99,0x96,0x00,
	0x13,0x33,0x33,0x33,0x38,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x44,0x44,0x44,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x58,0x88,0x88,0x88,0x88,0x00,0x65,0x55,0x55,0x55,0x55,0x00,
	0x11,0x11,0x11,0x11,0x15,0xFF,0x99,0x99,0x99,0x99,0x98,0xFF,
	0x13,0x33,0x33,0x33,0x5F,0xFD,0x96,0x66,0x66,0x66,0x6F,0xFD,
	0x13,0x33,0x33,0x35,0xFC,0x00,0x96,0x66,0x66,0x66,0xFF,0x00,
	0x13,0x33,0x33,0x5F,0xCB,0x00,0x96,0x66,0x66,0x6F,0xF9,0x00,
	0x1F,0xC3,0x35,0xFC,0x99,0x00,0x9F,0xF6,0x66,0xFF,0xB5,0x00,
	0x13,0xFF,0x5F,0xF9,0x58,0x00,0x96,0xFF,0x6F,0xFB,0x65,0x00,
	0x13,0x5F,0xFF,0x95,0x48,0x00,0x96,0x6F,0xFF,0xB6,0x65,0x00,
	0x13,0x36,0xF9,0x53,0x48,0x00,0x96,0x68,0xFB,0x66,0x65,0x00,
	0x13,0x44,0x66,0x44,0x48,0x00,0x96,0x66,0x86,0x66,0x65,0x00,
	0x58,0x88,0x88,0x88,0x88,0x00,0x65,0x55,0x55,0x55,0x55,0x00,
	0x11,0x11,0x11,0x11,0x15,0x00,0x99,0x99,0x99,0x99,0x96,0x00,
	0x13,0x33,0x33,0x33,0x38,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0xBB,0xBB,0xBB,0x68,0x00,0x96,0xCC,0xCC,0xCC,0x85,0x00,
	0x13,0xFF,0xFF,0xFF,0x88,0x00,0x96,0xFF,0xFF,0xFF,0x95,0x00,
	0x13,0x56,0x66,0x66,0x68,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x33,0x33,0x33,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x13,0x44,0x44,0x44,0x48,0x00,0x96,0x66,0x66,0x66,0x65,0x00,
	0x58,0x88,0x88,0x88,0x88,0x00,0x65,0x55,0x55,0x55,0x55
};

static DEF_CTL_RTN(checkBoxProc)
{
	grafPort	*p;
	penState	pen;
	int		u;

	if (!ctl)
		return -1;
	p = &ctl->wind->port;
	if (msg==wcMsgDraw)
	{
		pixMap		m =
			{24,30,{0,0},12,4,0,4,4,0,0,(int8*)checkBoxMerge};
		rect2dq		b;
		rgb48		oldfg, oldbk;
		SLRgn		clipr, *oldclip;
		const rgb48	*color;
		cSpec		cstor[17];
		cTab64		*ct = (cTab64*)&cstor;
		uint8		oldtm;

		if (!ctl->visible || !p || !p->map)
			return 0;
		oldfg = p->fgColor, oldbk = p->bkColor;
		oldtm = p->txMode;
		rPenNormal(p,&pen);

		oldclip = NULL;
		copyRgn(&oldclip,p->clipRgn);
		clipr.size = RECT_RGN_SIZE;
		clipr.bBox = ctl->bounds;
		sectRgn(&p->clipRgn,&clipr);

		color = ctl->data ? ctl->data:_defMCtlColors;
		RGBBack(p,&p->bkColor);
		rEraseRect(p,ctl->bounds.left,ctl->bounds.top,
			ctl->bounds.right,ctl->bounds.bottom);
		p->txMode = _imgContColors(ct,p->map->bpp==1 ? NULL:color+wCContent,&p->bkColor,
			(uint8)ctl->hilite>=kControlDisabledPart);
		m.pTable = ct;
		u = (ctl->bounds.top+ctl->bounds.bottom-12)>>1;
		p->bkColor = color[wCFrame];
		if ((uint8)ctl->hilite>=kControlDisabledPart)
			rgb48lerp(&p->bkColor,color,color+wCContent,.5*256);
		p->fgSolid = RGBToPen(p->map,&p->bkColor);
		FixRect(&b,ctl->bounds.left+2,u,ctl->bounds.left+14,u+12);
		p->pnSize.x = p->pnSize.y = 0x10000;
		rRect(p,&b);
		rBltPix(p,&m,((uint8)ctl->hilite-1u<254u)*12,PIN(ctl->value,0,2)*10,12,10,
			ctl->bounds.left+3,u+1,12,10,mTransparent);
		ctlBoxText(ctl,1+(const char*)ctl->title,*ctl->title,color,18);

		copyRgn(&p->clipRgn,oldclip);
		free(oldclip);

		rRestorePen(p,&pen,1);
		RGBFore(p,&oldfg);
		RGBBack(p,&oldbk);
		return 1;
	} else if (msg==wcMsgGetType)
		return ctlTypeCheckBox;
	return ctlGeneral(ctl,msg,data);
}

static uint8	radioBtnMerge[408] =
{ /* based on (but isn't) the Appearance 1.0 radioBtn control */
	0x00,0x0D,0xCF,0xFC,0xD0,0x00,0x00,0xDC,0xFF,0xCD,0x00,0x00,
	0x0E,0xB9,0x43,0x35,0x9B,0xE0,0xEB,0xCB,0xBB,0xBC,0xBE,0x00,
	0x0B,0x53,0x21,0x11,0x38,0xB0,0xBC,0xBA,0x99,0x88,0xBB,0x00,
	0xD9,0x31,0x11,0x22,0x35,0xBD,0xCB,0xA9,0x88,0x76,0x6B,0xD0,
	0xC4,0x21,0x12,0x23,0x45,0x8C,0xBA,0x98,0x87,0x65,0x58,0xC0,
	0xF3,0x11,0x22,0x34,0x56,0x8F,0xB9,0x88,0x76,0x55,0x46,0xF0,
	0xF3,0x12,0x23,0x45,0x56,0x8F,0xB9,0x87,0x65,0x54,0x46,0xF0,
	0xC5,0x12,0x34,0x55,0x67,0x9C,0xB8,0x76,0x55,0x44,0x57,0xC0,
	0xD9,0x33,0x45,0x56,0x78,0xBD,0xC8,0x65,0x54,0x44,0x6B,0xD0,
	0x0B,0x85,0x56,0x67,0x8A,0xB0,0xCB,0x65,0x44,0x56,0x8C,0x00,
	0x0E,0xBB,0x88,0x89,0xBB,0xE0,0xEC,0xB8,0x66,0x7B,0xCE,0x00,
	0x00,0x0D,0xCF,0xFC,0xD0,0x00,0x00,0xDC,0xFF,0xCD,0x00,0x00,
	0x0E,0xBB,0x98,0x88,0xBB,0xE0,0xEB,0xCB,0xBB,0xBC,0xBE,0x00,
	0x0B,0xA8,0x66,0x65,0x58,0xB0,0xBB,0xB9,0x99,0x88,0xBB,0x00,
	0xDB,0x87,0xBF,0xFA,0x43,0x9D,0xBB,0xAC,0xFF,0xC7,0x6B,0xD0,
	0xC9,0x6B,0xFF,0xFF,0xA1,0x5C,0xB9,0xCF,0xFF,0xFC,0x48,0xC0,
	0xF8,0x6F,0xFF,0xFF,0xF2,0x3F,0xB9,0xFF,0xFF,0xFF,0x56,0xF0,
	0xF8,0x6F,0xFF,0xFF,0xF2,0x3F,0xB9,0xFF,0xFF,0xFF,0x56,0xF0,
	0xC8,0x5A,0xFF,0xFF,0xA1,0x3C,0xB8,0xCF,0xFF,0xFC,0x46,0xC0,
	0xDB,0x54,0xAF,0xFA,0x33,0x9D,0xC8,0x7C,0xFF,0xC6,0x6B,0xD0,
	0x0B,0x83,0x12,0x21,0x35,0xB0,0xCB,0x64,0x55,0x46,0x8C,0x00,
	0x0E,0xB9,0x53,0x33,0x9B,0xE0,0xEC,0xB8,0x66,0x6B,0xCE,0x00,
	0x00,0x0D,0xCF,0xFC,0xD0,0x00,0x00,0xDC,0xFF,0xCD,0x00,0x00,
	0x0E,0xB9,0x43,0x35,0x9B,0xE0,0xEB,0xCB,0xBB,0xBC,0xBE,0x00,
	0x0B,0x53,0x21,0x11,0x38,0xB0,0xBC,0xBA,0x98,0x88,0xBB,0x00,
	0xD9,0x31,0x11,0x22,0x35,0xBD,0xCB,0xA9,0x88,0x76,0x6B,0xD0,
	0xC4,0x21,0x12,0x23,0x45,0x8C,0xBA,0x99,0x98,0x75,0x48,0xC0,
	0xF3,0x19,0xBB,0xBB,0xB6,0x8F,0xB9,0xCC,0xCC,0xCB,0x46,0xF0,
	0xF3,0x1C,0xFF,0xFF,0xC6,0x8F,0xB9,0xFF,0xFF,0xFF,0x46,0xF0,
	0xC5,0x12,0x45,0x66,0x67,0x9C,0xB8,0x76,0x55,0x44,0x57,0xC0,
	0xD9,0x33,0x45,0x56,0x78,0xB6,0xB8,0x65,0x54,0x44,0x6B,0xD0,
	0x0B,0x85,0x55,0x67,0x8A,0xB0,0xBB,0x64,0x44,0x56,0x8B,0x00,
	0x0E,0xBB,0x88,0x89,0xBB,0xE0,0xEB,0xB8,0x66,0x7B,0xBE,0x00,
	0x00,0x0D,0xCF,0xFC,0xD0,0x00,0x00,0xDC,0xFF,0xCD
};

static DEF_CTL_RTN(radioBtnProc)
{
	grafPort	*p;
	penState	pen;
	int		r;

	if (!ctl)
		return -1;
	p = &ctl->wind->port;
	if (msg==wcMsgDraw)
	{
		pixMap		m =
			{24,33,{0,0},12,4,0,4,4,0,0,(int8*)radioBtnMerge};
		rgb48		oldfg, oldbk;
		SLRgn		clipr, *oldclip;
		const rgb48	*color;
		cSpec		cstor[17];
		cTab64		*ct = (cTab64*)&cstor;
		uint8		oldtm;

		if (!ctl->visible || !p)
			return 0;
		oldfg = p->fgColor, oldbk = p->bkColor;
		oldtm = p->txMode;
		rPenNormal(p,&pen);

		oldclip = NULL;
		copyRgn(&oldclip,p->clipRgn);
		clipr.size = RECT_RGN_SIZE;
		clipr.bBox = ctl->bounds;
		sectRgn(&p->clipRgn,&clipr);

		color = ctl->data ? ctl->data:_defMCtlColors;
		RGBFore(p,&p->bkColor);
		if (rFillRect(p,ctl->bounds.left,ctl->bounds.top,
				ctl->bounds.right,ctl->bounds.bottom)>0)
		{
			uint8		hilite = (uint8)ctl->hilite;

			p->txMode = _imgContColors(ct,p->map->bpp==1
				? NULL:color+wCContent,&p->bkColor,
				hilite>=kControlDisabledPart);
			m.pTable = ct;
			r = (ctl->bounds.top+ctl->bounds.bottom-12)>>1;
			//p->bkColor = color[wCFrame];
			if (hilite>=kControlDisabledPart)
				rgb48lerp(&p->bkColor,color,color+wCContent,.5*256);
			rBltPix(p,&m,(ctl->hilite-1u<254u)*11,PIN(ctl->value,0,2)*11,12,12,
				ctl->bounds.left+2,r,12,12,mSrcCopy);
			ctlBoxText(ctl,1+(const char*)ctl->title,*ctl->title,color,18);
		}
		copyRgn(&p->clipRgn,oldclip);
		free(oldclip);

		p->txMode = oldtm;
		rRestorePen(p,&pen,1);
		RGBFore(p,&oldfg);
		RGBBack(p,&oldbk);
		return 1;
	} else if (msg==wcMsgGetType)
		return ctlTypeRadioBtn;
	r = ctlGeneral(ctl,msg,data);
	if (r>=0 && ctl->value)
	{
		ctlObj		*tctl;
		uint8		c;

		for(tctl=ctl->group,c=ctl->groupCount; tctl && c; tctl=tctl->next,--c)
			if (tctl->value)
				wSetControlValue(tctl,0);
	}
	return r;
}

/* scroll bar */

/* only functions and draws well at 16 pixels in perpendicular width/height */

 #define sb_xup			0
 #define sb_xdown		15
 #define sb_xthumb		30
 #define sb_xhilite		45
 #define sb_yvert		0
 #define sb_yhoriz		15

static uint8	scrollBarMerge[1426] =
{ /* loosely based on (but isn't) the Appearance 1.0 scrollBar */
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xF1,0x11,
	0x11,0x11,0x11,0x11,0x11,0x3F,0x11,0x11,0x11,0x11,0x11,0x11,
	0x13,0xF2,0x44,0x44,0x44,0x44,0x44,0x44,0xDF,0xAA,0xAA,0xAA,
	0xAA,0xAA,0xAA,0xA8,0xFA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0x8F,
	0x79,0x99,0x99,0x99,0x99,0x99,0x9A,0xFF,0xF1,0x22,0x22,0x22,
	0x22,0x22,0x22,0x5F,0x12,0x22,0x22,0x22,0x22,0x22,0x25,0xF4,
	0x77,0x77,0x77,0x77,0x77,0x77,0xAF,0xA8,0x88,0x88,0x88,0x88,
	0x88,0x85,0xFA,0x88,0x88,0x88,0x88,0x88,0x88,0x5F,0x9B,0xBB,
	0xBB,0xBB,0xBB,0xBB,0xBC,0xFF,0xF1,0x22,0x22,0x22,0x22,0x22,
	0x22,0x5F,0x12,0x22,0x22,0x22,0x22,0x22,0x25,0xF4,0x77,0x77,
	0x77,0x77,0x77,0x77,0xAF,0xA8,0x88,0x88,0x88,0x88,0x88,0x85,
	0xFA,0x88,0x88,0x88,0x88,0x88,0x88,0x5F,0x9B,0xBB,0xBB,0xBB,
	0xBB,0xBB,0xBC,0xFF,0xF1,0x22,0x22,0x22,0x22,0x22,0x22,0x5F,
	0x12,0x22,0x22,0x22,0x22,0x22,0x25,0xF4,0x77,0x77,0x77,0x77,
	0x77,0x77,0xAF,0xA8,0x88,0x88,0x88,0x88,0x88,0x85,0xFA,0x88,
	0x88,0x88,0x88,0x88,0x88,0x5F,0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,
	0xBC,0xFF,0xF1,0x22,0x22,0x25,0x52,0x22,0x22,0x5F,0x12,0x25,
	0x62,0x22,0x26,0x62,0x25,0xF4,0x77,0x77,0x77,0x77,0x77,0x77,
	0xAF,0xA8,0x88,0x88,0xAA,0x88,0x88,0x85,0xFA,0x88,0xAA,0x88,
	0x88,0xAA,0x88,0x5F,0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,0xBC,0xFF,
	0xF1,0x22,0x22,0x6F,0xF6,0x22,0x22,0x5F,0x12,0x2C,0xF6,0x22,
	0x6F,0xF2,0x25,0xF4,0x77,0x35,0x55,0x55,0x56,0x77,0xAF,0xA8,
	0x88,0x8A,0xFF,0xA8,0x88,0x85,0xFA,0x88,0xFF,0xA8,0x8A,0xFF,
	0x88,0x5F,0x9B,0xB5,0x77,0x77,0x77,0x9B,0xBC,0xFF,0xF1,0x22,
	0x26,0xFC,0xCF,0x62,0x22,0x5F,0x12,0x25,0xCF,0x66,0xFF,0x62,
	0x25,0xF4,0x77,0x7B,0xBB,0xBB,0xBB,0x87,0xAF,0xA8,0x88,0xAF,
	0xFF,0xFA,0x88,0x85,0xFA,0x88,0xAF,0xFA,0xAF,0xFA,0x88,0x5F,
	0x9B,0xBB,0xFF,0xFF,0xFF,0xFB,0xBC,0xFF,0xF1,0x22,0x6F,0xC5,
	0x5C,0xF6,0x22,0x5F,0x12,0x22,0x5C,0xFF,0xF6,0x22,0x25,0xF4,
	0x77,0x35,0x55,0x55,0x56,0x77,0xAF,0xA8,0x8A,0xFF,0xAA,0xFF,
	0xA8,0x85,0xFA,0x88,0x8A,0xFF,0xFF,0xA8,0x88,0x5F,0x9B,0xB5,
	0x77,0x77,0x77,0x9B,0xBC,0xFF,0xF1,0x22,0xCC,0x52,0x25,0xCC,
	0x22,0x5F,0x12,0x22,0x25,0xCF,0x62,0x22,0x25,0xF4,0x77,0x7B,
	0xBB,0xBB,0xBB,0x87,0xAF,0xA8,0x8F,0xFA,0x88,0xAF,0xF8,0x85,
	0xFA,0x88,0x88,0xAF,0xFA,0x88,0x88,0x5F,0x9B,0xBB,0xFF,0xFF,
	0xFF,0xFB,0xBC,0xFF,0xF1,0x22,0x55,0x22,0x22,0x55,0x22,0x5F,
	0x12,0x22,0x22,0x55,0x22,0x22,0x25,0xF4,0x77,0x77,0x77,0x77,
	0x77,0x77,0xAF,0xA8,0x8A,0xA8,0x88,0x8A,0xA8,0x85,0xFA,0x88,
	0x88,0x8A,0xA8,0x88,0x88,0x5F,0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,
	0xBC,0xFF,0xF1,0x22,0x22,0x22,0x22,0x22,0x22,0x5F,0x12,0x22,
	0x22,0x22,0x22,0x22,0x25,0xF4,0x77,0x77,0x77,0x77,0x77,0x77,
	0xAF,0xA8,0x88,0x88,0x88,0x88,0x88,0x85,0xFA,0x88,0x88,0x88,
	0x88,0x88,0x88,0x5F,0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,0xBC,0xFF,
	0xF1,0x22,0x22,0x22,0x22,0x22,0x22,0x5F,0x12,0x22,0x22,0x22,
	0x22,0x22,0x25,0xF4,0x77,0x77,0x77,0x77,0x77,0x77,0xAF,0xA8,
	0x88,0x88,0x88,0x88,0x88,0x85,0xFA,0x88,0x88,0x88,0x88,0x88,
	0x88,0x5F,0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,0xBC,0xFF,0xF1,0x22,
	0x22,0x22,0x22,0x22,0x22,0x5F,0x12,0x22,0x22,0x22,0x22,0x22,
	0x25,0xF4,0x77,0x77,0x77,0x77,0x77,0x77,0xAF,0xA8,0x88,0x88,
	0x88,0x88,0x88,0x85,0xFA,0x88,0x88,0x88,0x88,0x88,0x88,0x5F,
	0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,0xBC,0xFF,0xF3,0x55,0x55,0x55,
	0x55,0x55,0x55,0x7F,0x35,0x55,0x55,0x55,0x55,0x55,0x57,0xFD,
	0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAF,0x85,0x55,0x55,0x55,0x55,
	0x55,0x55,0xF8,0x55,0x55,0x55,0x55,0x55,0x55,0x5F,0xAC,0xCC,
	0xCC,0xCC,0xCC,0xCC,0xCC,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xF1,0x11,0x11,0x11,0x11,0x11,0x11,0x3F,
	0x11,0x11,0x11,0x11,0x11,0x11,0x13,0xF2,0x44,0x44,0x44,0x44,
	0x44,0x44,0xDF,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xA8,0xFA,0xAA,
	0xAA,0xAA,0xAA,0xAA,0xAA,0x8F,0x79,0x99,0x99,0x99,0x99,0x99,
	0x9A,0xFF,0xF1,0x22,0x22,0x22,0x22,0x22,0x22,0x5F,0x12,0x22,
	0x22,0x22,0x22,0x22,0x25,0xF4,0x77,0x77,0x77,0x77,0x77,0x77,
	0xAF,0xA8,0x88,0x88,0x88,0x88,0x88,0x85,0xFA,0x88,0x88,0x88,
	0x88,0x88,0x88,0x5F,0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,0xBC,0xFF,
	0xF1,0x22,0x22,0x22,0x22,0x22,0x22,0x5F,0x12,0x22,0x22,0x22,
	0x22,0x22,0x25,0xF4,0x77,0x77,0x77,0x77,0x77,0x77,0xAF,0xA8,
	0x88,0x88,0x88,0x88,0x88,0x85,0xFA,0x88,0x88,0x88,0x88,0x88,
	0x88,0x5F,0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,0xBC,0xFF,0xF1,0x22,
	0x22,0x22,0x6F,0x52,0x22,0x5F,0x12,0x22,0x5F,0x62,0x22,0x22,
	0x25,0xF4,0x77,0x77,0x37,0x37,0x77,0x77,0xAF,0xA8,0x88,0x88,
	0x8A,0xFA,0x88,0x85,0xFA,0x88,0x8A,0xFA,0x88,0x88,0x88,0x5F,
	0x9B,0xBB,0xB5,0xB5,0xBB,0xBB,0xBC,0xFF,0xF1,0x22,0x22,0x26,
	0xFF,0x52,0x22,0x5F,0x12,0x22,0x5C,0xF6,0x22,0x22,0x25,0xF4,
	0x77,0x77,0x5B,0x5B,0x77,0x77,0xAF,0xA8,0x88,0x88,0xAF,0xFA,
	0x88,0x85,0xFA,0x88,0x8A,0xFF,0xA8,0x88,0x88,0x5F,0x9B,0xBB,
	0xB7,0xF7,0xFB,0xBB,0xBC,0xFF,0xF1,0x22,0x22,0x6F,0xF6,0x22,
	0x22,0x5F,0x12,0x22,0x25,0xCF,0x62,0x22,0x25,0xF4,0x77,0x77,
	0x5B,0x5B,0x77,0x77,0xAF,0xA8,0x88,0x8A,0xFF,0xA8,0x88,0x85,
	0xFA,0x88,0x88,0xAF,0xFA,0x88,0x88,0x5F,0x9B,0xBB,0xB7,0xF7,
	0xFB,0xBB,0xBC,0xFF,0xF1,0x22,0x25,0xFF,0x62,0x22,0x22,0x5F,
	0x12,0x22,0x22,0x5C,0xF5,0x22,0x25,0xF4,0x77,0x77,0x5B,0x5B,
	0x77,0x77,0xAF,0xA8,0x88,0xAF,0xFA,0x88,0x88,0x85,0xFA,0x88,
	0x88,0x8A,0xFF,0xA8,0x88,0x5F,0x9B,0xBB,0xB7,0xF7,0xFB,0xBB,
	0xBC,0xFF,0xF1,0x22,0x25,0xCF,0x62,0x22,0x22,0x5F,0x12,0x22,
	0x22,0x6F,0xF5,0x22,0x25,0xF4,0x77,0x77,0x5B,0x5B,0x77,0x77,
	0xAF,0xA8,0x88,0xAF,0xFA,0x88,0x88,0x85,0xFA,0x88,0x88,0x8A,
	0xFF,0xA8,0x88,0x5F,0x9B,0xBB,0xB7,0xF7,0xFB,0xBB,0xBC,0xFF,
	0xF1,0x22,0x22,0x6C,0xF6,0x22,0x22,0x5F,0x12,0x22,0x26,0xFF,
	0x62,0x22,0x25,0xF4,0x77,0x77,0x5B,0x5B,0x77,0x77,0xAF,0xA8,
	0x88,0x8A,0xFF,0xA8,0x88,0x85,0xFA,0x88,0x88,0xAF,0xFA,0x88,
	0x88,0x5F,0x9B,0xBB,0xB7,0xF7,0xFB,0xBB,0xBC,0xFF,0xF1,0x22,
	0x22,0x26,0xCF,0x52,0x22,0x5F,0x12,0x22,0x5F,0xF6,0x22,0x22,
	0x25,0xF4,0x77,0x77,0x5B,0x5B,0x77,0x77,0xAF,0xA8,0x88,0x88,
	0xAF,0xFA,0x88,0x85,0xFA,0x88,0x8A,0xFF,0xA8,0x88,0x88,0x5F,
	0x9B,0xBB,0xB7,0xF7,0xFB,0xBB,0xBC,0xFF,0xF1,0x22,0x22,0x22,
	0x6C,0x52,0x22,0x5F,0x12,0x22,0x5C,0x62,0x22,0x22,0x25,0xF4,
	0x77,0x77,0x6B,0x6B,0x77,0x77,0xAF,0xA8,0x88,0x88,0x8A,0xFA,
	0x88,0x85,0xFA,0x88,0x8A,0xFA,0x88,0x88,0x88,0x5F,0x9B,0xBB,
	0xB9,0xF9,0xFB,0xBB,0xBC,0xFF,0xF1,0x22,0x22,0x22,0x22,0x22,
	0x22,0x5F,0x12,0x22,0x22,0x22,0x22,0x22,0x25,0xF4,0x77,0x77,
	0x78,0x78,0x77,0x77,0xAF,0xA8,0x88,0x88,0x88,0x88,0x88,0x85,
	0xFA,0x88,0x88,0x88,0x88,0x88,0x88,0x5F,0x9B,0xBB,0xBB,0xBB,
	0xBB,0xBB,0xBC,0xFF,0xF1,0x22,0x22,0x22,0x22,0x22,0x22,0x5F,
	0x12,0x22,0x22,0x22,0x22,0x22,0x25,0xF4,0x77,0x77,0x77,0x77,
	0x77,0x77,0xAF,0xA8,0x88,0x88,0x88,0x88,0x88,0x85,0xFA,0x88,
	0x88,0x88,0x88,0x88,0x88,0x5F,0x9B,0xBB,0xBB,0xBB,0xBB,0xBB,
	0xBC,0xFF,0xF3,0x55,0x55,0x55,0x55,0x55,0x55,0x7F,0x35,0x55,
	0x55,0x55,0x55,0x55,0x57,0xFD,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,
	0xAF,0x85,0x55,0x55,0x55,0x55,0x55,0x55,0xF8,0x55,0x55,0x55,
	0x55,0x55,0x55,0x5F,0xAC,0xCC,0xCC,0xCC,0xCC,0xCC,0xCC,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF
};

/* works both vertically and horizontally */
static uint8	wellPat[16] =
	{12,10,8,6,6,6,6,6,6,6,6,6,6,5,4,15};
static uint8	wellVShadow[32] =
{
	12,10,10,10,10,10,10,10,10,10,10,10,10,10,9,15,
	12,10,8,8,8,8,8,8,8,8,8,8,8,8,7,15
};
static uint8	wellHShadow[32] =
{
	12,12,
	10,10,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	10,8,
	9,7,
	15,15
};

typedef struct sBarLocs
{
	int		tx, ty, dx, dy;
	boolean		vert;
} sBarLocs;

static int calcScrollBarLocs(const ctlObj *ctl,
	sBarLocs *locs,
	int32 value,
	int *thumbSize)
{
	int32		u, h;
	int		hilite = (uint8)ctl->hilite;
	boolean		vert;

	if ((u=RECTANGLE_WIDTH(&ctl->bounds))<=0
			|| (h=RECTANGLE_HEIGHT(&ctl->bounds))<=0)
		return locs ? -1:0;
	if ((vert=h>u)!=0)
		u = h;
	if (locs)
		locs->vert = vert;
	if ((*thumbSize=u-32-ctl->max+ctl->min)<16) //(u*3>>2)-ctl->max+ctl->min
		*thumbSize = 16;
	if ((u-=30+*thumbSize)<0)
		u = 0;
	if (ctl->min>=ctl->max)
	{
		if (!locs)
			goto end;
		hilite = kControlInactivePart, *thumbSize = u = 16;
		goto calc;
	}
	if (locs)
	{
		h = MAX(value,ctl->min);
		if (h>ctl->max)
			h = ctl->max;
		u = ufixmuldiv(u,h-ctl->min,ctl->max-ctl->min)+15;
	calc:	if (vert)
		{
			locs->tx = ctl->bounds.left, locs->ty = u+ctl->bounds.top;
			locs->dx = ctl->bounds.left, locs->dy = ctl->bounds.bottom-16;
		} else
		{
			locs->tx = u+ctl->bounds.left, locs->ty = ctl->bounds.top;
			locs->dx = ctl->bounds.right-16, locs->dy = ctl->bounds.top;
		}
	} else
	{
		if (value<0 || (value-=(vert ? ctl->bounds.top:ctl->bounds.left)+15)<0)
			value = 0;
		return ufixmuldiv(value,ctl->max-ctl->min,u);
	}
end:	return hilite;
}

static DEF_CTL_RTN(scrollBarProc)
{
	grafPort	*p;
	penState	pen;
	sBarLocs	locs;
	int		hilite, thumbSize;

	if (!ctl)
		return -1;
	if (msg==wcMsgGetType)
		return ctlTypeScrollBar;
	p = &ctl->wind->port;
	if ((hilite=calcScrollBarLocs(ctl,&locs,ctl->value,&thumbSize))<0)
		return 0;
	if (msg==wcMsgDragMeas)
		return calcScrollBarLocs(ctl,NULL,locs.vert
			? ((const pt2d16*)data)->y:((const pt2d16*)data)->x,&thumbSize);
	if (msg==wcMsgMarkMeas)
	{
		pt2d16		pt;

		pt.x = locs.tx!=(int16)locs.tx ? locs.tx>>31^IMAX(int16):locs.tx;
		pt.y = locs.ty!=(int16)locs.ty ? locs.ty>>31^IMAX(int16):locs.ty;
		return *(int32*)&pt;
	}
	if (msg==wcMsgDraw)
	{
		pixMap		m =
			{91,31,{0,0},46,4,0,4,4,0,0,(int8*)scrollBarMerge};
		rgb48		oldfg;
		SLRgn		clipr, *oldclip;
		const rgb48	*colors;
		cSpec		cstor[17];
		cTab64		*ct = (cTab64*)&cstor;
		int		dw, dh, sy;

		if (!ctl->visible || !p || !p->map)
			return 0;
		oldfg = p->fgColor;
		rPenNormal(p,&pen);

		oldclip = NULL;
		copyRgn(&oldclip,p->clipRgn);
		clipr.size = RECT_RGN_SIZE;
		clipr.bBox = ctl->bounds;
		sectRgn(&p->clipRgn,&clipr);

		colors = ctl->data ? ctl->data:_defMCtlColors;
		p->txMode = _imgContColors(ct,p->map->bpp==1 ? NULL:colors+wCContent,&rgb_white,
			(uint8)hilite>=kControlDisabledPart);
		if ((uint8)hilite>=kControlDisabledPart)
		{
			rgb48add(&cstor[2].rgb,&cstor[2].rgb,0x1111);
			cstor[11] = cstor[10] = cstor[9] = cstor[8] = cstor[7] =
				cstor[6] = cstor[5] = cstor[4] = cstor[3] = cstor[2];
		}
		m.pTable = ct;
		sy = locs.vert ? sb_yvert:sb_yhoriz;
		if (locs.vert || (dw=locs.tx-ctl->bounds.left)>16)
			dw = 16;
		if (!locs.vert || (dh=locs.ty-ctl->bounds.top)>16)
			dh = 16;
		rBltPix(p,&m,((uint8)hilite==kControlUpButtonPart)*sb_xhilite+sb_xup,sy,dw,dh,
			ctl->bounds.left,ctl->bounds.top,dw,dh,mSrcCopy);
		if ((uint8)hilite<kControlDisabledPart)
		{
			int		sx = ((uint8)hilite==kControlLabelPart)*sb_xhilite+sb_xthumb;
			int		t, dxy, hw = (thumbSize-13)>>1;

			if (thumbSize<17)
				rBltPix(p,&m,sx,sy,16,16,
					locs.tx,locs.ty,
					locs.vert ? dw:16,
					locs.vert ? 16:dh,
					mSrcCopy);
			else if (locs.vert)
			{ /* stretch out "blank" portions */
				dxy = locs.ty;
				rBltPix(p,&m,sx,sy,16,3,locs.tx,dxy,dw,3,mSrcCopy);
				dxy+= 3;
				rBltPix(p,&m,sx,sy+3,16,1,locs.tx,dxy,dw,hw,mSrcCopy);
				dxy+= hw;
				rBltPix(p,&m,sx,sy+4,16,8,locs.tx,dxy,dw,8,mSrcCopy);
				dxy+= 8, t = thumbSize-14-hw;
				rBltPix(p,&m,sx,sy+12,16,1,locs.tx,dxy,dw,t,mSrcCopy);
				dxy+= t;
				rBltPix(p,&m,sx,sy+13,16,3,locs.tx,dxy,dw,3,mSrcCopy);
			} else
			{
				dxy = locs.tx;
				rBltPix(p,&m,sx,sy,3,16,dxy,locs.ty,3,dh,mSrcCopy);
				dxy+= 3;
				rBltPix(p,&m,sx+3,sy,1,16,dxy,locs.ty,hw,dh,mSrcCopy);
				dxy+= hw;
				rBltPix(p,&m,sx+4,sy,8,16,dxy,locs.ty,8,dh,mSrcCopy);
				dxy+= 8, t = thumbSize-14-hw;
				rBltPix(p,&m,sx+12,sy,1,16,dxy,locs.ty,t,dh,mSrcCopy);
				dxy+= t;
				rBltPix(p,&m,sx+13,sy,3,16,dxy,locs.ty,3,dh,mSrcCopy);
			}
		}
		rBltPix(p,&m,((uint8)hilite==kControlDownButtonPart)*sb_xhilite+sb_xdown,sy,16,16,
			locs.dx,locs.dy,16,16,mSrcCopy);
		m.pSize = m.bpp = 8;
		if (locs.vert)
		{
			int		dy;

			m.h = 2, m.w = m.pitch = 16;
			dy = ctl->bounds.top+16;
			if ((uint8)hilite<kControlDisabledPart)
			{
				dh = locs.ty-dy;
				m.baseAddr = (int8*)wellVShadow;
				rBltPix(p,&m,0,0,16,2,ctl->bounds.left,dy,16,MIN(dh,2),0);
				m.baseAddr = (int8*)wellPat;
				rBltPix(p,&m,0,0,16,1,ctl->bounds.left,dy+2,16,dh-2,0);
				dy = locs.ty+thumbSize, dh = locs.dy-dy;
				m.baseAddr = (int8*)wellVShadow;
				rBltPix(p,&m,0,0,16,2,ctl->bounds.left,dy,16,MIN(dh,2),0);
				dy+= 2, dh-= 2;
			} else
				dh = locs.dy-dy;
			m.baseAddr = (int8*)wellPat;
			rBltPix(p,&m,0,0,16,1,ctl->bounds.left,dy,dw,dh,0);
		} else
		{
			int		dx;

			m.w = m.pitch = 2, m.h = 16;
			dx = ctl->bounds.left+16;
			if ((uint8)hilite<kControlDisabledPart)
			{
				dw = locs.tx-dx;
				m.baseAddr = (int8*)wellHShadow;
				rBltPix(p,&m,0,0,2,16,dx,ctl->bounds.top,MIN(dw,2),16,0);
				m.baseAddr = (int8*)wellPat;
				m.w = m.pitch = 1, m.h = 16;
				rBltPix(p,&m,0,0,1,16,dx+2,ctl->bounds.top,dw-2,16,0);
				dx = locs.tx+thumbSize, dw = locs.dx-dx;
				m.baseAddr = (int8*)wellHShadow;
				m.w = m.pitch = 2, m.h = 16;
				rBltPix(p,&m,0,0,2,16,dx,ctl->bounds.top,MIN(dw,2),16,0);
				dx+= 2, dw-= 2;
			} else
				dw = locs.dx-dx;
			m.baseAddr = (int8*)wellPat;
			m.w = m.pitch = 1, m.h = 16;
			rBltPix(p,&m,0,0,1,16,dx,ctl->bounds.top,dw,dh,0);
		}
		copyRgn(&p->clipRgn,oldclip);
		free(oldclip);

		rRestorePen(p,&pen,1);
		RGBFore(p,&oldfg);
		return 1;
	}
	if (msg==wcMsgGetType)
		return ctlTypeScrollBar;
	if ((uint8)hilite>=kControlDisabledPart)
		return 0;
	if ((msg&~1)==wcMsgTest)
	{
		rect16		wbounds = ctl->bounds;
		pt2d16		pt;
		int		r;

		if (data)
			pt = *(const pt2d16*)data;
		else
			get_mouse_loc(&pt.x,&pt.y);
		if (locs.vert)
		{
			if (msg&1)
				OutsetRect16(&wbounds,64,8);
			if (pt.y<wbounds.top || pt.y>=wbounds.bottom)
				return 0;
			if (pt.y<ctl->bounds.top+16)
				{r = kControlUpButtonPart; goto rchk;}
			if (pt.y<locs.ty)
				{r = kControlPageUpButtonPart; goto rchk;}
			if (pt.y<locs.ty+thumbSize)
				{r = kControlLabelPart; goto rchk;}
			if (pt.y<locs.dy)
				{r = kControlPageDownButtonPart; goto rchk;}
			r = kControlDownButtonPart;
			goto rchk;
		}
		if (msg&1)
			OutsetRect16(&wbounds,8,64);
		if (pt.x<wbounds.left || pt.x>=wbounds.right)
			return 0;
		if (pt.x<ctl->bounds.left+16)
			{r = kControlUpButtonPart; goto rchk;}
		if (pt.x<locs.tx)
			{r = kControlPageUpButtonPart; goto rchk;}
		if (pt.x<locs.tx+thumbSize)
			{r = kControlLabelPart; goto rchk;}
		if (pt.x<locs.dx)
			{r = kControlPageDownButtonPart; goto rchk;}
		r = kControlDownButtonPart;
	rchk:	return ptInRect16(&wbounds,pt) ? r:0;
	}
	if (msg==wcMsgInBounds)
	{
		*(rect16*)data = ctl->bounds;
		return 1;
	}
	if (msg==wcMsgWheel && ctl->group)
	{
		ctl = ctl->group;
		return ctl->rtn(ctl,msg,data);
	}
	return ctlGeneral(ctl,msg,data);
}

static DEF_TRK_RTN(sBarTrack)
{
	if (part==kControlLabelPart)
		wSetControlValue(ctl,value);
};

/* popup */

static DEF_CTL_RTN(popUpProc)
{
	grafPort	*p;
	penState	pen;
	HMENU		menu = ctl->data;
	int		dn, t, u, ind;

	if (!ctl)
		return -1;
	p = &ctl->wind->port;
	ind = MAX(ctl->labelIndent,0);
	u = (ctl->bounds.top+ctl->bounds.bottom)>>1;
	if (msg==wcMsgDraw)
	{
		char		str[260*sizeof(wchar_t)];
		pt2dq		p0, p1, p2;
		rect16		frame;
		rgb48		oldfg, oldbk, oldoc, cont, embr;
		SLRgn		clipr, *oldclip;
	  	uint8		oldpm;
		const rgb48	*color;
		ufixed		qw;
		int		z, w, dvx, dvr;
		uint8		oldtm, hilite;

		if (!ctl->visible || !p || !p->map)
			return 0;
		oldfg = p->fgColor, oldbk = p->bkColor, oldoc = p->opColor;
		oldpm = p->pnMode, oldtm = p->txMode;
		color = _defMCtlColors;
		rPenNormal(p,&pen);

		if (GetMenuItemCount(menu)>0)
		{
			GetMenuStringW(menu,ctl->value,(wchar_t*)str+1,sizeof(str)-5,0);
			z = wcslen((const wchar_t*)str+1);
			z = WideCharToMultiByte(CP_MACCP,0,(wchar_t*)str+1,z,str,z,0,0);
			hilite = (uint8)ctl->hilite;
		} else
			*str = z = 0, hilite = kControlInactivePart;
		dn = hilite-1u<kControlDisabledPart-1u;
		cont = p->bkColor = color[wCContent];
		if (dn)
			cont.red>>= 1, cont.green>>= 1, cont.blue>>= 1;
		frame = ctl->bounds;
		frame.left+= ind;
		dvx = ctl->bounds.right-15;
		dvr = MAX(ctl->bounds.left,dvx);

		pushBtnCont(p,&frame,hilite,dn,color,&cont,&embr,3,1);

		oldclip = NULL;
		copyRgn(&oldclip,p->clipRgn);
		clipr.size = RECT_RGN_SIZE;
		clipr.bBox = ctl->bounds;
		sectRgn(&p->clipRgn,&clipr);

		if (hilite<kControlDisabledPart)
			t = .95*256;
		else
			t = hilite==kControlInactivePart ? .6*256:.5*256;

		rgb48lerp(&p->fgColor,color+2,color,t);
		RGBFore(p,&p->fgColor);

		FixPt(&p0,frame.right-12,u-5);
		FixPt(&p1,frame.right-8,u-1);
		FixPt(&p2,frame.right-16,u-1);
		rTriangle(p,p0,p1,p2);
		p0.y = (u+5)<<16|0x8000;
		p1.y = p2.y = (u+1)<<16|0x8000;
		rTriangle(p,p0,p1,p2);

		p->fgSolid = RGBToPen(p->map,p->map->bpp==1 && dn ? &rgb_white:color+1);
		dvr-= 3;
		t = dvr-ctl->bounds.left-ind-12;
		qw = (w=MAX(t,0))<<16;
		t = z;
		z = rClipText(p,str,z,&qw,0,0);
		if (z<t)
		{
			if (z>0)
				if (rTextWidth(p,str+z-1,1,0)>6)
					--z;
			str[z] = '.', str[z+1] = '.', str[z+2] = '.', z+= 3;
		}
		if (p->clipRgn)
		{ /* fudging bBox left and right is still allowed... */
			t = p->clipRgn->bBox.right;
			p->clipRgn->bBox.right = dvr+2;
			ctlBoxText(ctl,str,z,color,ind+8);
			p->clipRgn->bBox.right = t;
		}

		/* 15, 13, 10, 2, 0 */
		ctlBoxText(ctl,1+(const char*)ctl->title,*ctl->title,color,4);

		/* partial masking out of elipsis characters */
		p->pnMode = mBlend;
		p->opColor = system_colors[_system_gray15];
		FixPt(&p0,dvr-=2,frame.top+3);
		FixPt(&p1,dvr,frame.bottom-4);
		RGBFore(p,&cont);
		if (p->map->bpp>7)
		{
			rLine(p,p0,p1);
			p->opColor = system_colors[_system_gray33];
			p0.x+= 1<<16, p1.x+= 1<<16;
			rLine(p,p0,p1);
		}
		p->opColor = system_colors[_system_gray67];
		p0.x+= 1<<16, p1.x+= 1<<16;
		rLine(p,p0,p1);
		p->opColor = system_colors[_system_gray87];
		p0.x+= 1<<16, p1.x+= 1<<16;
		rLine(p,p0,p1);

		copyRgn(&p->clipRgn,oldclip);
		free(oldclip);

		p->pnMode = oldpm, p->opColor = oldoc;
		p->txMode = oldtm;
		rRestorePen(p,&pen,1);
		RGBFore(p,&oldfg);
		RGBBack(p,&oldbk);
		return 1;
	}
	if (msg==wcMsgGetType)
		return ctlTypePopUpMenu;
	if ((msg&~1)==wcMsgClick)
	{
		grafPort	*p;
		POINT		lpt;

		if ((uint8)ctl->hilite>=kControlDisabledPart || GetMenuItemCount(menu)<=0)
			return 0;
		p = &ctl->wind->port;
		lpt.y = ctl->bounds.top-p->origin.y;
		lpt.x = ctl->bounds.left+ind-p->origin.x;
		if (p->map)
			lpt.y-= p->map->pOrigin.y, lpt.x-= p->map->pOrigin.x;

		ClientToScreen(world_screen_window,&lpt);
		wHiliteControl(ctl,1);
		//if ((menu=GetSubMenu(ctl->data,0))==NULL)
		SetMenuDefaultItem(menu,ctl->value,0);
		if ((u=TrackPopupMenu(menu,TPM_LEFTALIGN,lpt.x,lpt.y,0,
				world_screen_window,NULL))!=0)
		{
			MSG		msg;

			if (PeekMessage(&msg,world_screen_window,WM_COMMAND,WM_COMMAND,PM_REMOVE))
			{
				TranslateMessage(&msg);
				if (!HIWORD(msg.wParam))
					ctl->value = LOWORD(msg.wParam);
				DispatchMessage(&msg);
			}
		}
		wHiliteControl(ctl,0);
		return u;
	}
	if (msg==wcMsgDispose)
	{
		DestroyMenu(ctl->data); ctl->data = NULL;
		return 0;
	}
	return ctlGeneral(ctl,msg,data);
}

/* text box */

static DEF_CTL_RTN(textBoxProc)
{
	textBoxObj	*tbox = ctl->data;
	int		r;

	if (!tbox)
		return 0;
	switch(msg)
	{
	case wcMsgDraw:
	{
		rect2dq		b;
		grafPort	*port = &ctl->wind->port;
		penState	oldstate;

		if (!port->map)
			return 0;
		if (tbox->flags&tBoxFrame)
		{
			rPenNormal(port,&oldstate);
			FixRect(&b,ctl->bounds.left-1,ctl->bounds.top-1,
				ctl->bounds.right+1,ctl->bounds.bottom+1);
			rRect(port,&b);
			rRestorePen(port,&oldstate,1);
		}
		return tBoxDraw(tbox);
	}
	case wcMsgGetType:
		return tbox->flags&tBoxFrame
			? ctlTypeEditText:ctlTypeStaticText;
	case wcMsgTest:
		return (uint8)ctl->hilite<kControlDisabledPart
			&& tbox->flags&(tBoxActive|tBoxListSelect)
			&& ptInRect16(&tbox->bounds,*(const pt2d16*)data)
			? kControlEditTextPart:0;
	case wcMsgClick:
	case wcMsgDblClick:
	{
		pt2d16		pt;

		if ((uint8)ctl->hilite>=kControlDisabledPart)
			return 0;
		if (data)
			pt = *(const pt2d16*)data;
		else
			get_mouse_loc(&pt.x,&pt.y);
		r = 0;
		msg&= 1;
		if (tBoxClick(ctl->data,pt,mEventState.modifiers,
				msg ? dblClickFlag:0)>0)
			r = tbox->flags&tBoxListSelect ? msg ? kControlListBoxDoubleClickPart
				:kControlListBoxPart:kControlEditTextPart;
		return r;
	}
	case wcMsgDispose:
		tBoxDispose(ctl->data); ctl->data = NULL;
		return 0;
	case wcMsgIdle:
		tBoxIdle(tbox);
		return 0;
	case wcMsgFocus:
		if ((int)data&wcFFlagResetPosition)
			tBoxSelect(tbox,-1,0);
		if ((int)data&wcFFlagGain)
		{
			if (tbox->flags&tBoxFocusState)
				return 1;
			tbox->flags|= tBoxFocusState;
		} else
		{
			if (!(tbox->flags&tBoxFocusState))
				return 1;
			tbox->flags&=~tBoxFocusState;
		}
		_tBoxDrawSelect(tbox,tbox->selStart,tbox->selEnd);
		return 2;
	case wcMsgInBounds: 
		*(rect16*)data = tbox->txbounds;
		return 1;
	case wcMsgSetBounds:
	{
		rect16		mbounds;

		tbox->bounds = *(rect16*)data;
		if (tbox->flags&tBoxFrame)
		{
			OutsetRect16(&tbox->bounds,2,2);
			mbounds.left = mbounds.top = mbounds.right = mbounds.bottom = 2;
		} else
			mbounds.left = mbounds.top = mbounds.right = mbounds.bottom = 0;
		tBoxSetMargins(tbox,&mbounds,!!ctl->value);
		return 1;
	}
	case wcMsgReadValue:
	{
		const char	*s = tBoxGetText(tbox);

		*(double*)data = ctl->value; /* default */
		return parse_number(s,s+tbox->length,(double*)data,1)!=NULL;
	}
	case wcMsgHover:
		if (tbox->flags&tBoxActive && !(tbox->flags&(tBoxDragState|tBoxListSelect)))
		{
	#ifdef TEXT_DRAGDROP
			SLRgn		*hRgn = tBoxSelRgn((textBoxObj*)ctl->data);

			r = ptInRgn(hRgn,loc.x,loc.y);
			free(hRgn);
			return r;
	#else
			return 1;
	#endif
		}
		return 0;
	case wcMsgWheel:
		r = (uint32)data>>16;
		if (r&controlKey)
			return 0;
		if (r&shiftKey)
			++r, data = (void*)((int16)data*63>>5);
		r&= wcMFlagDirectionMask;
		r = tBoxScroll(tbox,r==0 ? (int16)data:0,r==1 ? (int16)data:0);
		return (r>=0)+r;
	case wcMsgDragEnter:
	{
		const ctlDragData	*dragInfo = data;

	// printf("tbdrag %p enter\n",
	//  tbox);
		if ((uint8)ctl->hilite<kControlDisabledPart
				&& (tbox->flags&(tBoxModifiable|tBoxDragDrop))
				==(tBoxModifiable|tBoxDragDrop)
				&& dragInfo
				&& dragInfo->originType==ctlTypeEditText
				&& dragInfo->originPart==kControlEditTextPart)
			return kControlEditTextPart;
		return 0;
	}
	case wcMsgDragLeave:
	// printf("tbdrag %p leave\n",
	//  tbox);
		_tBoxDrawSelect(tbox,tbox->selStart,tbox->selEnd);
		return 0;
	}
	return -8;
}

static DEF_CTL_RTN(pictProc)
{
	if (msg==wcMsgGetType)
		return ctlTypePicture;
	if (msg==wcMsgDraw)
	{
		pixMap		*m;
		int8		*pic;
		int		r;
		uint8		mode;

		if (!ctl->visible || (pic=hLock(ctl->data))==NULL)
			return 0;
		m = renderQDPict((const uint8*)pic,hSize(ctl->data),NULL,NULL,&mode,
			&rgb_white,NULL);
		hUnlock(ctl->data);
		if (!m)
			return memory_error();
		r = rBltPix(&ctl->wind->port,m,0,0,m->w,m->h,
			ctl->bounds.left,ctl->bounds.top,
			RECTANGLE_WIDTH(&ctl->bounds),RECTANGLE_HEIGHT(&ctl->bounds),
			mode); //m->bpp>ctl->wind->port.map->bpp ? mDither:mSrcCopy);
		disposePixMap(m);
		return r;
	}
	if (msg==wcMsgDispose)
		{hFree(ctl->data); ctl->data = NULL;}
	return ctlGeneral(ctl,msg,data);
}

static DEF_CTL_RTN(iconProc)
{
	if (msg==wcMsgGetType)
		return ctlTypeIcon;
	if (msg==wcMsgDraw)
	{
		cSpec		cstor[3];
		cTab64		*ct = (cTab64*)&cstor;
		pixMap		m =
			{32,32,{0,0},32>>3,1,0,1,3};
		int		r;

		if (!ctl->visible || (m.baseAddr=hLock(ctl->data))==NULL)
			return 0;
		ct->flags = 0x8000;
		ct->count = 2;
		ct->colors->rgb = ctl->wind->port.bkColor;
		ct->colors[1].rgb = ctl->wind->port.fgColor;
		m.pTable = ct;
		r = rBltPix(&ctl->wind->port,&m,0,0,m.w,m.h,
			ctl->bounds.left,ctl->bounds.top,
			RECTANGLE_WIDTH(&ctl->bounds),RECTANGLE_HEIGHT(&ctl->bounds),mSrcCopy);
		hUnlock(ctl->data);
		return r;
	}
	if (msg==wcMsgDispose)
		{hFree(ctl->data); ctl->data = NULL;}
	return ctlGeneral(ctl,msg,data);
}

static DEF_CTL_RTN(pixMapProc)
{
	if (msg==wcMsgGetType)
		return ctlTypePixMap;
	if (msg==wcMsgDraw)
	{
		pixMap		*m;
		int		r;

		if (!ctl->visible || (m=ctl->data)==NULL)
			return 0;
		r = ctl->wind->port.map->bpp;
		r = rBltPix(&ctl->wind->port,m,0,0,m->w,m->h,
			ctl->bounds.left,ctl->bounds.top,
			RECTANGLE_WIDTH(&ctl->bounds),RECTANGLE_HEIGHT(&ctl->bounds),
			ctl->max&0xF0);
		return r;
	}
	if (msg==wcMsgClick)
		return (uint8)ctl->hilite<kControlDisabledPart;
	if (msg==wcMsgDispose)
		{disposePixMap(ctl->data); ctl->data = NULL;}
	return ctlGeneral(ctl,msg,data);
}

/* ---- high-level */

int wFindControl(const windowRef wind,
	pt2d16 loc,
	ctlObj **out)
{
	ctlObj		*ctl;
	int		r = 0;

	if (!wind || !out)
		return -1;
	for(ctl=wind->ctlList; ctl; ctl=ctl->next)
		if (ctl->visible && ptInRect16(&ctl->bounds,loc)
				&& (r=ctl->rtn(ctl,wcMsgTest,&loc))>0)
			{*out = ctl; break;}
	return r;
}

int wFindControlIndex(const windowRef win,
	pt2d16 loc,
	boolean test)
{
	ctlObj		*ctl;
	int16		ix;

	if (!win)
		return -1;
	for(ctl=win->ctlList,ix=0; ctl; ++ix,ctl=ctl->next)
		if (ctl->visible && ptInRect16(&ctl->bounds,loc)
				&& (!test || ctl->rtn(ctl,wcMsgTest,&loc)))
			return ix;
	return -1;
}

int32 wCountControls(windowRef wind)
{
	ctlObj		*ctl;
	long		c;

	if (!wind)
		return -1;
	for(c=0,ctl=wind->ctlList; ctl; ctl=ctl->next)
		++c;
	return c;
}

ctlObj *wGetIndexedControl(windowRef win,
	int item)
{
	ctlObj		*ctl;

	if (!win || item<0 || (ctl=win->ctlList)==NULL)
		return NULL;
	while(item--)
		if ((ctl=ctl->next)==NULL)
			break;
	return ctl;
}

int wGetControlIndex(const windowRef wind,
	const ctlObj *s)
{
	ctlObj		*ctl;
	long		c;

	if (!wind)
		return -1;
	for(c=0,ctl=wind->ctlList; ctl; ctl=ctl->next,++c)
		if (ctl==s)
			return c;
	return -1;
}

int wDrawControls(windowRef wind)
{
	ctlObj		*ctl;
	long		r = 0, c;

	if ((c=wCountControls(wind))<=0)
		return c;
	/* render in reverse */
	if (--c)
	{
		ctlObj		**list = _alloca(sizeof(ctlObj*)*c);
		long		i;

		for(i=0,ctl=wind->ctlList; ++i<=c;)
		{
			ctl = ctl->next;
			list[i-1] = ctl;
		}
		for(i=c; --i>=0;)
		{
			ctl = list[i];
			if (ctl->visible)
				r+= ctl->rtn(ctl,wcMsgDraw,0);
		}
	}
	ctl = wind->ctlList;
	if (ctl && ctl->visible)
		r+= ctl->rtn(ctl,wcMsgDraw,0);
	return r;
}

int wDraw1Control(ctlObj *ctl)
{
	if (!ctl)
		return -1;
	if (!ctl->visible || ctl->rtn(ctl,wcMsgDraw,0)<=0)
		return 0;
	update_screen_window_rect(ctl->wind,ctl->bounds.left-2,ctl->bounds.top-2,
		ctl->bounds.right+2,ctl->bounds.bottom+2);
	return 1;
}

static DEF_CTL_RTN(nullProc)
{
	if (msg==wcMsgGetType)
		return -1;
	if (msg+0u>wcMsgWTest)
		return -8;
	return ctlGeneral(ctl,msg,data);
}

ctlRoutine	f_ctlr[] =
{
	nullProc,
	pushBtnProc, checkBoxProc, radioBtnProc, scrollBarProc,
	popUpProc, pushBtnProc, textBoxProc, textBoxProc,
	pictProc, iconProc, pixMapProc, textBoxProc
};

ctlObj *wNewControl(windowRef wind,
	const rect16 *bounds,
	const uint8 *title,
	boolean visible,
	int32 value,
	int32 min,
	int32 max,
	int type,
	void *refCon)
{
	ctlObj		**sctl, *ctl;

	if (!wind || type+1u>GET_ARRAY_COUNT(f_ctlr))
		return NULL;
	for(sctl=&wind->ctlList; *sctl; sctl=&(**sctl).next)
		;
	if ((ctl=calloc(sizeof(ctlObj),1))==NULL)
		return NULL;
	*sctl = ctl;
	ctl->wind = wind;
	if (title)
		memcpy(ctl->title,title,*title+1);
	ctl->bounds = *bounds;
	ctl->value = value;
	ctl->min = min;
	ctl->max = max;
	ctl->refCon = refCon;
	ctl->txFont = ffSystem;
	ctl->visible = visible;
	ctl->rtn = f_ctlr[type+1];
	if (type==ctlTypeScrollBar)
		ctl->trk = sBarTrack;
	else if (ctl->rtn==textBoxProc)
	{
		rect16		mbounds = *bounds;
		textBoxObj	*tbox;

		if (type==ctlTypeEditText)
			OutsetRect16(&mbounds,2,2);
		tbox = tBoxNew(wind,&mbounds,0,0,(tBoxAutoScroll|tBoxActive|tBoxModifiable
			|(type&1 ? tBoxFrame:0))^value>>16);
		ctl->data = tbox;
		if (tbox)
		{
			if (type!=ctlTypeEditText)
				tBoxSetMargins(tbox,NULL,1);
			if (tbox->ctlScrollV)
				tbox->ctlScrollV->group = ctl;
			if (tbox->ctlScrollH)
				tbox->ctlScrollH->group = ctl;
		}
	/*	ctl->txFont = wind->port.txFace.font;
		ctl->txSize = wind->port.txFace.size;
		ctl->txStyle = wind->port.txFace.style;*/
	} else if (type==ctlTypePopUpMenu)
		ctl->labelIndent = max;
	if (ctl->visible)
		ctl->rtn(ctl,wcMsgDraw,0);
	return ctl;
}

int wDisposeControl(ctlObj *ctl)
{
	rect16		mbounds;
	windowRef	wind;
	int		r;

	if (!ctl || !ctl->rtn)
		return -1;
	r = 0;
	if ((wind=ctl->wind)!=NULL)
	{
		ctlObj		**sctl;

		for(sctl=&wind->ctlList; *sctl; sctl=&(**sctl).next)
			if (*sctl==ctl)
				{*sctl = ctl->next, ++r; break;}
	}
	if (ctl==wind->ctlFocus)
		wind->ctlFocus = NULL;
	if (ctl==wind->ctlDefault)
		wind->ctlDefault = NULL;
	if (ctl==wind->ctlCancel)
		wind->ctlCancel = NULL;
	if (ctl->data)
		ctl->rtn(ctl,wcMsgDispose,0);
	mbounds = ctl->bounds;
	OutsetRect16(&mbounds,2,2);
	InvalWindowRect(ctl->wind,&mbounds);
	ctl->rtn = NULL;
	ctl->next = NULL;
	ctl->wind = NULL;
	free(ctl);
	return r;
}

int wKillControls(windowRef wind)
{
	int		r;

	if (!wind)
		return -1;
	r = 0;
	while(wind->ctlList)
		r+= wDisposeControl(wind->ctlList);
	return r;
}

void wIdleFocus(windowRef wind)
{
	ctlObj		*ctl;

	if (!wind)
		return;
	if ((ctl=wind->ctlFocus)!=NULL && (uint8)ctl->hilite<kControlDisabledPart)
		ctl->rtn(ctl,wcMsgIdle,NULL);
}

int wIdleControls(windowRef wind)
{
	rect16		bounds;
	pt2d16		loc;
	ctlObj		*ctl;
	int		curs;

	if (!wind)
		return -1;
	get_mouse_loc(&loc.x,&loc.y);
	wIdleFocus(wind);
	if (wFindControl(wind,loc,&ctl)>0)
	{
		bounds = ctl->bounds;
		ctl->rtn(ctl,wcMsgInBounds,&bounds);
		/* for sake of textbox whitespace margins, the larger bounds makes the call */
		/* let hover determine pt in region, etc. */
		UnionRect16(&bounds,&ctl->bounds);
		if (ptInRect16(&bounds,loc) && (curs=ctl->rtn(ctl,wcMsgHover,NULL))>=0)
		{
			set_mouse_cursor(curs,0);
			return 1;
		}
	}
	set_mouse_cursor(0,0);
	return 0;
}

void wCtlNewFocus(ctlObj *ctl,
	boolean focusHomePosition)
{
	windowRef	wind = ctl->wind;

	if (wind->ctlFocus && wind->ctlFocus!=ctl)
	{
		wind->ctlFocus->rtn(wind->ctlFocus,wcMsgFocus,
			(void*)(focusHomePosition ? wcFFlagResetPosition:0));
		ctl->rtn(ctl,wcMsgFocus,(void*)wcFFlagGain);
	}
	wind->ctlFocus = ctl;
}

void wCtlSelectNext(ctlObj *ctl,
	boolean reverse)
{
	windowRef	win = ctl->wind;
	ctlObj		*oldCtl = ctl;
	int		ix = 0, ct = wCountControls(win);

	if (reverse)
		ix = wGetControlIndex(win,ctl);
	do
	{
		if (reverse)
		{
			if (--ix<0)
				ix+= ct;
			ctl = wGetIndexedControl(win,ix);
		} else
		{
			if ((ctl=ctl->next)==NULL)
				ctl = win->ctlList;
		}
		if ((uint8)ctl->hilite>=kControlDisabledPart)
			continue;
		if (((ctl->rtn(ctl,wcMsgGetType,NULL)&~1)==ctlTypeStaticText
				&& ctl->visible)) /* 2018.04.02 */
		{
			int		flags = ((textBoxObj*)ctl->data)->flags;

			if (flags&tBoxModifiable && !(flags&tBoxListSelect))
				goto set;
		}
		#if 0
		else if (ctl->visible>1) /* 2024.02.13 */
		{
			rect2dq		b;
			penState	oldstate;
			uint8		grayPat[8];

			rPenNormal(p,&oldstate);
			_get_gray_pattern(grayPat);
			rPenPat(p,grayPat);
			FixRect(&b,ctl->bounds.left-1,ctl->bounds.top-1,
				ctl->bounds.right+1,ctl->bounds.bottom+1);
			rRect(p,&b);
			rRestorePen(p,&oldstate,1);
			goto set;
		}
		#endif
	} while(ctl!=oldCtl);
	return;
set:	wCtlNewFocus(ctl,(uint8)ctl->value);
	if ((uint8)ctl->value)
		tBoxSelect(ctl->data,0,IMAX(int32));
}

int wCtlAnimHilite(ctlObj *ctl,
	uint8 speed)
{
	int		delay = speed*2000;

	if (!ctl || (uint8)ctl->hilite>=kControlDisabledPart)
		{ctl = NULL; return 0;}
	wHiliteControl(ctl,1);
	wMcsDelay(delay);
	wHiliteControl(ctl,0);
	wMcsDelay(MIN(100000,delay));
	return 1;
}

int wCtlRequestDragDrop(const windowRef wind,
	pt2d16 loc,
	ctlObj **out,
	ctlDragData *dragInfo)
{
	ctlObj		*ctl;
	int		r;

	if ((r=wFindControl(wind,loc,&ctl))<=0)
	{
		ctl = *out;
		if (ctl)
			ctl->rtn(ctl,wcMsgDragLeave,dragInfo);
		*out = NULL;
		return r;
	}
	if (*out==ctl)
		return 1;
	if (*out)
		(*out)->rtn(*out,wcMsgDragLeave,dragInfo);
	*out = ctl;
	return ctl->rtn(ctl,wcMsgDragEnter,dragInfo);
}

ctlObj *wControlEvent(windowRef wind,
	const eventRecord *event)
{
	ctlObj		*ctl = NULL;

	if (!wind || !event)
		return NULL;
	switch(event->what)
	{
	case nullEvent:
		wIdleControls(wind);
		return NULL;
	case mouseUp:
		if ((int)event->message&wheelFlag && (ctl=wind->ctlFocus)!=NULL)
		{ /* focus isn't always the one we need to 'wheel' */
		  /* check for hover bounds */
			rect16		b;
			pt2d16		loc;
			uint32		flags;

			get_mouse_loc(&loc.x,&loc.y);
			if (wFindControl(wind,loc,&ctl)<=0 || ctl->rtn(ctl,wcMsgInBounds,&b)<=0)
				ctl = wind->ctlFocus;
			flags = (uint16)((int)event->message&wheelDownFlag
					? ctlWheelScrollCount:-ctlWheelScrollCount);
			flags|= ((event->modifiers&wcMFlagModifiersMask)
				|((int)event->message&wheelVFlag ? wcMDirectionV
				 :(int)event->message&wheelHFlag ? wcMDirectionH:0))<<16;
			if (ctl->rtn(ctl,wcMsgWheel,(void*)flags)<=0)
				ctl = NULL;
		}
		break;
	case mouseDown:
	{
		pt2d16		loc = event->where;
		int		i;

		global_to_local_pt(&loc);
		if (wFindControl(wind,loc,&ctl)<=0)
			{ctl = NULL; break;}
		if (ctl!=wind->ctlFocus && (ctl->rtn(ctl,wcMsgGetType,NULL)&~1)==ctlTypeStaticText)
			wCtlNewFocus(ctl,(uint8)ctl->value);
		if ((i=ctl->rtn(ctl,(int)event->message&dblClickFlag
				? wcMsgDblClick:wcMsgClick,&loc))<=0)
			{ctl = NULL; break;}
		if (i==kControlListBoxDoubleClickPart)
			if (ctl->group && (uint8)ctl->group->hilite<kControlDisabledPart)
				{ctl = ctl->group; goto anim;}
		break;
	}
	case keyDown:
	case autoKey:
		switch((uint8)event->message)
		{
		case kEnterCharCode:
			goto defo;
		case kEscapeCharCode:
			ctl = wind->ctlCancel;
			goto anim;

		default:
			if ((ctl=wind->ctlFocus)!=NULL
					&& (ctl->rtn(ctl,wcMsgGetType,NULL)&~1)!=ctlTypeStaticText)
				ctl = NULL;
			else
			{
				if (ctl //&& ((textBoxObj*)ctl->data)->flags&tBoxListSelect
						&& (uint8)event->message==kReturnCharCode
						&& wind->ctlDefault)
				{
				defo:	ctl = wind->ctlDefault;
				anim:	if (!wCtlAnimHilite(ctl,defaultBtnAnimDelay))
						{ctl = NULL; break;}
				}
				if (!ctl)
					break;
				if ((uint8)event->message==kTabCharCode)
				{
					wCtlSelectNext(ctl,!!(event->modifiers&shiftKey));
					ctl = wind->ctlFocus;
				} else
					tBoxKey((textBoxObj*)ctl->data,(uint32)event->message,
						event->modifiers);
			}
		}
		break;
	case updateEvt:
		wBeginUpdate((windowRef)event->message);
		wDrawControls((windowRef)event->message);
		wEndUpdate((windowRef)event->message);
	}
	return ctl;
}

/* ---- basic setting routines */

int wSetControlBounds(ctlObj *ctl,
	const rect16 *bounds)
{
	if (!ctl || !bounds)
		return -1;
	if (!memcmp(&ctl->bounds,bounds,sizeof(*bounds)))
		return 0;
	ctl->bounds = *bounds;
	ctl->rtn(ctl,wcMsgSetBounds,bounds);
	return 1+wDraw1Control(ctl);
}

int wSetControlValue(ctlObj *ctl,
	int32 value)
{
	if (!ctl)
		return -1;
	value = PIN(value,ctl->min,ctl->max);
	if (value==ctl->value)
		return 0;
	ctl->value = value;
	return 1+wDraw1Control(ctl);
}

int wSetControlMinimum(ctlObj *ctl,
	int32 value)
{
	if (!ctl)
		return -1;
	ctl->min = value;
	value = PIN(ctl->value,value,ctl->max);
	if (value==ctl->value)
		return 0;
	ctl->value = value;
	return 1+wDraw1Control(ctl);
}

int wSetControlMaximum(ctlObj *ctl,
	int32 value)
{
	if (!ctl)
		return -1;
	ctl->max = value;
	value = PIN(ctl->value,ctl->min,value);
	if (value==ctl->value)
		return 0;
	ctl->value = value;
	return 1+wDraw1Control(ctl);
}

int wHiliteControl(ctlObj *ctl,
	uint16 hilite)
{
	if (!ctl)
		return -1;
	if (hilite==ctl->hilite)
		return 0;
	ctl->hilite = hilite;
	return 1+wDraw1Control(ctl);
}

int wSetControlTitle(ctlObj *ctl,
	const uint8 *title)
{
	if (!ctl)
		return -1;
	if (!memcmp(ctl->title,title,*title+1))
		return 0;
	memcpy(ctl->title,title,*title+1);
	return 1+wDraw1Control(ctl);
}

int wSelectControlText(ctlObj *ctl,
	int32 selStart,
	int32 selEnd)
{
	windowRef	win;

	if (!ctl || (win=ctl->wind)==NULL)
		return -1;
	if ((ctl->rtn(ctl,wcMsgGetType,0)&~1)!=ctlTypeStaticText)
		return 0;
	wCtlNewFocus(ctl,1);
	tBoxSelect(ctl->data,selStart,selEnd);
	return 1;
}

int32 wGetControlText(ctlObj *ctl,
	char *outStr,
	int32 z)
{
	textBoxObj	*tbox;
	const char	*s;

	if (!outStr || !ctl || (ctl->rtn(ctl,wcMsgGetType,0)&~1)!=ctlTypeStaticText)
		return -1;
	if (z<=0)
		return 0;
	if ((tbox=(textBoxObj*)ctl->data)==NULL || (s=tBoxGetText(tbox))==NULL)
		{*outStr = 0; return 0;}
	z = MIN(tbox->length,z);
	memcpy(outStr,s,z);
	return z;
}

int wSetControlText(ctlObj *ctl,
	const char *s,
	int32 z)
{
	textBoxObj	*tbox;
	int		r;

	if (!ctl || (ctl->rtn(ctl,wcMsgGetType,0)&~1)!=ctlTypeStaticText
			|| (tbox=(textBoxObj*)ctl->data)==NULL)
		return -1;
	if ((r=tBoxSetTextRaw(tbox,s,z))<=0)
		return r;
	if (tbox->styles)
	{ /* truncate styles to first */
		tStyleRun	*styles;

		tbox->styleCount = 1;
		if ((styles=realloc(tbox->styles,sizeof(tStyleRun)))!=NULL)
			tbox->styles = styles;
	}
	if (ctl->visible)
		tBoxUpdate(tbox,NULL,0);
	else
		tBoxRecalc(tbox,0);
	return 1;
}

int wSetControlVisible(ctlObj *ctl,
	boolean visible)
{
	rect16		obounds;
	windowRef	win;

	if (!ctl)
		return -1;
	if (visible==ctl->visible)
		return 0;
	ctl->visible = visible;
	if (visible)
		return 1+wDraw1Control(ctl);
	if (!ctl->wind)
		return 0;
	win = ctl->wind;
	if (win->ctlFocus==ctl)
		wCtlSelectNext(ctl,1);
	obounds = ctl->bounds;
	OutsetRect16(&obounds,3,3);
	OffsetRect16(&obounds,-win->port.origin.x,-win->port.origin.y);
	wInvalRect(ctl->wind,&obounds);
	return 0;
}
