/*
 mevent.c

   Revised: 2023.10.04
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.05.22
  Formerly mtimer, moved event code implemented in spanner.c & cseries here.
2014.08.21
  The Microseconds equivalent version is producing results 1.00247 times
faster than GetTickCount()*1000.
2015.01.07
  Turns out GetTickCount is the one returning an inaccurate value
(see perftime).
2015.04.07
  Major code optional (USE_MEVENTS or USE_MYDIALOGS).
2015.07.24
  Added WaitNextEvent.  Smooth sailing with near-0% CPU usage.
GetNextEvent's default max sleep time reduced to 3/16.625.
2017.11.23
  Workaround to fix right-click bug when activating app with a left click.
2019.01.16
  New time resolution form.
*/

#include <process.h>
#ifdef DEBUG_MKV_DOWN
 #include <stdio.h>
#endif

#define USE_HIGH_RES_TIMER	1
#if HAVE_OS
 #ifndef USE_WINEVENT_TIMER
  #define USE_WINEVENT_TIMER	1
 #endif
#endif

#ifdef USE_MYDIALOGS
 #define USE_MEVENTS
#endif

#if defined(WIN32) || defined(WIN64)
 #include <windows.h>
 #define ThdMutexLock(t)	EnterCriticalSection(&(t)->criticalSection)
 #define ThdMutexUnlock(t)	LeaveCriticalSection(&(t)->criticalSection)
 #define EvtMutexLock()		EnterCriticalSection(&eventSection)
 #define EvtMutexUnlock()	LeaveCriticalSection(&eventSection)
#elif defined(__MT__)
 #include <threads.h>
 #define ThdMutexLock(t)	mtx_lock(&(t)->mtx);
 #define ThdMutexUnlock(t)	mtx_unlock(&(t)->mtx);
#endif

#define __MWIND_PRIVATE

#include "mevent.h"
#include "mwind.h"

#if 0 /* same precision as GetTickCount, slower */
 #include <time.h>
 #define milliseconds()		clock()
#elif USE_WINEVENT_TIMER
 #define milliseconds()		timeGetTime()
#else
 #define milliseconds()		GetTickCount()
#endif

/* ---- constants */

/* standard EventQueue only carries 20 */
#define EVENTQMASK		31

/* ---- typedefs */

#if defined(ThdMutexLock) && defined(USE_MEVENTS)
typedef struct timerThd
{
#if defined(WINUSERAPI) && !USE_WINEVENT_TIMER
	CRITICAL_SECTION	criticalSection;
#else
	mtx_t		mtx;
#endif
#if USE_HIGH_RES_TIMER
	int64		ltick;
	int16		resolution;
#else
	int32		ltick;
#endif
#if USE_WINEVENT_TIMER
	UINT		ticEID;
#endif
	uint16		count, maximum;
	timerEntry	list[1];
} timerThd;
#endif

/* ---- globals */

#ifdef HAVE_OS
WNDCLASSEX	mEventWClass = {0};
const TCHAR	*mEventWDevName = NULL;
COLORREF	mEventWEraseColor;
HPALETTE	world_screen_pal = NULL;
HBITMAP		world_screen_dib = NULL;
HWND		world_screen_window = NULL;

eventRecord	mEventState = {0,btnState|rightBtnState};
uint8		mKeyMap[20];

pt2d16		mEventWMin = {0}, mEventWMax = {0};
uint8		mEventInWMLoop = 0;
int8		mEventMouseRelV = 0;
boolean		mEventDisableSSAV = 0, mEventMouseDetail = 0;
int32		mEventMouseRelW = 0, mEventMouseRelX = 0, mEventMouseRelY = 0;

int (__cdecl *mEventCmdRequest)(TCHAR *param);
void (__cdecl *mEventDropFilesFunc)(HDROP);

 #if defined(ThdMutexLock) && defined(USE_MEVENTS)
static uint64	mVBL;
static boolean	waitnext;
 #endif
int8		mGetEventSleepCycle;
uint32		mVBLInitMcs = 14667;
#endif

#ifdef WINUSERAPI
static DWORD	mcs_freq;
static double	mcs_freqsub;

static int CalcQPCRatios(void)
{
	LARGE_INTEGER	tt;

	tt.LowPart = 0;
	if (!QueryPerformanceFrequency(&tt) || (mcs_freq=tt.LowPart)==0)
	{
		mcs_freqsub = 1;
		return 0;
	}
	mcs_freqsub = 1000000./mcs_freq;
	return 1;
}

void Microseconds(UnsignedWide *mcs)
{
	LARGE_INTEGER	tt;
	uint64		n;

	if (!mcs_freq && !CalcQPCRatios())
	{
		*mcs = GetTickCount()*1000ll;
		return;
	}
	/* multi-core use may result in differing CPU results */
	//DWORD_PTR oldAffinity = SetThreadAffinityMask(GetCurrentThread(),1);

	(void)QueryPerformanceCounter(&tt);
#if defined(__POCC__) && envx86
	__asm
	{
		mov	eax,tt+4
		xor	edx,edx
		mov	[tt+4],edx
		div	mcs_freq
		mov	n+4,eax
		mov	eax,tt
		div	mcs_freq
		mov	tt,edx
		fild	qword [tt]
		fmul	mcs_freqsub
		fistp	qword [tt]
		mov	ecx,1000000
		mul	ecx
		add	eax,tt
		adc	edx,tt+4
		imul	ecx,n+4
		add	edx,ecx
		mov	ecx,mcs
		mov	[ecx],eax
		mov	[ecx+4],edx
	}
#else
	n = (uint64)tt.QuadPart/mcs_freq;
	*mcs = n*1000000u+(uint64)((tt.u.LowPart-(uint32)n*mcs_freq)*mcs_freqsub);
#endif
	//(void)SetThreadAffinityMask(GetCurrentThread(),oldAffinity);
}

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
UnsignedWide QPCtoMicroseconds(ULONGLONG tt)
{
#if defined(__POCC__) && envx86
(void)tt;
	__asm
	{
		push	ebx
		cmp	mcs_freq,0
		jnz	c

		call	CalcQPCRatios
		test	eax,eax
		jnz	c

		mov	eax,[esp+8] ;default to mcs
		mov	edx,[esp+8+4]
		jmp	e

	c:	mov	eax,[esp+8+4]
		xor	edx,edx
		mov	[esp+8+4],edx
		div	mcs_freq
		mov	ebx,eax
		mov	eax,[esp+8]
		div	mcs_freq
		mov	[esp+8],edx
		fild	qword [esp+8]
		fmul	mcs_freqsub
		fistp	qword [esp+8]
		mov	ecx,1000000
		mul	ecx
		add	eax,[esp+8]
		adc	edx,[esp+8+4]
		imul	ecx,ebx
		add	edx,ecx
	e:	pop	ebx
		ret
	}
#else
	uint64		n;

	if (!mcs_freq && !CalcQPCRatios())
		return tt;
	n = (uint64)tt/mcs_freq;
	return n*1000000u+(uint64)(((uint32)tt-(uint32)n*mcs_freq)*mcs_freqsub);
#endif
}
#elif __POCC_STDC_VERSION__ >= 201101L
#include <time.h>
/* no point of using this if pocrt.dll errors start coming up */
void Microseconds(UnsignedWide *mcs)
{
	struct timespec	ts;

	timespec_get(&ts,TIME_UTC);
	*mcs = ts.tv_sec*(uint64)1000000+ts.tv_nsec/1000;
}
#endif

/* ---- timer */

#if !TARGET_OS_MAC
#undef TickCount
uint32 TickCount(void)
{
/* translating native messages is a bad idea here, considering input latency */
#ifdef HAVE_OS
	return mEventState.when;
#else
	uint64		mcs;

	Microseconds(&mcs);
	return (uint32)(mcs/SMALL_TICK_TIME);
#endif
}
#endif

#if defined(ThdMutexLock) && defined(USE_MEVENTS)
int PrimeTimer(timerThread *thd,
	timerEntry *entry,
	int64 mcs)
{
	int64		n;
	timerThd	*tdata = (timerThd*)thd;
	int		r;

	if (!tdata->count || !tdata->maximum)
		return -1;
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	n = entry->mcsremaining;
	if (!int64n(n))
		{r = 0; goto end;} /* already primed */
	r = 1;
	(void)int64add(n,mcs);
	if (int64n(n))
		int64xset(n,0);
	entry->mcsremaining = n;
end:
#if !USE_WINEVENT_TIMER
	ThdMutexUnlock(tdata);
#endif
	return r;
}

timerEntry *InstallTimer(timerThread *thd,
	timerRoutine rtn)
{
	timerThd	*tdata;
	timerEntry	*t;

	if (!rtn || (tdata=(timerThd*)thd)==NULL
			|| tdata->count>=tdata->maximum)
		return NULL; /* bad param or maxed out */
	/* probably no need to lock; the count is incremented last */
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	t = tdata->list+tdata->count;
	int64xset(t->mcsremaining,-1);
	t->routine = rtn;
	++tdata->count;
#if !USE_WINEVENT_TIMER
	ThdMutexUnlock(tdata);
#endif
	return t;
}

int RemoveTimer(timerThread *thd,
	timerEntry *entry)
{
	timerThd	*tdata;
	timerEntry	*t;
	int		c, r;

	if (!entry || (tdata=(timerThd*)thd)==NULL
			|| (c=tdata->count)==0 || !tdata->maximum)
		return -1;
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	t = tdata->list, r = 1;
	do
		if (t==entry)
		{
			--tdata->count, ++t;
			memcpy(t-1,t,--c*sizeof(*t));
			goto end;
		}
	while(++t,--c);
	r = 0;
end:
#if !USE_WINEVENT_TIMER
	ThdMutexUnlock(tdata);
#endif
	return r;
}

#ifdef __MT__
static
#endif
void timerThreadIdle(timerThread *thd)
{
	timerThd	*tdata = (timerThd*)thd;
	timerEntry	*t;
	int		c;
#if USE_HIGH_RES_TIMER
	int64		mcs, mcsd;
#else
	int32		mcsd;
 #if !USE_WINEVENT_TIMER
	int32		tkc;
 #endif
#endif
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	if (!tdata->count)
		goto end;
#if USE_HIGH_RES_TIMER
	Microseconds((uint64*)&mcs);
	mcsd = mcs;
	(void)int64sub(mcsd,tdata->ltick);
	tdata->ltick = mcs;
	if (!(mcsd>0))
		goto end;
#elif !USE_WINEVENT_TIMER
	mcsd = tkc = milliseconds();
	mcsd-= tdata->ltick;
	tdata->ltick = tkc;
	if (mcsd<=0)
		goto end;
	mcsd*= 1000;
#else
	mcsd = 1000;
#endif
	for(t=tdata->list,c=tdata->count; c--; ++t)
		if (!int64n(t->mcsremaining))
		{
			(void)int64sub(t->mcsremaining,mcsd);
			if (int64n(t->mcsremaining))
				t->routine(t);
		}
end:
#if USE_WINEVENT_TIMER
	;
#else
	ThdMutexUnlock(tdata);
#endif
}

#if USE_WINEVENT_TIMER
static void CALLBACK ticEvtFunc(UINT id,
	UINT msg,
	DWORD_PTR user,
	DWORD_PTR dw1,
	DWORD_PTR dw2)
{ /* showing time differences very close to 1ms, no issue */
	timerThd	*tdata = (timerThd*)user;

	timerThreadIdle((timerThread*)tdata);
}
#elif defined(__MT__)
static void __cdecl timerThreadFunc(void *param)
{
	timerThd	*tdata = param;
//	int8		toggle = 0;
	MSG		msg;

	(void)PeekMessage(&msg,NULL,WM_USER,WM_USER,PM_REMOVE);
  #if USE_HIGH_RES_TIMER
	Microseconds((UnsignedWide*)&tdata->ltick);
  #else
	tdata->ltick = milliseconds()+1;
  #endif
	while(tdata->maximum)
	{ /*
	     A previous version of this was inappropriately
	    wrapped with timeBeginPeriod and timeEndPeriod.
	     They should only be called at program level,
	    at best only once each.
	  */
		timerThreadIdle(param);
		Sleep(1);
	}
	_endthread();
}
#endif

timerThread *CreateTimerThread(int maxCount)
{
	timerThd	*tdata;
#if defined(__MT__) && !USE_WINEVENT_TIMER
	uintptr_t	r;
#endif
	if (maxCount-1>65534u || (tdata=calloc(sizeof(timerThd)
			+sizeof(timerEntry)*(maxCount-1),1))==NULL)
		return NULL;
	tdata->maximum = (uint16)maxCount;
#if USE_WINEVENT_TIMER
	tdata->ticEID = timeSetEvent(1,1,ticEvtFunc,(DWORD_PTR)tdata,TIME_PERIODIC);
#else
	InitializeCriticalSection(&tdata->criticalSection);
 #if defined(__MT__)
	if ((r=_beginthread(timerThreadFunc,0,tdata))==-1)
	{
		DeleteCriticalSection(&tdata->criticalSection);
		free(tdata);
		return NULL;
	}
	/* Setting to 'highest' from 'above normal' fixes some inaccuracies.
	  Doing things with the Task Manager may affect the timer,
	  but moving generic windows (including our app's) around no longer does.
	*/
	if (SetThreadPriority((HANDLE)r,THREAD_PRIORITY_HIGHEST)
			|| SetThreadPriority((HANDLE)r,THREAD_PRIORITY_ABOVE_NORMAL))
		;
 #endif
#endif
#if USE_HIGH_RES_TIMER
	Microseconds((UnsignedWide*)&tdata->ltick);
	tdata->resolution =-1;
#else
	tdata->ltick = milliseconds()+1;
#endif
	return (timerThread*)tdata;
}

int EndTimerThread(timerThread *thd)
{
	timerThd	*tdata;

	if ((tdata=(timerThd*)thd)==NULL || !tdata->maximum)
		return -1;
#if !USE_WINEVENT_TIMER
	ThdMutexLock(tdata);
#endif
	/* this should quickly kill the thread from inside, should it run */
	tdata->count = tdata->maximum = 0;
#if USE_WINEVENT_TIMER
	timeKillEvent(tdata->ticEID);
#else
	ThdMutexUnlock(tdata);
	Sleep(1);
	DeleteCriticalSection(&tdata->criticalSection);
#endif
	if (tdata->resolution>=0)
		(void)timeEndPeriod(tdata->resolution);
	free(tdata);
	return 0;
}

int SetTimerResolution(timerThread *thd,
	int resolution)
{
#if USE_HIGH_RES_TIMER
	TIMECAPS	ptc;
	timerThd	*tdata;
	int		r = 1;

	if ((tdata=(timerThd*)thd)==NULL || !tdata->maximum)
		return -1;
	if (tdata->resolution>=0)
		(void)timeEndPeriod(tdata->resolution);
	if (resolution>=0 && timeGetDevCaps(&ptc,sizeof(ptc))==MMSYSERR_NOERROR)
	{
		if (resolution+0u>ptc.wPeriodMax)
			resolution = ptc.wPeriodMax;
		if (resolution+0u<ptc.wPeriodMin)
			resolution = ptc.wPeriodMin;
		r = timeBeginPeriod(resolution)==TIMERR_NOERROR;
	} else /* do nothing on failure */
		resolution =-1;
	tdata->resolution = (int16)(resolution!=(int16)resolution ? resolution>>31^IMAX(int16):resolution);
	return resolution>=0 && r;
#else
(void)thd; (void)resolution;
	return -8;
#endif
}

void mOSVBL_timer(timerEntry *this)
{ /* 16625 mcs (133/8 ms), the Classic Mac OS tick rate - see my_cseries.h */
	++mEventState.when;
	(void)int64addu32(this->mcsremaining,SMALL_TICK_TIME);
	if (int64n(this->mcsremaining))
		int64xset(this->mcsremaining,0);
	Microseconds(&mVBL);
#ifndef NO_MVBL_USER_MSG
	/* post user event to relieve WaitNextEvent waits */
	if (waitnext)
		(void)PostMessage(world_screen_window,WM_USER,0,0);
#endif
}

/* note that Windows won't set the cursor if main thread won't poll events! */
void mOSBusyCursor_timer(timerEntry *this)
{
	set_mouse_cursor(_curs_wait,_setcurs_immediate);
	/* let this expire */
}
#endif

/* ---- events */

#ifndef NO_GUI
#if defined(USE_MEVENTS) || defined(USE_MYDIALOGS)
static uint8	deadKey, menuDown, eventCount, eventWPos, eventRPos;

static eventRecord	eventQueue[EVENTQMASK+1];

#ifdef WINUSERAPI
static CRITICAL_SECTION	eventSection;
#endif

boolean mouse_still_down(void)
{
	MSG		msg;

	if (!(mGetEventSleepCycle+=0x40))
		Sleep(1);
	global_idle_proc();
	while(PeekMessage(&msg,world_screen_window,WM_MOUSEMOVE,WM_MBUTTONDBLCLK,PM_REMOVE))
	{
		(void)TranslateMessage(&msg);
		(void)DispatchMessage(&msg);
	}
	return !((mEventState.modifiers&mEventState.modifiers*2)&btnState);
}

#define MODKEYS_6L	(1<<(MKV_MOD&7)|1<<(MKV_LALT&7)|(1<<(MKV_LSHIFT&7) \
	|1<<(MKV_CAPSLOCK&7)|1<<(MKV_LCONTROL&7) \
	|1<<(MKV_RSHIFT&7)|1<<(MKV_RALT&7) \
	|1<<(MKV_RCONTROL&7))<<8|(1<<(MKV_RGUI&7))<<16)

uint8 *MGetKeys(uint8 outmap[20])
{
	uint8		vkmap[256];
	MSG		msg;

	if (!outmap)
		return NULL;
	while(PeekMessage(&msg,world_screen_window,WM_KEYDOWN,WM_SYSKEYUP,PM_REMOVE))
	{
		(void)TranslateMessage(&msg);
		(void)DispatchMessage(&msg);
	}
	/* only GetKeyboardState will deliver independent left/right modifier keys */
	/* the problem: keys like Enter and Return are inseparable */
	/* key up/down event reading, done above, is the only way to get true Return/Enter */
	*(int32*)(mKeyMap+6)&=~MODKEYS_6L;
	if (GetKeyboardState(vkmap))
	{
		if (vkmap[VK_LSHIFT]>>1)
			_MKSET(mKeyMap,MKV_LSHIFT);
		if (vkmap[VK_RSHIFT]>>1)
			_MKSET(mKeyMap,MKV_RSHIFT);
		if (vkmap[VK_CAPITAL]&1) /* sticky capsLock */
			_MKSET(mKeyMap,MKV_CAPSLOCK);
		if (vkmap[VK_LMENU]>>1)
			_MKSET(mKeyMap,MKV_LALT);
		if (vkmap[VK_RMENU]>>1)
			_MKSET(mKeyMap,MKV_RALT);
		if (GetKeyState(VK_LCONTROL)>>1)
			_MKSET(mKeyMap,MKV_LCONTROL);
		if (GetKeyState(VK_RCONTROL)>>1)
			_MKSET(mKeyMap,MKV_RCONTROL);
	}
	memcpy(outmap,mKeyMap,20);
	return outmap;
}

/* Do Not add PeekMessage/GetMessage here; mWndProc calls this */
uint16 updateEventState(void)
{
	uint16		state =
#ifdef MKV_OPT_APPS
		_MKDOWN(mKeyMap,MKV_OPT_APPS) ? optionKey:0;
#else
		_MKDOWN(mKeyMap,MKV_APPS) ? optionKey:0;
#endif
	*(int32*)(mKeyMap+6)&=~MODKEYS_6L;
	if (GetKeyState(VK_SHIFT)>>1)
	{
		state|= shiftKey;
		if (GetKeyState(VK_LSHIFT)>>1)
			_MKSET(mKeyMap,MKV_LSHIFT);
		if (GetKeyState(VK_RSHIFT)>>1)
			_MKSET(mKeyMap,MKV_RSHIFT);
	}
	if (GetKeyState(VK_CAPITAL)&1) /* sticky capsLock */
		state|= capsLock, _MKSET(mKeyMap,MKV_CAPSLOCK);
	if (GetKeyState(VK_CONTROL)>>1)
	{
		state|= controlKey;
		if (GetKeyState(VK_LCONTROL)>>1)
			_MKSET(mKeyMap,MKV_LCONTROL);
		if (GetKeyState(VK_RCONTROL)>>1)
			_MKSET(mKeyMap,MKV_RCONTROL);
	}
	if (menuDown)
		state|= menuKey;
	if (menuDown || GetKeyState(VK_MENU)>>1)
	{
		if (GetKeyState(VK_LMENU)>>1)
			_MKSET(mKeyMap,MKV_LALT);
		if (GetKeyState(VK_RMENU)>>1)
			_MKSET(mKeyMap,MKV_RALT);
	}
	mEventState.modifiers&= btnState|rightBtnState|middleBtnState;
	return mEventState.modifiers|= state;
}

static int evtDeQueue(eventRecord *event, /* must be valid */
	uint32 mask)
{
	int		r, c, i = eventRPos;

	if ((c=eventCount)==0 || !mask)
		return 0;
	EvtMutexLock();
	r = 0;
	do
	{
		eventRecord	*a = eventQueue+i;

		i = (i+1)&EVENTQMASK;
		if (1u<<a->what&mask)
		{
			*event = *a;
			eventWPos&= EVENTQMASK;
			if (--eventCount)
			{ /* remove from list */
				do
				{
					eventRecord	*n = eventQueue+i;

					i = (i+1)&EVENTQMASK;
					*a = *n, a = n;
				} while(--c);
			}
			--eventWPos;
			a->what = 0;
			r = 1;
			break;
		}
	} while(--c);
	EvtMutexUnlock();
	return r;
}

int PostMEvent(const eventRecord *event)
{
	int		r;

	if (!event)
		return -1;
	EvtMutexLock();
	r = 1;
	if (eventCount<=EVENTQMASK)
		++eventCount;
	else if (event->what==nullEvent) /* full - ignore excess null events */
		{r = 0; goto end;}
	else
		++eventRPos, eventRPos&= EVENTQMASK; /* otherwise overwrite old */
	eventQueue[eventWPos&EVENTQMASK] = *event, ++eventWPos;
end:	EvtMutexUnlock();
	return r;
}

boolean GetNextEvent(uint32 mask,
	eventRecord *event)
{
	MSG		msg;
	mEventInWMLoop|= 1;
	/* sleep first if necessary to lessen input lag */
	if (!(mGetEventSleepCycle+=0x20))
	{
		int64		mcs;

		Microseconds((uint64*)&mcs);
		mcs-= mVBL;
		if (mcs>=(int64)mVBLInitMcs)
			Sleep(1);
	}
	while(PeekMessage(&msg,world_screen_window,0,0,PM_REMOVE))
	{
		(void)TranslateMessage(&msg);
		(void)DispatchMessage(&msg);
	}
	mEventInWMLoop = 0;
	if (!event)
		goto end;
	if (evtDeQueue(event,mask))
		return 1;
	/* didn't get anymessages */
	*event = mEventState;
	event->what = nullEvent;
end:	return 0;
}

boolean WaitNextEvent(uint32 mask,
	eventRecord *event,
	uint32 waitSleep,
	const SLRgn *wakeRgn)
{
	MSG		msg;
	int32		next_time;

	msg.time =~0u;
	PeekMessage(&msg,world_screen_window,0,0,PM_NOREMOVE);
	next_time = msg.time+17;
	waitnext = 1;
	mEventInWMLoop|= 1;
	do
		if (GetMessage(&msg,world_screen_window,0,0))
		{
			if (waitSleep) /* our Tick counter can sometimes fail to post; 2023.08.11 */
				if (msg.message==WM_USER || (int32)msg.time>=next_time)
					--waitSleep, next_time = msg.time+16;
			(void)TranslateMessage(&msg);
			(void)DispatchMessage(&msg);
		}
	while(waitSleep && ptInRgn(wakeRgn,mEventState.where.x,mEventState.where.y)<=0);
	waitnext = 0;

	while(PeekMessage(&msg,world_screen_window,0,0,PM_REMOVE))
	{
		(void)TranslateMessage(&msg);
		(void)DispatchMessage(&msg);
	}
	mEventInWMLoop = 0;
	if (event)
	{
		if (evtDeQueue(event,mask))
			return 1;
		*event = mEventState;
		event->what = nullEvent;
	}
	return 0;
}

/* 2014.07.30 */
void FlushEvents(uint32 mask,
	uint32 stopMask)
{
	eventRecord	event;
	/* appears to be a negation mask (idk) */
	for(mask&=~stopMask; evtDeQueue(&event,mask); )
		;
}

/* ---- window code */

static int8	deactiveCounter;

/*
 what does not work (at least in Win8) is detecting if taskbar is visible:
  IsWindowVisible (taskbar window always visible in autohide)
  whether the taskbar bounds is too small or has changed (unchanged)
  ABM_GETSTATE (always 0)
  SPI_GETWORKAREA (unchanged)
*/
static int WindowIsClipped(HWND win)
{
	HWND		upper, top = GetTopWindow(NULL);

	if (!win)
		return -1;
	for(upper=top; upper && win!=upper; )
	{
		if (IsWindowVisible(upper))
			return 1;
		upper = GetNextWindow(upper,GW_HWNDNEXT);
		if (upper==top)
			return -1; /* recursion! */
	}
	return 0;
}

static BOOL WindowResize(HWND window,
	WPARAM wParam,
	LPARAM lParam)
{
	RECT		wrect, crect;
	int		width = LOWORD(lParam);
	int		height = HIWORD(lParam);

	GetWindowRect(window,&wrect);
	GetClientRect(window,&crect);
	if (width>mEventWMin.x && width<=mEventWMax.x
			&& height>mEventWMin.y && height<=mEventWMax.y)
	{
		width+= wrect.right-wrect.left-crect.right;
		height+= wrect.bottom-wrect.top-crect.bottom;
		SetWindowPos(window,NULL,0,0,width,height,
			SWP_NOMOVE|SWP_NOZORDER|SWP_NOCOPYBITS);
		return TRUE;
	}
	return FALSE;
}

static void HandleWindowResized(HWND window)
{
	RECT		crect;
	rgb48		rgbdefault;

	GetClientRect(window,&crect);
	if (crect.right==mOS_screen_w && crect.bottom==mOS_screen_h)
		return; /* unaltered; WM_EXITSIZEMOVE can happen merely dragging window */
	rgbdefault.red = GetRValue(mEventWEraseColor)*0x101;
	rgbdefault.green = GetGValue(mEventWEraseColor)*0x101;
	rgbdefault.blue = GetBValue(mEventWEraseColor)*0x101;
	wRSVideoRealloc(&rgbdefault);
	mEventState.message = (void*)(appScreenChangeMessage<<24);
	mEventState.what = osEvt;
	PostMEvent(&mEventState);
}

/*static void wmUpdateMetrics(void)
{
	screenMetricX[0] = GetSystemMetrics(SM_CXSCREEN)*(1./65535);
	screenMetricX[1] = GetSystemMetrics(SM_CXVIRTUALSCREEN)*(1./65535);
	screenMetricY[0] = GetSystemMetrics(SM_CYSCREEN)*(1./65535);
	screenMetricY[1] = GetSystemMetrics(SM_CYVIRTUALSCREEN)*(1./65535);
}*/

static LRESULT CALLBACK mWndProc(HWND window,
	UINT what,
	WPARAM wParam,
	LPARAM lParam)
{
	static boolean	hasleft;
	uintptr_t	r;
#if DEBUG_WMANAGER
	uint8		keys[160];

	MGetKeys(keys);
	if (_MKDOWN(keys,MKV_LSHIFT) && _MKDOWN(keys,MKV_RMENU) && _MKDOWN(keys,MKV_F12))
		__debugbreak();
#endif
#if 0
	if ((what==WM_SIZE || what==WM_MOVE) && haveSetPos) /* don't save a default until set */
		if (!IsIconic(window) && !IsZoomed(window))
			GetWindowRect(window,&mEventOptWRect);
#endif
	//printf("WM_+0x%X\n",
	//	what);
	switch(what)
	{
	case WM_CREATE:
		if (mEventDropFilesFunc)
		{
			DragAcceptFiles(window,TRUE);
			ChangeWindowMessageFilter(WM_DROPFILES,MSGFLT_ADD);
			ChangeWindowMessageFilter(0x0049,MSGFLT_ADD);
		}
		break;
	case WM_DROPFILES:
		if (mEventDropFilesFunc)
			mEventDropFilesFunc((HDROP)wParam);
		break;
	case WM_DESTROY:
		if (mEventDropFilesFunc)
			DragAcceptFiles(window,FALSE);
		break;
	case WM_CLOSE:
	case WM_QUIT:
		mEventState.what = closeEvent;
		PostMEvent(&mEventState);
		break;
	//case WM_PALETTECHANGED: -- system colors changed
	//	break;
	case WM_PAINT:
		{
			PAINTSTRUCT	ps;

			(void)BeginPaint(window,&ps);
			if (screen_desktop_window.port.map)
				update_screen_rect(0,0,
					screen_desktop_window.port.map->w,
					screen_desktop_window.port.map->h);
			(void)EndPaint(window,&ps);
		}
		break;
	case WM_SIZE:
		mEventInWMLoop|= 8;
		return WindowResize(window,wParam,lParam);
	case WM_SIZING:
		mEventInWMLoop|= 8;
		goto norm;
	case WM_EXITSIZEMOVE:
		mEventInWMLoop&=~2;
		HandleWindowResized(window);
		goto norm;
	case WM_INPUT:
	{
		unsigned	size = sizeof(RAWINPUT);
		static RAWINPUT	raw[sizeof(RAWINPUT)];

		if ((mEventState.modifiers&(rightBtnState|btnState|middleBtnState))
				!=(rightBtnState|btnState))
			goto tstmr;
		if (!hasleft)
		{ /* SIGH; we don't always get the message when leaving the window */
			RECT		b;
			POINT		nativept;

			if (mEventState.when&1
					|| !wbGetMouseLoc(world_screen_window,&nativept,&b)
					|| PtInRect(&b,nativept))
			{
			tstmr:	if (mEventMouseDetail)
					goto recinp;
				return 0;
			}
			hasleft = 1;
		}
		/* unfortunately have to use raw input to detect mouseUp outside of HCLIENT */
	recinp:	raw->header.dwType = 0;
		GetRawInputData((HRAWINPUT)lParam,RID_INPUT,raw,&size,
			sizeof(RAWINPUTHEADER));
		if (raw->header.dwType==RIM_TYPEMOUSE)
		{
			if (raw->data.mouse.usButtonFlags&RI_MOUSE_BUTTON_1_UP)
				mEventState.modifiers|= btnState;
			if (raw->data.mouse.usButtonFlags&RI_MOUSE_BUTTON_2_UP)
				mEventState.modifiers|= rightBtnState;
			if (raw->data.mouse.usButtonFlags&RI_MOUSE_BUTTON_3_UP)
				mEventState.modifiers&=~middleBtnState;
			if (!(raw->data.mouse.usFlags&MOUSE_MOVE_ABSOLUTE))
			{
				mEventMouseRelW = mEventState.when;
				mEventMouseRelX+= raw->data.mouse.lLastX;
				mEventMouseRelY+= raw->data.mouse.lLastY;
			}
		}
		return 0;
	}
	case WM_GETMINMAXINFO:
	{
		LPMINMAXINFO	lpmmi = (LPMINMAXINFO)lParam;
		RECT		wrect, crect;

		if (mEventWMax.x<=0 || mEventWMax.y<=0)
			goto norm;
		GetWindowRect(window,&wrect);
		GetClientRect(window,&crect);
		crect.right-= crect.left;
		crect.bottom-= crect.top;
		wrect.right-= wrect.left+crect.right;
		wrect.bottom-= wrect.top+crect.bottom;
		lpmmi->ptMinTrackSize.x = mEventWMin.x+wrect.right;
		lpmmi->ptMinTrackSize.y = mEventWMin.y+wrect.bottom;
		lpmmi->ptMaxSize.x = lpmmi->ptMaxTrackSize.x = MAX(mEventWMax.x,mOS_screen_w)+wrect.right;
		lpmmi->ptMaxSize.y = lpmmi->ptMaxTrackSize.y = MAX(mEventWMax.y,mOS_screen_h)+wrect.bottom;
		return 0;
	}
	case WM_NCHITTEST:
		if ((r=DefWindowProc(window,what,wParam,lParam))!=HTCLIENT)
			hasleft = 1;
		return r;
	case WM_NCMOUSEMOVE:
		hasleft = 1;
		goto norm;
	case WM_MOUSEMOVE:
		mEventState.where.x = (int16)LOWORD(lParam);
		mEventState.where.y = (int16)HIWORD(lParam);
		if ((~mEventState.modifiers&(rightBtnState|btnState))
				&& !(wParam&(MK_LBUTTON|MK_RBUTTON)))
		{
			mEventState.message = 0;
			mEventState.modifiers|= rightBtnState|btnState;
			mEventState.what = mouseUp;
			PostMEvent(&mEventState);
		}
		hasleft = 0;
		break;		/* We need to set the cursor */
	case WM_SETCURSOR:
		if (LOWORD(lParam)==HTCLIENT)
		{
			if (mEventWClass.hCursor)
				(void)SetCursor(mEventWClass.hCursor);
			return TRUE;
		}
		goto norm;
	//case WM_CHAR:
	//case WM_DEADCHAR: -- dead keys are accent modifiers (e.g., umlaut)
	//case WM_SYSDEADCHAR:
	case WM_ERASEBKGND:
	{ /* default method flashes or in Windows 8+ paints white */
		RECT		rc;

		(void)GetClientRect(window,&rc);
		(void)SetBkColor((HDC)wParam,mEventWEraseColor);
		(void)ExtTextOut((HDC)wParam,0,0,ETO_OPAQUE,&rc,0,0,0);
		return 1;
	}
	/* MSDN: be sure to return TRUE if processing XBUTTON events */
	case WM_RBUTTONDBLCLK:
	case WM_RBUTTONDOWN:
		mEventState.message = (void*)(what==WM_RBUTTONDBLCLK ? dblClickFlag:0);
		mEventState.modifiers&=~rightBtnState;
		goto mdn;
	case WM_LBUTTONDBLCLK:
	case WM_LBUTTONDOWN:
		mEventState.message = (void*)(what==WM_LBUTTONDBLCLK ? dblClickFlag:0);
		mEventState.modifiers&=~btnState;
	mdn:	mEventState.what = mouseDown;
		mEventState.where.x = (int16)LOWORD(lParam);
		mEventState.where.y = (int16)HIWORD(lParam);
		PostMEvent(&mEventState);
		break;
	case WM_NCRBUTTONUP:
	case WM_RBUTTONUP:
		mEventState.modifiers|= rightBtnState;
		goto mup;
	case WM_LBUTTONUP:
		mEventState.modifiers|= btnState;
	mup:	mEventState.message = 0;
		mEventState.where.x = (int16)LOWORD(lParam);
		mEventState.where.y = (int16)HIWORD(lParam);
		goto mue;
	case WM_MBUTTONUP:
		mEventState.modifiers&=~middleBtnState;
		mEventState.message = (void*)wheelClickFlag;
		goto mue;
	case WM_MBUTTONDOWN:
		mEventState.modifiers|= middleBtnState;
		break;
	case WM_MOUSEHWHEEL:
		mEventState.message = (void*)((wParam>>23&wheelDownFlag)|wheelHFlag);
		goto mue;
	case WM_MOUSEWHEEL:
		mEventState.message = (void*)((wParam>>23&wheelDownFlag)|wheelVFlag);
		mEventMouseRelV+= wParam&wheelDownFlag<<23 ? 1:-1;
	mue:	mEventState.what = mouseUp;
		PostMEvent(&mEventState);
		break;
	case WM_SYSKEYDOWN:
	case WM_KEYDOWN:
	case WM_SYSKEYUP:
	case WM_KEYUP:
	{ /* need to trap the ALT and WIN keys! .. no success on WIN keys */
		uint8		mvk, down = (uint8)(1&~what), kchar, dchar;
	#ifdef DEBUG_MKV_DOWN
		if (down && !(lParam&0x40000000))
			printf("native keydown %d\n",
				wParam);
	#endif
		if (wParam+0u>=sizeof(mkvtab_win) || (mvk=mkvtab_win[wParam])>=MKC)
			break;
		if (down && lParam&0x40000000)
		{ /* auto-repeat */
			mEventState.what = autoKey;
			goto kchr;
		}
		if (!(lParam&0x1000000))
			switch(wParam)
			{ /* numlock-off numpad translation */
			case VK_CLEAR: mvk = MKV_KP5; break;
			case VK_PRIOR: mvk = MKV_KP9; break;
			case VK_NEXT: mvk = MKV_KP3; break;
			case VK_END: mvk = MKV_KP1; break;
			case VK_HOME: mvk = MKV_KP7; break;
			case VK_LEFT: mvk = MKV_KP4; break;
			case VK_RIGHT: mvk = MKV_KP6; break;
			case VK_UP: mvk = MKV_KP8; break;
			case VK_DOWN: mvk = MKV_KP2; break;
			case VK_INSERT: mvk = MKV_KP0; break;
			case VK_DELETE: mvk = MKV_KPDECIMAL;
			}
		else if (wParam==VK_RETURN)
			mvk = MKV_ENTER;
		menuDown = (what&~1)==WM_SYSKEYDOWN && lParam&0x20000000;
		if (down)
			_MKSET(mKeyMap,mvk);
		else
			_MKCLR(mKeyMap,mvk);
		updateEventState();
		/* as an EventRecord post: state&=~eventModifiedSwitch(mvk); */

		mEventState.what = keyUp-down;
	kchr:	dchar = 0;
		if ((kchar=mKCHRproc(mvk,mEventState.modifiers,
				mEventState.what==keyDown ? &deadKey:NULL,
				mEventState.what==keyDown ? &dchar:NULL))==0)
			mEventState.what = nullEvent;
		else
		{
			if (dchar)
			{
				mEventState.message = (void*)(dchar|KBL<<8);
				PostMEvent(&mEventState);
			}
			mEventState.message = (void*)(mvk<<8|kchar);
			PostMEvent(&mEventState);
		}
		break;
	}
	case WM_COPYDATA:
	{
		COPYDATASTRUCT	*pcds = (COPYDATASTRUCT*)lParam;

		if (pcds->dwData!=3)
			break;
		if (mEventCmdRequest)
			return mEventCmdRequest(pcds->lpData);
		break;
	//case WM_CONTEXTMENU: /* if you want to disable contextual menu on titlebar */
	//	break;
	}
	case WM_ACTIVATE: /* individual window (includes app activation) */
		mEventState.modifiers&=~charCodeMask;
		updateEventState();
		mEventState.modifiers|= LOWORD(wParam)!=WA_INACTIVE
			&& !HIWORD(wParam) /* minimized flag */
			? activeFlag|rightBtnState:rightBtnState; /* always clear right */
		mEventState.message = window;
		mEventState.what = activateEvt;
		PostMEvent(&mEventState);
		goto norm;
	case WM_ACTIVATEAPP: /* any window between programs */
		/*
		   Since around Windows Vista, apps can lose focus inexplicably.
		   Even with registry changes, the Windows error service can temporarily
		 throw off the current focus with an invisible "Cancel" window/dialog.
		   It's annoying, and we need to know when we're actually deactivating.
		   Typically, we won't get more than one consecutive tick WM_USER.
		   The only way to avoid deactivate is to have the window always on top.
		*/
		if (wParam==WA_INACTIVE)
		{
			if (deactiveCounter<3)
				++deactiveCounter;
			if (mOS_screen_fs && window==world_screen_window)
			{
				++deactiveCounter;
				goto inact;
			}
		} else if (deactiveCounter>1)
		{
			if (mOS_screen_fs && window==world_screen_window)
			{
				(void)ShowWindow(world_screen_window,SW_RESTORE);
			//	SetWindowPos(world_screen_window,HWND_TOP,
			//		0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_SHOWWINDOW);
			}
			deactiveCounter = 0;
			mEventState.message =
				(void*)(resumeFlag|btnState|rightBtnState|suspendResumeMessage<<24);
			mEventState.what = osEvt;
			PostMEvent(&mEventState);
			//wmUpdateMetrics();
		}
		goto norm;
	case WM_USER:
		mEventInWMLoop&=~2;
		if (deactiveCounter>0 && deactiveCounter<4)
			if (++deactiveCounter==4)
			{
				if (mOS_screen_fs && WindowIsClipped(window)>0)
					(void)ShowWindow(world_screen_window,SW_MINIMIZE);
			inact:	mEventState.message = (void*)(suspendResumeMessage<<24);
				mEventState.what = osEvt;
				PostMEvent(&mEventState);
				memset(mKeyMap,0,sizeof(mKeyMap));
			}
		break;
	case WM_NCLBUTTONDOWN:
		mEventInWMLoop|= 2;
		goto norm;
	case WM_MOVING:
		if (!(mEventInWMLoop&4)) /* prevent warp to last click */
			GetWindowRect(window,(RECT*)lParam);
		mEventInWMLoop|= 4;
	case WM_MOVE:
		//wmUpdateMetrics();
		goto norm;
	case WM_SYSCOMMAND:
		switch(wParam&0xFFF0)
		{
	#ifdef SC_SCREENSAVE
		case SC_MONITORPOWER:
		case SC_SCREENSAVE:
			if (mEventDisableSSAV)
				return 0; /* only works when in fore, non-minimized */
			break;
	#endif
		case SC_KEYMENU: /* alt-space et al. */
			return 0;
		}
		//wmUpdateMetrics();
	case WM_NULL:
	default:
	norm:	return DefWindowProc(window,what,wParam,lParam);
	}
	return 0;
}

void revive_screen_saver(void)
{
	mEventDisableSSAV = 0;
}

void kill_screen_saver(void)
{
	mEventDisableSSAV = 1;
}

static boolean	initReg;
static HANDLE	simutex;

int wRegisterDev(HINSTANCE devInstance,
	HICON icon,
	HCURSOR curs,
	COLORREF bgColor,
	const TCHAR *className)
{
	int		si;

	if (!className) /* name required */
		return -1;
#ifdef UNICODE
	si = !memcmp(className,L".SI:",4*sizeof(*className)) ? 4*sizeof(*className):0;
#else
	si = !memcmp(className,".SI:",4) ? 4:0;
#endif
	mEventWEraseColor = bgColor;
	mEventWClass.cbSize = sizeof(WNDCLASSEX);
	mEventWClass.style = CS_OWNDC|CS_DBLCLKS;
	mEventWClass.lpfnWndProc = mWndProc;
	mEventWClass.cbWndExtra = mEventWClass.cbClsExtra = 0;
	mEventWClass.hInstance = devInstance;
	mEventWClass.hIcon = icon;
	mEventWClass.hCursor = curs;
	mEventWClass.hbrBackground = CreateSolidBrush(bgColor);
	mEventWClass.lpszMenuName = NULL;
	mEventWClass.lpszClassName = className;
	mEventWDevName = className+si;

	if (!initReg)
	{
		initReg = 1;
		InitializeCriticalSection(&eventSection);
		if (si)
		{
			simutex = CreateMutex(NULL,TRUE,className);
			if (GetLastError()==ERROR_ALREADY_EXISTS)
				return 0;
		}
	}
	return !!RegisterClassEx(&mEventWClass);
}

int wUnregisterDev(HINSTANCE devInstance,
	const TCHAR *className)
{
	wSetVideoMode(-1,0,0,0,0); /* make sure all class windows are closed */
	if (simutex)
	{
		ReleaseMutex(simutex);
		CloseHandle(simutex);
		simutex = NULL;
	}
	return !!UnregisterClass(mEventWClass.lpszClassName,devInstance);
}

int wRequestSwitchDev(const TCHAR *className,
	TCHAR *param)
{
	COPYDATASTRUCT	cdata;
	HWND		w;
	DWORD		processID;

	if ((w=FindWindow(className,0))!=NULL)
	{
		Sleep(1);
		GetWindowThreadProcessId(w,&processID);
		AllowSetForegroundWindow(processID);
		SetForegroundWindow(w); /* needs to be called first...? */
		cdata.dwData = 3;
		cdata.cbData = strlen(param);
		cdata.lpData = param;
		SendMessage(w,WM_COPYDATA,0,(LPARAM)(LPVOID)&cdata);
	}
	return !!w;
}
#endif
#endif
