/*
 mfont.c
  (minimal) font loading

  Created: 2014.02.28 12:43 EST
  Revised: 2023.10.04
   Author: Adam P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.29
  Bitmapped font loading now works.  Update: calloc size was incorrect.
2014.04.11
  Loading by ID done by cast of name param as a int16.
2014.04.17
  Added spExtra, made x fixed-point.
2014.04.21
  Fixed underlining bug that wrote outside destination buffer.
2014.05.20
  Zero width for -1 (should widTable be int8?).
2016.08.22
  Color font (fctb) loading.
*/

#include <enviro64.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_MODULE_FONTS
 #define HAVE_OS
#endif
#include "../resex/resex.h"
#include "mtext.h"

/*
 //1 bit maps for 1 bit to reduce memory load
*/

/* ---- constants */

#define MacFONDType		MK4CC('F','O','N','D')

/* system font characters */
#define MacSysCharCommand	17
#define MacSysCharCheck		18
#define MacSysCharDiamond	19
#define MacSysCharApple		20

/* a bit of specs from Fonts.r & Fonts.h */
/* these constants & structures summarize, but aren't used at this level of rendering */

/* map flags */
#define famFixedWidth		0x80 /* otherwise, proportional */
#define famIgnoreFractWidths	0x40
#define famIntegralExtra	0x20 /* otherwise, fixed-point */
#define famIgnoreFractEnable	0x10
#define famNoSpacingAdjust	0x8 /* otherwise, adjust char. spacing */
/* reserved flags */
/* table flags */
#define tblHasCharWidths	0x2
#define tblHasImageHeights	0x1

/* ---- typedefs */

typedef struct fontAssc/*(6)*/
{
	int16		size;
	int16		style;
	int16		ID;
} fontAssc;

#include <pshpack2.h>

typedef struct fontFamilyDesc
{
	uint8		famFlags, tblFlags;
	int16		familyID;
	uint16		firstChar, lastChar;
	int16		ascent, descent;
	int16		leading, maxWidth;
	/* offsets */
/*x10*/	int32		widthsTableOffset;
	int32		kerningTableOffset;
	int32		styleTableOffset;
	int16 rsv1;
	int16		exWidBold;
/*x20*/	int16		exWidItalic;
	int16		exWidUnderline;
	int16		exWidOutline;
	int16		exWidShadow;
	int16		exWidCondensed;
	int16		exWidExtended;
	int16 rsv2;
	int32 rsvIntl;
	int16		version;
	/* font association table */
/*x34*/	uint16		asscCount;
	fontAssc	assc[1];
	/* 32-bit offset table if version 2 */
	/* bounding box table: zcount.w style.w, Rect */
} fontFamilyDesc;

#include <poppack.h>

#define bmtNoFontExpansion	0x40
#define bmtFixedWidth		0x20 /* otherwise, proportional */
#define bmtColor		0x2 /* otherwise, black on white */
#define bmtSynthetic		0x1

#define bmfDepthMask		0xC
#define bmfHasCharWidths	0x2
#define bmfHasImageHeights	0x1

#define bmfValidType(v)		((v&0x9C)==0x90)
#define bmfValidFlag(v)		!(v&0x70)

typedef struct bmFont
{
	uint8		type; /* required signature: 1xx100xx */
	uint8		flag; /* required signature: x000xxxx */
	uint16		firstChar, lastChar;
	int16		maxWidth, maxKerning;
	int16		negDescent;
	int16		bmWidth, bmHeight;
	/* docs say negDescent is added as high-word if != -1, but really if >0 */
	uint16		OWTLoc;

	int16		ascent, descent, leading;
	/* color font test shows that this field is not a double rowBytes but a width */
	uint16		bmWidth16;
	uint16		bm[1];
	/* directly after is the loc. table (lastChar-firstChar+3 ct) */
	/* OWTLoc to offset/width table, then optional char widths/img heights */
} bmFont;

/* ---- globals */

ffFamily	*fmFonts[mOS_MaxFonts];
int16		ffDefault, ffFWDefault;
uint8		ffSystem = 0;

/* ---- code */

/* loading */

/* some code here initially based on code found in AlephOne/RenderOther/sdl_fonts.cpp */

static int loadFont(ResFile_Memory *rMap,
	ffFont *fm,
	void *h,
	int ID)
{
	uint32		ct[256];
	void		*fch;
	const uint8	*fcp;
	const uint8	*s, *_s;
	uint32		o, v, c, z, allocsize;
	int		cc, r;
	uint8		u;
	/* pack flags */
	if ((s=(const uint8*)hLock(h))==NULL)
		return -1;
	_s = s;
	if (!fm->ptSize)
	{
#ifdef MFONT_TTF
		fm->sfnt = h;
		r = TTF_OpenSFNT(fm,0);
		return r;
#else
		return 0;
#endif
	}
	z = hSize(h);
	u = s[1];
	if (!bmfValidType(*s) || !bmfValidFlag(u)
			|| z<=OFFSET(bmFont,bm))
		return 0;
	fm->flag = (*s&0x60)|(u&ffHasfctb)|(u<<2&12);
	s+= 2;
	fm->firstChar = va_be16(s);
	fm->lastChar = va_be16(s);
	if (fm->lastChar<fm->firstChar)
		return 0;
	fm->maxWidth = va_be16(s);
	fm->maxKerning = va_be16(s);
	v = (int16)va_be16(s);
	fm->bmWidth = va_be16(s);
	fm->height = va_be16(s);
	o = va_be16(s);
	if ((signed)v>=0) //!=-1u)
		o|= v<<16;
	o = o*2+16; /* actual offset is word count from OWT field */
	if (o>=z)
		goto err;
	fm->ascent = va_be16(s);
	fm->descent = va_be16(s);
	fm->leading = va_be16(s);
	/* rowWords is not a double pitch/rowBytes, but a multiple of 16 pixels */
	v = va_be16(s)*2;
	u = (u&bmfDepthMask)>>2;
	fm->flag|= u;
#if MFONT_NO_UNPACK1BPP
	fm->pitch = u ? v*8u:v;
#else
	fm->pitch = v*8u; //>>u;
#endif
	fm->bmSize = (fm->height*fm->pitch+3)&~3;
	c = fm->lastChar-fm->firstChar+3;
	r = 1u<<(1u<<u);
	cc = 0, fch = NULL;
	if (fm->flag&ffHasfctb)
	#ifdef WINUSERAPI
	 if (rMap!=NULL || (fch=GetModuleResource(GetModuleHandle(NULL),
			MK4CC('f','c','t','b'),NULL,ID,NULL))==NULL)
	#endif
		fch = ResFile_GetResource(rMap,MK4CC('f','c','t','b'),NULL,ID,NULL);

	*ct = (1u<<31)-1;
	if (fch)
	{
		cc = r;
		memset(ct+1,0,sizeof(ct)-4);
		fcp = (const uint8*)hLock(fch);
		if ((r=(fcp[6]<<8|fcp[7])+1)>cc)
			r = cc;
		if (r)
		{ /* process the colors on the stack to free the fctb ahead of calloc */
		  /* it will be in 10:11:10 form so 31-bit color can turn out right */
			uint32		*d = ct;

			if (fcp[4]&0x80)
				do
				{
					fcp+= 8;
					*d++ = ((fcp[2]<<8|fcp[3])<<31-16&0x7FE00000)
						|((fcp[4]<<8|fcp[5])<<21-16&0x001FFC00)
						|(fcp[6]<<8|fcp[7])>>16-10;
				} while(--r);
			else
			{
				do
				{
					fcp+= 8;
					d[fcp[1]] = ((fcp[2]<<8|fcp[3])<<31-16&0x7FE00000)
						|((fcp[4]<<8|fcp[5])<<21-16&0x001FFC00)
						|(fcp[6]<<8|fcp[7])>>16-10;
				} while(--r);
			}
		}
		hFree(fch);
	} else if (r>2)
	{ /* 2-8 bits require some filler to not render random junk colors */
	  /* QuickDraw renders native/random colors for depth, so not same behavior */
		uint32		*d = ct+1; /* don't overwrite first or last */

		cc = r;
		r-= 2;
		do
			d[1] = *d = (1023u/15)|((2047u/15)<<10)|((1023u/15)<<21);
		while(d+=2,r-=2);
		*d = 0;
	}
	fm->bmSize+= cc*4;
	fm->ctCount = (short)cc;

	/* 2B ea. for the location table (a int16 list) */
	/* plus 2x1B ea. for the width-pair table */
	allocsize = fm->bmSize+c*(sizeof(int16)*3);
	if (fm->bmSize<=0 || (fm->data=calloc(1,allocsize))==NULL)
	{
	err:	fm->height = fm->bmWidth = 0;
		r = fm->bmSize>0 ? -2:0;
		goto end;
	}
	v*= fm->height;
	switch(u)
	{
	case 0:
	#if MFONT_NO_UNPACK1BPP
		memcpy(fm->data,s,v);
	#else
		b1to8(fm->data,s,v);
	#endif
		break;
	case 1:	v*= 2; b2to8(fm->data,s,v);
		break;
	case 2:	v*= 4; b4to8(fm->data,s,v);
		break;

	case 3: v*= 8; memcpy(fm->data,s,v);
	}
	if (cc)
		{r = cc*4; memcpy((int8*)fm->data+fm->bmSize-r,ct,r);}
	s+= v;
	{ /* loc table, wid table; */
		int16		*d = (int16*)(fm->bmSize+(int8*)fm->data);

		do
			*d++ = (int16)va_be16(s);
		while(--c);
		fm->widTable = d;
		if ((v=(fm->lastChar-fm->firstChar+3)*2)>0)
		{
			s = _s+o;
			do
				*d++ = *(const int8*)s;
			while(++s,--v);
		}
	}
	r = 1;
end:
	return r;
}

ffFamily *ffLoad(ResFile_Memory *rMap,
	const uint8 *name,
	int ID)
{
	void		*ffh, *h;
	ffFamily	*ff = NULL;
	ffFont		*fm;
	int8		*s;
	int		r;
	uint16		c, i;
#ifdef WINUSERAPI
	if (rMap!=NULL || (ffh=GetModuleResource(GetModuleHandle(NULL),
			MacFONDType,name,ID,NULL))==NULL)
#endif
	if ((ffh=ResFile_GetResource(rMap,MacFONDType,name,ID,NULL))==NULL)
		return NULL;
	if ((s=hLock(ffh))==NULL)
		goto end;
	s+= OFFSET(fontFamilyDesc,asscCount);
	c = (uint16)(va_be16(s)+1);
	if (!c || (ff=calloc(OFFSET(ffFamily,font)
			+(OFFSET(ffFamily,font[1])-OFFSET(ffFamily,font[0]))*c,1))==NULL)
		goto end;
	fm = ff->font;
	ff->asscCount = 0;
	ff->ID = ID;
	i = 0;
	do
	{
		int		v;

		fm->ptSize = va_be16(s);
		if (*s)
			continue; /* no support for upper 8-bits of style */
		fm->style = s[1], s+= 2;
		v = va_be16(s);
	#ifdef MFONT_TTF
		if (!fm->ptSize)
		{
		#ifdef WINUSERAPI
			if (rMap!=NULL || (h=GetModuleResource(GetModuleHandle(NULL),
					MK4CC('s','f','n','t'),NULL,v,NULL))==NULL)
		#endif
			{
				if ((h=ResFile_GetResource(rMap,MK4CC('s','f','n','t'),NULL,v,NULL))==NULL)
					continue;
			}
		} else
	#endif
		{
	#ifdef WINUSERAPI
			if (rMap!=NULL || ((h=GetModuleResource(GetModuleHandle(NULL),
					MK4CC('N','F','N','T'),NULL,v,NULL))==NULL
					&& (h=GetModuleResource(GetModuleHandle(NULL),
					MK4CC('F','O','N','T'),NULL,v,NULL))==NULL))
	#endif
			if ((h=ResFile_GetResource(rMap,MK4CC('N','F','N','T'),NULL,v,NULL))==NULL
					&& (h=ResFile_GetResource(rMap,MK4CC('F','O','N','T'),
					NULL,v,NULL))==NULL)
				continue;
		}
		if ((r=loadFont(rMap,fm,h,v))>0)
		{
			++fm, ++ff->asscCount;
		#ifdef MFONT_TTF
			if (fm[-1].sfnt)
				continue; /* original data is kept in memory! */
		#endif
		} else /* this 'else' was missing, writing NULL to next slot or out of bounds */
		{
		#ifdef MFONT_TTF
			dprintf("failed to load font ID %d:%i (%d);g",
				ID,i,r);
			fm->sfnt = NULL;
		#endif
		}
		hFree(h); h = NULL;
	} while(++i<c);
end:
	hFree(ffh);
	if (!_heap_validate(NULL))
		dprintf("ffLoad: heap was broken");
	return ff;
}

int ffFree(ffFamily *ff)
{
	ffFont		*fm;
	int		c;

	if (!ff || (c=ff->asscCount)<0)
		return -1;
	if (c)
	{
		fm = ff->font;
		do
			if (fm->data)
			{
			#ifdef MFONT_TTF
				if (fm->sfnt)
				{
					TTF_CloseFont(fm); /* clears fm->data */
					hFree(fm->sfnt);
					fm->sfnt = NULL;
				} else
			#endif
				free(fm->data);
				fm->data = NULL;
			}
		while(++fm,--c);
	}
	c = ff->asscCount;
	ff->asscCount = 0;
	free(ff);
	return c;
}

/*
 QD Text finds the best size in divisibility for "ideal" scaling.
 This, however, goes for the closest to style or plain larger (down-scaling).
*/
ffFont *ffMatch(ffFamily *ff,
	int size,
	int8 styl)
{
	ffFont		*fm;
	ffFont		*fc, *fz; /* closest with style, closest without */
	int		c, v;
	boolean		ttz, ttc;

	if (!ff || (c=ff->asscCount)<=0)
		return NULL;
	if (size<=0)
		size = 12;
	fm = ff->font, fz = fc = NULL, styl&= 3, ttz = ttc = 0;
	do
	{
		if ((v=fm->style&3)==0 && !ttz)
			if (!fm->widTable)
				fz = fm, ttz = 1;
			else if (!fz || ABS(size-fm->ptSize)<=ABS(size-fz->ptSize))
				fz = fm;
		if (!fm->widTable)
		{
			if (v==styl)
				fc = fm, ttc = 1;
		} else if (fm->ptSize>=size)
		{
			if (v==styl && fm->ptSize==size)
				return fm; /* best case: match */
			/* closest style without going over */
			if (!ttc && fm->ptSize==size && !(v&~styl))
				fc = fm;
		}
	} while(++fm,--c);
	/* next case, closest size without style; worse case, first */
#ifdef MFONT_TTF
	fc = fc ? fc:fz ? fz:ff->font;
	if (!fc->widTable)
		(void)TTF_SetPtSize(fc,size);
	return fc;
#else
	return fc ? fc:fz ? fz:ff->font;
#endif
}

/* global loader */

#if !mOS_FontsLoadSequential
int ffIterLoad(int ID,
	const uint8 *name,
	void *rmap)
{
(void)name;
	if (ID+0u>32767 || fmFonts[ID])
		return 0; /* out of range or font already loaded in slot */
	return !!(fmFonts[ID]=ffLoad(rmap,NULL,ID));
}
#endif

int fmLoadAll(ResFile_Memory *rmap,
	int (__cdecl iterator)(const uint8 *name,
		int ID))
{
	RsrcInfo	info;
	int		i, succCount = 0;
	int		r, c;
#ifdef WINUSERAPI
	HANDLE		wMod = GetModuleHandle(NULL);

	c = CountModuleResources(wMod,MacFONDType);
	for(i=0; i<c; ++i)
		if ((r=IterateModuleResources(wMod,MacFONDType,rmap,ffIterLoad))>=0)
			succCount+= r;
#endif
	c = ResFile_CountResources(rmap,MacFONDType,&r);
	for(i=0; i<c; ++i)
	{
		int		r;

		(void)ResFile_GetIndResource(rmap,MacFONDType,~i,&info,&r);
		if (r>=0)
		{
			if (iterator && (r=iterator(info.name,info.ID))<0)
				break;
			if (!r)
				r = ffIterLoad(info.ID,NULL,rmap);
			succCount+= r;
		}
	}
	return succCount;
}

int fmLoadMaster(const char *name,
	int (__cdecl iterator)(const uint8 *name,
		int ID))
{
	ResFile_Memory	*fontRMap;
	FILE		*fontFile;
	int		r = 0;

	if (!name)
		return -1;
	if ((fontFile=fopen(name,"rb"))==NULL)
		return 0;
	if (ResFile_LoadMap(fontFile,0,&fontRMap,0)>=0)
		r = fmLoadAll(fontRMap,iterator);
	free(fontRMap);
	(void)fclose(fontFile);
	return r;
}

void fmUnloadAll(void)
{
	ffFamily	**ff = fmFonts;
	int		c = GET_ARRAY_COUNT(fmFonts);

	do
		if (*ff)
			{(void)ffFree(*ff); *ff = NULL;}
	while(++ff,--c);
}

int fmFlushCache(ffFamily *ff)
{
	int		r = 0;
#ifdef MFONT_TTF
	ffFont		*fm;
	int		c;

	if (!ff || (c=ff->asscCount)<0)
		return -1;
	if (c)
	{
		fm = ff->font;
		do
			if (fm->data)
				r+= TTF_FlushCache(fm->sfnt);
		while(++fm,--c);
	}
#endif
	return r;
}

/* ---- port settings */

int rTextFace(grafPort *p,
	int16 font,
	int16 size,
	uint16 style)
{
	if (!p)
		return -1;
	p->txFace.font = font;
	p->txFace.size = size;
	p->txFace.style = style;
	if ((p->txFace.fm=ffMatch(fmFonts[font>=GET_ARRAY_COUNT(fmFonts) ? ffDefault:font],
			size,(int8)style))==NULL)
		p->txFace.fm = ffMatch(fmFonts[ffDefault],size,(int8)style); /* last resort */
	return !!p->txFace.fm;
}

int rTextFont(grafPort *p,
	int16 font)
{
	if (!p)
		return -1;
	return rTextFace(p,font,p->txFace.size,p->txFace.style);
}

int rTextSize(grafPort *p,
	int16 size)
{
	if (!p)
		return -1;
	return rTextFace(p,p->txFace.font,size,p->txFace.style);
}

int rTextStyle(grafPort *p,
	uint16 style)
{
	if (!p)
		return -1;
	return rTextFace(p,p->txFace.font,p->txFace.size,style);
}

int rSetFScaleDisable(grafPort *p,
	int fnoscale)
{
	if (!p)
		return -1;
	if ((p->txFace.fflag&ffBMNoScale)==(fnoscale ? ffBMNoScale:0))
		return 0;
	p->txFace.fflag^= ffBMNoScale;
	return rTextFace(p,p->txFace.font,p->txFace.size,p->txFace.style);
}
