/*
 mgraf.h
  minimal OS layer graphics library

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#ifndef __MGRAF_H
#pragma once

#include "my_cseries.h"

/* ---- constants */

enum
{ /* pType indices, bits */
	pTypeIndexed,
	pTypeFixed,		/* pTable lock */
	pTypeRawRGB,		/* direct, no table - gamma ramp applies to device */

	pTypeAlphaBit	= 5,	/* has alpha channel (what's left at top) */
	pTypeLGreenBit,		/* green channel is 1 bit larger (i.e., 5:6:5) */
	pTypeLSBBit,		/* values in little endian ordering */

	pTypeAlpha	= 1<<pTypeAlphaBit,
	pTypeLGreen	= 1<<pTypeLGreenBit,
	pTypeLSB	= 1<<pTypeLSBBit,

	pTypeNativeEndian = !BIG_ENDIAN<<pTypeLSBBit
};

enum
{ /* dpi */
	hvResDefault	= 72<<16 /* if screen resolution is not available */
};

enum
{ /* pen modes */
	mSrcCopy,		/* no transparency by default */
	mSrcOr,			/* pixel-merge, OR bits indexed/AND direct */
	mSrcXor,		/* inversion */
	mSrcBic,		/* pixel-merge, AND bits indexed/OR direct */
	mSrcNotCopy,		/* not: invert src; reverse colors for solids/patterns */
	mSrcNotOr,
	mSrcNotXor,
	mSrcNotBic,
	mTransparent,		/* mask out index-zero or bkColor src pixels */
	mBlend,			/* interpolate, given opColor */
	mAdMax,			/* highest of src/dst components */
	mAdMin,			/* lowest of src/dst components */
	mAddPin,		/* add, no higher than channels in opColor */
	mAddOver,		/* add, no clipping */
	mSubPin,		/* subtract, no lower than channels in opColor */
	mSubOver,		/* subtract, no clipping */
	mSrcPat		= 0x10,	/* img flag: pattern with mx/mh position repeats */
	mTextGray	= 0x20,	/* text flag: render w. 50% fg/bg blend or gray pattern */
	mDither		= 0x40,	/* img flag: dither if dst. can't accurately represent */
	mMergeDown	= 0x80,	/* img flag: interpolate/merge on downscaling */
	mMergeUp	= 0x100	/* img flag: interpolate on scaling */
};

#define RECT_RGN_SIZE		OFFSET(SLRgn,x)

enum
{ /* color conversion routines */
	ccGray,
	ccDirect,
	ccDirectBS32,
	ccDirectBS16,
	ccNative	= 0x2,
	ccIndexed	= 0x4,	/* passed to remapper, supresses indexed output */

	remapDInvert	= 0x1,	/* inverts direct table output */
	remapReverse	= 0x8,	/* builds in reverse for logical cts */
	remapForce30	= 0x40,	/* forces 10:10:10 output */
	remap32bit	= 0x80	/* needs 32-bit remapping table over 16-bit */
};

enum
{ /* newPixMap depth hint bits */
	newDepthLGreenBit = 11,	/* green channel is 1 bit larger (direct depths only) */
	newDepthMSBBit,		/* force big endian (native by default) */
	newDepthLSBBit,		/* force little endian (native by default) */
	newDepForce4Bit,	/* next step above 1bpp is 4bpp; for Windows compatibility */

	newDepthLGreen	= 1<<newDepthLGreenBit,	
	newDepthMSB	= 1<<newDepthMSBBit,
	newDepthLSB	= 1<<newDepthLSBBit,
	newDepthForce4	= 1<<newDepForce4Bit
};

enum
{
	pPatRGBDither	= 0x61 /* magic number for auto-dither pattern */
};

enum
{ /* face flags */
	ffBMNoScale	= 0x1  /* don't adjust scales on bitmap size mismatch */
};

/* ---- typedefs */

#include <pshpack2.h>

typedef struct hsv64
{
	uint32		hue;
	uint16		sat, val;
} hsv64;

typedef struct rect2dq
{
	ifixed		top, left, bottom, right;
} rect2dq;

/* explicit scanline region (smallest size=empty; rects require redundancies) */
typedef struct SLRgn
{
	int32		size; /* total size, including header */
	rect16		bBox;
	int16		x[1]; /* lines end with 0x7FFF */
} SLRgn;

typedef struct pixMap/*(40+)*/
{ /* aligned like crazy */
	uint16		w, h;
	pt2d16		pOrigin;
	int32		pitch; /* bytes per line */
	uint8		pSize; /* 1, 2, 4, 8, 16 or 32; otherwise, codec type */
	uint8		pType; /* see above */
	uint8		bpp; /* 1-32 */
	uint8		minCmp; /* smallest bit count for a direct component */
/*x10*/	ufixed		hRes, vRes; /* horiz. & vert. ppi resolution */
	int8		*baseAddr; /* access to the pixel data (may be NULL) */
	void		*pHand; /* the applicable mhandle, if used for baseAddr */
	/* start of the applicable tables */
	cTab64		*pTable;
	/* invalidated every time the ct seed changes */
	uint8		*iTable; /* inverse table; minCmp=resolution */
} pixMap;

typedef struct pixPat
{ /* like QD, only the largest power of two contained is valid for the width */
  /* however, here, the height can be any non-zero value */
	void		*px; /* 8bpp or original pattern */
	void		*nx; /* native table/pattern (generated on-the-fly) */
	int32		xpitch;
	int32		pitch;
	uint16		w, h;
	/* if seed bit #31=1: [dstfmt.16]; upper byte of flags is source format */
	cTab64		ct; 
} pixPat;

typedef struct penState
{
	pt2dq		pnLoc, pnSize;
	pixPat		*pPat;
	uint8		pnMode;
} penState;

typedef struct ffFace
{
	struct ffFont	*fm;
	int16		font, size;
	uint16		style;
	uint8		fflag;
} ffFace;

typedef struct grafPort
{
	pixMap		*map;
	SLRgn		*clipRgn, *clipSave, *visRgn;
	pt2d16		origin;
	uint32		fgSolid, bkSolid;
	rgb48		fgColor, bkColor, opColor;
	uint8		pnMode, txMode;
	pixPat		*pPat;
	pt2dq		pnLoc, pnSize, frSize;
	ffFace		txFace;
	ifixed		chExtra, spExtra;
	ifixed		tabAlign, tabOrigin;
} grafPort;

typedef struct RGBLineInfo
{
	uint32		k[3];
	uint8		z[3], h[3];
} RGBLineInfo;

/* for indexed-native, gt is the inverse table cast as a 'long int' */
/* otherwise, the green channel bit count */
#define DEF_RGB_RTN(n)		uint32 n(int i, const void *gt, const rgb48 *s)
#define TYP_RGB_RTN(n)		uint32(*n)(int, const void*, const rgb48*)

typedef TYP_RGB_RTN(rgbRoutine);

#include <poppack.h>

/* ---- macros, inlines */

static __forceinline ifixed addu(ifixed a,
	ufixed b)
{
	ifixed		v = a+b;

	return v<a ? IMAX(v):v;
}

static __forceinline ifixed subu(ifixed a,
	ufixed b)
{
	ifixed		v = a-b;

	return v>a ? IMIN(v):v;
}

static __forceinline ufixed idim(ifixed a,
	ifixed b)
{
	return a>b ? a-b:b-a;
}

static __forceinline inline uint32 formDirect(const rgb48 *c,
	int g,
	int rb)
{
	int		h = 16-rb;

	return (c->red>>h<<g|c->green<<g>>16)<<rb|c->blue>>h;
}

#define formGray(r,g,b,u)	(u=r,u+=(((u+g)>>1)+b)>>1,(u>>1)+g)

#define pixMapIsDirect(m)	!!(uint8)((m)->pType&(-pTypeRawRGB&~pTypeLSB))

/* formPxFmt: produces a 14-bit hash to quicky tell different formats apart */
/* inverse table resolution isn't technically part of the pixel format, but
 changes in resolution affect color selection, sometimes rendering different results
*/
static __forceinline int formPxFmt(const pixMap *m)
	{return m->pSize<<8|((uint8)(m->minCmp ? m->minCmp:m->bpp)
	|(m->pType&(pTypeLSB|pTypeLGreen)));}

/* unaligned read macros */
#if env68k|envppc|envx86
 #define urb16(a)		(0+*(const uint16*)(a))
 #define urb32(a)		(0+*(const uint32*)(a))
 #define urw32(a)		(0+*(const uint32*)(a))
#else
 #define urb16(a)		((uint32)((const uint8*)(a))[!BIG_ENDIAN]<<8 \
	|((const uint8*)(a))[BIG_ENDIAN])
 #define urb32(a)		((uint32)((const uint8*)(a))[3*!BIG_ENDIAN]<<24 \
	|((const uint8*)(a))[2-BIG_ENDIAN]<<16 \
	|((const uint8*)(a))[1+BIG_ENDIAN]<<8 \
	|((const uint8*)(a))[3*BIG_ENDIAN])
 #define urw32(a)		((uint32)((const uint16*)(a))[!BIG_ENDIAN]<<16 \
	|((const uint16*)(a))[BIG_ENDIAN])
#endif

#define q2dsort2(a,b)		do{ pt2dq _u; \
	if(a.y>b.y) _u=a,a=b,b=_u; \
}while(0)

#ifndef QD_ROUNDED_LINES
 #define QD_ROUNDED_LINES 1
#endif

/* ---- globals */

extern const uint32	pSizeMasks[32], pMCmpMasks[16], pDepthK[32];
extern const uint8	pDepthX[32], pMinCmpX[32];
extern const rgbRoutine	rgbCCRoutine[8];
/* if !=0, rCLineList renders single pixel lines if pen is no larger than 1.0/1.0 */
extern boolean		mCLineSingle1px;

/* ---- prototypes */

/*
 newPixMap
	Allocates a pixMap, using calloc and the fields given.
	Pass zero for bpp to attempt to match the screen format.
	Call disposePixMap to dispose.
  Notes:
	The initial NULL pTable implies grayscale for indexed modes.
	The pitch is aligned to four bytes.
	The user may modify the structure to see fit.
	See the enum above for format hint bits in 'bpp'.
  Result:
	The map, or NULL on error.
*/
pixMap *newPixMap(int bpp,
	int width,
	int height);
/*
 disposePixMap
	Disposes of a newPixMap-created pixMap.
	Pixels are not disposed if the pHand field is NULL.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int disposePixMap(pixMap*);

/*
 pmAllocate
	Creates a movable native handle for a buffer.
	The contents of the buffer is undefined.
  Result:
	Returns 1 if a buffer was created.
*/
int pmAllocate(pixMap*);
/*
 pmFree
	Opposite of pmAllocate, deallocates the movable pixel buffer.
	pHand and baseAddr are then set to NULL.
	Nothing else is affected or disposed.
*/
int pmFree(pixMap*);
/*
 pmCopy
	Makes a copy of a pixMap.
	If the source is null, the resulting pixmap will be allocated empty.
  NOTE:
	The destination is free'd before making the duplicate.
  Result:
	Returns 1 on success; 0 on empty; negative on error.
*/
int pmCopy(const pixMap *src,
	pixMap **dst);
/*
 pmConvert
	Makes an in-place conversion of a pixmap.
	Resizes or allocates pixel space in pHand if necessary.
	If an erasure color is specified, any new void will be filled.
	If depth is <0, then the pixel format will remain unchanged.
	Otherwise, the new format will be the same as if newPixMap was called.
	Width and height must be valid.
  Result:
	Returns 1 on success; 0 on no change; negative on error.
*/
int pmConvert(pixMap *m,
	short depth,
	int w,
	int h,
	const cTab64 *ct, /* can be NULL */
	const rgb48 *eraseColor); /* can be NULL */

/*
 pmErase
	Fills a single-pixel pen for the entire bounds.
	The pixMap must be locked (baseAddr loaded) prior to the call.
  Result:
	Returns 1 on success; zero if unlocked or empty.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int pmErase(pixMap*,
	uint32 penc);
/*
 pmFilter
	Performs a direct filter for a number of scanlines.
	The ctable must be >= the size indicated by the green channel depth
	- or, for grayscale images, m->bpp.
	This is faster and more capable than using the mBlend mode.
  Notes:
	The pixMap must be locked prior to the call.
	Grayscale tables call for 1024 entries.
	Takes a bounding rectangle in absolute, unclipped coordinates.
  Result:
	Returns 1 on success; zero if unlocked or either ptr are empty.
*/
int pmFilter(pixMap *m,
	const cTab64 *t,
	const rect16*); /* can be NULL */

/*
 lockPixels
	Locks the pixmap baseAddr, updating the map if necessary.
  Result:
	The pixels base address, or NULL on error.
*/
int8 *lockPixels(pixMap *p);
/*
 unlockPixels
	Unlocks the pixmap baseAddr, updating the map if necessary.
  Result:
	Positive if unlocked, zero if nothing changed, negative on error.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int unlockPixels(pixMap *p);

/*
 openGrafPort
	Initializes a grafPort, with default values and a pixMap.
	Close with closeGrafPort.
  Note:
	This expects the port to not have a valid pixMap (will not dispose).
	This calls newPixMap; see the comment on that function above.
  Result:
	Positive on success; otherwise, an error.
*/
int openGrafPort(grafPort *p,
	int depth,
	int width,
	int height,
	boolean allocate);
/*
 closeGrafPort
	Disposes of the contents within a grafPort.
	This routine does not 'free' the ptr passed.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int closeGrafPort(grafPort*);

/*
 convertGrafPort
	Converts and recalculates a port via pmConvert, RGBFore, etc.
	Important: this function necessarily (re)allocates pixel data.
*/
int convertGrafPort(grafPort *p,
	int depth,
	int width,
	int height,
	const cTab64 *ct,
	const rgb48 *eraseColor);

/*
 pixMapToNativeDesc
	Allocates and defines a native platform bitmap/pixmap from a pixMap.
*/
#ifdef WINUSERAPI
BITMAPINFO *pixMapToNativeDesc(pixMap*);
#elif TARGET_OS_MAC
PixMapHandle pixMapToNativeDesc(pixMap*);
#endif

/* color selection & inverse tables */

/*
 RGBFore/RGBBack
	Sets the fore/back color, RGB and pen.
	If a pattern is loaded, the following ct entry will be set:
	 fore: last; back: first.
*/
void RGBFore(grafPort *p,
	const rgb48*);
void RGBBack(grafPort *p,
	const rgb48*);

/*
 RGBToPen
	Forms a multi-pixel for the passed pixmap format, given a color.
	It may build an inverse table for the pixMap.
  Result:
	Can only return zero on error.
*/
uint32 RGBToPen(pixMap *m,
	const rgb48*); /* ptr must be valid */
/*
 PenToRGB
	Converts a pen color to a 48-bit color.
  Note:
	This only takes the lower bits pixel of the passed value.
  Result:
	Returns 1 on success; negative on error.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int PenToRGB(pixMap *m,
	uint32 penc,
	rgb48 *rgb);
/*
 GetPixelRGB
	Converts a pixel, given a point, to a 48-bit color.
  Note:
	This may call lockPixels/unlockPixels.
  Result:
	Returns 1 on success; negative on error.
	Returns zero if out of bounds, where 'v' is left undefined.
*/
int GetPixelRGB(pixMap *m,
	int x,
	int y,
	rgb48 *v);
/*
 PenGetGray
	Makes a gray color solid from the fore and back solids.
	If it cannot find a suitable solid, 'solid' is undefined.
	Note that solids are pixel values, and are considered temporary.
  Result:
	Returns 1 if found; negative on error.
*/
int PenGetGray(grafPort *p,
	uint32 *solid);

/*
 pmGetRGBCC
	Calculates the color conversion code for a pixMap.
  Result:
	The index in the rgbCCRoutine table; negative on error (gt is undefined).
*/
int pmGetRGBCC(pixMap *m,
	void **gt); /* must be valid */

/*
 copyCTable
	Makes a copy of a color table, allocating 'out' if necessary.
	If 'out' is already allocated and 'in' is NULL, 'out' will be free'd.
*/
int copyCTable(cTab64 **out,
	const cTab64 *in);
/*
 allocCTable
	Makes a copy of colors, filling in zeros. Flags set to 0x8000.
*/
cTab64 *allocCTable(const rgb30 *colors, /* can be NULL (all black) */
	int count);

/*
 ctMakeITables
	Builds inverse & direct tables, given a color table.
	minCmp is where the channel resolution resides.
	The res. for a gray table is always 8, though minCmp=0.
  Result:
	The effective resolution (0=grays); otherwise a basic error code.
*/
int ctMakeITables(const cTab64 *ct,
	uint8 **itab, /* must be valid */
	int bpp,
	int res,
	boolean forceColor,
	boolean altAccuracy); /* don't make gray table */
int pmMakeITables(pixMap*,
	int res,
	boolean altAccuracy);

/*
 pmDefaultMacTable
	Depending on the bpp depth, loads the appropriate Mac OS table.
  Result:
	Returns 1 on success; zero if the pm has a direct or fixed clut.
*/
int pmDefaultMacTable(pixMap *m,
	int gray); /* pass negative for reverse order */

/*
 rGenRemappingTable
	Builds a remapping table for a given pixmap, given a color table.
	Other considerations are in the remap/cc flags enum above.
  Result:
	Returns a new pointer on success; NULL on failure.
*/
void *rGenRemappingTable(pixMap *m, /* must be valid */
	const cTab64 *ct,
	uint8 bpp,
	uint8 flag);

/*
 HSV2RGB/RGB2HSV
	Uses high-precision math to accurately convert HSV to RGB or vice versa.
	If either parameter is NULL, the result is undefined.
*/
void HSV2RGB(rgb48 *rgb,
	const hsv64 *hsv);
void RGB2HSV(hsv64 *hsv,
	const rgb48 *rgb);

/* patterns */

/*
 rMakePat
	Creates a pixPat out of an 8x8 monochrome map.
	Replaces pattern in 'pat' if !=NULL.
  Result:
	1 on success; negative on error.
*/
int rMakePat(pixPat **outpat,
	const uint8 inpat[8],
	const rgb48 *fore,
	const rgb48 *back);
#define rDisposePat(p)		rMakePat(&p,NULL,NULL,NULL);

int rMakeRGBPat(pixPat **outpat,
	const uint8 inpat[8],
	const rgb48 *color);

void rPenPat(grafPort *p,
	const uint8 inpat[8]);

/*
 rCopyPat
	Allocates a copy of a pattern.  Use rDisposePat when done.
*/
int rCopyPat(pixPat **out, /* must be valid */
	const pixPat *in);
/*
 rConvertPat
	Fixes up a pattern for drawing to a pixmap.
	This function may increase the depth of the original pattern.
  Result:
	Zero on no change; positive on expansion(s) made; negative on error.
*/
int rConvertPat(pixPat *pat, /* must be valid */
	pixMap *dm, /* must be valid */
	uint16 dflag);
/*
 rAllocPat
	(Re)allocates a pixPat to fit a color count, and (re)initializes fields.
	The pixel data is allocated to fit the depth, width and height.
	If colorCount is 0, it will be calculated for the depth.
	Any void is set to zeroes.
  Result:
	1 on success; 0 on empty; negative on failure.
*/
int rAllocPat(pixPat **outpat,
	uint8 depth,
	uint16 w,
	uint16 h,
	uint16 colorCount);

/*
 rSavePen
	Makes a save-state for a port, copying the pattern if it exists.
	Use rFreePen/rRestorePen when finished.
*/
int rSavePen(grafPort *p,
	penState *state);
#ifdef _CRTNOP
_CRTNOP
#endif
int rRestorePen(grafPort *p,
	penState *state,
	boolean freeState);
#define rFreePen(s) rRestorePen(NULL,s,1)
/*
 rPenNormal
	Sets a port's pen state to the default.
	The optional 'oldstate' allows you to save it, no additional allocations.
	Like rSavePen, if you won't restore the pen, use rFreePen.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rPenNormal(grafPort *p,
	penState *oldstate);

/*
 rSetClip
	Sets a port's clipping region and its aggregate.
	This must be called for windows to draw properly on screen.
	Offscreen ports may set clipRgn directly.
	If not set to copy, the previous is disposed and the pointer is owned.
  Result:
	Result is negative on error; 0 on empty; 1 on non-empty.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rSetClip(grafPort *p,
	SLRgn *srgn,
	boolean copy);

/*
 rAddClip
	Intersects a port's clipping region and its aggregate.
	Temporary and offscreen clips may call sectRgn with clipRgn directly.
	This function can be used to restore a clip by passing a wide-open srgn.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rAddClip(grafPort *p,
	const SLRgn *srgn);

/* graphics */

/*
 OutsetRect16
	Outsets a 16-bit rectangle, performing maximum clipping if necessary.
  Note:
	This does not perform minimum clipping (e.g., if either x or y are negative).
*/
void OutsetRect16(rect16 *b, /* must be valid */
	int x,
	int y);
void OffsetRect16(rect16 *b, /* must be valid */
	int x,
	int y);
/*
 SectRect16
	Calculates the smaller boundary within two rectangles.
	Returns 0 if empty.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int SectRect16(rect16 *d, /* must be valid */
	const rect16 *s);
#ifdef _CRTNOP
_CRTNOP
#endif
int UnionRect16(rect16 *d, /* must be valid */
	const rect16 *s);
void SubRect16(rect16 *d, /* must be valid */
	const rect16 *s);
/*
 FixRect
	Makes a fixed-point rectangle with +0.5 as fraction.
*/
void FixRect(rect2dq *b, /* must be valid */
	int left,
	int top,
	int right,
	int bottom);
void FixRect16(rect2dq *b, /* must be valid */
	const rect16 *r); /* can be NULL (set empty) */
/*
 FixPt
	Makes a fixed-point 2D point with +0.5 as fraction.
	The values are pinned to the maximum range.
*/
void FixPt(pt2dq*, /* must be valid */
	int x,
	int y);

int ptInRect16(const rect16*,
	pt2d16);

/* the results for the renderers:
 -2 = memory error (std utility error)
 -1 = parameter error or bad/uninitialized data
  0 = clipped-out or empty
  1 = within bounds (doesn't necessarily mean any pixels were changed)
*/

/* regions */

/*
 newRgn
	Creates an empty region.
	If not wideopen, bBox is set empty with maximus/minimus.
	Returns NULL on error.
*/
SLRgn *newRgn(boolean wideopen);

/*
 copyRgn
	Sets contents of a source region.
	If the source is empty, the destination is disposed and set to NULL.
  Result:
	Zero if empty/disposed.
	Positive if not empty and successful.
	Negative on error.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int copyRgn(SLRgn **drgn,
	const SLRgn *srgn);
/*
 disposeRgn
	Disposes contents of a region; the destination is set to NULL.
	Returns 0 on success; negative on error.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int disposeRgn(SLRgn **rrgn);

/*
 emptyRgn
 	Checks a region to see if it's empty (or NULL).
	With 0 passed to boundsOnly, can be used to validate a region.
  Result:
	Zero if the region has positive space within bounds.
	Positive if empty, according to the parameters given.
	Negative on error.

*/
int emptyRgn(const SLRgn *rgn,
	uint8 boundsOnly);

/*
 rectRgn
	Makes a rectangle region, implied (no data) or filled.
	The region will be created if need be.
*/
int rectRgn(SLRgn **rrgn,
	int16 left,
	int16 top,
	int16 right,
	int16 bottom,
	boolean fill);
/*
 arcRgn
	Makes an arc and/or clipping region for the oval arc.
	Returns 2 if only the tangent region is made.
*/
int arcRgn(SLRgn **rrgn, /* can be NULL */
	const rect2dq *ov, /* can be NULL */
	int16 ctop,
	int16 cbot,
	ufract startAngle,
	ufract arcAngle,
	boolean addCenter,
	SLRgn **tanrgn); /* saved tangent region; can be NULL */
#define ovalRgn(r,o,t,b) arcRgn(r,o,t,b,0,~0u,0,0)

/*
 roundRectRgn
	Makes a round rectangle region, identical to rRoundRect with no pen.
*/
int roundRectRgn(SLRgn **rrgn,
	const rect2dq *rc,
	ifixed rx,
	ifixed ry);

/*
 xorRgn
	Merges brgn and rrgn, inverting intersections.
*/
int xorRgn(SLRgn **rrgn,
	const SLRgn *brgn);
/*
 unionRgn
	Merges rrgn with brgn, keeping intersections.
*/
int unionRgn(SLRgn **rrgn,
	const SLRgn *brgn);
/*
 sectRgn
	Masks rrgn with brgn, keeping only intersections.
*/
int sectRgn(SLRgn **rrgn,
	const SLRgn *brgn);
/*
 diffRgn
	Masks argn and brgn, removing intersections.
*/
int diffRgn(SLRgn **rrgn,
	const SLRgn *argn,
	const SLRgn *brgn);

/*
 rgnXorLine
	Inserts the integral x-coords of a line into a region.
	Automatically adjusts the size and (re)allocates.
  Result:
	1 on success; 0 if nothing added; negative on error.
*/
int rgnXorLine(SLRgn **rrgn,
	int16 cliptop,
	int16 clipbot,
	pt2dq p0,
	pt2dq p1);

/*
 rectInRgn
	Scans a region to see if the passed rectangle intersects.
  Result:
	1 if true; 0 if false; negative on error.
*/
int rectInRgn(const SLRgn*,
	int left,
	int top,
	int right,
	int bottom);
int ptInRgn(const SLRgn*,
	int x,
	int y);
/*
 rgnCompact
	Removes any excess in a region, adjusting the bBox height if necessary.
	If the resulting region becomes empty, NULL will be returned.
	In the case of error, the region will be collapsed, and size will be odd.
*/
SLRgn *rgnCompact(SLRgn*,
	boolean linesOnly); /* remove only excess hollow lines above/below */
/*
 offsetRgn
	Shifts a region, given an horizontal and vertical offset.
	The resulting region is clipped and compacted.
  Result:
	1 if true; 0 if false; negative on error.
*/
int offsetRgn(SLRgn **rrgn,
	int dx,
	int dy);
/*
 insetRgn
	Shrinks a region by secting with itself displaced by +/-dx and +/-dy.
	The resulting region is clipped and compacted.
  Result:
	1 if true; 0 if false; negative on error.
*/
int insetRgn(SLRgn **rrgn,
	int dx,
	int dy);

/*
 bitmapToRgn
	Generates a compact region from a monochrome (1bpp) bitmap.
  Result:
	The region; NULL on error.
*/
int bitmapToRgn(SLRgn **rrgn,
	const uint8 *bm,
	int sp,
	const rect16 *bounds);
/*
 QDRgnToSLRgn
	Converts a QuickDraw region to an mgraf scanline region.
*/
int QDRgnToSLRgn(SLRgn **rrgn,
	const uint8 *s,
	int32 z);

/*
 rRegion
	Given the current pen mode, paints the area within a region.
  NOTE:
	This routine always fills if the pen is not completely zero.
	Use rFrameRegion for framing.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rRegion(grafPort *p,
	const SLRgn *rgn);
int rFrameRegion(grafPort *p,
	const SLRgn *rgn);

/*
 rTriangle
	Given the current drawing mode, paints the area within three points.
	This routine ignores the pen width & height.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rTriangle(grafPort *p,
	pt2dq p0,
	pt2dq p1,
	pt2dq p2);

/*
 rPolygon
	Given the current pen mode, paints a polygon.
	If the penw is positive, a line frame is drawn.
	(A "single-pixel" frame is drawn if only the penh is negative.)
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rPolygon(grafPort *p,
	const pt2dq *ep,
	int count);
/*
 rLine
	Paints a rectangular pen between two points.
	If the pen is negative-spaced, a "single-pixel" line is drawn.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rLine(grafPort *p,
	pt2dq p0,
	pt2dq p1);
/*
 rLineTo
	Same as rLine, but uses pnLoc as p0, setting pnLoc to p1 after drawing.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rLineTo(grafPort *p,
	ifixed p1x,
	ifixed p1y);
/*
 rCLineList
	Paints rectangular pen between a connected list of points.
	If the pen is negative-spaced, "single-pixel" lines are drawn.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rCLineList(grafPort *p,
	const pt2dq *ep, /* can be NULL */
	int ptc);
/*
 rOval
	Paints the area within an oval, given rectangular bounds.
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rOval(grafPort *p,
	const rect2dq *ov);
/*
 rArc
	Same as rOval, but renders only an arc, given an angle and circumference.
	If 0 is passed for 'pie', the arc is instead a chord, omitting the center.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rArc(grafPort *p,
	const rect2dq *ov,
	ufract startAngle,
	ufract circumference,
	boolean pie);
/*
 rRect
	Paints the area within a rectangle.
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rRect(grafPort *p,
	const rect2dq *rc);
/*
 rRoundRect
	Paints the area within a rectangle with round edges (rx/ry).
	If the pen is valid, only a frame is drawn, midway inset/outset.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rRoundRect(grafPort *p,
	const rect2dq *rc,
	ifixed rx,
	ifixed ry);
/*
 rFillRect
	Paints an integral rectangle, ignoring the current mode & pen size.
	The current pen color/pattern is used in a solid transfer.
	This is faster than rRect for solid painting.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rFillRectMode(grafPort *p,
	int left,
	int top,
	int right,
	int bottom,
	uint8 mode);
#define rFillRect(p,l,t,r,b) rFillRectMode(p,l,t,r,b,mSrcCopy)
#define rEraseRect(p,l,t,r,b) rFillRectMode(p,l,t,r,b,mSrcNotCopy)
/*
 rInvertRect
	Performs a solid mSrcXor inversion, ignoring the fg/bkColor & pPat.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rInvertRect(grafPort *p,
	int left,
	int top,
	int right,
	int bottom);

/*
 rBltPix
	Transfers a pixMap to a port.
	Scaling is performed if any widths or heights are inequal.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rBltPix(grafPort *p,
	pixMap *srcpm,
	int srcx,
	int srcy,
	int srcw,
	int srch,
	int dstx,
	int dsty,
	int dstw,
	int dsth,
	uint16 mode);
/*
 rClippedBltPix
	Same as rBltPix, but accepts an additional clipping region.
	If 'clip' is NULL, this routine behaves like rBltPix.
	Keep that in mind should clip actually be rendered NULL from becoming empty.
	(If necessarily blitting a region, first test with `if (clip)`.)
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rClippedBltPix(grafPort *p,
	pixMap *sm,
	int sx,
	int sy,
	int sw,
	int sh,
	int dx,
	int dy,
	int dw,
	int dh,
	uint16 mode,
	const SLRgn *clip);
/*
 rScrollBlt
	Copies a rectangular portion of a port to another location.
	If an eraseRgn is passed, the void from the former is calculated.
	This routine calls rectRgn, so set NULL if no erase region existed.
*/
#ifdef _CRTNOP
_CRTNOP
#endif
int rScrollBlt(grafPort *p,
	int sx,
	int sy,
	int dx,
	int dy,
	int w,
	int h,
	SLRgn **eraseRgn); /* can be NULL */

/*
 decompressImageFile
	This reads and renders a supported image file.
	Support for QuickDraw PICTs and GIFs are built-in.
*/
pixMap *decompressImageFile(const char *path,
	int *err,
	uint16 *srcWidth,
	uint16 *srcHeight,
	uint8 *mode);

/* ---- basic conversion, copy */

void __fastcall b1to8(uint8 *d, /* must be valid */
	const uint8 *s, /* must be valid */
	int byteCount);
/* reverse direction version */
void __fastcall rb1to8(uint8 *d,
	const uint8 *s,
	int c);
void __fastcall b2to8(uint8 *d,
	const uint8 *s,
	int byteCount);
void __fastcall b4to8(uint8 *d,
	const uint8 *s,
	int byteCount);

/*
 rCopyBits/rNotCopyBits/rOrBits/rXorBits/rAndBits
	Emphasis on 'bits,' these routines copy a 1 bpp line, aligned or unaligned.
*/
void rCopyBits(int8 *dst, /* must be valid */
	const int8 *src, /* must be valid */
	int dstBitOffset,
	int srcBitOffset,
	int bitCount);
void rNotCopyBits(int8 *dst, /* must be valid */
	const int8 *src, /* must be valid */
	int dx,
	int sx,
	int c,
	uint32 scm); /* source complement mask */
void rOrBits(int8 *dst, /* must be valid */
	const int8 *src, /* must be valid */
	int dx,
	int sx,
	int c,
	uint32 scm);
void rXorBits(int8 *dst, /* must be valid */
	const int8 *src, /* must be valid */
	int dx,
	int sx,
	int c,
	uint32 scm);
void rAndBits(int8 *a, /* must be valid */
	const int8 *b, /* must be valid */
	int dx,
	int sx,
	int c,
	uint32 scm);

#ifdef __MGRAF_INTERNAL
 #define RGN_STOP 32767
const int16 *_rgnGotoLine(const int16 *s,
	const int16 *e,
	int o,
	int c);
 #define _rawColorFBFore 0
 #define _rawColorFBBack (~0)
void _rgbSetFBSolid(grafPort *p,
	const rgb48 *v,
	int k);
void _pmFormStruct(pixMap *m,
	int bpp,
	int width,
	int height);
#endif

/*
 RGBDepthFromFields
	Finds an ideal depth for the masks in kh->k in RGB order.
	If the values are not supported in pixMap, then conversion will be signaled.
  Result:
	The ideal or required depth; 64 will be added if conversion is necessary. 
*/
int RGBDepthFromFields(RGBLineInfo *kh);

/*
 BGRUnpackRow
	Copies or converts a BGR-formatted row given a byte size and line profile.
	For indexed modes, the row size is required.
	For direct modes, use results from RGBDepthFromFields with proper masks.
*/
int BGRUnpackRow(uint8 *d,
	const uint8 *s,
	int width,
	int pByteSize,
	const RGBLineInfo *kh);

#define __MGRAF_H
#endif
