/*
 mimg.c
  mgraf imaging

  Revised: 2024.04.12
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.07
  Back & forth, realized colors must be 48-bit;
otherwise, QD-loaded colors won't match in gray conversion.
  Implemented rgbToPenColor and a full color & gray makeITable.
2014.04.13
  Fixed gray inverse table building.
2014.04.21
  Added a simple fix to get the closest cyan for inverse tables.
2014.04.23
  Gray inverse tables are now 10-bit, meeting 30-bit precision standards.
This allows for gray tables that cover smaller 16-bit ranges.
2019.02.16
  pmFilter optimizations.
2019.03.21
  Preliminary implementation for extended pattern rConvertPat.
*/

#include <string.h>

#define HAVE_OS
#define __MGRAF_INTERNAL
#include "mgraf.h"
#include "mblt.h"
#include "ftrig.h"

/* ---- globals */

const uint32	pSizeMasks[32] =
{
	1,~0u,0x55555555,0x11111111,0x11111111,0x1010101,0x1010101,0x1010101,
	0x1010101,0x10001,0x10001,0x10001,0x10001,0x10001,0x10001,0x10001,
	0x10001,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1
};
const uint32	pMCmpMasks[16] =
{ /* reciprocal table: raised 16 bits */
	0x1010000,0xFFFF0000u,0x55550000,0x24922493,0x11110000,0x8420843,0x4103CF4,0x204060D,
	0x1010000,0x803FA0,0x400FC4,0x2003E1,0x1000F1,0x80039,0x4000D,0x20003
};

const uint8	pDepthX[32] =
	{5,0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5};
const uint32	pDepthK[32] =
{
	~0u,1,3,15,15,255,255,255,255,65535,65535,65535,65535,65535,65535,65535,65535,
	~0u,~0u,~0u,~0u,~0u,~0u,~0u,~0u,~0u,~0u,~0u,~0u,~0u,~0u,~0u
};
const uint8	pMinCmpX[32] =
	{8,3,3,3,4,5,5,5,5,3,3,3,4,4,4,5,5,5,6,6,6,7,7,7,8,8,8,9,9,9,10,10};

const uint8	Mac1bppRGB12[2*2] =
	{0xF,0xFF};

/* ---- code */

static void cnvLine(void *d,
	int c,
	int pz,
	cnvParam *cnv)
{
	int32		*a = d;
	int		o = 0;

	cnv->hx =~0;
	if (pz>16)
	{
		for(; (c-=2)>=0; a+=2)
		{
			*a = cnv->rtn(cnv,o), ++o;
			a[1] = cnv->rtn(cnv,o), ++o;
		}
		if (c&1)
			*a = cnv->rtn(cnv,o);
		o = 0;
	} else if (pz>8)
	{
		int32		v;

		for(;(c-=2)>=0; ++a)
		{
			v = cnv->rtn(cnv,o)<<BIG_ENDIAN*16, ++o;
			v|= cnv->rtn(cnv,o)<<!BIG_ENDIAN*16, ++o;
			*a = v;
		}
		if (c&1)
			*(int16*)a = (int16)cnv->rtn(cnv,o);
		c = 0;
	} else
	{
		int32		v = 0;
		int		ib = 0;

		while(o<c)
		{
			v<<= pz, v|= cnv->rtn(cnv,o), ++o, ib+= pz;
			if (ib>31)
				*a = be32_native(v), ++a, ib-= 32;
		}
		if (ib) /* sizes here aligned to 32 bits */
			v<<= 32-ib, *a = be32_native(v);
	}
}

int getMainScreenDepth(void);

#if TARGET_OS_MAC
 /* TODO! */
#elif defined(WINUSERAPI)
 #define MAX_DIB_SIZE		(sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD))

int getMainScreenDepth(void)
{
	int8		b[MAX_DIB_SIZE];
	HDC		hdc;
	HBITMAP		hbm;
	BITMAPINFO	*info;
	int		bpp;

	info = memset(b,0,sizeof(b));
	info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);

	hdc = GetDC(NULL);
	hbm = CreateCompatibleBitmap(hdc,1,1); /* 0,0 would produce a mono dib */

	/* first call fills the header (because it's empty) */
	/* the second polls the palette */
	(void)GetDIBits(hdc,hbm,0,1,NULL,info,DIB_RGB_COLORS);
	(void)GetDIBits(hdc,hbm,0,1,NULL,info,DIB_RGB_COLORS);

	(void)DeleteObject(hbm);
	bpp = GetDeviceCaps(hdc,BITSPIXEL);
	(void)ReleaseDC(NULL,hdc);

	if (info->bmiHeader.biBitCount==16
			&& info->bmiHeader.biCompression==BI_BITFIELDS
			&& *(int32*)&info->bmiColors==0x7C00)
		return 15;
	return info->bmiHeader.biBitCount ? info->bmiHeader.biBitCount
		:bpp; /* last resort */
}

static void setQuadGrays(RGBQUAD *d,
	int c)
{
	int		x, dx;

	if (c<=0)
		return;
	dx = (255u<<16)/(c-1), x =~0;
	do
		d->rgbRed = d->rgbGreen = d->rgbBlue = x>>16;
	while(x-=dx,++d,--c);
}

BITMAPINFO *pixMapToNativeDesc(pixMap *m)
{
	BITMAPINFO	*info;
	uint32		pitch;

	if (!m || !m->pSize || (info=calloc(sizeof(BITMAPINFO)
			+(m->pSize>8 ? 3:256)*sizeof(RGBQUAD),1))==NULL)
		return NULL;
	pitch = m->pitch&0x7FFFFFFF;
	info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	info->bmiHeader.biWidth = (pitch<<3)/m->pSize;
	info->bmiHeader.biHeight =-m->h; /* negative for top-down bitmap */
	info->bmiHeader.biPlanes = 1;
	info->bmiHeader.biSizeImage = m->h*pitch;
	info->bmiHeader.biBitCount = m->pSize;
	if (m->pSize<9)
	{
		if (m->pTable)
			setQuadColors(info->bmiColors,m->pTable->colors,m->pTable->count);
		else
			setQuadGrays(info->bmiColors,1u<<m->bpp);
		return info;
	}
	/* can really only fix endian mismatch here for 24/32-bit */
	if (!((m->pType^pTypeNativeEndian)&(pTypeLSB|pTypeLGreen)))
		if ((m->pSize==16 && m->minCmp==5)
				|| (m->pSize==32 && m->minCmp==8))
			return info;
	info->bmiHeader.biCompression = BI_BITFIELDS;
	*(int32*)&info->bmiColors = RGBToPen(m,chromo_colors+_rgb_red);
	*(int32*)&info->bmiColors[1] = RGBToPen(m,chromo_colors+_rgb_green);
	((int32*)&info->bmiColors)[2] = RGBToPen(m,chromo_colors+_rgb_blue);
	return info;
}
#endif

/* pixmap */

void _pmFormStruct(pixMap *m,
	int bpp,
	int width,
	int height)
{
#ifdef WINUSERAPI
	HDC		hdc;
#endif
	unsigned	mincmp;
#ifdef WINUSERAPI
	hdc = GetDC(NULL);
	if (!m->hRes)
		m->hRes = GetDeviceCaps(hdc,LOGPIXELSX)<<16;
	if (!m->vRes)
		m->vRes = GetDeviceCaps(hdc,LOGPIXELSY)<<16;
	if (width<0)
		width = GetDeviceCaps(hdc,HORZRES);
	if (height<0)
		height = GetDeviceCaps(hdc,VERTRES);
	(void)ReleaseDC(NULL,hdc);
	if (bpp<=0)
		bpp = getMainScreenDepth();
#endif
	if (width>65535)
		width = 65535;
	if (height>65535)
		height = 65535;
	if (!m->vRes)
		m->vRes = hvResDefault;
	if (!m->hRes)
		m->hRes = hvResDefault;
	m->h = (uint16)height, m->w = (uint16)width;
	m->bpp = mincmp = (uint8)bpp;
	if (mincmp<9)
	{
		m->pSize = mincmp>4 ? 8:(mincmp|(bpp>>newDepForce4Bit&1))>2 ? 4:mincmp;
		m->pType = pTypeIndexed;
	} else
	{
		static const uint8	hiHintpType[8] =
		{
			pTypeRawRGB|pTypeNativeEndian,
			pTypeRawRGB|pTypeLGreen|pTypeNativeEndian,
			pTypeRawRGB,
			pTypeRawRGB|pTypeLGreen,
			pTypeRawRGB|pTypeLSB,
			pTypeRawRGB|pTypeLGreen|pTypeLSB,
			pTypeRawRGB|pTypeNativeEndian, /* both ==> default */
			pTypeRawRGB|pTypeLGreen|pTypeNativeEndian
		};
		uint8		typ;

		if (mincmp>=32)
		{
			m->pSize = 32;
			mincmp = 8;
			typ = pTypeAlpha|pTypeRawRGB|pTypeNativeEndian;
		} else
		{
			m->pSize = mincmp>16 ? 32:16;
			mincmp = pMinCmpX[mincmp];
			typ = hiHintpType[bpp>>newDepthLGreenBit&7];
			if (typ&pTypeLGreen && !(bpp&1) && m->bpp<m->pSize)
				++m->bpp;
		}
		m->pType|= typ;
	}
	m->minCmp = mincmp;
	if (width>0)
		m->pitch = ((width*m->pSize>>3)+3)&~3;
}

pixMap *newPixMap(int bpp,
	int width,
	int height)
{
	pixMap		*m;

	if ((m=calloc(sizeof(*m),1))==NULL)
		return NULL;
	_pmFormStruct(m,bpp,width,height);
	return m;
}

int disposePixMap(pixMap *m)
{
	int		r;

	if (!m)
		return -1;
	r = 0;
	if (m->iTable)
		{free(m->iTable); m->iTable = NULL, r+= 4;}
	if (m->pTable)
		{free(m->pTable); m->pTable = NULL, r+= 2;}
	r+= pmFree(m);
	free(m);
	return r;
}

int pmAllocate(pixMap *m)
{
	if (!m)
		return -1;
	if (!m->baseAddr && !m->pHand && m->h>0 && m->pitch>0)
		if ((m->pHand=hNew(m->pitch*m->h,0))!=NULL)
			return 1;
	return 0;
}

int pmFree(pixMap *m)
{
	if (!m || m->pitch<=0)
		return -1;
	if (m->pHand)
		{(void)hFree(m->pHand); m->pHand = NULL; return 1;}
	m->baseAddr = NULL;
	return 0;
}

int pmCopy(const pixMap *s,
	pixMap **d)
{
	pixMap		*m;
	int		r;
	boolean		srcLocked;

	if (!d)
		return -1;
	m = *d;
	if (!s)
	{
		pmFree(m);
		return 0;
	}
	if (m)
		disposePixMap(m);
	if ((m=malloc(sizeof(pixMap)))==NULL)
		return -2;
	*m = *s;
	m->pTable = NULL;
	m->iTable = NULL;
	if ((r=copyCTable(&m->pTable,s->pTable))==-2)
		goto merr;
	m->baseAddr = NULL;
	m->pHand = NULL;
	srcLocked = s->baseAddr!=NULL;
	if (s->pHand || srcLocked)
	{
		if ((r=pmAllocate(m))<=0)
		{
		merr:	if (!*d)
				disposePixMap(m);
			return r;
		}
		memcpy(lockPixels(m),
			srcLocked ? s->baseAddr:hLock(s->pHand),
			m->pitch*m->h);
		unlockPixels(m);
		if (!srcLocked)
			hUnlock(s->pHand);
	}
	*d = m;
	return 1;
}

int8 *lockPixels(pixMap *m)
{
	void		*a;

	if (!m || m->pitch<=0)
		return NULL; /* invalid parameter or read-only */
	if (m->baseAddr || !m->pHand)
		return m->baseAddr;
	if ((a=hLock(m->pHand))==NULL)
		return NULL;
	return m->baseAddr=a;
}

int unlockPixels(pixMap *m)
{
	void		*a;

	if (!m)
		return -1; /* invalid parameter */
	if ((a=m->pHand)==NULL)
		return 0;
	m->baseAddr = NULL;
	(void)hUnlock(a);
	return 1;
}

/* ---- image filtering */

int pmErase(pixMap *m,
	uint32 penc)
{
	uint32		*d;
	int32		z, c;

	if (!m || m->pitch<=0)
		return -1;
	if ((d=(uint32*)m->baseAddr)==NULL || (z=m->h)<=0 || (z*=m->pitch)<=0)
		return 0;
	penc&= ((2u<<(m->bpp-1))-1)*pSizeMasks[m->pSize&31];
	if (m->pSize<2)
		{memset(d,penc,z); return 1;}
	if ((c=z>>6)!=0)
		do
			d[15] = d[14] = d[13] = d[12] =
				d[11] = d[10] = d[9] = d[8] =
				d[7] = d[6] = d[5] = d[4] =
				d[3] = d[2] = d[1] = *d = penc;
		while(d+=16,--c);
	if ((c=z>>3&7)!=0)
		do
			d[1] = *d = penc;
		while(d+=2,--c);
	if (z&4)
		*d++ = penc;
	if (z&2)
		*(int16*)d = (int16)penc;
	return 1;
}

int pmFilter(pixMap *m,
	const cTab64 *ct,
	const rect16 *bounds)
{
	uint8		st[1024];
	void		*ba;
	const cSpec	*cs;
	uint32		kb, kg, kr, lf;
	int		c, slop, w, h, i, j, k;

	if (!m || m->pitch<=0)
		return -1;
	h = m->h, w = m->w;
	if (!ct || (ba=m->baseAddr)==NULL || (slop=m->pitch)<w)
		return 0;
	if (bounds)
	{
		c = MAX(bounds->top,0);
		lf = MAX(bounds->left,0);
		if (bounds->bottom+0u<h+0u)
			h = bounds->bottom;
		if (bounds->right+0u<w+0u)
			w = bounds->right;
		w-= lf, h-= c, ba = slop*c+(int8*)m->baseAddr;
	} else
		lf = 0;
	if (h<=0 || w<=0)
		return 0;
	if (m->pSize<9)
	{
		if (m->pSize<8)
			return -8;
		if (m->pSize<9 && !m->iTable && pmMakeITables(m,m->bpp,FALSE)<=0)
			return -2;
	}
	i = m->minCmp ? m->minCmp:10, j = i+(m->pType>>pTypeLGreenBit&1);
	kb = (1u<<i)-1, kg = (1u<<j)-1;
	if (ct->count+0u<=kg)
		return -1;
	kg<<= i, k = j+i, kr = kb<<k;
	/* the table builder ends up taking all the time when the area is small */
	if (m->pSize<9)
	{
		const uint8	*it = m->iTable;
		uint16		*t;
		uint8		*d;

		if ((t=(uint16*)rGenRemappingTable(m,ct,m->bpp,ccIndexed))==NULL)
			return -2;
		slop-= w;
		if (m->minCmp)
		{
			const uint16	*dt = (const uint16*)(it+(1u<<m->minCmp*3));

			c = 256, d = st;
			do
			{
				uint32		r, a = *dt;

				r = t[a&kb]&kb;
				r|= t[(a&kg)>>i]&kg;
				r|= t[(a&kr)>>k]&kr;
				*d = (uint8)it[r];
			} while(++d,++dt,--c);
		} else
		{
			const uint16	*dt = (const uint16*)(it+1024);

			c = 256, d = st;
			do
				*d = (uint8)it[t[*dt>>6]>>6];
			while(++d,++dt,--c);
		}
		free(t); t = NULL;
		d = lf+(uint8*)ba;
		do
		{
			if ((c=w)&1)
				*d = st[*d], ++d;
			if (c>>=1)
				do
					*d = st[*d], d[1] = st[d[1]];
				while(d+=2,--c);
		} while(d+=slop,--h);
		return 1;
	}
	cs = ct->colors, c = PIN(ct->count,0,256);
	memset(st,0,sizeof(st));
	if (m->pSize<17)
	{
		uint16		*t = (uint16*)st, *d = t;

		if (ct->flags&0x8000)
			do
				*d++ = (uint16)formDirect(&cs->rgb,i,j);
			while(++cs,--c);
		else
			do
				d[cs->value] = (uint16)formDirect(&cs->rgb,i,j);
			while(++cs,--c);
		d = lf+(uint16*)ba;
		slop-= w*2;
		do
		{
			c = w;
			do
			{
				uint32		a = *d;
				uint16		r;

				r = t[a&kb]&(uint16)kb;
				r|= t[(a&kg)>>i]&(uint16)kg;
				r|= t[(a&kr)>>k]&(uint16)kr;
				*d = r;
			} while(++d,--c);
		} while(d=(uint16*)((int8*)d+slop),--h);
	} else
	{
		uint32		*t = (uint32*)st, *d = t;

		if (ct->flags&0x8000)
			do
				*d++ = formDirect(&cs->rgb,i,j);
			while(++cs,--c);
		else
			do
				d[cs->value] = formDirect(&cs->rgb,i,j);
			while(++cs,--c);
		d = lf+(uint32*)ba;
		slop-= w*4;
		if (kb==0xFF && k==16)
			do
			{
				c = w;
				do
				{
					uint32		r, a = *d;

					r = st[(uint8)a*4+BIG_ENDIAN*3];
					r|= t[a>>8&255]&0xFF00;
					r|= t[a>>16&255]&0xFF0000;
					*d = r;
				} while(++d,--c);
			} while(d=(uint32*)((int8*)d+slop),--h);
		else
			do
			{
				c = w;
				do
				{
					uint32		r, a = *d;

					r = t[a&kb]&kb;
					r|= t[(a&kg)>>i]&kg;
					r|= t[(a&kr)>>k]&kr;
					*d = r;
				} while(++d,--c);
			} while(d=(uint32*)((int8*)d+slop),--h);
	}
	return 1;
}

/* ---- graphics ports */

int openGrafPort(grafPort *p,
	int bpp,
	int width,
	int height,
	boolean allocate)
{
	if (!p)
		return -1;
	memset(p,0,sizeof(*p));
	if ((p->map=newPixMap(bpp,width,height))==NULL)
		return 0;
	p->clipRgn = newRgn(1);
	p->clipSave = newRgn(1);
	p->visRgn = newRgn(1);
	p->pnSize.x = p->pnSize.y = 0x10000;
	p->frSize.x = p->frSize.y =~0;
	/* avoid unnecessary calls to makeITables */
	//RGBFore(p,&rgb_black);
	//RGBBack(p,&rgb_white);
	p->fgColor = rgb_black;
	p->fgSolid =-!pixMapIsDirect(p->map);
	p->bkColor = rgb_white;
	p->bkSolid =~p->fgSolid;
	if (allocate)
		return pmAllocate(p->map);
	return 1;
}

int closeGrafPort(grafPort *p)
{
	if (!p)
		return -1;
	if (p->map)
		{(void)disposePixMap(p->map); p->map = NULL;}
	rPenPat(p,NULL);
	disposeRgn(&p->clipRgn);
	disposeRgn(&p->clipSave);
	disposeRgn(&p->visRgn);
	p->pnSize.y = p->pnSize.x = MK4CC('D','E','A','D');
	return 1;
}

int rSetClip(grafPort *p,
	SLRgn *srgn,
	boolean copy)
{
	if (copy)
		copyRgn(&p->clipSave,srgn);
	else
	{
		disposeRgn(&p->clipSave);
		p->clipSave = srgn;
	}
	copyRgn(&p->clipRgn,p->visRgn);
	return sectRgn(&p->clipRgn,p->clipSave);
}

int rAddClip(grafPort *p,
	const SLRgn *srgn)
{
	sectRgn(&p->clipSave,srgn);
	copyRgn(&p->clipRgn,p->visRgn);
	return sectRgn(&p->clipRgn,p->clipSave);
}

/* table generation */

int pmDefaultMacTable(pixMap *m,
	int gray)
{
	const uint8	*t;
	int		bpp, count;

	if (!m)
		return -1;
	if (m->pType || (bpp=m->bpp)>8)
		return 0;
	if (m->pTable)
		{free(m->pTable); m->pTable = NULL;}
	if (m->iTable)
		{free(m->iTable); m->iTable = NULL;}
	if (gray)
		bpp|= 32;
	if (bpp==40 || bpp==36 || bpp==34 || bpp==33)
		t = NULL;
	else if (bpp==8)
		t = Mac8bppRGB12;
	else if (bpp==4)
		t = Mac4bppRGB24, bpp+= 16;
	else if ((bpp&15)==2)
		t = Mac2bppRGB24, bpp+= 16;
	else if ((bpp&15)==1)
		t = Mac1bppRGB12;
	else
		return 0;
	count = 1u<<(bpp&15);
	if ((m->pTable=malloc(8+8*count))==NULL)
		return -2;
	if (!t)
		ctMakeGrays(m->pTable->colors,gray<0 ? -count:count);
	else if (bpp&16)
		expandRGB24(t,m->pTable->colors,count,0);
	else
		expandRGB12(t,m->pTable->colors,count,0);
	m->pTable->seed = GetCTSeed();
	m->pTable->flags = 0x8000;
	m->pTable->count = (uint16)count;
	return 1;
}

int pmMakeITables(pixMap *m,
	int r,
	boolean altAccuracy)
{
	if (!m || m->bpp>8)
		return -1;
	m->pType&=~pTypeLGreen;
	if ((r=ctMakeITables(m->pTable,&m->iTable,m->bpp,r,0,altAccuracy))<0)
		return r;
	m->minCmp = (uint8)r;
	return 1;
}

int copyCTable(cTab64 **out,
	const cTab64 *in)
{
	void		*p;
	int32		z;

	if (!out)
		return -1;
	if (!in)
	{
		if (*out)
			{free(*out); *out = NULL;}
		return 0;
	}
	z = sizeof(cTab64)-sizeof(cSpec)+sizeof(cSpec)*PIN(in->count,0,256);
	if ((p=realloc(*out,z))==NULL)
	{
		if (!z)
			{*out = NULL; return 0;}
		return -2;
	}
	*out = p;
	memcpy(p,in,z);
	return 1;
}

cTab64 *allocCTable(const rgb30 *s,
	int c)
{
	cTab64		*t;
	cSpec		*d;

	c = PIN(c,0,256);
	t = calloc(sizeof(cSpec),c+1);
	t->flags = 0x8000;
	t->count = c;
	if (!s || c<=0)
		return t;
	d = t->colors;
	do
	{
		uint32		v = *s++;

		d->flags = v>>25&0x60;
		d->rgb.red = ((v>>20&1023)*1025)>>(20-16);
		d->rgb.green = ((v>>10&1023)*1025)>>(20-16);
		d->rgb.blue = ((v&1023)*1025)>>(20-16);
	} while(++d,--c);
	return t;
}

/* color selection */

int pmGetRGBCC(pixMap *m,
	void **gt)
{
	if (!m)
		return -1;
	if (pixMapIsDirect(m))
	{
		*gt = (void*)(m->minCmp+(m->pType>>pTypeLGreenBit&1));
		return (m->pType^pTypeNativeEndian)&pTypeLSB
			? ccDirectBS32+(m->pSize>>4&1):ccDirect;
	} else if ((*gt=m->iTable)==0)
	{
		if (pmMakeITables(m,m->bpp,FALSE)<=0)
			return -2;
		*gt = m->iTable;
	}
	return ccIndexed|ccNative|(!!m->minCmp);
}

uint32 RGBToPen(pixMap *m,
	const rgb48 *s)
{
	void		*gt;
	int		x;

	if ((x=pmGetRGBCC(m,&gt))<0)
		return 0;
	return rgbCCRoutine[x](m->minCmp,gt,s)*pSizeMasks[m->pSize&31];
}

int PenToRGB(pixMap *m,
	uint32 penc,
	rgb48 *v)
{
	const cTab64	*t;
	size_t		j;
	uint32		i, k;

	if (!m)
		return -1;
	i = m->minCmp&15;
	if (pixMapIsDirect(m))
	{
		k = (1u<<i)-1, j = i+(m->pType>>pTypeLGreenBit&1);
		v->blue = (uint16)((penc&k)*pMCmpMasks[i]>>16);
		penc>>= i;
		v->green = (uint16)((penc&(1u<<j)-1)*pMCmpMasks[j]>>16);
		v->red = (uint16)((penc>>j&k)*pMCmpMasks[i]>>16);
	} else if ((t=m->pTable)==NULL)
	{
		if (pmMakeITables(m,m->bpp,FALSE)<=0)
			return -2;
		v->blue = v->green = v->red = ((uint16*)m->iTable)[(penc&(1u<<m->bpp)-1)+512];
	} else
	{
		i = penc&(1u<<m->bpp)-1;
		if (t->flags&0x8000)
		{
			if (i>=t->count+0u)
				return 0; /* index out of range! */
		} else
		{
			for(j=0; j<t->count+0u; ++j)
				if (t->colors[j].value==i)
					goto ixc;
			return 0; /* not in the table */
		}
	ixc:	*v = t->colors[i].rgb;
	}
	return 1;
}

int GetPixelRGB(pixMap *m,
	int x,
	int y,
	rgb48 *v)
{
	const void	*s;
	uint32		u;
	boolean		hl;

	if (!m)
		return -1;
	if ((x-=m->pOrigin.x)+0u>=m->w+0u || (y-=m->pOrigin.y)+0u>=m->h+0u)
		return 0;
	hl = 0;
	if ((s=m->baseAddr)==NULL)
		if ((s=lockPixels(m))==NULL)
			return -1;
		else
			hl = 1;
	s = m->baseAddr+y*m->pitch;
	if (m->pSize==32)
	{
		u = ((const uint32*)s)[x];
		if ((m->pType^pTypeNativeEndian)&pTypeLSB)
			u = _reverse32(u);
	} else if (m->pSize==16)
	{
		u = ((const uint16*)s)[x];
		if ((m->pType^pTypeNativeEndian)&pTypeLSB)
			u = _reverse16(u);
	} else
		u = pDepthX[m->pSize], u = ((const uint8*)s)[x<<u>>3]>>(~x<<u&7);
	if (hl)
		(void)unlockPixels(m);
	return PenToRGB(m,u,v);
}

void _rgbSetFBSolid(grafPort *p,
	const rgb48 *v,
	int k)
{
	pixPat		*pat;
	pixMap		*m;
	uint32		c, col;

	if (!p || (m=p->map)==NULL || !v)
		return;
	(&p->fgColor)[k&1] = *v;
	(&p->fgSolid)[k&1] = col = RGBToPen(m,v);
	if ((pat=p->pPat)==NULL)
		return;
	if ((c=pat->ct.count)==0)
		return;
	if ((int16)pat->ct.flags>(int16)0x80FF)
		return; /* colorizing or no table to apply */
	pat->ct.seed = 0; /* only need to invalidate one color...but I'm tired */
	k = (c-1)&~k;
	pat->ct.colors[k].value = col&pDepthK[m->pSize];
	pat->ct.colors[k].rgb = *v; /* the decrement was missing!; 2019.03.26 */
}

void RGBFore(grafPort *p,
	const rgb48 *v)
{
	_rgbSetFBSolid(p,v,_rawColorFBFore);
}

void RGBBack(grafPort *p,
	const rgb48 *v)
{
	_rgbSetFBSolid(p,v,~0);
}

/* used to be in mtext */
int PenGetGray(grafPort *p,
	uint32 *solid)
{ /* get the average for the fore and back */
  /* if we can't get a close enough color match, gray the text (in an opaque way) */
  /* note that penToRGB can still fail (out-of-range index) */
	rgb48		back, grayc, graym;
	pixMap		*m;

	if ((m=p->map)==NULL)
		return -1;
	if (PenToRGB(m,p->fgSolid,&grayc) && PenToRGB(m,p->bkSolid,&back))
	{
		uint32		penc;
		/* another POCC bug: stuck in 16-bit ops. simply doing `(a+b+1)>>1` */
		penc = (back.red+grayc.red+1)>>1;
		grayc.red = (uint16)penc;
		penc = (back.green+grayc.green+1)>>1;
		grayc.green = (uint16)penc;
		penc = (back.blue+grayc.blue+1)>>1;
		grayc.blue = (uint16)penc;
		penc = RGBToPen(m,&grayc);
		if (PenToRGB(m,penc,&graym)
				&& idim(grayc.red,graym.red)*5
				+idim(grayc.green,graym.green)*6
				+idim(grayc.blue,graym.blue)*4<0x28000)
		{ /* close enough */
			*solid = penc;
			return 1;
		}
	}
	return 0;
}

/* ---- patterns */

int rMakePat(pixPat **outpat,
	const uint8 inpat[8],
	const rgb48 *fore,
	const rgb48 *back)
{
	pixPat		*pat;

	if (!outpat)
		return -1;
	if ((pat=*outpat)!=NULL)
	{
		free(pat->px); pat->px = NULL;
		free(pat->nx); pat->nx = NULL;
	}
	if (!inpat)
	{
		free(pat);
		*outpat = NULL;
		return 0;
	}
	if ((pat=realloc(pat,sizeof(pixPat)+sizeof(cSpec)))==NULL)
		return -2;
	*outpat = pat;
	pat->nx = NULL, pat->h = pat->w = pat->pitch = 8;
	pat->ct.seed = pat->xpitch = 0;
	pat->ct.count = 2, pat->ct.flags = 0x8000;
	if ((pat->px=malloc(64))==NULL)
		return -2;
#ifdef TEST_BITMAPTORGN
	{
		rect16		bounds;
		SLRgn		*rgn;

		*(int32*)&bounds.top = 0x10001;
		*(int32*)&bounds.bottom = 0x70007;
		rgn = NULL;
		bitmapToRgn(&rgn,inpat,1,&bounds);
		free(rgn);
	}
#endif
	b1to8(pat->px,inpat,8);
	if (!fore)
		fore = &rgb_black;
	if (!back)
		back = &rgb_white;
	*(int16*)&pat->ct.colors = 0;
	pat->ct.colors->rgb = *back;
	*(int16*)&pat->ct.colors[1] = be16_native(1);
	pat->ct.colors[1].rgb = *fore;
	return 1;
}

void rPenPat(grafPort *p,
	const uint8 inpat[8])
{
	(void)rMakePat(&p->pPat,inpat,&p->fgColor,&p->bkColor);
}

int rCopyPat(pixPat **out,
	const pixPat *in)
{
	pixPat		*d;
	int		z;

	if (!in)
		return 0;
	z = sizeof(pixPat)-sizeof(cSpec)+sizeof(cSpec)*in->ct.count;
	if ((d=malloc(z))==NULL)
		return -2;
	*out = d;
	z-= OFFSET(pixPat,pitch);
	/* really awful bug where source wasn't cast to byte offset pointer; 2019.03.04 */
	memcpy((int8*)d+OFFSET(pixPat,pitch),(int8*)in+OFFSET(pixPat,pitch),z);
	d->nx = NULL; /* redundant data */
	d->xpitch = 0;
	z = in->h*in->pitch;
	if ((d->px=malloc(z))==NULL)
		return -2;
	memcpy(d->px,in->px,z);
	return 1;
}

int rConvertPat(pixPat *pat,
	pixMap *dm,
	uint16 flag)
{
 /*
     We need to move beyond the remapped method, not just for performance
   but the fact that full direct depths can't fit in indexed pixels.
   In needed cases, the nx field will be used to store a native pattern.
   'xpitch' will store the native pattern pitch; if non-zero, rInitBlt
   will initialize the bltParam to use the extended pattern in drawing.
     Also, it would be good to switch to native pattern anyway on the case
   of frequent use or by request by the user.
 */
	void		*p;
	int		t, pz, seed, fmt = formPxFmt(dm);
	boolean		colorizing;

	if (fmt<0x900 && dm->pTable)
		seed = dm->pTable->seed&IMAX(seed); /* bit is #31 never on */
	else
		seed = fmt|1u<<31;
	flag|= 0x8000;
	if (pat->ct.seed==seed && (uint8)pat->ct.flags==(uint8)flag)
		return 0;
	if (!pat->h)
		return -1;
	colorizing = 1;
	pat->xpitch = 0;
	free(pat->nx); pat->nx = NULL;
	/* does the pattern have an original depth? */
	switch(t=pat->ct.flags>>8)
	{
	case pPatRGBDither:
	{
		rgb48		color;

		if (PenToRGB(dm,RGBToPen(dm,&pat->ct.colors[1].rgb),&color)>=0)
		{
			t = pat->ct.colors[1].rgb.red*2-color.red;
			pat->ct.colors[0].rgb.red = t+0u>0xFFFF ? (int)~t>>31&0xFFFF:t;
			t = pat->ct.colors[1].rgb.green*2-color.green;
			pat->ct.colors[0].rgb.green = t+0u>0xFFFF ? (int)~t>>31&0xFFFF:t;
			t = pat->ct.colors[1].rgb.blue*2-color.blue;
			pat->ct.colors[0].rgb.blue = t+0u>0xFFFF ? (int)~t>>31&0xFFFF:t;
		}
		pz = t = 1;
		goto sk;
	}
	case 0x80:
		colorizing = 0;
		flag|= 0x8000;
		goto sf;
	case 0x81:
		/* seriously need to implement faster same-depth pattern blitter */
		/* 1-bit variation at least to lower memory costs */
		t = pat->pitch<<3;
		if ((p=realloc(pat->px,t*pat->h))==NULL)
			return -2;
		rb1to8(p,p,pat->pitch*pat->h);
		goto s8nc;
	case 0x82:
		t = pat->pitch<<2;
		if ((p=realloc(pat->px,t*pat->h))==NULL)
			return -2;
		b2to8(p,p,pat->pitch*pat->h);
		goto s8nc;
	case 0x83: case 0x84:
		t = pat->pitch<<1;
		if ((p=realloc(pat->px,t*pat->h))==NULL)
			return -2;
		b4to8(p,p,pat->pitch*pat->h);
	s8nc:	pat->px = p;
		pat->pitch = t;
	case 0x85: case 0x86: case 0x87:
	case 0x88:
		flag|= 0x8800; /* non-colorizing 8 bit */
	sf:	pat->ct.flags = flag;
	//	if ((pat->nx=rGenRemappingTable(dm,&pat->ct,8,flag|(fmt>>6&remap32bit)))==NULL)
	//		return -2;
	//	break;
		pz = 8, t = pat->ct.flags>>8&0x1F; /* currently 8-bit active px */
		goto sk;

	case 0x7F: case 0x7E: case 0x7C: case 0x7B: case 0x79:
	case 0x78: case 0x76: case 0x75: case 0x73: case 0x72:
	case 0x70: case 0x6F: case 0x6D: case 0x6C: case 0x6A: case 0x69:
	case 0x68: case 0x67: case 0x66: case 0x65:
	case 0x64: case 0x63: case 0x62:
	default:
	{
		cnvParam	cnv;
		int8		*d;
		int		c;
		/* form format for flags depth */
		if ((pz=t&31)==0)
			return -1; /* can't have ctab and direct pixels (at least not here) */
		if (flag&remapReverse)
			flag^= remapReverse^remapDInvert;
	sk:	t&=~0x1F, t|= pMinCmpX[pz]|256u<<pDepthX[pz];
		if ((uint16)t==(uint16)fmt && pz>8)
			break;
		pz = dm->pSize;
		if (dm->pSize<=8)
		{
			pz = 8; /* remove when patterns use bit rotation instead of 8bpp */
			if (flag&ccIndexed)
				pz = 16;
		}
		pat->xpitch = pat->w<<pDepthX[pz&31]>>3;
		if ((pat->nx=malloc(pat->xpitch*pat->h))==NULL)
			return -2;
		cnv.s = pat->px;
		if (!colorizing && dm->bpp<9 && pz==8)
		{ /* preserve indicies in indexed mode (this is also faster) */
			uint8		t[2];

			c = pat->pitch*pat->h;
			memcpy(pat->nx,cnv.s,c);
			t[0] = pat->ct.colors[0].value;
			t[1] = pat->ct.colors[1].value;
			memrem(pat->nx,c,t);
		} else
		{
			if ((t=rInitCnv(&cnv,dm,&pat->ct,t,0x10000,16,remap32bit|flag))<0)
			{
				free(pat->nx); pat->nx = NULL;
				return t;
			}
			c = pat->h, d = pat->nx;
			do
				cnvLine(d,pat->w,pz,&cnv);
			while(d+=pat->xpitch,cnv.s=(const int8*)cnv.s+pat->pitch,--c);
			rEndCnv(&cnv);
		}
	}
	}
	pat->ct.seed = seed;
	return 1;
}

int rAllocPat(pixPat **outpat,
	uint8 depth,
	uint16 w,
	uint16 h,
	uint16 colorCount)
{
	pixPat		*pat;
	size_t		z;
	int32		pitch;

	if (!outpat)
		return -1;
	pitch = w;
	if (depth)
	{
		if (depth>31)
			depth = 24|pTypeAlpha;
		pitch = ((pitch<<pDepthX[depth&31])+7)>>3;
	}
	if (!colorCount && depth<9)
		colorCount = pDepthK[depth ? depth:8]+1; //1<<(1<<pDepthX[depth ? depth:8]);
	if ((pat=*outpat)!=NULL)
	{
		free(pat->px); pat->px = NULL;
		free(pat->nx); pat->nx = NULL;
	}
	z = sizeof(cSpec)*colorCount;
	if ((pat=realloc(pat,sizeof(pixPat)-sizeof(cSpec)+z))==NULL)
		return -2;
	*outpat = pat;
	pat->nx = NULL;
	pat->h = h;
	pat->w = w;
	pat->pitch = pitch;
	pat->ct.seed = pat->xpitch = 0;
	pat->ct.count = colorCount;
	pat->ct.flags = 0x8000|depth<<8;
	memset(pat->ct.colors,0,z);
	if ((z=pitch*h)==0)
		z = 4;
	if ((pat->px=malloc(z))==NULL)
		return -2;
	memset(pat->px,0,z);
	return 1;
}

int rMakeRGBPat(pixPat **outpat,
	const uint8 inpat[8],
	const rgb48 *color)
{
	pixPat		*pat;
	int		r;

	if ((r=rAllocPat(outpat,1,8,8,2))<=0)
		return r;
	pat = *outpat;
	pat->ct.flags = pPatRGBDither<<8;
	*(int64*)pat->px = *(int64*)inpat; /* unaligned copy */
	pat->ct.colors[1].value = 1;
	pat->ct.colors[1].rgb = pat->ct.colors[0].rgb = *color;
	return 1;
}

/* ---- color conversion */

DEF_RGB_RTN(RGBGray)
{
	uint32		u;
(void)gt;
	return (formGray(s->red,s->green,s->blue,u)^0x1FFFF)<<i>>17;
}
DEF_RGB_RTN(RGBDirect)
	{return formDirect(s,(int)gt,i);}
DEF_RGB_RTN(RGBBDirect32)
{
	uint32		v = formDirect(s,(int)gt,i);

	return _reverse32(v);
}
DEF_RGB_RTN(RGBBDirect16)
{
	uint32		v = formDirect(s,(int)gt,i);

	return _reverse16(v);
}
DEF_RGB_RTN(RGBGray8)
{
	uint32		u;
(void)gt; (void)i;
	return formGray(s->red,s->green,s->blue,u)>>1;
}
DEF_RGB_RTN(RGBDirect8)
{
(void)gt;
	return formDirect(s,i,i);
}
DEF_RGB_RTN(RGBGray8X)
{
	uint32		u;
(void)i;
	return ((const uint8*)gt)[formGray(s->red,s->green,s->blue,u)>>7];
}
DEF_RGB_RTN(RGBDirect8X)
	{return ((const uint8*)gt)[formDirect(s,i,i)];}

const rgbRoutine	rgbCCRoutine[8] =
{
	RGBGray, RGBDirect, RGBBDirect32, RGBBDirect16,
	RGBGray8, RGBDirect8, RGBGray8X, RGBDirect8X
};

/* ---- HSV conversion */

void HSV2RGB(rgb48 *rgb,
	const hsv64 *hsv)
{
	uint32		sr, sg, sb, m;

	if (!rgb || !hsv)
		return;
	sg = ufract32mul(hsv->hue,65535*4*6);
	if (sg<65535*4*2)
	{
		sb = 0;
		if (sg<65535*4)
			sr = 65535*4;
		else
			sr = 65535*4*2-sg, sg = 65535*4;
	} else if (sg<65535*4*4)
	{
		sb = sg-65535*4*2, sr = 0;
		if (sb<65535*4)
			sg = 65535*4;
		else
			sg = 65535*4*2-sb, sb = 65535*4;
	} else
	{
		sr = sg-65535*4*4, sg = 0;
		if (sr<65535*4)
			sb = 65535*4;
		else
			sb = 65535*4*2-sr, sr = 65535*4;
	}
	m = hsv->sat<<15, m+= (hsv->sat+1)>>1;
	sr = 65535^(ufract32mul(65535*4-sr,m)+1)>>1;
	sg = 65535^(ufract32mul(65535*4-sg,m)+1)>>1;
	sb = 65535^(ufract32mul(65535*4-sb,m)+1)>>1;
	m = hsv->val, m = m+(m>>15);
	rgb->red = (uint16)((sr*m+32786)>>16);
	rgb->green = (uint16)((sg*m+32786)>>16);
	rgb->blue = (uint16)((sb*m+32786)>>16);
}

void RGB2HSV(hsv64 *hsv,
	const rgb48 *rgb)
{ /* val = MAX; sat = val ? ~MIN:0 */
	int64		m, t;
	int32		v, s, h, sr, sg, sb;

	if (!rgb || !hsv)
		return;
	sr = rgb->red, sg = rgb->green, sb = rgb->blue;
	if (sr>sg)
		v = sr, s = sg;
	else
		v = sg, s = sr;
	if ((hsv->val=(uint16)(v=MAX(v,sb)))==0)
		goto unsat;
	s = MIN(s,sb);
	sr-= s, sg-= s, sb-= s;
	s = ufix16div(s,v);
	s-= s>>15, s^= 65535;
	hsv->sat = (uint16)s;
	/* get scale for hue */
	v = MAX(sr,sg);
	if ((v=MAX(v,sb))==0)
	{
	unsat:
		hsv->sat = 0, hsv->hue = 0;
		return;
	}
	/* rescale */
	int64sethl(m,0x2AAAAAAA,0xAAAAAAA9); /* 2**64/6-1 */
	(void)int64addu32(m,v);
	(void)int64divu32(m,v);
	t = m; (void)int64mulu32(t,sr); sr = (int32)int64hi(t);
	t = m; (void)int64mulu32(t,sg); sg = (int32)int64hi(t);
	t = m; (void)int64mulu32(t,sb); sb = (int32)int64hi(t);
	if (sr>=0x2AAAAAAA)
	{ /* red dominant: 300-60 deg */
		if (sb) /* 300-0 */
			h =-sb;
		else /* 0-60 */
			h = sg;
	} else if (sg>=0x2AAAAAAA)
	{ /* green dominant: 60-180 deg */
		if (sr) /* 60-120 */
			h = 0x55555555-sr;
		else /* 120-180 */
			h = 0x55555555+sb;
	} else
	{ /* blue dominant: 180-300 deg */
		if (sg) /* 180-240 */
			h = 0xAAAAAAAA-sg;
		else /* 240-300 */
			h = 0xAAAAAAAA+sr;
	}
	hsv->hue = h;
}

/* ---- depth and size changes */

int pmConvert(pixMap *m,
	short depth,
	int w,
	int h,
	const cTab64 *ct,
	const rgb48 *eraseColor)
{
	bltParam	f;
	pixMap		tm;
	int8		*b, *d;
	int32		oldsize, newsize;
	int		edgew, edgeh, r = 0, fmt, ifm;

	if (!m || (w|h)+0u>65535)
		return -1;
	memset(&f,0,sizeof(f));
	fmt = formPxFmt(m);
	if (depth<0 && m->bpp)
	{
		tm = *m;
		tm.pHand = NULL; /* no accidental copies that might get disposed */
		tm.pTable = NULL;
		tm.iTable = NULL;
		tm.w = (uint16)w;
		tm.h = (uint16)h;
		ifm = fmt;
	} else
	{
		memset(&tm,0,sizeof(tm));
		tm.hRes = m->hRes;
		tm.vRes = m->vRes;
		_pmFormStruct(&tm,depth&0x407F,w,h); /* allow routine to find dimensions, etc. */
		ifm = formPxFmt(&tm);
	}
	if (!((fmt&0x80FFFF)!=ifm || ((uint16)fmt<0x900 && ct && (!m->pTable
			|| ct->seed!=m->pTable->seed))))
		fmt = 0;
	if (!fmt && m->h==tm.h && m->pitch==tm.pitch)
		goto cdim; /* width may change, but allocation won't */
	if ((depth>=0 || ct) && m->iTable) /* remove iTab if new pTable */
		{free(m->iTable); m->iTable = NULL;}
	if ((uint16)ifm<0x900)
		if (ct)
		{
			if (depth>=0 || ct!=m->pTable)
				(void)copyCTable(&tm.pTable,ct);
		} else
			(void)pmDefaultMacTable(&tm,depth&(int16)0x8080);
	if (!m->pHand)
	{
		if (m->baseAddr)
		{ /* fixed ptr, not resizable; regard data and pitch as constant */
			if (!fmt && tm.pitch<=m->pitch && tm.h<=m->h)
			{ /* dimensions fit */
			cdim:	m->w = tm.w, m->h = tm.h;
				return 0;
			}
			return -1;
		}
		/* dispose of old tables, and allocate */
		//if (!(m->pType&pTypeFixed)) -- feature not available from bpp alone
		if (m->pTable && (depth>=0 || ct!=m->pTable))
			free(m->pTable);
		*m = tm;
		if ((r=pmAllocate(m))<=0)
			return r ? r:-2;
		if (eraseColor)
		{
			lockPixels(m);
			pmErase(m,RGBToPen(&tm,eraseColor));
			unlockPixels(m);
		}
		return 1;
	}
	(void)unlockPixels(m);
	if ((uint16)ifm<0x900 && (r=pmMakeITables(&tm,tm.bpp,FALSE))<0)
		goto end;
	if (fmt && (r=rInitCnv(&f.cnv,&tm,m->pTable,fmt,0x10000,16,remap32bit))<0)
		goto end;
	oldsize = m->pitch*m->h;
	newsize = tm.pitch*tm.h;
	tm.pHand = m->pHand;
	if (newsize>oldsize && !hResize(tm.pHand,newsize))
		{r =-2; goto end;}
	if (!oldsize)
	{
		edgew = tm.w;
		edgeh = tm.h;
		goto skip;
	}
	if ((f.a=lockPixels(&tm))==NULL)
		{r =-1; goto end;}
	edgew = tm.w-m->w; /* newer is larger if positive */
	edgeh = tm.h-m->h;
	w = edgew>0 ? m->w:tm.w;
	h = edgeh>0 ? m->h:tm.h;
	b = NULL;
	if (fmt && tm.pSize>m->pSize)
		b = calloc(tm.pitch,1); /* max. nearly 256KB, _alloca's not a good idea */
	else if (!f.cnv.rtn)
		w = w*tm.pSize>>3;
	d = f.a;
	if (tm.pitch>m->pitch)
	{ /* lines need to be copied in reverse order */
		f.cnv.s = d+h*m->pitch, d+= h*tm.pitch;
		do
		{
			f.cnv.s = (const int8*)f.cnv.s-m->pitch, d-= tm.pitch;
			if (b)
			{
				cnvLine(b,w,tm.pSize,&f.cnv);
				memcpy(d,b,tm.pitch);
			} else if (f.cnv.rtn)
				cnvLine(d,w,tm.pSize,&f.cnv);
			else
				memmove(d,f.cnv.s,w);
		} while(--h);
	} else
	{
		f.cnv.s = d;
		do
			if (b)
			{
				cnvLine(b,w,tm.pSize,&f.cnv);
				memcpy(d,b,tm.pitch);
			} else if (f.cnv.rtn)
				cnvLine(d,w,tm.pSize,&f.cnv);
			else
				memmove(d,f.cnv.s,w);
		while(f.cnv.s=(const int8*)f.cnv.s+m->pitch,d+=tm.pitch,--h);
	}
	if (b)
		free(b);
skip:	if (m->pTable && ct!=m->pTable)
		{free(m->pTable); m->pTable = NULL;}
	if (eraseColor && (edgew>0 || edgeh>0))
	{
	extern DEF_BLT_RTN(sldx1);
		f.sc = RGBToPen(&tm,eraseColor);
		w = m->w*tm.pSize; /* was <<5 */
		if ((h=edgeh>0 ? m->h:tm.h)!=0 && (edgew*=tm.pSize)>0)
			do
				sldx1(w,edgew,&f);
			while(f.a+=tm.pitch,--h);
		else
			f.a+= tm.pitch*h;
		w = tm.w*tm.pSize;
		if (edgeh>0)
			do
				sldx1(0,w,&f);
			while(f.a+=tm.pitch,--edgeh);
	}
	(void)unlockPixels(&tm);
	if (newsize<oldsize)
		(void)hResize(tm.pHand,newsize);
	*m = tm;
	r = 1;
end:	if (r<0)
	{ /* failed; cleanup anything made for the temp pixmap */
		if (tm.pTable)
			free(tm.pTable);
		if (tm.iTable)
			free(tm.iTable);
	}
	rEndCnv(&f.cnv);
	return r;
}

int convertGrafPort(grafPort *p,
	int depth,
	int w,
	int h,
	const cTab64 *ct,
	const rgb48 *eraseColor)
{
	int		r;

	if (!p)
		return -1;
	if ((r=pmConvert(p->map,depth,w,h,ct,eraseColor))<0)
		return r;
	if (depth>=0)
	{
		RGBFore(p,&p->fgColor);
		RGBBack(p,&p->bkColor);
	}
	return 1;
}
