/*
 mimg_file.c
  image file reading

  Revised: 2024.06.09
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2016.06.25
  Code moved from mimg.c.
2019.03.03
  To be on par with QuickTime, "empty" color for PICT files is now white.
2023.09.18
  Added support for uncompressed BMP files.
2023.09.19
  Added support for bitfield conversion (converts to 10:11:10).
2023.09.21
  Added support for MS BMP RLE compression and embedded JPEG.
2024.06.07
  Added support for PPM/PGM/PBM and DIB files.
*/

#include <stdio.h>
#include <string.h>
#include <io.h>

#include "mimg_file.h"
#include "ftrig.h"

#define	ReadOK(s,b,z)		(rf(s,z,b)==z)

/* ---- code */

int BGRUnpackRow(uint8 *d,
	const uint8 *s,
	int w,
	int z,
	const RGBLineInfo *kh)
{
	uint32		c;

	if (w<=0)
		return w>>31;
	switch(z)
	{
	case 0: case 1: /* 8 bit or native copy */
		memcpy(d,s,w);
		break;
	case 2: c = w;
		do
			*(uint16*)d = (s[0]|s[1]<<8);//&mask;
		while(d+=2,s+=z,--c);
		break;
	default:
	 {
		uint32		i = z&7;
		uint32		k = (0x100<<((i-1)*8))-1, v, u, r, cz;

		c = w;
		do
		{ /* get pixel and get the compatible form */
			v = (s[0]|s[1]<<8|s[2]<<16|s[3]<<24)&k;
			if (z&8)
			{ /* convert to max 10:11:10 */
				u = (v>>kh->h[0])&kh->k[0];
				for(cz=kh->z[0]; cz<10; cz+=cz)
					u|= u<<cz;
				r = (u>>(cz-10)<<10+11)&0x7FE00000;
				u = (v>>kh->h[1])&kh->k[1];
				for(cz=kh->z[1]; cz<11; cz+=cz)
					u|= u<<cz;
				r|= (u>>(cz-11)<<10)&0x001FFC00;
				u = (v>>kh->h[2])&kh->k[2];
				for(cz=kh->z[2]; cz<10; cz+=cz)
					u|= u<<cz;
				v = r|(u>>(cz-10)&0x3FF);
			}
			*(uint32*)d = v;
		} while(d+=4,s+=i,--c);
	 }
	}
	return w;
}

static int inline BFDecodeMask(uint32 *mask,
	uint8 *size)
{
	int		o;
	unsigned	k, z;

	for(k=*mask,o=cntlz32(k),k<<=o,z=0; k&1u<<31; ++z,k+=k)
		;
	*size = z+!z; /* zero will hang converter */
	*mask = (1<<z)-1; /* low mask */
	return 32-o-z;
}

int RGBDepthFromFields(RGBLineInfo *kh)
{
	uint32		lg, bz;

	kh->h[0] = BFDecodeMask(kh->k+0,kh->z+0);
	kh->h[1] = BFDecodeMask(kh->k+1,kh->z+1);
	kh->h[2] = BFDecodeMask(kh->k+2,kh->z+2);
	lg = kh->z[1]-kh->z[0], bz = kh->z[2];
	if (lg<2u && kh->z[0]==bz && !kh->h[2] && kh->h[1]==bz && kh->h[0]==bz+bz+lg)
		return (bz*3)+(lg*(newDepthLGreen+1)); /* a native format */
	/* non-native; upconvert */
	return 31|64|newDepthLGreen;
}

#if defined(QT_IMAGE) && !defined(NO_QT_IMAGE_FILES)
static int readCT8RGB(cSpec *d,
	sio_func_t rf,
	const void *s,
	int count,
	int cmcsize)
{
	uint8		b[4];
	int		i;

	if (cmcsize>4)
		return -1;
	for(i=0; i<count; ++i,++d)
	{
		if (!ReadOK(s,b,cmcsize))
			return -3;
		d->rgb.blue = *b*0x101;
		d->rgb.green = b[1]*0x101;
		d->rgb.red = b[2]*0x101;
	}
	return count;
}

int dcmpReadFileDIB(pixMap **m,
	sio_func_t rf,
	const void *s,
	BMdcmpHeader *dhd)
{
	uint8		hb[32];
	pixMap		*pm = NULL;
	uint8		*d, *b = NULL;
	size_t		allocatedSize;
	int		r, w, h, y, depth, rowOrder;

	switch(dhd->biCompression)
	{
	case 3: /* BI_BITFIELDS */
		if (dhd->headerSize<52)
		{
			if (dhd->headerPad<12)
				goto bad;
			dhd->headerPad-= 12;
			if (!ReadOK(s,hb,12))
				{r =-3; goto end;}
			dhd->hk.k[0] = le32_native(*(uint32*)hb);
			dhd->hk.k[1] = le32_native(*(uint32*)(hb+4));
			dhd->hk.k[2] = le32_native(*(uint32*)(hb+8));
		}
		dhd->depth = RGBDepthFromFields(&dhd->hk);
		if (dhd->depth<0)
			{r = dhd->depth; goto end;}
		//r = (2<<(bpp-1))-1;
		//if ((hk.k[0]<<hk.h[0]|hk.k[1]<<hk.h[1]|hk.k[2]<<hk.h[2])>r)
		//	{r =-1; goto end;} /* at least one mask surpasses bitfield */
	case 0: /* BI_RGB/indexed */
	case 1: /* BI_RLE8 */
	case 2: /* BI_RLE4 */
		break;
	case 4: /* BI_JPEG */
		if (dhd->headerSize==64 && dhd->bpp)
			{dhd->biCompression = 1; break;}
	#ifdef QT_IJG9
		if ((r=decompressJPEG(m,srd_fread,s))>0)
			return r;
		return -1;
	#endif
	default:
		return -8;
	}
	if (dhd->headerPad<0)
		{r =-1; goto end;}
	for(; (r=MIN(dhd->headerPad,32))>0; dhd->headerPad-=r)
		if (!ReadOK(s,hb,r))
			{r =-3; goto end;}
	r = 0;
	if (!m)
		goto end; /* verify header only */
	depth = dhd->depth;
	w = dhd->w;
	h = dhd->h;
	if ((pm=newPixMap(depth&0x83F,w,dhd->h))==NULL)
		{r =-2; goto end;}
	r = (((w*dhd->bpp)>>3)+3)&~3; /* 4-byte aligned in file */
	/* maximum unpacked run length is 255*4 */
	if ((b=malloc(MAX(r,1024)))==NULL || pmAllocate(pm)<=0
			|| (d=(uint8*)lockPixels(pm))==NULL)
	{
		pmFree(pm); pm = NULL;
		r =-2;
		goto end;
	}
	*m = pm;
	if (dhd->hRes)
	{ /* convert pixels/meters to dpi */
		pm->hRes = dhd->hRes;
		pm->vRes = dhd->vRes;
	}
	pm->pTable = dhd->ct;
	allocatedSize = pm->pitch*dhd->h;
	memset(d,0,allocatedSize);
	depth = (dhd->bpp|(depth&64))>>3; /* now src bytes per pixel */
	rowOrder =~dhd->rowOrder;
	y = (h-1)&rowOrder;
	d+= (allocatedSize-pm->pitch)&rowOrder;
	rowOrder|= 1;
	if (dhd->biCompression-1u>2u-1u)
	{ /* read image data to buffer */
		if (dhd->bpp<9)
			w = r;
		do
		{
			if (!ReadOK(s,b,r))
				{r =-3; break;}
			BGRUnpackRow(d,b,w,depth,&dhd->hk);
		} while(d+=pm->pitch*rowOrder,(y+=rowOrder)>=0);
	} else
	{ /* decompress RLE */
		const uint8	*a, *de = rowOrder<0 ? d+w:d+allocatedSize;
		unsigned	pz = pm->pSize;
		int		x = 0, i = 0, c, pad;

		depth&= 7;
		pad = depth<2; /* padding only in indexed RLE; rgb24rle24.bmp is likely invalid */
		do
		{
			if (!ReadOK(s,b,2))
				{r =-3; break;}
			i+= 2, r = b[1];
			if ((c=*b)==0) /* escape */
				switch(r)
				{
				case 0: x = 0; /* end of line */
					if ((y+=rowOrder)<0)
					{
						if (!ReadOK(s,b,1))
							{r =-3; break;}
						++i;
						if (*b==1)
							goto succ;
						else
							goto bad;
					}
					d = (uint8*)pm->baseAddr+y*pm->pitch;
					break;
				case 1: goto succ; /* end of bitmap */
				case 2: if (!ReadOK(s,b,2)) /* delta */
						{r =-3; break;}
					i+= 2;
					x+= *b, y+= b[1]*rowOrder;
					if (x+0u>=w+0u || y+0u>=h+0u)
						goto bad;
					d = (uint8*)pm->baseAddr+y*pm->pitch+(depth ? x*depth:x>>1);
					break;

				default: /* absolute run */
					c = r;
					if (d+(c*pz>>3)>de)
						goto bad;
					r = depth ? c*depth:(c+1)>>1;
					r+= pad&(i+r);
					if (!ReadOK(s,b,r))
						{r =-3; break;}
					i+= r, a = b, x+= c;
					switch(depth)
					{
					case 4:
						do
							*(uint32*)d = *a|a[1]<<8|a[2]<<16|a[3]<<24;
						while(d+=4,a+=4,--c);
					case 3:
						do
							*(uint32*)d = *a|a[1]<<8|a[2]<<16;
						while(d+=4,a+=3,--c);
						break;
					case 2:
						do
							*(uint16*)d = *a|a[1]<<8;
						while(d+=2,a+=2,--c);
						break;
					case 1:
						do
							*d++ = *a++;
						while(--c);
						break;
					case 0:
						if ((x-c)&1)
						{
							*d++|= *a>>4, --c;
							for(; c>1; c-=2,++a)
								*d++ = *a<<4|a[1]>>4;
							if (c)
								*d = *a<<4;
							break;
						}
						for(; c>1; c-=2)
							*d++ = *a++;
						if (c)
							*d = *a&0xF0;
						break;
					default:
						;
					}
					if (x+0u>w+0u)
						goto bad;
				}
			else
			{ /* a repeat run */
				x+= c;
				if (d+(c*pz>>3)>de)
					goto bad;
				if (depth>1)
				{
					if (!ReadOK(s,b,depth-1))
						{r =-3; break;}
					i+= depth-1;
				}
				switch(depth)
				{
				case 4: r|= b[2]<<8;
				case 3:	r|= b[1]<<8;
				case 2:	r|= *b<<8;
					if (depth==2)
						do
							*(int16*)d = (int16)r;
						while(d+=2,--c);
					else
						do
							*(int32*)d = r;
						while(d+=4,--c);
					break;
				case 1:
					do
						*d++ = (uint8)r;
					while(--c);
					break;
				case 0:
					if ((x-c)&1)
					{
						*d++|= r>>4, --c, r = r<<4|r>>4;
						for(; c>1; c-=2)
							*d++ = r;
					} else
						for(; c>1; c-=2)
							*d++ = r;
					if (c)
						*d = r&0xF0;
					break;
				default:
					vhalt("internal error");
				}
				if (x+0u>w+0u)
					{bad: r =-1; goto end;}
			}
		} while(y>=0);
	}
succ:	r = 1;
end:	if (!pm && dhd->ct)
		{free(dhd->ct); dhd->ct = NULL;}
	if (b)
		free(b);
	return r;
}

int imgReadFileDIB(pixMap **m,
	sio_func_t rf,
	const void *s,
	boolean hasBMPHeader)
{
	uint8		hb[14+124 +2]; /* OS/2 2 goes up to 64 bytes */
#define header (2+hb)
#define infoheader (header+14)
	BMdcmpHeader	dhd;

	int32		r;
	int		bfOffBits, ccount = 0;
	int		cmcsize = 4; /* RGBQUAD for modern 8bpp cmap */

	if (!hasBMPHeader)
	{
		*(uint16*)header = MK2CC('B','M');
		*(int32*)(header+10) = 0;
	} else if (!ReadOK(s,header,14))
		return -3;
	memset(&dhd,0,sizeof(dhd));
	if (*(const uint16*)header==MK2CC('B','A'))
	{ /* other signatures include: IC, PT, CI, CP */
		if (memcmp(header+2,"\x28\0\0\0\0\0\0\0\0\0\0\0",12))
			return -8;
		/* array, single */
		if (!ReadOK(s,header,14))
			return -3;
		dhd.headerPad =-14; /* image offset is absolute */
	}
	if (*(const uint16*)header!=MK2CC('B','M'))
		return -8; /* not valid or unsupported */
	if (!ReadOK(s,infoheader,4))
		return -3;
	bfOffBits = le32_native(*(int32*)(header+10));
	dhd.headerSize = le32_native(*(int32*)infoheader);
	if (*(const uint16*)header!=MK2CC('B','M') || dhd.headerSize-12u>124-12u)
		return -8; /* not valid or unsupported */
	if (!ReadOK(s,infoheader+4,dhd.headerSize-4))
		return -3;
	switch(dhd.headerSize)
	{
	case 12: /* BITMAPCOREHEADER */
		dhd.w = (int16)le16_native(*(uint16*)(infoheader+4));
		dhd.h = (int16)le16_native(*(uint16*)(infoheader+6));
		dhd.biPlanes = le16_native(*(uint16*)(infoheader+8));
		dhd.bpp = le16_native(*(uint16*)(infoheader+10));
		cmcsize = 3;
		break;
	case 108: /* BITMAPV4HEADER */
	case 124: /* BITMAPV5HEADER */
	case 64: /* BITMAPINFOHEADER+ */
		//ctype = le32_native(*(uint32*)(infoheader+56));
	case 56:
	case 52:
		dhd.hk.k[0] = le32_native(*(uint32*)(infoheader+40));
		dhd.hk.k[1] = le32_native(*(uint32*)(infoheader+44));
		dhd.hk.k[2] = le32_native(*(uint32*)(infoheader+48));
		/* biAlphaMask ignored for now */
	case 40: /* Win 3/BITMAPINFOHEADER, OS/2 2 */
		dhd.w = le32_native(*(uint32*)(infoheader+4));
		dhd.h = le32_native(*(uint32*)(infoheader+8));
		dhd.biPlanes = le16_native(*(uint16*)(infoheader+12));
		dhd.bpp = le16_native(*(uint16*)(infoheader+14));
		dhd.biCompression = le16_native(*(uint16*)(infoheader+16));
		/* biSizeImage ignored for now */
		dhd.hRes = le32_native(*(uint32*)(infoheader+24));
		dhd.vRes = le32_native(*(uint32*)(infoheader+28));
		if (dhd.hRes>0 && dhd.vRes>0)
		{ /* convert pixels/meters to dpi */
			dhd.hRes = (ifixed)(dhd.hRes*(0.0254*65536)+.5);
			dhd.vRes = (ifixed)(dhd.vRes*(0.0254*65536)+.5);
		}
		ccount = le32_native(*(uint32*)(infoheader+32));
		/* biClrImportant ignored for now */
		break;
	default:
		return -1;
	}
	switch(dhd.depth=dhd.bpp)
	{ /* Windows does not support 2-bit, but it is valid on other systems */
	case 1: case 2: case 4: case 8: break;
	case 0: if (dhd.biCompression==4)
	case 16: case 24: case 32: cmcsize = 0; break;
	default: return -1;
	}
	dhd.headerPad+= bfOffBits-dhd.headerSize-14;
	if (dhd.biPlanes!=1 || dhd.w<=0 || !dhd.h)
		return -1;
	dhd.rowOrder = r = dhd.h>>31;
	dhd.h = (dhd.h+r)^r;
	if (dhd.w>32767 || dhd.h>32767)
		return -25;

	if (cmcsize>0)
	{ /* read colormap */
		r = 1<<dhd.bpp;
		if (ccount-1u>255u) /* oversized palettes actually get clipped */
			ccount = r; /* assume max */
		dhd.headerPad-= ccount*cmcsize;
		if ((dhd.ct=calloc(8,MAX(ccount,r)+1))==NULL)
			return -2;
		dhd.ct->flags = 0x8000;
		dhd.ct->count = ccount;
		readCT8RGB(dhd.ct->colors,rf,s,ccount,cmcsize);
	}
	if (!hasBMPHeader)
		dhd.headerPad = 0;
	return dcmpReadFileDIB(m,rf,s,&dhd);
}
 #undef header
 #undef infoheader
#endif

/* PPM, PGM, PBM */
#ifndef NO_DCMP_PPM_IMAGE_FILES
int32 PPM_readint(sio_func_t rf,
	const void *s,
	uint8 *_com)
{
	uint8		b[1], com = *_com;
	uint32		v, t;

	while(ReadOK(s,b,1))
		switch(*b)
		{
		case '\t': case ' ': continue;
		case 10: case 13: case 11: case 12: com = 0; break;
		case '#': com = 1; break; /* comment - ignore rest of line */
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			if (!com)
				goto num;

		default: if (!com)
				return -8;
		}
	return -3; /* got nothing by EOF when required */
num:	v = *b-'0';
	while(ReadOK(s,b,1))
	{
		if ((t=*b-'0')>9u)
		{ /* potential ungetc */
			if (t=='#'-'0'+0u)
				com = 1;
			break;
		}
		v = v*10+t;
		if (v>65535)
			return -1;
	}
	*_com = com;
	return v;
}

#define DFROWRTN(n)	int32 n(sio_func_t rf, const void *s, uint8 *_d, uint8 *b, int w, ufixed m)
#define TPROWRTN(n)	int32(*n)(sio_func_t rf, const void *s, uint8 *_d, uint8 *b, int w, ufixed m)

static DFROWRTN(srd_Text1)
{
	uint8		*d, v;
	int32		r, c;

	d = _d, c = w, v = 1;
	do
	{
		r = PPM_readint(rf,s,b);
		if (r+0u>1u)
			return r<0 ? r:-1;
		if (v&0x80)
			*d++ = v+v+r, v = 1;
		else
			v+= v, v+= r;
	} while(--c);
	*d = v<<(8-(w&7));
	return w;
}

static DFROWRTN(srd_Text8)
{
	uint8		*d, k = b[2];
	int32		r, t, c;

	d = _d, c = w;
	do
	{
		r = PPM_readint(rf,s,b);
		t = ufix16mul(r,m);
		if ((r|t)+0u>255)
			return r<0 ? r:-1;
		*d++ = (uint8)t^k;
	} while(--c);
	return w;
}

static DFROWRTN(srd_TextG32)
{
	uint32		*d, x = 1<<b[1], k = (x<<b[1])|x|1;
	int32		r, t, c;

	d = (uint32*)_d, c = w;
	do
	{
		r = PPM_readint(rf,s,b);
		t = ufix16mul(r,m);
		if ((r|t)+0u>=x)
			return r<0 ? r:-1;
		*d++ = t*k;
	} while(--c);
	return w;
}

static DFROWRTN(srd_TextRGB16)
{
	uint16		*d;
	int32		cr, cg, cb, v, t, c, h = b[1], x = 1<<h;

	d = (uint16*)_d, c = w;
	do
	{
		cr = PPM_readint(rf,s,b);
		v = ufix16mul(cr,m)<<h;
		cg = PPM_readint(rf,s,b);
		v = (v|ufix16mul(cg,m))<<h;
		cb = PPM_readint(rf,s,b);
		t = ufix16mul(cb,m);
		if ((cr|cg|cb|t)+0u>=x+0u)
			return cr<0 ? cr:cg<0 ? cg:cb<0 ? cb:-1;
		*d++ = v|ufix16mul(t,m);
	} while(--c);
	return w;
}

static DFROWRTN(srd_TextRGB32)
{
	uint32		*d;
	int32		cr, cg, cb, v, t, c, h = b[1], x = 1<<h;

	d = (uint32*)_d, c = w;
	do
	{
		cr = PPM_readint(rf,s,b);
		v = ufix16mul(cr,m)<<h;
		cg = PPM_readint(rf,s,b);
		v = (v|ufix16mul(cg,m))<<h;
		cb = PPM_readint(rf,s,b);
		t = ufix16mul(cb,m);
		if ((cr|cg|cb|t)+0u>=x+0u)
			return cr<0 ? cr:cg<0 ? cg:cb<0 ? cb:-1;
		*d++ = v|ufix16mul(t,m);
	} while(--c);
	return w;
}

static DFROWRTN(cvt_Row1)
{
	uint8		*d;
	int		c;

	d = _d, c = w;
	do /* white to black */
		*d++ = ufix16mul(*b++,m)^255;
	while(--c);
	return w;
}

static DFROWRTN(cvt_Row2)
{
	uint32		*d;
	int		c;

	d = (uint32*)_d, c = w;
	do /* black to white */
		*d++ = ufix16mul(*b<<8|b[1],m)*(1<<20|1<<10|1); //^1023)*
	while(b+=2,--c);
	return w;
}

static DFROWRTN(cvt_RowRGB2)
{
	uint32		r, g, h = b[-1];
	uint16		*d;
	int		c;

	d = (uint16*)_d, c = w;
	do
	{
		r = ufix16mul(*b,m);
		g = ufix16mul(b[1],m);
		*d++ = (((r<<h)|g)<<h)|ufix16mul(b[2],m);
	} while(b+=3,--c);
	return w;
}

static DFROWRTN(cvt_RowRGB3)
{
	uint32		*d, r, g, h = b[-1];
	int		c;

	d = (uint32*)_d, c = w;
	do
	{
		r = ufix16mul(*b,m);
		g = ufix16mul(b[1],m);
		*d++ = (((r<<h)|g)<<h)|ufix16mul(b[2],m);
	} while(b+=3,--c);
	return w;
}

static DFROWRTN(cvt_RowRGB6)
{
	uint32		*d, r, g, h = b[-1];
	int		c;

	d = (uint32*)_d, c = w;
	do
	{
		r = ufix16mul(*b<<8|b[1],m);
		g = ufix16mul(b[2]<<8|b[3],m);
		*d++ = (((r<<h)|g)<<h)|ufix16mul(b[4]<<8|b[5],m);
	} while(b+=6,--c);
	return w;
}

uint8	ppmbinsiz[5] = {1, 2, 3, 3, 6};
TPROWRTN(ppmbinrtn[5]) = {cvt_Row1, cvt_Row2, cvt_RowRGB2, cvt_RowRGB3, cvt_RowRGB6};
TPROWRTN(ppmtxtrtn[5]) = {srd_Text8,srd_TextG32,srd_TextRGB16,srd_TextRGB32,srd_TextRGB32};

int imgReadFilePPM(pixMap **m,
	sio_func_t rf,
	const void *s)
{
	TPROWRTN(rtn);
	pixMap		*pm = NULL;
	uint8		hb[4], *b, *d;
	int32		w, h, r, maxv, z = 0, bz = 0, dmax;
	ufixed		scale;
	uint8		bpp, text, rgb, is16b;

	if (!ReadOK(s,hb,2))
		return -3;
	text = rgb = 0;
	if (*hb!='P')
		return -8;
	r = hb[1];

	*(int32*)hb = 0;
	dmax = 255, bpp = 8;
	switch(r)
	{
	case '3': text = 2; hb[2] = 255;
	case '6': rgb = 3;
		break;
	case '2': text = 2; hb[2] = 255;
	case '5': break;
	case '1': text = 1; bpp = 1; break; /* 1-bit */

	case '4': /* compressed PBM not supported */
	default: return -8;
	}

	w = PPM_readint(rf,s,hb);
	h = PPM_readint(rf,s,hb);
	maxv = text==1 ? 1:PPM_readint(rf,s,hb);
	if (maxv<0)
		return maxv;
	if (w<=0 || h<=0 || maxv<=0)
		return -1;
	if (maxv>65535)
		return -8;
	is16b = maxv>255;
	hb[1] = 32-cntlz32(maxv);
	if (hb[1]>9)
		hb[1] = 10;
	if (is16b|rgb)
	{
		dmax = (1<<hb[1])-1;
		if (hb[1]<6)
		{
			--rgb;
			if (dmax!=maxv)
				dmax = 31, hb[1] = 5; /* avoid precision banding */
		}
		bpp = hb[1], bpp+= bpp+bpp;
	}
	scale = ufix16div(dmax,maxv);
	z = is16b+rgb;
	rtn = ppmbinrtn[z];
	b = NULL;
	if (text)
		z = 0;
	else
	{
		z = w*ppmbinsiz[z];
		if (z<768)
			b = _alloca(z+8);
		else
			b = malloc(z+8), bz = z;
	}
	if ((!b && z) || (pm=newPixMap(bpp,w,h))==NULL)
		goto fail;
	if (pmAllocate(pm)<=0 || (d=(uint8*)lockPixels(pm))==NULL)
	{
		disposePixMap(pm);
	fail:	r =-2;
		goto end;
	}
	*m = pm;
	if (text)
	{
		rtn = bpp==1 ? srd_Text1:ppmtxtrtn[rgb ? rgb:is16b];
		do
			rtn(rf,s,d,hb,w,scale);
		while(d+=pm->pitch,--h);
	} else
	{
		*b = hb[1];
		do
		{
			if (!ReadOK(s,b+1,z))
				return -3;
			rtn(rf,s,d,b+1,w,scale);
		} while(d+=pm->pitch,--h);
	}
	r = 1;
end:	if (bz)
		free(b);
	return r;
}
#endif

static int testPICT(FILE *f)
{
	uint8		b[512];
	const int32	*s = (const int32*)b;
	int		c = 512/16;

	if (fread(b,512,1,f)<=0)
		return -3;
	if (*b-12u<125u-12 && !(b[1]|b[2]|b[3]))
		return -8; /* possible DIB */
	do
		if (*s|s[1]|s[2]|s[3])
			return 0;
	while(s+=4,--c);
	return 1;
}

pixMap *decompressImageFile(const char *path,
	int *err,
	uint16 *w,
	uint16 *h,
	uint8 *mode)
{
	FILE		*f = NULL;
	pixMap		*m = NULL;
	long long	z;
	int		r = 0;

	if (!path)
		{r =-1; goto end;}
	if ((f=fopen(path,"rb"))==0)
		{r =-43; goto end;}
	z = _filelength64(_fileno(f));
	if (z>0x7FFFFFF0)
		z = 0x7FFFFFF0; /* spooling not implemented, so clip */
	r = testPICT(f);
	if (r>0)
	{
		uint8		*b = malloc(z);

		r =-2;
		if (b)
		{
			r = 0;
			if (fread(b,z,1,f)<=0)
				r =-3;
			m = renderQDPict(b,(int)z,w,h,mode,&rgb_white,NULL);
			free(b);
		}
		goto end;
	}
	/* not a PICT file */
#if defined(QT_IMAGE) && !defined(NO_QT_IMAGE_FILES)
	if (r==-8)
	{
		fseek(f,0,SEEK_SET);
		r = imgReadFileDIB(&m,srd_fread,f,0);
		if (r>=0)
			goto succ;
		goto end;
	}
#endif
	r =-8;
#ifndef NO_DCMP_PPM_IMAGE_FILES
	fseek(f,0,SEEK_SET);
	r = imgReadFilePPM(&m,srd_fread,f);
	if (r!=-8)
 		goto succ;
#endif
	fseek(f,0,SEEK_SET);
#if defined(QT_IMAGE) && !defined(NO_QT_IMAGE_FILES)
	r = decompressGIF(&m,srd_fread,f,1,NULL);
 	fseek(f,0,SEEK_SET);
	if (r!=-8)
 		goto succ;
	r = imgReadFileBMP(&m,srd_fread,f);
 #ifdef QT_IJG9
	if (r!=-8)
 		goto succ;
 	fseek(f,0,SEEK_SET);
	if ((r=decompressJPEG(&m,srd_fread,f))>0)
	{
	succ:
 #else
	if (r>0)
	{
 #endif
		if (!m)
			{r = r<0 ? r:-1; goto end;}
		if (mode)
			*mode = mDither;
		if (w)
			*w = m->w;
		if (h)
			*h = m->h;
	}
#endif
end:	if (f)
		fclose(f);
	if (err)
		*err = r;
	return m;
}
