/*
 img_gif.c
  GIF8x image/animation decompression

  Created: 2016.06.23
  Revised: 2023.08.10

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2016.07.02
  Now correctly handles true color local table composite mapping.
2016.07.12
  Thread-safe.
*/

#include <string.h>
#include <errno.h>

#include "lzw.h"
#include "mgraf.h"
#include "mimg_file.h"


/* Code from here to end of file has been adapted from XPaint:           */
/* +-------------------------------------------------------------------+ */
/* | Copyright 1990, 1991, 1993 David Koblas.                          | */
/* | Copyright 1996 Torsten Martinsen.                                 | */
/* |   Permission to use, copy, modify, and distribute this software   | */
/* |   and its documentation for any purpose and without fee is hereby | */
/* |   granted, provided that the above copyright notice appear in all | */
/* |   copies and that both that copyright notice and this permission  | */
/* |   notice appear in supporting documentation.  This software is    | */
/* |   provided "as is" without express or implied warranty.           | */
/* +-------------------------------------------------------------------+ */

#define CM_RED			0
#define CM_GREEN		1
#define CM_BLUE			2

#define INTERLACE		0x40
#define HAS_COLORMAP		0x80

#define	ReadOK(s,b,z)		(rf(s,z,b)==z)

/* ---- code */

static int GetDataBlock(sio_func_t rf,
	const void *s,
	uint8 *buf)
{
	uint8		count;

	if (!ReadOK(s,&count,1) || (count && !ReadOK(s,buf,count)))
		return -3; /* error in reading DataBlock */
	return count;
}

static int DoExtension(sio_func_t rf,
	const void *s,
	int t,
	GIFState *state)
{
	uint8		b[256];
	int		c;

	if ((c=GetDataBlock(rf,s,b))<=0)
		return c;
	switch(t)
	{
	case 1: /* Plain Text Extension */
	case 0xfe: /* Comment Extension */
		break;
	case 0xff: /* Application Extension */
		if (c!=11 || memcmp(b,"NETSCAPE2.0",11))
			break;
		if ((c=GetDataBlock(rf,s,b))<=0)
			return c;
		if (c!=3 || b[0]!=1)
			break;
		state->info.loopCount = b[1]|b[2]<<8;
		break;
	case 0xf9: /* Graphic Control Extension */
		state->screen.Transparent = (b[0]&1)-1;
		state->info.disposal = b[0]>>2&7;
		state->info.inputFlag = b[0]>>1&1;
		if (c>2)
		{
			state->info.delayTime = b[1]|b[2]<<8;
			state->info.transparent = state->screen.Transparent+1;
			if (c>3)
				state->info.index = b[3];
		}
		if (c>3)
		{
			state->screen.Transparent|= b[3];
			state->info.index = b[3];
		}
	/* no other codes in the spec */
	}
	/* skip other subcodes */
	while(GetDataBlock(rf,s,b))
		;
	return 0;
}

static int ReadColorMap(sio_func_t rf,
	const void *s,
	int count,
	uint32 *cmap,
	GIFState *state)
{
	uint8		b[3];
	int		i;//, flag = 1;

	for(i=0; i<count; ++i,++cmap)
	{
		if (!ReadOK(s,b,3))
			return -3;
		*cmap = (((i!=state->screen.Transparent ? ~0xFF:0)
				|b[CM_RED])<<8|b[CM_GREEN])<<8|b[CM_BLUE];
	}
	return 0;
}

static int ReadImage(pixMap *image,
	sio_func_t rf,
	const void *s,
	int left,
	int top,
	int width,
	int height,
	int bpp,
	const uint32 *cmap,
	uint8 interlace,
	GIFState *state)
{
	int8		*p, *d;
	int		i, v, xpos, ypos, pass;
	int		right, bottom, pitch;
	uint8		c;
	boolean		is8b;

	/*
	**	Initialize the compression routines
	*/
	if (!ReadOK(s,&c,1))
		return -3; /* EOF / read error on image data */

	if ((i=LZWReadByte(rf,s,1,c,&state->lzw))<0)
		return i; /* error reading image */

	if (!image)
	{
		while(LZWReadByte(rf,s,0,c,&state->lzw)>=0)
			;
		return 0;
	}

	if ((p=lockPixels(image))==NULL || (pitch=image->pitch)<=0)
	{
		disposePixMap(image);
		return -1;
	}
	is8b = image->pSize==8;
	if (image->pTable)
	{
		cSpec		*color = image->pTable->colors;

		image->bpp = bpp, i = 1<<bpp;
		if (i>image->pTable->count)
			i = image->pTable->count;
		for(; --i>=0; ++color,++cmap)
		{
			uint32		entry = *cmap;

			color->rgb.red = (entry>>16&0xFF)*257;
			color->rgb.green = (entry>>8&0xFF)*257;
			color->rgb.blue = (entry&0xFF)*257;
		}
	}
	right = left+width;
	bottom = top+height;
	xpos = left, ypos = top, pass = 0;
	if (ypos>=state->screen.Height)
		goto end;
	for(d=p+ypos*pitch; (v=LZWReadByte(rf,s,0,c,&state->lzw))>=0; )
	{
		if (xpos+0u<pitch+0u && v!=state->screen.Transparent)
			if (is8b)
				d[xpos] = v;
			else
				((int32*)d)[xpos] = cmap[v];

		if (++xpos>=right)
		{
			xpos = left;
			if (interlace)
			{
				switch(pass)
				{
				case 0:
				case 1: ypos+= 8;
					break;
				case 2: ypos+= 4;
					break;
				case 3: ypos+= 2;
				}
				if (ypos>=bottom)
				{
					++pass;
					switch(pass)
					{
					case 1:	ypos = 4+top; /* 2022 */
						break;
					case 2:	ypos = 2+top;
						break;
					case 3:	ypos = 1+top;
						break;

					default:
						goto end;
					}
				}
			} else
				++ypos;
			if (ypos>=bottom)//state->screen.Height)
				break;
			d = p+ypos*pitch;
		}
	}
end:	return 1;
}

int decompressGIF(pixMap **output,
	sio_func_t rf,
	const void *s,
	int desiredFrame,
	GIFState **statep)
{
	GIFState	*state, *statet;
	uint32		localColorMap[256], *colorMap;
	uint8		buf[16];
	int		r, bpp;
	uint8		version, c;

	if (!rf || !s)
		return -1;
	if (desiredFrame<0)
	{
		desiredFrame =-desiredFrame;
		if (!statep)
			return -1; /* must have previous frame state! */
		if ((state=*statep)==NULL)
			return 0;
	} else
	{
		if (!ReadOK(s,buf,6))
			return -3; /* error reading magic number */
		version = buf[4];
		if (memcmp(buf,"GIF8",4) || buf[5]!='a' || (version!='7' && version!='9'))
			return -8; /* not a supported GIF file */
		if (!ReadOK(s,buf,7))
			return -3; /* failed to read screen descriptor */
		if (!desiredFrame)
			return 0;
		if (!statep)
			statep = &statet, *statep = NULL;
		if ((state=*statep)==NULL)
		{
			state = calloc(sizeof(*state),1);
			if (statep!=&statet)
				*statep = state;
		}
		state->screen.Width = *buf|buf[1]<<8;
		state->screen.Height = buf[2]|buf[3]<<8;
		state->screen.BPP = 1+(buf[4]&7);
		state->screen.ColorResolution = (buf[4]>>3&14)+1;
		state->screen.Background = buf[5];
		state->screen.AspectRatio = buf[6];
		state->screen.Transparent =~0;

		if (buf[4]&HAS_COLORMAP) /* Global Colormap */
			if (ReadColorMap(rf,s,1<<state->screen.BPP,state->screen.ColorMap,state))
			{ /* error reading global colormap */
				r =-3;
				goto end;
			}
	}
	state->info.loopCount =~0;
	state->info.delayTime =~0;
	*(int32*)&state->info.index = 0;
	for(;;)
	{
		pixMap		*image;

		if (!ReadOK(s,&c,1))
			{r =-3; goto end;} /* EOF / read error on image data */
		if (c==';')
		{ /* GIF terminator */
			if (state->screen.FrameCount<desiredFrame)
				{r =-7; goto end;} /* desired image index not found */
			r = output && *output;
			goto end;
		}
		if (c=='!')
		{ /* Extension */
			if (!ReadOK(s,&c,1))
				{r =-3; goto end;} /* EOF / read error on extension function code */
			DoExtension(rf,s,c,state);
			continue;
		}
		if (c!=',')
			continue; /* Not a valid start character */
		++state->screen.FrameCount;

		if (!ReadOK(s,buf,9))
			{r =-3; goto end;} /* couldn't read left/top/width/height */
		bpp = (buf[8]&7)+1;

		if (buf[8]&HAS_COLORMAP)
		{ /* local-not-global color table */
			if (ReadColorMap(rf,s,1<<bpp,localColorMap,state))
				{r =-3; goto end;} /* error reading local colormap */
			colorMap = localColorMap;
		} else
		{
			bpp = state->screen.BPP;
			colorMap = state->screen.ColorMap;
		}
		image = NULL;
		if (output)
		{
			if (*output)
			{ /* in-place transparency may accumulate 24bpp colors; */
			  /* it's actually an old technique of composite imaging to have */
			  /* multiple local cmaps produce images of more than 256 colors */
				if (state->screen.FrameCount>1 && buf[8]&HAS_COLORMAP)
					pmConvert(*output,32,state->screen.Width,
						state->screen.Height,NULL,NULL);
			} else if ((*output=alloc_image(8,state->screen.Width,
					state->screen.Height,NULL,NULL))==NULL)
				{r =-2; goto end;}
			else
				pmErase(*output,state->screen.Background*0x1010101);
			image = *output;
		}
		if ((r=ReadImage(image,rf,s,
				*buf|buf[1]<<8,
				buf[2]|buf[3]<<8,
				buf[4]|buf[5]<<8,
				buf[6]|buf[7]<<8,
				bpp,colorMap,buf[8]&INTERLACE,state))<0)
			break;
		if (desiredFrame==state->screen.FrameCount)
			goto end;
	}
	r = 1;
end:	if (r==-3 || statep==&statet)
		{free(state); *statep = NULL;}
	return r;
}
