/*
 mimg_jpeg
  JPEG image decompression

  Created: 2016.06.27
  Revised: 2023.09.24

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#include <stdio.h>

/* include must be placed here for default struct alignment */
#include "../jpeg9/jpeglib.h"

#include "mimg_file.h"

/* ---- code */

/* RGB24 to native endian 32-bit with full alpha */
static void unpack_rgb24(int8 *d,
	int w)
{
	const uint8	*s;

	d+= w*4, s = (const uint8*)(d-w);
	do
	{
		s-= 3;
		*(int32*)(d-=4) = (~0<<24)|(*s<<8|s[1])<<8|s[2];
	} while(--w);
}

/* 2022.04.10 */
/*
METHODDEF(noreturn_t)
jerror_debug(j_common_ptr cinfo)
{
	__debugbreak();
}
*/

int decompressJPEG(pixMap **output,
	sio_func_t rf,
	const void *s)
{
	struct jpeg_decompress_struct	cinfo;
	struct jpeg_error_mgr		jerr;

	pixMap		*image = NULL;
	int8		*d;
	int		r =-8, depth, pitch;
	/* seriously need to revise error code to return number instead of exit() */
	cinfo.err = jpeg_std_error(&jerr);
	//cinfo.err->error_exit = jerror_debug;
	jpeg_create_decompress(&cinfo);

	if (rf==srd_sread)
		jpeg_mem_src(&cinfo,((srd_ptr*)s)->src,((srd_ptr*)s)->remaining);
	else if (rf==srd_fread)
		jpeg_stdio_src(&cinfo,(FILE*)s);
	else /* other readers not easily supported by the JPEG-9 library */
		{r =-1; goto end;}
	if (jpeg_read_header(&cinfo,0)<JPEG_HEADER_OK)
		goto end;

	if (cinfo.out_color_space==JCS_GRAYSCALE)
		depth = 0x8088; /* 8-bit grayscale, least-significant-first */
	else if (cinfo.out_color_space==JCS_RGB)
	{ /* BITS_IN_JSAMPLE can go up to 12, and mgraf supports 10 bits/channel */
	  /* but the minimal library is compiled to produce only 24bpp */
		depth = 24;
	} else
	{ /* jpeg library automatically handles colorspace conversion */
	  /* ..but it may be a good idea to implement it natively */
		r =-2207;
		goto end;
	}
	r =-1; /* will be unchanged if corrupted data */
	jpeg_core_output_dimensions(&cinfo);

	if ((image=alloc_image((int16)depth,cinfo.output_width,
			output ? cinfo.output_height:1,
			NULL,NULL))==NULL)
		{r = cinfo.output_width|cinfo.output_height ? -2:0; goto end;}
	if ((d=lockPixels(image))==NULL)
	{ /* shouldn't happen; but if it does... */
		disposePixMap(image);
		goto end;
	}

	if (jpeg_start_decompress(&cinfo)<0)
		goto end;

	pitch = image->pitch;
	if (!output)
		depth = pitch = 0;
	for(; cinfo.output_scanline<cinfo.output_height;
			d+=pitch)
	{
		if (jpeg_read_scanlines(&cinfo,(uint8**)&d,1)<=0)
			break;
		if (depth==24)
			unpack_rgb24(d,cinfo.output_width);
	}

	(void)jpeg_finish_decompress(&cinfo);
	r = !!output;

end:	jpeg_destroy_decompress(&cinfo);
	if (output)
		*output = image;
	else if (image)
		disposePixMap(image);
	return r;
}
