/*
 mitab.c
  inverse table generation

   Author: Adam J. P.
  Revised: 2023.08.16

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#include "mgraf.h"

#include <string.h>

/* ---- code */

static void fillGrayTable(uint8 *d,
	const uint16 *s,
	int c,
	int16 *t)
{
	int16		*a;
	int		x, i, j;
	/* first, fill the remap */
	x = 0;
	do
		if (*(a=t+(s[x]>>6))<0) /* don't overwrite */
			*a = x;
	while(++x<c);
	/* second, bridge the gaps, rounding */
	x = j = 0;
	do
		if ((i=t[++x])>=0)
		{
			c = (x-j+1)>>1;
			memset(d+j,t[j],c);
			j+= c;
			memset(d+j,i,x-j);
			j = x;
		}
	while(x<1023);
	/* the rest is white, already filled by the caller */
}

static void formDirectGrays(uint16 *d,
	const cSpec *s,
	int c,
	int directct)
{
	uint32		u;

	if (directct)
		do
			*d++ = formGray(s->rgb.red,s->rgb.green,s->rgb.blue,u)>>1;
		while(++s,--c);
	else
		do
			d[s->value] = formGray(s->rgb.red,s->rgb.green,s->rgb.blue,u)>>1;
		while(++s,--c);
}

static int __cdecl compareDirect(const void *a,
	const void *b)
{
	return *(const uint32*)a-*(const uint32*)b;
}

/* low-accuracy constants */
/* for gray table case, 1024 entries, direct gray table and temporary buffer */
#define MIN_BUFFER_SIZE		(1024+256*2+2048)

/* high-accuracy constants */
#define ITAB_CMPRES		8
#define GI_RES			6
#define GI_ONE			(1<<GI_RES)
#define GI_MASK			(GI_ONE-1)
#define GI_TOLERANCE		(GI_ONE/4)
#define MAX_GI_LENGTH		(GI_TOLERANCE*3>>1)

typedef struct tc_entry/*(8)*/
{
	/* 'i' features two half/quarter-error bits (for color correction)
	  bit 14: dim green (postive error)
	  bit 15: gray (for matching ~zero/non-zero saturation)
	*/
	int16		r, g, b, i;
} tc_entry;

/* over 2KB in size */
typedef struct tc_table
{
	tc_entry	t[256];
	int16		c;
	uint8		g[(GI_ONE+MAX_GI_LENGTH+11)&~3]; /* gray intensity seek table */
} tc_table;

/* over 6KB in size */
typedef struct tc_split
{
	tc_table	r, g, b;
} tc_split;

static int __cdecl compare8_gray(const void *a,
	const void *b)
{ /* invert gray intensity so index has precedence, adding last-most black first */
	return ((((const tc_entry*)b)->i&GI_MASK<<8)^0xFF00)
		-((((const tc_entry*)a)->i&GI_MASK<<8)^0xFF00);
}

static void sort_tc_table(tc_table *t)
{
	qsort(t->t,t->c,sizeof(tc_entry),compare8_gray);

	{ /* gray intensity seek table with clipping */
		const tc_entry	*a = t->t;
		uint8		*w = t->g+GI_TOLERANCE; /* already zeroed w. calloc */
		int		i, j = 1, v, c;

		i = 0;
		do
			if ((c=(v=((const uint8*)&a->i)[!BIG_ENDIAN]&GI_MASK)-j)>0)
			{
				j = v;
				do
					*w++ = i;
				while(--c);
			}
		while(++a,++i<255);
		if ((c=-(int)w&3)!=0)
			do
				*w++ = 0xFE;
			while(--c);
		if ((c=sizeof(t->g)+t->g-w>>2)!=0)
			do
				*((int32*)w)++ = 0xFEFEFEFEu;
			while(--c);
	}
}

#define GEN_SEP_THRESHOLD	0xA0A
#define MAX_RES			5

int ctMakeITables(const cTab64 *ct,
	uint8 **itab,
	int bpp,
	int res,
	boolean forceColor,
	boolean accuracy)
{
	int		c;
	boolean		isColor;

	if (res<3)
		res = 3;
	else if (res>MAX_RES) /* 8 = 16MB */
		res = MAX_RES; /* can't handle 6 */
	if (accuracy)
	{
		//tc_entry	cya;
		tc_split	*ta;

		if ((ta=calloc(1,sizeof(*ta)))==NULL)
			return -2;
		isColor = 0, c = 0;
		//cya.i = 0;
		if (ct && (c=MIN(ct->count,256))!=0)
		{ /* build the RGB tables */
			const cSpec	*s = ct->colors;
			t_count		i = 0;//, cyr = IMAX(t_count);

			do
			{
				tc_entry	a;
				uint32		vr, vg, vb, v, g;

				if (s->flags&_cspec_flag_reserved)
					continue;
				vr = s->rgb.red, v = vr>>8;
				vg = s->rgb.green, vb = s->rgb.blue;
				/* ugly brown workaround to 15bpp red/green oscillation */
				/* additional red checks fix some game colors... a bit hacky */
				if (vr>vg && vg>vb && v<=106 && v!=40 && v!=63 && v!=95 && v!=90)
				{
					a.r = (vr-(vr>>res>>3))>>16-ITAB_CMPRES;
					a.g = (vg-(vg>>res>>3))>>16-ITAB_CMPRES;
					a.b = (vb-(vb>>res>>3))>>16-ITAB_CMPRES;
				} else
				{
					a.r = (vr-(vr>>res))>>16-ITAB_CMPRES;
					a.g = (vg-(vg>>res))>>16-ITAB_CMPRES;
					a.b = (vb-(vb>>res))>>16-ITAB_CMPRES;
				}
				v = (vr+vg+vg)<<2, g = vr+v, g+= vg, g+= vb, g+= vb; /* 5+9+2>>4 */
				/* gray intensity - make sure black & white gis are true */
				if ((v=g>>(4+8-GI_RES)&GI_MASK<<8)==0)
				{
					if (g)
						v = 1<<8;
				} else if (v==(GI_MASK<<8) && g!=0xFFFF0)
					v = (GI_ONE-2)<<8;
				v|= ct->flags&0x8000 ? i:s->value;
				if ((a.r^a.g)|(a.g^a.b))
					isColor = 1;
				else
					v|= 0x8000;
				if (a.g>a.r && a.g>a.b && a.g<=(a.r>a.b ? 3<<(ITAB_CMPRES-3)
						:1<<(ITAB_CMPRES-2)))
					v|= 0x4000;
				a.i = v;

				if (vr+GEN_SEP_THRESHOLD>vg && vr+GEN_SEP_THRESHOLD>vb)
					ta->r.t[ta->r.c++] = a;
				if (vg+GEN_SEP_THRESHOLD>vr && vg+(vg>vr+vr ? GEN_SEP_THRESHOLD*4/3:GEN_SEP_THRESHOLD)>vb)
					ta->g.t[ta->g.c++] = a;
				if (vb+GEN_SEP_THRESHOLD>vr && vb+GEN_SEP_THRESHOLD>vg)
					ta->b.t[ta->b.c++] = a;
				//g = (2<<ITAB_CMPRES)-a.b-a.g+a.r;
				//if (g<cyr)
				//	cyr = g, cya = a;
			} while(++s,++i<c);
		} else
		{ /* implied CT */
			memset(ta,0,sizeof(*ta));
			c = 1u<<PIN(bpp,1,8); /* ought be able to do 10 */
			ct = (cTab64*)&ta->g;
			((cTab64*)&ta->g)->flags = 0x8000, ((cTab64*)&ta->g)->count = c;
			ctMakeGrays(((cTab64*)&ta->g)->colors,c);
			((uint8*)&ta->r.t->i)[BIG_ENDIAN] = c-1;
		}
		sort_tc_table(&ta->r);
		if (!isColor && !forceColor)
		{ /* gray itable */
			uint16		*d;
			t_count		v;

			if ((d=realloc(*itab,1024+(ct->flags&0x8000 ? c*2:256*2)))==NULL)
				{free(ta); return -2;}
			*itab = (uint8*)d;
			memset(d,((const uint8*)&ta->r.t[c-1].i)[BIG_ENDIAN],1024);
			formDirectGrays(d+512,ct->colors,c,ct->flags&0x8000);
			v = ((const uint8*)&ta->r.t->i)[BIG_ENDIAN]; /* closest black entry */
			/* ta, used as scratch, needs to be >4KB */
			memset(ta,~0,2048);
			*(int16*)ta = v;
			fillGrayTable(*itab,512+(const uint16*)*itab,c,(int16*)ta);
			free(ta);
			return 0;
		}
	#if 0
		if (ta->g.c<256)
		{ /* as green has precedence over blue, make sure the cyan is in the green-dominant table */
			t_count		i;

			for(i=0; i<ta->g.c; ++i)
				if (ta->g.t[i].i==cya.i)
					goto ao;
			ta->g.t[ta->g.c++] = cya;
		}
	ao:
	#endif
		sort_tc_table(&ta->g);
		sort_tc_table(&ta->b);

		{ /* color itable */
			uint8		*d;
			unsigned	vr, vg, vb, cn, lm;

			if ((d=realloc(*itab,(1u<<res<<res<<res)
					+(ct->flags&0x8000 ? 512:c*2)))==NULL)
				{free(ta); return -2;}
			*itab = d, *d++ = (uint8)ta->r.t->i;
			vb = cn = 1u<<ITAB_CMPRES>>res, lm = vg = vr = 0;
			goto loop;
			do
			{
				vg = 0;
				do
				{
					lm = (((vr+vg+vg)<<2)+vg+vr)>>1, vb = 0;
					do
					{
						tc_table	*t;
						const tc_entry	*a, *e;
						t_count		i, y, z, gf;

					loop:	z = (lm+vb)>>(3+ITAB_CMPRES-GI_RES); /* gi index */
						/* it's best the leave this order alone */
						if (vr>=vg && vr>=vb)
							t = &ta->r;
						else if (vg>=vr && vg>=vb)
							t = &ta->g;
						else
							t = &ta->b;
						a = t->t+t->g[z], e = t->t+t->g[z+MAX_GI_LENGTH]; /* range */
						/* some counter-bias precalculations */
						gf = (vr-vg)|(vg-vb) ? 0:-0x8000;

						z = IMAX(z), i = 0;
						do
						{
							int16		x;

							if ((x=(t_count)vb-a->b)<0)
								x =-x, ++x;
							if (x>=z)
								continue;
							if ((y=(t_count)vr-a->r)<0)
								y =-y, y+= 2;
							x+= y;
							if ((y=(t_count)vg-a->g)<0)
							{
								y =-y, y+= 2;
								if ((t_count)(a->i&0x4000))
									y+= y>>1;
							}
							x+= y;
							if (x>=z)
								continue;
							if ((t_count)((gf^a->i)&0x8000) && vg>9
									&& !(t_count)(a->i&0x4000))
							{ /* should be a color/gray */
								x+= 2, x+= x>>2;
								if (x>=z)
									continue;
							}
							i = a->i;
							if ((z=x)==0)
								break;
						} while(++a<=e);
						*d++ = i;
					} while(vb+=cn,vb&=(1<<ITAB_CMPRES)-1);
				} while(vg+=cn,vg&=(1<<ITAB_CMPRES)-1);
			} while(vr+=cn,vr&=(1<<ITAB_CMPRES)-1);
			
			{ /* direct table */
				const cSpec	*s = ct->colors;
				uint8		*dt = d;

				if (ct->flags&0x8000)
				{
					do
						*((uint16*)d)++ = formDirect(&s->rgb,res,res);
					while(++s,--c);
				} else
				{
					memset(d,0,512);
					do
						((uint16*)d)[s->value] = formDirect(&s->rgb,res,res);
					while(++s,--c);
				}
				/* fix white to 0 if applicable */
				for(c=ta->g.c; c; )
					if (!(uint8)(vr=ta->g.t[--c].i))
					{
						if (!(~(vr>>8)&GI_MASK))
							dt[-1] = 0;
						break;
					}
				//if (!(ct->colors[0].flags&_cspec_flag_reserved)
				//		&& *(const uint32*)&ct->colors[0].rgb.red==~0ul
				//		&& ct->colors[0].rgb.blue==65535)
					
			}
		}
		free(ta);
	} else
	{
		uint32		sorted[256], v;
		int16		grays[1<<MAX_RES];
		const int16	*e;
		int16		*ta;
		unsigned	i, bm, gs, rs, tz, elc, elq, h;
		uint16		directct;

		elc = (rs=(gs=bm=1u<<res)<<res)<<res, --bm;
		elq = (2*bm+8-res)<<8; /* max color queue size */
		tz = elc+elq;
		//if (tz<MIN_BUFFER_SIZE)
		//	tz = MIN_BUFFER_SIZE;
		if ((ta=realloc(*itab,tz*2))==NULL)
			return -2;
		*itab = (uint8*)ta;
		memset(ta,~0,elc*2);
		memset(grays,~0,sizeof(grays));
		isColor = 0;
		h = 16-res;
		if (ct && (c=MIN(ct->count,256))!=0)
		{ /* build the RGB tables */
			uint32		*a = sorted;
			const cSpec	*s = ct->colors;
			unsigned	cr, cg, cb, j, k, x;

			i = 0, directct = ct->flags&0x8000;
			do
			{
				if (s->flags&_cspec_flag_reserved)
					continue; /* reserved */
				cr = s->rgb.red, cg = s->rgb.green, cb = s->rgb.blue;
				j = cr+cr+cg*5+cb;
				if ((k=j>>16-5+3&31)==0)
					k+= !!j;
				else if (k==31)
					k-= j!=0x7FFF8;
				x = directct ? i:s->value;
				if ((cr^cg)|(cg^cb))
					isColor = 1;
				else
				{
					//ta[((k<<r)|k)<<r|k] = i;
					grays[k] = x;
					//if (k!=31 && grays[k+1]<0)
					//	grays[k+(j>>10&1)] = x;
				}
				//cr+= (-(cr<0xFF80)&0x7F);
				//cg+= (-(cg<0xFF80)&0x7F);
				//cb+= (-(cb<0xFF80)&0x7F);
				/* inverting index so black is sorted last; white, first */
				*a++ = (((cr>>h<<res|cg>>h)<<res|cb>>h)<<5|k)<<8|(x^0xFF);
			} while(++s,++i<c+0u);
			c = a-sorted;
		} else
		{
			const cSpec	*s;
			uint32		*a, gm = rs+gs+1;
			int		cg;

			directct = (uint16)~0;
			c = 1u<<PIN(bpp,1,8); /* ought be able to do 10 */
			ct = (cTab64*)ta;
			((cTab64*)ta)->count = c;
			ctMakeGrays(((cTab64*)ta)->colors,c);
			i = 0, s = ct->colors, a = sorted;
			do
			{
				cg = s->rgb.green;
				*a++ = ((cg>>h)*gm<<5|(cg>>8&31))<<8|(i^0xFF);
			} while(++s,++i<c+0u);
			c = a-sorted;
		}

		qsort(sorted,c,4,compareDirect);
		*ta = (uint8)~*sorted; /* closest black entry */

		if (!isColor && !forceColor)
		{ /* gray itable */
			uint16		*d;

			elc = 1024;
			formDirectGrays(512+(uint16*)ta,ct->colors,c,directct);
			v = *ta;
			memset(ta,(uint8)~sorted[c-1],1024); /* fill with white */
			*(uint8*)ta = v;
			d = (512+256)+(uint16*)ta;
			memset(d,~0,2018);
			fillGrayTable((uint8*)ta,512+(const uint16*)ta,c,(int16*)d);
			res = 0;
			goto aend;
		}

		{ /* color itable */
			int16		*tq, *r, *w;
			int		j, x, k, rm = elc-rs, gm = rs-gs;

			/*
			    Fill indices first, tracking affected offsets to fill neighbors
			  in a diamond spread.  A circular queue is used to minimize memory
			  requirements.  To speed up this process, Quickdraw used fake
			  'boundary' indices to eliminate displacement checks, as well as
			  negate channel registers only once per iteration, resulting a
			  'dithered' ordering.  (This implementation is positive-first only.)
			*/
			i = 0, w = tq = elc+ta;
			do
			{
				v = sorted[i], x = v>>8+5;
				if (ta[x]<0)
					*w++ = x, ta[x] = (uint8)~v;
			} while(++i<c+0u);
			r = tq, e = tz+ta;
			do
			{
				if (r>=e)
					r = tq;
				v = *r++, x = ta[v];
				i = v&bm;
				if (i!=bm && ta[j=v+1]<0)
				{
					ta[j] = x;
					if (w>=e)
						w = tq;
					*w++ = j;
				}
				if (i && ta[j=v-1]<0)
				{
					ta[j] = x;
					if (w>=e)
						w = tq;
					*w++ = j;
				}
				i = v&gm;
				if (i!=gm && ta[j=v+gs]<0)
				{
					ta[j] = x;
					if (w>=e)
						w = tq;
					*w++ = j;
				}
				if (i && ta[j=v-gs]<0)
				{
					k = 1;
					ta[j] = x;
					if (w>=e)
						w = tq;
					*w++ = j;
				}
				if (v<rm+0u && ta[j=v+rs]<0)
				{
					ta[j] = x;
					if (w>=e)
						w = tq;
					*w++ = j;
				}
				if ((j=v-rs)>=0 && ta[j]<0)
				{
					ta[j] = x;
					if (w>=e)
						w = tq;
					*w++ = j;
				}
			} while(w!=r);

			w = r = ta, e = (int16*)((int8*)ta+elc);
			do
				*((uint8*)w)++ = (uint8)*r++;
			while(w<e);

			/* direct table */
			memset(w,0,256*2);
			i = 0;
			do
				v = sorted[i], w[(uint8)~v] = v>>8+5;
			while(++i<c+0u);
		}
		/* make sure grays are grays */
		rs+= gs, ++rs;
		h = res+res;
	#if 0
		for(i=0; i<32; ++i)
			if (grays[i]>=0)
			{
				int x = grays[i], k = e[x];
				printf(" gray %d:%02X:%d,%d,%d\n",
					i,x,k>>h,k>>res&bm,k&bm);
			}
	#endif
		/* unfortunately need to fix red-dominant, others to beat the alternate above */
		for(i=2,v=rs+rs; i<bm; v+=rs,++i)
		{
			int x = ((uint8*)ta)[v], cg, cb, k = e[x];

			cg = k>>res&bm, cb = k&bm;
			if (((k>>h)-cg)|((cb-cg)&~1))
			{
			//	printf("gray %d:%02X:%d,%d,%d\n",
			//		i,x,cr,k>>res&bm,cb);
				x =~0;
				if (grays[i]>=0)
					x = grays[i];
				else if (grays[i+1]>=0)
					x = grays[i+1];
				else
					for(k=i-1; k>0; --k)
						if (grays[k]>=0)
							{x = grays[k]; break;}
				if (x>=0)
				{
					k = e[x];
				//	printf(" fixed with %02X:%d,%d,%d\n",
				//		x,k>>h,k>>res&bm,k&bm);
					((uint8*)ta)[v] = x;
				}
			}
		}
	aend:	if ((ta=realloc(ta,elc+256*2))!=NULL)
			*itab = (uint8*)ta;
	}
	return res;
}
