/*
 mrgn.c

   Revised: 2024.04.14
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.05.15
  Split from mgraphic.
2014.05.23/4
  Added rectInRgn (tested, working), coordinate removal in rgnCompact (untested).
2015.05.06
  Added rectRgn (faster than two rgnXorLines).
2015.05.07
  rgnXorLine fills in IMAX for line endings (instead of copying old value).
2016.05.30
  Added emptyRgn and copyRgn.
  Added handling of implied rect regions to rgnXorLine.
2016.06.14
  Fixed rgnCompact and rgnOffsetXY.
*/

#define __MGRAF_INTERNAL

#include <stdlib.h>
#include <string.h>

#include "mgraf.h"
#include "ftrig.h"

/* ---- macros */

#if QD_ROUNDED_LINES
 #define rslopdiv lslopdiv
#endif

#define skipLines(s,e,h,l) \
	do \
		do \
			if(s>=e) \
				if(h==1) \
					break; \
				else \
					l; \
		while(*s++!=RGN_STOP); \
	while(--h);

/* ---- code */

static SLRgn *rgnAppend(SLRgn *rgn,
	const int16 *a,
	int32 az)
{
	void		*p;
	long		z;

	z = rgn->size+az;
	if ((p=realloc(rgn,z))==NULL)
		{free(rgn); return NULL;}
	rgn = p, rgn->size = z;
	memcpy((int8*)rgn+z-az,a,az);
	return rgn;
}

SLRgn *newRgn(boolean open)
{
	SLRgn		*rgn;

	if ((rgn=malloc(RECT_RGN_SIZE))==NULL)
		return NULL;
	rgn->bBox.top = rgn->bBox.left = open ? IMIN(rgn->bBox.left):IMAX(rgn->bBox.left);
	rgn->bBox.bottom = rgn->bBox.right = open ? IMAX(rgn->bBox.left):IMIN(rgn->bBox.right);
	rgn->size = RECT_RGN_SIZE;
	return rgn;
}

int disposeRgn(SLRgn **rrgn)
{
	SLRgn		*p;

	if (!rrgn)
		return -1;
	if ((p=*rrgn)!=0)
		free(p);
	*rrgn = NULL;
	return 0;
}

int copyRgn(SLRgn **rrgn,
	const SLRgn *srgn)
{
	SLRgn		*p;

	if (!rrgn)
		return -1;
	p = *rrgn;
	if (emptyRgn(srgn,1))
	{
		free(p);
		*rrgn = NULL;
		return 0;
	}
	if ((p=realloc(p,srgn->size))==NULL)
		return -2;
	*rrgn = p;
	/*if (!p || !srgn || srgn->size<12 || srgn->size>8*1048576)
		__debugbreak();*/
	memcpy(p,srgn,srgn->size);
	return 1;
}

int emptyRgn(const SLRgn *rgn,
	uint8 boundsOnly)
{
	const int16	*s, *e;
	int32		z;
	int		r, c, o, h, bl, br;

	if (!rgn)
		return 1;
	if ((z=rgn->size)&1)
		return -1;
	if (rgn->size<RECT_RGN_SIZE || rgn->bBox.bottom<=rgn->bBox.top)
		return 2;
	if (boundsOnly)
		return 0;
	s = rgn->x, e = (int16*)(z+(int8*)rgn);
	h = RECTANGLE_HEIGHT(&rgn->bBox);
	r = 3;
	bl = rgn->bBox.left;
	br = rgn->bBox.right;
	do
	{
		if (s>=e)
			return h==1 ? r:-1;
		while((o=*s++)!=RGN_STOP)
		{
			if (s>=e)
				return h==1 ? r:-1;
			c = *s++;
			if ((o=MIN(c,br)-MAX(o,bl))<0)
				return -1;
			else if (o)
				return 0; //r = 0;
			if (c==RGN_STOP) /* odd lines are valid */
				break;
		}
	} while(--h);
	return r;
}

int rectRgn(SLRgn **rrgn,
	int16 left,
	int16 top,
	int16 right,
	int16 bottom,
	boolean fill)
{
	SLRgn		*rgn;
	int32		z;
	int		h, fw;

	if (!rrgn)
		return -1;
	rgn = *rrgn, fw = right>left, h = bottom-top;
	if (fill)
	{
		if (h<=0)
			h = 0, fill = 0;
	} else
	{
		if (h<=0 || !fw)
		{
			free(rgn);
			*rrgn = NULL;
			return 0;
		}
		h = 0;
	}
	z = h;
	if (fw)
	{
		z+= h;
		if (right!=RGN_STOP)
			z+= h;
	}
	z = z*sizeof(*rgn->x)+OFFSET(SLRgn,x);
	if ((!rgn || rgn->size!=z) && (rgn=realloc(rgn,z))==NULL)
		return -2;
	*rrgn = rgn;
	rgn->size = z;
	rgn->bBox.left = left;
	rgn->bBox.top = top;
	rgn->bBox.right = right;
	rgn->bBox.bottom = bottom;
	if (fill)
	{ /* unlike QD, all lines have to be spelled out beyond one rectangle */
		int16		*d = rgn->x;

		if (fw)
			if (right==RGN_STOP)
			{
				uint32		v = (uint16)left<<16*BIG_ENDIAN
					|RGN_STOP<<16*!BIG_ENDIAN;

				do
					*(uint32*)d = v;
				while(d+=2,--h);
				return 3;
			} else
			{
				do
					*d = left, d[1] = right, d[2] = RGN_STOP;
				while(d+=3,--h);
				return 4;
			}
		if ((z=h>>1)!=0)
			do
				*(uint32*)d = RGN_STOP*0x10001u;
			while(d+=2,--z);
		if (h&1)
			*d = RGN_STOP;
		return 2;
	}
	return 1;
}

static void tangentPt(pt2dq *pt,
	pt2dq cp,
	ufixed rx,
	ufixed ry,
	ifract dx,
	ifract dy)
{
	ufixed		m;
	ufract		adx = ABS(dx), ady = ABS(dy);

	if (ady>adx)
	{
		pt->y = dy<0 ? subu(cp.y,ry):addu(cp.y,ry);
		m = ufixmuldiv(rx,adx,ady);
		pt->x = dx<0 ? subu(cp.x,m):addu(cp.x,m);
	} else
	{
		pt->x = dx<0 ? subu(cp.x,rx):addu(cp.x,rx);
		m = ufixmuldiv(ry,ady,adx);
		pt->y = dy<0 ? subu(cp.y,m):addu(cp.y,m);
	}
}

int arcRgn(SLRgn **rrgn,
	const rect2dq *ov,
	int16 ctop,
	int16 cbot,
	ufract startAngle,
	ufract arcAngle,
	boolean addCenter,
	SLRgn **tanRgn)
{
	pt2dq		cp;
	rect16		cl;
	SLRgn		*rgn, *trgn;
	int16		*d;
	uint64		dy0;
	int64		y0;
	ufixed		w, h;
	ifixed		rx, ry;
	int32		z;
	int		o, c, j;
	int16		btop;

	//if (!rrgn)
	//	return -1;
	if (!ov || ov->top>=ov->bottom || ov->left>=ov->right || !arcAngle)
		goto emp;
	/* rarely, some narrow fractions/large ratios manage to jutt a lot */
	cl.left = (int16)(ov->left>>16);
	/* without decrement, rounded full-width edge pixels may show up */
	/* with width check, decrement should not overflow */
	cl.right = (int16)((ov->right-1)>>16);
	if (cl.right<=cl.left)
		goto emp;
	/* cx +/- w~/(1-(y/h)**2) */
	w = ov->right-ov->left, h = ov->bottom-ov->top;
	rx = (w>>1)+(w&1), ry = (h>>1)+(h&1); /* any round not toward hollow left */
	cp.x = ov->left+rx, cp.y = ov->top+ry;
	btop = (int16)(ov->top>>16);
	if (!tanRgn)
	{
		trgn = NULL, tanRgn = &trgn;
		if (!rrgn)
			return 0;
	}
	if (arcAngle!=~0u && !*tanRgn)
	{
		pt2dq		t0, t1, sp, ep;
		ifract		cs, ss, ce, se;
		ufract		theta = startAngle, endAngle;
		int		ttop, tbot;
		int16		bbot = (int16)(ov->bottom>>16);
		/* translate the from starting and ending angles to get the edge points */
		cs = fractcos(theta);
		ss =-fractsin(theta);
		sp.x = cs<0 ? subu(cp.x,ufract30mul(1-cs,rx))
			:addu(cp.x,ufract30mul(1+cs,rx));
		sp.y = ss<0 ? subu(cp.y,ufract30mul(1-ss,ry))
			:addu(cp.y,ufract30mul(1+ss,ry));
		endAngle = theta+arcAngle;
		ce = fractcos(endAngle);
		se =-fractsin(endAngle);
		ep.x = ce<0 ? subu(cp.x,ufract30mul(1-ce,rx))
			:addu(cp.x,ufract30mul(1+ce,rx));
		ep.y = se<0 ? subu(cp.y,ufract30mul(1-se,ry))
			:addu(cp.y,ufract30mul(1+se,ry));
		/* hit the boundary edges so all applicable pixels intersect */
		ttop = tbot = sp.y>>16;
		o = ep.y>>16;
		if (o<ttop)
			ttop = o;
		if (o>tbot)
			tbot = o;
		/* compute the tangent region, encircling from the starting angle */
		tangentPt(&t0,cp,rx,ry,cs,ss);
		(void)rgnXorLine(tanRgn,btop,bbot,sp,t0);
		while(arcAngle>0)
		{
			ufract		turn = MIN(arcAngle,1<<29); /* max. 45deg */

			theta+= turn, arcAngle-= turn;
			/* add absolute octant */
			switch(theta>>29)
			{
			case 0: t1.x = ov->right, t1.y = cp.y; break; /* 0 */
			case 1: t1.x = ov->right, t1.y = ov->top; break; /* 45 */
			case 2: t1.x = cp.x, t1.y = ov->top, ttop = btop; break; /* 90 */
			case 3: t1.x = ov->left, t1.y = ov->top; break; /* 135 */
			case 4: t1.x = ov->left, t1.y = cp.y; break; /* 180 */
			case 5: t1.x = ov->left, t1.y = ov->bottom; break; /* 225 */
			case 6: t1.x = cp.x, t1.y = ov->bottom, tbot = bbot; break; /* 270 */
			default: t1.x = ov->right, t1.y = ov->bottom; break; /* 315 */
			}
			(void)rgnXorLine(tanRgn,btop,bbot,t0,t1);
			t0 = t1;
		}
		/* finish final tangent lines */
		tangentPt(&t1,cp,rx,ry,ce,se);
		(void)rgnXorLine(tanRgn,btop,bbot,t0,t1);
		(void)rgnXorLine(tanRgn,btop,bbot,t1,ep);
		if (addCenter)
		{
			(void)rgnXorLine(tanRgn,btop,bbot,ep,cp), t1 = cp;
			o = cp.y>>16;
			if (o<ttop)
				ttop = o;
			if (o>tbot)
				tbot = o;
		} else
			t1 = ep;
		(void)rgnXorLine(tanRgn,btop,bbot,t1,sp);
		if (!*tanRgn)
			goto emp;
		if (!rrgn)
			return 2;
		if (ttop>ctop)
			ctop = (int16)ttop;
		if (tbot<cbot)
			cbot = (int16)tbot;
	}
	/* find vertical range for arc */
	c = btop;
	o = (ov->bottom-1)>>16;
	cl.top = MAX(ctop,c);
	cl.bottom = MIN(cbot,o);
	if (cl.bottom<=cl.top)
	{
	emp:	if (rrgn && *rrgn)
			{free(*rrgn); *rrgn = NULL;}
		if (tanRgn==&trgn && trgn)
			free(trgn);
		return 0;
	}
	/* set up y increment & position */
	dy0 = rincdiv(&y0,ov->top,ov->bottom,(uint16)(cl.top-c));
	int64subhl(y0,0x40000000,0);
	j = cl.bottom-cl.top;
	z = OFFSET(SLRgn,x)+j*(3*sizeof(int16));
	if ((rgn=realloc(*rrgn,z))==NULL)
	{
		if (tanRgn==&trgn && trgn)
			free(trgn);
		return -2;
	}
	*rrgn = rgn;
	rgn->size = z;
	rgn->bBox = cl;
	d = rgn->x;
	do
	{
		o = rarcmul((int32)int64hi(y0),w);
		c = addu(cp.x,o)>>16, o = subu(cp.x,o)>>16;
		if (c>cl.right)
			c = cl.right;
		if (o<cl.left)
			o = cl.left;
	//	*d++ = IMIN(int16);
		if (c>o) /* add points */
		{
			*d++ = (int16)o;
			if (c!=RGN_STOP)
				*d++ = (int16)c;
		}
		*d++ = RGN_STOP;
		(void)int64add(y0,dy0);
	} while(--j);
	rgn->size = (int32)((int8*)d-(int8*)rgn);
	*rrgn = rgnCompact(rgn,0);
	if (*tanRgn)
	{
		j = sectRgn(rrgn,*tanRgn);
		if (tanRgn==&trgn)
			free(trgn);
		return j;
	}
	return !!*rrgn;
}

/* xor-rgn: blindly merge x-points */

int xorRgn(SLRgn **rrgn,
	const SLRgn *brgn)
{
	int16		b[96];
	SLRgn		*drgn, *argn;
	const int16	*sa, *sb, *ea, *eb;
	int16		*d;
	int32		za, zb;
	int		y, r, atop, btop, abot, bbot, cbot;

	if (!rrgn)
		return -1;
	if (!brgn || brgn->bBox.bottom<=brgn->bBox.top)
		return 0; /* no addition */
	if ((zb=brgn->size)<RECT_RGN_SIZE || zb&1)
		return -1;
	if ((argn=*rrgn)==NULL || (za=argn->size)<RECT_RGN_SIZE
			|| argn->bBox.bottom<=argn->bBox.top)
		return copyRgn(rrgn,brgn);
	if (za&1)
		return -1;
	if (za==RECT_RGN_SIZE)
	{
		if ((r=rectRgn(rrgn,argn->bBox.left,argn->bBox.top,argn->bBox.right,
				argn->bBox.bottom,1))<0)
			return r;
		argn = *rrgn;
	}
	if (zb==RECT_RGN_SIZE)
	{
		pt2dq		p0, p1;

		p0.y = brgn->bBox.top<<16, p1.y = brgn->bBox.bottom<<16;
		p0.x = p1.x = brgn->bBox.left<<16;
		r = rgnXorLine(rrgn,IMIN(int16),IMAX(int16),p0,p1);
		p0.x = p1.x = brgn->bBox.right<<16;
		r|= rgnXorLine(rrgn,IMIN(int16),IMAX(int16),p0,p1);
		return r;
	}

	/* grow the original enough to avoid fragmentation */
	if ((argn=realloc(argn,za+zb-OFFSET(SLRgn,x)))==NULL)
		argn = *rrgn;
	else
		*rrgn = argn;
	if ((drgn=malloc(RECT_RGN_SIZE))==NULL)
		return -2;
	drgn->size = RECT_RGN_SIZE;
	drgn->bBox = argn->bBox;
	UnionRect16(&drgn->bBox,&brgn->bBox);
	sa = argn->x, sb = brgn->x;
	ea = (int16*)((int8*)argn+za);
	eb = (int16*)((int8*)brgn+zb);
	atop = argn->bBox.top, btop = brgn->bBox.top;
	abot = argn->bBox.bottom, bbot = brgn->bBox.bottom;
	cbot = drgn->bBox.bottom;
	d = b;
	for(y=drgn->bBox.top; y<cbot; ++y)
	{
		int16		va, vb;

		if (d>b+GET_ARRAY_COUNT(b)-2)
		{
			if ((drgn=rgnAppend(drgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
				goto merr;
			d = b;
		}
		if (y<atop || y>=abot)
		{
			if (y>=btop && y<bbot)
			{ /* b-only line */
				if (sb[1]<*sb && *sb!=RGN_STOP)
					goto bad;
				goto cpyb;
			}
			goto el;
		}
		if (sa>=ea)
		{
			free(drgn);
			return -3;
		}
		if (sb<eb && y>=btop && y<bbot)
		{ /* both */
			if (sb[1]<*sb && *sb!=RGN_STOP)
			{
			bad:	free(drgn);
				return -1;
			}
			while(sa<ea && (va=*sa++)!=RGN_STOP)
			{ /* insert, sorted */
				for(;;)
				{
					if (sb>=eb || (vb=*sb)==RGN_STOP)
						{++sb; goto ala;}
					if (vb>=va)
						break;
					++sb, *d++ = vb;
					if (d>b+GET_ARRAY_COUNT(b)-2)
					{
						if ((drgn=rgnAppend(drgn,b,
								(int32)((int8*)d-(int8*)b)))==NULL)
							goto merr;
						d = b;
					}
				}
				*d++ = va;
				if (d>b+GET_ARRAY_COUNT(b)-2)
				{
					if ((drgn=rgnAppend(drgn,b,
							(int32)((int8*)d-(int8*)b)))==NULL)
						goto merr;
					d = b;
				}
			}
		cpyb:	while(sb<eb && (vb=*sb++)!=RGN_STOP)
			{
				*d++ = vb;
				if (d>b+GET_ARRAY_COUNT(b)-2)
				{
					if ((drgn=rgnAppend(drgn,b,
							(int32)((int8*)d-(int8*)b)))==NULL)
						goto merr;
					d = b;
				}
			}
		} else
			while(sa<ea && (va=*sa++)!=RGN_STOP)
			{
			ala:	*d++ = va;
				if (d>b+GET_ARRAY_COUNT(b)-2)
				{
					if ((drgn=rgnAppend(drgn,b,
							(int32)((int8*)d-(int8*)b)))==NULL)
						goto merr;
					d = b;
				}
			}
	el:	*d++ = RGN_STOP;
	}
	(void)copyRgn(rrgn,drgn);
	free(drgn);
	if ((*rrgn=rgnAppend(*rrgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
	{
	merr:	*rrgn = NULL;
		return -2;
	}
	return !!(*rrgn=rgnCompact(*rrgn,0));
}

/* unionRgn: merge positive spaces */

/* * * NOT FULLY TESTED * * */
int unionRgn(SLRgn **rrgn,
	const SLRgn *brgn)
{
	int16		b[96], rr[3];
	SLRgn		*drgn, *argn;
	const int16	*sa, *sb, *ea, *eb;
	int16		*d;
	size_t		za, zb;
	int		y, r, atop, btop, abot, bbot, cbot;

	if (!rrgn)
		return -1;
	if (!brgn || brgn->bBox.bottom<=brgn->bBox.top)
		return 0;
	if ((signed)(zb=brgn->size)<RECT_RGN_SIZE || zb&1)
		return -1;
	if ((argn=*rrgn)==NULL || (signed)(za=argn->size)<RECT_RGN_SIZE
			|| argn->bBox.bottom<=argn->bBox.top)
		return copyRgn(rrgn,brgn);
	if (za&1)
		return -1;
	if (za==RECT_RGN_SIZE)
	{
		if ((r=rectRgn(rrgn,argn->bBox.left,argn->bBox.top,argn->bBox.right,
				argn->bBox.bottom,1))<0)
			return r;
		argn = *rrgn;
		za = argn->size; /* 2024.04.02 */
	}

	/* grow the original enough to avoid fragmentation */
	if ((argn=realloc(argn,za+zb-OFFSET(SLRgn,x)))==NULL)
		argn = *rrgn;
	else
		*rrgn = argn;
	if ((drgn=malloc(RECT_RGN_SIZE))==NULL)
		return -2;
	drgn->size = RECT_RGN_SIZE;
	drgn->bBox = argn->bBox;
	UnionRect16(&drgn->bBox,&brgn->bBox);
	sa = argn->x, sb = brgn->x;
	ea = (int16*)((int8*)argn+za);
	eb = (int16*)((int8*)brgn+zb);
	if (zb==RECT_RGN_SIZE)
	{
		sb = rr, eb = rr+2;
		rr[0] = brgn->bBox.left;
		rr[1] = brgn->bBox.right;
		rr[2] = RGN_STOP;
		zb = (size_t)(-2*sizeof(int16)); /* to reset stack-based line */
	} else
		zb = 0;
	atop = argn->bBox.top, btop = brgn->bBox.top;
	abot = argn->bBox.bottom, bbot = brgn->bBox.bottom;
	cbot = drgn->bBox.bottom;
	d = b;
	for(y=drgn->bBox.top; y<cbot; ++y)
	{
		int16		va, vb, ua, ub;

		if (d>b+GET_ARRAY_COUNT(b)-4)
		{
			if ((drgn=rgnAppend(drgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
				goto merr;
			d = b;
		}
		if (y<atop || y>=abot)
		{
			if (y>=btop && y<bbot) /* b-only line */
				goto cpb;
			goto el;
		}
		if (sa>=ea)
		{
			free(drgn);
			return -3;
		}
		if (y>=btop && y<bbot)
		{ /* both */
			for(va=vb=ua=ub=RGN_STOP;;)
			{
				if (ua==RGN_STOP && (sa>=ea || (va=*sa++)==RGN_STOP))
					if (ub!=RGN_STOP)
					{
						*d++ = vb, vb = ub;
						goto lpb;
					} else
						goto cpb;
				if (ub==RGN_STOP)
				{
					if (sb>=eb || (vb=*sb++)==RGN_STOP)
					{
						sb = (const int16*)((int8*)sb+zb);
						if (ua!=RGN_STOP)
							*d++ = va, va = ua;
						goto lpa;
					}
					ub = sb>=eb ? RGN_STOP:*sb++;
				}
				if (ua==RGN_STOP)
					ua = sa>=ea ? RGN_STOP:*sa++;
				if (d>b+GET_ARRAY_COUNT(b)-4)
				{
					if ((drgn=rgnAppend(drgn,b,
							(int32)((int8*)d-(int8*)b)))==NULL)
						goto merr;
					d = b;
				}
				d+= 2;
				if (va<=ub)
				{
					d[-2] = va;
					if (ua==RGN_STOP || ub==RGN_STOP)
						break;
					if (vb<va)
						d[-2] = vb;
					if (ua>=vb && ub>ua) /* isect/adj a-b */
					{
						d[-1] = ub;
						ub = RGN_STOP;
					} else
						d[-1] = ua;
					ua = RGN_STOP;
				} else
				{
					d[-2] = vb;
					if (ua==RGN_STOP || ub==RGN_STOP)
						break;
					if (va<vb)
						d[-2] = va;
					if (ua<=vb && ua>ub) /* isect/adj b-a */
					{
						d[-1] = ua;
						ua = RGN_STOP;
					} else
						d[-1] = ub;
					ub = RGN_STOP;
				}
			}
		cpb:	while(sb<eb && (vb=*sb++)!=RGN_STOP)
			{
			lpb:	*d++ = vb;
				if (d>b+GET_ARRAY_COUNT(b)-4)
				{
					if ((drgn=rgnAppend(drgn,b,
							(int32)((int8*)d-(int8*)b)))==NULL)
						goto merr;
					d = b;
				}
			}
			sb = (const int16*)((int8*)sb+zb);
		} else
			while(sa<ea && (va=*sa++)!=RGN_STOP)
			{
			lpa:	*d++ = va;
				if (d>b+GET_ARRAY_COUNT(b)-4)
				{
					if ((drgn=rgnAppend(drgn,b,
							(int32)((int8*)d-(int8*)b)))==NULL)
						goto merr;
					d = b;
				}
			}
	el:	*d++ = RGN_STOP;
	}
	(void)copyRgn(rrgn,drgn);
	free(drgn);
	if ((*rrgn=rgnAppend(*rrgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
	{
	merr:	*rrgn = NULL;
		return -2;
	}
	return !!(*rrgn=rgnCompact(*rrgn,0));
}

int diffRgn(SLRgn **rrgn,
	const SLRgn *argn,
	const SLRgn *brgn)
{
	SLRgn		*crgn = NULL;

	if (!rrgn)
		return -1;
	copyRgn(&crgn,argn);
	xorRgn(&crgn,brgn);
	sectRgn(&crgn,argn);
	copyRgn(rrgn,crgn);
	free(crgn);
	return *rrgn!=0;
}

/* sectRgn: keep only overlapping positive spaces */

int sectRgn(SLRgn **rrgn,
	const SLRgn *brgn)
{
	int16		b[96];
	SLRgn		*drgn, *argn;
	const int16	*sa, *sb, *ea, *eb;
	int16		*d;
	int32		za, zb;
	int		y, r, atop, btop, abot, bbot, cbot, bx0, bx1;

	if (!rrgn)
		return -1; /* idiot proof */
	r = 0;
	if ((argn=*rrgn)==NULL)
		return 0;
	drgn = NULL;
	if (!brgn)
		goto empt;
	atop = argn->bBox.top, abot = argn->bBox.bottom;
	btop = brgn->bBox.top, bbot = brgn->bBox.bottom;
	cbot = MIN(abot,bbot);
	if (!SectRect16(&argn->bBox,&brgn->bBox))
		goto empt;
	if ((zb=brgn->size)<RECT_RGN_SIZE || (za=argn->size)<RECT_RGN_SIZE
			|| (zb|za)&1)
		return -1;
	if (zb==RECT_RGN_SIZE)
	{
		if (za<OFFSET(SLRgn,x)+3)
		{
			argn->size = RECT_RGN_SIZE;
			return 1;
		}
		if ((r=argn->bBox.top-atop)>0 || abot!=argn->bBox.bottom)
		{ /* crop vertical */
			sa = argn->x, ea = (int16*)((int8*)argn+za)-1;
			if (r>0)
			{
				skipLines(sa,ea,r,return -3);
				memcpy(argn->x,sa,(int8*)ea-(int8*)sa+2);
				r = (int32)((int8*)sa-(int8*)argn->x);
				za-= r;
				sa = argn->x;
				atop = argn->bBox.top;
			}
			if (abot!=argn->bBox.bottom && (r=argn->bBox.bottom-atop)>0)
			{
				skipLines(sa,ea,r,return -3);
				za = (int32)((int8*)sa-(int8*)argn);
				abot = argn->bBox.bottom;
			}
			if (argn->size!=za)
			{
				void		*p;

				argn->size = za;
				if ((p=realloc(argn,za))!=NULL)
					*rrgn = argn = p;
			}
		}
		return 1;
	}
	if (za==RECT_RGN_SIZE)
	{
		if ((r=rectRgn(rrgn,argn->bBox.left,argn->bBox.top,argn->bBox.right,
				argn->bBox.bottom,1))<0)
			return r;
		argn = *rrgn;
	}
	/* grow the original enough to avoid fragmentation */
	if ((argn=realloc(argn,MAX(za,zb)))==NULL)
		argn = *rrgn;
	else
		*rrgn = argn;
	if ((drgn=newRgn(0))==NULL)
		return -2;
	sa = argn->x, sb = brgn->x;
	ea = (int16*)((int8*)argn+za);
	eb = (int16*)((int8*)brgn+zb);
	drgn->bBox = argn->bBox;

	y = drgn->bBox.top;
	/* start a rgn */
	if ((sa=_rgnGotoLine(sa,ea,y-atop,abot-atop))==NULL)
		goto rerr; /* leave broken rgn */
	/* start b rgn */
	if ((sb=_rgnGotoLine(sb,eb,y-btop,bbot-btop))==NULL)
		{rerr: r = -3; goto err;} /* leave broken rgn */
	if (sb>=eb)
	{ /* empty */
	empt:	free(argn);
		*rrgn = NULL;
	err:	if (drgn)
			free(drgn);
		return r;
	}
	bx0 = *sb++;
	bx1 = sb>=eb ? IMAX(bx1):*sb++;
	d = b;
	for(; y<cbot; ++y)
	{
		int		ax0, ax1;

		if (sa>=ea)
			goto end; /* check for error */
		for(; (ax0=*sa++)!=RGN_STOP; ++sa)
		{
			if (sa>=ea)
				{++y; goto end;} /* check for error */
			ax1 = *sa;
			if (ax1<=bx0) /* a behind b */
				continue;
			do
			{
				while(ax0>=bx1) /* a ahead of b */
				{
					bx1 = IMAX(bx1);
					if (sb>=eb)
						{++y; goto end;}
					if (*sb==RGN_STOP)
					{ /* end of spans for b */
						bx0 = IMAX(bx0);
						do
							if (sa>=ea)
								{++y; goto end;}
						while(*sa++!=RGN_STOP);
						goto el;
					}
					bx0 = *sb++;
					if (sb>=eb || *sb==RGN_STOP)
						break;
					bx1 = *sb++;
				}
				if (ax0<bx0)
					ax0 = bx0;
				r = MIN(ax1,bx1);
				if (r>ax0)
				{
					*d = ax0, d[1] = r, d+= 2;
					if (d>b+GET_ARRAY_COUNT(b)-3)
					{
						if ((drgn=rgnAppend(drgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
							goto merr;
						d = b;
					}
				}
			} while(ax0=r,ax0<ax1);// && bx0<RGN_STOP);
		}
	el:	*d++ = RGN_STOP;
		if (d>b+GET_ARRAY_COUNT(b)-3)
		{
			if ((drgn=rgnAppend(drgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
				goto merr;
			d = b;
		}
		bx1 = IMAX(bx1);
		do
			if (sb>=eb)
				{++y; goto end;}
		while(*sb++!=RGN_STOP);
		if (sb>=eb)
			goto end;
		if (*sb==RGN_STOP)
			bx0 = IMAX(bx1);
		else
		{
			bx0 = *sb++;
			bx1 = sb>=eb ? IMAX(bx1):*sb++;
		}
	}
end:	drgn->bBox.bottom = y;

	(void)copyRgn(rrgn,drgn);
	free(drgn);
	if (*rrgn && (*rrgn=rgnAppend(*rrgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
	{
	merr:	*rrgn = NULL;
		return -2;
	}
/*	if (!_heap_validate(NULL))
		__debugbreak();*/
	return !!(*rrgn=rgnCompact(*rrgn,0));
}

int rgnXorLine(SLRgn **rrgn,
	int16 ctop,
	int16 cbot,
	pt2dq p0,
	pt2dq p1)
{
	SLRgn		*rgn;
	const int16	*f, *e;
	int16		*a, *s;
	int64		x, dx;
	int32		o, c, z, newa, newb, newc, nse;
	int32		y0i, y1i, x0i, x1i;

	if (!rrgn)
		return -1;
	q2dsort2(p0,p1);
	if ((y0i=p0.y>>16)<ctop)
		y0i = ctop;
	if ((y1i=p1.y>>16)>cbot)
		y1i = cbot;
	if ((c=y1i-y0i)<=0)
		return 0; /* no addition */
	nse = 0;
	if ((rgn=*rrgn)==NULL || (z=rgn->size)<RECT_RGN_SIZE
			|| rgn->bBox.bottom<=rgn->bBox.top)
	{ /* empty */
		ctop = cbot = (int16)y0i;
		goto rect;
	} else if (z&1)
		return -1;
	else if (z==RECT_RGN_SIZE)
	{ /* implied rect */
		int16		left, right;

		ctop = rgn->bBox.top, cbot = rgn->bBox.bottom;
	rect:	//left = right = 0;
		left = right = p0.x>>16;
		if (rgn)
			left = rgn->bBox.left, right = rgn->bBox.right;
		if ((z=rectRgn(rrgn,left,ctop,right,cbot,1))<0)
			return z;
		rgn = *rrgn, z = rgn->size;
	} else
	{ /* a necessary part of the stop code is making sure the stop exists at the end */
		e = (int16*)(rgn->size+(int8*)rgn);
		nse = (e[-1]!=RGN_STOP)*2;
	}
	/* performed backward from end for efficiency */
	dx = rslopdiv(&x,p0.x,p0.y,p1.x,p1.y,(uint16)(y1i-(p0.y>>16)));
	o = z;
	/* from-end insertion requires the new size be adequate */
	z+= c*sizeof(*a); /* inserted positions */
	if ((newa=rgn->bBox.top-y0i)>0) /* new lines above */
		z+= newa*sizeof(*a);
	if ((newb=y1i-rgn->bBox.bottom)>0) /* new lines below */
		z+= newb*sizeof(*a);
	if ((newc=MIN(rgn->bBox.bottom,y1i)-MAX(rgn->bBox.top,y0i))<0) /* middle intersection */
	{ /* gap - extents will be longer than actual */
		c = newb>>31, newb+= (newc^c)-c;
		c = newa>>31, newa+= (newc^c)-c;
	}
	z+= nse;
	if ((rgn=realloc(rgn,z))==NULL)
		return -2;
	*rrgn = rgn, rgn->size = z;
	/* adding from end to perform insertion on same & only buffer */
	f = rgn->x; /* cannot run before this (would be erroneous) */
	e = a = (int16*)(z+(int8*)rgn); /* new write */
	s = (int16*)(o+(int8*)rgn); /* old read */
	if (nse)
		*s++ = RGN_STOP;
	if (p0.x>p1.x)
		o = p1.x, c = p0.x;
	else
		o = p0.x, c = p1.x;
	/* fix bounds */
	if ((x0i=o>>16)<rgn->bBox.left)
		rgn->bBox.left = (int16)x0i;
	if ((x1i=c>>16)>rgn->bBox.right)
		rgn->bBox.right = (int16)x1i;
	if (y0i<rgn->bBox.top)
		rgn->bBox.top = (int16)y0i;
	if (y1i>rgn->bBox.bottom)
		rgn->bBox.bottom = (int16)y1i;
	/* below */
	if (newb<0) /* copy only */
	{
		int16		v;

		*--a = *--s; /* may or may not be stop point */
		for(;; *--a=v)
		{
			if (s<=f)
				if (++newb)
					goto bad;
				else
					break;
			else if ((v=*--s)==RGN_STOP)
				if (!++newb)
					{++s; break;}
			if (a<=f)
				goto bad;
		}
	} else if (newb) /* append */
		do
		{
			if ((a-=2)<f)
				goto bad;
			int64sub(x,dx);
			o = (int32)int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			*a = (int16)o, a[1] = RGN_STOP;
			if (o==RGN_STOP)
				++a;
		} while(--newb);
	if (newc<0) /* hollow non-y-intersection */
		do
		{
			if (--a<f)
				goto bad;
			*a = RGN_STOP;
		} while(++newc);
	else if (newc) /* insertion */
		do
		{
			if (a+1<=f || s<=f)
				goto bad;
			int64sub(x,dx);
			o = (int32)int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			/* endings require neg.-space, and xs can be non-IMIN/IMAX */
			/* be sure not to remove the start/end when compacting */
			if (s[-1]>=o) /* includes stop */
			{
				*--a = *--s;
				while(s>f && s[-1]>o && s[-1]!=RGN_STOP)
				{
					if (a<=f)
						goto bad;
					*--a = *--s;
				}
			}
			/* point of less or same - check for compact */
			if (o!=RGN_STOP)
				if (s>f && s[-1]==o)
					--s;
				else
				{
					if (a<=f)
						goto bad;
					*--a = (int16)o;
				}
			while(s>f && s[-1]!=RGN_STOP)
				*--a = *--s;
		} while(--newc);
	if (newa<0) /* copy only */
	{
		int16		v;

		*--a = *--s;
		for(;; *--a=v)
		{
			if (s<=f)
				if (++newa)
					goto bad;
				else
					break;
			else if ((v=*--s)==RGN_STOP)
				if (!++newa)
					{++s; break;}
			if (a<=f)
				goto bad;
		}
	} else if (newa) /* append */
		do
		{
			if ((a-=2)<f)
				goto bad;
			int64sub(x,dx);
			o = (int32)int64hi(x);
			if (o<x0i)
				o = x0i;
			else if (o>x1i)
				o = x1i;
			*a = (int16)o, a[1] = RGN_STOP;
			if (o==RGN_STOP)
				++a;
		} while(--newa);
	if (a<f)
	{ /* corrupt */
	bad:	rgn->bBox.top = rgn->bBox.bottom = 0;
		return -1;
	}
#ifndef MINIMAL_RGNCOMPACT
	while(a<e && *a==RGN_STOP)
		++rgn->bBox.top, ++a; /* top shouldn't ovfl */
	--e;
	while(a+1<e && *(uint32*)(e-1)==RGN_STOP*0x10001u)
		--rgn->bBox.bottom, --e; /* bottom shouldn't ovfl */
	++e;
#endif
	c = (int32)((int8*)e-(int8*)a);
	z = c+OFFSET(SLRgn,x);
	if (a!=f)
	{ /* compacting eliminated leading lines */
		memcpy(rgn->x,a,c);
#ifndef MINIMAL_RGNCOMPACT
		goto cpct;
	} else if (rgn->size!=z)
	{
	cpct:	rgn->size = z;
		if ((rgn=realloc(rgn,z))!=NULL || !z)
			*rrgn = rgn;
#endif
	}
	return 1;
}

int ptInRgn(const SLRgn *rgn,
	int x,
	int y)
{
	const int16	*s, *e;
	int32		o, c, h;

	if (!rgn || (c=rgn->size)<RECT_RGN_SIZE || c&1)
		return -1;
	if (x<rgn->bBox.left || x>=rgn->bBox.right || rgn->bBox.bottom<=rgn->bBox.top)
		return 0;
	s = rgn->x, e = (const int16*)(rgn->size+(const int8*)rgn);
	if (e<=s || (h=rgn->bBox.bottom-y)<=0 || (c=y-rgn->bBox.top)<0)
		return 0;
	if (c)
		skipLines(s,e,c,return -3);
	while(s<=e && (o=*s++)!=RGN_STOP && s<=e)
	{
		c = *s++;
		if (x>=o && x<c)
			return 1;
		if (c==RGN_STOP) /* odd lines are valid */
			break;
	}
	return 0;
}

int rectInRgn(const SLRgn *rgn,
	int left,
	int top,
	int right,
	int bottom)
{
	const int16	*s, *e;
	int		h;
	t_count		o, c;

	if (!rgn || (c=rgn->size)<RECT_RGN_SIZE || c&1)
		return -1;
	if (MAX(rgn->bBox.left,left)>=MIN(rgn->bBox.right,right))
		return 0;
	if (bottom>rgn->bBox.bottom)
		bottom = rgn->bBox.bottom;
	if (top<rgn->bBox.top)
		top = rgn->bBox.top;
	s = rgn->x, e = (const int16*)(rgn->size+(const int8*)rgn);
	if (e<=s || bottom<=top)
		return 0;
	if ((c=top-rgn->bBox.top)!=0) /* skip to .top line */
		skipLines(s,e,c,return -3);
	h = bottom-top;
	do
	{
		if (s>=e)
			return h==1 ? 0:-1;
		while((o=*s++)!=RGN_STOP)
		{
			if (s>=e)
				return h==1 ? 0:-1;
			c = *s++;
			if (MIN(c,right)>MAX(o,left))
				return 1;
			if (c==RGN_STOP) /* odd lines are valid */
				break;
		}
	} while(--h);
	return 0;
}

SLRgn *rgnCompact(SLRgn *rgn,
	boolean boundsOnly)
{
	int16		*s, *e;
	int32		z;

	if (!rgn)
		goto end;
	z = rgn->size;
	s = rgn->x, e = (int16*)(z+(int8*)rgn);
	if (rgn->bBox.bottom<=rgn->bBox.top || z<RECT_RGN_SIZE || z&1)
		{rgn->bBox.bottom = rgn->bBox.top; goto rct;}
	if (z<OFFSET(SLRgn,x)+4)
		goto rct;
	/* coords shouldn't overflow (unless the data's corrupted) */
	while(s<e && *s==RGN_STOP)
		++rgn->bBox.top, ++s;
	--e;
	while(s+1<e && *(uint32*)(e-1)==RGN_STOP*0x10001u)
		--rgn->bBox.bottom, --e;
	if (!boundsOnly)
	{ /* remove redundant x-coords, and again attempt to compact vertically */
		int		c, o, h;
		int16		nl, nr, v, *d, *a = s;

		while(a<e)
		{
			if (*a==RGN_STOP || *a!=a[1])
				{++a; continue;}
			d = a;
			while((a+=2)<e && (v=a[1])!=RGN_STOP && *a==v)
				;
			z = (int32)((int8*)a-(int8*)d);
			memcpy(d,a,(int8*)e-(int8*)a+2);
			a = d, ((int8*)e)-= z;
		}
		++e;
		while(s<e && *s==RGN_STOP)
			++rgn->bBox.top, ++s;
		--e;
		while(s+1<e && *(uint32*)(e-1)==RGN_STOP*0x10001u)
			--rgn->bBox.bottom, --e;
		/* finally, check to see if the region is a rectangle */
		/* error checking can be relaxed at this point */
		if ((h=RECTANGLE_HEIGHT(&rgn->bBox))==0)
			goto emp;
		a = s;
		/* should not begin with an empty line at this point */
		if ((nr=a[1])<(nl=*a))
			goto bad;
		a+= 2;
		if (nr==RGN_STOP)
			--a;
		do
		{
			for(;;)
			{
				if (a>e)
					if (h==1)
						break;
					else
					{
					bad:	rgn->size|= 1;
						goto end;
					}
				if ((o=*a++)==RGN_STOP)
					break;
				if ((c=*a++)!=nr || o!=nl)
					goto nrc;
				if (c==RGN_STOP)
					break;
			}
		} while(--h);
		rgn->bBox.left = nl, rgn->bBox.right = nr; /* important!; 2023.08.21 */
	rct:	z = 0;
		goto siz;
	} else if (rgn->bBox.bottom<=rgn->bBox.top)
	{
	emp:	free(rgn);
		return NULL;
	}
nrc:	++e;
	z = (int32)((int8*)e-(int8*)s);
	if (s>rgn->x)
		memcpy(rgn->x,s,z);
siz:	z+= OFFSET(SLRgn,x);
	if (rgn->size!=z)
	{
		void		*p;

		rgn->size = z;
		if ((p=realloc(rgn,z))!=NULL)
			rgn = p;
	}
end:	return rgn;
}

int offsetRgn(SLRgn **rrgn,
	int dx,
	int dy)
{
	SLRgn		*rgn;
	int16		*e, *s, *a;
	int32		z, c, v;

	if (!rrgn)
		return -1;
	if ((rgn=*rrgn)==NULL)
		return 0;
	if ((z=rgn->size)&1 || z<RECT_RGN_SIZE)
		return -1;
	/* adjust bounds, clip */
	v = rgn->bBox.left+dx, rgn->bBox.left = (int16)(v!=(int16)v ? v>>31^IMAX(int16):v);
	v = rgn->bBox.right+dx, rgn->bBox.right = (int16)(v!=(int16)v ? v>>31^IMAX(int16):v);
	v = rgn->bBox.bottom+dy, rgn->bBox.bottom = (int16)(v!=(int16)v ? v>>31^IMAX(int16):v);
	/* top last-most to retain top in v */
	v = rgn->bBox.top+dy, rgn->bBox.top = (int16)(v!=(int16)v ? v>>31^IMAX(int16):v);
	/* dispose if empty */
	if ((ABS(dx)|ABS(dy))>65535 || rgn->bBox.right<=rgn->bBox.left
			|| rgn->bBox.bottom<=rgn->bBox.top)
	{
		rgn->size = 0;
		free(rgn); *rrgn = NULL;
		return 0;
	}
	if (!(dx|dy))
		return 1;
	s = rgn->x;
	e = (int16*)(rgn->size+(const int8*)rgn);
	if (e<=s)
		goto end;
	if ((c=IMIN(int16)-v)>0)
		skipLines(s,e,c,goto bad);
	a = s; /* offset the coordinates in the new bounds */
	c = RECTANGLE_HEIGHT(&rgn->bBox);
	/* clipping will just lead to potential removal during compact */
	if (dx>0)
		do
			for(;;)
			{
				if (a>=e)
				{
					if (c==1)
						goto rsiz;
					goto bad;
				}
				/* we must also recognize that RGN_STOP is not a valid coord */
				if ((v=*a++)==RGN_STOP)
					break;
				v+= dx, a[-1] = (int16)MIN(v,RGN_STOP-1);
			}
		while(--c);
	else if (dx)
		do
			for(;;)
			{
				if (a>=e)
				{
					if (c==1)
						goto rsiz;
				bad:	rgn->size|= 1;
					return -1;
				}
				if ((v=*a++)==RGN_STOP)
					break;
				v+= dx, a[-1] = (int16)MAX(v,IMIN(int16));
			}
		while(--c);
rsiz:	z = (int32)((int8*)a-(int8*)s);
	if (s>rgn->x)
		memcpy(rgn->x,s,z);
	z+= OFFSET(SLRgn,x);
	rgn->size = z;
	rgn = rgnCompact(rgn,0);
end:	return (*rrgn=rgn)!=NULL;
}

int insetRgn(SLRgn **rrgn,
	int dx,
	int dy)
{ /* crude implementation */
	rect16		b;
	SLRgn		*rgn, *orgn = NULL;
	int16		*e, *s, *a;
	int32		c, v, u, o;

	if (!rrgn)
		return -1;
	if ((rgn=*rrgn)==NULL)
		return 0;
	if ((c=rgn->size)&1 || c<RECT_RGN_SIZE)
		return -1;
	/* dispose if empty or inset would inevitably make it empty */
	if (dx>65535 || dy>65535 || rgn->bBox.right<=rgn->bBox.left
			|| rgn->bBox.bottom<=rgn->bBox.top)
	{
		rgn->size = 0;
		free(rgn); *rrgn = NULL;
		return 0;
	}
	if (!(dx|dy))
		return 1;
	if (dx<-65535 || dy<-65535)
	{ /* wide open outset */
		rgn->size = RECT_RGN_SIZE;
		*(uint32*)&rgn->bBox = 0x80008000u;
		*(int32*)&rgn->bBox = 0x7FFF7FFF;
		if ((rgn=realloc(rgn,RECT_RGN_SIZE))!=NULL)
			*rrgn = rgn;
		return 1;
	}
	if (dx>0 && dy>0)
	{ /* first diagonal displacements */
		orgn = NULL;
		if (copyRgn(&orgn,rgn)==-2)
			return -2;
		offsetRgn(&orgn,-1,-1);
		sectRgn(&rgn,orgn);
		offsetRgn(&orgn,2,2);
		sectRgn(&rgn,orgn);
		offsetRgn(&orgn,-2,0);
		sectRgn(&rgn,orgn);
		offsetRgn(&orgn,2,-2);
		sectRgn(&rgn,orgn);
		if (orgn)
			free(orgn);
		--dx, --dy;
	}
	/* horizontal inset */
	/* adjust bounds, clip */
	b = rgn->bBox;
	v = b.left+dx, rgn->bBox.left = (int16)(v!=(int16)v ? v>>31^IMAX(int16):v);
	v = b.right-dx, rgn->bBox.right = (int16)(v!=(int16)v ? v>>31^IMAX(int16):v);
	s = rgn->x;
	e = (int16*)(rgn->size+(const int8*)rgn);
	if (e==s)
	{
		v = b.top+dy, rgn->bBox.top = (int16)(v!=(int16)v ? v>>31^IMAX(int16):v);
		v = b.bottom-dy, rgn->bBox.bottom = (int16)(v!=(int16)v ? v>>31^IMAX(int16):v);
		goto end;
	}
	o = dx>>31, dx+= o, dx^= o;
	if (dx)
	{
		a = s;
		c = RECTANGLE_HEIGHT(&b);
		do
		{ /* inset: inset pairs; flag empty pairs for removal */
		  /* outset: inset negative spaces */
			if (o)
			{
				if (a>=e)
				{
					if (c==1)
						break;
					goto bad;
				}
				if ((u=*a++)==RGN_STOP)
					continue;
				u-= dx, a[-1] = (int16)MAX(v,IMIN(int16));
			}
			for(;;)
			{
				if (a>=e)
				{
					if (c==1)
						break;
					goto bad;
				}
				if ((v=*a++)==RGN_STOP)
					break;
				if (a>=e)
				{
					if (c==1)
						break;
				bad:	rgn->size|= 1;
					return -1;
				}
				v+= dx;
				if ((u=*a++)==RGN_STOP)
				{ /* no end; offset only */
					a[-2] = (int16)MIN(v,RGN_STOP-1);
					break;
				}
				u-= dx;
				if (v>u) /* empty */
					u = v-=dx;
				a[-2] = v;
				a[-1] = u;
			}
		} while(--c);
		rgn->size = (int32)((int8*)a-(int8*)s)+OFFSET(SLRgn,x);
		if ((rgn=rgnCompact(rgn,0))==NULL)
			goto end;
	}
	/* vertical inset/out: sect/union with self displaced, from top and bottom */
	/* this is slow, sloppy and incomplete */
	if (dy>0)
	{ /* sect */
		orgn = NULL, o = 1;
		if (copyRgn(&orgn,rgn)==-2)
			return -2;
		do
		{
			offsetRgn(&orgn,0,o+(o>>1));
			sectRgn(&rgn,orgn);
			offsetRgn(&orgn,0,-(o+=o));
			sectRgn(&rgn,orgn);
		} while(dy>>=1);
		if (orgn)
			free(orgn);
	} else if ((dy=-dy)!=0)
	{ /* union */
		orgn = NULL, o = 1;
		if (copyRgn(&orgn,rgn)==-2)
			return -2;
		do
		{
			offsetRgn(&orgn,0,o+(o>>1));
			unionRgn(&rgn,orgn);
			offsetRgn(&orgn,0,-(o+=o));
			unionRgn(&rgn,orgn);
		} while(dy>>=1);
		if (orgn)
			free(orgn);
	}
end:	return (*rrgn=rgn)!=NULL;
}

int bitmapToRgn(SLRgn **rrgn,
	const uint8 *bm,
	int sp,
	const rect16 *bounds)
{
	int16		b[96], *d;
	SLRgn		*rgn;
	int		w, h, uh;

	if (!rrgn)
		return -1;
	if (!bm || sp<=0 || !bounds || (w=RECTANGLE_WIDTH(bounds))<=0
			|| (h=RECTANGLE_HEIGHT(bounds))<=0)
	{
		(void)disposeRgn(rrgn);
		return 0;
	}
	if ((rgn=realloc(*rrgn,RECT_RGN_SIZE))==NULL)
		return -2;
	*rrgn = rgn;
	rgn->size = RECT_RGN_SIZE;
	rgn->bBox = *bounds;
	bm+= sp*bounds->top+(bounds->left>>3);
	d = b, uh = bounds->left&7;

	/* the process is rather simple: add the x for each change, starting with 1 */
	do
	{
		const uint8	*s = bm;
		int		c, j;
		int16		x;
		int8		v, u;

		if (d>b+GET_ARRAY_COUNT(b)-2)
		{
			if ((rgn=rgnAppend(rgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
				goto merr;
			d = b;
		}
	//	*d++ = IMIN(*d);
		c = w, x = bounds->left, u = 0;
		v = (int8)(*s<<uh), j = 8-uh;
		goto bl;
		do
		{
			v = *s, j = 8;
		bl:	if (c<j)
				j = c;
			c-= j;
			do
				if ((int8)(u^v)<0)
				{
					u = v;
					if (d>b+GET_ARRAY_COUNT(b)-2)
					{
						if ((rgn=rgnAppend(rgn,b,
								(int32)((int8*)d-(int8*)b)))==NULL)
							goto merr;
						d = b;
					}
					*d++ = x;
				}
			while(v+=v,++x,--j);
		} while(++s,c>0);
		*d++ = RGN_STOP;
	} while(bm+=sp,--h);

	if ((rgn=rgnAppend(rgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
	{
	merr:	*rrgn = NULL;
		return -2;
	}
	/* it's impossible for pixels to produce multiple x-coordinates */
	return !!(*rrgn=rgnCompact(rgn,1));
}

int QDRgnToSLRgn(SLRgn **rrgn,
	const uint8 *s,
	int32 z)
{
	rect16		frame;
	int16		b[96], *d, *a;
	const uint8	*e;
	SLRgn		*rgn;
	size_t		lo, vo, vz;
	int16		lv, v;

	if (!rrgn)
		return -1;
	if (!s || z<10)
	{
		(void)disposeRgn(rrgn);
		return 0;
	}
	frame.top = (int16)(*s<<8|s[1]);
	frame.left = (int16)(s[2]<<8|s[3]);
	frame.bottom = (int16)(s[4]<<8|s[5]);
	frame.right = (int16)(s[6]<<8|s[7]);
	if ((rgn=realloc(*rrgn,RECT_RGN_SIZE))==NULL)
		return -2;
	*rrgn = rgn;
	rgn->size = RECT_RGN_SIZE;
	rgn->bBox = frame;
	if (z==10 || frame.right<=frame.left || frame.bottom<=frame.top)
		return 1; /* rect rgn */
	/* for SLRgn, vertical frame coordinates are bounds-dependent */
	rgn->bBox.top = lv = v = (int16)(s[8]<<8|s[9]);
	e = s-2+z; /* received source pointer is past size field */
	s+= 10, vz = lo = 0, a = NULL;
	goto nextrun;
	while(s<e)
	{
		int		c;
		int16		x;

		v = (int16)va_be16(s);
		if (v==0x7FFF)
			break;
		if ((c=v-lv)<=0)
		{
			rgn->size|= 1;
			return -1;
		}
		while(--c)
		{ /* unpack run */
			z = rgn->size+vz;
			if ((rgn=realloc(rgn,z))==NULL)
			{
				(**rrgn).size|= 1;
				return -1;
			}
			*rrgn = rgn, rgn->size = z;
			memcpy((int8*)rgn+z-vz,vo+(int8*)rgn,vz);
		}
		lv = v;
	nextrun:
		/* copy the next packed line for vertical run */
		if (v>=frame.bottom)
			break;
		vo = rgn->size;
		d = b;
		a = (int16*)((int8*)rgn+lo);
		//vz-= sizeof(int16)*2;
		while(s<e && (x=(int16)va_be16(s))!=0x7FFF)
		{
			if ((signed)vz>0)
			{ /* emphasis on *inversion* points: XOr with last run */
				while(*a<x)
				{
					*d++ = *a++, vz-= 2;
					if (d>b+GET_ARRAY_COUNT(b)-2)
					{
						lo = (int8*)a-(int8*)rgn;
						if ((rgn=rgnAppend(rgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
							goto merr;
						a = (int16*)((int8*)rgn+lo);
					}
				}
				if (*a==x)
					{++a, vz-= 2; continue;}
			}
			*d++ = x;
			if (d>b+GET_ARRAY_COUNT(b)-2)
			{
				lo = (int8*)a-(int8*)rgn;
				if ((rgn=rgnAppend(rgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
					goto merr;
				a = (int16*)((int8*)rgn+lo);
			}
		}
		*d++ = RGN_STOP;
		if ((rgn=rgnAppend(rgn,b,(int32)((int8*)d-(int8*)b)))==NULL)
		{
		merr:	*rrgn = NULL;
			return -2;
		}
		vz = rgn->size-vo, lo = vo; //+sizeof(int16);
	}
	return !!(*rrgn=rgnCompact(rgn,1));
}

const int16 *_rgnGotoLine(const int16 *s,
	const int16 *e,
	int o, /* y offset from s */
	int c) /* total line count */
{
	if (o<=0)
		return s;
	if ((c-=o)<o)
	{ /* offset is past midpoint; assuming it's faster to scan from end */
		if (e<=s)
			return NULL;
		if (e[-1]==RGN_STOP) /* final stop may or may not be present */
			--e;
		do
		{
			if (e-5>s)
			{ /* optimization: likely less than three points */
				if (e[-1]==RGN_STOP)
					{--e; continue;}
				if (e[-2]==RGN_STOP)
					{e-= 2; continue;}
				e-= 3;
				if (*e==RGN_STOP)
					continue; /* most likely case */
			}
			do
				if (--e<=s)
					return c==1 ? s:NULL;
			while(*e!=RGN_STOP);
		} while(--c);
		return e+1;
	}
	do
	{
		if (s+5<e)
		{ /* optimization: likely less than three points */
			if (*s==RGN_STOP)
				{++s; continue;}
			if (s[1]==RGN_STOP)
				{s+= 2; continue;}
			s+= 3;
			if (s[-1]==RGN_STOP)
				continue; /* most likely case */
		}
		do
			if (s>=e)
				return o==1 ? s:NULL;
		while(*s++!=RGN_STOP);
	} while(--o);
	return s;
}


/* roundRectRgn: a rectangle with corner arcs sected */

#if 1
#include "mblt.h"

int roundRectRgn(SLRgn **rrgn,
	const rect2dq *rc,
	ifixed rx,
	ifixed ry)
{
	bltParam	f;
	RRectParam	rr;
	SLRgn		*rgn;
	int16		*d;
	ufixed		w;
	ifixed		cx;
	int32		c, o, h;

	if (rx<0)
		rx = 0;
	if (ry<0)
		ry = 0;
	o = !!(rx|ry);
	h = rc->bottom>>16;
	if ((rc->left>>16)>=(rc->right>>16))
	{
		disposeRgn(rrgn);
		return 0;
	}
	c = rectRgn(rrgn,rc->left>>16,rc->top>>16,rc->right>>16,h,o);
	rgn = *rrgn;
	if (c<=0 || rgn->size<=RECT_RGN_SIZE)
		return c;
	memset(&f,0,sizeof(f));
	f.oz = 1;
	f.p = 0;
	f.a = (int8*)(rc->top>>16);
	f.e = (int8*)h;
	w = rc->right-rc->left;
	cx = rc->left+(w>>1)+(w&1); /* any frac round not toward hollow left */

	rInitRoundRect(&rr,&f,0,w,rc->top,rc->bottom,rx,ry,0);
	d = rgn->x;
	do
	{
		if (f.a>=rr.ua)
			if (f.a>=rr.ue)
				rr.ua = (int8*)-1, rr.y0 = rr.y1;
			else
			{
				o = rc->left>>16, c = rc->right>>16;
				goto uo;
			}
		o = rarcmul(int64hi(rr.y0),rr.rw)+rr.mw;
		(void)int64add(rr.y0,rr.dy);
		c = addu(cx,o)>>16, o = subu(cx,o)>>16;
	uo:	*d = o, d[1] = c;
	} while(d+=3,rEndLine(&f,f.e));
	return 1;
}
#endif
