/*
 mtext.c
  minimal (left-to-right) text measurement and rendering

   Revised: 2024.05.23
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.05.09
  Split from mfont.c.
  Put text font/style in grafPort for drawing case.
  Added chExtra; tabAlign (not in QDText).
2016.05.18
  Closer-to-QD post-image oblique shearing.
2016.08.22
  Got color (fctb) text working.
2022.04.10
  Implemented _style_overline.
*/

#include <string.h>

#include "mtext.h"
#include "mblt.h"

#include "ftrig.h"

/* ---- globals */

/* underlining thickness - excluding neighbors on first line */
/* for old QD style, set this to 0 */
ufixed		_ffPointScale = 1*65536./12+.5;
ufixed		_ffUlThick = 0.94*65536./12+.5;
ufixed		_ffOlThick = 0.94*65536./12+.5;

/* ---- macros, inlines */

__forceinline static int fixIndex(int i,
	const ffFont *fm)
{
	if (i+0u>fm->lastChar || (i-=fm->firstChar)<0 || fm->widTable[i*2+1]<0)
		i = fm->lastChar-fm->firstChar+1;
	return i;
}

/* returns negative when char does not exist in font */
__forceinline static int charWidth(int ch,
	ffFont *fm)
{
#ifdef MFONT_TTF
	if (!fm->widTable)
	{
		ffGlyph		*glyph = TTF_GetGlyph(fm,ch,NULL,0);

		return glyph ? glyph->advance:-1;
	}
#endif
	return fm->widTable[fixIndex(ch,fm)*2+1];
}

/* ---- measurement */

static ifixed fixScale(const ffFace *face,
	const ffFont *fm)
{ /* calculate scale for when source size mismatches */
	int		fz, z;

	if ((fz=fm->ptSize)==0 || face->fflag&ffBMNoScale)
		return 0x10000;
	z = face->size;
	if (z<=0)
		z = 12;
	return z==fz ? 0x10000:(z<<16|fz>>1)/fz;
}

int rGetFontInfo(ffFace *face,
	int16 *maxWidth,
	int16 *ascent,
	int16 *descent,
	int16 *leading,
	int16 *maxKerning)
{
	const ffFont	*fm;
	ufixed		scale;
	int		ow;

	if (!face)
		return -1;
	if (!face->fm)
		face->fm = ffMatch(fmFonts[face->font>=GET_ARRAY_COUNT(fmFonts) ? 0:face->font],
			face->size,face->style);
	if ((fm=face->fm)==NULL)
		return -8;
	scale = fixScale(face,fm);
	ow = (face->style&~fm->style)>>_style_outline_bit&1;
	if (maxWidth)
		*maxWidth = ufix16mul(fm->maxWidth,scale);
	if (maxKerning)
		*maxKerning = vfix16mul(fm->maxKerning,scale);
	if (ascent)
		*ascent = vfix16mul(fm->ascent,scale)+ow;
	if (descent)
		*descent = vfix16mul(fm->descent,scale)+ow;
	if (leading)
		*leading = vfix16mul(fm->leading,scale);
	return 1;
}

int rTextWidth(grafPort *p,
	const char *s,
	int z,
	ufixed x)
{
	ffFont		*fm;
	ifixed		chExtra;
	ufixed		scale, spExtra, sb;
	int		i, w, cw;
	int8		styl;

	if (!p || (fm=p->txFace.fm)==NULL || !fm->data || !fm->height)
		return -1;
	if (z<=0 && (!z || (z=strlen(s))==0))
		return 0;
	scale = fixScale(&p->txFace,fm);
	if (!fm->widTable)
	{
#ifdef MFONT_TTF
		scale = 1;
		sb<<= 16;
#else
		return -1;
#endif
	}
	spExtra = p->spExtra;
	chExtra = p->chExtra;
	styl = p->txFace.style&~fm->style;
	sb = ((styl&_style_bold)+(styl>>_style_outline_bit&1))*scale;
	w = 0;
	do
	{
		i = *(const uint8*)s;
		if ((cw=charWidth(i,fm))>0)
			cw = cw*scale+sb;
		cw+= chExtra;
		if (cw<0)
			cw = 0;
		x+= cw;
		if (i==' ')
			x+= spExtra;
		w+= x>>16, x&= 0xFFFF;
	} while(++s,--z);
	return w;
}

int rClipText(grafPort *p,
	const char *tx,
	int z,
	ufixed *txw,
	ufixed *minWidth,
	ufixed x)
{
	ffFont		*fm;
	const char	*s;
	ufixed		tabAlign, tabOrigin, chExtra, spExtra;
	ufixed		maxw, lw, scale, bw;
	int		i, cw;
	int8		styl;

	if (!p || (fm=p->txFace.fm)==NULL || !fm->data || !fm->height || !txw || (s=tx)==NULL)
		return -1;
	lw = 0;
	if ((z<=0 && (!z || (z=strlen(s))==0)) || (maxw=*txw)==0)
		goto end;
	scale = fixScale(&p->txFace,fm);
	styl = p->txFace.style&~fm->style;
	bw = ((styl&_style_bold)+(styl>>_style_outline_bit&1))*scale;
	if (!fm->widTable)
	{
#ifdef MFONT_TTF
		scale = 1;
#else
		return -1;
#endif
	}
	chExtra = p->chExtra, spExtra = p->spExtra&~(chExtra&IMIN(ifixed));
	tabAlign = p->tabAlign, tabOrigin = p->tabOrigin;
	do
	{
		lw = x, i = *(const uint8*)s, ++s;
		if (i==' ')
			x+= spExtra;
		else if (i==9 && tabAlign>0)
		{
			x-= tabOrigin;
			x+= tabAlign;
			if (x<lw)
				goto ovfl;
			x-= x%tabAlign;
			x+= tabOrigin;
			goto rechk;
		}
		//x+= chExtra;
		//if (x<lw)
		//	goto ovfl;
		if ((cw=charWidth(i,fm))>0)
			cw = cw*scale+bw;
		cw+= chExtra;
		if (cw<0)
			cw = 0;
		if ((x+=cw)<lw)
			{ovfl: x =~0u, --s; goto end;}
	rechk:	if (x>maxw)
			goto end;
	} while(--z);
	lw = x;
end:	*txw = x;
	if (minWidth)
		*minWidth = lw;
	return s-tx;
}

textBreakCode rTextBreak(grafPort *p,
	const char *start,
	const char *pos,
	int *length,
	ufixed *txw,
	ufixed w)
{
	ffFont		*fm;
	const char	*s = pos, *ps;
	ufixed		tabAlign, tabOrigin, chExtra, spExtra;
	ufixed		maxw, lw, pw, nw, scale;
	int		z, i, cw, sb, r = tNoBreak;
	int8		styl;

	if ((fm=p->txFace.fm)!=NULL && (z=*length)>0 && (maxw=*txw)!=0)
	{
		r = tBreakWord;
		chExtra = p->chExtra, spExtra = p->spExtra&~(chExtra&IMIN(ifixed));
		tabAlign = p->tabAlign, tabOrigin = p->tabOrigin;
		scale = fixScale(&p->txFace,fm);
		if (!fm->widTable)
		{
	#ifdef MFONT_TTF
			scale = 1;
	#else
			return -1;
	#endif
		}
		styl = p->txFace.style&~fm->style;
		sb = ((styl&_style_bold)+(styl>>_style_outline_bit&1))*scale;
		nw = pw = w, ps = NULL;
		do
		{
			lw = w, i = *(const uint8*)s, ++s;
			if (i-10u<4u)
			{ /* feed control */
				if (i!='\f')// && (i==10 || (s-2>=start && s[-2]!=13)))
				{ /* line-ending */
					r = tBreakLine;
					goto end;
				}
			} else if (i==9 && tabAlign>0)
			{
				w-= tabOrigin;
				w+= tabAlign;
				if (w<lw)
					{w =~0u, r = tBreakOverflow; goto end;}
				w-= w%tabAlign;
				w+= tabOrigin;
				goto rechk;
			}
			if ((cw=charWidth(i,fm))>0)
				cw = cw*scale+sb;
			cw+= chExtra;
			if (cw<0)
				cw = 0;
			if ((w+=cw)<lw)
		//	if (w<lw)
				{w =~0u, r = tBreakOverflow; goto end;}
			if (i==' ')
				w+= spExtra, ps = s, pw = nw;
			else
				nw = w;
			//w+= chExtra;
		rechk:	if (w<lw)
				{w =~0u, r = tBreakOverflow; goto end;}
		} while(w<=maxw && --z);
		if (!z)
			r = tNoBreak;
		else if (ps==s) /* walk forward to next non-space */
			for(w=nw; z && *s==' '; ++s,--z)
				;
		else if (ps && ps!=start+1)
			s = ps, w = pw; /* after last space covered */
		else
			--s, w = lw, r = tBreakChar;
	}
end:	*txw = w;
	*length = s-pos;
	return r;
}

int32 rTextSkipEmpty(const char *s,
	int32 len,
	int32 i,
	boolean reverse)
{
	if (i>0 && i<len)
		if (s[i-1]==13 && s[i]==10)
			i+= reverse ? -1:1;
	return i;
}

int32 rTextScanEOL(const char *text,
	int32 len,
	int32 i,
	boolean reverse)
{
	const char	*s = text+i, *e;
	int		dir;
	uint8		c;

	if (i>=0 && i<len)
	{
		dir = 1, e = text+len;
		if (reverse)
			dir =-dir, e = text;
		do
			if ((c=(uint8)*s-10u)<4u && c!='\f'-10u)
			{
				s-= dir<0 && s>text && !c && s[-1]==13;
				return s-text;
			}
		while(s+=dir,s!=e);
	}
	return -1;
}

boolean rTextStyleMatch(const tStyleRun *sra,
	const tStyleRun *srb,
	boolean hasLineEndings)
{
	uint8		sp;

	if ((*(const int32*)&sra->font^*(const int32*)&srb->font)
			|(*(const int32*)&sra->color.green^*(const int32*)&srb->color.green)
			|| (sra->color.red^srb->color.red)|(sra->style^srb->style))
		return 0;
	sp = sra->special^srb->special;
	if (!hasLineEndings)
		sp&=~_special_align_mask;
	return !sp;
}

/* ---- rendering */

static void bltOblique(int8 *a,
	size_t sp,
	size_t w, /* must be >0 */
	int h,
	int y) /* must be |y|<h */
{
	//if (h<=0) --rInitBlt takes care of height check
	//	return;
	++y;
	do
	{
		int8		*d, *s, *e;
		int		c, z;

		if ((c=y>>1)<0)
			e = a+c+w, c =-c, d = a, s = a+c;
		else
			e = s = a, d = a+c;
		if ((z=-c+w)>0)
			memmove(d,s,z);
		memset(e,0,c); /* to clear empty space */
	} while(--y,a+=sp,--h);
}

static void bltUnderline(int8 *d,
	size_t sp, /* must be >1 */
	size_t w, /* must be >0 */
	int h,
	int y, /* total height with thickness must be allocated */
	int uh)
{
	int		c, p;
	int8		*a, l, v, u;

	if (y<0)
		return;
	++y, p =-sp;
	a = d+sp*y;
	if (++y>=h)
		sp = uh = 0; /* height is too small for next-line reads */
	v = a[p]|*a|a[sp], l = 0;
	for(c=w; --c; )
		++a, u = a[p]|*a|a[sp], a[-1]|= !(v|l|u), l = v, v = u;
	*a|= !(v|l);
	/* following lines need to ignore the pixels we just made */
	while(--uh>0)
	{
		a = d+sp*y;
		if (++y>=h)
			sp = uh = 0;
		v = *a|a[sp], l = 0;
		for(c=w; --c; )
			++a, u = *a|a[sp], a[-1]|= !(v|l|u), l = v, v = u;
		*a|= !(v|l);
	}
}

static void bltOutline(int8 *a,
	size_t sp,
	size_t w,
	int h)
{ /* use of the Bits routines isn't particularly fast */
  /* ...but no unaligned reads/writes on systems that don't support that :) */
	int8		*b;
	size_t		z, c8;

	z = sp*h;
	if ((b=malloc(z+sp))==NULL)
		return;
	memcpy(b,a,z);
	/* 3px bold, horizontal */
	c8 = z*8-8;
	rOrBits(b,a,0,8,c8,0);
	rOrBits(b,a,8,0,c8,0);
	/* 3px bold, vertical */
	c8 = (z-sp-sp+w)*8;
	rOrBits(b+sp,a,0,0,c8,0);
	rOrBits(b,a+sp,0,0,c8,0);
	/* finish corners */
	rOrBits(b+sp,a,8,0,c8-8,0);
	rOrBits(b+sp,a,0,8,c8-8,0);
	rOrBits(b,a+sp,8,0,c8-8,0);
	rOrBits(b,a+sp,0,8,c8-8,0);
	rXorBits(a,b,0,0,c8+sp*8,0);
	free(b);
}

static void bltLinethrough(int8 *a,
	size_t sp,
	size_t w,
	int h,
	int y,
	int z)
{
	z+= !z;
	if (y<0) //(y-=z>>1)<0)
		z+= y, y = 0;
	if (z>0)
		memset(a+y*sp,1,sp*z-1);
}

/* expects buffer to have enough capacity */
/* bold merging is adjusted to scale (not 1px as with QD) */
static ifixed bltChar(ffFont *fm,
	int8 *a,
	int dpitch,
	int i,
	int j,
	int x, /* plus maxKerning */
	ifixed sy,
	ifixed scale,
	ifixed iscale,
	int bw,
	uint8 sb)
{
	const uint16	*t;
	const uint8	*s;
#if MFONT_NO_UNPACK1BPP
	uint8		*b = NULL;
	int		xw = 0;
#endif
	ifixed		w;
	int		bot, ow, spitch;
#ifdef MFONT_TTF /* shouldn't get NULL here without TTF */
	ffGlyph		*glyph = NULL;

	if (!fm->widTable)
	{
		ifixed		kerningX;
		int		top;

		if ((glyph=TTF_GetGlyph(fm,i,NULL,1))==NULL)
			return 0;
		ow = glyph->advance, bw<<= 16;
		if ((w=glyph->maxx-glyph->left)>glyph->w)
			w = glyph->w; /* should be 8bpp */
		s = glyph->bitmap;
		kerningX = 0;
		if (j)
			TTF_GetKerning(fm,j,i,&kerningX,NULL);
		if ((x+=(kerningX>>16)+glyph->left)<0)
			x = 0;
		top = glyph->top;
		bot = glyph->h+top;
		if (bot>fm->height)
			bot = fm->height;
		bot-= top;
		spitch = glyph->w;
		if (top<0)
			sy-= top<<16, top = 0;
		if (sy>0xFFFF)
		{
			top-= sy>>16;
			if (top<0)
				sy-= glyph->top<<16, top = 0;
			else
				sy = 0;
		}
		if ((sy>>16)+0u>=bot+0u)
		{
			if (w+0u>ow+0u)
				w = ow;
			ow+= bw;
			return ow;
		}
		a+= dpitch*top;
	#if MFONT_NO_UNPACK1BPP
		spitch = (spitch+7)>>3;
		xw = (spitch+16)>>3;
		b = _alloca(xw*8*2);
	#endif
		i = 0;
	} else
#endif
	{
		i = fixIndex(i,fm);
		if ((ow=fm->widTable[i*2+1])<=0)
			return 0;
		s = fm->data;
		t = (uint16*)(s+fm->bmSize);
		w = t[i+1]-t[i];
		x+= fm->widTable[i*2];
		bot = fm->height;
		spitch = fm->pitch;
		i = t[i];
	#if MFONT_NO_UNPACK1BPP
		if (!(fm->flag&ffDepthMask) && w>0)
		{
			xw = (w+16)>>3;
			b = _alloca(xw*8*2);
			s+= i>>3;
			i&= 7;
		} else
	#endif
		 s+= i;
	}
	if (w+0u>ow+0u)
		w = ow;
	ow+= bw;
	if (w>0)
	{
		const uint8	*r;
		ifixed		hscale = iscale>>1;

		if (iscale==0x10000)
		{
			sy>>= 16, s+= sy*spitch;
			do
			{
				int8		*d = x+a;
				int		c = w;

				r = s;
			#if MFONT_NO_UNPACK1BPP
				if (b)
				{
					b1to8(b,s,xw);
					r = b+i;
				}
			#endif
				if (sb)
				{
					*d++|= *r++;
					if (--c>0)
					{
						do
							*d++|= *r|r[-1], ++r;
						while(--c);
					}
					*d|= r[-1];
				} else
					do
						*d++|= *r++;
					while(--c);
				s+= spitch;
			} while(a+=dpitch,++sy<bot);
			i = 0;
		} else if (iscale>0x10000)
		{
		#if MFONT_NO_UNPACK1BPP
			const uint8	*l = NULL;
		#endif
			int		j = sy>>16, k;

			for(--w; (k=j)<bot; )
			{
				const uint8	*r2;
				int8		*d = x+a;
				ifixed		sx = hscale;
				int		t;

				sy+= iscale, j = sy>>16;
				r2 = r = s+k*spitch;
				if (j>++k && k<bot)
					r2+= spitch;
			#if MFONT_NO_UNPACK1BPP
				if (b)
				{
					if (r!=l)
					{
						b1to8(b,r,xw);
						b1to8(b+xw*8,r2,xw);
					}
					l = r;
					r = b+i;
					r2 = b+xw*8+i;
				}
			#endif
				k = sx>>16;
				if (sb)
				{
					*d++|= r[k]|r2[k];
					++k, t = (sx+=iscale)>>16;
					if (t>k && t<=w)
						d[-1]|= r[k]|r2[k];
					while((k=t)<=w)
					{
						sx+= iscale;
						*d++|= r[k]|r[k-1]|r2[k]|r2[k-1];
						++k, t = sx>>16;
						if (t>k && t<=w)
							d[-1]|= r[k]|r[t-1]|r2[k]|r2[t-1];
					}
					*d|= r[w]|r2[w];
				} else
				{
					for(; k<=w ; k=t)
					{
						*d++|= r[k]|r2[k];
						++k, t = (sx+=iscale)>>16;
						if (t>k && k<=w)
							d[-1]|= r[k]|r2[k]; /* was [t]; 2023.07.28 */
					}
					d[-1]|= r[w]|r2[w];
				}
				a+= dpitch;
			}
			i = 0;
		} else
		{
		#if MFONT_NO_UNPACK1BPP
			const uint8	*l = NULL;
		#endif
			ifixed		c = w<<16;
			int		k;

			while((k=sy>>16)<bot)
			{
				int8		*d = x+a;
				ifixed		sx = hscale;

				r = s+k*spitch;
			#if MFONT_NO_UNPACK1BPP
				if (b)
				{
					if (r!=l)
						b1to8(b,r,xw);
					l = r;
					r = b+i;
				}
			#endif
				if (sb)
				{
					while(sx<0x10000)
						k = sx>>16, *d++|= r[k], sx+= iscale;
					while(sx<c)
					{
						k = sx>>16;
						*d++|= r[k]|r[k-1], sx+= iscale;
					}
					*d|= r[w-1];
				} else
					while(sx<c)
						*d++|= r[sx>>16], sx+= iscale;
				sy+= iscale, a+= dpitch;
			}
		}
	}
	return ow*scale;
}

static void bltGrayPat(int8 *b,
	int dp,
	int h,
	int odd)
{
	int32		pat = _be32_native(0x01000100)^(-odd&0x01010101);
	int		z = dp+3;

	do
	{
		int32		*d = (int32*)b;
		int		c;

		if ((c=z>>3)!=0)
			do
				*d+= *d&pat, d[1]+= d[1]&pat;
			while(d+=2,--c);
		if (z&4)
			*d+= *d&pat;
	} while(pat^=0x01010101,b+=dp,--h);
}

/*
 this routine not only translates the packed fctb to native but colorizes with fore/bk
*/
static void genColorizedTable(void *t,
	const uint32 *s,
	int c,
	grafPort *p)
{
	rgb48		col;
	rgbRoutine	rtn;
	pixMap		*m = p->map;
	void		*gt;
	uint32		v, fg, bk;
	int		rb = m->minCmp ? m->minCmp:10;

	rtn = rgbCCRoutine[pmGetRGBCC(m,&gt)];
	fg = (p->fgColor.red<<31-16&0x7FE00000)
		|(p->fgColor.green<<21-16&0x001FFC00)
		|p->fgColor.blue>>16-10;
	bk = (p->bkColor.red<<31-16&0x7FE00000)
		|(p->bkColor.green<<21-16&0x001FFC00)
		|p->bkColor.blue>>16-10;
	if (m->bpp>16)
	{
		uint32		*d = t;

		do
		{
			v = (*s&bk)|fg;
			col.red = (uint16)(v>>31-16);
			col.green = (uint16)(v>>21-16);
			col.blue = (uint16)(v<<16-10);
			*d = rtn(rb,gt,&col);
		} while(++s,++d,--c);
	} else
	{
		uint16		*d = t;

		do
		{
			v = (*s&bk)|fg;
			col.red = (uint16)(v>>31-16);
			col.green = (uint16)(v>>21-16);
			col.blue = (uint16)(v<<16-10);
			*d = rtn(rb,gt,&col);
		} while(++s,++d,--c);
	}
}

ifixed rDrawText(grafPort *p,
	const char *s,
	int z,
	int dy,
	ifixed qx)
{
	bltParam	f;
	ffFont		*fm;
	SLRgn		*clip;
	char		*buf, *d;
	uint32		oldpenc;
	ifixed		tabAlign, tabOrigin, odisp, sk, scale, iscale, qscale, bwq;
	ufixed		w, spExtra, chExtra;
	int		gray, i, j, x, h, ik, k, o, ox, dx, sp, kr, bw, uh, ow;
	int		sy; /* source-x/y - may be altered for oblique adjustment */
	int		firstKerning;
	int16		ascent, descent, leading;
	uint8		styl, sb;

	if (!p || (fm=p->txFace.fm)==NULL || !fm->data || !fm->height || ABS(dy)>65535)
		return IMIN(ifixed);
	if (z<=0 && (!z || !s || (z=strlen(s))==0) || (clip=p->clipRgn)==NULL)
		return qx;
	ox = qx;
	scale = fixScale(&p->txFace,fm);

	/* bold/oblique adjustment for when the closest bmf doesn't match */
	styl = p->txFace.style&~fm->style;
	ow = styl>>_style_outline_bit&1;
	bw = styl&_style_bold;
	sb = (uint8)bw, bw+= ow;

	//rGetFontInfo(&p->txFace,NULL,&ascent,&descent,&leading,NULL);
	iscale = 0x10000;
	ascent = fm->ascent;
	descent = fm->descent;
	leading = fm->leading;
	if (scale!=0x10000)
	{
		iscale = ufix16div(0x10000,scale);
		ascent = vfix16mul(ascent,scale);
		descent = vfix16mul(descent,scale);
		leading = vfix16mul(leading,scale);
	}
	qscale = scale;
	bwq = bw*scale;
	if (!fm->widTable)
	{
#ifdef MFONT_TTF
		const ffGlyph	*firstGlyph;

		TTF_GetKerning(fm,0,*s,&firstKerning,NULL);
		firstKerning>>= 16;
		if ((firstGlyph=TTF_GetGlyph(fm,*s,NULL,0))!=NULL)
			firstKerning+= firstGlyph->minx;
		scale = 1;
#else
		return -1;
#endif
	} else
		firstKerning = fm->maxKerning;
	ascent+= ow, descent+= ow;
	tabAlign = p->tabAlign;
	tabOrigin = p->tabOrigin;
	chExtra = p->chExtra, spExtra = p->spExtra&~(chExtra&IMIN(ifixed));

	/* seek to first visible character */
	kr = firstKerning*scale>>16;
	if (!(styl&_style_italic) || (ik=(leading-ascent-2)>>1)>0)
		ik = 0;
	o = p->origin.x; /* 2023.06.29 */
	o = MAX(clip->bBox.left,o)+ik-1-ABS(kr);
	o = o!=(int16)o ? o>>31^IMAX(int16):o<<16;
	//if (clip->bBox.left<0)
	//	__debugbreak();
	odisp = 0;
	if (qx<o)
	{
		ifixed		lw;
		int		cw;

		do
		{
			lw = qx, i = *(const uint8*)s, ++s;
			if (i==9 && tabAlign>0)
			{
				qx-= tabOrigin;
				qx+= tabAlign;
				if (qx<lw)
					{--z; break;}
				qx-= qx%tabAlign;
				qx+= tabOrigin;
				if (qx<lw)
					{--z; break;}
				continue;
			}
			if (qx<lw)
				{--z; break;}
			if ((cw=charWidth(i,fm))>0)
				cw = cw*scale+bwq;
			cw+= chExtra;
			if (cw<0)
				cw = 0;
			if ((qx+=cw)<lw)
				{--z; break;}
			if (i==' ')
				qx+= spExtra;
		//	qx+= chExtra;
		} while(--z && qx<o && qx>=lw);
		odisp = qx-lw;
		--s, ++z, qx = lw;
	}
	x = qx>>16;
	if (x+MIN(kr,0)-ow>clip->bBox.right)
		return qx;
	dy-= ascent, h = 0;
	uh = 0;
	if (styl&_style_underline)
	{ /* underline is not part of the scale in QD */
		uh = (_ffUlThick*p->txFace.size)>>16;
		uh+= uh<=0;
		h = (ascent+leading+MAX(descent,uh))<<16;
	}
	k = fm->height*qscale;
	if (h<k)
		h = k;
	h+= 0x8000, h>>= 16, h+= ow+ow, dy-= ow;
	oldpenc = p->fgSolid;
	gray = p->txMode&mTextGray ? !PenGetGray(p,&p->fgSolid):0;

	/* making use of pattern bit for fast remapping */
	if ((k=rInitBlt(&f,p,fm,NULL,dy,dy+h,p->txMode|mSrcPat))<0)
		{k = qx; goto end;}
	w = (f.cl.right-x)<<16; /* shouldn't overflow */

	/* adjusting the tab origin is tricky; could overflow */
	//if (tabAlign>0)
	//	tabOrigin = (tabOrigin+(qx&~0xFFFF))%tabAlign;
	w = addu(w,ow<<16);

	//p->tabOrigin&= 0xFFFF; /* origin x is 0 for this rClipText call */
	p->tabOrigin =-odisp;
	z = rClipText(p,s,z,&w,NULL,qx&=0xFFFF);
	p->tabOrigin = tabOrigin;

	if (z<=0 || (w=(w>>16)+ow-ik)<=0)
		{k = qx; goto end;} /* clipped out */
	if ((sy=(k-3/*-ow-ow*/)&~1)<0)
		sy = 0;
	/* negative displacements allowed, used in some fonts */
	dx = x-ow;
	tabOrigin-= x<<16;
	if ((o=kr)<0)
		dx+= o, w-= o, o = 0; /* was dx-= o; fixes Symbol; 2023.04.17 */
	/* ClipText or the input still isn't 100% */
	if (odisp && tabAlign>0)
		w+= (tabAlign-1)>>16;

	sp = (w+4+((chExtra+spExtra)>>16))&~3;
	if ((buf=calloc(sp,1+(qscale>>16)+(h-=sy)))==NULL)
		{k =-2; goto end;}
	d = buf+(sp+1)*ow;
	sk = sy*iscale+(iscale>>1), j = 0;
	if (tabAlign>0)
		do
		{
			i = *(const uint8*)s, ++s;
			if (i==9)
			{
				//printf("qx %X, origin %X, x %d, ow %d\n",
				//	qx,tabOrigin,x,ow);
				qx-= tabOrigin;
				qx+= tabAlign;
				if (qx<0)
					break;
				qx-= qx%tabAlign;
				qx+= tabOrigin;
				if (qx<0)
					break;
				continue;
			}
			qx+= bltChar(fm,d,sp,i,j,((ufixed)qx>>16)+o,sk,scale,iscale,bw,sb);
			if (i==' ')
				qx+= spExtra;
			qx+= chExtra;
		} while(j=i,--z);
	else
		do
		{
			i = *(const uint8*)s, ++s;
			qx+= bltChar(fm,d,sp,i,j,((ufixed)qx>>16)+o,sk,scale,iscale,bw,sb);
			if (i==' ')
				qx+= spExtra;
			qx+= chExtra;
		} while(j=i,--z);
	if (styl&_style_italic)
		bltOblique(d,sp,w,h-ow,ascent-sy+leading);
	if (uh)
		bltUnderline(d-ow,sp,w,h-ow,ascent-sy,uh);
	if (styl&_style_outline)
		bltOutline(buf,sp,w,h);
	if (styl&_style_strikethrough)
		bltLinethrough(buf,sp,w,h,((ascent*8+descent*3+8+(fm->height*qscale>>16))>>4)-sy,
			(p->txFace.size*_ffPointScale+_ffPointScale)>>16);
	if (styl&_style_overline)
		bltLinethrough(buf,sp,w,h,((ascent+descent+(fm->height*qscale>>16))>>5)-sy,
			(p->txFace.size*_ffOlThick)>>16);
	f.cnv.s = buf, f.vx = f.vy = 0;

	if (gray) /* color will definitely mess this up.. */
		bltGrayPat(buf,sp,h,(x^ox>>16)&1);
	if (fm->ctCount>0)
	{ /* overriding fore color altogether */
		int		z = fm->ctCount*4;

		f.cnv.t = _alloca(z);
		genColorizedTable(f.cnv.t,(const uint32*)((const int8*)fm->data+fm->bmSize-z),
			fm->ctCount,p);
	}
	f.cnv.ry = 0x10000;
	rBlt(&f,sp,w-((styl&_style_italic)>>1),dx,0,k-sy,0);
	if (x+(qx>>16)>=IMAX(int16))
		k = IMAX(ifixed);
	else
		k = (x<<16)+qx;
	free(buf);
end:	p->fgSolid = oldpenc;
	rEndBlt(&f); /* not actually necessary, given the pattern mode is used */
	return k;
}
