/*
 mttf.c
  minimal TrueType (no PostScript) Font layer using FreeType2
  adapted from SDL_ttf.c

  Created: 2016.08.14
  Revised: 2023.08.21

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#include "../freetype/include/ft2build.h"
#include FT_FREETYPE_H
#ifdef TTF_OUTLINING
 #include FT_OUTLINE_H
#endif
#include FT_GLYPH_H
#include FT_MODULE_H

#include "mtext.h"

#ifndef MFONT_TTF
 #error Must define MFONT_TTF for this file!
#endif

//#define DEBUG_OUTGLYPH

/* ---- macros */

#define FT_FIXED(x)		((x)<<10)
#define FT_FLOOR(x)		((x)>>6)
#define FT_CEIL(x)		((x+63)>>6)

/* ---- typedefs */

/* ---- globals */

static FT_Library	library;
static int8		TTF_initialized = 0;

/* ---- code */

static int translate_ft_error(FT_Error err,
	int defn)
{
	switch(err)
	{ /* rough translation to utility error codes... */
	case FT_Err_Ok: return defn;
	case FT_Err_Unimplemented_Feature:
	case FT_Err_Unknown_File_Format:
	case FT_Err_Cannot_Render_Glyph: return -8;
	case FT_Err_Out_Of_Memory: return -2;
	case FT_Err_Stack_Overflow:
	case FT_Err_Stack_Underflow: return -9;
	case FT_Err_Invalid_Opcode:
	case FT_Err_Invalid_CodeRange: return -14;
	case FT_Err_Invalid_Glyph_Index:
	case FT_Err_Invalid_Character_Code: return -7;
	case FT_Err_Invalid_Frame_Read:
	case FT_Err_Invalid_Stream_Read: return -3;
	default: return -1;
	}
}

int TTF_FlushGlyph(ffGlyph *glyph)
{
	int		r = 0;

	if (glyph->bitmap)
		{free(glyph->bitmap); ++r;}
#if 1
	glyph->bitmap = NULL;
	glyph->cchar = glyph->index = 0;
	glyph->ptSize = 0;
	glyph->h = glyph->w = 0;
#else
	memset(glyph,0,sizeof(*glyph));
#endif
	return r;
}

int TTF_FlushCache(TTF_Font *ttf)
{
	int		r = 0, i;

	for(i=0; i<GET_ARRAY_COUNT(ttf->cache); ++i)
		if (ttf->cache[i].w)
			r+= TTF_FlushGlyph(&ttf->cache[i]);
	r+= TTF_FlushGlyph(&ttf->scratch);
	return r;
}

ffGlyph *TTF_GetGlyph(ffFont *fm,
	uint32 ch,
	int *rerr,
	boolean get_bmap)
{
	TTF_Font	*ttf = fm->data;
	ffGlyph		*outglyph;
	FT_Face		face;
	FT_GlyphSlot	glyph;
#ifdef TTF_OUTLINING
	FT_Outline	*outline;
#endif
	FT_Glyph_Metrics	*metrics;

	FT_Error	err;

	if (fm->widTable)
		{outglyph = NULL, err =-1; goto end;} /* not TrueType! */
	err = 0;
	if (ch<256)
		outglyph = &ttf->cache[ch];
	else
		outglyph = &ttf->scratch;
	if (outglyph->ptSize!=fm->ptSize || outglyph->cchar!=ch)
		TTF_FlushGlyph(outglyph);
	else if (outglyph->bitmap || (!get_bmap && outglyph->w))
		goto end;

	face = ttf->face;

	if (!outglyph->index)
		outglyph->index = FT_Get_Char_Index(face,ch);
	if ((err=FT_Load_Glyph(face,outglyph->index,FT_LOAD_DEFAULT))!=0)
		{outglyph = NULL; goto end;}

	glyph = face->glyph;
	metrics = &glyph->metrics;
#ifdef TTF_OUTLINING
	outline = &glyph->outline;
#endif
	/* metrics always loaded */
	if (!outglyph->w)
	{ /* freetype rounds the glyph bearings anyway (no point in saving fraction) */
		outglyph->minx = FT_FLOOR(metrics->horiBearingX);
		outglyph->maxy = FT_FLOOR(metrics->horiBearingY);
		if (FT_IS_SCALABLE(face))
		{
			outglyph->maxx = outglyph->minx+FT_CEIL(metrics->width);
			outglyph->miny = outglyph->maxy-FT_CEIL(metrics->height);
			outglyph->top = fm->ascent-FT_FLOOR(metrics->horiBearingY);
		} else
		{ /* Freetype 2 fills in many of the font metrics with 0.. */
			outglyph->maxx = outglyph->minx+FT_CEIL(metrics->horiAdvance);
			outglyph->miny = outglyph->maxy-fm->height;
			outglyph->top = 0;
		}
		outglyph->advance = FT_FIXED(metrics->horiAdvance);
		outglyph->w = 1;
	}

	if (get_bmap && !outglyph->bitmap && outglyph->minx!=outglyph->maxx)
	{
		FT_Bitmap	*src;
	#ifdef TTF_OUTLINING
		FT_Glyph	olglyph = NULL;

		if (ttf->outline>0 && glyph->format!=FT_GLYPH_FORMAT_BITMAP)
		{
			FT_Stroker	stroker;

			FT_Get_Glyph(glyph,&olglyph);
			if ((err=FT_Stroker_New(library,&stroker))!=0)
				{outglyph = NULL; goto end;}
			FT_Stroker_Set(stroker,ttf->outline*64,FT_STROKER_LINECAP_ROUND,
				FT_STROKER_LINEJOIN_ROUND,0); 
			FT_Glyph_Stroke(&olglyph,stroker,1); /* delete the original glyph */
			FT_Stroker_Done(stroker);
			/* Render the glyph */
			if ((err=FT_Glyph_To_Bitmap(&olglyph,FT_RENDER_MODE_MONO,0,1))!=0)
			{
				FT_Done_Glyph(olglyph);
				outglyph = NULL;
				goto end;
			}
			src = &((FT_BitmapGlyph)olglyph)->bitmap;
		} else
	#endif
		{
			if ((err=FT_Render_Glyph(glyph,FT_RENDER_MODE_MONO))!=0)
				{outglyph = NULL; goto end;}
			src = &glyph->bitmap;
		}

		outglyph->h = src->rows;
		if ((outglyph->w=src->pitch<<3)==0)
			{outglyph->w = 1; goto set;} /* in case of freetype bugs */
		if (outglyph->w>src->width)
			outglyph->w = src->width; /* minimize memory usage */

		assert(src->pixel_mode==FT_PIXEL_MODE_MONO);

		if (src->rows>0)
		{
			uint8		*d, *s;
		#if MFONT_NO_UNPACK1BPP
			int		c, rz = (outglyph->w+7)>>3;
			int		size = rz*src->rows;
		#else
			int		size = (outglyph->w*src->rows+15)&~7;
			int		c, rz = (outglyph->w+7)>>3;
		#endif
			if ((outglyph->bitmap=malloc(size))==NULL)
			{ /* TODO: flush other entries until there is enough memory */
				if (outglyph->w)
					err = FT_Err_Out_Of_Memory;
				outglyph = NULL;
				goto end;
			}
			c = src->rows, d = outglyph->bitmap, s = src->buffer;
			do
		#if MFONT_NO_UNPACK1BPP
				memcpy(d,s,rz);
			while(d+=rz,s+=src->pitch,--c);
		#else
				b1to8(d,s,rz);
			while(d+=outglyph->w,s+=src->pitch,--c);
		#endif
		}
	#ifdef TTF_OUTLINING
		if (olglyph)
			FT_Done_Glyph(olglyph);
	#endif
		/* sfnt bitmaps carry differing left than bearing */
		outglyph->left = CLIP_INT16(glyph->bitmap_left);
		if (outglyph->left>32767u)
			outglyph->left = outglyph->minx;
	#ifdef DEBUG_OUTGLYPH
		dprintf("rendered glyph for %d/%d: h %d, w %d, t %d, l %d, mnx %d, mxx %d;g",
			ch,fm->ptSize,outglyph->h,outglyph->w,outglyph->top,glyph->bitmap_left,
			outglyph->minx,outglyph->maxx);
	#endif
	}
set:	outglyph->cchar = ch;
	outglyph->ptSize = fm->ptSize;
end:	if (rerr)
		*rerr = translate_ft_error(err,0);
	return outglyph;
}

void TTF_CloseFont(ffFont *fm)
{
	TTF_Font	*ttf;

	if (!fm || fm->widTable || (ttf=fm->data)==NULL)
		return;
	TTF_FlushCache(ttf);
	if (ttf->face)
		(void)FT_Done_Face(ttf->face);
	free(ttf);
	fm->data = NULL;
}

int TTF_SetPtSize(ffFont *fm,
	int ptsize)
{
	FT_Face		face;
	TTF_Font	*ttf;
	FT_Error	err;

	if (!fm || (ttf=fm->data)==NULL || fm->widTable)
		return -1;
	if (fm->ptSize==ptsize)
		return 0;
	err = 0;
	if (!TTF_initialized)
		return -8;
	face = ttf->face;

	if (FT_IS_SCALABLE(face))
	{
		FT_Fixed	scale;
	  	/* *64 = 72 dpi */
		if ((err=FT_Set_Char_Size(ttf->face,0,ptsize*64,0,0))!=0)
		{
			TTF_CloseFont(fm);
			return -10;
		}
		scale = face->size->metrics.y_scale;
		fm->ascent  = FT_CEIL(FT_MulFix(face->ascender,scale));
		fm->descent =-FT_CEIL(FT_MulFix(face->descender,scale));
		fm->height = FT_CEIL(FT_MulFix(face->height,scale));
	  	fm->maxWidth = FT_CEIL(FT_MulFix(face->max_advance_width,scale));
	} else
	{ /* pre-rendered */
		if (ptsize>=ttf->face->num_fixed_sizes) /* ptSize and not index? */
			ptsize = ttf->face->num_fixed_sizes-1;
		err = FT_Set_Pixel_Sizes(face,
				face->available_sizes[ptsize].height,
				face->available_sizes[ptsize].width);
	  	/* Freetype 2 doesn't have full fixed-sized/bitmapped font support */
	  	fm->ascent = fm->height = face->available_sizes[ptsize].height;
	  	fm->descent = 0;
	  	fm->maxWidth = 0; /* not available */
	}
	fm->ptSize = ptsize;
	if ((fm->leading=fm->height-fm->ascent-fm->descent)<0) //-1 /* baseline */ )<0)
		fm->leading = 0;
	return translate_ft_error(err,1);
}

int TTF_OpenSFNT(ffFont *fm,
	int face_index)
{
	FT_Face		face;
	TTF_Font	*ttf;
	FT_Error	err;
	int		i;

	if (!fm || !fm->sfnt || fm->ptSize || fm->data)
		return -1;
	if (!TTF_initialized)
	{
		if (FT_Init_FreeType(&library))
			return -8;
		TTF_initialized = 1;
	}
	if ((ttf=calloc(sizeof(*ttf),1))==NULL)
		return -2;
	fm->data = ttf;
	ttf->args.flags = FT_OPEN_MEMORY|FT_OPEN_DRIVER;
	ttf->args.memory_base = (const uint8*)hLock(fm->sfnt);
	ttf->args.memory_size = hSize(fm->sfnt);
	ttf->args.driver = FT_Get_Module(library,
		ttf->args.memory_size>4 && !memcmp(ttf->args.memory_base,"OTTO",4)
		? "cff":"truetype");

	if ((err=FT_Open_Face(library,&ttf->args,face_index,&ttf->face))!=0)
	{
		TTF_CloseFont(fm);
		return translate_ft_error(err,1);
	}
	face = ttf->face;
	fm->firstChar = FT_Get_First_Char(face,NULL);
	for(i=0; i<face->num_charmaps; ++i)
	{
		FT_CharMap	charmap = face->charmaps[i];

		if (charmap->encoding==FT_ENCODING_APPLE_ROMAN)
			{(void)FT_Set_Charmap(face,charmap); break;}
	}
	return TTF_SetPtSize(fm,12);
}

int TTF_GetKerning(ffFont *fm,
	int prevch,
	int nextch,
	ifixed *kerningX,
	ifixed *kerningY)
{
	TTF_Font	*ttf;
	FT_Vector	delta;
	int		err;

	if (!fm || (ttf=fm->data)==NULL || fm->widTable)
		return -1;
	if (!TTF_initialized)
		return -8;
	delta.x = delta.y = 0;
	err = FT_Get_Kerning(ttf->face,prevch,nextch,ft_kerning_default,&delta)==0;
	if (kerningX)
		*kerningX = FT_FIXED(delta.x);
	if (kerningY)
		*kerningY = FT_FIXED(delta.y);
	return err;
}
