/*
 mwind.c

  Created: 2014.08.17
  Revised: 2024.03.30

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
*/

#define __MWIND_PRIVATE
#include "mevent.h"
#include "mwind.h"
#include "mtext.h"

#include <string.h>
#if DEBUG
 #include <stdio.h>
#endif

#define FIX_DESK_ORIGIN

#define NO_RGN_PTR		(SLRgn*)(-1)

/* ---- globals */

const rgb48	_defMWinColors[3] =
{
	{0,0,0}, /* frame color */
	{0,0,0}, /* title color */
	{61166,61166,61166} /* content (was 86.667%) */
};

windowRecord	screen_desktop_window;
/* the order here is deliberate for a wPaintBehind hack */
#define frontWindow	screen_desktop_window.below
#define bottomWindow	screen_desktop_window.above

GrafPtr		_curPort;

static int16	closeOne = 1;
static boolean	screen_lock;

#if DEBUG
static void checkWindowChain(void)
{
	windowRef	win;
	int		wincount = 1;
#if DUMPCHAIN
	printf("# window chain (bottom up):\n");
#endif
	if ((win=bottomWindow)!=NULL)
	{
		do
		{
		#if DUMPCHAIN
			printf("# %* %08p - %.*s\n",
				wincount,win,*win->title,win->title+1);
		#endif
		} while(++wincount<32 && (win=win->above)!=NULL);
		assert(wincount<32);
	}
#if DUMPCHAIN
	printf("# window chain (top down):\n");
#endif
	if ((win=frontWindow)!=NULL)
	{
		do
		{
		#if DUMPCHAIN
			printf("# %* %08p - %.*s\n",
				wincount,win,*win->title,win->title+1);
		#endif
		} while(--wincount>=-1 && (win=win->below)!=NULL);
		if (wincount<0)
			printf("# unmatching window chain below list");
	}
		
}
#endif

/* ---- code */

void _defMWinModalDraw(const rect16 *box,
	const rgb48 *color,
	boolean fill)
{
	rect2dq		b;
	pt2dq		pt[3];
	cSpec		cstor[17];
	cTab64		*ct = (cTab64*)&cstor;
	CGrafPtr	p = &screen_desktop_window.port;

	if (!color)
		color = _defMWinColors;
	_imgContColors(ct,color+wCContent,NULL,0);
	cstor->rgb = p->fgColor;
	rPenNormal(p,NULL);
	p->fgSolid = RGBToPen(p->map,color+wCFrame);
	FixRect(&b,box->left-3,box->top-3,box->right+3,box->bottom+3);
	rRect(p,&b);
	/* -1:bl-tl-tr */
	p->fgSolid = RGBToPen(p->map,&cstor[3].rgb);
	FixPt(pt,box->left-1,box->bottom);
	FixPt(pt+1,box->left-1,box->top-1);
	FixPt(pt+2,box->right,box->top-1);
	rLine(p,*pt,pt[1]);
	rLine(p,pt[1],pt[2]);
	/* -1:tr-br-bl */
	p->fgSolid = RGBToPen(p->map,&cstor[7].rgb);
	FixPt(pt,box->right,box->bottom);
	FixPt(pt+1,box->left-1,box->bottom);
	rLine(p,pt[2],*pt);
	rLine(p,*pt,pt[1]);
	/* -2:bl-tl-tr */
	p->fgSolid = RGBToPen(p->map,&cstor[2].rgb);
	FixPt(pt,box->left-2,box->bottom+1);
	FixPt(pt+1,box->left-2,box->top-2);
	FixPt(pt+2,box->right+1,box->top-2);
	rLine(p,*pt,pt[1]);
	rLine(p,pt[1],pt[2]);
	/* -2:tr-br-bl */
	p->fgSolid = RGBToPen(p->map,&cstor[10].rgb);
	FixPt(pt,box->right+1,box->top-2);
	FixPt(pt+1,box->right+1,box->bottom+1);
	FixPt(pt+2,box->left-2,box->bottom+1);
	rLine(p,*pt,pt[1]);
	rLine(p,pt[1],pt[2]);
	if (fill)
	{
		RGBBack(p,color+wCContent);
		rEraseRect(p,box->left,box->top,box->right,box->bottom);
	}
	RGBFore(p,&cstor->rgb);
	if (!screen_lock)
		update_screen_rect(box->left-3,box->top-3,box->right+3,box->bottom+3);
}

DEF_WIN_RTN(modalProc)
{
(void)data;
	if (msg==wcMsgDraw)
	{
		if (!win->visible)
			return 0;
		_defMWinModalDraw(&win->bounds,win->themeData,0);
		return 1;
	} else if (msg==wcMsgCalc)
	{
		win->frame = win->bounds;
		OutsetRect16(&win->frame,3,3);
	}
	return -8;
}

winRoutine	f_winr[] =
{
	NULL,		/* doc */
	modalProc	/* modal */
			/* single-pixel frame */
};

windowRef wFrontWindow(void)
	{return frontWindow;}

/* in this implementation, only draws the frame at most; wBeginUpdate erases */
int wPaintOne(windowRef win,
	const rect16 *inval)
{
	SLRgn		br;

	if (!win)
		return -1;
	br.bBox = screen_desktop_window.bounds;
	if (!win->visible || !SectRect16(&br.bBox,inval))
		return 0;
	br.size = RECT_RGN_SIZE;
	//offsetRgn();
	br.bBox.top-= win->bounds.top;
	br.bBox.left-= win->bounds.left;
	br.bBox.bottom-= win->bounds.top;
	br.bBox.right-= win->bounds.left;
	wInvalRgn(win,&br);
	return 1;
}

int wPaintBehind(windowRef behind,
	const rect16 *inval)
{
	rect16		clobbered;
	windowRef	win;
#if 0
	windowRef	*updl = NULL, *p;
#endif
	int		c;

	win = &screen_desktop_window;
	if (!win->above)
		return 0;
	clobbered = *inval;
	SectRect16(&clobbered,&screen_desktop_window.bounds);

	screen_lock = 1; /* no need to copy buffer to native screen until finished */

	for(c=0; win && win!=behind; win=win->above)
	{
		if (!win->visible)
			continue;
		wPaintOne(win,&clobbered);
	}

	screen_lock = 0;

	update_screen_rect(clobbered.left,clobbered.top,
		clobbered.right,clobbered.bottom);
	return c;
}

#ifdef _CRTNOP
_CRTNOP
#endif
static int updateWindow(windowRef win)
{
	rect16		b;
	pixMap		*m, *sm;

	if (!win->visible)
		return 0;
	b = screen_desktop_window.bounds;
	SectRect16(&b,&win->bounds);
	if ((m=win->port.map)==NULL)
		return -1;
	m->w = RECTANGLE_WIDTH(&b);
	m->h = RECTANGLE_HEIGHT(&b);
	m->pOrigin.x =-b.left;
	m->pOrigin.y =-b.top;
	sm = screen_desktop_window.port.map;
	if (memcmp(&m->pitch,&sm->pitch,OFFSET(pixMap,pHand)-OFFSET(pixMap,pitch))
			|| (sm->bpp<9 && sm->pTable
			&& (!m->pTable || m->pTable->seed!=sm->pTable->seed)))
	{
		if (m->iTable)
			{free(m->iTable); m->iTable = NULL;}
		memcpy(&m->pitch,&sm->pitch,OFFSET(pixMap,pHand)-OFFSET(pixMap,pitch));
		(void)copyCTable(&m->pTable,sm->bpp<9 ? sm->pTable:NULL);
	}
	win->frame = win->bounds;
	if (win->rtn)
		(void)win->rtn(win,wcMsgCalc,NULL);
	return wPaintOne(win,&win->frame);
}

void wCalcVis(windowRef win)
{
	SLRgn		br, *vRgn;
	windowRef	top = frontWindow;

	br.size = RECT_RGN_SIZE;
	if (!top || !win)
		return;
	vRgn = win->port.visRgn;
	br.bBox = screen_desktop_window.bounds;
	copyRgn(&vRgn,&br);
	do
	{
		if (top==win)
			break;
		br.bBox = top->frame;
		diffRgn(&vRgn,vRgn,&br);
	} while((top=win->below)!=NULL);
	win->port.visRgn = vRgn;
	copyRgn(&win->port.clipRgn,vRgn);
	sectRgn(&win->port.clipRgn,win->port.clipSave);
}

void wSendBehind(windowRef sending,
	windowRef behind)
{
	if (!sending || sending==&screen_desktop_window || sending==behind)
		return; /* idiot proofing */
	if (!frontWindow) /* uninitialized FrontWindow - assume empty */
		{frontWindow = bottomWindow = sending; return;}
	if (behind && behind!=(windowRef)-1l && behind->below==sending)
		return; /* already behind this window */
	/* remove, first (will be re-inserted into the list) */
	if (sending->above)
		sending->above->below = sending->below;
	if (sending->below)
		sending->below->above = sending->above;
	sending->below = sending->above = NULL;
	if (behind==(windowRef)-1l)
	{ /* bring to front */
		if (sending==frontWindow)// || !frontWindow)
			return; /* already front window */
		assert(!frontWindow->above);
		frontWindow->above = sending;
		sending->below = frontWindow;
		frontWindow = sending;
		updateWindow(sending); /* only the new front window to be updated */
	} else
	{
		if (!behind) /* to bottom */
			if ((behind=bottomWindow)==NULL)
				return; /* !!! */
		sending->above = behind;
		if ((sending->below=behind->below)!=NULL)
			behind->below->above = sending;
		behind->below = sending;
		wCalcVis(behind);
		wPaintBehind(sending,&sending->frame);
	}
	wCalcVis(sending);
}

int wOpenWindow(windowRef newWin,
	windowRef behind,
	int16 procID)
{
	if (!newWin)
		return -1;
	if (procID>=0)
		newWin->rtn = (procID&=15)<GET_ARRAY_COUNT(f_winr) ? f_winr[procID]:NULL;
	if (openGrafPort(&newWin->port,screen_desktop_window.port.map->bpp,
			RECTANGLE_WIDTH(&newWin->bounds),RECTANGLE_HEIGHT(&newWin->bounds),0)==0)
		return 0;

	rTextFont(&newWin->port,1); /* applFont */
	//newWin->port.fgColor = screen_desktop_window.port.fgColor;
	newWin->port.txMode = mTransparent;

	wSendBehind(newWin,behind);
	updateWindow(newWin);
	RGBBack(&newWin->port,(newWin->themeData ? (const rgb48*)newWin->themeData:_defMWinColors)+2);
	wPaintOne(newWin,&screen_desktop_window.bounds);
#if DEBUG
	checkWindowChain();
#endif
	return 1;
}

void wCloseWindow(windowRef win)
{
	if (!win || win==&screen_desktop_window)
		return; /* idiot proofing */
	win->visible = 0;
	disposeRgn(&win->clipSave);
	disposeRgn(&win->updateRgn);

	wKillControls(win);
	wPaintBehind(win,&win->frame);

	/* remove from list */
	if (win->above)
		win->above->below = win->below;
	if (win->below)
		win->below->above = win->above;
	closeGrafPort(&win->port);

	//if (win->palette)
	//	{free(win->palette); win->palette = NULL;}

	/* recalc bottom */
	if (!win->below)
	{ /* bottom's below ==NULL; the desktop's above should be the bottom window */
		assert(bottomWindow==win);
		bottomWindow = win->above;
	}
	/* recalc FrontWindow */
	win->below = win->above = NULL;
	if ((win=bottomWindow)!=NULL)
		while(win->above)
			win = win->above;
	/* avoid recursion (shouldn't really need this line) */
	frontWindow = win==&screen_desktop_window ? NULL:win;
	//if (win->palette)
	//	wActivatePalette(win);
#if DEBUG
	checkWindowChain();
#endif
}

void wDisposeWindow(windowRef win)
{
//	closeOne = 1; -- messes up window/clip/updateEvt redraw also messing up window frames
	wCloseWindow(win);
	free(win);
}

int wShowWindow(windowRef win,
	boolean show)
{
	if (!win)
		return -1;
	if (show)
	{
		if (!win->visible)
			win->visible = 1;
		return updateWindow(win);
	}
	/* hide */
	if (win->visible)
	{
		win->visible = 0;
		wPaintBehind(win,&win->frame);
		return 1;
	}
	return 0;
}

void wBeginUpdate(windowRef win)
{
	rgb48		bkc;
	pt2d16		oldor, newor;
	SLRgn		*clip, br;
	int		r;

	if (win->updateRect.left>=win->updateRect.right
			|| win->updateRect.top>=win->updateRect.bottom)
		return;
#ifdef FIX_DESK_ORIGIN
	oldor = screen_desktop_window.port.origin;
	newor.x = newor.y = 0;
	screen_desktop_window.port.origin = newor;
#endif
	if (win->rtn)
		(void)win->rtn(win,wcMsgDraw,NULL);
	bkc = win->port.bkColor;
	RGBBack(&win->port,(win->themeData ? (const rgb48*)win->themeData:_defMWinColors)+wCContent);
	//rEraseRect(p,MAX(win->updateRect.left,0)+win->bounds.left,
	//	MAX(win->updateRect.top,0)+win->bounds.top,
	//	MIN(win->updateRect.right,win->port.map->w)+win->bounds.left,
	//	MIN(win->updateRect.bottom,win->port.map->h)+win->bounds.top);
	r = rEraseRect(&win->port,win->updateRect.left,win->updateRect.top,
		win->updateRect.right,win->updateRect.bottom);
	win->port.bkColor = bkc;
#ifdef FIX_DESK_ORIGIN
	screen_desktop_window.port.origin = oldor;
#endif
	if (r<=0)
		return;
	clip = win->port.clipRgn;
	copyRgn(&win->clipSave,clip); /* 2024.03.28 */
	//win->clipASave = clip->bBox;
	br.bBox = win->updateRect;
	br.bBox.top+= win->port.origin.y;
	br.bBox.left+= win->port.origin.x;
	br.bBox.bottom+= win->port.origin.y;
	br.bBox.right+= win->port.origin.x;
#if 1
	br.size = RECT_RGN_SIZE;
	sectRgn(&win->port.clipRgn,&br);
#else
	r = clip->bBox.top;
	SectRect16(&clip->bBox,&br.bBox);
	if (clip->size!=RECT_RGN_SIZE)
		clip->bBox.top = r;
#endif
}

void wEndUpdate(windowRef win)
{
	SLRgn		*clip;

	if (win->visible==2)
		win->visible = 1; /* get update events again */
	if (win->updateRect.left>=win->updateRect.right
			|| win->updateRect.top>=win->updateRect.bottom
			|| !win->visible
			|| (clip=win->port.clipRgn)==NULL)
		return;
	if (win->clipSave)
	{ /* 2024.03.28 */
		copyRgn(&win->port.clipRgn,win->clipSave);
		free(win->clipSave); win->clipSave = NULL;
	}
	//clip->bBox = win->clipASave;
	if (!screen_lock)
		update_screen_rect(win->updateRect.left+win->bounds.left,
			win->updateRect.top+win->bounds.top,
			win->updateRect.right+win->bounds.left,
			win->updateRect.bottom+win->bounds.top);
	win->updateRect.right = win->updateRect.bottom = IMIN(int16);
}

void wInvalRgn(windowRef win,
	const SLRgn *dirtyRgn)
{
	eventRecord	event;

	if (!win || !win->visible || !dirtyRgn)
		return; /* not going to bother if not visible here */
	UnionRect16(&win->updateRect,&dirtyRgn->bBox);
	unionRgn(&win->updateRgn,dirtyRgn);
	if (win->visible!=2) /* a value to not send more than one updateEvt */
	{
		win->visible = 2;
		event = mEventState;
		event.what = updateEvt;
		event.message = win;
		PostMEvent(&event);
	}
}

void wInvalRect(windowRef win,
	const rect16 *dirtyRect)
{
	SLRgn		br;

	br.size = RECT_RGN_SIZE;
	if (!dirtyRect)
		return;
	br.bBox = *dirtyRect;
	wInvalRgn(win,&br);
}

void wValidRgn(windowRef win,
	const SLRgn *updatedRgn)
{
	if (!updatedRgn)
		return;
	SubRect16(&win->updateRect,&updatedRgn->bBox);
	diffRgn(&win->updateRgn,win->updateRgn,updatedRgn);
}

void wValidRect(windowRef win,
	const rect16 *updatedRect)
{
	SLRgn		br;

	br.size = RECT_RGN_SIZE;
	if (!updatedRect)
		return;
	br.bBox = *updatedRect;
	wValidRgn(win,&br);
}

void update_screen_window_rect(windowRef win,
	int left,
	int top,
	int right,
	int bottom)
{
	int		ox, oy;

	if (!win || !win->visible
			|| win->port.map->baseAddr!=screen_desktop_window.port.map->baseAddr)
		return;
	ox =-win->bounds.left, oy =-win->bounds.top;
	if ((ox+=win->port.origin.x)>0)
		left = subu(left,ox), right = subu(right,ox);
	else
		left = addu(left,-ox), right = addu(right,-ox);
	if ((oy+=win->port.origin.y)>0)
		top = subu(top,oy), bottom = subu(bottom,oy);
	else
		top = addu(top,-oy), bottom = addu(bottom,-oy);
	update_screen_rect(left,top,right-left,bottom-top);
}

void update_screen_window_clip(windowRef win)
{
	const SLRgn	*clip;
	int		o, top, left, bottom, right;

	if (!win || !win->visible
			|| win->port.map->baseAddr!=screen_desktop_window.port.map->baseAddr
			|| (clip=win->port.clipRgn)==NULL)
		return;
	top = clip->bBox.top;
	left = clip->bBox.left;
	bottom = clip->bBox.bottom;
	right = clip->bBox.right;
	o = win->port.origin.x-win->bounds.left, left-= o, right-= o;
	o = win->port.origin.y-win->bounds.top, top-= o, bottom-= o;
	update_screen_rect(left,top,right-left,bottom-top);
}

void MSetPort(GrafPtr port)
{
	_curPort = port;
}

void GetPort(GrafPtr *port)
{
	*port = _curPort;
}

void wModalDialog(wModalFilter filter,
	int16 *itemHit)
{
	DialogRef	dlg;
	GrafPtr		oldport;

	if ((dlg=GetWindowDialog(FrontWindow()))==NULL)
		{*itemHit =-1; return;}
	GetPort(&oldport);
	MSetPort(&dlg->port);
	while(1)
	{
		EventRecord	event;
		ctlObj		*ctl;

		(void)WaitNextEvent(mDownMask|mUpMask|keyDownMask|autoKeyMask|updateMask|activMask,
				&event,2,NULL);
		if (mEventState.what==closeEvent)
		{
			event.what = closeEvent;
			*itemHit = closeOne^=3; /* alternates between 2 and 1; 2017.06.02 */
			if (!filter || !filter(dlg,&event,itemHit))
				;
			break;
		}
		if (filter && filter(dlg,&event,itemHit))
			break;
		if ((ctl=wControlEvent(dlg,&event))!=NULL)
		{
			*itemHit = wGetControlIndex(dlg,ctl)+1;
			break;
		}
	}
	MSetPort(oldport);
}

/* ---- for some rudimentary timing functions */

void wMcsDelay(int32 mcs)
{
	UnsignedWide	ltime, time;

	Microseconds(&ltime);
	do
	{
		global_idle_proc();
		Microseconds(&time);
	} while((int32)int64lo(time)-(int32)int64lo(ltime)<mcs);
}

/* ---- template */

windowRef wNewWindowFromTemplate(const wcTemplate *item,
	windowRef behind)
{
	windowRef	win, parent;
	int		v;

	if (!item || item->bounds.top>=item->bounds.bottom
			|| (win=calloc(1,sizeof(*win)))==NULL)
		return NULL;
	win->bounds = item->bounds;
	win->goAway = item->hilite;
	win->refCon = (void*)item->maximum;
	memcpy(win->title,item->title,item->title ? *item->title+1:0);
	parent = &screen_desktop_window;

	/* recenter */
	v = RECTANGLE_WIDTH(&win->bounds);
	win->bounds.right = (int16)(v+(win->bounds.left=((RECTANGLE_WIDTH(&parent->bounds)-v)>>1)
		+parent->bounds.left));
	v = RECTANGLE_HEIGHT(&win->bounds);
	win->bounds.bottom = (int16)(v+(win->bounds.top=((RECTANGLE_HEIGHT(&parent->bounds)-v)>>1)
		+parent->bounds.top));
	if (!wOpenWindow(win,behind,item->type))
	{
		free(win);
		return NULL;
	}

	rTextFont(&win->port,ffSystem);
	for(++item; item->bounds.top<item->bounds.bottom; ++item)
	{
		ctlObj		*ctl;

		if ((ctl=wNewControl(win,&item->bounds,
				(item->type&~1)==ctlTypeStaticText ? 0:item->title,
				1,0,0,item->maximum,item->type,0))==NULL)
			continue;
		if ((item->type&~1)==ctlTypeStaticText)
		{
			ctl->value = 1;
			if (item->title)
			{
				textBoxObj	*tbox = ctl->data;

				tBoxInsert(tbox,0,(const char*)item->title+1,*item->title,NULL,0,0);
				tbox->scrollH = tbox->scrollV = 0;
				tbox->maxWidth = IMAX(ifixed);
			}
		}
		if (item->hilite)
			wHiliteControl(ctl,item->hilite);
	}
	return win;
}

/* ---- palette management */

#if 0
/* mind entry bits 6-7 (some may be 'locked' or 'animated') */
static void wReplaceColors(const cSpec *colors,
	int c)
{
	uint8		used[256];

	memset(used,0,sizeof(used));
	/* one method is to use the iTable for closest colors to replace */
	/* then search until exhaustion, and rebuild iTab */
}

void wActivatePalette(windowRef win)
{
}
#endif
