/*
 pict.c
  PICT resource renderer

  Created: 2012.12.29
  Revised: 2024.06.09

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2013.01.01 08:12 EST
  MAX instead of MIN for w/h so all of the parchment image in ViV is rendered.
Using the slow way to finish a clipped line fixed the missing final span bug.
2013.01.04 03:14 EST
  after implementing all of the rest of the opcodes found in PictUtils.r
(not rendered, however), made the image parser its own function for PixPat codes.
  forced to use GWorld instead of raw PixMap because QT always draws to port
(the dst parameter is counter-intuitive if it can't be used alone).
2013.07.16 07:01 EDT
  final fixes, etc. for post-QT image rendering (needed for multiple QT images).
2016.06.11
  PnLocHFrac only applies to text!
2021.07.06
  fixed RLE32 starting offset in alpha/4-cmpCount case.
2023.08.19
  comment proc in call for thread safety.
2024.05.29
  finised implementation for BMP in PICT, and more minimal depth for QT PICTs.
*/

#include <string.h>
#include <stdlib.h>

#define __MGRAF_INTERNAL

#ifdef TARGET_OS_MAC /* need the SDK to implement QT imaging on other platforms */
 #include "my_cseries.h"
 #define lockPixels(p) (**p).baseAddr /* alloc_image buffer w/o GWorld always a calloc */
 #define pmPitch(p) ((**(p)).rowBytes&0x3FFF)
 #define QT_IMAGE
 #include <QuickDraw.h>
#else
 #include "mgraf.h"
 #include "mtext.h"
 #define DisposePixMap disposePixMap
 #define pmPitch(p) (0+(p)->pitch)
#endif

#ifdef QT_IMAGE
 #include "mimg_file.h"
#endif

#include "ftrig.h"

/* ---- globals */

const rgb48	QDClassicColors[8] =
{
	{0},
	{0xFCFC,0xF3F3,0x0505},
	{0xF2F2,0x0808,0x8484},
	{0xDDDD,0x0808,0x0606},
	{0x0202,0xABAB,0xEAEA},
	{0,0x8080,0x1111},
	{0,0,0xD4D4},
	{0xFFFF,0xFFFF,0xFFFF}
};
/* this is the correct formula */
#define QD1ColorToRGB(v)	QDClassicColors[(((v&1)-1)|v)>>6&7]

/* ---- code */

static int unpack_rle8(const uint8 *src,
	const uint8 *end,
	uint8 *dst,
	int row_bytes,
	int dst_pitch,
	int rows)
{
	const uint8	*s = src;
	boolean		word_size = row_bytes>250;

	if (!dst)
		dst_pitch = 0; /* skipping */
	do
	{
		int		z;
		const uint8	*se;
		uint8		v, *d = dst, *de = dst_pitch+d;

		if (word_size)
			z = *s<<8|s[1], s+= 2;
		else
			z = *s++;
		se = s+z;
		while(s<se)
		{
			int8		c;

			if (src>=end)
				return -1;
			if ((c=*s++)<0)
			{ /* RLE */
				c =-c, ++c, v = *s++;
				if (d+c>de)
				{ /* clip, skip */
					do
						if (d<de)
							*d++ = v;
					while(--c);
					c = 0;
					break;
				}
				do
					*d++ = v;
				while(--c);
				c = 0;
			} else
			{ /* uncompressed run */
				++c;
				if (d+c>de) /* will overflow */
				{ /* clip, skip */
					do
						if (d<de)
							*d++ = *s++;
						else
							++s;
					while(--c);
					c = 0;
					break;
				}
				do
					*d++ = *s++;
				while(--c);
			}
		}
		s = se;
	} while(dst+=dst_pitch,--rows);

	return s-src;
}

static int unpack_rle16(const uint8 *src,
	const uint8 *end,
	uint8 *dst,
	int row_bytes,
	int dst_pitch,
	int rows)
{
	const uint8	*s = src;
	boolean		word_size = row_bytes>250;

	if (!dst)
		dst_pitch = 0; /* skipping */
	do
	{
		int		z;
		const uint8	*se;
		uint16		v, *d = (uint16*)dst, *de = d;

		if (word_size)
			z = *s<<8|s[1], s+= 2;
		else
			z = *s++;
		se = s+z, ((uint8*)de)+= dst_pitch;
		while(s<se)
		{
			int8		c;

			if (src>=end)
				return -1;
			if ((c=*s++)<0)
			{ /* RLE */
				c =-c, ++c, v = va_be16(s);
				if (d+c>de) /* will overflow */
				{ /* clip, skip */
					do
						if (d<de)
							*d++ = v;
					while(--c);
					break;
				}
				do
					*d++ = v;
				while(--c);
				c = 0;
			} else
			{ /* uncompressed run */
				++c;
				if (d+c>de) /* will overflow */
				{ /* clip, skip */
					do
						if (d<de)
							*d++ = *s<<8|s[1];
					while(s+=2,--c);
					c = 0;
					break;
				}
				do
					*d++ = va_be16(s);
				while(--c);
			}
		}
		s = se;
	} while(dst+=dst_pitch,--rows);

	return s-src;
}

static int unpack_rle32(const uint8 *src,
	const uint8 *end,
	uint8 *dst,
	int row_bytes,
	int dst_pitch,
	int rows,
	int cmp_count)
{
	const uint8	*s = src;
	int		offset = cmp_count<4 ? (BIG_ENDIAN ? 1:2):BIG_ENDIAN ? 0:3;
	boolean		word_size = row_bytes>250;

//	assert(cmp_count>1 && cmp_count<5);
	if (!dst)
		dst_pitch = 0; /* skipping */
	do
	{
		int		z, cc = cmp_count;
		const uint8	*se, *re;
		uint8		v, *d = dst, *de = dst_pitch+d;

		if (word_size)
			z = *s<<8|s[1], s+= 2;
		else
			z = *s++;
		se = s+z, re = row_bytes+d, d+= offset;//4-cc:cc-1;
		while(s<se)
		{
			int8		c;

			if (src>=end)
				return -1;
			if ((c=*s++)<0)
			{ /* RLE */
				c =-c, ++c, v = *s++;
				do
				{
					if (d<de)
						*d = v;
					if ((d+=4)>=re)
					{
						if (cc<=1)
							{s = se; continue;}
						--cc, d-= row_bytes, d+= BIG_ENDIAN ? 1:-1;
					}
				} while(--c);
				c = 0;
				continue;
			}
			++c; /* uncompressed run */
			do
			{
				if (d<de)
					*d = *s;
				++s;
				if ((d+=4)>=re)
				{
					if (cc<=1)
						{s = se; continue;}
					--cc, d-= row_bytes, d+= BIG_ENDIAN ? 1:-1;
				}
			} while(--c);
		}
		s = se;
	} while(dst+=dst_pitch,--rows);
	return s-src;
}

static int unpack_pixels(const uint8 *src,
	const uint8 *end,
	uint8 *dst,
	uint16 depth,
	uint16 pack_type,
	int row_bytes,
	int dst_pitch,
	int rows)
{
	int		low_pitch = MIN(row_bytes,dst_pitch), data_size = 0;

	if ((low_pitch<=0 && dst_pitch) || rows<=0) /* following calls expect valid values */
		return 0;
	if (row_bytes<8 || pack_type==1)
	{ /* uncompressed */
		const uint8	*p = src;
		uint8		*q;

		data_size = row_bytes*rows;
		if ((q=dst)==NULL || data_size<=0)
			goto end;
		do /* SDL AO doesn't byteswap in rb<8 case */
#if !BIG_ENDIAN
			if (depth==16)
			{
				uint16		*d = (uint16*)q;
				const uint8	*s = p;
				int		c;

				if ((c=row_bytes>>1)>0)
					do
						*d++ = va_be16(s);
					while(--c);
				c = 0;
			} else if (depth==24 || depth==32)
			{
				uint32		*d = (uint32*)q;
				const uint8	*s = p;
				int		c;

				if ((c=row_bytes>>2)>0)
					do
						*d++ = va_be32(s);
					while(--c);
				c = 0;
			} else
#endif
			 memcpy(q,p,low_pitch);
		while(p+=row_bytes,q+=dst_pitch,--rows);
		rows = 0;
		goto end;
	}
	/* compressed */
	if (depth<=8) /* indexed */
		data_size = unpack_rle8(src,end,dst,row_bytes,dst_pitch,rows);
	else
	{ /* direct */
		if (!pack_type)
			if (depth<=16)
				pack_type = 3;
			else if (depth==24 || depth==32)
				pack_type = 4;
		if (pack_type==3) /* 15bpp */
			data_size = unpack_rle16(src,end,dst,row_bytes,dst_pitch,rows);
		else if (pack_type==4) /* 24/32bpp */
			data_size = unpack_rle32(src,end,dst,row_bytes,dst_pitch,rows,depth>>3);
		else
		{
			dprintf("unpack_pixels: bad pack type %d, depth %d;g",
				pack_type,depth);
			return -1;
		}
	}
	if (data_size<0)
		dprintf("unpack_pixels: data position exceeded end;g");
end:	return data_size;
}

static const uint8 *parse_ctab(const uint8 *src,
	const uint8 *end,
	cTab64 *ct,
	short depth)
{
	int		c, z;
	/* count is parsed from PICT *unsigned* */
	/* -1 (for 0) would actually load some 65536... and probably crash */
	/* (don't assume zero entries will work!) */
	c = (src[6]<<8|src[7])+1;
	z = c*8+16;
	if (src+z>=end)
		return NULL;
	if (ct)
	{
		ct->flags|= 0x8000;
	//	ct->count = c; /* need to check this */
		z = 256;
		if (depth)
			z = 1u<<depth;
		else if (ct->flags>0x80FF)
			z = ct->count;
		ct->seed = GetCTSeed();
	} else
		z =-1;
	return parseColorsBE(ct ? ct->colors:NULL,src+8,c,z,src[4]&0x80);
}

static int translate_QDmode(uint8 mode,
	boolean tx)
{
	static uint8	arithTab[8] =
		{mBlend,mAddPin,mAddOver,mSubPin,mTransparent,mAdMax,mSubOver,mAdMin};

	uint8		dither = mode&0x40;

	mode&= 63;
	if (mode&0x20) /* arithmetic */
	{
		if (mode==50) /* hilite (not currently available in mgraf) */
			return mSrcXor;
		if (mode==49) /* grayishTextOr */
			mode = mTransparent|mTextGray;
		else
			mode = arithTab[mode&7];
	} else /* basic */
		mode&= 7; /* mgraf patterns currently only work with 8-bit source */
	if (tx && mode==mSrcOr)
		mode = mTransparent;
	return mode|dither;
}

static int translate_QDFace(uint8 mode,
	ifixed *stylExt)
{
	*stylExt = ((mode>>6&1)-(mode>>5&1))<<16; /* extended, condensed */
	/* shadow not supported, but outlining is part of QD shadowing */
	return (mode&(_style_outline*2-1))|(mode>>1&_style_outline);
}

/* for the fields that openPort doesn't initialize */
/* also to refine data to better match QD PICT rendering */
static void initPortQDFields(grafPort *p,
	const rect16 *bounds)
{
	p->pnSize.x =~0;
	p->frSize.x = p->frSize.y = 1<<16;
	p->txMode = mTransparent;
	p->origin.x = bounds->left;
	p->origin.y = bounds->top;
	rTextFont(p,ffDefault);
}

static int expandPort(grafPort **port,
	const rect16 *fullBounds,
	rect16 *srcRect,
	const rect16 *oldSrcRect,
	const rect16 *newDstRect,
	SLRgn *clip,
	uint8 *mode,
	int depth,
	const rgb48 *eraseColor)
{
	grafPort	*p = *port;
	pixMap		*image = NULL;
	int		fullRight = fullBounds->right;
	int		fullBottom = fullBounds->bottom;
	int		r, fullDepth;
	int		m = p ? *mode&(mSrcPat-1):mSrcCopy;

	if (oldSrcRect)
	{
		if (oldSrcRect->right>fullRight)
			fullRight = oldSrcRect->right;
		if (oldSrcRect->bottom>fullBottom)
			fullBottom = oldSrcRect->bottom;
	}
	if (newDstRect)
	{
		if (newDstRect->right>fullRight)
			fullRight = newDstRect->right;
		if (newDstRect->bottom>fullBottom)
			fullBottom = newDstRect->bottom;
	}
	if (srcRect)
	{
		if (srcRect->right>fullRight)
			fullRight = srcRect->right;
		if (srcRect->bottom>fullBottom)
			fullBottom = srcRect->bottom;
	}
	fullDepth = p && (image=p->map)!=NULL ? image->bpp:32;
	if (fullDepth!=depth || (m && m!=mTransparent)) //m>mSrcXor)
		fullDepth = 32;
	if (p && oldSrcRect)
	{ /* need to maintain current port settings */
		rect16		oldcb;
		SLRgn		*clip;
		pixPat		*oldpa;
		int32		oldcz;
		uint32		oldfs;
		pixMap		oldpm = *image;

		if (m>=mSrcXor) /* was >mSrcXor */
		{ /* keeping old in stack eliminating need to create a new pixmap */
		  /* setting actual to empty to allocate new with erasure */
		  /* reduces memory requirements */
			image->baseAddr = NULL;
			image->pHand = NULL;
			image->pTable = NULL;
			image->h = 0;
		}
		if ((r=convertGrafPort(p,fullDepth,
				fullRight-fullBounds->left,
				fullBottom-fullBounds->top,
				oldpm.pTable,/*NULL*/
				eraseColor ? eraseColor:&rgb_black))<0)
		{
			if (m>=mSrcXor)
			{
				pmFree(p->map); /* just in case any allocation */
				/* dispose old pixels and table */
				hFree(oldpm.pHand);
				free(oldpm.pTable);
			}
			goto end;
		}
		if (m==mSrcCopy)
			return 1;
		if (m>=mSrcXor)
		{
			p->origin.x = fullBounds->left;
			p->origin.y = fullBounds->top;
			rBltPix(p,&oldpm,0,0,oldpm.w,oldpm.h,
				oldSrcRect->left,oldSrcRect->top,oldpm.w,oldpm.h,
				fullDepth>=24 ? m:*mode);
			/* dispose old pixels and table */
			hFree(oldpm.pHand);
			free(oldpm.pTable);
			r = 3;
			goto end;
		}
		/* can simply paint eraseColor in-place */
		clip = p->clipRgn;
		oldcz = clip->size;
		oldcb = clip->bBox;
		clip->size = RECT_RGN_SIZE;
		clip->bBox = *fullBounds;
		oldpa = p->pPat;
		oldfs = p->fgSolid;
		p->pPat = NULL;
		p->fgSolid = RGBToPen(p->map,eraseColor);
		rFillRectMode(p,0,0,oldpm.w,oldpm.h,m);
		if ((fullBounds->left^p->origin.x)|(fullBounds->top^p->origin.y))
		{
			SLRgn		*eraseRgn;

			p->origin.x = fullBounds->left;
			p->origin.y = fullBounds->top;
			eraseRgn = NULL;
			rScrollBlt(p,p->origin.x,p->origin.y,oldpm.w,oldpm.h,
				oldSrcRect->left,oldSrcRect->top,&eraseRgn);
			rRegion(p,eraseRgn);
		}
		clip->size = oldcz;
		clip->bBox = oldcb;
		p->pPat = oldpa;
		p->fgSolid = oldfs;
		r = 2;
		goto end;
	}
	if (!alloc_image(fullDepth,fullRight-fullBounds->left,
				fullBottom-fullBounds->top,port,
				eraseColor))
			return -2;
	p = *port;
	initPortQDFields(p,fullBounds);
	copyRgn(&p->clipRgn,clip);
	r = 1;
end:	*mode&= mDither;
	return r;
}

/* returns new src ptr */
/* note: b&w pics may render incorrectly */
static const uint8 *parse_bits(const uint8 *src,
	int size,
	grafPort **port,
	pixPat **pat,
	rect16 *srcRect,
	rect16 *fullBounds,
	SLRgn *clip,
	uint8 *mode,
	uint8 hasClip,
	boolean hasCTab,
	boolean *expanded,
	const rgb48 *eraseColor)
{
	rect16		dstRect, oldSrcRect = *srcRect;
	SLRgn		*clipRgn = NULL;
	const rect16	*clipBounds;
	const uint8	*end = src+size;
	platform_pixmap	image = NULL;
	long		data_size = 0;
	int		t, row_bytes, depth, width, height, img_mode; /* for this one image */
	uint16		pack_type, cmpCount;
	boolean		tmp_img = 0;

	row_bytes = va_be16(src); /* data rowBytes, bounds */
	srcRect->top = *src<<8|src[1];
	srcRect->left = src[2]<<8|src[3];
	srcRect->bottom = src[4]<<8|src[5];
	srcRect->right = src[6]<<8|src[7], src+= 8;
	if (src>=end)
		return NULL;
	width = RECTANGLE_WIDTH(srcRect);
	height = RECTANGLE_HEIGHT(srcRect);
	if (row_bytes&0x8000)
	{
		src+= 2;	// pmVersion - irrelevant (stored in PICT)
		pack_type = va_be16(src);
		src+= 4+4+4+2;	// dataSize, hRes, vRes, pixelType (chunky,chunkyPlanar,planar,0x10=RGBDirect)
		depth = va_be16(src);
		if (depth!=1 && depth!=2 && depth!=4 && depth!=8 && depth!=16 && depth!=32)
		{
			dprintf("parse_bits: bad pixel size (%d);g",
				depth);
			return NULL;
		}
		cmpCount = va_be16(src);
		src+= 2+4+4+4; // cmpSize, planeBytes, pmTable, pmReserved
	} else
		pack_type = 0, cmpCount = depth = 1;
	if (port)
	{
		if (*port && (image=(**port).map)!=NULL)
		{ /* more than one image - expand port, and blit */
			if (expanded && !*expanded)
			{
				if (expandPort(port,fullBounds,srcRect,&oldSrcRect,NULL,clip,
						mode,depth,eraseColor)<0)
					return NULL;
				*expanded = 1;
			}
			/* note that zero width/height is valid (render nothing) */
			if ((image=alloc_image(depth,MAX(width,1),MAX(height,1),NULL,
					eraseColor))==NULL)
				return NULL;
			tmp_img = 1;
		} else if ((image=alloc_image(depth,MAX(width,1),MAX(height,1),port,
				eraseColor))==NULL)
			return NULL;
	} else if (pat && rAllocPat(pat,depth,width,height,0)<0)
		return NULL;
	if (!(row_bytes&0x8000))
		hasCTab = 0;
	row_bytes&= 0x3FFF;
	if (hasCTab && (src=parse_ctab(src,end,image ? image->pTable
			:pat ? &(**pat).ct:NULL,image ? image->bpp:0))==NULL)
		{img_mode = 0; goto end;}
	img_mode = 0;
	if (!pat)
	{ /* blt src/dst */
		srcRect->top = *src<<8|src[1];
		srcRect->left = src[2]<<8|src[3];
		srcRect->bottom = src[4]<<8|src[5];
		srcRect->right = src[6]<<8|src[7], src+= 8;
		dstRect.top = *src<<8|src[1];
		dstRect.left = src[2]<<8|src[3];
		dstRect.bottom = src[4]<<8|src[5];
		dstRect.right = src[6]<<8|src[7];
		src+= 8+2;
		img_mode = translate_QDmode(src[-1],0);
	}
	if (src>=end)
		{src = NULL; goto end;}
	if (!tmp_img && port && memcmp(srcRect,&dstRect,8))
	{
		(**port).map = NULL;
		closeGrafPort(*port); free(*port);
		*port = NULL;
		if (expandPort(port,fullBounds,srcRect,NULL,&dstRect,clip,mode,32,
				eraseColor)<0)
		{
			DisposePixMap(image);
			return NULL;
		}
		tmp_img = 1;
	}
	clipBounds = NULL;

	if (hasClip)
	{
		t = *src<<8|src[1]; /* skip clip region */
		if (t>9)
		{
			clipBounds = (rect16*)(src+2);
			QDRgnToSLRgn(&clipRgn,src,t);
		}
		src+= t;
	}
	//dprintf(" %spackBits(%d,%d,%d,%d%s), rowb=%d, depth=%d/%d%s;g",
	//	hasCTab ? "color ":"",srcRect->top,srcRect->left,srcRect->bottom,srcRect->right,
	//	hasClip ? " (clipped)":"",row_bytes,depth,pack_type,image ? "":" - not rendering");
	if (src>=end)
		{src = NULL; goto end;}

	if ((data_size=unpack_pixels(src,end,image ? (uint8*)lockPixels(image)
			:pat ? (**pat).px:NULL,
			depth==32 && cmpCount==3 ? 24:depth,pack_type,row_bytes,
			image ? pmPitch(image)
			:pat ? (**pat).pitch:0,height))<0)
	{
	//	if (*image)
	//		{dispose_screen_image(*image); *image = NULL;}
		src = NULL; goto end;
	}
	if (image && srcRect->top|srcRect->bottom && srcRect->left|srcRect->right)
			//&& (**port).map!=image)
	{
	#if TARGET_OS_MAC
		(**tmp_image).bounds = *(const Rect*)srcRect;
	#else
		width = srcRect->right, height = srcRect->bottom;
		width-= image->pOrigin.x=srcRect->left;
		height-= image->pOrigin.y = srcRect->top;
		if (clipBounds)
		{
			t = be16_native(clipBounds->right)-srcRect->left;
			if (t<width)
				width = t;
			t = be16_native(clipBounds->bottom)-srcRect->top;
			if (t<height)
				height = t;
		}
		if (width>image->w)
			dprintf("parse_bits: source image suggests larger source width;g");
		else
			image->w = width;
		if (height>image->h)
			dprintf("parse_bits: source image suggests larger source height;g");
		else
			image->h = height;
	#endif
	}
end:	if (tmp_img)
	{
		rClippedBltPix(*port,image,
			0,0,width,height,dstRect.left,dstRect.top,
				RECTANGLE_WIDTH(&dstRect),
				RECTANGLE_HEIGHT(&dstRect),img_mode,
				clipRgn);
		free(clipRgn);
		if (mode)
			*mode&= mDither;
		DisposePixMap(image);
		*srcRect = *fullBounds;
	} else if (port)
	{
		if (clipRgn)
			sectRgn(&clipRgn,clip);
		else
			copyRgn(&clipRgn,clip);
		if (clipRgn && clipRgn->size>RECT_RGN_SIZE)
		{ /* erase clipped-out area */
			SLRgn		brgn;
			penState	oldps;
			uint32		oldfs = (**port).fgSolid;

			brgn.size = RECT_RGN_SIZE;
			brgn.bBox = clipRgn->bBox;
			xorRgn(&clipRgn,&brgn);
			rPenNormal(*port,&oldps);
			(**port).fgSolid = RGBToPen(image,eraseColor);
			rRegion(*port,clipRgn);
			rRestorePen(*port,&oldps,1);
			(**port).fgSolid = oldfs;
		}
		free(clipRgn);
		copyRgn(&(**port).clipRgn,clip);
		if (mode)
			*mode = img_mode;
	}
	return src ? src+data_size:NULL;
}

#ifdef QT_IMAGE
 #if 0//TARGET_OS_MAC
  #include <ImageCompression.h>

static ImageDescriptionHandle make_image_description(uint32 data_size,
	uint32 codec,
//	short framecount,
	const uint8 *name, /* Pascal */
	short depth,
	short width,
	short height)
{
	ImageDescriptionHandle	h;
	ImageDescription	*d;

	if ((h=(ImageDescriptionHandle)NewHandleClear(sizeof(ImageDescription)))==NULL)
		return NULL;
	d = *h;
	d->idSize = sizeof(*d);
	d->cType = codec;
	d->width = width;
	d->height = height;
	d->dataSize = data_size;
//	d->frameCount = 0;
	d->depth = depth;
	memcpy(d->name,name,32);

	return h;
}
 #else
static int decompressImage(grafPort **port,
	const uint8 *src,
	int32 idSize,
	uint32 dataSize,
	uint32 codec,
	const rect16 *srcRect,
	int dx,
	int dy,
	const rgb48 *opColor,
	uint8 mode)
{
	BMdcmpHeader	dhd;
	pixMap		*imap = NULL;
	srd_ptr		ptr;
	int32		r;

	ptr.src = src+idSize;
	ptr.remaining = dataSize-idSize;
	memset(&dhd,0,sizeof(dhd));
	dhd.w = RECTANGLE_WIDTH(srcRect);
	dhd.h = RECTANGLE_HEIGHT(srcRect);
	dhd.depth = dhd.bpp = src[82]<<8|src[83];
	if (dhd.w<=0 || dhd.h<=0 || dhd.depth+0u>32u)
		return -1;
	dhd.hRes = be32_native(*(const uint32*)(src+36));
	dhd.vRes = be32_native(*(const uint32*)(src+40));
	r = 0;
	switch(codec)
	{
	case MK4CC('W','R','L','E'):
		dhd.headerSize = 64;
		src+= 86;
		r = (int16)(src[-2]<<8|src[-1]);
		if (dhd.bpp>1 && dhd.bpp<9)
		{
			r = 1<<dhd.bpp;
			if ((dhd.ct=calloc(8,r+1))==NULL)
				return -2;
			dhd.ct->count = r;
			src = parse_ctab(src,ptr.src,dhd.ct,dhd.depth);
		} else if (++r>0)
			src+= 8*r;
		dhd.biPlanes = 1;
		dhd.rowOrder =-(src[20]==0); /* QT 7 only reads this byte */
		if ((const uint8*)ptr.src>src+16) /* if there's any space for bmCmp */
		{
			src = ptr.src;
			src-= idSize&1;
			dhd.biCompression = src[-12]<<8|src[-13];//src[10]<<8|src[11];
			/* some formats like bitfields have no way to encode, per QT PictViewer */
		}
		r = dcmpReadFileDIB(&imap,srd_sread,&ptr,&dhd);
		break;
	/* note that QT uses the PICT color table, not the GIF ctab */
	case MK4CC('g','i','f',' '):
		r = decompressGIF(&imap,srd_sread,&ptr,1,NULL);
		break;
	#ifdef QT_IJG9
	case MK4CC('j','p','e','g'):
		r = decompressJPEG(&imap,srd_sread,&ptr);
		break;
	#endif

	default:
		return -2003;
	}
	if (imap)
	{
		/* there's the potential of bad data/bounds producing read faults */
		/* in this case, QT succeeds but cuts off remaining PICT rendering */
		/* clipping here fixes this problem */
		if (dhd.w>imap->w)
			dhd.w = imap->w;
		if (dhd.h>imap->h)
			dhd.h = imap->h;
		if (!*port)
		{
			grafPort	*p;

			p = calloc(1,sizeof(**port));
			if (!p)
			{
				disposePixMap(imap);
				return -2;
			}
			*port = p;
			openGrafPort(p,dhd.depth,dhd.w,dhd.h,0);
			p->opColor = *opColor;
			disposePixMap(p->map);
			p->map = imap;
		} else
		{
			rBltPix(*port,imap,srcRect->left,srcRect->top,
				dhd.w,dhd.h,
				dx,dy,
				dhd.w,dhd.h,
				mode);
			disposePixMap(imap);
		}
	} else if (dhd.ct)
		free(dhd.ct);
	if (dhd.hRes>0 && dhd.vRes>0)
	{
		imap = (**port).map;
		imap->hRes = dhd.hRes;
		imap->vRes = dhd.vRes;
	}
	return r;
}
 #endif
#endif

static int vectGFX(grafPort **port,
	const rect16 *bounds,
	SLRgn *clip,
	rect16 *gfxRect,
	uint8 *mode,
	const rgb48 *eraseColor,
	const rgb48 *opColor,
	boolean drawing)
{
	grafPort	*p;
	int		width, height, t;

	if (*port)
	{ /* need to accomodate depth, mode */
		pixMap		*m = (**port).map;

		width = RECTANGLE_WIDTH(bounds);
		height = RECTANGLE_HEIGHT(bounds);
		if (m->w<width || m->h<height || (m->bpp<24 && drawing))
			return expandPort(port,bounds,NULL,gfxRect,NULL,clip,mode,32,
				eraseColor);
		return 0;
	}
	width = height = 1;
	if (gfxRect->top<gfxRect->bottom || drawing)
	{
		width = RECTANGLE_WIDTH(bounds);
		height = RECTANGLE_HEIGHT(bounds);
		if (clip)
		{
			t = RECTANGLE_WIDTH(&clip->bBox);
			if (t<4096 && t>width)
				width = MIN(t,width*2);
			t = RECTANGLE_HEIGHT(&clip->bBox);
			if (t<4096 && t>height)
				height = MIN(t,height*2);
		}
	}
	if (!alloc_image(32,width,height,port,eraseColor))
		return -2;
	p = *port;
	initPortQDFields(p,bounds);
	copyRgn(&p->clipRgn,clip);
	p->opColor = *opColor;
	if (gfxRect->top>=gfxRect->bottom)
	{
		gfxRect->left = gfxRect->top = 0;
		gfxRect->bottom = height;
		gfxRect->right = width;
	}
	return 1;
}

static void opMode(grafPort *port,
	uint16 opcode,
	pixPat *fillPat,
	pixPat *bkPat)
{
	switch(opcode&7)
	{
	case 0: if (port->pnSize.x<0) /* frame */
			port->pnSize.x = 0x10000;
		if (port->pnSize.y<0)
			port->pnSize.y = 0x10000;
		break;
	case 4: port->pPat = fillPat;
	case 1: port->pnSize.x = port->pnSize.y =~0;
		break;
	case 2: port->pPat = bkPat;
		port->pnMode = mSrcCopy;
		port->fgSolid = port->bkSolid;
		break;
	case 3: port->pPat = NULL;
		port->pnMode = mSrcXor, port->fgSolid = 0;
		/* remember to, if hilite gets implemented, reset hilite after invert */
	}
}

static void polyRtn(grafPort *port,
	uint16 opcode,
	const uint8 *s,
	int count,
	pixPat *fillPat,
	pixPat *bkPat)
{ /* UInt16 size; Rect bbox; Point[] */
//	rect16		b;
	penState	oldps;
	pt2dq		*d, *polygon;
	uint32		oldfs = port->fgSolid; /* mode may temporarily switch solid */
	int		c;

/*	b.top = (int16)(*s<<8|s[1]);
	b.left = (int16)(s[2]<<8|s[3]);
	b.bottom = (int16)(s[4]<<8|s[5]);
	b.right = (int16)(s[6]<<8|s[7]);
	printf("pict: poly b %d,%d,%d,%d, count %d\n",
		b.top,b.left,
		b.bottom,b.right,count);*/
	if ((c=count)<=0)
		return;
	if ((polygon=malloc(c*sizeof(pt2dq)))==NULL)
		return;
	d = polygon, s+= 8;
	do
	{
		d->x = ((int8)s[2]<<8|s[3])<<16|0x8000;
		d->y = ((int8)*s<<8|s[1])<<16|0x8000;
	} while(s+=4,++d,--c);

	/* the poly bounds are for measurements only; they do not clip rendering */
	rSavePen(port,&oldps);
	port->pPat = oldps.pPat;
	opMode(port,opcode,fillPat,bkPat);
	rPolygon(port,polygon,count);
	port->pPat = NULL; /* not disposing of pattern allocated above this function */
	port->fgSolid = oldfs;
	free(polygon);
	rRestorePen(port,&oldps,1);
}

static void rgnRtn(grafPort *port,
	uint16 opcode,
	const uint8 *s,
	int32 size,
	pixPat *fillPat,
	pixPat *bkPat)
{
	penState	oldps;
	SLRgn		*region = NULL;
	uint32		oldfs = port->fgSolid;

	QDRgnToSLRgn(&region,s,size);
	rSavePen(port,&oldps);
	opMode(port,opcode,fillPat,bkPat);
	if (opcode!=0x080) /* FrameRgn */
		rRegion(port,region);
	else if (port->pnSize.x>0 && port->pnSize.y>0) /* QD actually uses DiffRgn */
		rFrameRegion(port,region);
	port->pPat = NULL; /* not disposing of fillPat */
	port->fgSolid = oldfs;
	free(region);
	rRestorePen(port,&oldps,1);
}

static ufract QDAngleToU32(int angle)
{
	ufract		r;

	angle%= 360;
	if (angle<0)
		angle+= 360;
	r = 0;
	if (angle>=180)
		r = 0x80000000u, angle-= 180;
	return r+vfract30div(angle,90);
}

static void rectRtn(grafPort *port,
	uint16 opcode,
	const uint8 *src,
	rect16 *saveRect,
	pt2dq ovz,
	pixPat *fillPat,
	pixPat *bkPat)
{
	rect2dq		bq;
	penState	oldps;
	uint32		oldfs = port->fgSolid;
	ufract		circ;

	if (src && !(opcode&8))
	{
		saveRect->top = (int8)*src<<8|src[1];
		saveRect->left = (int8)src[2]<<8|src[3];
		saveRect->bottom = (int8)src[4]<<8|src[5];
		saveRect->right = (int8)src[6]<<8|src[7];
	}
	FixRect(&bq,saveRect->left,saveRect->top,saveRect->right,saveRect->bottom);
#if 1 /* need test image */
	if (!(opcode&7))
	{
		bq.top = addu(bq.top,0x10000);
		bq.left = addu(bq.left,0x10000);
		bq.bottom = subu(bq.bottom,0x10000);
		bq.right = subu(bq.right,0x10000);
	}
#endif
	rSavePen(port,&oldps);
	opMode(port,opcode,fillPat,bkPat);
	switch(opcode>>4)
	{
	case 3: rRect(port,&bq); break;
	case 4: rRoundRect(port,&bq,ovz.x,ovz.y); break;
	case 5: rOval(port,&bq); break;
	case 6:	src+=~opcode&8;
		if ((signed)(circ=(int8)src[2]<<8|src[3])<0)
			break;
		if (circ>359)
			circ =~0u;
		circ = QDAngleToU32(circ);
		rArc(port,&bq,QDAngleToU32(90-((int8)*src<<8|src[1]))-circ,circ,1);
	}
	port->pPat = NULL; /* not disposing of fillPat */
	port->fgSolid = oldfs;
	rRestorePen(port,&oldps,1);
}

#if 0 /* I still don't know how numerator/denominator play in */
static void QD_TxRatio(pt2d16 *numer,
	pt2d16 *denom)
{
	reduceFrac(&numer->x,&denom->x);
	reduceFrac(&numer->y,&denom->y);
}
#endif

platform_pixmap renderQDPict(const uint8 *start,
	int32 size,
	uint16 *srcWidth,
	uint16 *srcHeight,
	uint8 *prefMode,
	const rgb48 *eraseColor,
	void(*commentProc)(int kind,
		int size,
		const uint8 *data))
{
	grafPort	*port = NULL;
	pixPat		*fillPat, *bkPat;
	rect16		saveRect, bounds, srcRect, dstRect, maxClip;
	rgb48		color, opColor;
	SLRgn		*clip;
	pt2dq		ovz;
	const uint8	*src;
	ifixed		stylExt;
	int		pic_width, pic_height, data_size;
	int		txpx, txpy, t, polyct;
	uint16		opcode = 0, type, pnLocHFrac, qdchex;
	uint8		vers2, mode;
	boolean		renderedQT, imgHasQT, expanded, oldclns = mCLineSingle1px;

	if ((src=start)==NULL || size<12)
		return NULL;
	mode = mSrcCopy;
	ovz.x = ovz.y = 0;
	memset(&saveRect,0,sizeof(saveRect));
	memset(&srcRect,0,sizeof(srcRect));
	memset(&opColor,0,sizeof(opColor));
	pnLocHFrac = 0x8000, txpx = txpy = polyct = stylExt = 0, qdchex = 0;
	src+= 2; /* low 16 bits of size */

	/* To account for all pixels, the effective bounds is the union of clipped points.
	  But that's not performed here.

	  What *is* required is starting with an empty clip as QuickDraw does.
	*/
	clip = NULL;
	*(int32*)&maxClip.bottom = *(int32*)&maxClip = 0;
	fillPat = bkPat = NULL;
	mCLineSingle1px = 1;

	bounds.top = *src<<8|src[1];
	bounds.left = src[2]<<8|src[3];
	bounds.bottom = src[4]<<8|src[5];
	bounds.right = src[6]<<8|src[7], src+= 8;
	pic_width = RECTANGLE_WIDTH(&bounds);
	pic_height = RECTANGLE_HEIGHT(&bounds);
	if (pic_width<0)
		pic_width = 0;
	if (pic_height<0)
		pic_height = 0;
	if (srcWidth)
		*srcWidth = pic_width;
	if (srcHeight)
		*srcHeight = pic_height;
	if (!pic_width || !pic_height)
		return NULL;

	/* the version 2 sequence is supposed to be 0x1102 (and it works), but Apple eventually
	  went with 0x0011 so an old type-1 parser would take it as 0x00,0x11 = NOP,Version,
	  followed by a high byte outside [1,2]
	*/
	vers2 = renderedQT = imgHasQT = expanded = 0;
	while(src<start+size)
	{ /* parse the opcodes */
		if (vers2)
		{
			src = start+((src-start+1)&~1); /* word align */
			opcode = va_be16(src);
		} else
			opcode = *src++;
		//dprintf("%04x;g",
		//	opcode);
		switch(opcode)
		{
		case 0x011:	// VersionOp
			vers2 = *src++;
			if ((uint8)(vers2-1)>1u)
			{
				dprintf("renderQDPict: bad parsed version (%u);g",
					vers2);
				goto end;
			}
			vers2&= 2;
		case 0:		// NOP
			break;
		case 1:		// ClipRgn
			if ((t=*src<<8|src[1])>9)
			{ /* load up the clip, and expand the largest boundary */
			  	QDRgnToSLRgn(&clip,src+2,t);
				if (clip)
					UnionRect16(&maxClip,&clip->bBox);
				if (port && copyRgn(&port->clipRgn,clip)>0
						&& clip->size>RECT_RGN_SIZE)
				{ /* clip data won't be used again; keep frame */
					void		*tmp;

					clip->size = RECT_RGN_SIZE;
					if ((tmp=realloc(clip,RECT_RGN_SIZE))!=NULL)
						clip = tmp;
				}
			}
			src+= t;
			break;
		case 0x070:	// FramePoly
		case 0x071:	// PaintPoly
		case 0x072:	// ErasePoly
		case 0x073:	// InvertPoly
		case 0x074:	// FillPoly
			t = *src<<8|src[1];
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,1)>=0)
				polyRtn(port,opcode,src+2,polyct=(t-10)>>2,fillPat,bkPat);
			src+= t;
			break;
		case 0x075:	// reserved poly
		case 0x076:
		case 0x077:
		case 0x085:	// reserved rgn
		case 0x086:
		case 0x087:
			src+= *src<<8|src[1];
			break;
		case 0x080:	// FrameRgn
		case 0x081:	// PaintRgn
		case 0x082:	// EraseRgn
		case 0x083:	// InvertRgn
		case 0x084:	// FillRgn
			t = *src<<8|src[1];
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,1)>=0)
				rgnRtn(port,opcode,src+2,t,fillPat,bkPat);
			src+= t;
			break;
			/*
			    SameRgn/SamePoly were not implemented properly in QuickDraw.
			  QuickTime does nothing; in System 7, FrameSamePoly results in
			  seemingly random lines (and a crash).  The Classic QD source
			  indicates that the code loaded a Handle regardless of bit #3.
			    These opcodes should be never be used.
			*/
		case 0x078:	// FrameSamePoly
		case 0x079:	// PaintSamePoly
		case 0x07A:	// EraseSamePoly
		case 0x07B:	// InvertSamePoly
		case 0x07C:	// FillSamePoly
		case 0x07D:	// reserved
		case 0x07E:
		case 0x07F:
		case 0x088:	// FrameSameRgn
		case 0x089:	// PaintSameRgn
		case 0x08A:	// EraseSameRgn
		case 0x08B:	// InvertSameRgn
		case 0x08C:	// FillSameRgn
		case 0x08D:	// reserved
		case 0x08E:
		case 0x08F:
			break;
		case 10:	// FillPat
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
				rMakePat(&fillPat,src,&port->fgColor,&port->bkColor);
			src+= 8;
			break;
		case 9:		// PnPat
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
				rMakePat(&port->pPat,src,&port->fgColor,&port->bkColor);
			src+= 8;
			break;
		case 2:		// BkPat
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
				rMakePat(&bkPat,src,&port->bkColor,&port->fgColor);
		case 0x010:	// TxRatio
			src+= 8;
			break;
		case 3:		// TxFont
			vectGFX(&port,&bounds,clip,&srcRect,&mode,
				eraseColor,&opColor,0);
			(void)rTextFont(port,*src<<8|src[1]);
			src+= 2;
			break;
		case 4:		// TxFace
			vectGFX(&port,&bounds,clip,&srcRect,&mode,
				eraseColor,&opColor,0);
			(void)rTextStyle(port,translate_QDFace(src[1],&stylExt));
			if (port)
				port->chExtra = qdchex*port->txFace.size+stylExt;
			src+= 2;
			break;
		case 5:		// TxMode
			if (vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
				port->txMode = translate_QDmode(src[1],1);
			src+= 2;
			break;
		case 6:		// SpExtra
			if (vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
				port->spExtra = va_be32(src);
			else
				src+= 4;
			break;
		case 7:		// PnSize
			if (vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
			{
				port->pnSize.x = ((int8)*src<<8|src[1])<<16;
				port->pnSize.y = ((int8)src[2]<<8|src[3])<<16;
				if (port->pnSize.x<=0 || port->pnSize.y<=0)
					port->pnSize.x = port->pnSize.y =~0;
				else if (port->pnSize.x==0x10000 && port->pnSize.y==0x10000)
					port->pnSize.x = port->pnSize.y =~0<<16;
			}
			src+= 4;
			break;
		case 8:		// PnMode
			if (vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
				port->pnMode = translate_QDmode(src[1],0), src+= 2;
			break;
		case 11:	// OvSize - Point
			ovz.y = (va_be16(src)<<15)-0x8000;
			ovz.x = (va_be16(src)<<15)-0x8000;
			break;
		case 12:	// Origin - delta dh/dv - adjusts pict frame
			if (port) /* this is needed */
				vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0);
			OffsetRect16(&bounds,(int16)((int8)*src<<8|src[1]),
				(int16)((int8)src[2]<<8|src[3]));
			if (port)
			{ /* ClarisWorks adjusts the origin every time it places an image... */
				port->origin.x = bounds.left;
				port->origin.y = bounds.top;
			}
			/* move text position? */
			src+= 4;
			break;
		case 14:	// FgColor
		case 15:	// BgColor
			/* For an active draw, format is destination fmt for a CPort, */
			/* but for a classic QuickDraw port, this is a CMYKRGBIN field */
			/* where QD really only reads the CMY and N (normal) bits. */
			/* QuickTime seems to parse this as a classic color anyway. */
			vectGFX(&port,&bounds,clip,&srcRect,&mode,
				eraseColor,&opColor,0);
			t = va_be32(src);
			color = QD1ColorToRGB(t);
			_rgbSetFBSolid(port,&color,-(opcode&1));
			break;
		case 13:	// TxSize
			vectGFX(&port,&bounds,clip,&srcRect,&mode,
				eraseColor,&opColor,0);
			(void)rTextSize(port,*src<<8|src[1]);
			if (port)
				port->chExtra = qdchex*port->txFace.size+stylExt;
			src+= 2;
			break;
		case 0x012:	// BkPixPat
		{
			pixPat		**ppat;

			ppat = &bkPat;
			goto prspat;
		case 0x013:	// PnPixPat
			if (vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
				{ppat = &port->pPat; goto prspat;}
		case 0x014:	// FillPixPat
			ppat = &fillPat;
		prspat:	type = va_be16(src);
			if (type==2)
			{ // ditherPattern - a type 2 pattern a la MakeRGBPat
				color.red = src[8]<<8|src[9];
				color.green = src[10]<<8|src[11];
				color.blue = src[12]<<8|src[13];
				rMakeRGBPat(ppat,src,&color);
				src+= 8+6; // Pattern/RGBColor
			} else
			{
				if (type!=1)
					dprintf("renderQDPict: unusual pattern type (%u);g",
						type);
			/* we're always drawing to a color port, so classic Pattern is ignored */
				src+= 8;
				if ((src=parse_bits(src,size,NULL,ppat,
						&srcRect,&bounds,clip,
						NULL,0,1,NULL,NULL))==NULL)
					goto end;
			}
			break;
		}
		case 0x015:	// PnLocHFrac - uint16 - really TxLocHFrac
			pnLocHFrac = va_be16(src);
			break;
		case 0x016:	// ChExtra - IM ch.4: product of 4:12 fixed-pt and size
			qdchex = va_be16(src);
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0))
				port->chExtra = qdchex*port->txFace.size+stylExt;
			break;
		case 0x017:	// reserved
		case 0x018:
		case 0x019:
			break;
		case 0x01A:	// RGBFgCol
		case 0x01B:	// RGBBkCol
			vectGFX(&port,&bounds,clip,&srcRect,&mode,
				eraseColor,&opColor,0);
			color.red = *src<<8|src[1];
			color.green = src[2]<<8|src[3];
			color.blue = src[4]<<8|src[5];
			_rgbSetFBSolid(port,&color,-(opcode&1));
			src+= 6;
		case 0x01C:	// HiliteMode
		case 0x01E:	// DefHilite (reset hiliteColor)
			break;
		case 0x01F:	// OpColor
			//if (!vers2)
			//	dprintf("renderQDPict: RGB color op in version #1 PICT;g");
			opColor.red = *src<<8|src[1];
			opColor.green = src[2]<<8|src[3];
			opColor.blue = src[4]<<8|src[5];
			if (port)
				port->opColor = opColor;
		case 0x01D:	// HiliteColor
			src+= 6;
			break;
		case 0x020:	// Line
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,port ? !(uint16)port->pnSize.x:0)>=0)
			{
				FixPt(&port->pnLoc,(int8)src[2]<<8|src[3],(int8)*src<<8|src[1]);
				rLineTo(port,((int8)src[6]<<8|src[7])<<16|0x8000,
					((int8)src[4]<<8|src[5])<<16|0x8000);
			}
			src+= 8;
			break;
		case 0x021:	// LineTo - Point
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,port ? !(uint16)port->pnSize.x:0)>=0)
				rLineTo(port,((int8)src[2]<<8|src[3])<<16|0x8000,
					((int8)*src<<8|src[1])<<16|0x8000);
			src+= 4;
			break;
		case 0x022:	// ShortLine - Point/byte dh,dv
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,port ? !(uint16)port->pnSize.x:0)>=0)
			{
				FixPt(&port->pnLoc,(int8)src[2]<<8|src[3],(int8)*src<<8|src[1]);
				rLineTo(port,((int8)src[4]<<16)+port->pnLoc.x,
					((int8)src[5]<<16)+port->pnLoc.y);
			}
			src+= 6;
			break;
		case 0x023:	// ShortLineFrom - byte dh,dv
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,port ? !(uint16)port->pnSize.x:0)>=0)
				rLineTo(port,((int8)*src<<16)+port->pnLoc.x,
					((int8)src[1]<<16)+port->pnLoc.y);
			src+= 2;
			break;
		case 0x028:	// LongText
			/* none of the pict text routines alter the pnLoc */
			txpx = (int8)src[2]<<8|src[3];
			txpy = (int8)*src<<8|src[1];
			src+= 4;
			goto text;
		case 0x029:	// DHText
			txpx+= (int8)*src, ++src;
			goto text;
		case 0x02B:	// DHVText
			txpx+= (int8)*src, ++src;
		case 0x02A:	// DVText
			txpy+= (int8)*src, ++src;
		text:	if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,1)>=0)
				rDrawText(port,(const char*)src+1,*src,txpy,txpx<<16|pnLocHFrac);
			src+= *src++; // pstring
			break;
		case 0x02D:	// LineJustify
			type = va_be16(src);
			if (type!=8)
			{
				dprintf("renderQDPict: bad line-justify code (%u);g",
					type);
				goto end;
			}
			src+= 4+4; // spacing/extra just-spacing
			break;
		case 0x030:	// FrameRect
		case 0x031:	// PaintRect
		case 0x032:	// EraseRect
		case 0x033:	// InvertRect
		case 0x034:	// FillRect
		case 0x035:	// reserved
		case 0x036:
		case 0x037:
		case 0x038:	// FrameSameRect
		case 0x039:	// PaintSameRect
		case 0x03A:	// EraseSameRect
		case 0x03B:	// InvertSameRect
		case 0x03C:	// FillSameRect
		case 0x03D:	// reserved
		case 0x03E:
		case 0x03F:
		case 0x040:	// FrameRRect
		case 0x041:	// PaintRRect
		case 0x042:	// EraseRRect
		case 0x043:	// InvertRRect
		case 0x044:	// FillRRect
		case 0x045:	// reserved
		case 0x046:
		case 0x047:
		case 0x048:	// FrameSameRRect
		case 0x049:	// PaintSameRRect
		case 0x04A:	// EraseSameRRect
		case 0x04B:	// InvertSameRRect
		case 0x04C:	// FillSameRRect
		case 0x04D:	// reserved
		case 0x04E:
		case 0x04F:
		case 0x050:	// FrameOval
		case 0x051:	// PaintOval
		case 0x052:	// EraseOval
		case 0x053:	// InvertOval
		case 0x054:	// FillOval
		case 0x055:	// reserved
		case 0x056:
		case 0x057:
		case 0x058:	// FrameSameROval
		case 0x059:	// PaintSameROval
		case 0x05A:	// EraseSameROval
		case 0x05B:	// InvertSameROval
		case 0x05C:	// FillSameROval
		case 0x05D:	// reserved
		case 0x05E:
		case 0x05F:
		case 0x060:	// FrameArc
		case 0x061:	// PaintArc
		case 0x062:	// EraseArc
		case 0x063:	// InvertArc
		case 0x064:	// FillArc
		case 0x065:	// reserved
		case 0x066:
		case 0x067:
		case 0x068:	// FrameSameArc - angle start/circum.
		case 0x069:	// PaintSameArc
		case 0x06A:	// EraseSameArc
		case 0x06B:	// InvertSameArc
		case 0x06C:	// FillSameArc
		case 0x06D:	// reserved
		case 0x06E:
		case 0x06F:
			if (!renderedQT && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,1)>=0)
				rectRtn(port,opcode,src,&saveRect,ovz,fillPat,bkPat);
			src+=~opcode&8;
			if (opcode>0x05F)
				src+= 4;
			break;
		case 0x09A:	// DirectBitsRect
		case 0x09B:	// DirectBitsRgn
			src+= 4; /* whole PixMap header (includes baseAddr) */
		case 0x090:	// BitsRect
		case 0x091:	// BitsRgn
		case 0x098:	// PackBitsRect
		case 0x099:	// PackBitsRgn
			if ((src=parse_bits(src,size,renderedQT ? NULL:&port,NULL,
					&srcRect,&bounds,clip,
					&mode,opcode&1,opcode<0x09A,&expanded,
					eraseColor))==NULL)
				goto end;
			break;
		case 0x02E:	// glyphState: outlinePref.b, preserveGlyph.b
				//	fractEnable.b, fScaleDisabled.b
			t = va_be16(src);
			if (t>3 && vectGFX(&port,&bounds,clip,&srcRect,&mode,
					eraseColor,&opColor,0)>=0)
				rSetFScaleDisable(port,src[2+3]);
			src+= t;
			break;
		case 0x02C:	// FontName - potentially complicated 'association' list
		case 0x02F:	// reserved
		case 0x092:	// reserved
		case 0x093:
		case 0x094:
		case 0x095:
		case 0x096:
		case 0x097:
		case 0x09C:	// reserved
		case 0x09D:
		case 0x09E:
		case 0x09F:
			src+= va_be16(src);
			break;
		case 0x0A0:	// ShortComment
			if (commentProc)
				commentProc(*src<<8|src[1],0,0);
			src+= 2;
			break;
		case 0x0A1:	// LongComment
			t = src[2]<<8|src[3];
			if (commentProc)
				commentProc(*src<<8|src[1],t,src+4);
			src+= t, src+= 4;
		case 0x0A2:
		case 0x0A3:
		case 0x0A4:
		case 0x0A5:
		case 0x0A6:
		case 0x0A7:
		case 0x0A8:
		case 0x0A9:
		case 0x0AA:
		case 0x0AB:
		case 0x0AC:
		case 0x0AD:
		case 0x0AE:
		case 0x0AF:
		case 0x0B0:
		case 0x0B1:
		case 0x0B2:
		case 0x0B3:
		case 0x0B4:
		case 0x0B5:
		case 0x0B6:
		case 0x0B7:
		case 0x0B8:
		case 0x0B9:
		case 0x0BA:
		case 0x0BB:
		case 0x0BC:
		case 0x0BD:
		case 0x0BE:
		case 0x0BF:
		case 0x0C0:
		case 0x0C1:
		case 0x0C2:
		case 0x0C3:
		case 0x0C4:
		case 0x0C5:
		case 0x0C6:
		case 0x0C7:
		case 0x0C8:
		case 0x0C9:
		case 0x0CA:
		case 0x0CB:
		case 0x0CC:
		case 0x0CD:
		case 0x0CE:
		case 0x0CF:
			break;
		case 0x0D0:
		case 0x0D1:
		case 0x0D2:
		case 0x0D3:
		case 0x0D4:
		case 0x0D5:
		case 0x0D6:
		case 0x0D7:
		case 0x0D8:
		case 0x0D9:
		case 0x0DA:
		case 0x0DB:
		case 0x0DC:
		case 0x0DD:
		case 0x0DE:
		case 0x0DF:
		case 0x0E0:
		case 0x0E1:
		case 0x0E2:
		case 0x0E3:
		case 0x0E4:
		case 0x0E5:
		case 0x0E6:
		case 0x0E7:
		case 0x0E8:
		case 0x0E9:
		case 0x0EA:
		case 0x0EB:
		case 0x0EC:
		case 0x0ED:
		case 0x0EE:
		case 0x0EF:
		case 0x0F0:
		case 0x0F1:
		case 0x0F2:
		case 0x0F3:
		case 0x0F4:
		case 0x0F5:
		case 0x0F6:
		case 0x0F7:
		case 0x0F8:
		case 0x0F9:
		case 0x0FA:
		case 0x0FB:
		case 0x0FC:
		case 0x0FD:
		case 0x0FE:
			t = 4+va_be32(src);
			src+= t;
			break;
		case 0x0FF:	// OpEndPic
			goto end;

		default:
			if (opcode<0x8000)
			{
				if (opcode==0xC00)	// HeaderOp
				{
					int16		v;

					switch((int8)*src<<8|src[1])
					{
					case -1: /* fill.w=-1, fixed-point bounding box */
						break;
					case -2: /* fill.w, fixed-point hRes, vRes, srcRect */
						t = (int16)(src[14]<<8|src[15]);
						v = (int16)(src[12]<<8|src[13]);
						OffsetRect16(&bounds,t-bounds.left,v-bounds.top);
						if (srcWidth)
						{
							t = (int16)(src[18]<<8|src[19])-t;
							*srcWidth = MAX(t,0);
						}
						if (srcHeight)
						{
							t = (int16)(src[16]<<8|src[17])-v;
							*srcHeight = MAX(t,0);
						}
					}
				}
				src+= (opcode>>8)*2;
			} else if (opcode<0x8100)
				;
			else if ((opcode&~1)==0x8200)
			{ /* QuickTime */
			#ifdef QT_IMAGE
				int		error;
			#endif
				uint32		opcode_size, opcode_start, matte_size, mask_size;
				uint32		codec;
				int32		id_start, id_size;
				int		iwid, ihgt;

				opcode_size = va_be32(src);
				opcode_start = src-start;
				src+= 26;	// version/matrix - taking contained offset
				dstRect.left = *src<<8|src[1], src+= 2+2;
				dstRect.top = va_be16(src);
				matte_size = 0;
				if (!(opcode&1)) // completely guessing on 0x8201
				{
					src+= 6;
					matte_size = va_be32(src);
				}
				src+= 10;	// matteRec/mode (QT appears to ignore mode too)
				srcRect.top = *src<<8|src[1];
				srcRect.left = src[2]<<8|src[3];
				srcRect.bottom = src[4]<<8|src[5];
				srcRect.right = src[6]<<8|src[7];
				src+= 8;	// drect,accuracy
				if (!(opcode&1))
				{
					src+= 4;
					mask_size = va_be32(src);

					if (matte_size)
						data_size = va_be32(src)-4+matte_size, src+= data_size;
					src+= mask_size;
				}
				id_start = src-start;
				id_size = va_be32(src);
				codec = *(const int32*)src;
				/* QT ignores vendor/appl; header is consistent */
				src+= 4 +4+2+2+2+2+4+4+4; // resvd1/2/dataRefIx/ver/rev/vendor/temporalQual/spatialQual
				iwid = be16_native(*(const uint16*)src);
				ihgt = be16_native(((const uint16*)src)[1]);
				src+= 4 +8; //hRes/vRes
				data_size = va_be32(src);
			#if 0
				if (!imgHasQT)
				{
					dprintf("QT PICT;g");
					imgHasQT = 1;
				}
				dprintf(" img @ %d,%d: width %d/%d, height %d/%d, depth %u, codec '%-4.4s'-%.*s;g",
					dstRect.left,dstRect.top,iwid,dhd.w,ihgt,dhd.h,dhd.depth,
					&codec,MIN(src[2],31),src+3);
			#endif
				/* some PICTs have a message appended in case the QT image cannot be parsed */
			#ifdef QT_IMAGE
				/* instead of scaling, QT just computes minima bounds */
			 #if 1
				t = RECTANGLE_WIDTH(&srcRect);
				if (iwid>t)
					iwid = t;
				t = RECTANGLE_HEIGHT(&srcRect);
				if (ihgt>t)
					ihgt = t;
				srcRect.bottom = srcRect.top+ihgt;
				srcRect.right = srcRect.left+iwid;
			 #endif
				dstRect.bottom = dstRect.top+ihgt;
				dstRect.right = dstRect.left+iwid;
				if (id_start+id_size>size)
					{error =-39; goto qterr;}
				/* the QT call */
				error = 0;
				if ((!port && !memcmp(&srcRect,&dstRect,8) && !memcmp(&srcRect,&bounds,8))
						|| vectGFX(&port,&bounds,clip,&srcRect,&mode,
						eraseColor,&opColor,1)>=0)
					error = decompressImage(&port,
							start+id_start,id_size,
							opcode_start+opcode_size,
							codec,
							&srcRect,
							dstRect.left,dstRect.top,
							&opColor,
							mSrcCopy);
				if (error<0)
				{
				qterr:	dprintf("renderQDPict: error #%d for QT image;g",
						error);
					if (error==-2003)
					{
						dprintf(" can't find handler;g");
						goto qtskip;
					} else if (error==-8969)
					{
						dprintf(" bad format;g");
						goto qtskip;
					}
				}
				renderedQT = 1; /* override "need to install QT" message */
				//return image;
			qtskip:
			#endif
				src = start+opcode_start+opcode_size;
			} else if (opcode>=0x8100)
				src+= va_be32(src);
		}
	}
end:	if (opcode!=0xFF)
		dprintf("renderQDPict: abrupt termination - opcode %04X;g",
			opcode);
	if (prefMode)
		*prefMode = mode;
	if (fillPat)
		rDisposePat(fillPat);
	if (bkPat)
		rDisposePat(bkPat);
	mCLineSingle1px = oldclns;
	if (port)
	{ /* extract pixmap and close port */
		platform_pixmap	pixmap = port->map;
		port->map = NULL;
		closeGrafPort(port); free(port);
		if (!_heap_validate(NULL))
			dprintf("renderQDPict: something broke heap space");
		return pixmap;
	}
	return NULL;
}
