/*
 main.c
  ResEx tool shell layer

  Revised: 2023.08.15

2016.07.13

2014.02.28
  added 8027, 802E-F, C001-3, FWID, KCN#, KSWP, MENA, alrx, dftb, dlgx,
dlle, fctb, flac, hovr, kcl4/8, kcs8, mcfg, mgrp, mime, mitq, pcki, pslt,
resf, rsmi, stg#, src#, tab#, thga, thn#, thnd, thnr, trk#, xmnu & wind.
  fixed 802A - rComment, from Finder info (the only use I saw).
2021.09.09
  handling of MacBinary II files, including updating.
2022.09.14
  only type option, and PICT to pct.
*/

#include "my_cseries.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <io.h>

#include "resex.h"
#include "interpreter.h"
#include "rType.h"

/* ---- globals */

char		temporary[512];

/* ---- library */

#define NO_GUI
#define NO_MGRAF
#define NO_MEMEXT
#include "../utility/interpreter.c"

#define NO_TABLES
#include "../cseries.c"

double_t scal10(double_t v,
	int e)
{
	exit(1); /* should not get here */
}

int dprintf(const char *form,
	...)
{
	va_list		args;
	int		r;

	va_start(args,form);
	r = vfprintf(stderr,form,args);
	va_end(args);
	return r;
}

static int MB_RF_update(int fd,
	int size)
{
	int64		r, o = _lseek64(fd,0,SEEK_CUR);
	uint8		header[128];
	uint16		crc;

	(void)_lseek(fd,0,SEEK_SET);
	_read(fd,header,128);
	header[87] = (uint8)(size>>24);
	header[88] = (uint8)(size>>16);
	header[89] = (uint8)(size>>8);
	header[90] = (uint8)size;
	crc = MacBinaryGenCRC(header);
	header[124] = (uint8)(crc>>8);
	header[125] = (uint8)crc;
	(void)_lseek(fd,0,SEEK_SET);
	r = _write(fd,header,128);
	(void)_lseek64(fd,o,SEEK_SET);
	return r;
}

/* ---- main code */

static const char *getMapTypeName(ResFile_Memory *m)
{
	switch(m->type)
	{
	case TYPE_RFMACOS: return "Mac OS";
	case TYPE_RFIIGS: return "Apple IIgs";
	case TYPE_RSWIN32: return "Windows";
	default: return "???";
	}
}

void listMap(ResFile_Memory *m,
	uint32 type,
	uint8 printOffs)
{
	RsrcInfo	info;

	ResFile_TypeEntry	*l;

	const char	*s;
	int32		t;
	int		i, j, c, err;

	if (m->map.totalCount<=0)
		return;
	puts("=======================");
	printf(m->options&FAIL_RECONSTRUCT ?
"## abnormal map - list reconstructed\n"
"   index     size  %s\n\n"
		:" ix/abs    ID     size  %s flags name\n\n",
		printOffs ? "  offset":"");
	for(i=t=0,l=m->map.typeEntries; ++i<=m->map.typeCount; ++l)
	{
		if (type && type!=l->typeCode)
			continue;
		err = 0, c = ResFile_CountResources(m,l->typeCode,&err);
		if (err)
		{
			printf("## listMap: error getting type #%d\n",
				i);
			continue;
		}
		printf("- type #%-2d- ",
			i);
		if (m->type!=TYPE_RFIIGS)
			fputc('\'',stdout);
		print_tchars(stdout,(char*)&l->typeCode,4,'\'',1);
		if (m->type!=TYPE_RFIIGS)
			fputc('\'',stdout);
		printf((s=rTypeGetDesc(rTypeGetIndex(l->typeCode)))!=NULL ? " (%s)\n":"\n",
			s+4);
		for(j=0; j<c; ++t,++j)
		{
			err = 0, ResFile_GetIndResource(m,
				l->typeCode,~j,&info,&err);
			if (err)
			{
				printf("## listMap: error getting #%d\n",
					j);
				continue;
			}
			printf("%3d/%-3d ",
				j,t);
			printf(info.ID>65535 ? "$%X":"%6d",
				info.ID,info.sizeOnDisk);
			printf(printOffs ? " % 8d  %7.6X":" % 7d",
				info.sizeOnDisk,info.offsetOnDisk);
			printf(m->type!=TYPE_RFMACOS ? "  %04X":"  %02X",
				info.flag);
			if (*info.name)
			{
				fputs(" \"",stdout);
				print_tchars(stdout,(char*)info.name+1,*info.name,'"',1);
				fputc('"',stdout);
			}
			fputc('\n',stdout);
		}
	}
}

void printRsrc(ResFile_Memory *m,
	const RsrcInfo *info,
	FILE *tout)
{
	char		b[32];
	int32		z = 0, t = info->sizeOnDisk;

	fseek(m->filePtr,info->offsetOnDisk,SEEK_SET);
	while((t-=z)>0)
	{
		z = MIN(t,sizeof(b));
		fread(b,1,z,m->filePtr);
		fputs(" \"",tout);
		print_tchars(tout,b,z,'"',1);
		fputs("\"\n",tout);
	}
}

static void writeRsrcFile(ResFile_Memory *m,
	const RsrcInfo *info,
	int tx,
	int ix,
	FILE *tout)
{
	char		b[512];
	FILE		*f;
	int32		z = 0, t = info->sizeOnDisk;

	if (tout)
	{
		sprintf(temporary+2+*(int16*)&temporary,"%d-%d.%s",
			tx,ix,tx==RC_TYPE_TEXT ? "txt":"dat");
		fprintf(tout,"\n%d %d \"%s\"",
			ix,tx,temporary+2);
	} else
		sprintf(temporary+2+*(int16*)&temporary,"PICT %d ID %d.pct",
			ix,info->ID);
	if ((f=fopen(temporary+2,"wb"))==NULL)
	{
		fprintf(stderr,"failed to open '%s' for writing\n",
			temporary+2);
		return;
	}

	fseek(m->filePtr,info->offsetOnDisk,SEEK_SET);
	memset(b,0,sizeof(b));
	if (!tout)
		fwrite(b,1,MIN(sizeof(b),512),f);
	for(z=0; (t-=z)>0; )
	{
		z = MIN(t,sizeof(b));
		fread(b,1,z,m->filePtr);
		fwrite(b,1,z,f);
	}
}

void portMap(ResFile_Memory *m,
	const char *ddir,
	uint32 type,
	uint8 extdata)
{
	static FILE	*rcout;
	RsrcInfo	info;

	ResFile_TypeEntry	*l;

	int32		t;
	int		tx, i, j, c, laslg, err;

	i = ddir ? strlen(ddir):0;
	if (i>400)
	{
		fprintf(stderr,"## dest path too long\n");
		exit(1);
	}
	temporary[2+i] = '/';
	*(int16*)&temporary = ++i;
	if (m->map.totalCount<=0)
		return;
	if (m->options&FAIL_RECONSTRUCT)
		printf("## abnormal map - list reconstructed\n");
	fflush(stdout);
	memcpy(temporary+2,ddir,i-1);
	if (!rcout && !(extdata&4))
	{
		memcpy(temporary+2+i,"out.rc",7);
		rcout = fopen(temporary+2,"w");
		if (fputs(
"/*\n"
" out.rc\n"
"  generated by resEx\n"
"*/\n",
				rcout)<0)
		{
			fprintf(stderr,"## failed to start output file\n");
			exit(1);
		}
	}
	for(i=t=0,l=m->map.typeEntries; ++i<=m->map.typeCount; ++l)
	{
		if (type && type!=l->typeCode)
			continue;
		err = 0, c = ResFile_CountResources(m,l->typeCode,&err);
		if (err)
		{
			printf("## portMap: error getting type #%d\n",
				i);
			continue;
		}
		if ((tx=rTypeGetIndex(l->typeCode))<0)
			continue;
		if (rcout)
		{
			fprintf(rcout,"\n// type ",
				i);
			if (m->type!=TYPE_RFIIGS)
				fputc('\'',rcout);
			print_tchars(rcout,(char*)&l->typeCode,4,'\'',1);
			if (m->type!=TYPE_RFIIGS)
				fputc('\'',rcout);
		}
		//printf((s=rTypeGetDesc(tx))!=NULL ? " (%s)\n":"\n",
		//	s+4);
		tx+= 256, laslg = 0;
		/* info and short resource data (<256 bytes) stored in rcdata list */
		if (rcout)
			fprintf(rcout,"\n%d rcdata {%d,",
				tx,c);
		for(j=0; j<c; ++j)
		{
			err = 0, ResFile_GetIndResource(m,
				l->typeCode,~j,&info,&err);
			if (err)
			{
				fprintf(stderr,"## portMap: error getting #%d\n",
					j);
				continue;
			}
			++t;
			if (rcout)
			{
				fprintf(rcout,"\n %d,\"\\%o\",",
					(uint16)info.ID,*info.name);
				if (*info.name)
				{
					fputc('"',rcout);
					print_tchars(rcout,1+(const char*)info.name,*info.name,'"',1);
					fputs("\",",rcout);
				}
				if (info.sizeOnDisk>255)
				{
					fprintf(rcout,"\"\\%o\\0\",\n",
						info.flag&255);
					laslg = c; continue;
				}
				fprintf(rcout,"\"\\%o\\%o\",",
					info.flag&255,info.sizeOnDisk);
			//	fputc('\n',rcout);
				printRsrc(m,&info,rcout);
			} else if (info.sizeOnDisk>255)
				{laslg = c; continue;}
		}
		if (rcout)
			fputs("}\n",rcout);
		if (!extdata)
			laslg = 0;
		for(j=0; j<laslg; )
		{
			err = 0, ResFile_GetIndResource(m,
				l->typeCode,~j,&info,&err);
			++j;
			if (err)
				continue;
			if (info.sizeOnDisk>255 || extdata&4)
				writeRsrcFile(m,&info,tx,j,rcout);
		}
		if (rcout)
			fputc('\n',rcout);
	}
	if (rcout)
		fprintf(rcout,"\n// %d resource(s) processed for file\n",
			t);
}

typedef struct options
{
	const char	*ddir;
	long int	offset;
	uint32		type;
	boolean		rsrc, info, list, update;
	uint8		port;
} options;

int process(const char *in,
	options *opt)
{
	FILE		*f = NULL;
	ResFile_Memory	*m = NULL;
	int32		dlen;
	int		fd =-1, err = 0, r = strlen(in), mb = 0;
	char		*n = malloc(r+15); /* excess for native RF appendage */

	if (!n)
		return -2;
	memcpy(n,in,r); n[r] = 0, r = 0;
	if (opt->rsrc)
	{
#if TARGET_OS_MAC
		if ((fd=_open(n,O_RDWR|O_BINARY|O_RSRC))==NULL)
			goto ferrno;
		if (fd<0)
			{err = fd; goto ferr;}
		f = fdopen(fd,"rb");
#endif
		memcpy(n+r,"/resource",10);
		r = 0;
	}
	if (!f)
	{
		f = fopen(n,"rb");
		fd = _fileno(f);
	}
	if (!f)
		goto ferrno;
#if TARGET_OS_MAC
	//Debugger();
#endif
	if (opt->offset<0)
	{
		opt->offset = 0;
		if ((r=MacBinaryDetect(fd,&dlen,NULL))>0)
		{
			opt->offset = r+((dlen+127)&~127);
			mb = 1;
			if (!opt->update)
				puts("# MacBinary file\n");
		}
	}
	if ((err=ResFile_LoadMap(f,TOPT_MAKERAW|TOPT_VERBOSE,&m,opt->offset))!=0)
	{
		fprintf(stderr,"### ResFile_LoadMap %s (%d)\n",
			err<0 ? "error":"warning",err);
		if (err<0)
			goto end;
	}
	if (opt->update)
	{
		f = freopen(n,"rb+",f);
		if (!f)
		{
		ferrno:	err = errno;
	#if TARGET_OS_MAC
		ferr:
	#endif
			fprintf(stderr,"### file error (%d)\n",
				err);
			if (err==ENOENT)
				fprintf(stderr," - '%s' could not be found\n",
					n);
			else if (err==EACCES)
				fprintf(stderr," - could not (re-)open '%s'\n",
					n);
			else if (err==EINVAL)
				fprintf(stderr," - illegal characters name '%s'\n",
					n);
			r = 1;
			goto end;
		}
		fd = _fileno(f);
		m->filePtr = f;
		if ((err=ResFile_Update(m))!=0)
		{
			fprintf(stderr,"### update %s (%d)\n",
				err<0 ? "error":"warning",err);
			goto end;
		}
		(void)MB_RF_update(fd,m->mapPos+m->mapLen);
	}
	if (opt->port|opt->info)
	{
		int		enz, ovz = ResFile_OverheadSize(m,&enz);

		printf(
"# file: '%s'\n"
"# RF type: %s\n",
			n,getMapTypeName(m));
		if (m->type==TYPE_RFMACOS) /* don't know what of other types */
			printf("# attributes: %04X%s\n",
				m->map.attributes,
				m->map.attributes&RF_READ_ONLY_ATTR ? " (read-only)":"");
		printf(
"# map length: %d; per-rsrc overhead: %d+%d\n"
"# type count: %d; total: %d\n\n",
			m->mapLen,ovz,enz,m->map.typeCount,m->map.totalCount);
		if (opt->list)
			listMap(m,opt->type,opt->list-1);
		if (opt->port)
			portMap(m,opt->ddir,opt->type,opt->port&6);
	}

end:	if (f)
		fclose(f);
	if (n)
		free(n);
	if (m)
		free(m);
	return r;
}

int main(int argc,
	char **argv)
{
	options		opt;
	char		**a = argv+1, *s;

	if (argc<1)
		return 0;
	memset(&opt,0,sizeof(opt));
	opt.offset =~0;
	for(; --argc && *(s=*a)=='-'; ++a)
		if (!(++s)[1])
			if (*s=='i')
				opt.info^= 1;
			else if (*s=='l')
				opt.list = 1;
			else if (*s=='L')
				opt.list = 2;
			else if (*s=='o')
			{
				s = *++a;
				if (!--argc || !readanyinteger(s,strlen(s),&opt.offset))
				{
					fprintf(stderr,argc ? "### bad offset value '%s'\n"
						:"### no offset value\n",
						s);
					return 1;
				}
			} else if (*s=='r')
				opt.rsrc^= 1;
			else if (*s=='t')
			{
				s = *++a;
				if (!--argc)
				{
					fprintf(stderr,"## missing type value\n");
					return 1;
				}
				opt.type = *(uint32*)s;
				if (!*(int8*)&opt.type)
					*(int8*)&opt.type = ' ';
				if (!((int8*)&opt.type)[1])
					((int8*)&opt.type)[1] = ' ';
				if (!((int8*)&opt.type)[2])
					((int8*)&opt.type)[2] = ' ';
				if (!((int8*)&opt.type)[3])
					((int8*)&opt.type)[3] = ' ';
			} else if (*s=='u')
				opt.update^= 1;
			else if (*s=='w')
				opt.port = opt.port ? 0:1;
			else if (*s=='W')
				opt.port = opt.port ? 0:2;
			else if (*s=='d')
			{
				s = *++a;
				if (!--argc)
				{
					fprintf(stderr,"## missing dest\n");
					return 1;
				}
				opt.ddir = s;
			} else
				goto unk;
		else if (!s[2])
			if (*s=='W' && s[1]=='P')
			{
				opt.port = opt.port ? 0:4;
				opt.type = MK4CC('P','I','C','T');
			} else
				goto unk;
		else
		{
		unk:	fprintf(stderr,"### unknown option '%s'\n",
				s-1);
			goto usage;
		}
	opt.info|= opt.list;
	if (argc<1)
	{
	usage:	fprintf(stderr,
"# usage: %s [options] files\n"
"#\n"
"# -i            print info\n"
"# -l            print info and listing\n"
"# -L            .. listing with offsets\n"
"# -o            manual offset in file for RF\n"
"# -r            native RF (/resource)\n"
"# -t            this rsrc type only; padded w/ spaces\n"
"# -u            update\n"
"# -W            port to RC\n"
"# -w            port to RC (no data files)\n"
"# -WP           extract PICTs as .pct files\n"
"# -d            extraction destination dir.\n",
			*argv);
		return 1;
	}
	//cs_initialize(NULL);
	for(; argc; --argc,++a)
		if (process(*a,&opt))
			return 1;
	return 0;
}
