/*
 resex.c
  resource extraction (and recovery)

  Created: 2010.01.17
  Revised: 2023.09.25
   Author: Adam J. P.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

2016.07.22
*/

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <wchar.h>
#include <errno.h>

#include "my_cseries.h"

#include "error_codes.h"
#include "resex.h"

#define MAC_TYPES_OFFSET	(16+6)
#define MAC_RSRCENT_SIZE	12
#define IIGS_RSRCENT_SIZE	20

/* ---- general functions */

/* for reading any length, storing 1, 2, 4 or 8 bit integer
 (can also be used to ignore preceding bytes)
 -- Passing the negative of the length will force 32 bits
*/

/* a "remaining" macro; subtracts size */
#define sreadx(f,v,r)		(((r)-=sizeof(v))<0?_e_read_error:-(fread(&(v),sizeof(v),1,f)<=0))

#define sreadintxBE(f,v,r)	(((r)-=sizeof(v))<0?_e_read_error:freadintBE(f,&(v),sizeof(v)))

/*
static int writeintBE(FILE *f,
	long v,
	int z)
{
	if (!z)
		return 0;
	while(--z>=0)
		if (fputc(v>>8*z,f)<0)
			return errno==ENOENT ? errHitFileEOF:_e_read_error;
	return 1;
}*/

/* ---- utility */

/*
 getMapType
	Finds a type in the map.
  Result:
	The index, or -1.
*/
static int32 getMapType(ResFile_Memory *m,
	int32 type)
{
	ResFile_TypeEntry	*t = m->map.typeEntries;

	uint16		c;

	if ((c=m->map.typeCount)==0)
		return -1;
	do
		if (t->typeCode==type)
			return (int32)(t-m->map.typeEntries);
	while(++t,--c);

	return -1;
}

int ResFile_CountResources(ResFile_Memory *m,
	int32 type,
	int *r)
{
	int32		x;
	int		err = 0, c = 0;

	if (!m || !m->filePtr)
		err = _e_general_error;
	else if ((x=getMapType(m,type))<0)
		err = errNoRsrcMapEntry;
	else
		c = m->map.typeEntries[x].count;
	if (r)
		*r = err;
	return c;
}

int ResFile_OverheadSize(ResFile_Memory *m,
	int *enz)
{
	if (!m)
		return _e_general_error;
	if (enz)
		*enz = 0;
	switch(m->type)
	{
	case TYPE_RFMACOS:
		if (enz)
			*enz = MAC_RSRCENT_SIZE;
		return 4; /* size */
	case TYPE_RFIIGS:
		if (enz)
			*enz = IIGS_RSRCENT_SIZE;
	default: return 0;
	}
}

/*
 rsrcSeek
	fseeks to byte #2 of the resource.
  Result:
	The first two bytes skipped (be16_native), or error (negative).
*/
static int rsrcSeek(const ResFile_Memory *m,
	const ResFile_RsrcEntry *res)
{
	FILE		*f = m->filePtr;
	long int	offset = m->fileStart+m->dataPos+res->dataOffset;
	/* top byte is invalid - as part of length, checked <16MB */
	if ((m->type==TYPE_RFMACOS && res->dataOffset+4>m->mapPos)
			|| fseek(f,offset,SEEK_SET)<0)
		return errSeekBadPosition;
	return 0;
}

/*
 rsrcLen
	Calculates the length.
	Any additional reads may be performed for this operation.
  Result:
	The length or EOF error (-1).
*/
static int32 rsrcLen(const ResFile_Memory *m,
	const ResFile_RsrcEntry *res)
{
	int32		x;

	if (m->type==TYPE_RFMACOS || m->type==TYPE_RSWIN32)
	{ /* though Mac OS mhandles are 8MB max., allow for all sizes under 16MB */
		FILE	*f = m->filePtr;

		x = 0;
		freadintBE(f,&x,4);
		if (m->type==TYPE_RSWIN32)
		{
			int32		hz;

			freadintBE(f,&hz,4);
			hz = be32_native(hz);
			fseek(f,hz-8,SEEK_CUR);
			return be32_native(x);
		}
		if (x>0xFFFFFF)
		{
			if (m->options&TOPT_VERBOSE)
				(void)fprintf(stderr,"### rsrcSeek error @ %06X - bad stored size\n",
					ftell(f));
			return errReadBadEntryMap;
		}
		return x;
	}
	/* get the total absolute index and verify that it's in the map */
	x = (int32)(res-(ResFile_RsrcEntry*)(m->map.typeCount+m->map.typeEntries));
	if ((x-=m->map.totalCount)>=0)
		return _e_general_error;
	return ((int32*)(m->memSize+(int8*)m))[x];
}

/*
 getResource
	Locates the resource, given a name or ID (if name==NULL).
  Result:
	The descriptor pointer in the map, or NULL.
*/
static ResFile_RsrcEntry *getResource(ResFile_Memory *m,
	const ResFile_TypeEntry *t,
	const uint8 *name,
	int32 ID,
	int32 *ix)
{
	const uint8	*l = (const uint8*)(m->map.typeCount+m->map.typeEntries);

	ResFile_RsrcEntry	*res = t->firstRes+(ResFile_RsrcEntry*)l;

	uint16		i;

	if (name)
	{
		l+= sizeof(ResFile_RsrcEntry)*m->map.totalCount-1;
		for(i=0; i<t->count; ++i,++res) //was i<=t->count
		{
			const uint8	*s;
			int		o;

			if ((o=res->flag&RESNAME_OFFSET_MASK)!=0
					&& *(s=l+o)==*name
					&& !memcasecmp(1+(const char*)s,
					1+(const char*)name,*name))
				goto gotit;
		}
		i = 0;
	} else
		for(i=0; i<t->count; ++i,++res) //was i<=t->count
			if (res->ID==ID)
			{
			gotit:	if (ix)
					*ix = i;
				return res;
			}
	return NULL;
}

/*
 calcTypeOffsets
	This tiny function performs the task of renumbering firstRes indices.
	Call it whenever the list has been changed or imported.
*/
static void calcTypeOffsets(ResFile_TypeEntry *t,
	int c)
{
	uint32		o = 0;

	if (c<=0)
		return;
	do
		t->firstRes = (uint16)o, o+= t->count;
	while(++t,--c);
}

/*
 updateHeader
	Updates map lengths and calculates the file map offset to names.
  Result:
	The offset to names.
*/
static int32 updateHeader(ResFile_Memory *m)
{
	if (m->type!=TYPE_RFMACOS)
		return 0;
	/* contiguous */
	m->mapLen = m->memSize-OFFSET(ResFile_Memory,map)+MAC_TYPES_OFFSET;
	return MAC_TYPES_OFFSET+2+6
		+sizeof(ResFile_TypeEntry)*m->map.typeCount
		+sizeof(ResFile_RsrcEntry)*m->map.totalCount;
}

/* ---- Mac OS-specific functions */

/* unfortunately, Mac OS maps use 16-bit offsets throughout */
/* anything past max namesOffset + rsrc entry nameOffset is unreachable;
  add max name length, makes an offset for data that's completely unused
*/
#define MAX_MACOSRF_MAP		(65535*2+256)

static int MacRF_checkheader(ResFile_Memory *m,
	const uint8 *h)
{
	uint32		t[4];
	/* data pos/len read only for verification */
	const int32	*a = (const int32*)h;
	int32		dataLen;

	m->dataPos = va_be32(a);
	m->mapPos = va_be32(a);
	dataLen = (int32)va_be32(a);
	m->mapLen = (int32)va_be32(a);
	if (m->dataPos<16 || m->mapPos<m->dataPos
			|| dataLen!=m->mapPos-m->dataPos
			|| m->mapLen<30 || m->mapLen>MAX_MACOSRF_MAP)
		return 0; /* not a working Mac OS resource fork */

	/* the first 16 bytes of the map must match the heading 16 bytes */
	(void)fseek(m->filePtr,m->fileStart+m->mapPos,SEEK_SET);
	if (fread(t,sizeof(t),1,m->filePtr)!=1)
		return errno==ENOENT ? errHitFileEOF:_e_read_error;
	/* allow mismatch if the map length indicates 'empty' */
	/* give a warning if blank (applies to some improper maps made for Aleph One) */
	return !(*t|t[1]|t[2]|t[3]) ? 2:!memcmp(t,h,16) || m->mapLen==30;
}

/* temporary under 32 bit Mac OS (but ends up being stored on disk):
  mapChainHandle.l,
  refNum.w,
  attributes.b,
  reserved.b,
  typesOffset.w - normally 0x001C,
  namesOffset.w
*/

static int MacRF_loadmap(ResFile_Memory *m)
{
	uint8		b[12];

	ResFile_TypeEntry	*typeent;

	FILE		*f = m->filePtr;
	/* ignoring mptr/ref, remaining byte count avoid possible overrun (bad length) */
	long		r = m->mapLen-MAC_TYPES_OFFSET;
	int		v, c, err, uerr = 0, firstAdj;
	uint16		storetypesoffs, storenamesoffs;

	/* skip the stale stuff, and read the fixed first three values */
	if (fseek(f,m->fileStart+MAC_TYPES_OFFSET+m->mapPos,SEEK_SET)<0
			|| fread(b,6,1,f)!=1)
		return errno==ENOENT ? errHitFileEOF:_e_read_error;
	m->map.attributes = *b<<8|b[1];
	storetypesoffs = b[2]<<8|b[3];
	storenamesoffs = b[4]<<8|b[5];

	/* check and calculate lengths */

	/* the case of a bad names offset shouldn't happen;
	  copy map length (erases possible names, though...)
	*/
	if (storenamesoffs>m->mapLen
			|| storenamesoffs<MAC_TYPES_OFFSET)
	{ /* names surpasses end */
		storenamesoffs = (uint16)m->mapLen, uerr = 3;
		if (m->options&TOPT_VERBOSE)
			(void)fprintf(stderr,"## MacRF_loadmap: bad names offset\n");
		//return errReadBadEntryMap;
	}
	if (storetypesoffs<28 || storetypesoffs>=storenamesoffs-1)
	{
		if (m->options&TOPT_VERBOSE)
			(void)fprintf(stderr,"### MacRF_loadmap: bad types offset\n");
		return errReadBadEntryMap;
	}
	/* onto the types list, which starts with a DBF-based count.w */
	(void)fseek(f,m->fileStart+m->mapPos+storetypesoffs,SEEK_SET);
	r = storenamesoffs-storetypesoffs;
	if ((err=sreadintxBE(f,m->map.typeCount,r))<0)
		return err;
	if ((c=++m->map.typeCount)==0)
		return 0;  /* no resources */
	typeent = m->map.typeEntries;
	/* now allows any first reslist offset, but must be 12-byte sequential */
	firstAdj = 0;
	do
	{
		if ((err=sreadx(f,typeent->typeCode,r))<0
				|| (err=sreadintxBE(f,typeent->count,r))<0
				|| (err=sreadintxBE(f,typeent->firstRes,r))<0)
			return err;
		if ((!firstAdj && (firstAdj=typeent->firstRes)==0)
				|| typeent->firstRes<firstAdj)
			uerr = errReadBadEntryMap;
		m->map.totalCount+= ++typeent->count;
		/* no weird offsets */
		v = typeent->firstRes-firstAdj; /* now the reslist index */
		typeent->firstRes = (uint16)(v/12u); /* unsigned will red-flag negatives */
		if (v!=typeent->firstRes*MAC_RSRCENT_SIZE)
			uerr = errReadBadEntryMap, typeent->firstRes = (uint16)-1;
	} while(++typeent,--c);

	if (uerr<0 && m->options&TOPT_VERBOSE)
		(void)fprintf(stderr,"## MacRF_loadmap: reslist has bad data\n");

	if (r>=MAC_RSRCENT_SIZE) /* rrrghghgh... should be true either way */
	{
		ResFile_RsrcEntry	*desc = (ResFile_RsrcEntry*)typeent;

		long int	o = m->fileStart+m->mapPos+storetypesoffs+firstAdj;

		if ((o-=ftell(f))<0)
		{
			if (m->options&TOPT_VERBOSE)
				(void)fprintf(stderr,"## MacRF_loadmap: type/res lists intersect!\n");
			return errReadBadEntryMap;
		}
		if (o)
			r-= o, fseek(f,o,SEEK_CUR);
		c = 0, o = m->mapLen-storenamesoffs;
		do
		{ /* stored on disk: ID.s16, nameo.16, attr.8, datao.24, handle.32 */
			if (fread(b,MAC_RSRCENT_SIZE,1,f)!=1)
				return err;
			desc->ID = *(int8*)b<<8|b[1];
			if ((desc->flag=0xFFFF&(1+(b[2]<<8|b[3])))!=0
					&& desc->flag>=o+0u)
			{
				desc->flag = 0;
				if (!uerr)
					uerr = 3;
			}
			desc->flag|= b[4]<<18;
			desc->dataOffset = (b[5]<<8|b[6])<<8|b[7];
		} while(++desc,r-=MAC_RSRCENT_SIZE,++c!=m->map.totalCount
			&& r>=MAC_RSRCENT_SIZE);

		if (c!=m->map.totalCount)
		{
			if (m->options&TOPT_VERBOSE)
				(void)fprintf(stderr,"## MacRF_loadmap: bad reslist count\n");
			if (!uerr)
				uerr = 3;
		}
		r = m->fileStart+m->mapPos+storenamesoffs-ftell(f);
		if (r) /* truncated resource info? */
		{
			(void)fseek(f,r,SEEK_CUR);
			if (r<0)
			{
				if (m->options&TOPT_VERBOSE)
					(void)fprintf(stderr,"## MacRF_loadmap: res/names lists intersect!\n");
				if (!uerr)
					uerr = 3;
			}
		}
		/* finally, read names */
		if ((c=m->mapLen-storenamesoffs)>0)
			if (fread(desc,c,1,f)<=0)
				if (!uerr) /* hit EOF into names - not always critical */
					uerr = 1;
	}
	return err>=0 ? uerr:err;
}

/* a really hacky function,
 but necessary since ResEdit doesn't provide anything like it
*/
static int MacRF_makemap(ResFile_Memory **a)
{
	ResFile_Memory	*m = *a;
	FILE		*f = m->filePtr;
	long		pos;
	int32		len;
	int		err = 0;
	int16		count =~0;

	/* well, I guess this'll just overflow, and ResEdit will complain of dual IDs */
	/*int16		onecount =-1; // new type every time this overflows */

	if (m->dataPos<16)
		m->dataPos = 0x100;
	pos = m->dataPos;
	if (fseek(f,0,SEEK_END)<0) /* set mapPos to the file's EOF */
		return errno==ENOENT ? errHitFileEOF:_e_read_error;
	if (m->mapPos<pos+0u) /* never really tested range errors... */
		m->mapPos = ftell(f);
	m->map.typeCount = 1;
	m->map.typeEntries->typeCode = 0x3F3F3F3F; /* ???? */

	do
	{
		/* leading byte should be zero, but... it's useless, really */
		(void)fseek(f,m->fileStart+pos+1,SEEK_SET);
		if (freadintBE(f,&len,3)<0)
			{err = 3; break;} /* maybe, cleanly hit EOF? */
		if ((m=realloc(*a,sizeof(ResFile_Memory)
				+sizeof(ResFile_RsrcEntry)*(++count+1)))==NULL)
			{err = errOutOfMemory; break;}
		*a = m;
		{
			ResFile_RsrcEntry	*desc = count
				+(ResFile_RsrcEntry*)(1+m->map.typeEntries);

			desc->ID = count;
			desc->flag = 0;
			desc->dataOffset = pos-m->dataPos;
		}
	} while((pos+=4+len)+0u<m->mapPos && count<32767);

	m->map.totalCount = count;
	m->map.typeEntries->count = (uint16)count;
	m->mapLen = MAC_TYPES_OFFSET+sizeof(ResFile_TypeEntry)
		+sizeof(ResFile_RsrcEntry)*++count;

	return err;
}

static int MacRF_update(ResFile_Memory *m)
{
	uint32		h[8];
	int		fd = _fileno(m->filePtr), err = 0;
	uint32		v, t = 0, n = updateHeader(m);
	uint16		c;

	if ((int32)n>m->mapLen)
		n = m->mapLen; /* dumb-dumb dummy */

	{
		uint8		*d = 4+(uint8*)h;

		*h = be32_native(m->dataPos);
		h[1] = be32_native(m->mapPos);
		v = m->mapPos-m->dataPos;
		h[2] = be32_native(v);
		h[3] = be32_native(m->mapLen);
		h[5] = h[4] = 0; /* stale map handle, refnum */
		d = 22+(uint8*)h;
		v = m->map.attributes;
		*d++ = (uint8)(v>>8), *d++ = (uint8)v;
		*d++ = 0, *d++ = 0x1C; /* types offs (0x1C if ya know what's good for ya) */
		*d++ = (uint8)(n>>8), *d++ = (uint8)n;
		c = m->map.typeCount-1;
		*d++ = (uint8)(c>>8), *d++ = (uint8)c;
	}

	(void)_lseek(fd,m->fileStart,SEEK_SET);
	if (_write(fd,h,16)!=16)
		return errno==ENOSPC ? errVolumeFull:_e_write_error;
	/* fileStart was missing here!; 2014.02.11 */
	(void)_lseek(fd,m->fileStart+m->mapPos,SEEK_SET);
	if (_write(fd,h,t=30)!=30)
		return errno==ENOSPC ? errVolumeFull:_e_write_error;
	if (++c)
	{
		const ResFile_TypeEntry	*typeent = m->map.typeEntries;
		const ResFile_RsrcEntry	*resinfo;

		int		firstAdj = c*8+2;

		do
		{
			*h = typeent->typeCode;
			v = typeent->count-1;
			/* limitation: count can interfere w. offsets */
			((uint8*)h)[4] = (uint8)(v>>8), ((uint8*)h)[5] = (uint8)v;
			v = typeent->firstRes*MAC_RSRCENT_SIZE+firstAdj;
			if (v>65535)
				err = 3;
			((uint8*)h)[6] = (uint8)(v>>8), ((uint8*)h)[7] = (uint8)v;
			t+= 8;
			if (_write(fd,h,8)!=8)
				return errno==ENOSPC ? errVolumeFull:_e_write_error;
		} while(++typeent,--c);

		resinfo = (const ResFile_RsrcEntry*)typeent;
		h[2] = 0; /* stale (temporary) resource handle */
		do
		{
			v = resinfo->ID;
			((uint8*)h)[0] = (uint8)(v>>8), ((uint8*)h)[1] = (uint8)v;
			v = (resinfo->flag&RESNAME_OFFSET_MASK)-1;
			((uint8*)h)[2] = (uint8)(v>>8), ((uint8*)h)[3] = (uint8)v;
			((uint8*)h)[4] = (uint8)(resinfo->flag>>18);
			v = resinfo->dataOffset;
			((uint8*)h)[5] = (uint8)(v>>16);
			((uint8*)h)[6] = (uint8)(v>>8);
			((uint8*)h)[7] = (uint8)v;
			if (_write(fd,h,MAC_RSRCENT_SIZE)!=MAC_RSRCENT_SIZE)
				return errno==ENOSPC ? errVolumeFull:_e_write_error;
		} while(++resinfo,(t+=MAC_RSRCENT_SIZE)<n);

		if (t>m->mapLen+0u)
			err = 3; /* written length beyond stored map length! */
		else if (t<m->mapLen+0u)
		{
			t =-t, t+= m->mapLen;
			if (_write(fd,resinfo,t)!=t) /* write names */
				return errno==ENOSPC ? errVolumeFull:_e_write_error;
		}
	}
	return err;
}

/* ---- IIgs-specific functions */

/* 14MB is the usable Apple IIgs RAM limit */
#define MAX_IIGSRF_MAP		((8<<20)-1)
#define rResNames		0x8014

static int IIGS_checkheader(ResFile_Memory *m,
	const uint8 *h)
{
	uint8		t[22];

	m->mapPos = ((h[7]<<8|h[6])<<8|h[5])<<8|h[4];
	m->mapLen = ((h[11]<<8|h[10])<<8|h[9])<<8|h[8];

	if (h[0]|h[1]|h[2]|h[3] || m->mapPos-1u<12u-1
			|| m->mapLen<32 || m->mapLen>MAX_IIGSRF_MAP)
		return 0; /* not a working Mac OS resource fork */
	/* map offset/length copy must match */
	(void)fseek(m->filePtr,m->fileStart+m->mapPos+6,SEEK_SET);
	if (fread(t,sizeof(t),1,m->filePtr)!=1)
		return errno==ENOENT ? errHitFileEOF:_e_read_error;
	m->map.totalCount = t[18]|t[19]<<8|t[20]<<16|t[21]<<24;
	return memcmp(t,h+4,8) ? 0:t[8]|t[9]<<8;
}

static int32 makeHexType(uint16 x)
{
	const char	*t = "0123456789ABCDEF";

	return MK4CC(t[x>>12],t[x>>8&15],t[x>>4&15],t[x&15]);
}

static uint32 remapIIGSFlag(uint16 flag)
{
	int32		r = flag<<(13-14+18)&RES_FIXED_FLAG<<18;

	r|= flag<<(12-11+18)&(RES_CONVERTER_FLAG|RES_ABS_LOAD_FLAG)<<18;
	r|= flag<<(4-7+18)&(RES_PROTECTED_FLAG|RES_PRELOAD_FLAG)<<18;
	r|= flag<<(10-4+18)&(RES_NO_BANK_CROSS_FLAG|RES_NO_SPECIAL_MEM_FLAG
		|RES_PAGE_ALIGNED_FLAG)<<18;
	r|= flag<<(5-9+18)&RES_ALLOW_PURGE_FLAG<<18; /* purge level bit #1 */
	r|= flag<<(7-8+18)&0x2000000; /* purge level bit #0 */
	return r;
}

static ResFile_Memory *addIIGSNames(ResFile_Memory *m,
	int32 t)
{
	uint8		b[6];

	ResFile_RsrcEntry	*l, *res;

	FILE		*f = m->filePtr;
	int32		i, c, tc, v, ID, z, o = 1, nt;

	if (t<0)
		return m;
	z = m->memSize, o = 1;
	/* get indexed resources; IDs in the 0x1xxxx range correspond to types */
	for(i=0,tc=m->map.typeEntries[t].count; i<tc; ++i)
	{
		l = i+m->map.typeEntries[t].firstRes
			+(ResFile_RsrcEntry*)(m->map.typeCount+m->map.typeEntries);
		if ((l->ID&0xFFFF0000)!=0x10000
				|| (nt=getMapType(m,(uint16)l->ID))<0)
			continue;
		/* parse the rResNames resource */
		/* reject absurd counts */
		if (rsrcSeek(m,l)<0 || fread(b,6,1,f)!=1 || b[5])
			continue;
		c = b[2]|b[3]<<8|b[4]<<16;
		do
		{
			if (fread(b,5,1,f)!=1)
				break;
			ID = ((b[3]<<8|b[2])<<8|b[1])<<8|*b;
			if ((v=b[4])==0 || (res=getResource(m,nt+m->map.typeEntries,
					NULL,ID,NULL))==NULL)
				continue; /* blank or no corresponding rsrc */
			{ /* append and attribute offset */
				/* remember to save the ptr offset! */
				int32		ro = (int)((int8*)res-(int8*)m);

				++z;
				if ((m=realloc(m,z+v))==NULL)
					return NULL;
				res = (void*)(ro+(int8*)m);
			}
			((int8*)m)[z-1] = (int8)v;
			if (fread(z+(int8*)m,v,1,f)!=1)
				break; /* read error */
			res->flag|= o;
			++o, o+= v, z+= v, m->memSize = z;
			if (o>=0x3FFFF)
				return m; /* maxed out */
		} while(--c);
	}
	return m;
}

/* this routine assumes the typeCodes are in order */
static int IIGS_loadmap(ResFile_Memory **data,
	uint32 o)
{
	ResFile_Memory		*m = *data;
	ResFile_TypeEntry	*t;
	ResFile_RsrcEntry	*res;

	FILE		*f = m->filePtr;
	int32		*zl, *dz, z, r, c, lcode;
	const int8	*e;

	if ((c=m->map.totalCount)<=0)
		return 0;
	if (fseek(f,o+m->fileStart+m->mapPos,SEEK_SET)<0)
		return errno==ENOENT ? errHitFileEOF:_e_read_error;
	if ((dz=zl=calloc(m->map.totalCount,4))==NULL)
		return _e_memory_error;
	res = (ResFile_RsrcEntry*)(c+m->map.typeEntries);
	e = m->memSize+(const int8*)m;
	t = m->map.typeEntries, lcode =-1;
	do
	{
		uint8		b[20];
		uint16		code;

		if (fread(b,sizeof(b),1,f)!=1)
			{r = errno==ENOENT ? errHitFileEOF:_e_read_error; goto end;}
		if ((int8*)(1+res)>e)
		{
			if (m->options&TOPT_VERBOSE)
				(void)fprintf(stderr,"## IIGS_loadmap: buffer overflow!\n");
			r = errReadBadEntryMap;
			goto end;
		}
		code = *b|b[1]<<8;
		res->ID = b[2]|b[3]<<8|b[4]<<16|b[5]<<24;
		res->dataOffset = b[6]|b[7]<<8|b[8]<<16|b[9]<<24;
		res->flag = remapIIGSFlag(b[10]|b[11]<<8);
		*dz++ = b[12]|b[13]<<8|b[14]<<16|b[15]<<24;
		if (lcode!=code)
		{ /* add type */
			t->typeCode = code, t->count = 0;
			++t, ++m->map.typeCount, lcode = code;
		}
		++t[-1].count;
	} while(++res,--c);

	/* shift the processed entries into place, and resize */
	z = m->map.totalCount*sizeof(ResFile_RsrcEntry);
	memmove(t,m->map.totalCount+m->map.typeEntries,z);
	m->memSize = (int32)(z+(int8*)t-(int8*)m);

	/* calculate the offsets */
	calcTypeOffsets(m->map.typeEntries,m->map.typeCount);
	r = 0;
	/* read the names... adjust data, and tighten the map */
	if ((m=addIIGSNames(m,getMapType(m,rResNames)))==NULL)
		{r = _e_memory_error; goto end;}
	/* make the types readable */
	c = m->map.typeCount, t = m->map.typeEntries;
	do
		t->typeCode = makeHexType((uint16)t->typeCode);
	while(++t,--c);
	/* and finally, append the lengths list (4-byte aligned) */
	o = (m->memSize+3)&~3, z = m->map.totalCount*4;
	m->memSize = o+z;
	if ((m=realloc(m,m->memSize))==NULL)
		{r = _e_memory_error; goto end;}
	*data = m;
	memcpy(o+(int8*)m,zl,z);
end:	free(zl);
	return r;
}

/* ---- Windows 32-bit resource file */

static int Win32_checkheader(ResFile_Memory *m,
	const uint8 *h)
{
	uint8		t[16];
	int32		z, hz;
	/* heading designed to be an invalid rsrc in Win16 format */
	if (*(const int32*)h|(((const int32*)h)[1]^_le32_native(32))
			|(((const int32*)h)[2]^_le32_native(65535))
			|(((const int32*)h)[3]^_le32_native(65535)))
		return 0;
	(void)fseek(m->filePtr,m->fileStart+16,SEEK_SET);
	if (fread(t,sizeof(t),1,m->filePtr)!=1)
		return errno==ENOENT ? errHitFileEOF:_e_read_error;
	if (*(const int32*)t|((const int32*)t)[1]|((const int32*)t)[2]
			|((const int32*)t)[3])
		return 0;
	/* this resource file format does not use a type table; instead,
	  resources have a general 32-byte header, all concatenated inefficiently
	  so the type list would have to be built manually
	*/
	m->map.totalCount = 0;
	/* count the types; will hit EOF when finished */
	/* a zero count is normal */
	while((z=fread(t,1,sizeof(t),m->filePtr))==sizeof(t))
	{
		z = le32_native(*(const int32*)t); /* data size */
		hz = le32_native(((const int32*)t)[1]);
		if (z<0 || hz<32)
			return 0;
		++m->map.totalCount;
		/* aligned to next 4-byte boundary */
		fseek(m->filePtr,(z+hz-16+3)&~3,SEEK_CUR);
	}
	if (z)
		return 0;
	/* max out to allocate enough */
	m->map.typeCount = MIN(m->map.totalCount,65535);
	m->mapLen = m->map.typeCount*50+50;
	m->mapPos = ftell(m->filePtr);
	return m->dataPos=32;
}

static int Win32_genmap(ResFile_Memory **data)
{
	uint8		b[12];

	ResFile_Memory		*m = *data;
	ResFile_TypeEntry	*t;
	ResFile_RsrcEntry	*res;

	FILE		*f = m->filePtr;
	int32		r, c, z, hz, tz, i, j, ID, o, next;

	m->mapLen = 0;
	m->map.typeCount = 0;
	if ((c=m->map.totalCount)<=0)
		return 0;
	r = 0;
	if (fseek(f,m->fileStart+m->dataPos,SEEK_SET)<0)
		return errno==ENOENT ? errHitFileEOF:_e_read_error;
	tz = (sizeof(ResFile_TypeEntry)+sizeof(ResFile_RsrcEntry))*c;
	memset(m->map.typeEntries,0,tz);
	res = (ResFile_RsrcEntry*)(c+m->map.typeEntries);
	i = 0;
	do
	{
		if (fread(b,sizeof(b),1,f)!=1) /* should not error */
			{r = errno==ENOENT ? errHitFileEOF:_e_read_error; goto end;}
		z = le32_native(*(const int32*)b);
		hz = le32_native(((const int32*)b)[1]);
		r = le32_native(((const int32*)b)[2]);
		if (z<0 || hz<32)
			return -1;
		next = ftell(f)-sizeof(b);
		o = next-m->dataPos;
		next+= z+hz;
		/* match type to count; add if not found */
		/* non-ordinal types are wchar_t cstr */
		if ((int16)r+1)
		{
			if (r>>=16)
				do
					r = fgetc(f), r|=fgetc(f)<<8;
				while(r);
			t = m->map.typeEntries+m->map.typeCount;
			r =-1;
		} else
		{
			r = (uint32)r>>16;
			for(j=0,t=m->map.typeEntries; j<m->map.typeCount; ++j,++t)
			{
				ID = be32_native(t->typeCode);
				if (ID==r && (int32)ID>=0)
					goto gotit;
			}
		}
		t->typeCode = be32_native(r);
		t->firstRes = i;
		++m->map.typeCount;
	gotit:	t->count = t->count>65534 ? 65535:t->count+1;
		/* can be an ID or a wchar_t cstr but not both */
		fread(&ID,sizeof(ID),1,f);
		ID = le32_native(ID);
		if ((int16)ID+1)
		{
		/*	uint8		*d;

			ID = (uint32)ID>>16;
			m = realloc(m,tz+OFFSET(ResFile_Memory,map)+256);
			d = tz+OFFSET(ResFile_Memory,map)+(uint8*)m;*/
			j = 0;
			do
				;//d[j+1] = r>126 ? 127:r;
			while(++j,(r=fgetwc(f))!=0);
			//res = i+(ResFile_RsrcEntry*)(c+m->map.typeEntries);
			ID =-1;
		} else
			ID = (uint32)ID>>16;
		res->ID = ID;
		res->dataOffset = o;
		++res;
		fseek(m->filePtr,next&~3,SEEK_SET);
	} while(++i<c);
	t = m->map.typeEntries+m->map.typeCount;
	/* compact list */
	z = c*sizeof(ResFile_RsrcEntry);
	memmove(t,m->map.typeEntries+m->map.totalCount,z);
	m->memSize = (int32)(z+(int8*)t-(int8*)m);
	m = realloc(m,m->memSize);
	*data = m;
end:	return r;
}

/* ---- map functions */

int ResFile_LoadMap(FILE *f,
	uint16 options,
	ResFile_Memory **data,
	uint32 fileStart)
{
	uint8		h[16];
	ResFile_Memory	*m, b;
	long int	oldpos;
	int32		size, typeso = 0;
	int		err = 0;

	if (!data)
		return _e_general_error;
	*data = 0;
	if (!f)
		return _e_general_error;
	memset(&b.options,0,sizeof(b)-OFFSET(ResFile_Memory,options));
	b.filePtr = f;
	b.fileStart = fileStart;

	/* save position */
	oldpos = ftell(f);

	/* bit #31 will be 1 if some data's already loaded due to looking for a format;
	   bit #30 will be 1 if failure occured and we're trying to reconstruct
	*/
	options&= TOPT_PUBLIC_BITS;

	(void)fseek(f,fileStart,SEEK_SET);
	if (fread(h,sizeof(h),1,f)!=1) /* the fork should be >= this size! */
		{err = errno==ENOENT ? errHitFileEOF:_e_read_error; goto raw;}

	/* type 0 = implied identification */
	/* note that damaged RFs may fail */
	if ((!(uint8)options || (uint8)options==TYPE_RFMACOS)
			&& (size=MacRF_checkheader(&b,h))>0)
	{
		b.type = TYPE_RFMACOS;
		if (size>1)
			err = 5;
	} else if ((!(uint8)options || (uint8)options==TYPE_RFIIGS)
			&& (typeso=IIGS_checkheader(&b,h))>22)
		b.type = TYPE_RFIIGS;
	else if ((!(uint8)options || (uint8)options==TYPE_RSWIN32)
			&& (typeso=Win32_checkheader(&b,h))>31)
		b.type = TYPE_RSWIN32;
	else
		err = _e_handler_error;
raw:	if (err<0)
	{
		if (!(options&TOPT_MAKERAW))
			goto end;
		options|= FAIL_RECONSTRUCT;
	}
	/* IIGS rsrc count limit in RAM <419430 */
	size = (b.mapLen-22u)/20u;
	if (b.mapLen>0xFFFFFFu || b.map.totalCount>MIN(size,65535))
	{
		if (options&TOPT_VERBOSE)
			(void)fprintf(stderr,"### ResFile_LoadMap: bad mapLen or resCount\n");
		err = errReadBadEntryMap;
		goto end;
	}
	size = b.mapLen-MAC_TYPES_OFFSET-6+OFFSET(ResFile_Map,typeCount);
	b.memSize = size+= OFFSET(ResFile_Memory,map);
	if (size<sizeof(b))
		size = sizeof(b);
	if ((m=realloc(0,size))==NULL)
		{err = errOutOfMemory; goto end;}

	b.options = options;
	*data = m, *m = b;

	if (!(options&FAIL_RECONSTRUCT))
		if (b.type==TYPE_RFMACOS)
			err = MacRF_loadmap(m);
		else if (b.type==TYPE_RFIIGS)
			err = IIGS_loadmap(data,typeso), m = *data;
		else if (b.type==TYPE_RSWIN32)
			err = Win32_genmap(data), m = *data;
		else if (b.type)
			err = _e_handler_error;
	if (err<0 && !(~options&(TOPT_MAKERAW|FAIL_RECONSTRUCT)))
	{ /* failure can still occur right before here */
		if (b.type==TYPE_RFMACOS)
			err = MacRF_makemap(data);
	}
end:	/* restore position */
	(void)fseek(f,oldpos,SEEK_SET);

	return err;
}

/*
 NOTE: this always compacts on byte boundary
*/
static int RF_compact(ResFile_Memory *m)
{
/* TODO: sort offsets to prevent overwrite */
(void)m;
#if 0
	ResFile_RsrcEntry	*l;

	int8		*b, *p;
	FILE		*f = m->filePtr;
	long int	o;
	int32		c, z;
	int		err;

	if ((c=m->map.totalCount)<=0)
		return 0;
	l = (ResFile_RsrcEntry*)(m->map.typeCount+m->map.typeEntries);
	o = m->dataPos;
	b = NULL, z = 0;
	do
		if (o<l->dataOffset)
		{
			int	len = rsrcLen(m,l)+4;

			if (len>z)
			{
				if ((p=realloc(b,z))==NULL)
					{err = _e_memory_error; break;}
				b = p;
			}
			/* not very efficient, but gets the job done */
			fseek(f,m->fileStart+m->dataPos,SEEK_SET);
			if (fread(b,z,1,f)!=1)
				{err = _e_read_error; break;}
			m->dataPos = o;
			fseek(f,m->fileStart+o,SEEK_SET);
			if (fwrite(b,z,1,f)!=1)
				{err = _e_write_error; break;}
		}
	while(--c);
	m->flag&=~RF_COMPACT_FLAG;
	if (b)
		free(b);
#endif
	return 0;
}

int ResFile_Update(ResFile_Memory *m)
{
	FILE		*f;
	long int	oldpos;
	int		err;

	if (!m || (f=m->filePtr)==NULL)
		return _e_general_error;
	if ((oldpos=ftell(f))<0)
		return oldpos;
	err = 0;
	if (m->type==TYPE_RFMACOS)
	{
		int		err2 = 0;

		if (m->flag&RF_COMPACT_FLAG)
			err2 = RF_compact(m);
		if ((err=MacRF_update(m))>=0)
			err = err2;
	} else if (m->type)
		err = _e_handler_error;
	(void)fseek(f,oldpos,SEEK_SET);

	return err;
}

/* ---- resource reading */

/*
 loadResource
	Reads the resource from the file.
  Result:
	An mhandle, locked if flagged in res.
	If an error occurs, the code is stored in err.
  Notes:
	Decompression is not handled.
*/
static void *loadResource(const ResFile_Memory *m,
	const ResFile_RsrcEntry *res,
	int *r)
{
	void		*h = NULL;
	int32		len;

	if (!res)
		{*r = errNoRsrcMapEntry; return NULL;}
	if (res->flag&RES_COMPRESSED_FLAG<<18)
		*r = errDecompressErr; /* decompression not implemented */
	else if ((len=rsrcSeek(m,res))<0 || (len=rsrcLen(m,res))<0)
		*r = len;
	else if ((h=hNew(len,1))==NULL)
		*r = errOutOfMemory;
	else
	{
		int64		o = ftell(m->filePtr)-m->fileStart, mp = m->mapPos;
		int8		*p;

		if (o<mp && o+len>mp)
		{
			if (m->options&TOPT_VERBOSE)
				(void)fprintf(stderr,"### loadResource: rsrc intersects map!\n");
			*r = errReadBadEntryMap;
		}
		if ((p=hLock(h))==NULL)
		{
			*r =-1;
			hFree(h); h = NULL;
		}
		if (fread(p,len,1,m->filePtr)!=1)
			*r = errno==ENOENT ? errHitFileEOF:_e_read_error;
		else if (!(res->flag&RES_LOCKED_FLAG))
			hUnlock(h);
	}
	return h;
}

void *ResFile_GetResource(ResFile_Memory *m,
	int32 type,
	const uint8 *name,
	int ID,
	int *r)
{
	void		*h = NULL;
	int32		x;
	int		err;

	if (!m || !m->filePtr)
		err = _e_general_error;
	else if ((x=getMapType(m,type))<0)
		err = errNoRsrcMapEntry;
	else
		h = loadResource(m,getResource(m,m->map.typeEntries+x,name,ID,NULL),&err);
	if (r)
		*r = err;
	return h;
}

void *ResFile_GetIndResource(ResFile_Memory *m,
	int32 type,
	int ix,
	RsrcInfo *ri,
	int *r)
{
	ResFile_TypeEntry	*t;

	void		*h = NULL;
	uint32		i = ix^ix>>31;
	int		x, err = 0;

	if (ri)
		ri->offsetOnDisk = 0;
	if (!m || !m->filePtr)
		err = _e_general_error;
	else if ((x=getMapType(m,type))<0
			|| i>=(t=x+m->map.typeEntries)->count+0u)
	{
		if (ix>=0)
			err = errNoRsrcMapEntry;
	} else
	{
		const uint8	*l = (const uint8*)(m->map.typeCount+m->map.typeEntries);

		const ResFile_RsrcEntry	*res = i+t->firstRes+(ResFile_RsrcEntry*)l;

		if (ri)
		{
			ri->ID = res->ID;
			ri->flag = (uint16)(res->flag>>18);
			*ri->name = 0;
			if ((x=res->flag&RESNAME_OFFSET_MASK)!=0)
			{
				l+= sizeof(ResFile_RsrcEntry)*m->map.totalCount+x-1;
				memcpy(ri->name,l,*l+1);
			}
			ri->sizeOnDisk = (x=rsrcSeek(m,res))<0 ? 0
				:rsrcLen(m,res);
			ri->offsetOnDisk = ftell(m->filePtr);
		}
		if (ix>=0)
			h = loadResource(m,res,&err);
	}
	if (r)
		*r = err;
	return h;
}

int ResFile_GetResourceIndex(ResFile_Memory *m,
	int32 type,
	const uint8 *name,
	int ID)
{
	int32		x;

	if (!m || !m->filePtr)
		return _e_general_error;
	if ((x=getMapType(m,type))<0)
		return errNoRsrcMapEntry;
	(void)getResource(m,m->map.typeEntries+x,name,ID,&x);
	return x;
}

#ifdef NEW_AND_INCOMPLETE
/*
 ResFile_RemoveResource
	Finds and removes a specified rsrc (type/zero-based index) in the map.
	You must call ResFile_Update to affect the change.
  Note:
	Passing a negative index is invalid, and will result in _e_general_error.
  Result:
  	Error (negative), or zero.
*/
/* removal and addition should realloc (ResFile_Memory** param) */
int ResFile_RemoveResource(ResFile_Memory *m,
	int32 type,
	int i)
{
	ResFile_TypeEntry	*t;

	int32		x;
	int		err = 0;

	if (i<0 || !m || !m->filePtr)
		return _e_general_error;
	if ((x=getMapType(m,type))<0
			|| i>=(t=x+m->map.typeEntries)->count)
		err = errNoRsrcMapEntry;
	else
	{
		ResFile_RsrcEntry	*l = (ResFile_RsrcEntry*)(m->map.typeCount+m->map.typeEntries);
		ResFile_RsrcEntry	*res = i+t->firstRes+l;
		/* remove the entry, adjust the counts, & flag the map for writeback */
	#if 0
		if (m->type==TYPE_RFIIGS)
		{ /* for IIGS storage, the length on the end also has to be removed */
			x = res-l;
			if ((x-=m->map.totalCount)<0)
			{
				int32		*a = 1+x+(int32*)(m->memSize+(int8*)m);

				memmove(a-1,a,m->memSize+(int8*)m-(int8*)a);
				m->memSize-= 4;
			}
		}
	#endif
		++res; /* moving down from next */
		memmove(res-1,res,m->memSize+(int8*)m-(int8*)res);
		--m->map.totalCount;
		if ((--t->count)<=0)
		{ /* remove type */
		}
		m->memSize-= sizeof(ResFile_RsrcEntry);
		//m->mapLen-= sizeof(ResFile_RsrcEntry); -- updateHeader takes care of this
		calcTypeOffsets(m->map.typeEntries,m->map.typeCount);
	}
	return 0;
}
#endif

/*
 compression:
  start code: 0xA89F6572
  header size: 0xA=type (8 or 9); 0xB=flag (bit #0 on);
   0xC uncmp len; 0x10=tmp coef; 0x11=???; 0x12=dcmp ID; can be zero
  tmp len: (tmp coef+1)*(cmp len)>>8)+4
  adj end: uncmp len-cmp len + 0x12

  dcmp param: cmp data, adj end, tmp buf, tmp len
*/
