/*
 resex.h
  resource extraction (and recovery)

2014.02.09
  maximized type lists by making reslist offsets indices instead.
*/

#ifndef __INTERPRETER_H
 #include "interpreter.h"
#endif
#ifndef __MY_CSERIES_H
 #include "my_cseries.h"
#endif

/* ---- constants */

/* options/types */

/* the resfork/resfile type -- Windows not implemented in resex */
#define TYPE_RFMACOS		1
#define TYPE_RFIIGS		2
#define TYPE_RSWIN32		3

/* TOPT_MAKERAW
  If loading the resfile's info fails, scan sequentially, try to make one.
  Note that raw map creation only works w. byte-aligned Mac OS RFs.
  Be careful in using this; any possibly valid data may be overridden.
*/
#define TOPT_MAKERAW		0x100

/* TOPT_VERBOSE
  Prints more detailed errors to stderr.
*/
#define TOPT_VERBOSE		0x800

#define TOPT_PUBLIC_BITS	0xFFF
#define FAIL_RECONSTRUCT	0x8000	/* flagged by _Load - reconstructed */

#define RF_READ_ONLY_ATTR	0x8000

#define RF_COMPACT_FLAG		0x1	/* resize on update */

/* individual resource flags - upper bits (.flag>>18) */

#define RES_COMPRESSED_FLAG	0x1
#define RES_CHANGED_FLAG	0x2	/* remember to update this in-file */
#define RES_PRELOAD_FLAG	0x4	/* PDR #6 - preload on open (not implemented) */
#define RES_PROTECTED_FLAG	0x8	/* PDR #7 - not implemented */
#define RES_LOCKED_FLAG		0x10	/* PDR #15 - mhandle locked on load */
#define RES_ALLOW_PURGE_FLAG	0x20	/* PDR #8/9 - allow automatic disposal to save memory */
#define RES_SYSTEM_HEAP_FLAG	0x40	/* load into system map (obsolete) */
/* remapped IIgs 5+ flags */
#define RES_FIXED_FLAG		0x2000	/* PDR #14; don't move in fork (promotes fragmentation) */
#define RES_CONVERTER_FLAG	0x1000	/* PDR #11 */
#define RES_ABS_LOAD_FLAG	0x800	/* PDR #10 */
#define RES_NO_BANK_CROSS_FLAG	0x400	/* PDR #4; load w. 65536-byte bank alignment */
#define RES_NO_SPECIAL_MEM_FLAG	0x200	/* PDR #3 */
#define RES_PAGE_ALIGNED_FLAG	0x100	/* PDR #2; load w. 256-byte page alignment */

/* ---- typedefs */

typedef struct ResFile_RsrcEntry/*(12)*/
{
	int32		ID;
	uint32		flag; /* [flags.14] [nameoffs.18; 0=none] */
	uint32		dataOffset;	/* from data start */
} ResFile_RsrcEntry;
#define RESNAME_OFFSET_MASK	0x3FFFF

typedef struct ResFile_TypeEntry/*(8)*/
{
	int32		typeCode;	/* a four-char or hex code */
	uint16		count;		/* whole */
	uint16		firstRes;	/* first absolute resource in list */
} ResFile_TypeEntry;

typedef struct ResFile_Map/*(8+)*/
{
	int32		totalCount;	/* of all resources */
	uint16		attributes;	/* bit #15: read-only */
	uint16		typeCount;	/* whole */

	ResFile_TypeEntry	typeEntries[1];

	/* followed by resource info entries, names (p-strings) */
	/* in non-Mac OS case, lengths may follow */
} ResFile_Map;

typedef struct ResFile_Memory/*(20+)*/
{
	struct FILE	*filePtr;
	uint16		options;	/* end-value of options */
	uint8		type, flag;
	uint32		fileStart;	/* alt. start of file/fork (normally zero) */
	int32		memSize;
	uint32		dataPos, mapPos;
	int32		mapLen;
	ResFile_Map	map;
} ResFile_Memory;

typedef struct RsrcInfo/*(270)*/
{
	int32		ID;
	uint32		sizeOnDisk, offsetOnDisk; /* excluding leading bytes */
	uint16		flag;
	uint8		name[256];	/* pascal string */
} RsrcInfo;

/* ---- prototypes */

/*
 Returned error codes can be from either error_codes.h or interpreter.h.
   _e_handler_error may be returned if the file format could not be managed.
 A positive result is a warning or a count.
*/

/*
 ResFile_LoadMap
	Loads the entire map.
  Notes:
	This function does not open a file; it takes an open file.
	When finished, call free for the map.  fclose is left to the caller.
  Result:
  	Zero (no err), error (neg.) or warning (positive) if unusual data encountered.
*/
int ResFile_LoadMap(struct FILE *f,
	uint16 options, /* low 8 bits = force type (leave 0 to find out what it is) */
	ResFile_Memory **map,
	uint32 filestart);

/*
 ResFile_Update
	Writes whatever needs to be written of the resource file.
  Result:
  	Zero on success; negative on error.
*/
int ResFile_Update(ResFile_Memory *data);

/*
 ResFile_CountResources
	Finds and returns the whole-number count of a resource type.
	Use the ResFile_Memory struct for navigating the types.
*/
int ResFile_CountResources(ResFile_Memory *m,
	int32 type,
	int *err); /* can be NULL */

/*
 ResFile_GetResource
	Finds and loads a specified resource in the file.
	If name != NULL, then only the name is used, otherwise the ID is used.
  Result:
  	The handle or NULL, with *err set to zero or error (negative).
*/
void *ResFile_GetResource(ResFile_Memory *m,
	int32 type,
	const uint8 *name, /* Pascal string; can be NULL */
	int ID,
	int *err); /* can be NULL */
/*
 ResFile_GetIndResource
	Loads a resource of type and zero-based(!) index.
  Result:
  	The handle or NULL, with *err set to zero or error (negative).
	offsetOnDisk will be 0 by default in case the rsrc doesn't exist.
  Notes:
	Pass the binary complement (~) of the index to belay loading.
	Optionally, a ptr for info can be passed to retrieve more info.
*/
void *ResFile_GetIndResource(ResFile_Memory *m,
	int32 type,
	int ix, /* see comment above */
	RsrcInfo *info, /* can be NULL */
	int *r); /* can be NULL */

/*
 ResFile_GetResourceIndex
	Finds the index for a rsrc by name (if !=NULL) or ID.
  Result:
	The zero-based index (of type), or error code (negative).
*/
int ResFile_GetResourceIndex(ResFile_Memory *m,
	int32 type,
	const uint8 *name, /* Pascal string; can be NULL */
	int ID);

/*
 ResFile_OverheadSize
	Calculates the overhead size for any resource in-file.
  Result:
	The size, or error (negative).
*/
int ResFile_OverheadSize(ResFile_Memory *m,
	int *entrySize); /* can be NULL */

#ifdef WINUSERAPI
/* shellRc protos */
int CountModuleResources(HMODULE module,
	int32 type);
void *GetModuleIndResource(HMODULE module,
	int32 type,
	int ix,
	RsrcInfo *info,
	int *r);
void *GetModuleResource(HMODULE module,
	int32 type,
	const uint8 *name,
	int ID,
	int *r);
int GetModuleResourceIndex(HMODULE module,
	int32 type,
	const uint8 *name,
	int ID);
int IterateModuleResources(HMODULE module,
	int32 type,
	void *userData,
	int func(int ID,
		const uint8 *name,
		void *userData));
#endif
