/*
 main.c
  executable shell (basic player)

  Revised: 2023.08.26

 When compiled for Windows:
   escape key: stop
   M: toggle mute
   space: pause/resume
   , or .: volume
   - or +: rate
   double-quote: type new file to load
   i: print info (if not already displayed)
   f: cycle through filter options

 Note: there's a small delay in key response time.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2015.04.07
  MT implementation, since 64-bit Pelles C appears to break single-threaded...
2021.07.27
  assume mpeg by file extension.
2022.11.17
  added rsrc fork / companion file search code.
2023.03.21
  added console file drag & drop.
*/

#include <stdio.h>
#include <conio.h>
#include <io.h>
#include <string.h>

#include <math.h>
#include <enviro64.h>
#include "snddec.h"
#include "../resex/resex.h"

#if 0//def __MT__
 #include "mevent.h"
 #define MAXTIMERS		32
timerThread	*global_timer_thread;
#endif

#define DEFAULT_RAWRATE		10000

/* ---- library */

#include <sys/stat.h>

int fcanopen(const char *path)
{
	struct _stat	info;

	if (!path)
		return -1;
	if (_stat(path,&info))
		return 0;
	return !_S_ISDIR(info.st_mode) && info.st_mode&_S_IREAD;
}

#include <stdarg.h>

int dprintf(const char *form,
	...)
{
	va_list		args;
	int		r;

	va_start(args,form);
	r = vfprintf(stderr,form,args);
	va_end(args);
	return r;
}

#define NO_GUI
#define NO_MGRAF
#include "../utility/interpreter.c"

#define NO_TABLES
#include "../cseries.c"

/* ---- main code */

char		temporary[512];
char		*progname;

static int8 stopdammit, quiet, dispinfo;

static void printDuration(const sndFormat *f,
	int userSet);
static void printSndInfo(const sndFormat *f,
	uint8 origChanCount);

static long readBufferCmdHeader(FILE *s,
	sndFormat *f,
	uint8 *bn)
{
	uint8		b[64];

	if (!s || !f)
		return -1;
	if (fread(b,1,sizeof(b),s)<22)
		return -3;
	return parseBufferCmdHeader(b,f,NULL,bn);
}

#ifndef NO_SNDMGR
 #include <stdlib.h>
 #include <string.h>
 #include <windows.h>	//no longer needed to compile

static void sndMgrCloseVerbose(sndManager *smgr)
{
	sndChannel	*ch;
	uint8		err = 0;

	if (!smgr)
		return;
	for(ch=sndMgrLeadChannel(smgr); ch; ch=ch->next)
		err|= ch->sndFrame.err&0x80;
	if (err&128)
		printf("\n## errors detected in decoding\n");
	sndMgrClose(smgr);
}

ufixed		fvol = 0;
boolean		takeparam = 1;
char		*takeparambuf, *takeparamarr[64];

static void setChSettings(sndChannel *ch,
	int32 vol,
	ufixed rmul,
	double filter)
{
	sndCommand	cmd;

	cmd.cmd = sndRLVolumeCmd;
	cmd.param2 = (void*)vol;
	(void)sndDoCommand(ch,&cmd,1);

	cmd.cmd = sndRateMultiplierCmd;
	cmd.param1 = 0x80000000u; /* to ignore */
	cmd.param2 = (void*)rmul;
	(void)sndDoCommand(ch,&cmd,1);

	if (filter>=0)
	{
		cmd.cmd = sndFilterCmd;
		cmd.param2 = (void*)(uint32)(filter*4294967296.);
		(void)sndDoCommand(ch,&cmd,1);
	}
}

//#define ASYNC_KEYS

uint32 fixAccumulatedRate(uint32 rmul)
{
	if (rmul>=0xFFFD)
	{
		if (rmul<=0x10002)
			rmul = 0x10000;
		else if (rmul>=0x1FFFD && rmul<=0x20002)
			rmul = 0x20000;
	} else if (rmul>=0x3FFD)
	{
		if (rmul<=0x4002)
			rmul = 0x4000;
		else if (rmul>=0x7FFD && rmul<=0x8002)
			rmul = 0x8000;
	} else if (rmul>=0x1FFD && rmul<=0x2002)
		rmul = 0x2000;
	else if (rmul>=0x3446 && rmul<=0x344B)
		rmul = 0x3449;

	return rmul;
}

static void sndWaitLoop(sndManager *mgr,
	sndChannel *ch,
	int32 vol,
	ufixed rmul,
	double filter,
	uint8 origChanCount)
{
	double		posRate = 0;
	uint64		lastdur = 0;
	int		r;
	int8		mutev = sndEnableLeft|sndEnableRight;
	boolean		notChangedSetting = rmul==0x10000 && vol==0x1000100;
#ifdef WINUSERAPI
 #ifdef ASYNC_KEYS
	int8		pauset = 0, incrt = 0, decrt = 0, incvt = 0, decvt = 0, mutet = 0;
 #endif
#endif
	uint8		oldVoiceMode = ch->voiceMode, change;//, u = 0;

	if (takeparam==2)
		quiet = 0;
	while(sndChannelPlaying(ch))
	{
		change = 0;
	#ifdef WINUSERAPI
 		if (_kbhit())
		{
			if ((r=_getch())==0 || r==0xE0)
				{(void)_getch(); continue;} /* arrow key */
			switch(r)
			{
			case 27: stopdammit = 1/*, ch->actFlags|= chStopped*/; goto end; /* esc */
			case ' ': ch->voiceMode^= oldVoiceMode; break;
			case '-': rmul = rmul*(1/1.0293022366)+.5; goto set;
			case '+': rmul = rmul*1.0293022366+.5; goto set;
			case '0': rmul = 0x10000; goto set;
			case ',': vol-= 0x200020; goto set;
			case '.': vol+= 0x200020; goto set;
			set:	rmul = fixAccumulatedRate(rmul);
				setChSettings(ch,vol,rmul,filter);
				notChangedSetting = 0;
				break;
			case 'm':
				mutev^= sndEnableLeft|sndEnableRight;
				(void)sndMgrMute(mgr,mutev);
				break;
			case 8: /* backspace */
				if (!quiet)
				{ /* replay */
					takeparam = 3, quiet = 1;
					printf("\r%80 \r");
					return;
				}
			case '"':
				takeparam = 1;
				goto end;
			case 'I':
			case 'i':
				if (dispinfo)
					break;
				printf("\r%80 \r");
				dispinfo = 1+(r=='I');
				printSndInfo(&ch->sndFrame.format,
					origChanCount);
				putc('\n',stdout);
				break;
			case 'f':
				if (filter<0)
					filter = 0;
				if ((filter+=.13)>=1)
					filter = 0;
				setChSettings(ch,vol,rmul,filter);
				notChangedSetting = 0;
				break;

			default:
				goto skip;
			}
			change = 1;
	 	}
	skip:
	#endif
		sndMgrProcWait(mgr);
#if 1
		if ((dispinfo || !quiet) && oldVoiceMode>sndVoiceLocked)
		{
			uint64		dur;

			if (posRate<=0)
				posRate = (ch->sndFrame.format.rate
					+ch->sndFrame.format.rateFrac*(1/4294967296.))
					*(ch->sndFrame.format.chanCount&127);
			dur = ch->sndFrame.sampElapsed/posRate;
			change|= dur>lastdur && dur>1;
			if (change)
			{
				uint64		mins;
				uint32		hrs;
				int32		lvol = vol>>16;

				dur>>= 1; /* 16-bit out; done here to update on .5 sec */
				mins = dur/60, hrs = mins/60;
				mins-= hrs*60;
				printf(notChangedSetting ? "  Elapsed: %d:%02d:%02d%44  "
					:lvol!=(int16)vol || filter>0 ?
					"  Elapsed: %d:%02d:%02d  Rate: %.4f  Vol: %d/%d%% F=%.2f        "
					:"  Elapsed: %d:%02d:%02d  Rate: %.4f  Vol: %d%%              ",
					hrs,(uint32)mins,(uint32)(dur-mins*60),//u,
					round(rmul*(10000/65536.))/10000,
					lvol*25/64,(int16)vol*25/64,filter);
				putc('\r',stdout);
				//++u, u&= 7;
			}
			lastdur = dur;
		}
#endif
	}
end:	(void)sndMgrWait(mgr);
	printf("\r%80  \r");
}

static int playFile(FILE *in,
	sndFormat *f,
	int32 k,
	int32 vol,
	ufixed rmul,
	int32 prefRate,
	double filter,
	uint8 origChanCount)
{
	sndManager	*mgr = NULL;
	sndChannel	*ch = NULL;
	int		r, rate;

	rate = f->rate+(f->rateFrac>>31);
	rate = prefRate ? prefRate
		:rate>44100 || rmul!=0x10000 ? 0
		//:rate<16000 ? 24000
		:/*rate<24000 ? 32000:*/44100;
	if (!mgr && (mgr=sndMgrOpen(rate,
			f->chanCount>1 || (int16)vol!=(vol>>16),&r,1))==NULL)
	{
		(void)fprintf(stderr,"### unable to open sndManager (%d)\n",
			r);
		goto err;
	}
	if (dispinfo&2)
		printf(" Out rate: %d%40 \n",
			rate ? rate:sndMgrGetMixRate(mgr));
	if (fvol)
		sndMgrSetVolume(mgr,fvol,fvol,0,0,0);
	if ((ch=sndNewChannel(mgr,sndInitStereo|(filter ? sndInitExtendedFilter:0),NULL))==NULL)
	{
		(void)fprintf(stderr,"### unable to create sndChannel\n");
	err:	r = 0;
		goto end;
	}
	setChSettings(ch,vol,rmul,filter);
	if ((r=sndPlayFile(in,ch,f,k,quiet || !dispinfo))<0)
		goto end;
	sndWaitLoop(mgr,ch,vol,rmul,filter,origChanCount);
	r = 1;
	(void)sndPlayFileStop(ch);
end:	(void)sndMgrCloseVerbose(mgr);
	return r;
}

/* ---- testing */

#define TEST_RATE	40960 /* fixed-point rate conversion error: 4.634mcs */
#define SINE_SRATE	48000

typedef struct playLogParam
{
	UnsignedWide	time[20];
	int16		*samples;
	int8		count, max;
} playLogParam;

/* this fails part-time */
static void playLog(sndChannel *ch,
	sndCommand *cbcmd)
{
	sndCommand	cmd;
	playLogParam	*param = (playLogParam*)ch->userInfo;
(void)cbcmd;
	if (!param || param->count>=param->max)
		return;
	Microseconds(param->time+param->count);
	++param->count;
	cmd.cmd = sndRawCmd;
	cmd.param1 = ch->sndFrame.format.rate;
	cmd.param2 = param->samples;
	(void)sndDoCommand(ch,&cmd,0);
	cmd.cmd = sndCallBackCmd;
	cmd.param2 = (void*)playLog;
	(void)sndDoCommand(ch,&cmd,0);
}

static void sndFreqScaleWait(sndChannel *ch,
	int *rem,
	int ampc)
{
	static uint8	scale[] = {60,62,64,65,67,69,71,72};
			/* 95.99*.834**x-x**1.5*.05 */
	static uint8	ampt[] = {95,80,66,55,46,38,31,26,21,17,14,11,8,6,4};
	sndCommand	cmd;
	int		i, j, k;

	if (ampc+0u>sizeof(ampt))
		ampc = sizeof(ampt);
	for(i=0; i<sizeof(scale); ++i)
	{
		cmd.cmd = sndFreqCmd;
		cmd.param2 = (void*)scale[i];
		(void)sndDoCommand(ch,&cmd,0);
		for(j=0; j<(i<sizeof(scale)-1 ? 6:ampc); ++j)
		{
			cmd.cmd = sndAmpCmd;
			cmd.param1 = ampt[j];
			(void)sndDoCommand(ch,&cmd,0);
			cmd.cmd = sndWaitCmd;
			k = (++*rem-3)>>31, *rem&= k;
			cmd.param1 = 166-k;
			(void)sndDoCommand(ch,&cmd,0);
		}
	}
	while(sndChannelPlaying(ch))
		sndMgrProcWait(sndMgrFromChannel(ch));
}

static int testSndMgr(int a,
	int hz)
{
	sndFormat	fmt;
	sndCommand	cmd;
	sndManager	*mgr = NULL;
	sndChannel	*ch = NULL;
	playLogParam	param;
	int64		n;
	int		r = 0, sr = hz || !a ? SINE_SRATE:TEST_RATE;

	if (a<0)
		return -1;
	if ((mgr=sndMgrOpen(sr,0,&r,0))==NULL)
	{
		(void)fprintf(stderr,"### unable to open sndManager (%d)\n",
			r);
		goto err;
	}
	if (fvol)
		sndMgrSetVolume(mgr,fvol,fvol,0,0,0);
	if ((ch=sndNewChannel(mgr,0,NULL))==NULL)
	{
		(void)fprintf(stderr,"### unable to create sndChannel\n");
	err:	r = 0;
		goto end;
	}
	memset(&param,0,sizeof(param));
	if ((param.samples=calloc(sizeof(int16),sr))==NULL)
		goto end;
	if (!a && hz<55)
		hz = 220;
	if (hz)
	{
	#define WAVE_AMP .25
		double		r = hz*3.1415926535897932*2/SINE_SRATE;
		double		theta = 0;
		int16		*d = param.samples;
		int		i = 0;

		do
		{
			d[i] = (int16)((WAVE_AMP*32768)*sin(theta)-.5);
			//if (i<512)
			//	printf(!(i&7) ? "\n%d,":"%d,",
			//		d[i]);
		} while(theta+=r,++i<SINE_SRATE);
	}

	--a;
	param.max = MIN(a,GET_ARRAY_COUNT(param.time)-1)+1;

	memset(&fmt,0,sizeof(fmt));
	fmt.rate = fmt.frameCount = sr;
	fmt.chanCount = 1;
	(void)sndFixCmpType(BIG_ENDIAN ? CODEC_PCM_BE_S16:CODEC_PCM_LE_S16,&fmt);
	sndFixCodecFrame(&fmt);
	ch->userInfo = &param;
	ch->baseNote = 57;
	cmd.cmd = sndFormatCmd;
	cmd.param2 = &fmt;
	(void)sndDoCommand(ch,&cmd,1);
	if (a<0)
	{
		int		rem = 0;

		if (!quiet)
			(void)printf("playing sine voice scale...\n");
		cmd.cmd = sndVoiceCmd;
		cmd.param1 = sr;
		cmd.param2 = param.samples;
		(void)sndDoCommand(ch,&cmd,1);
		cmd.cmd = sndLoopCmd;
		cmd.param1 = 0;
		cmd.param2 = (void*)sr;
		(void)sndDoCommand(ch,&cmd,1);
		sndFreqScaleWait(ch,&rem,6);
		if (!quiet)
			(void)printf("playing soft square scale...\n");
		cmd.cmd = sndToneCmd;
		cmd.param1 = 400|sndToneSquare<<16;
		(void)sndDoCommand(ch,&cmd,0);
		sndFreqScaleWait(ch,&rem,6);
		if (!quiet)
			(void)printf("playing soft ramp scale...\n");
		cmd.cmd = sndToneCmd;
		cmd.param1 = 400|sndToneRamp<<16;
		(void)sndDoCommand(ch,&cmd,0);
		sndFreqScaleWait(ch,&rem,~0);
		sndMgrProcWait(mgr);
		/* need to remove voice since it's going to be disposed */
		cmd.cmd = sndQuietCmd;
		(void)sndDoCommand(ch,&cmd,1);
	} else
	{
		cmd.cmd = sndCallBackCmd;
		cmd.param2 = (void*)&playLog;
		if (!quiet)
			if (hz)
				(void)printf("playing %d Hz sine wave for %ds...\n",
					hz,param.max);
			else
				(void)printf("timing silent sample time %ds @ %d Hz...\n",
					param.max,TEST_RATE);
		(void)sndDoCommand(ch,&cmd,1);
		while(sndChannelPlaying(ch))
		{
			sndMgrProcWait(mgr);
		#if 0 /* forces completion but ruins from-cb timing */
			if (ch->qHead==ch->qTail && param.count<param.max)
				sndDoCommand(ch,&cmd,1);
		#endif
		}
		//if (param.count<param.max)
		//	printf("failed to complete callback-only\n");
		printf("total callback time: %ds\n",
			param.count);
		if (param.count>2)
		{
			n = param.time[1];
			(void)int64add(n,n);
			(void)int64sub(n,*param.time);
			(void)int64sub(n,param.time[2]);
			/* low -71.079ms (<29ms non-silent) */
			printf("overhead time: %lld mcs\n", 
				n);
			n = param.time[param.count-1];
			(void)int64sub(n,param.time[1]);
			/* lowest time for my system: 1.0000041 (1.000001 short) */
			if (param.count>3) 
				printf("per second time: %.8gs\n",
					int64tox(n)/(1000000*(param.count-2)));
		}
	}
	r = 1;
	free(param.samples); param.samples = NULL;
end:	(void)sndMgrClose(mgr);
	return r;
}

static long loadSndPlay(FILE *f,
	long z,
	const char *vfile,
	int32 vol,
	ufixed rmul,
	int8 mc)
{
	FILE		*fvfile = NULL;
	ResFile_Memory	*fvmap = NULL;
	void		*fvh = NULL;
	sndManager	*smgr = NULL;
	sndChannel	*ch[64];
	uint8		*p;
	long		t;
	int		r;

	memset(ch,0,sizeof(ch));
	if (mc-1u>63)
		mc = mc<=0 ? 1:64;
	if ((p=calloc(z,1))==NULL)
		return -2;
	if ((t=fread(p,1,z,f))<z)
	{
		(void)fprintf(stderr,"## EOF error reading %d bytes\n",
			z);
		r =-3;
		goto end;
	}
	z = t;
	if ((smgr=sndMgrOpen(0,1,&r,0))==NULL)
	{
		(void)fprintf(stderr,"### unable to open sndManager (%d)\n",
			r);
		goto err;
	}
	if (fvol)
		sndMgrSetVolume(smgr,fvol,fvol,0,0,0);
	t = mc;
	do
	{
		if ((ch[t-1]=sndNewChannel(smgr,sndInitStereo,NULL))==NULL)
		{
			(void)fprintf(stderr,"### unable to open sndChannel (-207)\n");
			goto err;
		}
		setChSettings(ch[t-1],mc>1 ? vol>>2&0x3FFF3FFF:vol,rmul/t,0);
	} while(--t);
	if (vfile && ResFile_LoadMap(fvfile=fopen(vfile,"rb"),0,&fvmap,0)>=0)
	{
		fvh = ResFile_GetIndResource(fvmap,MK4CC('s','n','d',' '),0,NULL,NULL);
		t = mc;
		do
			sndPlaySndList(ch+t-1,(const uint8*)hLock(fvh),hSize(fvh),0,1);
		while(--t);
	}
	if ((r=sndPlaySndList(ch,p,z,1,0))<=0)
	{
		if (r)
			printf("Error #%d attempting to play SoundList\n",
				r);
		goto err;
	}
	if (mc>1)
	{
		sndCommand	cmd;

		t = mc;
		do
		{
			cmd.cmd = sndFreqDurationCmd;
			cmd.param1 = t*398;
			cmd.param2 = (void*)(t+40);
			(void)sndDoCommand(ch[t-1],&cmd,0);
		} while(--t);
	} else if (!quiet)
	{
		printf("Playing: %d command(s)...\n",
			r);
		if (r==1 && dispinfo)
		{
			sndMgrProcWait(smgr);
			if ((**ch).sndFrame.format.rate)
				printSndInfo(&(**ch).sndFrame.format,0);
		}
	}
	sndWaitLoop(smgr,*ch,vol,rmul,0,0);
	if (mc>1)
		do
		{
			t = mc, r = 0;
			do
				if (sndChannelPlaying(ch[t-1]))
					++r;
			while(--t);
		} while(r);
	if (!quiet)
		printf("\rFinished.\n");
err:	r =-128;
end:	sndMgrCloseVerbose(smgr);
	if (smgr)
		(void)sndMgrClose(smgr); /* automatically disposes channels */
	if (fvfile)
		fclose(fvfile);
	if (fvmap)
		free(fvmap);
	if (fvh)
		hFree(fvh);
	if (p)
		free(p);
	return r;
}
#endif

static long procSoundList(FILE *s,
	const char *vfile,
	sndFormat *f,
	uint8 *bn,
	long z,
	int32 vol,
	ufixed rmul,
	int8 mc)
{
	ufixed		rateMultiplier = 0x10000;
	long int	starto, datao = 0;
	int32		v, c, m;

	if (!s || !f)
		return -1;
	starto = ftell(s);
	v = fgetc(s)<<8;
	if ((v|=fgetc(s))<0)
		return -3;
	if (v-1u>1u)
	{
		(void)fprintf(stderr,"### bad soundlist format (%d)\n",
			v);
		return -206; /* badFormat */
	}
	if (v==1)
	{ /* parse modifiers */
		c = fgetc(s)<<8, v = m = 0;
		if ((c|=fgetc(s))<0)
			return -3;
		if ((c=(c^0x8000)-0x8000)<=0)
			if (!c)
			{
				if (!quiet)
					(void)fprintf(stderr,"## no modifiers\n");
				if (!z)
					return -128;
			} else
			{
				(void)fprintf(stderr,"### bad modifier count (%d)\n",
					c);
				return -206;
			}
		v = fgetc(s)<<8;
		if ((v|=fgetc(s))<0)
			return -3;
		if (v!=5 && !z)
		{
			(void)fprintf(stderr,"### only sampled synth supported for exporting\n");
			return -206;
		}
		m = fgetc(s)<<16, m|= fgetc(s)<<8;
		if ((m|=fgetc(s))<0)
			return -3;
		if ((v=fgetc(s))<0)
			return -3;
		m = m<<8|v;
		(void)fseek(s,6*c-6,SEEK_CUR);
	#if 0 /* not necessary */
		if (!(m&0x80))
		{
			fprintf(stderr,"### no channel init (%d)\n",
				m);
			return -206; /* no channel init */
		}
	#endif
		/* what the heck are srate flags? */
	} else /* refCount */
		(void)fseek(s,2,SEEK_CUR);
	c = fgetc(s)<<8;
	if ((c|=fgetc(s))<0)
		return -3;
	if ((c=(c^0x8000)-0x8000)<=0)
		return c ? -5:-1;
	if (z)
	{
		v = ftell(s)-starto;
		if ((z-=v)<=0)
			return -5;
#ifndef NO_SNDMGR
		fseek(s,starto,SEEK_SET);
		return loadSndPlay(s,z+v,vfile,vol,rmul,mc);
#endif
	}
	/* parse sndCommands */
	do
	{
		m = fgetc(s)<<8;
		if ((m|=fgetc(s))<0)
			return -3;
		switch(m)
		{ /* #15 on = param2 is an absolute offset (off=ptr, invalid here) */
		case 86: /* rateMultiplier - apply afterward */
			(void)fseek(s,2,SEEK_CUR); /* param1 - SM-ignored */
			v = fgetc(s)<<16, v|= fgetc(s)<<8;
			if ((v|=fgetc(s))<0)
				return -3;
			rateMultiplier = v<<8;
			if ((v=fgetc(s))<0)
				return -3;
			rateMultiplier|= v;
			break;
		case 0x8050: /* soundCmd */
		case 0x8051: /* bufferCmd */
			if (datao)
			{
				(void)fprintf(stderr,"## more than one bufferCmd/soundCmd\n");
				goto skip;
			}
			(void)fseek(s,2,SEEK_CUR); /* param1 - ignored */
			v = fgetc(s)<<16, v|= fgetc(s)<<8;
			if ((v|=fgetc(s))<0)
				return -3;
			datao = v<<8;
			if ((v=fgetc(s))<0)
				return -3;
			datao|= v;
		case sndNullCmd:
			break;

		default:
			(void)fprintf(stderr,"## unsupported SndCommand for exporting (#%d)\n",
				m);
		skip:	(void)fseek(s,6,SEEK_CUR);
		}
	} while(--c);

	if (!datao)
		return -5;
	(void)fseek(s,datao+starto,SEEK_SET);
	if ((starto=readBufferCmdHeader(s,f,bn))<=0)
		return starto;
	if (rateMultiplier!=0x10000)
	{ /* normally, this and volumeCmd would be resampled through a mixer */
		uint64		rate;

		(void)mulu64u32(rate,f->rate<<16|f->rateFrac>>16,rateMultiplier);
		f->rate = (uint32)int64hi(rate), f->rateFrac = int64lo(rate);
	}
	return datao+starto;
}

static int readExtHeader(FILE *s,
	sndFormat *f,
	uint8 *bn,
	uint8 m)
{
	uint8		b[64];
	int		r;

	if (!s || !f)
		return -1;
	memset(b,0,22);
	(void)fread(b+22,1,sizeof(b)-22,s);
	b[7] = (m&1)+1, b[8] = 0x80, b[20] = 0xFF, b[21] = 0x3C;
	r = parseBufferCmdHeader(b,f,NULL,bn);
	if (r<=0)
		return r;
	r-= 22;
	/* ROM ext. headers contain a TOTAL 4-byte count, not numFrames */
	/* (at least in the stereo case) */
	if (m&4)
	{
		f->frameCount = b[22]<<24|b[23]<<16|b[24]<<8|b[25];
		if (f->depth<8 || f->frameCount>0x200000u)
			return -1;
		f->frameCount = ((f->frameCount<<2)-42)
			/(((f->depth+7)>>3)*f->chanCount); //&127);
		if (f->frameCount>0x3FFFFFFF)
			return -1;
	}
	return !parseExt80Rate(b+26,f) ? -1:r;
}

static int readDOCSSampleHeader(FILE *s,
	sndFormat *f,
	int32 z)
{
	uint8		b[10];
	long		c;

	if (fread(b,10,1,s)<=0)
		return -3;
	for(c=0,z-=10; z>0 && fgetc(s)>0; --z)
		++c;
	return parseSoundSampleHeader(b,f,c);
}

static int readDetectRaw(FILE *s)
{
	uint8		b[8];
	size_t		i;

	(void)fseek(s,0,SEEK_SET);
	if (fread(b,sizeof(b),1,s)!=1)
		return 0;
	if (*b-0x7Cu>15u || b[7]-0x7Cu>20u)
		return 0;
	i = 1;
	do
		if (b[i]-0x7Cu>12+i)
			return 0;
	while(++i<7);
	return 1;
}

static int outWAVE(FILE *in,
	sndFormat *f,
	const char *outn,
	int k,
	int8 wr,
	int ix)
{
extern int _sndStreamOpenClose(FILE *in,
  sndFormat *f,
  int32 k,
  int quietWarnings);

	FILE		*out;
	const uint8	*h;
	uint8		*buf;
	long		t, z = 0;
	int32		packo = sndMaxFrameSize(f);
	int		r = 0;

	if (f->streamType)
	{
		(void)fprintf(stderr,"### only playback currently supported for codec\n");
		return -1;
		//_sndStreamOpenClose(in,f,0,0);
	}
	/* allocate the buffers */
	if ((buf=calloc(packo+f->maxRead*f->blockSize,f->streamType ? 512:1))==NULL)
	{
		(void)fprintf(stderr,"### failed to allocate conversion buffers\n");
		return 1;
	}
	if (wr && !outn)
	{
		(void)fprintf(stderr,"### no destination path\n",
			outn);
		r = 1;
		goto end;
	}
	/* set up the output file */
	h = setupWAVEHeader(f,&z);
	if (!wr)
		out = NULL;
	else
	{
		if (ix)
		{
			char		*s;
			int		z;

			if ((s=strrchr(outn,'.'))!=NULL)
				z = (int)(s-outn);
			else
				z = (int)strlen(outn);
			(void)sprintf(temporary,"%.*s-%d.wav",
				z,outn,ix+1);
			outn = temporary;
		}
		if ((out=fopen(outn,"wb"))==NULL)
		{
			(void)fprintf(stderr,"### failed to open '%s' for writing\n",
				outn);
			r = 1;
			goto end;
		} else
		{
			(void)fseek(out,0,SEEK_SET);
			(void)fwrite(h,z,1,out);
		}
	}
	t = 0;
	/* some formats don't have starting samples in their packets */
	/* the universal method for seek would be to just decompress and skip */
	while((z=decompressSndFrame(buf,buf+packo,f,
			readSndFrame(buf+packo,in,f))-buf)>0)
	{
		if (k>0)
			{k-= f->maxRead; continue;}
		z = writeSndSamples(buf,z,out,f,STYPE_WAVE32,
			!out ? sio_null:swr_fwrite);
		if (z<=0)
			{r|= 1; break;}
		t+= z;
	}
	if (out)
	{
		h = setupWAVEHeader(f,&t);
		(void)fseek(out,0,SEEK_SET);
		(void)fwrite(h,t,1,out);
	} else
		(void)printf("\n % 7lu sample bytes accounted for\n\n",
			t);
	if (r)
		(void)fprintf(stderr,"## error(s) occurred during %s (%d)\n",
			out ? "conversion":"verification",r);
	r = 0;
end:	free(buf);
	return r;
}

static long long parseDuration(const char *n)
{
	long long	k;
	long int	t;

	if (!n || (n=readanyint64(n,n+strlen(n),&k))==NULL || k+0u>(uint32)~0u)
		return -1;
	if (*n==':')
	{
		++n, k*= 60;
		if ((n=readanyinteger(n,strlen(n),&t))==NULL || t<0)
			return -1;
		k+= t;
		if (*n==':')
		{
			++n, k*= 60;
			if (*n=='-' || (n=readanyinteger(n,strlen(n),&t))==NULL || t<0)
				return -1;
			k+= t;
		}
		k*= 1000;
		if (*n=='.')
			for(t=1000; t; t/=10,k+=(*n-'0')*t)
				if (*++n-'0'+0u>9)
					break;
				else if (*n=='.')
					return -1;
	}
	if (*n!='.' && !isspace(*n) && *n)
		return -1;
	return k;
}

static void printAnyDuration(uint32 dur)
{
	uint32		mins;

	if ((mins=dur/60000)!=0)
	{
		uint32		hrs = mins/60;

		dur-= mins*60000, mins-= hrs*60;
		printf(" %d:%02d:%c%g",
			hrs,mins,dur/10000+'0',dur%10000/1000.);
	} else
		printf(" %g %s",
			dur/1000.,dur==1000 ? "second":"seconds");
}

static void printDuration(const sndFormat *f,
	int userSet)
{
	uint32		dur;

	if ((dur=sndFrameToTime(IMAX(int32),f,0))==0)
		return;
	(void)printf(userSet ? userSet<0 ? "\n Duration: up to":" - for":"\n Duration:");
	printAnyDuration(dur);
	putc('\n',stdout);
}

static void printSndInfo(const sndFormat *f,
	uint8 origChanCount)
{
	static char	*types[] =
	{
		"PCM",
		"NPCM",
		"DPCM",
		"ADPCM",
		"lossless",
		"lossy/dynamic",
		"wave table",
		"tone"
	};

	const char	*n;
	double		bitrate, rate = f->rate+f->rateFrac*(1/4294967296.);
	uint8		stype = f->type, sdepth = f->mSampleSize*8;
	uint8		ctype = stype&15, ccount;

	(void)printf(f->depth!=sdepth ? "    Depth: %d/%d bits\n"
		:"    Depth: %d bits\n",
		f->depth,sdepth);
	bitrate = (f->blockSize*(f->chanCount&127)*f->mSampleSize)*rate/(f->frameSize*125);
	if (ctype || f->streamType)
		if (f->streamBitrate)
		{
			if (f->streamType==f->codec)
				(void)printf("    Codec: %s; %dB/sample\n",
						(n=sndGetCodecName(f->streamType))==NULL ? "???"
						:n,f->frameSize);
			else
				(void)printf("    Codec: %s (%s output); %dB/sample\n",
						(n=sndGetCodecName(f->streamType))==NULL ? "???"
						:n,sndGetCodecName(f->codec),f->frameSize);
			if (f->streamBitrate>1)
				bitrate = f->streamBitrate/8000.;
		} else if (f->streamType)
			(void)printf("    Codec: %s; %d:%dB raw\n",
				(n=sndGetCodecName(f->streamType))==NULL
				&& (n=sndGetCodecName(f->codec))==NULL ? "???"
				:n,f->frameSize,f->blockSize);
		else
		{
			(void)printf("    Codec: %s\n",
				(n=sndGetCodecName(f->codec))==NULL ? "???":n);
			bitrate = 0;
		}
	ccount = f->chanCount&127;
	(void)printf(f->mSampleSize && bitrate
		&& (!f->streamType || f->streamBitrate) /* uncompressed rate not helpful */
		?
"     Type: %s%s%s\n"
"     Rate: %s%dx%.*f Hz%s, %.1f kbps\n"
:"     Type: %s%s%s\n"
"     Rate: %s%dx%.*f Hz%s\n",
		stype&STYPE_FP ? stype&STYPE_ISGN ? "fp/fixed scale ":"floating point "
		:stype&STYPE_ISGN ? "unsigned ":bitrate ? "signed ":"",
		ctype || sdepth==8 || f->streamType ? "":stype&STYPE_LSB ? "(LE) ":"(BE) ",
		ctype>=GET_ARRAY_COUNT(types) ? "???":types[ctype],
		f->chanCount&128 ? "r":"",ccount,f->rateFrac ? 6:0,rate,
		origChanCount==1 && ccount>1 ? " (mono in description)":"",
		bitrate);
}

typedef enum
{
	_none,
	_bufC,
	_extM,
	_extS,
	_romM,
	_romS,
	_sndL,
	_sndS,
	_rawM,
	_rawS
} sndType;

int __cdecl main(int argc,
	char **argv)
{
	sndFormat	format;
	ResFile_Memory	*rMap;
	FILE		*in;
	char const	*n, *vfile;
	char		**a, *s, *ft;
	double		filter = 0;
	long long	k;
	long int	prefRate;
	long		o, tz, dur, z, rz, vol, ssz;
	long		rk;
	int32		r = 0;
	ufixed		rmul;
	int		rIndex;
	sndType		m;
	uint8		rLoop, play, dcvt, dchk, bn, origChanCount;
	int8		mc, rsrcfile;

	progname = *argv;
	while(takeparam&1)
	{
	 takeparam&= 2;
	 a = argv+1;
	 rMap = NULL;
	 vfile = NULL;
	 ft = NULL;
	 prefRate = 0;
	 if (argc<1)
		return 0;
	 o =-1, tz =-1, dur =-1, z = IMAX(long), vol = 0x1000100;
	 k = 0, r = 0;
	 rmul = 0x10000;
	 rIndex = 0;
	 m = _none;
	 rLoop = 0, play = 1, dcvt = 0, dchk = 0, bn = 0;
	 mc = 0, rsrcfile = 0;
	 memset(&format,0,sizeof(format));
	 origChanCount = 0;

	 while(argc>1 && *(s=*a)=='-')
		if (--argc,++a,!(++s)[1])
			if (*s=='k')
				goto ktime;
			else if (*s=='d')
			{
				if (*a==NULL || !readanyinteger(*a,(long)strlen(*a),&dur) || dur<0)
				{
					(void)fprintf(stderr,"### bad duration value '%s'\n",
						*a);
					argc = 0;
					break;
				}
				--argc, ++a;
			} else if (*s=='v')
			{
				long int	rv;

				if (*a==NULL || (n=readanyinteger(*a,(long)strlen(*a),&vol))==NULL
						|| vol<0)
					goto badvol;
				rv = vol;
				if (*n++==',' && !readanyinteger(n,(long)strlen(n),&rv))
				{
				badvol:	(void)fprintf(stderr,"### bad volume value\n");
					argc = 0;
					break;
				}
				vol = (vol*64+12)/25, rv = (rv*64+12)/25;
				vol = vol<<16|rv;
				--argc, ++a;
			} else if (*s=='r')
			{
				if (*a==NULL || (rmul=(ufixed)(atof(*a)*65536+.5))==0)
				{
					(void)fprintf(stderr,"### bad rate value\n");
					argc = 0;
					break;
				}
				--argc, ++a;
			} else if (*s=='a')
				rLoop^= 3;
			else if (*s=='q')
				quiet = 1;
			else if (*s=='p')
				play = 2;
			else if (*s=='i')
				dispinfo = 1, play&=~1;
			else if (*s=='I')
				dispinfo = 2, play&=~1;
			else if (*s=='z')
				dchk = 1, play = 0;
			else if (*s=='c' || *s=='w')
				dcvt = 1, play = 0;
		#ifndef NO_SNDMGR
			else if (*s=='t')
				return testSndMgr(*a ? atoi(*a):-1,0)<=0;
			else if (*s=='f')
				return testSndMgr(2,*a ? atoi(*a):0)<=0;
			else if (*s=='s')
				return testSndMgr(0,0)<=0;
			else if (*s=='V' || *s=='m')
			{
				--argc;
				vfile = NULL, mc = *s=='m' ? 8:0;
				if (fcanopen(*a)>0)
					vfile = *a;
				else if (!argc)
				{
					(void)fprintf(stderr,"### expected file path\n");
					goto usage;
				} else
					(void)fprintf(stderr,"## '%s' is not accessible\n",
						*a);
				++a;
			}
		#endif
			else
				goto unk;
		else if (!strcmp(s,"ploop"))
			play = 6;
		else if (!strcmp(s,"fv"))
			fvol = 1<<16;
		else if (!strcmp(s,"hv"))
			fvol = 1<<15;
		else if (!strcmp(s,"filt"))
			filter =-1;
		else if (!strcmp(s,"filtn"))
		{
			if (*a==NULL || (filter=(atof(*a)))<0)
			{
				(void)fprintf(stderr,"### bad rc value\n");
				argc = 0;
				break;
			}
			--argc, ++a;
		} else if (!strcmp(s,"bc"))
			{m = _bufC; goto co;}
		else if (!strcmp(s,"xm"))
			{m = _extM; goto co;}
		else if (!strcmp(s,"xs"))
			{m = _extS; goto co;}
		else if (!strcmp(s,"cm"))
			{m = _rawM; goto co;}
		else if (!strcmp(s,"cs"))
			{m = _rawS; goto co;}
		else if (!strcmp(s,"rm"))
			{m = _romM; goto co;}
		else if (!strcmp(s,"rs"))
			{m = _romS; goto co;}
		else if (!strcmp(s,"sl"))
			{m = _sndL; goto co;}
		else if (!strcmp(s,"ss"))
		{
			m = _sndS;
		co:	if (*a==NULL || (n=readanyinteger(*a,(long)strlen(*a),&o))==NULL || o<0)
			{
				(void)fprintf(stderr,"### bad offset value\n");
				argc = 0;
				break;
			}
			if (m>_romS && (*n++==',' || m>=_rawM))
			{
				ssz = 2147483647;
				if (*n==',' && m>=_rawM)
					++n;
				else if ((n=readanyinteger(n,(long)strlen(n),&ssz))==NULL || ssz<=0)
				{
					(void)fprintf(stderr,"### bad size value in '%s'\n",
						*a);
					argc = 0;
					break;
				}
				z = ssz;
				if (m>=_rawM)
				{
					format.codec = CODEC_PCM_U8;
					format.chanCount = (uint8)(m+1-_rawM);
					ssz = 22050;
					if (*n==',')
						if (*++n!=',' && (n=readanyinteger(n,
								(long)strlen(n),&ssz))==NULL)
						{
							(void)fprintf(stderr,"### bad rate value in '%s'\n",
								*a);
							argv = 0;
							break;
						}
					format.rate = ssz;
					if (*n++==',')
						format.codec = MK4CC(*n,n[1],n[2],n[3]);
				}
			} else if (m==_sndL)
				z = 0;
			--argc, ++a;
		} else if (!strcmp(s,"smr"))
		{
			if (*a==NULL || (n=readanyinteger(*a,(long)strlen(*a),&prefRate))==NULL
					|| prefRate<=0)
				goto badvol;
			--argc, ++a;
		} else if (!strcmp(s,"seek"))
		{
		ktime:	k = parseDuration(*a);
			if ((uint64)k+0u>(uint32)~0u)
			{
				(void)fprintf(stderr,"### bad seek value '%s'\n",
					*a);
				argc = 0;
				break;
			}
			--argc, ++a;
		} else
		{
		unk:	(void)fprintf(stderr,"### unknown option '%s'\n",
				s-1);
			goto usage;
		}

	 if (argc<2)
	 {
	 usage:	(void)fprintf(stderr,
"# usage: %s [options] infile\n"
"#\n"
"# -bc <offset>   bufferCmd\n"
"# -xm <offset>   extended, mono\n"
"# -xs <offset>   extended, stereo\n"
"# -rm <offset>   ROM extended, mono\n"
"# -rs <offset>   ROM extended, stereo\n"
"# -sl <ofs[,sz]> Mac SoundList - virtual SM with sz+play\n"
"# -ss <ofs[,sz]> DOC rSoundSample\n"
"# -cm <o,z,r,4c> raw-codec, mono\n"
"# -cs <o,z,r,4c> raw-codec, stereo\n"
"# -a             process all, not one (rsrc fork case)\n"
"# -k | -seek <t> starting time ([h:]mm:ss)/milliseconds\n"
"# -d <ms>        max. duration (milliseconds)\n"
"# -q             quiet messages except -i\n"
"# -i             print info\n"
"# -c | -w        convert (to WAVE; path follows infile)\n"
"# -z             check for errors (only)\n"
#ifndef NO_SNDMGR
"# -p             play (default on single path)\n"
"# -ploop         play with loop\n"
"# -v             set playback volume\n"
"# -r             set playback rate multiplier\n"
"# -t <secs>      time sound playback (silent)\n"
"# -f <int Hz>    test sine wave\n"
"# -s             test frequency scales\n"
"# -V             load voice file for soundList\n"
"# -hv | -fv      play half/full-volume for device\n"
"# -m             test multiple sound channels\n"
"# -smr           force sound output sample rate\n"
"# -filt          use cubic filter\n"
#endif
			,*argv);
		return 1;
	 }
#ifdef __MT__
	//global_timer_thread = CreateTimerThread(MAXTIMERS);
#endif
	 ft = strrchr(*a,'.');
	 if ((in=fopen(*a,"rb"))==NULL)
	 {
		(void)fprintf(stderr,"### failed to open '%s' for reading\n",
			*a);
		return 1;
	 }
	 if (m>=_rawM)
		goto rraw;
	 if (!_filelength64(_fileno(in)))
	 { /* no data; check for companion fork file */
		char	*testpath, *fs;

		fclose(in); in = NULL;
		rk = strlen(*a);
		testpath = malloc(rk+14);
		if ((fs=strrchr(*a,'/'))==NULL)
			fs = *a;
		/* reusing ft since main data file type is no longer relevant */
		if ((ft=strrchr(*a,'\\'))==NULL)
			ft = *a;
		if (ft>fs)
			fs = ft;
		ft = NULL;
		rk = fs-*a;
		memcpy(testpath,*a,rk);
		strcpy(testpath+rk,"/resource.frk");
		strcat(testpath,fs);
		in = fopen(testpath,"rb");
		if (in==NULL)
		{
			strcpy(testpath+rk,"/.rsrc");
			strcat(testpath,fs);
			in = fopen(testpath,"rb");
		}
		free(testpath);
		if (in==NULL)
		{
			(void)fprintf(stderr,"### '%s' is empty\n",
				*a);
			return 1;
		}
		rsrcfile = 1;
	  }
	  if (o<0)
	  {
		RsrcInfo	info;
		int		err, mpegext;

		if (rsrcfile)
			z = 0;
		else
		{
			mpegext = ft && (!_stricmp(ft,".mpg") || !_stricmp(ft,".mpeg")
				|| !_stricmp(ft,".mp2") || !_stricmp(ft,".mp3"));
			if ((z=readAudioHeader(in,&format,mpegext ? quiet ? -1:0:1))!=0
					&& z!=-1)
			{
				if (format.streamType==MK4CC('s','n','d',' '))
				{
					o = z;
					z = ((streamwave_base*)format.variables)->initSize;
					m = _sndL;
					goto floop;
				}
				/* MP4 player can force stereo out, so we want to save description */
				origChanCount = format.chanCount&127;
				rz = z;
				goto goti;
			}
		}
 		if ((z=ResFile_LoadMap(in,0,&rMap,0))>=0)
		{
			printf(" - rsrc file\n");
		nextR:	err =-192;
			(void)ResFile_GetIndResource(rMap,MK4CC('s','n','d',' '),~rIndex,&info,&err);
			if (!err && info.offsetOnDisk)
			{
				if (!quiet)
					printf(" -- 'snd ' ID %d '%.*s'\n",
						info.ID,*info.name,1+info.name);
				m = _sndL, o = info.offsetOnDisk;
				err = 1, z = info.sizeOnDisk;
				goto gots;
			}
			(void)ResFile_GetIndResource(rMap,MK4CC('8','0','2','4'),~rIndex,&info,&err);
			if (!err && info.offsetOnDisk)
			{
				if (!quiet)
					printf(" -- rSoundSample ID %d '%.*s'\n",
						info.ID,*info.name,1+info.name);
				m = _sndS, o = info.offsetOnDisk;
				err = 1, z = info.sizeOnDisk;
			}
		gots:	if (err<=0)
				z = err==-192 || !err ? -5:err;
			if (z==-5 && rIndex)
				{rLoop&= 2; goto end;}
		} else if (z==_e_handler_error && !rsrcfile)
		{
			if (!readDetectRaw(in))
				{rz = z = 0; goto goti;}
			if (!quiet)
				printf(" - implied raw/uint8\n");
			m = _rawM;
			format.codec = CODEC_PCM_U8;
			format.chanCount = 1;
			format.rate = DEFAULT_RAWRATE;
			o = 0;
			goto rmax;

		rraw:	if (z==2147483647)
			{
			rmax:	(void)fseek(in,0,SEEK_END);
				z = ftell(in);
				(void)fseek(in,0,SEEK_SET);
			}
			(void)sndFixCmpType(format.codec,&format);
			if (!format.frameSize)
				format.frameSize = 1;
			sndFixCodecFrame(&format);
			format.frameCount = z/(format.frameSize);//*format.chanCount);
			(void)mulu64u32(format.totalCount,format.frameSize,format.frameCount);
			//printf("codec '%-4.4s', %llu frames, %ld per frame\n",
			//	(const char*)&format.codec,format.totalCount,format.frameSize);
		}
	 }
floop:	 rz = z;
	 if (o>=0)
	 {
		(void)fseek(in,o,SEEK_SET);
 		if (m>_bufC)
		{
	 		if (m==_sndL)
			{
				if (!play)
					z = 0;
				if (z && dispinfo && !quiet)
					printf("Parsing SoundList within '%s'%s...\n",
						*a,rsrcfile ? " (rsrc fork)":"");
				rz = procSoundList(in,vfile,&format,&bn,z,vol,rmul,mc);
				if (rz==-128)
					goto end;
	 		} else if (m==_sndS)
				rz = readDOCSSampleHeader(in,&format,z);
	 		else if (m<_rawM)
				rz = readExtHeader(in,&format,&bn,(uint8)m);
			else
				rz = 0;
		} else
			rz = readBufferCmdHeader(in,&format,&bn);
	 }
goti:	 rk = 0;
	 if (rz>0 && k && (rk=sndTimeToFrame(k,&format))<0)
		rz = rk;
	 if (rz>0 && format.frameCount==~0u)
		rz =-1;
	 if (rz<=0 && (rz || m<_rawM))
		goto err;
	 if ((format.chanCount&127)>2)
	 {
		if (dispinfo) /* at least print info if requested; 2021.05.02 */
			printSndInfo(&format,origChanCount);
		(void)fprintf(stderr,"### implementation cannot handle more than two channels\n");
		stopdammit = 1;
		goto end;
	}
	tz = MAX(o,0)+rz;
	/* check positive each time in case of overflow */
	if (dur>0 && (dur=sndTimeToFrame(dur,&format))>0)
	{
		dur-= rk;
		if ((int64)dur<(int64)format.frameCount)
			format.frameCount = dur;
	}
	/* for higher PCM seek precision */
	 if (rk>0 && format.maxRead>1)
	 { /* none of the formats in this case need pre-decomp */
	//	t = k%format.maxRead, tz+= t*format.blockSize, k-= t;
		tz+= rk*format.blockSize;
		if ((int64)rk>(int64)format.frameCount)
			format.frameCount = 0;
		else
			format.frameCount-= rk;
		rk = 0;
	 }
	/* initialize variables, seek */
	 (void)decompressSndFrame((uint8*)format.variables,NULL,&format,0);
	 (void)fseek(in,tz,SEEK_SET);
	 if (play)
	 {
	 #ifndef NO_SNDMGR
		if (!quiet)
			printf(m ? "\nPlaying sound within '%s'%s...\n":"\nPlaying '%s'...\n",
				*a,rsrcfile ? " (rsrc fork)":"");
		if (!takeparam)
		{
			if (!quiet || dispinfo)
				printDuration(&format,dur>0 ? 1:m>=_rawM ? -1:0);
			if (dispinfo)
			{
				printSndInfo(&format,origChanCount);
				if (bn)
					printf("Base Note: %u\n",
						bn);
				putc('\n',stdout);
			}
		}
		if (!quiet && k)
		{
			fputs("From",stdout);
			printAnyDuration((uint32)k);
			puts(" in\n");
		}
		rz = playFile(in,&format,k,vol,rmul,prefRate,filter,origChanCount);
	 #endif
		if (rz<=0)
		{
			if (!rz)
				(void)fputs("## failed to set up the sound output\n",stderr);
			else
			err:	(void)fprintf(stderr,!rz ? "### unsupported or corrupted file\n"
					:rz==-1 ? "### bad header format or parameter\n"
					:rz==-2 ? "### out of memory\n"
					:rz==-3 ? "### read/EOF error\n"
					:rz==-5 ? "### no sound\n"
					:rz==-8 ? "### unsupported codec\n"
					:"### error(s) occurred parsing sound header (%d)\n",
					rz);
			stopdammit = 1;
		} else
	 #ifndef NO_SNDMGR
			if (!(play&4) && !quiet)
				if (!format.dcmpRtn)
					printf("\nDecoding of codec not supported.\n\n");
				else
					printf("\nDone playing.\n\n");
	 #else
			printf("### Sound playback not implemented for this platform.\n\n");
	 #endif
		goto end;
	 } else if (!quiet)
	 {
		if (dcvt|dchk)
			printf("Decoding '%s'...\n",
				*a);
		if (dispinfo)
		{
			printDuration(&format,dur>0);
			printSndInfo(&format,origChanCount);
		}
		if (!(dcvt|dchk))
			(void)fputc('\n',stdout);
	 }
	 r = dcvt|dchk ? outWAVE(in,&format,a[1],k,dcvt,rIndex):0;
end:	 if (rMap)
	 {
		if (!stopdammit)
		{
			if (rLoop&1)
			{
				++rIndex;
				goto nextR;
			}
			if (play&4)
			{ /* DANGER, WILL ROBINSON!! INFINITE ANNOYANCE ALERT!! */
				rIndex = 0, rLoop|= rLoop>>1&1;
				//_clrscr();
				if (!(play&8))
					(void)printf("\n looping...\n");
				play|= 8, quiet = 1;
				goto nextR;
			}
		}
		free(rMap);
		rMap = NULL;
	 } else if (r>=0 && !stopdammit && play&4)
	 {
		//_clrscr();
		if (!(play&8) && !quiet)
			(void)printf("\n looping...\n");
		play|= 8, quiet = 1;
		goto floop;
	 }
	 (void)fclose(in);
 #ifdef __MT__
	//(void)EndTimerThread(global_timer_thread);
 #endif
 #ifdef WINUSERAPI
	 if (takeparam==1)
	 { /* getting a quote for a file */
		int		i, f, q;

		takeparambuf = realloc(takeparambuf,999);
		if (!takeparambuf)
			goto perr;
		*takeparamarr = progname;
		takeparamarr[1] = takeparambuf;
		putc('"',stdout);
		for(i=f=0,q=1,argc=2;;q&=2)
		{
			if ((r=_getch())==0)
				continue;
			if (r=='"')
			{
				takeparambuf[i++] = 0;
				takeparamarr[1+argc] = takeparambuf+i;
				break;
			}
			if (r=='-' && q==1)
			{ /* user manually entering switches, not a file name */
				q = 2; /* set the bit that can't be cleared */
			}
			if (r==27)
				return 0; /* cancel */
			if (r==0xE0 || argc==sizeof(takeparamarr)/sizeof(*takeparamarr)
					|| i>=998)
				return 1; /* special, overflow */
			/* append char */
		#if 0
			if (((i+63)&~63)>(i&~63))
				takeparambuf = realloc(takeparambuf,(i+63)&~63);
		#endif
			if (!takeparambuf)
			{
			perr:	fputs("### internal error: param realloc failed\n",stderr);
				return 1;
			}
			if (r==' ' && q==2)
			{ /* terminate and add next param */
				takeparambuf[i++] = 0;
				takeparamarr[argc++] = takeparambuf+i;
			} else if (r==8) /* backspace */
			{
				if (i<=f)
					continue;
				--i;
			} else /* append character */
				takeparambuf[i++] = r;
			putc(r,stdout);
		}
		argv = takeparamarr;
		puts("\"\n");
		//for(i=0; i<argc; ++i)
		//	printf("%s\n",takeparamarr[i]);
		r = 0;
	 }
 #else
	 takeparam = 0;
 #endif
	}
	if (takeparambuf)
		{free(takeparambuf); takeparambuf = NULL;}
	return r;
}
