/*
 snddec.h
  sound decoder package

  Created: 2014.01.13 23:39 EST
  Revised: 2024.05.16

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#ifndef __SNDDEC_H
 #define __SNDDEC_H

#ifdef __cplusplus
extern "C"
{
#endif

 #include <stdlib.h>
 #include <pshpack2.h>

/* ---- constants */

#define CTYPE_PCM		0 /* linear PCM */
#define CTYPE_NPCM		1 /* non-linear PCM */
#define CTYPE_DPCM		2 /* delta-modulated PCM (may use non-linear tables) */
#define CTYPE_ADPCM		3 /* Adaptive Differential PCM */
#define CTYPE_PACKED		4 /* lossless */
#define CTYPE_DYNAMIC		5 /* lossy psychoacoustic, other */
#define CTYPE_WTAB		6 /* wave table */
#define CTYPE_TONE		7 /* generated waveform */

/* ---- codec four-char-codes */
/* -- PCM */
#define CODEC_PCM		MK4CC('l','p','c','m')
#define CODEC_PCM_U8		MK4CC('r','a','w',' ') /* offset-binary: 0=127.5 */
#define CODEC_PCM_DEFAULT	MK4CC('N','O','N','E')
#define CODEC_PCM_BE_S16	MK4CC('t','w','o','s') /* int16 MSB */
#define CODEC_PCM_LE_S16	MK4CC('s','o','w','t') /* int16 LSB */
#define CODEC_PCM_BE_S24	MK4CC('i','n','2','4') /* int24 MSB */
#define CODEC_PCM_BE_S32	MK4CC('i','n','3','2') /* int32 MSB */
#define CODEC_PCM_LE_S32	MK4CC('2','3','n','i') /* int32 LSB */
#define CODEC_PCM_BE_F32	MK4CC('f','l','3','2') /* float MSB */
#define CODEC_PCM_BE_F64	MK4CC('f','l','6','4') /* double MSB */
/* -- non-linear PCM */
#define CODEC_ALAW		MK4CC('a','l','a','w') /* CCITT 12-bit alaw 2:1 */
#define CODEC_ULAW		MK4CC('u','l','a','w') /* CCITT 12-bit [micro]law 2:1 */
/* -- table DPCM */
/* Macintosh Audio Compression/Expansion */
#define CODEC_MACE_3		MK4CC('M','A','C','3') /* 9-bit/pair MACE 3:1 */
#define CODEC_MACE_6		MK4CC('M','A','C','6') /* 10-bit MACE 6:1 */
/* Apple IIgs Audio Compression/Expansion */
#define CODEC_ACE_2		MK4CC('A','C','E','2') /* ACE 2:1 */
#define CODEC_ACE_8		MK4CC('A','C','E','8') /* ACE 8:3 */
/* -- ADPCM */
#define CODEC_AIFC_IMA		MK4CC('i','m','a','4') /* 16-bit no-multiply IMA 64:17 */
/* the CD/XA codecs are h/w decompressed using the Enhanced Apple Sound Chip (EASC) */
#define CODEC_CDXA_4		MK4CC('c','d','x','4') /* 16-bit BRR CD/XA 4:1 (56:15) */
#define CODEC_CDXA_2		MK4CC('c','d','x','2') /* 16-bit BRR CD/XA 2:1 */

#define CODEC_ADP4		MK4CC('A','D','P','4')
/* 4XM is LE IMA but with all-channels-in-one packets (pred.w[ch],ix.w[ch]); */
/* same w. DK4, but header is (samp.w,ix.w)[ch] */
/* nybbles are swapped for ISS (pred.w,ix.w)[ch] */
/* EA EACS is 32-bit (ix.l[ch],pred.l[ch]):ads=6, and APC has no header */

/* -- MPEG */
#define CODEC_MPEG_AUDIO	MK4CC('m','p','g','a') /* default */
#define CODEC_MPEG_L1P		MK4CC('.','m','p','1')
#define CODEC_MPEG_L2P		MK4CC('.','m','p','2')
#define CODEC_MPEG_L3P		MK4CC('.','m','p','3') /* MPEG CBR/VBR - QT 4.1+ */
#define CODEC_AAC		MK4CC('a','a','c',' ') /* LC-AAC by default */
#define CODEC_LC_AAC		MK4CC('a','a','c','l')

/* other */
#define CODEC_DVCA		MK4CC('d','v','c','a') /* DV Audio */
#define CODEC_QDESIGN		MK4CC('Q','D','M','C') /* QDesign music */
#define CODEC_QDESIGN2		MK4CC('Q','D','M','2') /* QDesign2 music */
#define CODEC_QUALCOMM_PV	MK4CC('Q','c','l','p') /* QUALCOMM PureVoice */
#define CODEC_GSM		MK4CC('G','S','M',' ')
#define CODEC_G726_ADPCM	MK4CC('G','7','2','6')
#define CODEC_DWVW		MK4CC('D','W','V','W') /* TX16W Typhoon Delta-W.-Var.-Word */
#define CODEC_SDX2		MK4CC('S','D','X','2') /* sqrt-delta, 3DO / MAC */
#define CODEC_SPEEX		MK4CC('s','p','e','x')
#define CODEC_VORBIS		MK4CC('v','o','r','b')
#define CODEC_OPUS		MK4CC('o','p','u','s')
#define CODEC_ATRAC		MK4CC('a','t','r','c')
#define CODEC_AMR		MK4CC('s','a','m','r')
#define CODEC_SONIC		MK4CC('S','O','N','C')
#define CODEC_WMA1		MK4CC('w','m','a','1')	/* Windows Media Audio v1 */
#define CODEC_WMA2		MK4CC('w','m','a','2')	/* Windows Media Audio v2 */
#define CODEC_VOXWARE_RT24	MK4CC('r','t','2','4')
#define CODEC_VOXWARE_RT29	MK4CC('r','t','2','9')
#define CODEC_AC3		MK4CC('a','c','-','3') /* Dolby AC-3 / A/52 */
#define CODEC_ILBC		MK4CC('i','l','b','c') /* Internet Low Bitrate Codec */

/* VOC SoundBlaster Pro ADPCM; used internally until an official fourcc is known */
#define CODEC_SBPRO_ADPCM_4	MK4CC('S','B','P',4) /* 4-bit */
#define CODEC_SBPRO_ADPCM_3	MK4CC('S','B','P',3) /* 2.6-bit */
#define CODEC_SBPRO_ADPCM_2	MK4CC('S','B','P',2) /* 2-bit */

/* lossless compression */
#define CODEC_ALAC		MK4CC('a','l','a','c') /* Apple Lossless Audio Codec */
#define CODEC_FLAC		MK4CC('f','l','a','c') /* Free Lossless Audio Codec */
#define CODEC_TAK		MK4CC('t','a','k',' ') /* Tom's verlustfreier Audiokompressor */
#define CODEC_MONKEYS_AUDIO	MK4CC('A','P','E',' ')
#define CODEC_WAVPACK		MK4CC('W','V','P','K')

/* unofficial little endian ordering types */
#define CODEC_PCM_LE_F32	MK4CC('2','3','l','f') /* float LSB */
#define CODEC_PCM_LE_F64	MK4CC('4','6','l','f') /* double LSB */

/* Microsoft WAVE Format types - ID 0=unknown */
/* most of these are here merely for completeness */
#define CODEC_MS_PCM		MK4CC('m','s',0,1)
#define CODEC_MS_ADPCM		MK4CC('m','s',0,2)	/* 16-bit Microsoft ACM ADPCM 253:64 */
#define CODEC_MS_FLOAT		MK4CC('m','s',0,3)
#define CODEC_MS_VSELP		MK4CC('m','s',0,4)	/* Compaq Computer Corporation */
#define CODEC_IBM_CVSD		MK4CC('m','s',0,5)	/* IBM Corporation */
#define CODEC_MS_ALAW		MK4CC('m','s',0,6)
#define CODEC_MS_ULAW		MK4CC('m','s',0,7)
#define CODEC_MS_DTS		MK4CC('m','s',0,8)
#define CODEC_MS_DRM		MK4CC('m','s',0,9)
#define CODEC_OKI_ADPCM		MK4CC('m','s',0,16)	/* 12-bit OKI */
#define CODEC_DVI_IMA		MK4CC('m','s',0,17)	/* 16-bit DVI IMA 4:1 (255:64) */
#define CODEC_MEDIASPACE	MK4CC('m','s',0,18)	/* Videologic MediaSpace ADPCM */
#define CODEC_SIERRA_ADPCM	MK4CC('m','s',0,19)	/* Sierra Semiconductor Corp */
#define CODEC_G723_ADPCM	MK4CC('m','s',0,20)	/* Antex Electronics Corp. */
#define CODEC_DSP_DIGISTD	MK4CC('m','s',0,21)	/* DSP Solutions, Inc. */
#define CODEC_DSP_DIGIFIX	MK4CC('m','s',0,22)
#define CODEC_DIALOGIC_OKI	MK4CC('m','s',0,23)
#define CODEC_MEDIAVISION	MK4CC('m','s',0,24)
#define CODEC_CU_HP		MK4CC('m','s',0,25)	/* CU codec - Hewlett-Packard Company */
#define CODEC_YAMAHA_ADPCM	MK4CC('m','s',0,32)
#define CODEC_SONARC		MK4CC('m','s',0,33)	/* Speech Compression */
#define CODEC_TRUESPEECH	MK4CC('m','s',0,34)	/* DSP Group TrueSpeech[TM] */
#define CODEC_ECHOSC1		MK4CC('m','s',0,35)	/* Echo Speech Corp. */
#define CODEC_AUDIOFILE_AF36	MK4CC('m','s',0,36)	/* Audiofile, Inc. */
#define CODEC_APTX		MK4CC('m','s',0,37)	/* Audio Processing Technology */
#define CODEC_AUDIOFILE_AF10	MK4CC('m','s',0,38)
#define CODEC_PROSODY_1612	MK4CC('m','s',0,39)	/* Aculab plc */
#define CODEC_LRC		MK4CC('m','s',0,40)	/* Merging Technologies S.A. */
#define CODEC_DOLBY_AC2		MK4CC('m','s',0,48)
#define CODEC_MS_GSM_610	MK4CC('m','s',0,49)	/* GSM 6.10 - Microsoft Corp */
#define CODEC_MSNAUDIO		MK4CC('m','s',0,50)
#define CODEC_ANTEX_ADPCME	MK4CC('m','s',0,51)	/* Antex Electronics Corp. */
#define CODEC_CONTROLRES_VQLPC	MK4CC('m','s',0,52)	/* Control Resources Ltd. */
#define CODEC_DIGIREAL		MK4CC('m','s',0,53)	/* DSP Solutions, Inc. */
#define CODEC_DIGIADPCM		MK4CC('m','s',0,54)
#define CODEC_CONTROLRES_CR10	MK4CC('m','s',0,55)	/* Control Resources Ltd. */
#define CODEC_NMS_VBXADPCM	MK4CC('m','s',0,56)	/* Natural MicroSystems */
#define CODEC_ROLAND_RDAC	MK4CC('m','s',0,57)
#define CODEC_ECHOSC3		MK4CC('m','s',0,58)	/* Echo Speech Corp. */
#define CODEC_ROCKWELL_ADPCM	MK4CC('m','s',0,59)	/* Rockwell Intl. */
#define CODEC_ROCKWELL_DIGITALK	MK4CC('m','s',0,60)	/* Rockwell Intl. */
#define CODEC_XEBEC		MK4CC('m','s',0,61)	/* Xebec Multimedia Solutions Ltd */
#define CODEC_G721_ADPCM	MK4CC('m','s',0,64)	/* Antex Electronics Corp. */
#define CODEC_G728_CELP		MK4CC('m','s',0,65)	/* Antex Electronics Corp. */
#define CODEC_G723_1		MK4CC('m','s',0,66)	/* Microsoft G.723.1 */
#define CODEC_MS_MPEG		MK4CC('m','s',0,80)	/* Microsoft custom MPEG */
#define CODEC_RT24		MK4CC('m','s',0,82)	/* InSoft Inc. */
#define CODEC_PAC		MK4CC('m','s',0,83)	/* InSoft Inc. */
#define CODEC_MPEG_L3		MK4CC('m','s',0,85)	/* MPEG - CBR-only, pre-QT 4.1 */
#define CODEC_LUCENT_G723	MK4CC('m','s',0,89)	/* Lucent Technologies */
#define CODEC_CIRRUS		MK4CC('m','s',0,96)	/* Cirrus Logic */
#define CODEC_ESPCM		MK4CC('m','s',0,97)	/* ESS Technology */
#define CODEC_MS_G722_ADPCM	MK4CC('m','s',0,101)	/* APICOM */
#define CODEC_MS_DSAT		MK4CC('m','s',0,102)
#define CODEC_DSAT_DISPLAY	MK4CC('m','s',0,103)
#define CODEC_SOFTSOUND		MK4CC('m','s',0,128)	/* Softsound, Ltd. */
#define CODEC_MS_RT24		MK4CC('m','s',0,130)	/* Microsoft Corp. */
#define CODEC_G729A		MK4CC('m','s',0,131)	/* AT&T Labs, Inc. */
#define CODEC_DF_GSM610		MK4CC('m','s',0,134)	/* DataFusion Systems (Pty) (Ltd) */
#define CODEC_ISIAUDIO		MK4CC('m','s',0,136)	/* Iterated Systems, Inc. */
#define CODEC_ONLIVE		MK4CC('m','s',0,137)	/* OnLive! Technologies, Inc. */
#define CODEC_MULTITUDE_FT_SX20	MK4CC('m','s',0,138)	/* Multitude, Inc. */
#define CODEC_CONVEDIA_G729	MK4CC('m','s',0,140)	/* Convedia Corp. */
#define CODEC_SBC24		MK4CC('m','s',0,145)	/* Siemens Business Communications */
#define CODEC_DOLBY_AC3_SPDIF	MK4CC('m','s',0,146)
#define CODEC_MEDIASONIC_G723	MK4CC('m','s',0,147)
#define CODEC_PROSODY_8K	MK4CC('m','s',0,148)	/* Aculab plc */
#define CODEC_ZYXEL_ADPCM	MK4CC('m','s',0,151)	/* ZyXEL Communications, Inc. */
#define CODEC_LPCBB		MK4CC('m','s',0,152)	/* Philips */
#define CODEC_WRPR		MK4CC('m','s',0,244)
#define CODEC_AAC_RAW1		MK4CC('m','s',0,255)
#define CODEC_RHETOREX_ADPCM	MK4CC('m','s',1,0)	/* Rhetorex, Inc. */
#define CODEC_PHILIPS_CELP	MK4CC('m','s',1,32)
#define CODEC_ACELP_NET		MK4CC('m','s',1,48)	/* Sipro Lab Telecom ACELP.net */
#define CODEC_SIPRO_KELVIN	MK4CC('m','s',1,53)
#define CODEC_MS_WMA1		MK4CC('m','s',1,96)
#define CODEC_MS_WMA2		MK4CC('m','s',1,97)
#define CODEC_WMAPRO		MK4CC('m','s',1,98)
#define CODEC_WMALOSSLESS	MK4CC('m','s',1,99)
#define CODEC_UNISYS_NAP_ADPCM	MK4CC('m','s',1,112)
#define CODEC_MPEG2_AAC		MK4CC('m','s',1,128)
#define CODEC_DTS_DS		MK4CC('m','s',1,144)
#define CODEC_ADPCM_CT		MK4CC('m','s',2,0)
#define CODEC_UHER_ADPCM	MK4CC('m','s',2,16)
#define CODEC_SANYO_LDADPCM	MK4CC('m','s',2,37)
#define CODEC_ATRAC3		MK4CC('m','s',2,112)
#define CODEC_IMC		MK4CC('m','s',4,1)
#define CODEC_IAC_2		MK4CC('m','s',4,2)	/* Indeo(R) v2-32 */
#define CODEC_VME_VMPCM		MK4CC('m','s',6,128)
#define CODEC_CJ_LIGHTWAVE	MK4CC('m','s',8,174)	/* ClearJump Lightwave Lossless */
#define CODEC_LH		MK4CC('m','s',17,0)	/* Lernout & Hauspie */
#define CODEC_GSM_MS		MK4CC('m','s',21,0)
#define CODEC_TRUESPEECH_X	MK4CC('m','s',21,1)	/* another TrueSpeech ID */
#define CODEC_AAC_LATM		MK4CC('m','s',22,2)
#define CODEC_DVM		MK4CC('m','s',32,0)	/* FAST Multimedia AG, AC3 */
#define CODEC_DTS		MK4CC('m','s',32,1)
#define CODEC_AAC_WAV		MK4CC('m','s','A','C')
#define CODEC_AAC_PM		MK4CC('m','s','p','m')
#define CODEC_AAC_ADTS		MK4CC('m','s',129,128)

#define STYPE_ISGN		0x80 /* in: invert sign (U8, U16, U32); out: U8 for S8 */
#define STYPE_LSB		0x40 /* natively or stored little endian */
#define STYPE_FP		0x20 /* floating point format (float/double) */
#define STYPE_MAX16		0x10 /* maximum storage depth of 16 bits (i.e., WAVE) */
/* values to pass as an stype parameter */
/* ISGN in this case means u8 for 8-bit */
#define STYPE_STDMAX16		(STYPE_ISGN|STYPE_LSB|STYPE_MAX16)
#define STYPE_WAVE32		STYPE_STDMAX16
#define STYPE_WAVE64		(STYPE_ISGN|STYPE_LSB|STYPE_FP)

#define STYPE_ENDIAN_NATIVE	(!BIG_ENDIAN<<6)

/* sound command numbers */

enum
{ /* most of these are in Mac OS order, and some are not backward-compatible */
	sndNullCmd,
	sndAllocCmd,			/* allocates channel's unpacking buffer */
	sndFreeCmd,			/* frees channel's unpacking buffer (silence till init) */
	sndQuietCmd,			/* ends current sound */
	sndFlushCmd,			/* clears queue */
	sndReInitCmd,			/* param2=initFlag */
	sndFormatCmd,			/* load a raw sndFormat */
	sndRawCmd,			/* buffer new samples/current fmt.; param1=frameCount */
	sndSoundCmd,			/* 8 - Mac OS MSB soundCmd - samplePtr not supported */
	sndBufferCmd,			/* Mac OS MSB bufferCmd - samplePtr not supported */
	sndWaitCmd,			/* param1=half-ms queue delay */
	sndPauseCmd,			/* pause queue */
	sndResumeCmd,			/* resume queue */
	sndCallBackCmd,			/* calls param2 as sndCallBack */
	sndSyncCmd,			/* wait until next out-block (helps sync channels) */
	sndStopCmd,			/* stop playback and queue until resume command */
	sndLRVolumeCmd,			/* 16 - left volume in high word param2, right in low */
	sndRLVolumeCmd,			/* right volume in high word param2, left in low */
	sndRateCmd,			/* set sample rate of current format - int/frac */
	sndRateMultiplierCmd,		/* param2=ufixed multiplier; param1=pre-fraction */
	sndFreqDurationCmd,		/* param1=half-ms; param2=MIDI note, amp in high byte */
	sndRestCmd,			/* param1=half-ms delay, with completion */
	sndFreqCmd,			/* param2=MIDI note */
	sndAmpCmd,			/* param1=MIDI amplitude */
	sndToneCmd,			/* 24 - param1.w=timbre; type/res. in high bytes */
	sndNoiseCmd,			/* param1.w=timbre */
	sndMWaveCmd,			/* param2=ptr; param1=size.w, type/res. in high bytes */
	sndPhaseCmd,			/* set phase for current tone/wave */
	sndOutRateCmd,			/* set out-rate bypassing in-rate and multiplier */
	sndVoiceCmd,			/* set raw voice for current format */
	sndLoopCmd,			/* set loop for current voice - start/end */
	sndFilterCmd			/* set filter level (0=off/linear) */
	/* getVolumeCmd: ->LRVolume */
	/* getRateMultiplierCmd: ->sndFrame.rateMul */
};

enum
{ /* new channel/reinit channel and sndMgrMute flags */
	/* volume muting bits */
	sndEnableLeft		= 1,
	sndEnableRight		= 2,
	sndInitPanMask		= 3,
	/* number of channels to expect; defaults to mono if none set */
	sndInitStereo		= 0x40,
	sndInitChanCountMask	= 0xC0,
	/* rate conversion bits */
	/* note that turning on filtering will demand more from the command queue
	  and promote latency */
	sndInitLinearInterp	= 0,	/* default */
	sndInitNoInterpolation	= 4,	/* request zero-order-hold */
	sndInitExtendedFilter	= 8,	/* request lowpass */
	sndInitFilterMask	= 0xC
};

enum
{ /* voice modes */
	sndVoiceDisabled,		/* disables channel playback completely */
	sndVoiceLocked,			/* disables switching of sampling context */
	sndVoiceDefault,		/* enables all context switching */
	sndVoiceAlwaysBuffer		/* soundCmd acts like bufferCmd */
};

enum
{
	sndToneSquare,			/* positive leading */
	sndTonePulse1_8,		/* 1/8 thresholds, negative leading */
	sndTonePulse2_8,
	sndTonePulse3_8,
	sndTonePulse4_8,
	sndTonePulse5_8,
	sndTonePulse6_8,
	sndTonePulse7_8,
	sndToneTriangle,
	sndToneRamp,
	sndToneDualHarmonic,		/* 8-step 1x+2x sinusoid */
	sndToneHalfSine
};

/* actFlags */
/* these should only be treated read-only at user level */
/* use sound commands for proper results */
#define chPaused		1
#define chStopped		2
#define chCompletion		4
#define chFreqSet		8
#define chDurAct		0x10
#define chDurAux		0x20
#define chDisposed		0x40
#define chBusy			0x80

/* ---- typedefs */

#define DEF_MIX_RTN(n)		void n(int32*d, const void*p, sndMixerFrame*m, int c)
#define TYP_MIX_RTN(n)		void(*n)(int32*d, const void*p, struct sndMixerFrame*, int c)

#define DEF_DCMP_RTN(n)		uint8* n(uint8*d, const uint8*s, sndFormat*f, long z)
#define TYP_DCMP_RTN(n)		uint8*(*n)(uint8*d, const uint8*s, struct sndFormat*f, long z)

typedef TYP_MIX_RTN(mixerRoutine);

typedef uint8 *(*mixerDcmpRtn)(uint8 *d,
	const uint8 *s, /* can be NULL */
	struct sndFormat *f,
	long z);

typedef struct sndFormat/*(92+)*/
{
	uint32		codec;		/* four-char-type */
	uint32		rate, rateFrac;	/* sample rate and fraction */
	uint8		chanCount;	/* channel count - bit #7 = reverse ordering */
	uint8		type;		/* [isgn.1] [LSB.1] [fp.1] [max16.1] [cmptype.4] */
	uint8		depth;		/* codec precision in bits */
	uint8		mSampleSize;	/* unpacked single-channel byte size (1, 2, 4 or 8) */
/*0x10*/uint32		frameCount;	/* number of uncompressed block frames */
	int32		frameSize;	/* size of one multi-channel frame in native PCM */
	uint16		packSize;	/* compressed size for one channel in bytes */
	uint16		blockSize;	/* packSize*chanCount, w. any necessary alignment */
	/* the user can set this field after the codec has been initialized */
	uint32		maxRead;	/* maximum block read count per readSndFrame */
/*0x20*/uint64		totalCount;	/* total number of decompressed samples */
	TYP_DCMP_RTN(dcmpRtn);		/* set to NULL if struct has changed */
	uint32		streamType;	/* dynamic load/dcmp; always set to zero by default */
	uint32		streamBitrate;	/* in bits */
	int16		variables[20];	/* interframe data; MACE needs 6/chan; CD/XA 8/chan */
} sndFormat;

typedef struct adpcm_base/*(16)*/
{
	int8		coef[10];	/* applicable coefficients */
	uint16		kx;		/* applicable pre-sample mask/index */
	int16		l0, l1;		/* out-sample and modifier */
} adpcm_base;

typedef struct genwave_base
{
	uint32		phase;
	uint16		rsize;
	uint8		type, amp;
	int32		l0;
	uint32		rstep;
	int16		perm[8];
} genwave_base;

typedef struct streamwave_base
{
	void		*stream;
	int64		initOff, initSize;
	uint32		accum; /* reserved */
	int		bstream;
	int32		CBRFrameSize; /* 0 for VBR */
	int32		descData[1];
} streamwave_base;

typedef struct sndMixerFrame
{
	TYP_MIX_RTN(mixerRtn); /* loaded automatically - set to NULL */

	uint8		*cmpPos;
	uint8		*dcmpBuf; /* calloc allocated - must be >sndEndFrameCount*2 */
	int32		dcmpSize; /* byte count at dcmpBuf */
	sndFormat	format;
	ufixed		rateMul;
	uint32		LRAmp; /* ifix16 left/right amplitudes */
	/* these are automatically calculated */
	uint64		sampElapsed;
	uint32		pos; /* byte position in dcmpBuf */
	uint32		filterL, filterR;
	float		filterS[2];
	ufixed		cnvRate; /* src/dst */
	uint16		cnvFrc; /* pos. fraction; start at d/s, zero, or 0.5 if s>=2*d */
	uint8		err;
} sndMixerFrame;

typedef struct sndCommand/*(10+)*/
{
	uint16		cmd;
	uint32		param1; /* expanded for 32-bit raw frameCounts */
	void		*param2;
} sndCommand;

typedef struct sndChannel
{
	struct sndChannel	*next;

	void		*private;
	void		*userInfo, *streamEntry;
	void(*defaultCallBack)(struct sndChannel*,
		sndCommand*);
	uint32		LRVolume;
	sndMixerFrame	sndFrame;
	uint8		*durStart, *loopStart;
	int32		loopCount, durEnd, restEnd;
	uint32		waitCnt;
	uint16		waitCntF, waitCntCF;
	uint8		initFlag, actFlags;
	uint8		voiceMode;
	uint8		baseNote;
	int16		qLength, qHead, qTail;
	sndCommand	cmdInProg;
	sndCommand	queue[128];
} sndChannel;

typedef struct sndManager
{
	int8		private[1];
} sndManager;

typedef void(*sndCallBack)(struct sndChannel*,
	struct sndCommand*);

#ifndef _SIOF_DEFINED
 #include "csstream.h"
#endif

/* ---- globals */

extern const adpcm_base	brr_default; /* PSX values */
extern const int16	Alaw16btable[256], ulaw16btable[256];

/* ---- prototypes */

/*
 sndMaxFrameSize
	Computes the maximum size for one multi-channel sample frame.
	This is meant for the allocating of space for readSndFrame calls.
  Result:
	The byte size; zero on error.
*/
int sndMaxFrameSize(const sndFormat *f);

/*
 decompressSndFrame
	Decompresses memory-stored data into nativized PCM.
	This also decrements the frame count by maxFrame until zero.
	When out of source data, the destination does not advance.
	Pass NULL for the source to initialize the codec's variables.
  Result:
	The destination end pointer; NULL on error.
*/
uint8 *decompressSndFrame(uint8 *d, /* can be NULL */
	const uint8 *s, /* can be NULL */
	sndFormat *f,
	long inByteSize);
/*
 sndFixCmpType
	Updates the sound format to fit the passed codec OSType.
  Result:
	Zero on success; -8 if the codec is not supported.
*/
int sndFixCmpType(uint32 codec,
	sndFormat *f);
/*
 sndFixCodecFrame
	This is used to finish sound description parameters made from scratch.
	Used internally, you'll probably never need to use this.
*/
void sndFixCodecFrame(sndFormat *f);

/*
 sndGetCodecName
	Returns the name for a codec type.
	If it fails to identify the type, NULL is returned.
*/
const char *sndGetCodecName(int32 type);

/*
 parseBufferCmdHeader
	Parses a Mac OS buffer command header: standard, extended or compressed.
  Result:
	Zero or an error bitmask, with bit #3 or higher being more serious.
*/
int32 parseBufferCmdHeader(const uint8 *s,
	sndFormat *f,
	int32 *loop, /* can be NULL */
	uint8 *baseNote); /* can be NULL */
/*
 parseSoundSampleHeader
	Parses an Apple IIgs rSoundSample header.
  NOTE:
	The caller must pass the length, as it's not stored in the header.
*/
int parseSoundSampleHeader(const uint8*,
	sndFormat*,
	int32 length);
/*
 parseExt80Rate
	Parses an "Extended" 80-bit floating point sample rate.
	The value is stored in the codec frame's 64-bit fixed-point representation.
  Result:
	1 if valid; 0 if out of range; undefined if NaN.
*/
int parseExt80Rate(const uint8*,
	sndFormat*);
/*
 sndMIDINoteToRate
	Produces a rate from a note, given baseNote.
  NOTE:
	Zero is parsed as a normal note.
*/
uint32 sndMIDINoteToRate(uint8 note,
       uint8 baseNote);

/*
 sndTimeToFrame
	Converts a millisecond time to a frame count, rounding.
  Result:
	The frame count; negative on error.
  NOTE:
	The current implementation only handles constant bitrates.
*/
int32 sndTimeToFrame(uint32 ms,
	const sndFormat *f);
/*
 sndFrameToTime
	Converts a frame index to milliseconds time.
	The index is clipped to the remaining count.
*/
uint32 sndFrameToTime(uint32 x,
	const sndFormat *f,
	int round); /* treat rate rounded to nearest whole */

/* reading/writing */

/*
 readAudioHeader
	Identifies and reads a supported audio file header (AIFF-C/WAVE/W64).
	The full sample rate & fraction is rendered.
	In the case of a streaming file such as MPEG, a bit rate is provided.
  Result:
	The offset to the first sample in the file; non-positive on error.
*/
long int readAudioHeader(struct FILE *s,
	sndFormat *f,
	int strict); /* validate headers to spec / no frame scans */

/*
 readSndFrame
	Performs a block read for a sound format, given maxRead.
  Result:
	Byte size read to be passed to decompressSndFrame; otherwise, an error.
*/
long readSndFrame(uint8 *d,
	struct FILE*,
	const sndFormat *f);

/*
 writeSndSamples
	Writes PCM samples to an out-stream.
	It also performs any necessary conversions to fit the stype parameter.
  Result:
	The total out-length; the sum of what's returned by the write function.
*/
long writeSndSamples(const void *p,
	long byteSize,
	void *outStream,
	const sndFormat *f,
	uint8 stype, /* use the STYPE constants above */
	sio_func_t wf);

/*
 setupWAVEHeader
	Sets up a 32-bit WAVE header according to the codec type & passed size.
	This should be called before writing samples and after finishing.
	Because WAVE rates cannot have a fraction, the encoded rate is rounded.
  Result:
	The header (static global-allocated; do not dispose); NULL on error.
	'size' will be set to the header size.
*/
const uint8 *setupWAVEHeader(const sndFormat *f,
	long *size); /* can be NULL */

/* ---- sound mixer */

/*
 sndMixSamples
	Adds stereo samples, performing decompression if necessary.
	Interpolation is also performed if enabled in rflag.
  Result:
	The remaining count.
*/
int sndMixSamples(int32 *d, /* must be valid */
	sndMixerFrame*, /* must be valid */
	uint32 rflag, /* upper 8-bits=init-flags, low 24=out-rate */
	int dcmpMinCount, /* buffer must be adequate for this count */
	int outMaxCount);
#ifndef NO_LOWLEVEL_MIXER
/*
 sndUpdateMixerRtn
	Updates the conversion rate and internal mixer variables.
*/
void sndUpdateMixerRtn(sndMixerFrame *m,
	uint32 rflag);
/*
 sndMixFill
	Adds a stereo repeat of the last sample.
	The last is the current (if pos!=0), or final in dcmpBuf, if it exists.
*/
void sndMixFill(int c,
	uint32 rflag, /* upper 8-bits=init-flags */
	int32 *d, /* must be valid */
	const sndMixerFrame*); /* must be valid */
/*
 s32to16
	Performs an in-place clipping, of 32 to 16 bits.
	If rflag indicates mono, every other entry is skipped.
*/
void __fastcall s32to16(int32 c,
	uint32 rflag, /* upper 8-bits=init-flags */
	int32 *b); /* must be valid */
#endif

/* ---- sound manager */

/*
 sndNewChannel
	Creates and inserts a new channel into a manager's list.
	Use the init-flags constants above; will be mono without sndInitStereo.
	Channels are initially enabled; use sndReInitCmd to disable left/right.
*/
sndChannel *sndNewChannel(sndManager*,
	uint8 initFlags,
	void(*defaultCallBack)(sndChannel*,
		sndCommand*));
/*
 sndDisposeChannel
	Disposes a sound channel, removing it from its manager's list.
  Result:
	Returns zero on success; negative on error.
*/
int sndDisposeChannel(sndChannel*);
/*
 sndChannelResize
	Resizes the queue array of a sndChannel.
	A negative count cancels the resize.
  Result:
	Zero on success; negative on error.
*/
int sndChannelResize(sndChannel **ch,
	int count);

/*
 sndChannelPlaying
	Returns a positive value if a sound is playing or the queue has commands.
	Inactivity due to being pause won't necessarily return zero.
  Result:
	Returns zero if NULL, empty, completed or stopped.
	Bit #0 = PCM/compressed samples playing.
	Bit #1 = queue has commands.
	Bit #2 = wait or duration.
*/
int sndChannelPlaying(sndChannel*);

/*
  sndMgrFromChannel
	Returns a channel's manager, if applicable; NULL on error.
*/
sndManager *sndMgrFromChannel(sndChannel*);

/*
 sndDoCommand
	In the immediate case, the command is inserted & processed from the start.
	In the delayed case, the command is merely inserted at the queue end.
  Result:
	Zero on success; a 'full' error code if the queue is full; -1 if qLength<=0.
*/
int sndDoCommand(sndChannel *ch,
	sndCommand *cmd,
	int8 immediate);

/*
 sndPlaySndList
	Plays a Mac OS SoundList (resource data).
	This routine will resize the channel if necessary.
  Result:
	Zero on empty; positive for number of commands processed; negative on error.
*/
int sndPlaySndList(sndChannel **ch,
	const uint8 *start,
	long z,
	int8 mayresize,
	int8 immediate);

/*
 sndPlayFile
	Plays a file, given the parameters.
	Use fopen and readAudioHeader for 'inStream' and 'format'.
	Use sndPauseCmd and sndResumeCmd for pause/play.
	Passing the binary complement (~) for frameSeek starts play paused.
	This function may internally reduce the sample rate in extreme cases.
  Notes:
	This routine uses the userInfo field; it must be NULL prior to calling.
	It completely takes over the channel, and will flush the queue.
  Result:
	Positive on success; zero if seek surpassed end; otherwise, error.
*/
int sndPlayFile(struct FILE *inStream,
	sndChannel *ch,
	sndFormat *format,
	int32 frameSeek,
	int quietWarnings);
/*
 sndPlayFileStop
	Stops and disposes of the frame set up by sndPlayFile.
  Notes:
	This routine flushes the command queue to ensure stop.
	As sndPlayFile does not open a file, neither does this close.
*/
int sndPlayFileStop(sndChannel *ch);

/*
 sndMgrOpen
	Opens a virtual sound manager, given a sample rate.
	If the rate is non-positive, the default rate is used.
	The out-rate is pinned to the range [1000,100000].
  Result:
	A system-compatible output (e.g., wave mapper); NULL on error.
*/
sndManager *sndMgrOpen(int outRate,
	int stereo,
	int *err, /* can be NULL */
	int maxStreamThreads);
/*
 sndMgrClose
	Waits until last packet is finished, frees everything.
	All of its channels are disposed.
  Result:
	The number of elements disposed; negative on error.
*/
int sndMgrClose(sndManager*);
/*
 sndMgrMute
	Sets a manger's mute flags, updating its channels in accordance.
	Use the sndEnableLeft/Right constants above.  (Zero for full mute.)
*/
int sndMgrMute(sndManager*,
	uint8 LREnable);
/*
 sndMgrGetVolume/sndMgrSetVolume
	Gets or sets the manager volumes for front, back and low-fi channels.
	Returns zero on success.
*/
int sndMgrGetVolume(sndManager *smgr,
	ufixed *L,
	ufixed *R,
	ufixed *BL,
	ufixed *BR,
	ufixed *LF);
int sndMgrSetVolume(sndManager *smgr,
	ufixed L,
	ufixed R,
	ufixed BL,
	ufixed BR,
	ufixed LF);
/*
 sndMgrWait
	If a manager is active, wait one sound packet.
	This is useful for waiting until all sound has completed.
*/
int sndMgrWait(sndManager*);

/*
 sndMgrProcWait
	Use this to ensure that the manager gets cycles.
  Result:
	Zero when finished; 1 when full; otherwise, 2.
*/
void sndMgrProcWait(sndManager*);
/*
 sndMgrMixRate
	Sets the mixing rate, detuning the results & timing.
	Passed rate must be positive.
  Result:
	Old rate if positive; negative on error.
*/
int sndMgrMixRate(sndManager*,
	int rate);

int sndMgrGetMixRate(sndManager *smgr);

/*
 sndMgrLeadChannel
	Returns the leading channel in list.
	Crawling is most reliable in the thread that opened the sndManager.
*/
struct sndChannel *sndMgrLeadChannel(sndManager*);

#ifdef WINUSERAPI
HWAVEOUT sndMgrGetHWO(sndManager*);
#endif

/*
 stream thread functions
*/

#ifdef __MT__
typedef struct sndTaskEntry
{
	void (*routine)(struct sndTaskEntry*);
	void		*userInfo;
	int8		primed;
} sndTaskEntry;

typedef void (*sndStreamRoutine)(sndTaskEntry*);

sndTaskEntry *sndInstallTask(sndManager *thd,
	sndStreamRoutine rtn);
int sndRemoveTask(sndManager *thd,
	sndTaskEntry *entry);
#endif

/* ---- individual codecs */

/* either Sony (mx>8) or SNES(SPC-700) (mx<=8) */
uint8 *brr_decode(uint8 *d,
	const uint8 *s,
	int chanCount,
	int sampleByteSize, /* not counting the shift/filter byte */
	adpcm_base *f);

/* non-multiplier (lossy conditional downshifts, but faster) */
uint8 *imanm_decode(uint8 *d,
	const uint8 *s,
	int chanCount,
	int sampleByteSize, /* not counting the shift/filter byte */
	adpcm_base *f);
uint8 *ima_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f);
uint8 *ms_adpcm_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f); /* adpcm_base[2] */
uint8 *oki_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f);
uint8 *adpcm_yamaha_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f); /* adpcm_base[2] */
uint8 *ctadpcm_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f); /* adpcm_base[2] */
uint8 *sbpadpcm_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f); /* adpcm_base[2] */

uint8 *MACE3_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	int16 *f); /* int16[2] */
uint8 *MACE6_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	int16 *f); /* int16[6] */
 #include <poppack.h>

#ifdef __cplusplus
}
#endif
#endif
