/*
 sndfile.c
  file reading/writing

  Revised: 2023.08.26

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.04.24/5
  Added WAVE file parsing.
2016.06.25
  Moved [data/file] stream functions to aptly-named stream.c.
2019.09.30
  Make sure codec field is in place for AIFF.
2020.04.19
  MPEG support via mpg123.  Needs more preloading for live playback.
2021.04.21
  Handle cases for COMM after SSND in AIFF-C.
  Added CAF.
2021.07.27
  Negative 'strict' option to silence decoder warnings.
2021.07.31
  Added basic Extensible Format decoding. (Assumes mono/LR.)
2021.08.03-4
  Added .au file reading.
2021.08.06-7
  Added .voc file reading.
2022.02.04
  Got Opus playback working... decoding on x64 failing, though.
2022.02.07
  Opus fixed. (Pelles C optimizer failed on CELT and DRC.)
2022.03.02
  Added recognition of an Audition fixed-point-on-FP32 type.
2022.09.30
  Fixed 64-bit bug using fseek for file size with WAVE, others.
*/

#include <stdio.h>
#include <string.h>
#include <io.h>

#include "enviro64.h"
#include "snddec.h"
#ifdef SNDDEC_MPEG_DECODE
 #include "../mpg123/mpgaud/mpgaud.h"
#endif
#ifdef SNDDEC_OGG_DECODE
 #include "../ogg/include/vorbisfile.h"
 #ifdef SNDDEC_OPUS_DECODE
  #include "../ogg/include/opusfile.h"
  extern const OpusFileCallbacks *_op_get_file_callbacks(FILE*, int);
 #endif
#endif
#ifdef SNDDEC_MP4_DECODE
 #undef INT32_MIN
 #undef INT64_MIN
 #undef INT32_MAX
 #undef INT64_MAX
 #include "../faad2/frontend/mp4read.h"
 #include "../faad2/include/neaacdec.h"

 typedef struct mp4stream
 {
	//FILE *in;
	mp4config_t	config;
	NeAACDecHandle	hDecoder;

	mp4AudioSpecificConfig	mp4ASC;

	int32		frameSize;
	uint32		decoded;
	boolean		quiet;
 } mp4stream;
#endif

#define WAV_ROUND_ODDCHUNK 1

/* ---- basic functions */

int _sndStreamOpenClose(FILE *in,
	sndFormat *f,
	int32 k,
	int quietWarnings)
{
	streamwave_base	*s = (streamwave_base*)f->variables;

	switch(f->streamType)
	{
	case 0: if (k)
			fseek(in,sndTimeToFrame(k<0 ? ~k:k,f)*f->blockSize,SEEK_CUR);
		break;
#ifdef SNDDEC_MPEG_DECODE
	case CODEC_MPEG_L3P:
		if (in)
		{
			mpg123_handle	*mpgh;

			if ((mpgh=mpg123_new(NULL,NULL))==NULL)
				return -2;
			_lseek64(_fileno(in),s->initOff,SEEK_SET);
			if (mpg123_open_fd(mpgh,_fileno(in))!=MPG123_OK)
				return -3;
			s->stream = mpgh;
			if (quietWarnings)
				mpg123_param(mpgh,MPG123_ADD_FLAGS,MPG123_QUIET,0);
			if (k>0)
			{
				mpg123_seek(mpgh,k,SEEK_SET);
				k = 0;
			}
		} else if (s->stream)
		{
			mpg123_delete(s->stream);
			s->stream = NULL;
		}
		break;
#endif
#ifdef SNDDEC_MP4_DECODE
	case CODEC_AAC:
	{
		mp4stream	*stream = s->stream;

		if (in)
		{
			NeAACDecConfigurationPtr nconfig;
			unsigned long	rate, cc, t;

			fseek(in,s->initOff,SEEK_SET);
			if (!*((streamwave_base*)f->variables)->descData)
			{
				if (_mp4read_open(in,&stream->config,1))
					return -3;
			} else
			{
				stream->config.bits = f->depth;
				stream->config.samples = f->totalCount;
				stream->config.samplerate = f->rate;
				stream->config.channels = f->chanCount;
				stream->config.frame.maxsize = ((streamwave_base*)f->variables)->CBRFrameSize/f->blockSize;
				stream->config.bitbuf.data = malloc(MAX(stream->config.frame.maxsize,
					stream->config.buffersize));
			}
			stream->config.verbose.errors = !quietWarnings;
			stream->hDecoder = NeAACDecOpen();
			nconfig = NeAACDecGetCurrentConfiguration(stream->hDecoder);
			nconfig->outputFormat = FAAD_FMT_16BIT;
			nconfig->downMatrix = 1;
			nconfig->verbose = !quietWarnings;
			NeAACDecSetConfiguration(stream->hDecoder,nconfig);

			cc = f->chanCount;
			stream->frameSize = 1024;
			stream->quiet = !!quietWarnings;
			rate = f->rate;
			if (stream->config.asc.size)
			{
				if (NeAACDecInit2(stream->hDecoder,stream->config.asc.buf,
						stream->config.asc.size,
						&rate,&f->chanCount)<0)
					goto fail;
				/* this call is still necessary for correct frame sizing */
				/* still doesn't work well w. mono files created with FAAC */
				if (NeAACDecAudioSpecificConfig(stream->config.asc.buf,
						stream->config.asc.size,&stream->mp4ASC)>=0)
				{
					if (stream->mp4ASC.frameLengthFlag==1)
						stream->frameSize = 960;
					if (stream->mp4ASC.sbr_present_flag==1
							|| stream->mp4ASC.forceUpSampling)
					{
						stream->frameSize*= 2;
						if (stream->mp4ASC.sbr_present_flag == -1
								&& !stream->config.frame_init)
						{ /* some cases are bugged in length */
							t = f->frameCount*2;
							if (t<f->frameCount)
								t =~0u;
							f->frameCount = t;
							f->totalCount*= 2;
						}
					}
				}
			} else
			{ /* UNTESTED! */
				//printf("mp4a raw\n");
			fseek(in,s->initOff,SEEK_SET);
				mp4read_frame(in,&stream->config);
				if (NeAACDecInit(stream->hDecoder,
						(uint8*)stream->config.bitbuf.data,
						stream->config.frame.maxsize,&rate,&f->chanCount)<0)
				{
				fail:	NeAACDecClose(stream->hDecoder);
					mp4read_close(NULL,&stream->config);
					return -1;
				}
			}
			stream->config.frame_init = 1;

			if (!_heap_validate(NULL))
				dprintf("_sndStreamOpenClose: heap was damaged somewhere");

			f->rate = rate;
			k&=~(k>>31); // k = MAX(k,0);
			if (stream->config.samples)
				mp4read_seek(in,&stream->config,
					(uint64)k*stream->config.buffersize
					/stream->config.samples);
			t = f->chanCount&127;
			if (t>cc)
			{ /* no mono option? */
				cc = t/cc;
				f->blockSize*= cc;
				f->frameSize*= cc;
			}
		} else if (stream)
		{
			NeAACDecClose(stream->hDecoder);
			mp4read_close(NULL,&stream->config);
			free(stream);
			s->stream = NULL;
		}
		break;
	}
#endif
#ifdef SNDDEC_OGG_DECODE
	case CODEC_VORBIS:
		if (in)
		{
			OggVorbis_File	*vf;

			fseek(in,s->initOff,SEEK_SET);
			if ((vf=malloc(sizeof(*vf)))==0)
				return -2;
			if (ov_open_callbacks(in,vf,NULL,0,OV_CALLBACKS_NOCLOSE)<0)
				return -3;
			s->stream = vf;
			if (k>0)
			{
				ov_time_seek(vf,k/1000.);
				k = 0;
			}
			if (f->rate>96000 && quietWarnings)
			{
				vorbis_synthesis_halfrate(vf->vi,1);
				f->rate>>= 1;
			}
		} else if (s->stream)
		{
			ov_clear(s->stream);
			free(s->stream);
			s->stream = NULL;
		}
		break;
#endif
 #if defined(SNDDEC_OPUS_DECODE)
	case CODEC_OPUS:
		if (in)
		{
			OggOpusFile	*of;
			int		ret;

			fseek(in,s->initOff,SEEK_SET);
			of = op_open_callbacks(in,_op_get_file_callbacks(in,0),NULL,0,&ret);
			if (!of)
				return ret==OP_EREAD ? -3:-1;
			s->stream = of;
			if (k>0)
			{
				op_pcm_seek(of,k/48);
				k = 0;
			}
		} else if (s->stream)
		{
			op_free(s->stream);
			s->stream = NULL;
		}
		break;
#endif
	default: return -8;
	}
	return 0;
}

long readSndFrame(uint8 *d,
	FILE *s,
	const sndFormat *f)
{
	if (!d || !s || !f)
		return -1;
	switch(f->streamType)
	{
	case 0: return fread(d,1,f->blockSize*MIN(f->frameCount,f->maxRead),s);
#ifdef SNDDEC_MPEG_DECODE
	case CODEC_MPEG_L3P:
	{
		mpg123_handle	*stream = ((streamwave_base*)f->variables)->stream;
		uint8		*bx;
		size_t		count = 0;
		int		r;

		if ((r=mpg123_decode_frame(stream,NULL,&bx,&count))==MPG123_OK)
			memcpy(d,bx,count);
		else if (r==MPG123_NEW_FORMAT)
			if ((r=mpg123_decode_frame(stream,NULL,&bx,&count))==MPG123_OK)
				memcpy(d,bx,count);
		return count;
	}
#endif
#ifdef SNDDEC_MP4_DECODE
	case CODEC_AAC:
	{
		NeAACDecFrameInfo	frameInfo;

		mp4stream	*stream = ((streamwave_base*)f->variables)->stream;
		void		*sample_buffer;
		long		dur;
		unsigned int	sample_count;

		if (!_heap_validate(NULL))
			dprintf("readSndFrame: heap was broken");
		do
		{
			/*printf("frame %d/%d, 0x%llx/0x%llx\n",
				stream->config.frame.current,stream->config.frame.ents,
				stream->config.frame.data[stream->config.frame.current].offset,
				stream->config.datasize);*/
			if (mp4read_frame(s,&stream->config))
				return 0;

if (!_heap_validate(NULL))
 dprintf("readSndFrame: heap was broken");

			sample_buffer = NeAACDecDecode(stream->hDecoder,&frameInfo,
				stream->config.bitbuf.data,
				stream->config.bitbuf.size);
			if (!sample_buffer)
			{
				dprintf("readSndFrame: unable to decode AAC stream\n;g");
				return 0;
			}
		} while(!frameInfo.samples && stream->config.frame.current<=1);

if (!_heap_validate(NULL))
 dprintf("readSndFrame: heap was broken");

		dur = frameInfo.samples/frameInfo.channels;
		stream->decoded+= dur;
		if (stream->decoded>stream->config.samples)
			dur+= stream->config.samples-stream->decoded;
		if (dur>stream->frameSize)
			dur = stream->frameSize;

		if (stream->config.samplerate!=f->rate)
			sample_count = frameInfo.samples;
		else
		{
			sample_count = (unsigned int)(dur*frameInfo.channels);
			if (sample_count>frameInfo.samples)
				sample_count = frameInfo.samples;
		}
		sample_count*= sizeof(int16);
		if (frameInfo.error>0)
		{
			if (!stream->quiet)
				fprintf(stderr,"AAC decode: %s\n",
					NeAACDecGetErrorMessage(frameInfo.error));
			sample_count = 0;
		} else if (sample_count>0)
	        	memcpy(d,sample_buffer,sample_count);

if (!_heap_validate(NULL))
 dprintf("readSndFrame: heap was broken");

		return sample_count;
	}
#endif
#ifdef SNDDEC_OGG_DECODE
	case CODEC_VORBIS:
	{
		streamwave_base	*sb = (streamwave_base*)f->variables;
		OggVorbis_File	*vf = sb->stream;

		if (!vf)
			return 0;
		return ov_read(vf,(char*)d,1024,BIG_ENDIAN,2,1,&sb->bstream);
	}
#endif
#if defined(SNDDEC_OPUS_DECODE)
	case CODEC_OPUS:
	{
		streamwave_base	*sb = (streamwave_base*)f->variables;
		OggOpusFile	*of = sb->stream;
		int		z;

		if (!of)
			return 0;
		z = op_read(of,(int16*)d,1024,&sb->bstream)*f->frameSize;
		return z;
	}
#endif
	default: return 0;
	}
}

/* ---- header parsing */

static long int parseFileFORM(FILE *s,
	sndFormat *f)
{ /* FORM: AIFx */
	uint8		b[40];
	long		o =-1;
	const int32	*r;
	int32		t, z, c, sc = 0, so = 0, sz = 0;
	uint8		cmp;

	if (fread(b,8,1,s)!=1)
		return -3;
	r = (const int32*)&b;
	if ((t=(int32)va_be32(r))<=0)
		return 0;
	cmp = *r==MK4CC('A','I','F','C');
	if (!cmp && *r!=MK4CC('A','I','F','F'))
		return 0;
	for(t+=4; (t-=8)>0; t-=z)
	{
		long		a = 0;

		if (fread(b,8,1,s)!=1)
			return -3;
		if ((z=be32_native(((const int32*)b)[1]))<0)
			return -1;
		switch(*(const int32*)b)
		{
		case MK4CC('C','O','M','M'):
			if (z<18)
				return -1;
			if ((a=fread(b,1,40,s))<18)
				return -3;
			if (*b|b[6] || !b[7]) /* bad channel count, depth */
				return -1;
			/* chnCnt.w/frameCnt.l/depth.w/rate.x80 */
			f->chanCount = b[1];
			r = (const int32*)(b+2);
			f->frameCount = va_be32(r);
			/* more than one COMM, bad frame count or rate */
			if (f->depth || !parseExt80Rate(b+8,f))
				return -1;
			f->depth = b[7];
			c = 0;
			if (cmp) /* codec.l cname.pstr */
				if (z>18+3)
					c = *(const int32*)(b+18);
				else if (a>18+3 && *(const int32*)(b+18)==MK4CC('N','O','N','E')
						&& b[22]) /* COMM with bad length! */
					c = MK4CC('N','O','N','E'), z = b[22]+38-15;
			if ((c=sndFixCmpType(c,f))<0)
				return c;
			f->blockSize = f->packSize*f->chanCount;
			if (o==-1)
				o = 0;
			goto next;
		case MK4CC('S','S','N','D'):
			if (o>0) /* more than one SSND */
				return -1;
			if (z<8 || (a=fread(b,1,8,s))<8)
				return -1;
			/* data offset */
			r = (const int32*)b, o = (long)va_be32(r), c = (int32)va_be32(r);
			if (o<0 || c<0) /* QT returns -208 on bad offsets */
				return -1;
			so = o, sz = z;
			o+= ftell(s);
		default:
		next:	++z, z&=~1; /* must align! (AIFF-C.9.26.91.pdf); 2014.01.27 */
			(void)fseek(s,z-a,SEEK_CUR);
		}
	}
	if (t<-1) /* -1 valid in regards to alignment */
		return -3;
	if (!o)
		o = ftell(s), sc = 0; /* empty AIFF-C files are valid; fake offset */
	else if (f->blockSize) /* I/O block size (for alignment) ignored (QT does) */
		sc = (sz-8-so+f->blockSize-1)/f->blockSize;
	/* may have to adjust frame count if too short */
	if (sc+0u<f->frameCount)
		f->frameCount = sc; //return -1;
	sndFixCodecFrame(f);
	if (!f->codec)
		if (!f->type && f->depth==8)
			f->codec = CODEC_PCM;
	if (int64z(f->totalCount))
		int64setl(f->totalCount,sc/f->chanCount);
	return o;
}

static const uint8	extCommonGUID[14] =
	{/* 2 ms codec bytes */	0,0,0,0,0x10,0, 0x80,0,0,0xAA,0,0x38,0x9B,0x71};
static const uint8	extATRAC3GUID[16] =
{
	0xBF,0xAA,0x23,0xE9,0x58,0xCB,0x71,0x44,0xA1,0x19,0xFF,0xFA,1,0xE4,0xCE,0x62
};

static int parseWAVfmt(FILE *s,
	sndFormat *f,
	size_t *a,
	int64 z)
{
	uint8		b[32];

	if ((*a=fread(b,1,20,s))<16)
		return -3;
	if ((b[11]&0x80)|b[3]|b[7]|b[15] || !b[2])
		return -1; /* bad stored frame size, chan-count, rate or depth */
	/* fmt.w, chnCnt.w, rate.l, sfsize.l, balgn.w, depth.w */
	f->chanCount = b[2];
	f->rate = le32_native(((const int32*)b)[1]);
	/* more than one 'fmt ', bad frame count or rate */
	if (f->depth) // || (f->blockSize=le32_native(((const int32*)b)[2]))<0)
		return -1;
	f->blockSize = b[12]|b[13]<<8, f->depth = b[14];
	f->packSize = f->blockSize/f->chanCount;
	if ((b[1]<<8|*b)==0xFFFE)
	{ /* Extensible */
		if (z<16+22 || (b[16]|b[17]<<8)<22)
			return -1; /* bad size for extensible */
		z = b[18]|b[19]<<8; /* valid bits */
		if (z>0 && z<f->depth)
			f->depth = z; /* TODO: zero out non-valid bits in decoder */
		if (fread(b+16,1,4,s)!=4) /* channel listing */
			return -3;
		if (fread(b,1,16,s)!=16)
			return -3;
		*a+= 4+16;
		if (memcmp(b+2,extCommonGUID,14))
		{
			if (memcmp(b,extATRAC3GUID,16))
				return -8;
			*b = _be32_native(CODEC_ATRAC3)&255;
			b[1] = _be32_native(CODEC_ATRAC3)>>8&255;
		}
	} else if (z>19)
	{
		f->frameSize = b[18]|b[19]<<8;
		if (z==20 && f->frameSize==1 && (b[17]<<8|b[16]|*b<<16|b[1]<<24)==0x10002
				&& f->depth==32)
			*b = 3, f->type = STYPE_ISGN;
	}
	return sndFixCmpType(MK4CC('m','s',b[1],*b),f);
}

static void fixFrameCount(sndFormat *f,
	int64 dz)
{
	int64		t;

	int64addu32(dz,f->blockSize-1); /* round-up */
	int64divu32(dz,f->blockSize);
	if (!int64z(f->totalCount))
	{
		int64mulu32(dz,f->frameSize);
		t = f->totalCount; int64mulu32(t,f->mSampleSize*f->chanCount);
		if (int64gt(dz,t))
			dz = t;
		int64addu32(dz,f->frameSize-1); /* round-up */
		int64divu32(dz,f->frameSize);
	}
	f->frameCount = int64tou(dz);
}

long int parseFileRIFF(FILE *s,
	sndFormat *f,
	int is64b) /* has 64-bit extensions (not to be confused with W64) */
{ /* RIFF: WAVE (32-bit) */
	uint8		b[32];
	fpos_t		pos;
	int64		t, dz, z;
	long int	o =-1;
	size_t		a;
	int32		c;

	if (fread(b,8,1,s)!=1)
		return -3;
	if ((a=le32_native(*(const int32*)b))==0)
		return 0;
	t = a;
	if (((const int32*)b)[1]!=MK4CC('W','A','V','E'))
		return 0;
	z = dz = 0, f->totalCount = 0, a = 0;
	if (is64b)
	{ /* 64-bit - need signature; the first tag */
		if ((a=fread(b,1,32,s))<32)
			return -3;
		if (*(const int32*)b!=MK4CC('d','s','6','4')
				|| (c=le32_native(((const int32*)b)[1]))<16)
			return -1;
		z = c;
		int64sethl(t,le32_native(((const int32*)b)[3]),le32_native(((const int32*)b)[2]));
		int64sethl(dz,le32_native(((const int32*)b)[5]),le32_native(((const int32*)b)[4]));
		int64sethl(f->totalCount,le32_native(((const int32*)b)[7]),
			le32_native(((const int32*)b)[6]));
	}
	for(t+=4; t-=8,t-=z, t>0;)
	{
		(void)fgetpos(s,&pos);
	 #if WAV_ROUND_ODDCHUNK
		z+= z&1;
	 #endif
		pos.off+= z-a;
		(void)fsetpos(s,&pos);
		if ((a=fread(b,1,8,s))!=8)
		{
			if (a-1u<4u)
				goto end; /* some WAVEs have extra benign bytes */
			if (!int64z(dz) && o>0 && f->chanCount)
				break; /* RIFF size larger than file but got samples */
			return -3;
		}
		a = le32_native(((const int32*)b)[1]);
		z = a;
		a = 0;
		switch(*(const int32*)b)
		{
		case MK4CC('f','m','t',' '):
			if (z<16)
				return -1;
			if ((c=parseWAVfmt(s,f,&a,z))<0)
				return c;
			continue;
		case MK4CC('d','a','t','a'):
			if (/*!f->blockSize ||*/ o>0) /* no format or more than one 'data' */
				return -1;
			if (is64b)
				z = dz;
			else
				dz = z; /* size as a last resort (no 'fact' or 'ds64') */
			o = ftell(s);
			((streamwave_base*)f->variables)->initSize = dz;
			continue;
		case MK4CC('f','a','c','t'):
			if (!int64z(f->totalCount))
				return -1; /* multiple instances */
			if ((a=fread(b,1,4,s))<4)
				return -3;
			//int64sethl(f->totalCount,0,le32_native(*(const int32*)b));
			continue;
		case MK4CC('b','e','x','t'):
			/* and some WAVEs have bad tag lengths at the end */
			/* if we have enough info, skip tags to avoid potential EOF error */
			if (!int64z(dz) && o>0 && f->chanCount)
				goto end;
		default:
			; /* anything about alignment? */
		}
	}
	if (int64n(t)<0)
		return -3;
	/* clip data size to actual available data */
	//fseek(s,0,SEEK_END); -- fails for 64 bit!
	pos.off = _filelength64(_fileno(s));
	dz+= o;
	if (pos.off<dz)
		dz = pos.off;
	dz-= o;
end:	sndFixCodecFrame(f);
	if (int64z(dz) || f->frameSize<=0 || !f->blockSize || !f->mSampleSize || !f->chanCount)
		return -1;
	fixFrameCount(f,dz);
	return o;
}

long int parseFileW64(FILE *s,
	sndFormat *f)
{ /* riff: wave (64-bit) */
	int32		w[9];
	int64		t, dz, z;
	long		o =-1;
	size_t		a;
	int32		c;

	if (fread(w,4,9,s)!=9)
		return -3;
	int64sethl(t,le32_native(w[4]),le32_native(w[3]));
	if (w[5]!=MK4CC('w','a','v','e'))
		return 0;
	int64qset(z,0); int64qset(dz,0); int64qset(f->totalCount,0); a = 0;
	for(int64subu32(t,4*10-24); int64subu32(t,24),int64sub(t,z),int64p(t)>0; a+=24)
	{
		fpos_t		pos;

		(void)fgetpos(s,&pos);
	 #if WAV_ROUND_ODDCHUNK
		z+= z&1;
	 #endif
		pos.off+= z-a;
		if ((a=pos.off&7)!=0)
			a = 8-a, pos.off+= a, z-= a;
		(void)fsetpos(s,&pos);
		if ((a=fread(w,4,6,s))!=6)
		{
			if (a-1u<4u)
				goto end; /* some WAVEs have extra benign bytes */
			return -3;
		}
		int64sethl(z,le32_native(w[5]),le32_native(w[4]));
		a = 0;
		switch(*w)
		{
		case MK4CC('f','m','t',' '):
			if (!int64gehl(z,0,16))
				return -1;
			if ((c=parseWAVfmt(s,f,&a,z))<0)
				return c;
			continue;
		case MK4CC('d','a','t','a'):
			if (!f->blockSize || o>0) /* no format or more than one 'data' */
				return -1;
			dz = z;
			o = ftell(s);//+8;
			((streamwave_base*)f->variables)->initSize = dz;
			continue;
		case MK4CC('f','a','c','t'):
			if (!int64z(f->totalCount))
				return -1; /* multiple instances */
			if ((a=fread(w,1,8,s))<8)
				return -3;
			//int64sethl(f->totalCount,le32_native(w[1]),le32_native(*w));
		default:
			;
		}
	}
	if (int64n(t)<0)
		return -3;
end:	sndFixCodecFrame(f);
	if (int64z(dz) || f->frameSize<=0 || !f->blockSize || !f->mSampleSize || !f->chanCount)
		return -1;
	fixFrameCount(f,dz);
	return o;
}

/* Core Audio format file */
long int parseFilecaff(FILE *s,
	sndFormat *f)
{
	uint8		b[32];
	int64		dz, z, u;
	long int	o =-1;
	size_t		a;
	int32		c;

	if (fread(b,4,1,s)!=1)
		return -3;
	if (be16_native(*(const int16*)b)!=1)
		return 0;
	int64qset(z,0); int64qset(dz,0); int64qset(f->totalCount,0);
	((streamwave_base*)f->variables)->bstream = 0;
	for(a=0; ;)
	{
		fpos_t		pos;

		(void)fgetpos(s,&pos);
		pos.off+= z-a;
		(void)fsetpos(s,&pos);
		if ((a=fread(b,1,12,s))!=12)
		{
			if (!a)
				goto end;
			return -3;
		}
		int64sethl(z,be32_native(((const int32*)b)[1]),
			be32_native(((const int32*)b)[2]));
		a = 0;
		switch(*(const int32*)b)
		{
		case MK4CC('d','e','s','c'):
			if (!int64gehl(z,0,32))
				return -1;
			if ((a=fread(b,1,32,s))<32)
				return -3;
			int64sethl(u,be32_native(*(const int32*)b),
				be32_native(((const int32*)b)[1]));
			f->rate = *(double*)&u;
			f->rateFrac = (*(double*)&u-f->rate)*4294967296.;
			c = be32_native(((const int32*)b)[6]);
			if (c-1u>127-1)
				return -1;
			f->chanCount = (int8)c;
			/* more than one 'desc' */
			if (f->depth)
				return -1;
			c = be32_native(((const int32*)b)[7]);
			if (c>64u)
				return -1; /* bad depth */
			f->depth = (int8)c;
			f->blockSize = ((streamwave_base*)f->variables)->CBRFrameSize
				= be32_native(((const int32*)b)[4]); /* mBytesPerPacket */
			f->packSize = f->blockSize/f->chanCount;
			f->frameSize = be32_native(((const int32*)b)[5]); /* mFramesPerPacket */
			if ((c=sndFixCmpType(((const int32*)b)[2],f))<0)
				return c;
			if ((f->type&~STYPE_MAX16)==CTYPE_PCM)
			{
				if (b[15]&1)
					f->type|= STYPE_FP;
				if (b[15]&2)
					f->type|= STYPE_LSB;
			}
			continue;
		case MK4CC('d','a','t','a'):
			if (o>0) /* more than one 'data' */
				return -1;
			dz = z-4; /* skip edit count */
			o = ftell(s)+4;
			if (z<4)
				return -3;
			((streamwave_base*)f->variables)->initSize = dz;
			continue;
		case MK4CC('k','u','k','i'): /* MP4 edts atom equivalent */
			*((streamwave_base*)f->variables)->descData = ftell(s);
			continue;
		case MK4CC('p','a','k','t'):
			((streamwave_base*)f->variables)->bstream = ftell(s);
			continue;

		default:
			; /* anything about alignment? */
		}
	}
end:	sndFixCodecFrame(f);
	if (int64z(dz) || f->frameSize<=0 || !f->blockSize || !f->mSampleSize || !f->chanCount)
		return -1;
	fixFrameCount(f,dz);
	if (f->totalCount && ((streamwave_base*)f->variables)->bstream)
	{ /* IMA can call for a remaining count to remove */
		fseek(s,((streamwave_base*)f->variables)->bstream+20,SEEK_SET);
		if (fread(b,1,4,s)==4)
			f->frameCount = f->totalCount-= be32_native(*(int32*)b);
	}
	return o;
}

const static int32	auEnc[] =
{
	~0, /* reserved */
	CODEC_ULAW,
	CODEC_PCM_U8,
	CODEC_PCM_BE_S16,
	CODEC_PCM_BE_S24,
	CODEC_PCM_BE_S32,
	CODEC_PCM_BE_F32,
	CODEC_PCM_BE_F64,
	~0, /* fragment */
	~0, /* DSP program */
	CODEC_PCM_U8, /* fixed-point */
	CODEC_PCM_BE_S16, /* fixed-point */
	CODEC_PCM_BE_S24, /* fixed-point */
	CODEC_PCM_BE_S32, /* fixed-point */
	~0,
	~0,
	~0,
	~0,
	CODEC_PCM_BE_S16, /* emphasis */
	CODEC_PCM_BE_S16, /* DRC */
	CODEC_PCM_BE_S16 /* emphasis+DRC */
};

long int parseFileAU(FILE *s,
	sndFormat *f)
{
	uint8		b[20];
	int64		fend;
	long int	o;
	int		c;

	if (fread(b,20,1,s)!=1)
		return -3;
	if ((c=be32_native(((const int32*)b)[4]))-1u>127-1)
		return -1;
	f->chanCount = c;
	c = be32_native(((const int32*)b)[2]);
	if (c+0u>=GET_ARRAY_COUNT(auEnc) || (c=auEnc[c])==~0)
		return -8;
	if ((c=sndFixCmpType(c,f))<0)
		return c;
	if ((c=be32_native(((const int32*)b)[3]))<=0)
		return -1;
	f->rate = c;
	sndFixCodecFrame(f);
	o = (be32_native(*(const int32*)b)+3)&~3;
	c = be32_native(((const int32*)b)[1]);
	/* get actual size by seeking to offset+size or end if "undefined" */
	//(void)fgetpos(s,&pos);
	if ((uint32)c==0xFFFFFFFFu)
	{
		//fseek(s,0,SEEK_END);
		fend = _filelength64(_fileno(s));
	} else
	{
		fend = o, fend+= (uint32)c;
		//(void)fsetpos(s,&pos);
	}
	//(void)fgetpos(s,&pos);
	fixFrameCount(f,fend-o);
	f->totalCount = f->frameCount;
	f->totalCount*= f->frameSize, f->totalCount/= f->chanCount*f->mSampleSize;
	return o; /* 4 byte boundary */
}

const static char	vocSig[] = "tive Voice File\32\32";

const static int32	vocEnc[] =
{
	CODEC_PCM_U8,
	CODEC_SBPRO_ADPCM_4,
	CODEC_SBPRO_ADPCM_3,
	CODEC_SBPRO_ADPCM_2,
	CODEC_PCM_LE_S16,
	~0,
	CODEC_ALAW,
	CODEC_ULAW,
	CODEC_ADPCM_CT
};

/* NOTE: A Creative Voice file is a stream of blocks;
	the file can have more than one sample rate and silence blocks.
	 This implementation enables continuous streaming,
	but only the first block of samples is handled from here.
*/
int parseVOCFileBlocks(FILE *s,
	sndFormat *f);
int parseVOCFileBlocks(FILE *s,
	sndFormat *f)
{
	double		r, extrate = 0;
	int32		z, c;

	while((c=fgetc(s))>=0)
	{
		z = fgetc(s), z|= fgetc(s)<<8, z|= fgetc(s)<<16;
		switch(c) /* parse block */
		{
		case 0: break; /* separator - won't always have size */
		case 1: /* sound data */
			c = fgetc(s);
			if (c<0)
				return -3;
			if (!f->rate)
			{
				r = 1e6/(256-c);
				if (extrate)
					r = extrate;
				f->rateFrac = (r-(f->rate=(int)r))*4294967296.;
			}
			c = fgetc(s);
			f->chanCount = 1;
			z-= 2;
			goto setenc;
		case 8: /* extended format */
			c = fgetc(s), c|= fgetc(s)<<8;
			fgetc(s);
			extrate = 256e6/(65536-c);
			if ((c=fgetc(s))+0u>126) /* channel count-1 */
				return c<0 ? -3:-1;
			f->chanCount = c+1;
			break;
		case 9: /* extended data */
			if (!f->rate)
			{
				c = fgetc(s), c|= fgetc(s)<<8, c|= fgetc(s)<<16, c|= fgetc(s)<<24;
				extrate = c;
				f->depth = fgetc(s);
				if ((c=fgetc(s))+1u>126) /* channel count */
					return c<0 ? -3:-1;
				f->chanCount = c;
				if (extrate<=0 || f->depth>32)
					return -1;
				f->rate = extrate; /* not sure where to put this */
			} else
				fseek(s,6,SEEK_CUR);
			c = fgetc(s), c|= fgetc(s)<<8;
			if (fseek(s,4,SEEK_CUR))
				return -3;
			z-= 12;
		setenc:	if (c>=GET_ARRAY_COUNT(vocEnc)-1)
				if (c==0x200)
					c = 8; /* CT position */
				else
					return c<0 ? -3:-8;
			if ((c=sndFixCmpType(vocEnc[c],f))<0)
				return c;
			sndFixCodecFrame(f);
			if (z<1)
				return -1;
			fixFrameCount(f,z);
			f->maxRead = 256;
		case 2:	return 0; /* continue data */
		case 3: /* silence */
			z = fgetc(s), z|= fgetc(s)<<8, ++z;
			if ((c=fgetc(s))<0) /* sample rate for period */
				return -3;
			r = 1e6/(256-c);
			if (f->rate) /* adjust period to set rate */
				z = z*(f->rate+f->rateFrac*(1/4294967296.))/r;
			else
				f->rateFrac = (r-(f->rate=(int)r))*4294967296.;
			f->type = CTYPE_TONE;
			f->depth = 16;
			f->frameCount = z;
			memset(f->variables,0,sizeof(f->variables));
			return 0;
		case 4: /* marker */
		case 5: /* string */
		case 6: /* loop count */
		case 7: /* loop end */
			if (fseek(s,z,SEEK_CUR))
				return -3;
			break;
		//case 5: /* string */
		//	while(fgetc(s)>=0)
		//		;
		//	break;

		default: return -1; /* unknown block type */
		}
	}
	return -1;
}

long int parseFileVOC(FILE *s,
	sndFormat *f)
{
	uint8		b[22];
	int		c;

	if (fread(b,22,1,s)!=1)
		return -3;
	if (memcmp(b,vocSig,18))
		return 0;
	c = b[18]|b[19]<<8; /* version */
	if ((b[20]|b[21]<<8)^(~c+0x1234)) /* verify */
		return 0;

	if ((c=parseVOCFileBlocks(s,f))<0)
		return c;
	return ftell(s);
}

/* Scripted Resource Format; this would be in resex but it is proprietary */
long int parseFileSrf(FILE *s,
	sndFormat *f)
{
	uint8		b[20];
	int32		z, hz;

	if (fread(b,8,1,s)!=1)
		return -3;
	z = be32_native(*(int32*)b); /* offset to trailer */
	if (z<12)
		return -1;
	hz = be32_native(((int32*)b)[1]); /* init header size */
	if (hz<12)
		return 0;
	//while(hz>0)
	{
		if (fread(b,20,1,s)!=1) /* tag.l, type.l, kind.l, offs.l, size.l */
			return -3;
		if (be32_native(((int32*)b)[1])<4
				&& be32_native(((int32*)b)[2])<128)
		{ /* first sound */
			memset(f,0,sizeof(*f));
			f->streamType = MK4CC('s','n','d',' ');
			((streamwave_base*)f->variables)->initSize = be32_native(((int32*)b)[4]);
			return be32_native(((int32*)b)[3]);
		}
		//hz-= 12;
	}
	return 0;
}

/* minimal 8-bit header */
long int parseFileMH8B(FILE *s,
	sndFormat *f,
	int32 b)
{
	int32		z, c;

	if (fread(&z,4,1,s)!=1)
		return -3;
	z = le32_native(z);
	if (z<0)
		return -1;
	f->chanCount = 1;
	f->rate = (uint16)(b>>16);
	if ((c=sndFixCmpType(CODEC_PCM_U8,f))<0)
		return c;
	sndFixCodecFrame(f);
	f->totalCount = f->frameCount = z;

	return 8;
}

#ifdef SNDDEC_MPEG_DECODE
long int parseFileMPEG(FILE *s,
	sndFormat *f,
	int strict);
long int parseFileMPEG(FILE *s,
	sndFormat *f,
	int strict)
{
	struct mpg123_frameinfo	mi;

	fpos_t		pos;
	int8		b[10];
	mpg123_handle	*mpgh;
	long		r;
	int		refnum, chanc, enc;

	if (!s)
		return -1;
	if (strict>0)
	{
		(void)fgetpos(s,&pos);
		/* identify and skip tag, verify first frame */
		if (fread(b,1,10,s)!=10)
			return -3;
		if ((*(int32*)b&MK4CC(255,255,255,0))==MK4CC('I','D','3',0))
		{
			(void)fseek(s,(((((b[6]<<7)|b[7])<<7)|b[8])<<7)|b[9],SEEK_CUR);
			if (fread(b,1,4,s)!=4)
				return -3;
		}
		(void)fsetpos(s,&pos);
		r = be32_native(*(int32*)b);
		if ((r&0xFFE00000)!=0xFFE00000 // signature
				|| !(r&0x00060000) // layers 1-3 only
				|| (r&0xF000)==0xF000 // valid bitrate
				|| (r&0xC00)==0xC00) // valid frequency
			return 0;
		if (strict>1 && !(r&0xF000))
			return 0; /* level 2: reject free format */
	}
	/* safe on original calling thread and re-entrant */
	if ((r=mpg123_init())!=MPG123_OK)
		return 0;
	/* auto format detection */
	if ((mpgh=mpg123_new(NULL,NULL))==NULL)
		return 0;
	if (strict<0)
		mpg123_param(mpgh,MPG123_ADD_FLAGS,MPG123_QUIET,0);
	refnum = _fileno(s);
	if (mpg123_open_fd(mpgh,refnum)!=MPG123_OK)
		goto close;

	r = 0, chanc = enc = 0;
	mpg123_getformat(mpgh,&r,&chanc,&enc);
	if (r<8000 || chanc-1u>1u)
		goto close; /* silently go out into the night... */

	memset(f,0,sizeof(*f));
	if (chanc==3) /* 'channels' may be represented as mono/stereo flag bits */
		chanc = 2;
	f->chanCount = chanc;
	f->mSampleSize = mpg123_encsize(enc);
	if (!f->mSampleSize)
	{
	close:	mpg123_delete(mpgh);
		return 0;
	}
	f->codec = enc&MPG123_ENC_FLOAT_64 ? (BIG_ENDIAN ? CODEC_PCM_BE_F64:CODEC_PCM_LE_F64)
		:enc&MPG123_ENC_FLOAT_32 ? (BIG_ENDIAN ? CODEC_PCM_BE_F32:CODEC_PCM_LE_F32)
		:f->mSampleSize>2 ? (BIG_ENDIAN ? CODEC_PCM_BE_S32:CODEC_PCM_LE_S32)
		:f->mSampleSize>1 ? enc&MPG123_ENC_SIGNED
		? (BIG_ENDIAN ? CODEC_PCM_BE_S16:CODEC_PCM_LE_S16)
		:(BIG_ENDIAN ? CODEC_PCM_BE_S16:CODEC_PCM_LE_S16)
		:enc==MPG123_ENC_ULAW_8 ? CODEC_ULAW
		:enc==MPG123_ENC_ALAW_8 ? CODEC_ALAW
		:enc&MPG123_ENC_SIGNED ? CODEC_PCM:CODEC_PCM_U8;
	f->rate = r;
	f->depth = f->mSampleSize<<3;
	/* no immediate codec values for u24/u32, so they're handled here */
	if (f->depth==32 && !(enc&(MPG123_ENC_FLOAT|MPG123_ENC_SIGNED)))
		f->type|= STYPE_ISGN;
	f->streamType = CODEC_MPEG_L3P;
	f->maxRead = 1; //MPG_MAX_SAMPLES*2;
	f->frameSize = f->blockSize = f->mSampleSize*chanc;
	f->frameCount = mpg123_length(mpgh);
	r = mpg123_framepos(mpgh);
	mpg123_info(mpgh,&mi);
	f->streamBitrate = (mi.abr_rate ? mi.abr_rate:mi.bitrate)*8000;
	((streamwave_base*)f->variables)->initOff = r;
	mpg123_delete(mpgh);
	return r+!r;
}
#endif

#ifdef SNDDEC_OGG_DECODE
long int parseFileOGG(FILE *s,
	sndFormat *f,
	int strict)
{
	OggVorbis_File	vf;
	vorbis_info	*vi;
	double		ti;
	int		ret = 0;

	((streamwave_base*)f->variables)->initOff = ftell(s);
	if (f->codec==CODEC_OPUS
			|| (ret=ov_open_callbacks(s,&vf,NULL,0,OV_CALLBACKS_NOCLOSE))<0)
	{
	#ifdef SNDDEC_OPUS_DECODE
		OggOpusFile	*of;

		if (ret==OV_EREAD)
			return -3;
		if (f->codec==CODEC_VORBIS)
			return 0;
		(void)fseek(s,((streamwave_base*)f->variables)->initOff,SEEK_SET);
		of = op_open_callbacks(s,_op_get_file_callbacks(s,0),NULL,0,&ret);
		if (of)
		{
			f->streamType = CODEC_OPUS;
			f->streamBitrate = op_bitrate(of,-1)*8;
			f->rate = 48000;
			f->chanCount = op_channel_count(of,-1);
			f->frameCount = f->totalCount = op_pcm_total(of,-1);
			op_free(of);
			goto oggp16;
		}
		return 0;
	#endif
	}

	vi = ov_info(&vf,-1);
	if (!vi || (f->rate=vi->rate)<=0 || vi->channels-1u>126)
		return ret==OV_EREAD ? -3:-1;
	f->chanCount = vi->channels;
	f->streamType = CODEC_VORBIS;
	ti = ov_time_total(&vf,-1);
	if (ti<0)
		return -1;
	if (strict>0)
		f->streamBitrate = ov_bitrate(&vf,-1)*8;
	else
		f->streamBitrate = vi->bitrate_nominal*8;
	f->frameCount = f->totalCount = ti*f->rate+.125;
	ov_clear(&vf);
#ifdef SNDDEC_OPUS_DECODE
oggp16:
#endif
	f->codec = BIG_ENDIAN ? CODEC_PCM_BE_S16:CODEC_PCM_LE_S16;
	f->type = CTYPE_PCM;
	f->mSampleSize = 2;
	f->depth = 16;
	f->maxRead = 1;
	f->frameSize = f->blockSize = f->chanCount*sizeof(int16);
	return 1;
}
#endif

#ifdef SNDDEC_MP4_DECODE
static void setupCommonAAC(sndFormat *f,
	mp4stream *stream)
{
	f->codec = BIG_ENDIAN ? CODEC_PCM_BE_S16:CODEC_PCM_LE_S16;
	f->type = CTYPE_PCM;
	f->mSampleSize = 2;
	f->maxRead = 1;
	f->streamType = CODEC_AAC;
	if ((f->streamBitrate=stream->config.bitrateavg)==0)
	{ /* calculate the average bitrate using data size against maximum */
		double		t;

		if ((t=stream->config.samples)!=0)
		{
			t = (double)stream->config.datasize*f->rate*8/t;
			if (t<=stream->config.bitratemax)
				f->streamBitrate = t+.125;
		}
	}
	f->streamBitrate*= 8;
}

long int parseFileMP4(FILE *s,
	sndFormat *f,
	int strict)
{
	mp4stream	*stream;
	long int	o;
(void)strict;
	stream = calloc(1,sizeof(mp4stream));
	((streamwave_base*)f->variables)->initOff = ftell(s);
	stream->config.verbose.errors = strict>=0;
	stream->config.verbose.header = strict==8;
	o = _mp4read_open(s,&stream->config,0);
	if (o<0)
	{
		if (o==-8)
		{
			uint16	type = stream->config.format;

			switch(type)
			{
			case 0x4305:
				f->codec = CODEC_PCM_BE_S16; goto fix;
			case 0x5034:
				f->codec = CODEC_PCM_LE_S16; goto fix;
			case 0xA1A3:
				f->codec = CODEC_ALAW; goto fix;
			case 0xB1A3:
				f->codec = CODEC_ULAW; goto fix;
			case 0xA1AC:
				f->codec = CODEC_ALAC; goto fix;
			case 0x3AC3:
				f->codec = CODEC_MACE_3; goto fix;
			case 0x3AC6:
				f->codec = CODEC_MACE_6; goto fix;
			case 0x5124:
				f->codec = CODEC_PCM_BE_S24; goto fix;
			case 0x5132:
				f->codec = CODEC_PCM_BE_S32; goto fix;
			case 0x1532:
				f->codec = CODEC_PCM_LE_S32; goto fix;
			case 0xF132:
				f->codec = CODEC_PCM_BE_F32; goto fix;
			case 0xF164:
				f->codec = CODEC_PCM_BE_F64; goto fix;
			case 0x408B:
				f->codec = CODEC_VORBIS; goto fix;
			case 0x0B25:
				f->codec = CODEC_OPUS; goto fix;

			case 0x13A4:
				f->codec = CODEC_AIFC_IMA;
			fix:	f->rate = stream->config.samplerate;
				f->rateFrac = stream->config.sampleratefrac;
				f->chanCount = stream->config.channels;
				f->depth = stream->config.bits;
				sndFixCmpType(f->codec,f);
				sndFixCodecFrame(f);
				f->totalCount = stream->config.samples;
				f->frameCount = f->totalCount/f->blockSize;
				o = stream->config.datasize/f->blockSize;
				if (o && o+0u<f->frameCount)
					f->frameCount = o;
				o = stream->config.dataofs;
			}
			if (o<0 && strict>=0)
				fprintf(stderr,"### MP4 file uses codec 0x%X\n",
					type);
		}
		free(stream);
		return o;
	}
	((streamwave_base*)f->variables)->stream = stream;
	//stream->in = s;
	f->rate = stream->config.samplerate;
	f->chanCount = stream->config.channels;
	f->depth = stream->config.bits;
	f->frameSize = f->blockSize = sizeof(int16)*f->chanCount;
	f->frameCount = f->totalCount = stream->config.samples;
	setupCommonAAC(f,stream);
	mp4read_close(NULL,&stream->config);
	return 1;
}

 #ifdef NEW_AND_BROKEN
static int unpakt(FILE *s,
	uint32 c,
	uint32 z,
	uint32 **data)
{
	uint32		*d, ofs = 0;
	int		r;

	*data = d = malloc((c+1)*sizeof(*data));
	if (!d)
		return -2;
	do
	{
		r = fgetc(s);
		if (r&0x80)
			r = (r<<8&0x7F00)|fgetc(s);
		if (r==EOF)
			return -3;
		*d++ = ofs;
		ofs+= r*z;
	} while(--c);
	*d = ofs;
	return 0;
}

static int setupCaffAAC(FILE *s,
	sndFormat *f,
	long int o)
{
	int8		b[24];
	mp4stream	*stream;
	int32		descOfs = *((streamwave_base*)f->variables)->descData;
	int32		r, pkt = ((streamwave_base*)f->variables)->bstream;

	if (descOfs<8 || pkt<8)
		return -1;
	fseek(s,pkt,SEEK_SET);
	if (fread(&b,1,sizeof(b),s)!=sizeof(b))
		return -3;
	if (*(int32*)b)
		return -1; /* packet count must fit into an uint32 */
	stream = calloc(1,sizeof(mp4stream));
	((streamwave_base*)f->variables)->stream = stream;
	stream->config.frame.maxsize = 1024; /* default */
	stream->config.frame.ents = be32_native(*(int32*)(b+4));
	if (stream->config.frame.ents && !((streamwave_base*)f->variables)->CBRFrameSize) /* VBR: unpack */
		if ((r=unpakt(s,stream->config.frame.ents,((streamwave_base*)f->variables)->packetsPerFrame
				&stream->config.frame.data))<0)
			goto fail;
	int64sethl(f->totalCount,be32_native(*(int32*)(b+8)),be32_native(*(int32*)(b+12)));
	fseek(s,descOfs-4,SEEK_SET);
	r = 0;
	fread(&r,1,sizeof(r),s);
	r = _mp4read_esds_atom(s,be32_native(r),&stream->config);
	fseek(s,o,SEEK_SET);
	if (r<0)
	{
	fail:	if (stream->config.frame.data)
			free(stream->config.frame.data);
		free(stream); ((streamwave_base*)f->variables)->stream = NULL;
		return r;
	}
	if (!stream->config.frame.ents)
	{ /* CBR only */
		if (!stream->config.frame.maxsize)
			stream->config.frame.maxsize = 1024;
		stream->config.frame.ents = f->totalCount/stream->config.frame.maxsize;
	}
	//f->totalCount-= be32_native(*(int32*)(b+16)); /* primer */
	f->totalCount-= be32_native(*(int32*)(b+20)); /* remaining */
	f->frameCount = f->totalCount;
	stream->config.samples = (uint32)MIN(f->frameCount,0xFFFFFFFF);
	stream->config.dataofs = o;
	stream->config.datasize = ((streamwave_base*)f->variables)->initSize;
	setupCommonAAC(f,stream);
	f->frameSize = f->blockSize = sizeof(int16)*f->chanCount;
	return o;
}
 #endif
#endif

long int readAudioHeader(FILE *s,
	sndFormat *f,
	int strict)
{
	int32		b;
	long int	o;

	if (!s || !f)
		return -1;
	if (fread(&b,sizeof(b),1,s)!=1)
		return -3;
	memset(f,0,sizeof(*f));
	o = 0;
	if (b==MK4CC('F','O','R','M'))
		o = parseFileFORM(s,f);
	else if (b==MK4CC('R','I','F','F')
			|| b==MK4CC('R','F','6','4'))
		o = parseFileRIFF(s,f,b^MK4CC('R','I','F','F'));
	else if (b==MK4CC('r','i','f','f'))
		o = parseFileW64(s,f);
	else if (b==MK4CC('c','a','f','f'))
		o = parseFilecaff(s,f);
	else if (b==MK4CC('.','s','n','d'))
		o = parseFileAU(s,f);
	else if (b==MK4CC('C','r','e','a'))
		o = parseFileVOC(s,f);
	else if (b==MK4CC('s','r','f','1'))
		o = parseFileSrf(s,f);
	else if (b==MK4CC(3,0,17,43) || b==MK4CC(3,0,22,56))
		o = parseFileMH8B(s,f,le32_native(b));
#ifdef SNDDEC_OGG_DECODE
	else if (b==MK4CC('O','g','g','S'))
		goto ogg;
#endif
#ifdef SNDDEC_MP4_DECODE
	else if (!(b&MK4CC(255,0,0,0)))
	{
		b = fgetc(s);
		(void)fseek(s,0,SEEK_SET);
		if (b) /* RF or other formats will follow a null, MOV/MP4 never will */
			o = parseFileMP4(s,f,strict);
	}
#endif
#ifdef SNDDEC_MPEG_DECODE
	else
		goto trympg;
	if (!o)
		goto trympg;
	/* because MPEG is still not handled as an internal codec */
	switch(f->codec)
	{
	#if 0//def SNDDEC_MP4_DECODE
	case CODEC_AAC:
		o = setupCaffAAC(s,f,o);
		break;
	#endif
	case CODEC_MPEG_AUDIO:
	case CODEC_MPEG_L1P:
	case CODEC_MPEG_L2P:
	case CODEC_MPEG_L3P:
	case CODEC_MPEG_L3:
	trympg:	(void)fseek(s,o,SEEK_SET);
		o = parseFileMPEG(s,f,strict);
	}
#endif
#ifdef SNDDEC_OGG_DECODE
	if (!o || f->codec==CODEC_VORBIS || f->codec==CODEC_OPUS
			|| f->codec==MK4CC('m','s',103,79))
	{
	ogg:	(void)fseek(s,o,SEEK_SET);
		o = parseFileOGG(s,f,strict);
	}
#endif
	if (o<=0)
		return o;
	if ((!f->chanCount || !f->depth || !f->rate || !o)
			&& f->streamType!=MK4CC('s','n','d',' '))
		return -1;
	return o;
}
