/*
 sndmgr.c
  sound channel player, command processor

  Created: 2014.03.21 02:09 EDT
  Revised: 2024.04.15
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.21
  separated from smixer.
2014.03.22
  finished working prototype; gaps (mixer); not everything's been tested.
2014.03.23
  success!!  multi-threaded OR single (manual sndMgrProcess calls).
2014.03.24
  seamless interpolation, end-sample silence (instead of zeroes).
2014.04.04
  pause now repeats current sample at ->pos (instead of zeroes).
deactivation can still be performed by setting bit #1 of actFlags.
2014.08.09
  sndFilePlay callback now uses double-buffering, as it should have.
2014.08.14
  sndPlayFileStop clears userInfo first (as it's supposed to).
not doing so promotes crashes, hangs.
2014.12.05
  using the delayed buffering approach (2ms timer) used in mpegdecmac.c
to stop all the sndFilePlay callback crashing.
2015.04.07
  out-of-range fraction for sndRateMultiplierCmd does not set.
2015.05.30
  mutex lock only on command enQueue/deQueue. (may fix asynch CBs.)
2019.02.15
  loops are cut off at non-wait queue end, restarted on a wait/duration.
2021.07.27
  reduce pulse wave tones slightly; square emulation has been too loud.
  (high pass still not implemented and probably never needed.)
2021.08.02
  added defaultCallBack to reduce porting stress.
2022.01.08
  streamEntry instead of userInfo (which should be for the *user*).
*/

//#define DEBUG_QUEUE 1

#include <stdio.h>

#include <enviro64.h>
#include "ftrig.h"
#include "snddec.h"

#if defined(WIN32) || defined(WIN64)
 #include <windows.h>
 #include <mmsystem.h>
 #include <process.h>	//threads

 #define CloseObject		CloseHandle
 #define WaitForSemaphore	WaitForSingleObject
 /* these call for win_ce_semaphore.c ... CleanUp has bugs with its mask conditionals */
 #ifdef _WIN32_WCE
  #if _WIN32_WCE<300
   #define CreateSemaphore	CreateSemaphoreCE
   #define ReleaseSemaphore	ReleaseSemaphoreCE
   #undef CloseObject
   #undef WaitForSemaphore
   #define CloseObject		CloseSynchHandle
   #define WaitForSemaphore	WaitForSemaphoreCE
  #endif
 #endif
 #define MgrMutexLock(m)	EnterCriticalSection(&(m)->criticalSection)//; printf("lock %d\n",__LINE__)
 #define MgrMutexUnlock(m)	LeaveCriticalSection(&(m)->criticalSection)//; printf("unlock %d\n",__LINE__)
#endif

/* ---- constants */

#define TERMINATE_LOOP_IF_ZERO_DELAY

/* the thread loop is so lean it ends up most often calling Sleep */
/* the Semphore method can also break up easily */
#define NO_SEMAPHORE
#define QUEUED_WAVE
//#define WAVE_DUMP

#ifdef WAVE_DUMP
 #include <io.h>
 #include <fcntl.h>
static FILE		*fwout;
static int		fwoutn;
static long int		_lastFmtSize;
static sndFormat	format;
 #define lastFmtSize		&_lastFmtSize
#else
 #define lastFmtSize		NULL
#endif

/*
 The Apple Sound Chip and EASC have a 1024-stereo sample buffer,
but has also a 'half-full' indicator - 512 samples, that is, the ideal.
 However, the Sound Manager buffers 4096 samples in loading; this weakens
loading latency, but makes loading stereo DVI_IMA possible in one frame
(2040*2).  But that's not the potential max. here.  64-bit samples can
trump that; maxRead(512)*8 makes 4096.  4096*stereo 16-bit makes 16384.
*/
/* 2014.07.27: 4096 doesn't seem to work w. stereo 8-bit (out-of-bounds dcmp writes) */
#define chBufCount		(MAX(sndEndFrameCount,8192)+1)
#if defined(__MT__) && defined(SNDMGR_PRECISION)
 #define sndEndFrameCount	1024	/* can afford smaller at above-normal... */
#else
 #define sndEndFrameCount	2048
#endif
#define sndBufferCount		4	/* Semaphore method with small buffer needs 3+ */
#define defaultVolume		0x1000100
/* not meant to handle rates above 96000 */
#define defaultOutRate		48000	/* was 44100 */

/* ---- manager structure */

typedef struct sndMgrData
{
	sndChannel	*sndChannels; /* a 'next' chain, until NULL */
	uint32		rflag;
 #ifdef WINUSERAPI
  #ifdef __MT__
	HANDLE		thread;
   #ifndef NO_SEMAPHORE
	HANDLE		hSemaphore;
   #endif
  #endif
	HWAVEOUT	hWAudio;
	WAVEHDR		hWHeader[sndBufferCount];

	CRITICAL_SECTION	criticalSection;
 #endif
	int32		out[sndBufferCount][sndEndFrameCount*2];
 #ifdef __MT__
	UINT		ticEID;
	uint16		tcount, tmaximum;
	sndTaskEntry	tlist[1];
 #endif
} sndMgrData;

/* ---- globals */

static int8		toneTables[12][8] =
{
	{  63,  63,  63,  63, -63, -63, -63, -63},
	{ -63,  63,  63,  63,  63,  63,  63,  63},
	{ -63, -63,  63,  63,  63,  63,  63,  63},
	{ -63, -63, -63,  63,  63,  63,  63,  63},
	{ -63, -63, -63, -63,  63,  63,  63,  63},
	{ -63, -63, -63, -63, -63,  63,  63,  63},
	{ -63, -63, -63, -63, -63, -63,  63,  63},
	{ -63, -63, -63, -63, -63, -63, -63,  63},
	{   0,  63, 127,  63,   0, -63,-127, -63},
	{  12,  38,  64,  90, -90, -64, -38, -12},
	{   0, 109,  63, -18,   0,  18, -63,-109},
	{ -12,  41,  86, 117, 127, 117,  86,  41}
};

/* ---- code */

static void sndSetAmp(sndFormat *f, /* expects format type to be valid */
	int amp)
{
	genwave_base	*b = (genwave_base*)&f->variables;
	const int8	*t;

	if (amp>255)
		amp = 255;
	b->amp = (uint8)amp;
	if ((f->type&15)!=CTYPE_TONE || b->type>sndToneHalfSine)
		return;
	t = toneTables[b->type];
	b->perm[0] = (int16)(t[0]*amp);
	b->perm[1] = (int16)(t[1]*amp);
	b->perm[2] = (int16)(t[2]*amp);
	b->perm[3] = (int16)(t[3]*amp);
	b->perm[4] = (int16)(t[4]*amp);
	b->perm[5] = (int16)(t[5]*amp);
	b->perm[6] = (int16)(t[6]*amp);
	b->perm[7] = (int16)(t[7]*amp);
}

static inline void updateChMixerAmp(sndChannel *ch)
{
	int32		v = ch->LRVolume;
	uint8		flag = (uint8)(((sndMgrData*)ch->private)->rflag>>24&ch->initFlag);

	if (!(flag&sndEnableLeft))
		v&= 0xFFFF;
	if (!(flag&sndEnableRight))
		v&=~0xFFFF;
	ch->sndFrame.LRAmp = v;
}

/* commands */

#if DEBUG_QUEUE
const char	*sndMgrCmdName[] =
{
	"null",
	"alloc",
	"free",
	"quiet",
	"flush",
	"reInit",
	"format",
	"raw",
	"sound",
	"buffer",
	"wait",
	"pause",
	"resume",
	"callBack",
	"syncCmd",
	"stopCmd",
	"LRVolume",
	"RLVolume",
	"rateCmd",
	"rateMul",
	"freqDur",
	"restCmd",
	"freqCmd",
	"amp",
	"tone",
	"noise",
	"mWave",
	"phase",
	"outRate",
	"voice",
	"loop",
	"filter"
};
#endif

static int enQueue(sndChannel *ch,
	const sndCommand *cmd)
{
	int		r = 0;
	int16		x;

	if (!cmd)
		return -1;
#if DEBUG_QUEUE
	{
		uint64		mcs;

		Microseconds(&mcs);
		printf("en[% 3d/% 3d] % 10s % 3d @ %f\n",
			ch->qHead,ch->qTail,
			cmd->cmd<GET_ARRAY_COUNT(sndMgrCmdName) ? sndMgrCmdName[cmd->cmd]:"?",
			cmd->cmd,
			mcs/1000000.);
	}
#endif
	if ((x=ch->qLength)<=0)
		{r =-1; goto end;}
	if (ch->qTail+0u>=x)
		ch->qTail = 0;
	if (ch->qHead>0)
		x = ch->qHead;
	if (ch->qTail==--x)
		r =-203; /* full */
	else /* post-incr. from tail */
		ch->queue[x=ch->qTail] = *cmd, ch->qTail = ++x>=ch->qLength ? 0:x;
end:	return r;
}

static int deQueue(sndChannel *ch,
	sndCommand *cmd) /* must be valid */
{
	int		z, x, r = 0;

	if ((z=ch->qLength)<=0)
		{r =-1; goto end;}
	/* post increment from head */
	/* first, correct head/tail if out of bounds */
	if ((x=ch->qHead)+0u>=z)
		ch->qHead = 0;
	if (ch->qTail+0u>=z)
		ch->qTail = 0;
	if (x!=ch->qTail)
	{
		*cmd = ch->queue[x], ch->qHead = (int16)(++x>=z ? 0:x), r = 1;
#if DEBUG_QUEUE
		uint64		mcs;

		Microseconds(&mcs);
		printf("de[% 3d/% 3d] % 10s % 3d @ %f\n",
			ch->qHead,ch->qTail,
			cmd->cmd<GET_ARRAY_COUNT(sndMgrCmdName) ? sndMgrCmdName[cmd->cmd]:"?",
			cmd->cmd,
			mcs/1000000.);
#endif
	}
end:	return r;
}

static int processCommands(sndChannel *ch,
	sndCommand *cmd,
	int c)
{
	int32		l[2];
	int32		r = 0;

	do
	{
		if (c && (r=deQueue(ch,cmd))<=0)
			goto end;
		switch(cmd->cmd)
		{
		case sndAllocCmd:
			if (!(ch->actFlags&chBusy) && !ch->sndFrame.dcmpBuf)
				ch->sndFrame.dcmpBuf = calloc(sizeof(int16)*2,chBufCount);
			break;
		case sndFreeCmd:
			if (!(ch->actFlags&chBusy) && ch->sndFrame.dcmpBuf)
				{free(ch->sndFrame.dcmpBuf); ch->sndFrame.dcmpBuf = NULL;}
			break;
		case sndQuietCmd:
			ch->sndFrame.pos = ch->sndFrame.dcmpSize = 0;
			ch->sndFrame.format.frameCount = 0;
			ch->sndFrame.cmpPos = NULL;
			ch->sndFrame.mixerRtn = NULL;
			ch->actFlags&=~chCompletion;
		case sndNullCmd:
			break;
		case sndFlushCmd:
			ch->qHead = ch->qTail = 0;
			r = 0; /* nothing to process now */
			goto end;
		case sndReInitCmd:
			ch->initFlag = (uint8)cmd->param2;
			if (ch->initFlag&sndEnableRight) /* initial bit */
				ch->initFlag|= sndInitStereo; /* mono-left or right */
			ch->initFlag^= sndInitPanMask;
			ch->sndFrame.filterR = ch->initFlag&sndInitExtendedFilter ? 0x3FFFF:0x1FFFF;
			goto upmix;
		case sndFormatCmd:
			if (!cmd->param2 || ch->voiceMode<sndVoiceDefault)
				goto inv;
			ch->durStart = ch->sndFrame.cmpPos = NULL;
			memcpy(&ch->sndFrame.format,cmd->param2,sizeof(ch->sndFrame.format));
			ch->sndFrame.format.frameCount = 0;
			if (sndFixCmpType(ch->sndFrame.format.codec,&ch->sndFrame.format)<0
					|| ch->sndFrame.format.type&8)
			{
				ch->sndFrame.format.codec = 0;
				goto badf;
			}
			ch->sndFrame.format.maxRead =
				(ch->sndFrame.format.type&15)>CTYPE_NPCM ? 1:512;
			if (!ch->sndFrame.format.blockSize)
				sndFixCodecFrame(&ch->sndFrame.format);
			ch->sndFrame.mixerRtn = NULL;
			r = 0;
			break;
		case sndLoopCmd:
			if (!ch->durStart || !ch->sndFrame.format.blockSize)
				break;
			*l = (int32)cmd->param1;
			l[1] = (int32)cmd->param2;
			if ((*l|l[1])<0 || *l+16>=l[1])
				*l = l[1] = 0;
			r = ch->durEnd/ch->sndFrame.format.blockSize;
			if (l[1]>r)
				l[1] = r;
			ch->loopStart = ch->durStart+*l*ch->sndFrame.format.blockSize;
			ch->loopCount = (l[1]-*l)*ch->sndFrame.format.frameSize;
			ch->durEnd = l[1]*ch->sndFrame.format.frameSize;
			ch->restEnd = ch->sndFrame.format.frameCount-ch->durEnd;
			r = 0;
			break;
		case sndRawCmd:
		case sndVoiceCmd:
			if (!cmd->param1)
				break;
			if (!cmd->param2)
				goto inv;
			if (ch->voiceMode<sndVoiceDefault)
			{
				if (ch->voiceMode==sndVoiceDisabled)
					break;
				/* locked */
				if ((ch->sndFrame.format.type&15)>=CTYPE_WTAB)
					goto badf;
			}
			if (!ch->sndFrame.format.rate || !ch->sndFrame.format.codec)
				goto badf;
			ch->durStart = cmd->param2;
			ch->durEnd = cmd->param1;
			ch->loopStart = NULL;
			ch->loopCount = ch->restEnd = 0;
			if (cmd->cmd==sndVoiceCmd)
				break;
			ch->sndFrame.format.frameCount = ch->durEnd;
			ch->sndFrame.cmpPos = ch->durStart;
			ch->actFlags|= chCompletion;
			r = 1;
			goto end;
		case sndBufferCmd: /* bufferCmd plays to completion but also resets rate */
		case sndSoundCmd:
			if (ch->voiceMode<sndVoiceDefault)
			{
				if (ch->voiceMode==sndVoiceDisabled)
					break;
				/* locked */
				if ((ch->sndFrame.format.type&15)>=CTYPE_WTAB)
				{
				badf:	r =-223;
					goto end;
				}
				r = ch->actFlags&chDurAux ? 0:-205; /* no error code if playing */
				goto end;
			}
			if ((r=parseBufferCmdHeader(cmd->param2,&ch->sndFrame.format,
					l,&ch->baseNote))<0)
				goto end;
			ch->durStart = r+(uint8*)cmd->param2;
			ch->durEnd = ch->sndFrame.format.frameCount;
			ch->loopStart = NULL;
			ch->loopCount = 0;
			if (l[1]>0 && ch->sndFrame.format.blockSize)
			{
				r = ch->sndFrame.format.frameCount/ch->sndFrame.format.blockSize;
				if (l[1]>r)
					r = l[1];
				ch->loopStart = ch->durStart+*l*ch->sndFrame.format.blockSize;
				ch->loopCount = (l[1]-l[0])*ch->sndFrame.format.frameSize;
				ch->durEnd = l[1]*ch->sndFrame.format.frameSize;
				ch->restEnd = ch->sndFrame.format.frameCount-ch->durEnd;
			}
			r = 0;
			if (cmd->cmd==sndBufferCmd || ch->voiceMode==sndVoiceAlwaysBuffer)
			{
				ch->sndFrame.cmpPos = ch->durStart;
				ch->actFlags|= chCompletion;
				r = 1;
			}
			/* initialize decompression variables */
			(void)decompressSndFrame(ch->sndFrame.dcmpBuf,NULL,&ch->sndFrame.format,0);
			ch->sndFrame.mixerRtn = NULL;
			goto end;
		case sndMWaveCmd:
		{
			genwave_base	*b = (genwave_base*)&ch->sndFrame.format.variables;

			if ((ch->sndFrame.format.type&15)!=CTYPE_WTAB)
			{
			extern DEF_DCMP_RTN(snddecRenderWave);
				if (ch->voiceMode<=sndVoiceLocked)
					break;
				(void)snddecRenderWave(0,0,&ch->sndFrame.format,0);
				ch->loopCount = 0;
				ch->sndFrame.format.dcmpRtn = NULL;
				ch->baseNote = 57;
			}
			*(void**)&b->perm = cmd->param2; /* pointer uses rest of space */
			b->rsize = (uint16)cmd->param1;
			b->type = (uint8)(cmd->param1>>16);
			r = cmd->param1>>24&15;
			b->rstep = 65536>>r;
			ch->sndFrame.format.rate = (b->rsize+1)*220<<r;
			ch->sndFrame.mixerRtn = NULL;
			r = 0;
			break;
		}
		case sndToneCmd: /* wave generation factor, separate from volume */
		case sndNoiseCmd:
		{
			genwave_base	*b = (genwave_base*)&ch->sndFrame.format.variables;

			if ((ch->sndFrame.format.type&15)!=CTYPE_TONE)
			{
			extern DEF_DCMP_RTN(snddecGenTone);
			extern DEF_DCMP_RTN(snddecGenNoise);
				if (ch->voiceMode<=sndVoiceLocked)
					break;
				if (cmd->cmd==sndNoiseCmd)
					(void)snddecGenNoise(0,0,&ch->sndFrame.format,0);
				else
					(void)snddecGenTone(0,0,&ch->sndFrame.format,0);
				ch->loopCount = 0;
				ch->sndFrame.format.dcmpRtn = NULL;
				ch->baseNote = 57;
			}
			if ((ch->sndFrame.format.type&15)==CTYPE_TONE && cmd->param1)
			{
				b->rsize = (uint16)cmd->param1;
				b->type = (uint8)(cmd->param1>>16);
				r = (cmd->param1>>24&15)^8;
				ch->sndFrame.format.rate = 220<<r;
				b->rstep = 1u<<(32-r);
				sndSetAmp(&ch->sndFrame.format,b->amp);
				ch->sndFrame.mixerRtn = NULL;
				r = 0;
			}
			break;
		}
		case sndPhaseCmd:
		{
			genwave_base	*b = (genwave_base*)&ch->sndFrame.format.variables;

			if ((ch->sndFrame.format.type&15)>=CTYPE_WTAB)
				b->phase = cmd->param1<<16;
			break;
		}
		case sndAmpCmd: /* wave factor, separate from volume */
			if ((ch->sndFrame.format.type&15)>=CTYPE_WTAB)
				sndSetAmp(&ch->sndFrame.format,cmd->param1&0xFF);
			break;
		case sndRateCmd:
			ch->sndFrame.format.rate = cmd->param1;
			ch->sndFrame.format.rateFrac = (uint32)cmd->param2;
			break;
		case sndOutRateCmd:
		if ((ch->sndFrame.format.type&15)<CTYPE_WTAB) /* an rmul for generated */
		{ /* temporarily bypass rates for a custom conversion rate */
			uint32		saveRate = ch->sndFrame.format.rate;
			uint32		saveRFrac = ch->sndFrame.format.rateFrac;
			ufixed		saveRMul = ch->sndFrame.rateMul;

			ch->sndFrame.format.rate = cmd->param1;
			ch->sndFrame.format.rateFrac = (uint32)cmd->param2;
			ch->sndFrame.rateMul = 1<<16;
			sndUpdateMixerRtn(&ch->sndFrame,
				((sndMgrData*)ch->private)->rflag&(ch->initFlag<<24|0xFFFFFF));
			ch->sndFrame.format.rate = saveRate;
			ch->sndFrame.format.rateFrac = saveRFrac;
			ch->sndFrame.rateMul = saveRMul;
			break;
		}
		case sndRateMultiplierCmd:
			if (!ch->sndFrame.dcmpSize && (uint16)cmd->param1==cmd->param1)
				ch->sndFrame.cnvFrc = (uint16)cmd->param1;
			ch->sndFrame.rateMul = (ufixed)cmd->param2;
			ch->sndFrame.mixerRtn = NULL;
			break;
		case sndFreqCmd:
			ch->actFlags|= chFreqSet;
			goto fc;
		case sndFreqDurationCmd:
			ch->actFlags&=~chFreqSet; /* freqDuration erases old frequency */
		fc:	ch->sndFrame.mixerRtn = NULL;
			ch->actFlags|= chDurAct;
			ch->sndFrame.rateMul = sndMIDINoteToRate((uint8)cmd->param2,
				ch->baseNote);
			if (cmd->cmd!=sndFreqDurationCmd)
				break;
			if ((ch->sndFrame.format.type&15)>=CTYPE_WTAB
					&& (uint32)cmd->param2&255<<24)
				sndSetAmp(&ch->sndFrame.format,(uint32)cmd->param2>>24);
			goto wc;
		case sndRestCmd:
			ch->actFlags&=~(chDurAct|chFreqSet);
			ch->durEnd+= ch->restEnd, ch->restEnd = 0;
			//ch->durStart = NULL;
			if ((ch->sndFrame.format.type&15)>=CTYPE_WTAB)
				ch->sndFrame.format.frameCount = 0;
		case sndWaitCmd:
		{
			uint32		rem;

		wc:	r = ch->waitCnt+umuldivrem(cmd->param1,
				((sndMgrData*)ch->private)->rflag&0xFFFFFF,2000,&rem);
			if (r+0u>=ch->waitCnt)
			{
				rem+= ch->waitCntF;
				if (rem>=2000)
					rem-= 2000, r+= r!=-1;
				ch->waitCntF = (uint16)rem;
				ch->waitCnt = (uint32)r;
				//printf("waitcntf %d\n",
				//	rem);
			} else /* overflow */
				ch->waitCnt =~0u;
			ch->actFlags|= chDurAux; /* yes, also applies to restCmd */
			if (ch->waitCnt && ch->actFlags&(chFreqSet|chDurAct))
				if ((ch->sndFrame.format.type&15)>=CTYPE_WTAB)
					ch->sndFrame.format.frameCount = 65536;
				else if ((cmd->cmd==sndFreqDurationCmd /* starts at beginning */
						|| (ch->loopCount
						&& !ch->sndFrame.cmpPos)) /* not already playing */
						&& ch->voiceMode>=sndVoiceDefault)
				{
					//printf("wait start\n");
					ch->actFlags|= chDurAct;
					ch->sndFrame.cmpPos = ch->durStart;
					ch->sndFrame.format.frameCount = ch->durEnd;
				}
			/* must be updated ahead of loop for cnvRate */
			if (!ch->sndFrame.mixerRtn)
				sndUpdateMixerRtn(&ch->sndFrame,
					((sndMgrData*)ch->private)->rflag&(ch->initFlag<<24|0xFFFFFF));
			r = 1;
			goto end;
		}
		case sndStopCmd:
			ch->actFlags|= chStopped;
			r = 2;
			goto end;
		case sndPauseCmd:
			ch->actFlags|= chPaused;
			r = 2;
			goto end;
		case sndResumeCmd:
			ch->actFlags&=~(chStopped|chPaused);
			break;
		case sndCallBackCmd:
			if (cmd->param2)
				((sndCallBack)cmd->param2)(ch,cmd);
			else if (ch->defaultCallBack)
				ch->defaultCallBack(ch,cmd);
			break;
		case sndRLVolumeCmd:
			ch->LRVolume = (uint16)((uint32)cmd->param2>>16)
				|(uint32)cmd->param2<<16;
			goto upmix;
		case sndLRVolumeCmd:
			ch->LRVolume = (uint32)cmd->param2;
		upmix:	updateChMixerAmp(ch);
			break;
		case sndFilterCmd:
			ch->sndFrame.mixerRtn = NULL;
			if ((ch->sndFrame.filterL=(uint32)cmd->param2)==0)
			{
				ch->initFlag&=~sndInitExtendedFilter;
				ch->sndFrame.filterR = 0x1FFFF;
			} else
			{
				ch->initFlag|= sndInitExtendedFilter;
				ch->sndFrame.filterR = 0x3FFFF;
			}
			break;

		inv:
		default:
			r =-50;
			goto end;
		}
	} while(c && --c);
end:	if (r==1)
		ch->cmdInProg = *cmd;
	return r;
}

int sndDoCommand(sndChannel *ch,
	sndCommand *cmd,
	int8 immediate)
{
	sndMgrData	*mgr;
	int		err;

	if (!ch || !cmd)
		return -1;
	//printf("actFlags %02X\n",
	//	ch->actFlags);
	if ((mgr=ch->private)!=0)
		MgrMutexLock(mgr);
	if (immediate)
		err = processCommands(ch,cmd,0);
	else
		err = enQueue(ch,cmd);
	if (mgr)
		MgrMutexUnlock(mgr);
	return err;
}

/* fill channel block using sndMixSamples;
  advance queue when samples or current period is up;
  if all else fails, fill lastmost silence point
*/

#define NOSAMPLES(ch) ch->sndFrame.format.frameCount<=0 \
	&& !ch->sndFrame.cmpPos \
	&& !ch->sndFrame.pos

static void sndChRender(int32 *out,
	sndChannel *ch)
{
	sndCommand	cmd;
	uint32		rflag, rem, w, v;
	int		r;

	rem = sndEndFrameCount;
	ch->actFlags|= chBusy;
	rflag = ((sndMgrData*)ch->private)->rflag&(ch->initFlag<<24|0xFFFFFF);
	if (ch->actFlags&chStopped)
	{ /* no processing; fill silence only */
		sndMixFill(rem,rflag,
			out+2*(sndEndFrameCount-rem),
			&ch->sndFrame);
		goto end;
	}
	do
	{
		if (ch->actFlags&chPaused)
			goto fill;
		if (!ch->waitCnt && !(ch->actFlags&chCompletion))
		{
			r = processCommands(ch,&cmd,~0);
			/* because of reInit, rflag must be recalculated */
			rflag = ((sndMgrData*)ch->private)->rflag&(ch->initFlag<<24|0xFFFFFF);
			if (r<=0) /* queue has emptied */
			{
				ch->cmdInProg.cmd = sndNullCmd;
			#ifdef TERMINATE_LOOP_IF_ZERO_DELAY
				if (!ch->waitCnt)
					ch->sndFrame.format.frameCount = 0;
			#endif
			}
		}
		if ((w=ch->waitCnt)==0 || w>=rem)
		fill:	w = rem;
		if (NOSAMPLES(ch) || !ch->sndFrame.rateMul)
		{
			sndMixFill(w,rflag,
				out+2*(sndEndFrameCount-rem),
				&ch->sndFrame);
			rem-= w;
			if (!(ch->actFlags&chPaused))
				ch->waitCnt = ch->waitCnt>w ? ch->waitCnt-w:0;
		} else
		{
			if (ch->waitCnt && ch->sndFrame.cnvRate)
			{
				int64		t;

				mulu64u32(t,w,ch->sndFrame.cnvRate);
				/* don't overflow dcmp buffer */
				if (t>(sndEndFrameCount<<16)-1)
					t = sndEndFrameCount<<16;
				t+= ch->waitCntCF;
				ch->waitCntCF = (uint16)int64lo(t);
				v = int64lo(t)>>16;
			} else
				v = w;
			w = rem;
			rem = sndMixSamples(out+2*(sndEndFrameCount-rem),&ch->sndFrame,
				rflag,v,rem);
			w-= rem;
			if (ch->sndFrame.err&2)
			{ /* stalled */
				ch->sndFrame.err&= 128;
				ch->sndFrame.pos = 0;
				ch->sndFrame.format.frameCount = 0;
			}
			if (!(ch->actFlags&chPaused))
				ch->waitCnt = ch->waitCnt>w ? ch->waitCnt-w:0;
			if (NOSAMPLES(ch))
			{
				//printf("complete\n");
				ch->actFlags&=~chCompletion;
				if (ch->actFlags&chDurAct && ch->loopCount && ch->waitCnt)
				{ /* loop restart still isn't consistent or correct */
					//printf("loop\n");
					ch->sndFrame.cmpPos = ch->loopStart;
					ch->sndFrame.format.frameCount = ch->loopCount;
				}
			}
		}
	} while(rem>0);
end:	ch->actFlags&=~chBusy;
}

/* channel creation/disposal */

sndChannel *sndNewChannel(sndManager *smgr,
	uint8 initFlags,
	void(*defaultCallBack)(sndChannel*,
		sndCommand*))
{
	sndMgrData	*mgr;
	sndChannel	**p, *ch;

	if ((mgr=(sndMgrData*)smgr)==NULL)
		return NULL;
	for(p=&mgr->sndChannels; *p;)
		if (*(p=&(**p).next)==mgr->sndChannels)
			return NULL; /* recursion! */
	if ((ch=calloc(sizeof(sndChannel),1))==NULL)
		return NULL;
	if ((ch->sndFrame.dcmpBuf=calloc(sizeof(int16)*2,chBufCount))==NULL)
		{free(ch); return NULL;}
	//ch->sndFrame.pos = MAXIMUM_FILTER_SIZE;
	ch->initFlag = initFlags|sndInitPanMask;
	ch->voiceMode = sndVoiceDefault;
	ch->private = mgr;
	ch->qLength = GET_ARRAY_COUNT(ch->queue);
	ch->sndFrame.rateMul = 0x10000;
	ch->sndFrame.filterR = initFlags&sndInitExtendedFilter ? 0x3FFFF:0x1FFFF;
	ch->LRVolume = defaultVolume;
	ch->defaultCallBack = defaultCallBack;
	updateChMixerAmp(ch);
	MgrMutexLock(mgr); /* usually not needed; but to be safe */
	*p = ch;
	MgrMutexUnlock(mgr);
	return ch;
}

int sndDisposeChannel(sndChannel *ch)
{
	sndMgrData	*mgr;
	sndChannel	**p;

	if (!ch || (mgr=ch->private)==NULL)
		return -1;
	if (ch->streamEntry)
	{ /* tsk tsk */
		//printf("killing with active stream! %p\n",
		(void)sndPlayFileStop(ch);
	}
	for(p=&mgr->sndChannels; *p;)
		if (*p==ch)
			goto found;
		else if (*(p=&(**p).next)==mgr->sndChannels)
			return -50; /* recursion! */
	return -205; //-7;
found:	MgrMutexLock(mgr);
	ch->voiceMode = sndVoiceDisabled;
	*p = ch->next;
	ch->actFlags|= chDisposed;
	if (ch->sndFrame.dcmpBuf)
		{free(ch->sndFrame.dcmpBuf); ch->sndFrame.dcmpBuf = NULL;}
	MgrMutexUnlock(mgr);
	ch->private = NULL;
	free(ch);
	return 0;
}

int sndChannelResize(sndChannel **_ch,
	int count)
{
	sndMgrData	*mgr;
	sndChannel	*ch, **p;

	if (!_ch || (ch=*_ch)==NULL || (mgr=ch->private)==NULL || ch->actFlags&chDisposed)
		return -1;
	if (count>IMAX(ch->qLength))
		count = IMAX(ch->qLength);
	for(p=&mgr->sndChannels; *p;)
		if (*p==ch)
			goto found;
		else if (*(p=&(**p).next)==mgr->sndChannels)
			return -50; /* recursion! */
	return -205; //-7;
found:	if (count<0)
		return -128;
	MgrMutexLock(mgr);
	ch->qHead = ch->qTail = 0;
	if ((ch=realloc(ch,OFFSET(sndChannel,queue)+sizeof(sndCommand)*count))!=NULL)
	{
		*p = ch;
		*_ch = ch;
		ch->qLength = count;
	}
	MgrMutexUnlock(mgr);
	return ch ? 0:-2;
}

int sndChannelPlaying(sndChannel *ch)
{
	sndMgrData	*mgr;
	int		r;

	if (!ch || ch->actFlags&chStopped)
		return 0;
	mgr = ch->private;
	if (mgr)
		MgrMutexLock(mgr); /* necessary for a locked snapshot with no tearing */
	r = ch->sndFrame.cmpPos /* is processing samples */
		|| (ch->sndFrame.dcmpSize && ch->sndFrame.pos); /* has outgoing samples */
	if (ch->qHead!=ch->qTail && ch->qLength>0) /* has queue */
		r|= 2;
	r|= (ch->waitCnt ? 4:0);
	if (mgr)
		MgrMutexUnlock(mgr);
	return r;
}

sndManager *sndMgrFromChannel(sndChannel *ch)
{
	sndManager	*mgr;

	if (!ch || (mgr=ch->private)==NULL || ch->actFlags&chDisposed)
		return NULL;
	return mgr;
}

int sndPlaySndList(sndChannel **_ch,
	const uint8 *start,
	long z,
	int8 mayresize,
	int8 immediate)
{
	sndChannel	*ch;
	sndCommand	cmd;
	const uint8	*s, *e;
	int32		c, v, vers;
	int		r;

	if (!_ch || (ch=*_ch)==NULL || (s=start)==NULL)
		return -1;
	if (z<4)
		return -3;
	e = s+z;
	vers = va_be16(s);
	if (vers-1u>1u)
		return -206; /* badFormat */
	if (vers==1)
	{ /* version 1: parse modifiers */
		c = va_be16(s), v = 0;
		if ((c=(c^0x8000)-0x8000)<0)
			return -206;
		if (!c)
		{
			if (ch->sndFrame.format.depth)
				goto nomod;
			return -206; /* no voice, no synth */
		}
		v = va_be16(s); /* synth (low byte only?) (...no, it's 16-bit) */
		if (v!=1 && v!=3 && v!=5)
			return -50;
		if (ch->voiceMode==sndVoiceDefault && v!=5)
		{
			if (v==1)
			{
				cmd.cmd = sndToneCmd;
				cmd.param1 = 100|sndToneSquare<<16;
				(void)sndDoCommand(ch,&cmd,1);
			} else if (v==3)
			{
				cmd.cmd = sndMWaveCmd;
				cmd.param1 = 0;
				cmd.param2 = 0;
				(void)sndDoCommand(ch,&cmd,1);
				//ch->initFlag|= sndInitNoInterpolation;
			}
			ch->voiceMode = sndVoiceLocked;
		}
		/* init flags typically ignored */
		s+= 4+6*c-6;
		/* what the heck are srate flags? */
	} else /* version 2: implied synth */
	{
		if (ch->voiceMode==sndVoiceDefault)
			ch->voiceMode = sndVoiceAlwaysBuffer;
		s+= 2; /* refCount */
	}
nomod:	if (s+2>e)
		return -3;
	c = va_be16(s);
	if ((c=(c^0x8000)-0x8000)<=0)
		return c ? -5:0;
	//printf("count %d + voice %d, queue %d/%d\n",
	//	c,vc,ch->qHead,ch->qTail);
	if (mayresize && c>=ch->qLength)
	{ /*
	    Neither SndPlay nor ResEdit expand the queue buffer.
	    SndPlay instead passes false for noWait, making the async param partly useless
	   when the queue becomes full during play.
	  */
		(void)sndChannelResize(_ch,c+1); /* what to do with an error here... */
		ch = *_ch;
	}
	r = 0;
	do
	{
		if (s>e-8)
		{
			r =-3;
			break;
		}
		v = (int)va_be16(s);
		cmd.param1 = va_be16(s);
		cmd.param2 = (void*)va_be32(s);
		switch(v)
		{
		case 82:
		{ /* rateCmd - fixed point * 22254.545455 */
			uint64	hl;

			if (vers==2) /* vers. 2 pretty much skips all but soundCmd/bufferCmd */
				continue;
			v = sndOutRateCmd; /* actually an output rate */
			(void)mulu64u32(hl,(ufixed)cmd.param2,(ufixed)(22254.545455*65536+.5));
			cmd.param1 = (uint32)int64hi(hl);
			cmd.param2 = (void*)int64lo(hl);
			break;
		}
		case 86:
			if (vers==2)
				continue;
			v = sndRateMultiplierCmd;
			cmd.param1 = 0x80000000u;
			break;
		case 46:
			if (vers==2)
				continue;
			v = sndRLVolumeCmd;
		case 40: /* freqDurationCmd */
			if (vers==2)
				continue;
			/* amp field is actually ignored by my tests */
			cmd.param2 = (void*)(uint8)cmd.param2;
			v = sndFreqDurationCmd;
			break;
		case 44: /* timbreCmd - does nothing outside of square/note synth */
			cmd.param1 = (uint8)cmd.param1;
			cmd.param1 = (cmd.param1*cmd.param1>>6)+8;
			cmd.param1 = PIN(cmd.param1,25,1016)|sndToneSquare<<16;
		case 43: /* ampCmd - does nothing outside of square/note synth */
		case 42: /* freqCmd */
		case 41: /* restCmd */
			if (vers==2)
				continue;
			v+= sndRestCmd-41;
			break;
		case sndCallBackCmd: cmd.param2 = 0; /* ignored, of course */
		case sndWaitCmd:
		case sndNullCmd:
		case sndQuietCmd:
		case sndFlushCmd:
		case sndReInitCmd:
		case sndPauseCmd: /* not a great idea allowing all of these... */
		case sndResumeCmd:
			if (vers==2)
				continue;
			break;

			/*
			   The get commands should never be put in a soundList resource
			  as they eventually hang the Mac OS after playback.
			   Here, they're recognized and ignored.
			*/
		case 24: /* availableCmd */
		case 25: /* versionCmd */
		case 45: /* getAmpCmd */
		case 47: /* getVolumeCmd */
		case 61: /* phaseCmd (waveTable) - shift.w/channelPtr.l (not portable) */
		case 83: /* continueCmd - obsolete */
		case 85: /* getRateCmd */
		case 87: /* getRateMultiplierCmd */
		case 90: /* sizeCmd */
		case sndSyncCmd:
			continue;

		default:
			if ((v-=0x8050)<2u)
			{
				v+= sndSoundCmd;
				cmd.param2 = (void*)(start+(int32)cmd.param2);
				if ((const uint8*)cmd.param2>e-22)
					{r =-3; goto end;}
				break;
			}
			if (v==0x803C-0x8050)
			{
				v = sndMWaveCmd; /* no high param1 bits=8-bit */
				cmd.param2 = (void*)(start+(int32)cmd.param2);
				if ((const uint8*)cmd.param2>e-cmd.param1)
					{r =-3; goto end;}
				cmd.param1|= 1<<24;
				break;
			}
			r =-50;
			goto end;
		}
		cmd.cmd = (uint16)v;
		if ((v=sndDoCommand(ch,&cmd,immediate))<0)
			return v;
		++r, immediate = 0;
	} while(--c);

end:	return r;
}

/* sound file playback */

#ifdef __MT__
 #include "mevent.h"
extern timerThread	*global_timer_thread;
#endif

typedef struct playFileData
{
	FILE		*in;
	sndFormat	format;
	int32		mxc;
#ifdef __MT__
	sndTaskEntry	*tstream;
#endif
	sndChannel	*chan;
	uint8		*wh, *alt;
	uint8		buf[1];
} playFileData;

#ifdef __MT__
static void playFileCallBack(sndChannel *ch,
	sndCommand *_cmd);
#endif

static int loader(playFileData *pfd)
{
	sndCommand	cmd;
	sndChannel	*ch;
	streamwave_base	*s;
	long		bz, o;

	o = bz = 0;
	if (!pfd->wh)
		return -1;
	s = (streamwave_base*)pfd->format.variables;
	if (pfd->format.streamType)
		o = s->accum;
	//printf("loading... %d\n",
	//	o);
	if (!o)
		pfd->wh = pfd->wh==pfd->buf ? pfd->alt:pfd->buf; /* swap buffer */
	if (pfd->format.streamType)
	{
		do
		{
			if ((bz=readSndFrame(pfd->wh+o,pfd->in,&pfd->format))==0)
			{
				if (!o)
					goto done;
				break;
			}
			o+= bz, s->accum = o;
		} while(o<16384);
		//printf("here we go %d/%d\n",
		//	o,pfd->format.maxRead);
		bz = o/pfd->format.frameSize;
		s->accum = 0;
	} else if (!pfd->format.frameCount)
		goto done;
	else
	{
		if (pfd->format.type==CTYPE_ADPCM)
			o = (long)ftell(pfd->in);
		if ((bz=fread(pfd->wh,pfd->format.blockSize, /* raw dcmp block size */
				MIN(pfd->mxc,pfd->format.frameCount),pfd->in))==0)
		{
			if (pfd->format.type==CTYPE_ADPCM && pfd->format.blockSize>4
					&& (long)ftell(pfd->in)-o>4)
			{ /* short ADPCM block */
				pfd->format.frameCount = bz = 1;
				goto sw;
			}
		done:	pfd->wh = NULL;
			return 0;
		}
	}
sw:	ch = pfd->chan;
	if (bz>pfd->format.frameCount)
		bz = pfd->format.frameCount;
	pfd->format.frameCount-= bz;
	cmd.cmd = sndRawCmd;
	cmd.param1 = bz;
	cmd.param2 = pfd->wh;
	(void)sndDoCommand(pfd->chan,&cmd,0);
	cmd.cmd = sndCallBackCmd;
	cmd.param1 = 0;
	cmd.param2 = (void*)playFileCallBack;
	(void)sndDoCommand(ch,&cmd,0);
	return 1;
}

#ifdef __MT__
static void autoloader(sndTaskEntry *param)
{
	playFileData	*pfd;

	if ((pfd=param->userInfo)==NULL)
		return;
	if (!pfd->chan)
		return;
	if (!pfd->chan->streamEntry)
		return;
	(void)loader(pfd);
}
#endif

static void playFileCallBack(sndChannel *ch,
	sndCommand *_cmd)
{
	playFileData	*pfd;
(void)_cmd;
	if ((pfd=ch->streamEntry)==NULL || !pfd->format.blockSize
			|| !pfd->format.maxRead)
		return;
#ifdef __MT__
	{
		sndTaskEntry	*task;

		if ((task=pfd->tstream)!=0)
			task->primed = 1;
	}
#else
	loader(pfd);
#endif
}

int _sndStreamOpenClose(FILE *in,
	sndFormat *f,
	int32 k,
	int quietWarnings);

int sndPlayFile(FILE *in,
	sndChannel *ch,
	sndFormat *f,
	int32 k,
	int quietWarnings)
{
	playFileData	*pfd;
	sndCommand	cmd;
	int32		z, v;

	if (!in || !f || !ch || !f->blockSize || ch->streamEntry)
		return -1;
(void)quietWarnings;
	if ((v=_sndStreamOpenClose(in,f,k,quietWarnings))<0)
		return v;
	/* needs to be at least 32KB for 96KHz samples */
	for(z=f->maxRead*f->blockSize; z<32768; z+=z)
		;
	if (z<=0)
		return -1;
	v = f->maxRead*f->frameSize>>1;
	v = MAX(z,v);
	if ((pfd=calloc(1,v+sizeof(playFileData)))==NULL)
		return -2;
	ch->streamEntry = pfd;
	/* alternate must be in a separate allocation
	  (Access Exception can occur if any pages cross writing when sndMgr reads)
	*/
	if ((pfd->alt=calloc(1,v))==NULL)
	{
		free(pfd);
		return -2;
	}
#ifdef __MT__
	(void)setvbuf(in,NULL,_IOFBF,32768);
	pfd->chan = ch;
	if ((pfd->tstream=sndInstallTask((sndManager*)ch->private,autoloader))!=NULL)
		pfd->tstream->userInfo = pfd;
#endif
	pfd->in = in;
	pfd->format = *f;
	pfd->mxc = z/f->blockSize;
	/* skip any leading frames now */
	if ((v=k)<0)
		k =~k;
#if 0 /* now more efficiently handled by _sndStreamOpenClose */
	for(; k; --k)
		if (decompressSndFrame(pfd->buf,pfd->buf,&pfd->format,
				readSndFrame(pfd->buf,in,&pfd->format))-pfd->buf<=0)
			return 0; /* ended before we even started */
#endif
	if (v<0)
	{
		cmd.cmd = sndPauseCmd;
		(void)sndDoCommand(ch,&cmd,1);
	}
	cmd.cmd = sndFlushCmd;
	(void)sndDoCommand(ch,&cmd,1);
	cmd.cmd = sndFormatCmd;
	cmd.param2 = &pfd->format;
	(void)sndDoCommand(ch,&cmd,1);
	pfd->wh = pfd->buf;
	(void)loader(pfd);
	playFileCallBack(pfd->chan,NULL);
	return 1;
}

int sndPlayFileStop(sndChannel *ch)
{
	sndCommand	cmd;
	playFileData	*pfd;
	int		r;

	if (!ch || !ch->private)
		return -1;
	if ((pfd=ch->streamEntry)==NULL)
		return 0;
	r = pfd->format.frameCount>0;
	/* at least one decompressor (IMA) might on occasion bomb without this */
	cmd.cmd = sndFlushCmd;
	sndDoCommand(ch,&cmd,1);
	cmd.cmd = sndQuietCmd;
	sndDoCommand(ch,&cmd,1);
#ifdef __MT__
	(void)sndRemoveTask((sndManager*)ch->private,pfd->tstream);
	pfd->tstream = NULL;
#endif
	pfd->format.frameCount = 0;
	(void)_sndStreamOpenClose(NULL,&pfd->format,0,0);
	ch->streamEntry = NULL;
	free(pfd->alt); pfd->alt = NULL;
	free(pfd);
	return r;
}

/* sound manager top level */

int sndMgrMute(sndManager *smgr,
	uint8 lrdisable)
{
	sndMgrData	*mgr;
	sndChannel	*ch;

	if ((mgr=(sndMgrData*)smgr)==NULL)
		return -1;
	MgrMutexLock(mgr);
	mgr->rflag&=~(sndInitPanMask<<24);
	mgr->rflag|= (lrdisable&=sndInitPanMask)<<24;
	if ((ch=mgr->sndChannels)!=NULL)
		do
			updateChMixerAmp(ch);
		while((ch=ch->next)!=NULL);
	MgrMutexUnlock(mgr);
	return lrdisable;
}

static inline void sndMgrRender(sndMgrData *mgr,
	int i)
{
	sndChannel	*ch = mgr->sndChannels;
	int32		*b = mgr->out[i];

	memset(b,0,sizeof(*mgr->out));
	do
	{
		if (ch->voiceMode==sndVoiceDisabled)
			continue;
		sndChRender(b,ch);
	} while((ch=ch->next)!=NULL);
	s32to16(sndEndFrameCount,mgr->rflag,b);
}

#ifdef WINUSERAPI
 #if defined(__MT__) && !defined(NO_SEMAPHORE)
/* this should work on Win32 and up */
/* ReleaseSemaphore will not work from a callback on older systems */
static void CALLBACK sndWaitSync(HWAVEOUT hwa,
	UINT msg,
	DWORD_PTR dwu,
	DWORD dw1,
	DWORD dw2)
{
	if (msg!=WOM_DONE)
		return;
	ReleaseSemaphore(((sndMgrData*)dwu)->hSemaphore,1,NULL);
}
 #elif defined(QUEUED_WAVE)
static void CALLBACK sndWaitSync(HWAVEOUT hwa,
	UINT msg,
	DWORD_PTR dwu,
	DWORD dw1,
	DWORD dw2)
{
	sndMgrData	*mgr = (sndMgrData*)dwu;
(void)hwa; (void)dw1; (void)dw2;
	if (msg!=WOM_DONE)
		return;
	MgrMutexLock(mgr);
	--*(int32*)&mgr->hWHeader->dwUser;
	MgrMutexUnlock(mgr);
}
 #endif

HWAVEOUT sndMgrGetHWO(sndManager *smgr)
{
	sndMgrData	*mgr;

	if ((mgr=(sndMgrData*)smgr)==NULL || !(mgr->rflag&0xFFFFFF))
		return NULL;
	return mgr->hWAudio;
}

static int sndMgrProcess(sndMgrData *mgr)
{
	WAVEHDR		*hdr;
	int		x;

	if (!mgr->sndChannels)
		return 0;
 #if defined(__MT__) && !defined(NO_SEMAPHORE)
	x = (int32)mgr->hWHeader[1].dwUser;
	hdr = mgr->hWHeader+x;
	//MgrMutexLock(mgr);
	sndMgrRender(mgr,x);
	//MgrMutexUnlock(mgr);
	if (mgr->hWHeader->dwUser)
		WaitForSemaphore(mgr->hSemaphore,INFINITE);
	waveOutWrite(mgr->hWAudio,hdr,sizeof(*hdr));
	*(uint32*)&mgr->hWHeader[1].dwUser = (x+1u)%sndBufferCount;
	mgr->hWHeader->dwUser = (DWORD_PTR)1;
	return 1;
 #elif defined(QUEUED_WAVE)
	if ((uint32)mgr->hWHeader->dwUser>=sndBufferCount)
		return 1;
	x = (int32)mgr->hWHeader[1].dwUser;
	hdr = mgr->hWHeader+x;
	MgrMutexLock(mgr);
	++*(int32*)&mgr->hWHeader->dwUser;
	sndMgrRender(mgr,x);
	MgrMutexUnlock(mgr);
	*(uint32*)&mgr->hWHeader[1].dwUser = (x+1u)%sndBufferCount;
  #ifdef WAVE_DUMP
	_write(fwoutn,hdr->lpData,hdr->dwBufferLength);
  #endif
	if (waveOutWrite(mgr->hWAudio,hdr,sizeof(*hdr))!=MMSYSERR_NOERROR)
		hdr->dwFlags|= WHDR_DONE;
	return 2;
 #else
	{ /* basic loader (no callback) */
		int		i = 0;

		x = 2, hdr = mgr->hWHeader;
		do
			if (hdr->dwFlags&WHDR_DONE)
			{
				MgrMutexLock(mgr);
				hdr->dwFlags&=~WHDR_DONE;
				sndMgrRender(mgr,i);
				MgrMutexUnlock(mgr);
			#ifdef WAVE_DUMP
				fwrite(hdr->lpData,hdr->dwBufferLength,1,fwout);
			#endif
				if (waveOutWrite(mgr->hWAudio,hdr,sizeof(*hdr))!=MMSYSERR_NOERROR)
					hdr->dwFlags|= WHDR_DONE;
				x = 1;
			}
		while(++hdr,++i<sndBufferCount);
	}
	return x;
 #endif
}

int sndMgrWait(sndManager *smgr)
{
	sndMgrData	*mgr;
	uint32		rate;

	if ((mgr=(sndMgrData*)smgr)==NULL || (rate=mgr->rflag&0xFFFFFF)==0)
		return -1;
	if (!(mgr->rflag&0x80000000u))// || !mgr->sndChannels)
		return 0;
	Sleep(1+sndEndFrameCount*1000u/rate);
	return 1;
}

 #ifdef __MT__
sndTaskEntry *sndInstallTask(sndManager *thd,
	sndStreamRoutine rtn)
{
	sndMgrData	*mgr = (sndMgrData*)thd;
	sndTaskEntry	*t;
//printf(" install %p,%p,%d\n",
// thd,rtn,mgr->tcount);
	if (!rtn || !mgr || mgr->tcount>=mgr->tmaximum)
		return NULL; /* bad param or maxed out */
	/* probably no need to lock; the count is incremented last */
	MgrMutexLock(mgr);
	t = mgr->tlist+mgr->tcount;
	t->routine = rtn;
	++mgr->tcount;
	MgrMutexUnlock(mgr);
	return t;
}

int sndRemoveTask(sndManager *thd,
	sndTaskEntry *entry)
{
	sndMgrData	*mgr = (sndMgrData*)thd;
	sndTaskEntry	*s;
	int		c, r;
//printf(" remove %p,%p,%d\n",
// thd,entry,mgr->tcount);
	if (!mgr || !entry || (c=mgr->tcount)==0 || !mgr->tmaximum)
		return -1;
	MgrMutexLock(mgr);
	s = mgr->tlist, r = 1;
	do
		if (s==entry)
		{
			s->primed = 0;
			++s;
			if (--c)
				memcpy(s-1,s,c*sizeof(*s));
			mgr->tcount = (uint16)c;
			goto end;
		}
	while(++s,--c);
	r = 0;
end:	MgrMutexUnlock(mgr);
	return r;
}

static void __stdcall streamsThreadFunc(UINT id,
	UINT msg,
	DWORD_PTR user,
	DWORD_PTR dw1,
	DWORD_PTR dw2)
{
	sndMgrData	*mgr = (sndMgrData*)user;
(void)id; (void)msg; (void)dw1; (void)dw2;
	if (mgr->tmaximum)
	{
		sndTaskEntry	*s;
		int		c;

		MgrMutexLock(mgr);
		if ((c=mgr->tcount)!=0)
			for(s=mgr->tlist; c--; ++s)
				if (s->primed)
				{
				//__try{
					s->routine(s);
				//}
				//__finally{}
					s->primed = 0;
				}
		MgrMutexUnlock(mgr);
	}
}

static void __cdecl sndMgrThread(void *smgr)
{
	while(((sndMgrData*)smgr)->rflag&0x80000000u)
	#ifdef NO_SEMAPHORE
	{
		static int8	retry; /* may stutter on older systems without more effort */
		int		r;

		if ((r=sndMgrProcess(smgr))!=0 && !(retry+=0x10))
			r = sndMgrProcess(smgr);
		Sleep(r^3);
	}
	#else
		if (!sndMgrProcess(smgr))
			Sleep(2); /* semaphore waits */
	#endif
	Sleep(2);
	_endthread();
}

void sndMgrProcWait(sndManager *smgr)
{
	if (smgr)
		Sleep(2);
}
 #else
void sndMgrProcWait(sndManager *smgr)
{
	if (smgr)
		Sleep(2-!!sndMgrProcess((sndMgrData*)smgr));
}
 #endif

int sndMgrMixRate(sndManager *smgr,
	int rate)
{
	sndMgrData	*mgr;
	int		r;

	if ((mgr=(sndMgrData*)smgr)==NULL || !(mgr->rflag&0x80000000u) || rate<=0)
		return -1;
	rate = PIN(rate,1000,100000);
	MgrMutexLock(mgr);
	r = mgr->rflag;
	mgr->rflag = (r&0xFF000000u)|rate;
	MgrMutexUnlock(mgr);

	return r;
}

int sndMgrGetMixRate(sndManager *smgr)
{
	sndMgrData	*mgr;

	if ((mgr=(sndMgrData*)smgr)==NULL || !(mgr->rflag&0x80000000u))
		return -1;
	return mgr->rflag&0xFFFFFF;
}

int sndMgrGetVolume(sndManager *smgr,
	ufixed *L,
	ufixed *R,
	ufixed *BL,
	ufixed *BR,
	ufixed *LF)
{
	DWORD		wlevel;
	int		r;
(void)BL; (void)BR; (void)LF;
	if (!smgr)
		return -32767;
	r = waveOutGetVolume(sndMgrGetHWO(smgr),&wlevel);
	if (r!=MMSYSERR_NOERROR)
		return r;
	if (L)
		r = (uint16)wlevel, *L = r+(r>>15);
	if (R)
		r = (uint32)wlevel>>16, *R = r+(r>>15);
	return 0;
}

int sndMgrSetVolume(sndManager *smgr,
	ufixed L,
	ufixed R,
	ufixed BL,
	ufixed BR,
	ufixed LF)
{
(void)BL; (void)BR; (void)LF; /* only stereo currently supported */
	if (L>65536)
		L = 65535;
	else if (L>32768)
		--L;
	if (R>65536)
		R = 65535;
	else if (R>32768)
		--R;
	return waveOutSetVolume(sndMgrGetHWO(smgr),(R<<16)|L);
}

int sndMgrClose(sndManager *smgr)
{
	sndMgrData	*mgr;
	sndChannel	*ch;
	WAVEHDR		*hdr;
	int		i, r = 0;

	if ((mgr=(sndMgrData*)smgr)==NULL || !(mgr->rflag&0x80000000u))
		return -1;
	MgrMutexLock(mgr);
	mgr->rflag = 0;
	MgrMutexUnlock(mgr);
	/* helps fix a crash when closing during playback */
	if ((ch=mgr->sndChannels)!=NULL)
		do
			ch->actFlags|= chStopped;
		while((ch=ch->next)!=NULL);

	/* thread will close itself */
//#ifdef __MT__
//	CloseHandle(mgr->thread);
//#endif

	while((ch=mgr->sndChannels)!=NULL)
		(void)sndDisposeChannel(ch);

 #ifdef __MT__
	if (mgr->tmaximum)
	{
		mgr->tcount = mgr->tmaximum = 0;
		(void)timeKillEvent(mgr->ticEID);
	}
  #ifndef NO_SEMAPHORE
	if (mgr->hSemaphore)
	{
		HANDLE		sem = mgr->hSemaphore;

		ReleaseSemaphore(sem,1,NULL);
		Sleep(8);
		mgr->hSemaphore = NULL;
		CloseObject(sem);
		++r;
	}
  #endif
 #endif
	Sleep(1);
	for(i=0,hdr=mgr->hWHeader; i<sndBufferCount; ++i,++hdr)
		while(hdr->dwFlags&WHDR_PREPARED
				&& waveOutUnprepareHeader(mgr->hWAudio,hdr,sizeof(*hdr))
				==WAVERR_STILLPLAYING)
			Sleep(2);

	if (mgr->hWAudio)
	{
		while(waveOutClose(mgr->hWAudio)==WAVERR_STILLPLAYING)
			Sleep(2);
		mgr->hWAudio = NULL, ++r;
	}
 #ifdef WAVE_DUMP
	{
		const uint8	*h;
		int64		llz = _tell64(fwoutn)-_lastFmtSize;
		long int	z = MIN(llz,IMAX(z));

		_lseek(fwoutn,0,SEEK_SET);
		h = setupWAVEHeader(&format,&z);
		_write(fwoutn,h,z);
		fclose(fwout); fwout = NULL;
		fwoutn =-1;
	}
 #endif
	DeleteCriticalSection(&mgr->criticalSection);
	free(mgr);
	return r;
}

sndManager *sndMgrOpen(int outRate,
	int stereo,
	int *err,
	int maxStreamThreads)
{
	WAVEFORMATEX	wfmt;
 #ifndef WAVE_DUMP
	sndFormat	format;
 #endif
	const uint8	*h;
	sndMgrData	*mgr = NULL;
	WAVEHDR		*hdr;
	int32		*d;
	MMRESULT	mmr;
	intptr_t	r =-1;
	int		i;

	memset(&format,0,sizeof(format));
	format.chanCount = stereo ? 2:1;
	format.frameCount = sndEndFrameCount;
	format.rate = outRate<=0 ? defaultOutRate:PIN(outRate,1000,100000);
	(void)sndFixCmpType(CODEC_PCM_LE_S16,&format);
	if ((h=setupWAVEHeader(&format,lastFmtSize))==NULL)
		goto end;
	memcpy(&wfmt,h+20,sizeof(wfmt)); /* assuming native LE... */
	wfmt.cbSize = sizeof(WAVEFORMATEX); //0;
	/* allocate */
#ifdef __MT__
	maxStreamThreads = PIN(maxStreamThreads,0,65535);
#else
	maxStreamThreads = 0;
#endif
	if ((mgr=calloc(sizeof(sndMgrData)+sizeof(mgr->tlist)*(maxStreamThreads-1),1))==NULL)
		{r =-2; goto end;}
	mgr->rflag = format.rate|(stereo ? -1<<24:~sndInitStereo<<24);
	/* sndMgrClose needs this created even on failure */
	InitializeCriticalSection(&mgr->criticalSection);

	if ((mmr=waveOutOpen(&mgr->hWAudio,WAVE_MAPPER,&wfmt,
 #if (defined(__MT__) && !defined(NO_SEMAPHORE)) || defined(QUEUED_WAVE)
			(DWORD_PTR)sndWaitSync,(DWORD_PTR)mgr,CALLBACK_FUNCTION
 #else
			0,(DWORD_PTR)mgr,CALLBACK_NULL
 #endif
			))!=MMSYSERR_NOERROR)
	{
		if (mmr==MMSYSERR_BADDEVICEID || mmr==MMSYSERR_NODRIVER)
			r =-32767;
		else if (mmr==MMSYSERR_NOMEM)
			r =-2;
		goto close;
	}
 #if defined(__MT__) && !defined(NO_SEMAPHORE)
	if ((mgr->hSemaphore=CreateSemaphore(NULL,sndBufferCount-1,sndBufferCount,NULL))==NULL)
		goto close;
 #endif

	i = 0, hdr = mgr->hWHeader, d = *mgr->out;
	do
	{
		hdr->lpData = (LPSTR)d;
		hdr->dwBufferLength = sndEndFrameCount<<format.chanCount;
		mmr = waveOutPrepareHeader(mgr->hWAudio,hdr,sizeof(*hdr));
		hdr->dwFlags|= WHDR_DONE;
		if (mmr!=MMSYSERR_NOERROR)
		{
		close:	(void)sndMgrClose((sndManager*)mgr);
			mgr = NULL;
			goto end;
		}
	} while(++hdr,d+=sndEndFrameCount*2,++i<sndBufferCount);
 #ifdef WAVE_DUMP
	if ((fwout=fopen("fwout.wav","wb"))!=NULL)
		_write(fwoutn=_fileno(fwout),h,_lastFmtSize);
 #endif
 #ifdef __MT__
	if ((r=_beginthread(sndMgrThread,0,mgr))==-1)
		goto end;
	mgr->thread = (HANDLE)r;
	(void)SetThreadPriority((HANDLE)r,
		outRate>=96000 ? THREAD_PRIORITY_HIGHEST:THREAD_PRIORITY_ABOVE_NORMAL);

	if (maxStreamThreads>0)
	{
		mgr->tmaximum = MIN(maxStreamThreads,65535);
		mgr->ticEID = timeSetEvent(1,1,streamsThreadFunc,(DWORD_PTR)mgr,TIME_PERIODIC);
	}
 #endif
	r = 0;
end:	if (err)
		*err = r;
	return (sndManager*)mgr;
}

struct sndChannel *sndMgrLeadChannel(sndManager *smgr)
{
	sndMgrData	*mgr;

	if ((mgr=(sndMgrData*)smgr)==NULL || !(mgr->rflag&0x80000000u))
		return NULL;
	return mgr->sndChannels;
}

 #undef CreateSemaphore
 #undef ReleaseSemaphore
 #undef CloseObject
 #undef WaitForSemaphore
#endif /* defined(WINUSERAPI) */
