/*
 sndmixer.c
  sample mixing code, stereo output

  Created: 2014.01.26 09:29 EST
  Revised: 2024.04.16
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.03.23
  interpolation end-sample problem solved: writes final sample to front.
2014.03.25
  mix routines no longer return remaining count;
all length clipping is done with sndMixSamples now.
  more 1.0 volume detection (for performance).
*/

#include <stdlib.h>
#include <string.h>
#include "my_cseries.h"
#include "snddec.h"
#include "ftrig.h"

/* ---- typedefs */

typedef struct lri16
{
	int16		l, r;
} lri16;

/* ---- macros */

#define monoMix(v,u,l,r)	((v)*(l)+(u)*(r))>>9

/* ---- code */

/*
 Process:
  1.	Decode (via decompressSndFrame).
  2.	Conform buffer to 16-bit samples.
  3.	Add to output, performing rate conversion & L/R amp. if necessary.
  4.    Repeat steps 1-3 until output block or cmpSamples are finished.

  Caller would make the mixing calls, and clip the result (s32to16).
*/

/* self-depth conversion routines */

/* to-larger overwrite: works backward from end */
#if defined(__POCC__) && envx86
__declspec(naked)
#endif
static size_t wrChannelu8(const void *p,
	size_t z,
	void *stream)
{
#if defined(__POCC__) && envx86
	__asm
	{
		push	esi
		mov	ecx,[esp+12]
		mov	esi,[esp+16]
		mov	esi,[esi]
		lea	edx,[esi+ecx*2]
		add	esi,ecx
		shr	ecx,1
		jnc	ao

		dec	esi
		mov	al,[esi]
		mov	ah,al
		xor	ax,0x8000
		sub	edx,2
		mov	[edx],ax
		test	ecx,ecx
		jz	eo

	ao:	sub	esi,2
		mov	ax,[esi]
		mov	al,ah
		shl	eax,16
		mov	al,[esi]
		mov	ah,al
		xor	eax,0x80008000
		sub	edx,4
		mov	[edx],eax
		dec	ecx
		jnz	ao

	eo:	mov	eax,[esp+12]
		add	eax,eax
		pop	esi
		ret
	}
#else
	const uint8	*s;
	int16		*d = *(int16**)stream;
	int		c;
	unsigned	v;
(void)p;
	s = (uint8*)d+z, d+= z;
	if ((c=z>>1)!=0)
		do
		{
			v = *--s, *--d = (int16)((v|v<<8)^0x8000);
			v = *--s, *--d = (int16)((v|v<<8)^0x8000);
		} while(--c);
	if (z&1)
		v = s[-1], d[-1] = (int16)((v|v<<8)^0x8000);
	/* No need to update the stream; sndMixSamples only makes one call;
	  so does writeSndSamples; in-stack downsampling only occurs for 16-bit.
	*/
	return z+z;
#endif
}

/* ---- mixer routines */

/* rate conversion */

#define endRateCnv \
	m->pos=(uint32)((const uint8*)s-m->dcmpBuf), m->cnvFrc=(uint16)(rf);
#define startRateCnv(t) \
	const t*s=p; \
	uint32 rt=((uint16*)&m->cnvRate)[!BIG_ENDIAN]; \
	uint32 rl=((uint16*)&m->cnvRate)[BIG_ENDIAN]; \
	uint32 rf=m->cnvFrc; \
	int32 v, lv=m->LRAmp;

/* -- interpolation: linear (default) */

static __forceinline int lerp(int m,
	int a,
	int b)
{
	return a+((b-a)*m>>15);
}

/* cubic (not used) */
/* there needs to be always at least 4 samples loaded for this */
static __forceinline double cerp(double x,
	int a,
	int b,
	int c,
	int d)
{
	return b + 0.5 * x*(c - a + x*(2*a - 5*b + 4*c - d
		+ x*(3*(b - c) + d - a)));
}

/* stereo source */

#if envx86
extern DEF_MIX_RTN(rateCnvIntrp2);
extern DEF_MIX_RTN(rateCnvIntrp2m);
#else
static DEF_MIX_RTN(rateCnvIntrp2)
{
	startRateCnv(lri16)
	if (lv==0x1000100)
		do
		{
			v = rf>>1;
			*d+= lerp(v,s->l,s[1].l), d[1]+= lerp(v,s->r,s[1].r);
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = rf>>1;
			*d+= lerp(v,s->l,s[1].l)*lv>>8, d[1]+= lerp(v,s->r,s[1].r)*rv>>8;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

static DEF_MIX_RTN(rateCnvIntrp2m)
{
	startRateCnv(lri16)
	if (lv==0x1000100)
		do
		{
			v = rf>>1;
			v = (lerp(v,s->l,s[1].l)+lerp(v,s->r,s[1].r))>>1;
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = rf>>1;
			v = monoMix(lerp(v,s->l,s[1].l),lerp(v,s->r,s[1].r),lv,rv);
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}
#endif

/* mono source */

static DEF_MIX_RTN(rateCnvIntrp1)
{
	startRateCnv(int16)
	if (lv==0x1000100)
		do
		{
			v = lerp(rf>>1,*s,s[1]);
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = lerp(rf>>1,*s,s[1]);
			*d+= v*lv>>8, d[1]+= v*rv>>8;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

static DEF_MIX_RTN(rateCnvIntrp1m)
{
	startRateCnv(int16)
	if ((lv=((lv>>16)+(int16)lv)>>1)==0x100)
		do
		{
			v = lerp(rf>>1,*s,s[1]);
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
		do
		{
			v = lerp(rf>>1,*s,s[1])*lv>>8;
			*d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	endRateCnv
}

/* -- filtered */

static DEF_MIX_RTN(rateCnvIntrf2)
{
	startRateCnv(lri16)
	double		ff, fl = m->filterL*(-1/4294967296.)+1, fsl, fsr, fvl, fvr;
(void)v;
	fsl = m->filterS[0], fsr = m->filterS[1];
	fvl = (int16)lv*(1./256), fvr = (int16)(lv>>16)*(1./256);
	do
	{
		ff = rf*(1./65536);
		fsl = (cerp(ff,s->l,s[1].l,s[2].l,s[3].l)-fsl)*fl+fsl;
		fsr = (cerp(ff,s->r,s[1].r,s[2].r,s[3].r)-fsr)*fl+fsr;
		*d+= fsl*fvl, d[1]+= fsr*fvr;
		rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
	} while(d+=2,--c);
	m->filterS[0] = fsl, m->filterS[1] = fsr;
	endRateCnv
}

static DEF_MIX_RTN(rateCnvIntrf2m)
{
	startRateCnv(lri16)
	double		ff, fl = m->filterL*(-1/4294967296.)+1, fsl, fsr, fvl, fvr;

	fsl = m->filterS[0], fsr = m->filterS[1];
	fvl = (int16)lv*(.5/256), fvr = (int16)(lv>>16)*(.5/256);
	do
	{
		ff = rf*(1./65536);
		fsl = (cerp(ff,s->l,s[1].l,s[2].l,s[3].l)-fsl)*fl+fsl;
		fsr = (cerp(ff,s->r,s[1].r,s[2].r,s[3].r)-fsr)*fl+fsr;
		v = fsl*fvl+fsr*fvr;
		*d+= v, d[1]+= v;
		rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
	} while(d+=2,--c);
	m->filterS[0] = fsl, m->filterS[1] = fsr;
	endRateCnv
}

/* mono source */

static DEF_MIX_RTN(rateCnvIntrf1)
{
	startRateCnv(int16)
	double		ff, fl = m->filterL*(-1/4294967296.)+1, fsl, fvl, fvr;
(void)v;
	fsl = m->filterS[0];
	fvl = (int16)lv*(1./256), fvr = (int16)(lv>>16)*(1./256);
	do
	{
		ff = rf*(1./65536);
		fsl = (cerp(ff,s[0],s[1],s[2],s[3])-fsl)*fl+fsl;
		*d+= fsl*fvl, d[1]+= fsl*fvr;
		rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
	} while(d+=2,--c);
	m->filterS[0] = fsl;
	endRateCnv
}

static DEF_MIX_RTN(rateCnvIntrf1m)
{
	startRateCnv(int16)
	double		ff, fl = m->filterL*(-1/4294967296.)+1, fsl, fvl;

	fsl = m->filterS[0];
	fvl = (int16)lv*(.5/256)+(int16)(lv>>16)*(.5/256);
//printf("fsl %f\n",fl);
	do
	{
		ff = rf*(1./65536);
		fsl = (cerp(ff,s[0],s[1],s[2],s[3])-fsl)*fl+fsl;
		v = fsl*fvl;
		*d+= v, d[1]+= v;
		rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
	} while(d+=2,--c);
	m->filterS[0] = fsl;
	endRateCnv
}

/* -- no interpolation */

/* stereo source */

static DEF_MIX_RTN(rateCnvNoIntrp2)
{
	startRateCnv(lri16)
	if (lv==0x1000100)
		do
		{
			*d+= s->l, d[1]+= s->r;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		v = (int16)lv, lv>>= 16;
		do
		{
			*d+= s->l*lv>>8, d[1]+= s->r*v>>8;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

static DEF_MIX_RTN(rateCnvNoIntrp2m)
{
	startRateCnv(lri16)
	if (lv==0x1000100)
		do
		{
			v = (s->l+s->r)>>1, *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = monoMix(s->l,s->r,lv,rv), *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

/* mono source */

static DEF_MIX_RTN(rateCnvNoIntrp1)
{
	startRateCnv(int16)
	if (lv==0x1000100)
		do
		{
			v = *s, *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
	{
		int32 rv=(int16)lv; lv>>= 16;
		do
		{
			v = *s, *d+= v*lv>>8, d[1]+= v*rv>>8;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	}
	endRateCnv
}

static DEF_MIX_RTN(rateCnvNoIntrp1m)
{
	startRateCnv(int16)
	if ((lv=((lv>>16)+(int16)lv)>>1)==0x100)
		do
		{
			v = *s, *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	else
		do
		{
			v = *s*lv>>8, *d+= v, d[1]+= v;
			rf+= rl, s+= rt+(rf>>16), rf&= 0xFFFF;
		} while(d+=2,--c);
	endRateCnv
}

/* no rate conversion */

#define startOneCnv(h) \
	const int16 *s=p; \
	int32 v, lv=m->LRAmp; \
	m->pos+=c<<h;

static DEF_MIX_RTN(mixNoRateCnv2)
{
	startOneCnv(2)
	if (lv==0x1000100)
	{
		if (c&1)
			*d+= *s, d[1]+= s[1], s+= 2, d+= 2;
		if (c>>=1)
			do
				*d+= *s, d[1]+= s[1], d[2]+= s[2], d[3]+= s[3];
			while(s+=4,d+=4,--c);
	} else if (lv)
	{
		v = (int16)lv, lv>>= 16;
		do
			*d+= *s*lv>>8, d[1]+= s[1]*v>>8;
		while(s+=2,d+=2,--c);
	}
}

static DEF_MIX_RTN(mixNoRateCnv2m)
{
	startOneCnv(2)
	if (lv==0x1000100)
	{
		if (c&1)
			v = (*s+s[1])>>1, *d+= v, d[1]+= v, s+= 2, d+= 2;
		if (c>>=1)
			do
			{
				v = (*s+s[1])>>1, *d+= v, d[1]+= v;
				v = (s[2]+s[3])>>1, d[2]+= v, d[3]+= v;
			} while(s+=4,d+=4,--c);
	} else if (lv)
	{
		int32		rv = (int16)lv;

		lv>>= 16;
		do
			v = monoMix(*s,s[1],lv,rv), *d+= v, d[1]+= v;
		while(s+=2,d+=2,--c);
	}
}

static DEF_MIX_RTN(mixNoRateCnv1)
{
	startOneCnv(1)
	if (lv==0x1000100)
	{
		if (c&1)
			v = *s, *d+= v, d[1]+= v, ++s, d+= 2;
		if (c>>=1)
			do
			{
				v = *s, *d+= v, d[1]+= v;
				v = s[1], d[2]+= v, d[3]+= v;
			} while(s+=2,d+=4,--c);
	} else if (lv)
	{
		int32		rv = (int16)lv;

		lv>>= 16;
		do
			v = *s, *d+= v*lv>>8, d[1]+= v*rv>>8;
		while(++s,d+=2,--c);
	}
}

static DEF_MIX_RTN(mixNoRateCnv1m)
{
	startOneCnv(1)
	if ((lv=((lv>>16)+(int16)lv)>>1)==0x100)
	{
		if (c&1)
			v = *s, *d+= v, d[1]+= v, ++s, d+= 2;
		if (c>>=1)
			do
			{
				v = *s, *d+= v, d[1]+= v;
				v = s[1], d[2]+= v, d[3]+= v;
			} while(s+=2,d+=4,--c);
	} else if (lv)
		do
			v = *s*lv>>8, *d+= v, d[1]+= v;
		while(++s,d+=2,--c);
}

/* the decompression and mixer routine manager */

static TYP_MIX_RTN(cnvRtnList[sndInitStereo>>2<<1]) =
{
	[0] = rateCnvIntrp1m,
	[4] = rateCnvIntrp2m,
	[sndInitNoInterpolation>>2|0] = rateCnvNoIntrp1m,
	[sndInitNoInterpolation>>2|4] = rateCnvNoIntrp2m,
	[sndInitExtendedFilter>>2|0] = rateCnvIntrf1m,
	[sndInitExtendedFilter>>2|4] = rateCnvIntrf2m,
	[sndInitFilterMask>>2|0] = rateCnvIntrf1m,
	[sndInitFilterMask>>2|4] = rateCnvIntrf2m,
	[sndInitStereo>>2|0] = rateCnvIntrp1,
	[sndInitStereo>>2|4] = rateCnvIntrp2,
	[sndInitStereo>>2|sndInitNoInterpolation>>2|0] = rateCnvNoIntrp1,
	[sndInitStereo>>2|sndInitNoInterpolation>>2|4] = rateCnvNoIntrp2,
	[sndInitStereo>>2|sndInitExtendedFilter>>2|0] = rateCnvIntrf1,
	[sndInitStereo>>2|sndInitExtendedFilter>>2|4] = rateCnvIntrf2,
	[sndInitStereo>>2|sndInitFilterMask>>2|0] = rateCnvIntrf1,
	[sndInitStereo>>2|sndInitFilterMask>>2|4] = rateCnvIntrf2,

	[8|0] = mixNoRateCnv1m,
	[8|4] = mixNoRateCnv2m,
	[8|sndInitNoInterpolation>>2|0] = mixNoRateCnv1m,
	[8|sndInitNoInterpolation>>2|4] = mixNoRateCnv2m,
	[8|sndInitStereo>>2|0] = mixNoRateCnv1,
	[8|sndInitStereo>>2|4] = mixNoRateCnv2,
	[8|sndInitExtendedFilter>>2|0] = rateCnvIntrf1m,
	[8|sndInitExtendedFilter>>2|4] = rateCnvIntrf2m,
	[8|sndInitFilterMask>>2|0] = rateCnvIntrf1m,
	[8|sndInitFilterMask>>2|4] = rateCnvIntrf2m,
	[8|sndInitNoInterpolation>>2|sndInitStereo>>2|0] = mixNoRateCnv1,
	[8|sndInitNoInterpolation>>2|sndInitStereo>>2|4] = mixNoRateCnv2,
	[8|sndInitStereo>>2|sndInitExtendedFilter>>2|0] = rateCnvIntrf1,
	[8|sndInitStereo>>2|sndInitExtendedFilter>>2|4] = rateCnvIntrf2,
	[8|sndInitStereo>>2|sndInitFilterMask>>2|0] = rateCnvIntrf1,
	[8|sndInitStereo>>2|sndInitFilterMask>>2|4] = rateCnvIntrf2,
};

void sndUpdateMixerRtn(sndMixerFrame *m,
	uint32 rflag)
{
	m->cnvRate = ufix32muldiv(m->format.rate,m->format.rateFrac,
		m->rateMul,rflag&0xFFFFFF);
	m->cnvRate+= !m->cnvRate; /* zero reserved for non-conversion */
	m->mixerRtn = cnvRtnList[(uint8)(((m->format.chanCount&127)>1)<<2)
		|(uint8)((m->cnvRate==0x10000)<<3)
		|(rflag>>26&(sndInitFilterMask|sndInitStereo)>>2)];
}

int sndMixSamples(int32 *d,
	sndMixerFrame *m,
	uint32 rflag,
	int outReq,
	int rem)
{
	uint32		z;
	uint8		st;

	if (!m->dcmpBuf || !m->format.blockSize)
		return -1;
	st = ((m->format.chanCount&127)>1)+1;
	if (!m->pos)
	{ /* decompression phase */
		uint8		*p, *s = m->dcmpSize+m->dcmpBuf, *a = s;
		const uint8	*e;
		size_t		k = (!m->cmpPos)-1;

		for(e=a+(outReq<<(m->format.chanCount&127)); a<e; m->cmpPos+=z&k)
		{
			z = m->format.blockSize*MIN(m->format.frameCount,
				m->format.maxRead);
			if ((p=decompressSndFrame(a,m->cmpPos,&m->format,z))<=a)
			{
				if (!p)
					m->err|= 0x80;
				if (!m->cmpPos)
					m->err|= m->err*2+1;
				m->cmpPos = NULL;
				break;
			}
			m->err&= 127;
			a = p;
		}
		/* sndwrite_null if already in native s16 format */
		/* endian conversion is done in decompression; no need to check that */
		m->dcmpSize+= writeSndSamples(s,(long)(a-s),&s,&m->format,
			STYPE_STDMAX16,m->format.mSampleSize==2 ? sio_null
			:m->format.mSampleSize==1 ? wrChannelu8:swr_memcpy); 
	}
	/* resampling phase, including mono/stereo and rate conversion */
	if (!m->mixerRtn)
		sndUpdateMixerRtn(m,rflag);
	/* Must clip the count, adjusting with the rate, so the last read is in bounds. */
	/* This must be done so the interpolation routines get the correct samples. */
	/* This can also be a tricky process, but using division is faster than conditional
	   checking per iteration, as used previously.
	*/
	z = (m->dcmpSize-m->pos)>>st;
	if (m->cnvRate)
		z = rfix16div(z,m->cnvRate,m->filterR); /* required: one whole + any fraction */
	if (rem>z)
	{
		int32	dif = rem-z;

		rem = z, z = dif;
	} else
		z = 0;
	if (rem>0)
	{
		m->sampElapsed-= m->pos;
		m->mixerRtn(d,m->pos+m->dcmpBuf,m,rem);
		m->sampElapsed+= m->pos;
	}
	rem = z;
	if (m->pos>=m->dcmpSize)
	{
		m->pos = m->dcmpSize = 0;
	} else if (z)
	{ /* must do a full copy due to possible rateMultiplier changes */
	  /* when FIR is implemented, the pre-pos will also be pushed back */
		m->dcmpSize-= m->pos;
		memcpy(m->dcmpBuf,m->dcmpBuf+m->pos,m->dcmpSize);
		m->pos = 0;
	}
	/* if delay forward needs more samples */
	//m->err|= m->dcmpSize<4*2; //(m->dcmpSize<m->filterSize)*2;
	return rem;
}

void sndMixFill(int c,
	uint32 rflag,
	int32 *d,
	const sndMixerFrame *m)
{
	int32		v, u, lv, rv;
	const int16	*s;

	if (!m->dcmpBuf || m->dcmpSize<2 || c<=0)
		return;
	v = m->pos ? m->pos:m->dcmpSize-2;
	s = (int16*)(m->dcmpBuf+v);
	lv = m->LRAmp, rv = (int16)lv, lv>>= 16;
	if (rflag&sndInitStereo<<24)
		if ((m->format.chanCount&127)>1)
			v = s[-1]*lv>>8, u = *s*rv>>8;
		else
			v = *s, u = v*rv>>8, v = v*lv>>8;
	else
	{
		if ((m->format.chanCount&127)>1)
			v = monoMix(s[-1],*s,lv,rv);
		else
			v = *s*((lv+rv)>>1)>>8;
		u = v;
	}
	if (!(v|u))
		return;
	if (c&1)
		*d+= v, d[1]+= u, d+= 2;
	if ((c>>=1)==0)
		return;
	do
		*d+= v, d[1]+= u, d[2]+= v, d[3]+= u;
	while(d+=4,--c);
}

/* clipping */

#if defined(__POCC__) && envx86
__declspec(naked)
#endif
void __fastcall s32to16(int32 c,
	uint32 rflag,
	int32 *b)
{
#if defined(__POCC__) && envx86
	__asm
	{
		push	esi
		push	edi
		test	ecx,ecx
		jle	eo

		mov	esi,[esp+12]	;+16]
		mov	edi,esi
	#if sndInitStereo==0x40
		add	edx,edx
		jns	mo
	#else
		test	edx,sndInitStereo<<24
		jz	mo
	#endif
	so:	mov	eax,[esi]
	;	movsx	ebx,ax
	;	cdq
	;	xor	dx,0x7FFF
	;	cmp	eax,ebx
	;	cmovnz	eax,edx		;686+

		movsx	edx,ax
		cmp	eax,edx
		cdq
		jz	s1

		mov	eax,edx
		xor	eax,0x7FFF
	s1:	mov	[edi],ax
		mov	eax,[esi+4]
		movsx	edx,ax
		cmp	eax,edx
		cdq
		jz	s2

		mov	eax,edx
		xor	eax,0x7FFF
	s2:	mov	[edi+2],ax
		add	esi,8
		add	edi,4
		dec	ecx
		jnz	so

		pop	edi
		pop	esi
		ret	4

	mo:	mov	eax,[esi]
		movsx	edx,ax
		cmp	eax,edx
		cdq
		jz	m1

		mov	eax,edx
		xor	eax,0x7FFF
	m1:	mov	[edi],ax
		add	esi,8
		add	edi,2
		dec	ecx
		jnz	mo

	eo:	pop	edi
		pop	esi
		ret	4
	}
#else
	int16		*d;
	int32		v;

	if (c<=0)
		return;
	d = (int16*)b;
	if (rflag&sndInitStereo<<24)
		do
		{
			v = *b, *d = (int16)(v==(int16)v ? v:v>>31^0x7FFF);
			v = b[1], d[1] = (int16)(v==(int16)v ? v:v>>31^0x7FFF);
		} while(b+=2,d+=2,--c);
	else /* already mono - just take one and skip the other */
		do
			v = *b, *d++ = (int16)(v==(int16)v ? v:v>>31^0x7FFF);
		while(b+=2,--c);
#endif
}
