/*
 sndparse.c
  parsing, identification

  Revised: 2022.05.23
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2015.06.14
  Split from "snddec.c".
2021.04.21
  Added more codec types.
*/

#include <enviro64.h>
#include <string.h>
#include "snddec.h"

/* ---- globals */

const static int32	cmpID2type[7] =
{
	0,
	CODEC_ACE_2,CODEC_ACE_8,
	CODEC_MACE_3,CODEC_MACE_6,
	CODEC_CDXA_2,CODEC_CDXA_4
};

const static uint32	noteTable31nr[12] =
{
	0x00000000,0x0F38F92D,0x1F59AC3C,0x306FE0A3,
	0x428A2F98,0x55B8108F,0x6A09E667,0x7F910D76,
	0x965FEA54,0xAE89F996,0xC823E075,0xE3437E72
};

/* ---- code */

uint32 sndMIDINoteToRate(uint8 note,
       uint8 baseNote)
{
	unsigned	expfreq = 264+note-baseNote;
	int		den = expfreq/12u;
	int		rem = expfreq-den*12u;

	den-= 22+15;
	if ((den=-den)<0)
		return ~0;
	if (den>31 || !note)
		return (den==32)&(!!note);
	return (((noteTable31nr[rem]>>den)+1)>>1)+(0x80000000u>>den);
}

int sndFixCmpType(uint32 c,
	sndFormat *f)
{
	int		fail = 0;

	if (!f)
		return -1;
	/* first, check for alternate types */
	if (c==CODEC_MS_ALAW || c==MK4CC('A','L','A','W'))
		{c = CODEC_ALAW; goto alaw;}
	else if (c==CODEC_MS_ULAW || c==MK4CC('U','L','A','W'))
		{c = CODEC_ULAW; goto alaw;}
	else if (!c)
	{
		if (f->depth<=0) /* get appropriate AIFF-C type */
			f->depth = 16, c = CODEC_PCM_BE_S16;
		else
			c = CODEC_PCM;
		f->type&=~15, f->type|= CTYPE_PCM;
		goto idpack;
	} else
		++fail;
	/* now make sure the type, depth, packet, etc. values are correct */
	if (c==CODEC_PCM || c==CODEC_PCM_DEFAULT)
		f->type&=~15, f->type|= CTYPE_PCM;
	else if (c==CODEC_ALAW || c==CODEC_ULAW)
	alaw: f->frameSize = 1, f->packSize = 1, f->depth = 12, f->type = CTYPE_NPCM;
	else if (c==CODEC_MACE_3)
		f->frameSize = 6, f->packSize = 2, f->depth = 9, f->type = CTYPE_DPCM;
	else if (c==CODEC_MACE_6)
		f->frameSize = 6, f->packSize = 1, f->depth = 10, f->type = CTYPE_DPCM;
	else if (c==CODEC_ACE_2)
		f->frameSize = 2, f->packSize = 1, f->depth = 8, f->type = CTYPE_ADPCM;
	else if (c==CODEC_ACE_8)
		f->frameSize = 8, f->packSize = 3, f->depth = 8, f->type = CTYPE_ADPCM;
	else if (c==CODEC_DVI_IMA)
	{
		if (!f->packSize)
			f->packSize = 1024;
		if (!f->frameSize)
			f->frameSize = f->packSize*2-7; /* WAVE: packSize*2-7 */
		f->depth = 16, f->type = STYPE_LSB|CTYPE_ADPCM;
	} else if (c==CODEC_AIFC_IMA)
		f->frameSize = 64, f->packSize = 34, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_CDXA_4)
		f->frameSize = 28, f->packSize = 15, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_CDXA_2) /* unsure about this */
		f->frameSize = 2, f->packSize = 2, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_PCM_BE_S24)
		f->depth = 24;
	else if (c==CODEC_PCM_BE_S32)
		f->depth = 32;
	else if (c==CODEC_PCM_LE_S32)
		f->type = STYPE_LSB, f->depth = 32;
	else if (c==CODEC_PCM_BE_F32)
		f->type = STYPE_FP, f->depth = 32;
	else if (c==CODEC_PCM_LE_F32)
	{
		f->type = (f->type&STYPE_ISGN)|STYPE_FP|STYPE_LSB;
		f->depth = 32;
	} else if (c==CODEC_PCM_BE_F64)
		f->type = STYPE_FP, f->depth = 64;
	else if (c==CODEC_PCM_LE_F64)
		f->type = STYPE_FP|STYPE_LSB, f->depth = 64;
	else if (c==CODEC_PCM_BE_S16)
	{
		if (!f->depth) // f->depth!=8
			f->depth = 16;
	} else if (c==CODEC_PCM_LE_S16)
	{
		f->type = STYPE_LSB;
		if (!f->depth) // f->depth!=8
			f->depth = 16;
	} else if (c==CODEC_MS_ADPCM /* Microsoft ADPCM - usu. 1012/512 in WAVEs */
			|| c==MK4CC('m','s',0,225)) /* really ?? */
	{
		c = CODEC_MS_ADPCM;
		f->depth = 16;
		f->type = STYPE_LSB|CTYPE_ADPCM;
		if (!f->packSize)
			f->packSize = 256;
		f->frameSize = f->packSize*2-12;
	} else if (c==MK4CC('F','L','3','2')) /* for CSound, SoundHack, etc. */
		c = CODEC_PCM_BE_F32, f->type = STYPE_FP, f->depth = 32;
	else if (c==CODEC_OKI_ADPCM)
		f->frameSize = 512, f->packSize = 256, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_ADPCM_CT)
	{
		if (!f->packSize)
			f->packSize = 1;
		f->frameSize = 2*f->packSize, f->depth = 16, f->type = CTYPE_ADPCM;
	} else if (c==CODEC_AAC || c==CODEC_AAC_ADTS || c==CODEC_PHILIPS_CELP
			|| c==CODEC_ACELP_NET || c==CODEC_SIPRO_KELVIN)
	{
		if (!f->depth)
			f->depth = 16;
		f->type = CTYPE_DYNAMIC;
	} else if (c==MK4CC('m','s',22,0))
		c = CODEC_AAC_ADTS, f->type = CTYPE_DYNAMIC;
	else if (c==CODEC_MEDIASPACE || c==CODEC_YAMAHA_ADPCM)
		f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_G726_ADPCM || c==MK4CC('m','s',0,100))
		c = CODEC_G726_ADPCM, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_GSM /* MS packSize=65; MSN=41 */
			|| c==CODEC_MS_GSM_610)
		f->packSize = 33, f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==MK4CC('m','s',1,80)
			|| c==CODEC_QUALCOMM_PV)
	{
		c = CODEC_QUALCOMM_PV;
		f->packSize = 35, f->depth = 16, f->type = 15;
	} else if (c==CODEC_MPEG_L3 || c==CODEC_MPEG_L3P /* an MPEG reader is needed for this */
			|| c==CODEC_DOLBY_AC2 || c==CODEC_AC3
			|| c==CODEC_WMA1 || c==CODEC_WMA2)
		f->depth = 16, f->type = CTYPE_DYNAMIC;
	else if (c==MK4CC('m','p','4','a'))
		f->type = CTYPE_DYNAMIC, c = CODEC_AAC;
	else if (c==CODEC_ADP4)
		f->depth = 16, f->type = CTYPE_ADPCM;
	else if (c==CODEC_MS_WMA1)
		c = CODEC_WMA1;
	else if (c==CODEC_MS_WMA2)
		c = CODEC_WMA2;
	else if (c==MK4CC('m','s',0,116) /* some less official codes... */
			|| c==MK4CC('m','s',24,28))
	{
		c = CODEC_VOXWARE_RT24;
		f->type = CTYPE_DYNAMIC;
		if (!f->depth)
			f->depth = 16;
	} else if (c==MK4CC('m','s',0,117)) /* also, Meta-Sound */
		c = CODEC_VOXWARE_RT29;
	else if (c==MK4CC('m','s',0xF1,0xAC))
		c = CODEC_FLAC;
	else if (c==MK4CC('m','s',161,9))
		c = CODEC_SPEEX;
	else if (c==CODEC_CJ_LIGHTWAVE || c==CODEC_WMALOSSLESS || c==CODEC_ALAC)
	{
		if (!f->depth)
			f->depth = 16;
		f->type = c==CODEC_CJ_LIGHTWAVE ? 15:CTYPE_PACKED;
	} else if (c==MK4CC('m','s','l','u'))
		{c = CODEC_ULAW; goto alaw;}
	else if (c==CODEC_RHETOREX_ADPCM /* 8K */
			|| c==CODEC_UHER_ADPCM
			|| c==CODEC_SANYO_LDADPCM || c==CODEC_UNISYS_NAP_ADPCM
			|| c==CODEC_GSM_MS || c==CODEC_TRUESPEECH_X /* ? */
			|| c==CODEC_TRUESPEECH) /* ? */
	{
		f->type = CTYPE_ADPCM;
		f->depth = 16;
	} else if (c==CODEC_OPUS)
	{
		f->depth = 16;
		f->type = CTYPE_DYNAMIC;
		f->frameSize = 960; /* default */
	} else if (c==MK4CC('f','L','a','C') || c==CODEC_FLAC)
	{
		if (!f->depth)
			f->depth = 16;
		c = CODEC_FLAC;
		f->type = CTYPE_PACKED;
	} else if (c==CODEC_ATRAC3)
	{
		if (!f->depth)
			f->depth = 16;
		f->type = CTYPE_DYNAMIC;
	} else if (c==CODEC_SBPRO_ADPCM_4 || c==CODEC_SBPRO_ADPCM_3
			|| c==CODEC_SBPRO_ADPCM_2)
	{
		f->depth = 8;
		f->type = CTYPE_ADPCM;
		f->packSize = 2; /* first byte the reference sample */
		f->frameSize = c==CODEC_SBPRO_ADPCM_4 ? 2*2
			:c==CODEC_SBPRO_ADPCM_3 ? 3*2
			:4*2;
	} else if (c==CODEC_CU_HP)
	{
		f->depth = 16;
		f->type = CTYPE_ADPCM;
	} else
		++fail;
	if (f->depth<8)
		if (c==CODEC_PCM_U8 || !c || c==CODEC_PCM_DEFAULT)
			f->depth = 8;
		else
		{ /* default: 16 bits */
			if (c!=CODEC_PCM_BE_S16 && c!=CODEC_PCM_LE_S16)
				return -8;
			f->depth = 16;
		}
idpack:	if (!f->packSize && (f->packSize=(f->depth+7)>>3)==0)
		f->packSize = 1;
	if (c==CODEC_MS_PCM)
		if (f->depth==8)
			c = CODEC_PCM_U8;
		else
			c = CODEC_PCM_LE_S16, f->type = STYPE_LSB;
		/* other depths are not WAVE-supported */
	else if (c==CODEC_MS_FLOAT)
	{ /* isgn used for fixed-point-adjusted fp */
		f->type = (f->type&STYPE_ISGN)|STYPE_FP|STYPE_LSB;
		if (f->depth==64)
			c = CODEC_PCM_LE_F64;
		else
			c = CODEC_PCM_LE_F32;
	} else if (fail>1 && c!=CODEC_PCM_U8 && c!=CODEC_PCM_DEFAULT)
		f->type|= 15;
	f->codec = c;
	f->dcmpRtn = NULL;
	return fail;
}

void sndFixCodecFrame(sndFormat *f)
{
	int		a, b;

	if (!f)
		return;
	if (f->codec==CODEC_PCM_U8)
		f->type = STYPE_ISGN;
	if (!f->blockSize)
		f->blockSize = f->packSize*(f->chanCount&127);
	if (!f->frameSize)
		f->frameSize = 1;
	for(a=(f->depth+7)>>3,b=1; b<a; )
		b+= b;
	f->mSampleSize = (uint8)b;
	f->frameSize*= f->chanCount*b;
	f->maxRead = (f->type&15)>CTYPE_NPCM ? 1:512; //144
}

int32 parseBufferCmdHeader(const uint8 *s,
	sndFormat *f,
	int32 *loop,
	uint8 *baseNote)
{
	uint32		c, o = (uint32)-(int32)s;
	uint8		h;
	/* must read header type first */
	/* samplePtr should be NULL! */
	if (!s || !f || *(int32*)s)
		return -1;
	if ((h=s[20])-1u<0xFEu-1u)
		return -223;
	s+= 4;
	memset(f,0,sizeof(*f));
	if ((signed)(c=va_be32(s))<0)
		return -3; /* bit #31 is invalid for a stored bufferCmd */
	if (baseNote)
		*baseNote = s[21-8];
	if (!h)
	{ /* stdSH: count is #mono frames, uint8 */
		f->frameCount = c, f->codec = CODEC_PCM_U8;
		f->depth = 8, f->packSize = 1;
		c = 1;
	} /* extSH/cmpSH: count is #channels, frame count follows basenote */
	if (c-1u>1u)
		return -223;
	f->chanCount = (uint8)c;
	c = va_be32(s);
	/* rate: max(c,1Hz); QuickTime doesn't even allow rates below 1KHz */
	if (c<(1<<16))
		return -1;
	f->rate = c>>16; /* ignoring float80 rate, as Sound Manager 3.4 does */
	f->rateFrac = c<<16;
	if (loop)
	{
		int32		start = be32_native(*(int32*)s);
		int32		end = be32_native(((int32*)s)[1]);

		if ((start|end)<0 || start+16>=end)
			start = end = IMIN(end); /* range is n/a or bad */
		*loop = start;
		loop[1] = end;//+1;
	}
	s+= 10; /* loopStart.l/loopEnd.l/enc.b/baseNote.b */
	if (!h)
		goto end;
	f->frameCount = va_be32(s);
	s+= 10+4; /* AIFFRate.t/markerChunk.p */
	if (h==0xFF)
	{
		s+= 8; /* insChunks.p/AESRecording.p/reserved.. */
		//f->packSize = 0;
		c = 0; /* figure out type by depth */
	} else
	{
		c = *(int32*)s;
		s+= 4+14; /* reserved.l/stateVars.p/leftOverSamples.p/cmpID.w */
		if (!c)
		{ /* cmpID values: 0=none/implied */
		  /* -1=fixed (default MAC3 for 8-bit) */
		  /* -2=variable-sized */
			if ((signed)(c=(int16)be16_native(((const int16*)s)[-1]))<0)
				c = s[5]==8 ? 3:0;
			if (c>6)
				return -8; //-223
			c = cmpID2type[c];
		}
		f->packSize = be16_native(*(const int16*)s)>>3;
		s+= 4;
	}
	f->depth = s[1];
	/* without a codec, 8 bit is always unsigned */
	if ((!c || c==CODEC_PCM_DEFAULT) && f->depth==8)
		c = CODEC_PCM_U8;
	if (sndFixCmpType(c,f)<0 || *s)
		return -1; /* bad codec data or depth */
	s+= h==0xFF ? 16:2;
end:	sndFixCodecFrame(f);
	return (int32)o+(int32)s;
}

int parseSoundSampleHeader(const uint8 *s,
	sndFormat *f,
	int32 z)
{
	int8		stereo;

	if (!s || !f)
		return -1;
	if (z<=0)
		return -5;
	memset(f,0,sizeof(*f));
	stereo = !!(s[6]&15);
	if ((f->rate=s[9]<<8|s[8])==0)
		f->rate = 11127; /* default */
	f->chanCount = stereo ? 0x82:1; /* ordering reversed in IIgs AIFF-C files */
	f->frameCount = z>>stereo;
	(void)sndFixCmpType(CODEC_PCM_U8,f);
	sndFixCodecFrame(f);
	return 10;
}

int parseExt80Rate(const uint8 *s,
	sndFormat *f)
{
	uint64		v;
	uint32		a, b;
	int16		x;

	if (!s)
		return -1;
	x = *s<<8|s[1], s+= 2;
	a = va_be32(s), b = va_be32(s);
	/* normalize */
	if (!a)
	{
		if ((a=b)==0)
			if (x!=0x7FFF)
				goto zero;
			else /* INF */
				goto ovfl;
		b = 0, x-= 32;
	}
	int64sethl(v,a,b);
	if (!(~x&0x7FFF)) /* NaN */
		return 0;
	if (x<0) /* fixed-point value is unsigned */
		goto zero;
	for(; (signed)int64hi(v)>=0; --x)
		uint64add(v,v);
	x =-x, x+= 0x401E;
	if (x<0)
		{ovfl: f->rate = f->rateFrac =~0u; return 0;}
	else if (x>63u)
	zero:	int64xset(v,0); /* zero or underflow */
	uint64shr(v,x);
	f->rate = (uint32)int64hi(v), f->rateFrac = int64lo(v);
	return !!f->rate;
}

static const char *msName[] =
{
	NULL,
	[_be32_native(CODEC_MS_PCM)&255] = "PCM",
	[_be32_native(CODEC_MS_ADPCM)&255] = "MS ADPCM",
	[_be32_native(CODEC_MS_FLOAT)&255] = "floating point",
	[_be32_native(CODEC_MS_VSELP)&255] = "MS VSELP",
	[_be32_native(CODEC_IBM_CVSD)&255] = "IBM CVSD",
	[_be32_native(CODEC_MS_ALAW)&255] = "A-law",
	[_be32_native(CODEC_MS_ULAW)&255] = "u-law",
	[_be32_native(CODEC_MS_DTS)&255] = "MS DTS",
	[_be32_native(CODEC_MS_DRM)&255] = "MS DRM",
	[_be32_native(CODEC_OKI_ADPCM)&255] = "OKI",
	[_be32_native(CODEC_DVI_IMA)&255] = "DVI IMA 4:1",
	[_be32_native(CODEC_MEDIASPACE)&255] = "MediaSpace",
	[_be32_native(CODEC_SIERRA_ADPCM)&255] = "Sierra",
	[_be32_native(CODEC_G723_ADPCM)&255] = "Antex G723",
	[_be32_native(CODEC_DSP_DIGISTD)&255] = "DSP DigiStd",
	[_be32_native(CODEC_DSP_DIGIFIX)&255] = "DSP DigiFix",
	[_be32_native(CODEC_DIALOGIC_OKI)&255] = "Dialogic OKI",
	[_be32_native(CODEC_MEDIAVISION)&255] = "Media Vision", /* Media Vision, Inc. */
	[_be32_native(CODEC_CU_HP)&255] = "HP CU Mobile Voice",
	[_be32_native(CODEC_YAMAHA_ADPCM)&255] = "Yamaha",
	[_be32_native(CODEC_SONARC)&255] = "SC Sonarc",
	[_be32_native(CODEC_TRUESPEECH)&255] = "DSP TrueSpeech",
	[_be32_native(CODEC_ECHOSC1)&255] = "Echo Speech SC1",
	[_be32_native(CODEC_AUDIOFILE_AF36)&255] = "AudioFile AF36",
	[_be32_native(CODEC_APTX)&255] = "APTX",
	[_be32_native(CODEC_AUDIOFILE_AF10)&255] = "AudioFile AF10", /* Audiofile, Inc. */
	[_be32_native(CODEC_PROSODY_1612)&255] = "Aculab Prosody 1612",
	[_be32_native(CODEC_LRC)&255] = "MT LRC", /* Merging Technologies */
	[_be32_native(CODEC_DOLBY_AC2)&255] = "Dolby AC2",
	[_be32_native(CODEC_DOLBY_AC3_SPDIF)&255] = "Dolby AC3-S/PDIF",
	[_be32_native(CODEC_PROSODY_8K)&255] = "Aculab Prosody 8kbps",
	[_be32_native(CODEC_MEDIASONIC_G723)&255] = "MediaSonic G.723",
	[_be32_native(CODEC_MS_GSM_610)&255] = "MS GSM 6.10",
	[_be32_native(CODEC_MSNAUDIO)&255] = "MSN Audio", /* Microsoft Corp. */
	[_be32_native(CODEC_ANTEX_ADPCME)&255] = "Antex ADPCME",
	[_be32_native(CODEC_CONTROLRES_VQLPC)&255] = "CR VQLPC", /* Control Resources */
	[_be32_native(CODEC_DIGIREAL)&255] = "DSP DigiReal",
	[_be32_native(CODEC_DIGIADPCM)&255] = "DSP DigiADPCM",
	[_be32_native(CODEC_CONTROLRES_CR10)&255] = "CR 10", /* Control Resources */
	[_be32_native(CODEC_NMS_VBXADPCM)&255] = "NMS VBXADPCM",
	[_be32_native(CODEC_ROLAND_RDAC)&255] = "Roland RDAC",
	[_be32_native(CODEC_ECHOSC3)&255] = "Echo Speech Codec 3",
	[_be32_native(CODEC_ROCKWELL_ADPCM)&255] = "Rockwell ADPCM",
	[_be32_native(CODEC_ROCKWELL_DIGITALK)&255] = "Rockwell DigiTalk",
	[_be32_native(CODEC_XEBEC)&255] = "Xebec Multimedia",
	[139] = "G.721 Infocom ITS A/S",
	[_be32_native(CODEC_G721_ADPCM)&255] = "Antex G721",
	[_be32_native(CODEC_G728_CELP)&255] = "Antex G728 CELP",
	[_be32_native(CODEC_G723_1)&255] = "MS G723.1",
	[_be32_native(CODEC_CONVEDIA_G729)&255] = "Convedia G729",
	[_be32_native(CODEC_MS_MPEG)&255] = "MS MPEG",
	[_be32_native(CODEC_RT24)&255] = "InSoft RT24",
	[_be32_native(CODEC_PAC)&255] = "InSoft PAC",
	[_be32_native(CODEC_MPEG_L3)&255] = "MPEG Layer III",
	[_be32_native(CODEC_LUCENT_G723)&255] = "Lucent G723",
	[_be32_native(CODEC_CIRRUS)&255] = "Cirrus Logic",
	[_be32_native(CODEC_ESPCM)&255] = "ESS ESPCM",
	[100] = "G726",
	[_be32_native(CODEC_MS_G722_ADPCM)&255] = "APICOM G722",
	[_be32_native(CODEC_MS_DSAT)&255] = "MS DSAT", /* Microsoft Corp. */
	[_be32_native(CODEC_DSAT_DISPLAY)&255] = "MS DSAT Display", /* Microsoft Corp. */
	[116] = "Voxware RT24",
	[117] = "Voxware RT29",
	[_be32_native(CODEC_SOFTSOUND)&255] = "SoftSound",
	[122] = "Voxware SC3",
	[123] = "Voxware SC3",
	[_be32_native(CODEC_MS_RT24)&255] = "MS RT24",
	[_be32_native(CODEC_G729A)&255] = "AT&T G729A",
	[132] = "MVI MV12", /* Motion Pixels, USA */
	[133] = "DF G726", /* DataFusion Systems (Pty) (Ltd), S.Africa */
	[_be32_native(CODEC_DF_GSM610)&255] = "DataFusion GSM 610",
	[_be32_native(CODEC_ISIAUDIO)&255] = "ISIAudio",
	[_be32_native(CODEC_ONLIVE)&255] = "OnLive",
	[_be32_native(CODEC_MULTITUDE_FT_SX20)&255] = "Multitude FT SX20",
	[_be32_native(CODEC_SBC24)&255] = "SBC24",
	[_be32_native(CODEC_ZYXEL_ADPCM)&255] = "ZyXEL ADPCM",
	[_be32_native(CODEC_LPCBB)&255] = "Philips LPCBB",
	[153] = "Studer Pro Packed",
	[160] = "Phony Talk Audio", /* Malden Electronics */
	[161] = "Racal GSM", /* Racal Recorder GSM Audio */
	[162] = "Racal G720A",
	[163] = "Racal G723.1",
	[164] = "Racal ACELP", /* Racal Tetra ACELP Audio */
	[176] = "AAC",
	/* obsolete types (but still recognized) */
	/* [98] = "Voxware" */
	[105] = "VW Aligned",
	[112] = "VW AC8",
	[113] = "VW AC10",
	[114] = "VW AC16",
	[115] = "VW AC20",
	[118] = "VW RT29HW",
	[119] = "VW VR12",
	[120] = "VW VR18",
	[121] = "VW TQ40",
	[129] = "VW TQ60"
};

const char *sndGetCodecName(int32 c)
{
	uint32		msIDtest = be32_native(c)-('m'<<24|'s'<<16);

	if (msIDtest<GET_ARRAY_COUNT(msName))
		return msName[msIDtest];
	switch(c)
	{
	case 0: return "<none>";
	case CODEC_ALAW: return "A-law";
	case MK4CC('m','s','l','u'):
	case CODEC_ULAW: return "mu-law"; /* micro character difficult to make here */
	case CODEC_MACE_3: return "MACE 3:1";
	case CODEC_MACE_6: return "MACE 6:1";
	case CODEC_ACE_2: return "ACE 2:1";
	case CODEC_ACE_8: return "ACE 8:3";
	case CODEC_AIFC_IMA: return "IMA 4:1 (64:17)";
	case CODEC_CDXA_4: return "CD/XA 4:1";
	case CODEC_CDXA_2: return "CD/XA 2:1";
	case CODEC_DVCA: return "DV Audio";
	case CODEC_QDESIGN: return "QDesign music";
	case CODEC_QDESIGN2: return "QDesign2 music";
	case MK4CC('m','s',1,80):
	case CODEC_QUALCOMM_PV: return "Qualcomm PureVoice";
	case CODEC_DWVW: return "DWVW";
	case CODEC_GSM: return "Apple GSM";
	case CODEC_G726_ADPCM: return "G726";
	case CODEC_ADP4: return "ADP4";
	case CODEC_MPEG_L3P: return "MPEG Layer II/III";
	case CODEC_AAC_RAW1: return "AAC";
	case MK4CC('m','s',22,0):
	case CODEC_AAC_ADTS: return "AAC ADTS";
	case CODEC_AAC_LATM: return "AAC LATM";
	case CODEC_PCM:
	case CODEC_PCM_DEFAULT: return "linear PCM";
	case CODEC_PCM_U8: return "offset binary";
	case CODEC_PCM_BE_S16:
	case CODEC_PCM_LE_S16:
	case CODEC_PCM_BE_S24:
	case CODEC_PCM_BE_S32:
	case CODEC_PCM_LE_S32: return "integer";
	case CODEC_PCM_BE_F32:
	case CODEC_PCM_BE_F64:
	case CODEC_PCM_LE_F32:
	case CODEC_PCM_LE_F64: return "floating point";
	case CODEC_RHETOREX_ADPCM: return "Rhetorex";
	case CODEC_WMA1: return "WMA v1";
	case CODEC_WMA2: return "WMA v2";
	case CODEC_WMAPRO: return "WMA Pro";
	case CODEC_WMALOSSLESS: return "WMA";
	case CODEC_ADPCM_CT: return "Creative Labs";
	case CODEC_IMC: return "IMC";
	case CODEC_IAC_2: return "Indeo(R) IAC v2";
	case CODEC_VME_VMPCM: return "VME VMPCM"; /* AT&T Labs, Inc. */
	case CODEC_LH: return "Lernout & Hauspie";
	case CODEC_DTS_DS: return "DTS DS";
	case CODEC_DTS: return "DTS";
	case MK4CC('m','s',32,2): return "RA 1/2 14.4";
	case MK4CC('m','s',32,3): return "RA 1/2 28.8";
	case MK4CC('m','s',32,4): return "RA G2/8";
	case MK4CC('m','s',32,5): return "RA 3/4/5";
	case MK4CC('m','s',32,6): return "RA 10 AAC";
	case MK4CC('m','s',32,7): return "RA 10 AAC+";
	case CODEC_VOXWARE_RT24: return "Voxware RT24";
	case CODEC_VOXWARE_RT29: return "Voxware RT29";
	case MK4CC('m','s',161,9):
	case CODEC_SPEEX: return "Speex";
	case CODEC_ALAC: return "ALAC";
	case CODEC_FLAC: return "FLAC";
	case CODEC_TAK: return "TAK";
	case CODEC_MONKEYS_AUDIO: return "Monkey's Audio";
	case CODEC_SDX2: return "SDX2";
	case CODEC_AAC: return "AAC";
	case CODEC_LC_AAC: return "LC-AAC";
	case CODEC_MPEG_AUDIO: return "MPEG";
	case CODEC_MPEG_L1P: return "MPEG Layer I";
	case CODEC_MPEG_L2P: return "MPEG Layer II";
	case CODEC_VORBIS: return "Vorbis";
	case CODEC_OPUS: return "Opus";
	case CODEC_ATRAC: return "ATRAC";
	case CODEC_ATRAC3: return "ATRAC3";
	case CODEC_AMR: return "AMR";
	case CODEC_SONIC: return "Sonic";
	case CODEC_AC3: return "Dolby AC3";
	case CODEC_ILBC: return "iLBC";
	case CODEC_UHER_ADPCM: return "UHER"; /* UHER informatic GmbH */
	case CODEC_CJ_LIGHTWAVE: return "Litewave";
	case CODEC_GSM_MS: return "GSM 1500";
	case CODEC_SANYO_LDADPCM: return "Sanyo LD";
	case CODEC_UNISYS_NAP_ADPCM: return "Unisys NAP";
	case CODEC_PHILIPS_CELP: return "Philips CELP";
	case CODEC_ACELP_NET: return "ACELP.NET";
	case CODEC_SIPRO_KELVIN: return "ACELP.KELVIN";
	case CODEC_TRUESPEECH_X: return "TrueSpeech"; /* deprecated */
	case MK4CC('m','s',1,1): return "IRAT";
	case MK4CC('m','s',1,17): return "Vivo G.723";
	case MK4CC('m','s',1,18): return "Vivo Siren";
	case MK4CC('m','s',1,49): return "Sipro ACELP4800";
	case MK4CC('m','s',1,50): return "Sipro ACELP8v3";
	case MK4CC('m','s',1,51): return "Sipro G729";
	case MK4CC('m','s',1,52): return "Sipro G729A";
	case MK4CC('m','s',1,81): return "Qualcomm Halfrate";
	case MK4CC('m','s',1,85): return "TUBGSM"; /* Ring Zero Systems, Inc. */
	case MK4CC('m','s',2,2): return "CT FastSpeech8";
	case MK4CC('m','s',2,3): return "CT FastSpeech10";
	case MK4CC('m','s',2,32): return "Quarterdeck";
	case MK4CC('m','s',2,48): return "ILinkVC";
	case MK4CC('m','s',2,64): return "AS RawSport";
	case MK4CC('m','s',2,65): return "ESSTAC3";
	case MK4CC('m','s',3,0): return "FM Towns Snd";
	case MK4CC('m','s',4,0): return "BTV Digital"; /* Brooktree Digital */
	case MK4CC('m','s',16,0): return "Olivetti GSM";
	case MK4CC('m','s',16,1): return "Olivetti ADPCM";
	case MK4CC('m','s',16,2): return "Olivetti CELP";
	case MK4CC('m','s',16,3): return "Olivetti SBC";
	case MK4CC('m','s',16,4): return "Olivetti OPR";
	case MK4CC('m','s',20,0): return "Norris";
	case MK4CC('m','s',20,1): return "ATISI"; /* AT&T Labs, Inc. (?) */
	case MK4CC('m','s',2,22): return "DVAudio";
	case MK4CC('m','s','g','O'): return "Vorbis (mode 1)";
	case MK4CC('m','s','g',80):
	case MK4CC('m','s','g',81):
	case MK4CC('m','s','g',111):
	case MK4CC('m','s','g',112):
	case MK4CC('m','s','g',113): return "Vorbis (ACM)"; /* incompatible */
	case MK4CC('m','s','p','m'): return "FAAD AAC";

	case CODEC_SBPRO_ADPCM_4: return "SBPro 4-bit";
	case CODEC_SBPRO_ADPCM_3: return "SBPro 2.6-bit";
	case CODEC_SBPRO_ADPCM_2: return "SBPro 2-bit";

	default: return NULL;
	}
}
