/*
 interpreter.c
  Version 1.9.9f3

  Created: 2008.08.11
   Author: Adam J. P.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 See history.txt for the complete version history.

 Important Language Notes:
   In legacy ordering, not only are most operations left-to-right, but run-on
    multiplication can be right-to-left; e.g.,: `~/2~/3` = 1.8612097182042.
    Because the initial 'else' value is taken, `a?(2):1?4:3` will always = 4 as
    the first interrogation - a?(2):1 - yields a non-zero value.
   Always use PEMDAS+ mode for modern language compatibility.

   Deep ternary conditionals (`a?b?c:d:e`) are not handled without parentheses.

   Unlike JavaScript, >>> performs an signed downshift; unlike Lua, >> performs a
    scalb(a,-b) when compiled for floating-point. Logical ops like `a|b` integerize.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifndef NO_TABLES
 #define HAVE_IUC_TABLES
#endif

#include "interpreter.h"
#include "eval_op.h"

#include "my_cseries.h"

#ifdef env68k
 #pragma segment Interpreter
#endif

#ifndef ALLOW_RUNON_WSP
 #define ALLOW_RUNON_WSP	0
#endif
#ifndef STRICT_ESCAPES
 #define STRICT_ESCAPES		0
#elif STRICT_ESCAPES>1 /* request override new-line char with modern LF */
 #undef CHR_NL
 #define CHR_NL 10
#endif

/* upper 8 bits are for smaller values, +/-0x7F */
typedef uint16		evalent;

#define ENTS_PER_EVAL		(sizeof(evalvalue)/sizeof(evalent))

/* EXT_EVAL: full value follows, lowest-significant first */
#if EVAL_FLOATING
 #ifndef native_eval
  #define native_eval		double_t
 #endif
 #define EXT_EVAL		0xFE		/* large range */
 typedef uint16		t_ceval;
#else
 #define native_eval		evalvalue
 #define EXT_EVAL		(int8)0x80	/* +/-0x7F */
 typedef int16		t_ceval;
#endif

/* ---- macros */

#ifndef SKIP_NBSP
 #undef CHR_NB
#endif

#if 0//def CHR_NB
 #define skipwsp(s,e)		do{uint8 c=*(s)++; \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0&&(c+=' '-CHR_NB)!=0) \
		{--(s); break;}}while(s<e)
 #define rskipwsp(s,f)		do{uint8 c=*--(s); \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0&&(c+=' '-CHR_NB)!=0) \
		{++(s); break;}}while(s>f)
#else
 #define skipwsp(s,e)		do{uint8 c=*(s)++; \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0) {--(s); break;}}while(s<e)
 #define rskipwsp(s,f)		do{uint8 c=*--(s); \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0) {++(s); break;}}while(s>f)
#endif

#define skipdig(s,e)		do{uint8 c=*(s)++; \
	if ((c+=-'_')!=0 && (c+='_'-'0')>(uint8)('9'-'0')) \
	{--(s); break;}}while(s<e)

/* ---- globals */

const char	*_interp_defbasechars =
	"0123456789ABCDEFGHIJKLMNOPQRSTUVXYZ@#$`!%&()*+,./:;<=>?[\\]^{|}~";

#ifndef NO_TABLES
 #include "iuctable.h"

#define EURO	UC_EURO
#define HELI	UC_HELIP
#define DG	UC_DAGGR
#define PRML	UC_PERML
#define INEQ	UC_INEQ
#define ALSQ	UC_ALSQU
#define ARSQ	UC_ARSQU
#define ALDQ	UC_ALDQU
#define ARDQ	UC_ARDQU
#define SQ	UC_LSQUO
#define DQ	UC_LDQUO
#define BULL	UC_BULLT
#define ND	UC_NDASH
#define TRMK	UC_TM
#define LOGO	UC_LOGOP
#define INFN	UC_INF
#define LSEQ	UC_LSOEQ
#define GTEQ	UC_GTOEQ
#define PARD	UC_PARD
#define SUMM	UC_SUM
#define PROD	UC_PROD
#define INTG	UC_INTGR
#define RADC	UC_RAD
#define ALEQ	UC_ALMEQ
#define INCR	UC_INCR
#define LZNG	UC_LOZNG
#define FRSL	UC_FRACS
#define LO9S	UC_LO9SQ
#define LO9D	UC_LO9DQ
#define fiLG	UC_fiLIG
#define flLG	UC_flLIG
#define ADAI	UC_ADIAE-0x1000
#define ODAI	UC_ODIAE-0x1000

#define UNDS	UCFUNDSC
#define NBSP	IC_NBSP

#define SP	IC_SP
#define	LF	IC_LF
#define	VT	IC_VT
#define	FS	IC_FS
#define HS	IC_SOH
#define ST	IC_STX
#define EN	IC_ENQ
#define BS	IC_BKSP
#define D0	UCFDIG_0
#define LA	UCFLAT_A
#define LK	UCFLAT_A+'K'-'A'
#define LU	UCFLAT_A+'U'-'A'
#define AG	UC_AGRAV-0x1000
#define EG	UC_EGRAV-0x1000
#define IG	UC_IGRAV-0x1000
#define OG	UC_OGRAV-0x1000
#define UG	UC_UGRAV-0x1000
#define PI	UC_PI-0x1000
#define A	1
#define C	2
#define D	3
const int16	_interpUCType[256] =
{	/*0    1    2    3    4    5    6    7    8    9    A    B    C    D    E    F*/
	0x20,HS+0,ST+0,ST+1,HS+1,EN+0,EN+1,BS-1,BS+0,SP+1,LF+0,VT+0,VT-1,LF+1,0x0E,0x0F,
	0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,FS+0,FS+1,FS+2,FS+3,
	SP+0,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,
	D0+0,D0+1,D0+2,D0+3,D0+4,D0+5,D0+6,D0+7,D0+8,D0+9,0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,
	0x40,LA+0,LA+1,LA+2,LA+3,LA+4,LA+5,LA+6,LA+7,LA+8,LA+9,LK+0,LK+1,LK+2,LK+3,LK+4,
	LK+5,LK+6,LK+7,LK+8,LK+9,LU+0,LU+1,LU+2,LU+3,LU+4,LU+5,0x5B,0x5C,0x5D,0x5E,UNDS,
	0x60,LA+0,LA+1,LA+2,LA+3,LA+4,LA+5,LA+6,LA+7,LA+8,LA+9,LK+0,LK+1,LK+2,LK+3,LK+4,
	LK+5,LK+6,LK+7,LK+8,LK+9,LU+0,LU+1,LU+2,LU+3,LU+4,LU+5,0x7B,0x7C,0x7D,0x7E,0x7F,
#ifdef mac /* MacRoman default */
	ADAI,ARNG,CCED,EG+A,NTLD,ODAI,UG+D,AG+A,AG+0,AG+C,ADAI,ATLD,ARNG,CCED,EG+A,EG+0,
	EG+C,EG+D,IG+A,IG+0,IG+C,IG+D,NTLD,OG+A,OG+0,OG+C,ODAI,OTLD,UG+A,UG+0,UG+C,UG+D,
	DG+0,0xB0,0xA2,0xA3,0xA7,BULL,0xB6,SHPS,0xAE,0xA9,TRMK,ACUT,DIAE,INEQ,AELG,OSLS,
	INFN,0xB1,LSEQ,GTEQ,0xA5,0xB5,PARD,SUMM,PROD,PI+0,INTG,FORD,MORD,OMGA,AELG,OSLS,
	0xBF,0xA1,0xAC,RADC,FHOO,ALEQ,INCR,0xAB,0xBB,HELI,NBSP,AG+0,ATLD,OTLD,OELG,OELG,
	ND+0,ND+1,DQ+0,DQ+1,SQ+0,SQ+1,0xF7,LZNG,YDIA,YDIA,FRSL,EURO,ALSQ,ARSQ,fiLG,flLG,
	DG+1,0xB7,LO9S,LO9D,PRML,AG+C,EG+C,AG+A,EG+D,EG+0,IG+A,IG+C,IG+D,IG+0,OG+A,OG+C,
	LOGO,OG+0,UG+A,UG+C,UG+0,DTLI,CRCM,SMTL,MCRN,BREV,DTAB,RNAB,CDLL,DBAC,OGNK,CARN
	/* characters like broken bar and vulgar 1/2 are not in the set */
	/* in latest set, D5=Euro (originally currency U+00A4) and B5=omega (vs ohm) */
#else /* Windows-1252 default (warning: not double-checked in revision) */
	EURO,0x81,LO9S,FHOO,LO9D,HELI,DG+0,DG+1,CRCM,PRML,SCAR,ALSQ,OELG,0x8D,ZCAR,0x8F,
	0x90,SQ+0,SQ+1,DQ+0,DQ+1,BULL,ND+0,ND+1,SMTL,TRMK,SCAR,ARSQ,OELG,0x9D,ZCAR,YDIA,
	NBSP,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,DIAE,0xA9,FORD,ALDQ,0xAC,0xAD,0xAE,MCRN,
	0xB0,0xB1,SUP2,SUP3,ACUT,0xB5,0xB6,0xB7,CDLL,SUP1,MORD,ARDQ,F1_4,F1_2,F3_4,0xBF,
	AG+0,AG+A,AG+C,ATLD,ADAI,ARNG,AELG,CCED,EG+0,EG+A,EG+C,EG+D,IG+0,IG+A,IG+C,IG+D,
	ETH_,NTLD,OG+0,OG+A,OG+C,OTLD,ODAI,0xD7,OSLS,UG+0,UG+A,UG+C,UG+D,YACU,THRN,SHPS,
	AG+0,AG+A,AG+C,ATLD,ADAI,ARNG,AELG,CCED,EG+0,EG+A,EG+C,EG+D,IG+0,IG+A,IG+C,IG+D,
	ETH_,NTLD,OG+0,OG+A,OG+C,OTLD,ODAI,0xF7,OSLS,UG+0,UG+A,UG+C,UG+D,YACU,THRN,YDIA
	/* characters like radical and inequal are not in the set */
#endif
#undef D0
#undef LA
#undef LK
#undef LU
#undef AG
#undef EG
#undef IG
#undef OG
#undef UG
#undef SC
#undef PI
#undef A
#undef C
#undef D
};

/* ---- required code and basic parsing */

int strcasecmp(const char *a,
	const char *b)
{
	const int16	*t = _interpUCType;
	unsigned	c, d;

	if (!a || !b)
		return IMIN(int);
	do
		if ((c=(uint8)*a++)==0)
			return -*b;
	while(t[c]==t[d=(uint8)*b++]);
	return c-d;
}

int strncasecmp(const char *a,
	const char *b,
	size_t z)
{
	const int16	*t = _interpUCType;
	unsigned	c, d;

	if (!a || !b)
		return IMIN(int);
	if (!z)
		return 0;
	do
		if ((c=(uint8)*a++)==0)
			return -*b;
	while(t[c]==t[d=(uint8)*b++] && --z);
	return c-d;
}

int memcasecmp(const char *a,
	const char *b,
	size_t z)
{
	const int16	*t = _interpUCType;
	unsigned	c, d;

	if (!a || !b)
		return IMIN(int);
	if (!z)
		return 0;
	while(t[c=(uint8)*a++]==t[d=(uint8)*b++] && --z)
		;
	return z ? c-d:0;
}

const char *parse_wordbreak(const char *s,
	const char *e,
	uint8 *h,
	int16 min,
	int16 max)
{
	int16		i;

	if (!s)
		return 0;
	skipwsp(s,e);
	if (s==e)
		return s;
	if (h)
	{
		unsigned	c, v = 0;
		uint8		f = *s&0x3F;

		do
			if ((i=_interpUCType[c=(uint8)*s++])>max || i<min)
				{--s; break;}
		while(v+=c,s<e);
		*h = (uint8)(f|v<<6);
	} else
		do
			if ((i=_interpUCType[(uint8)*s++])>max || i<min)
				{--s; break;}
		while(s<e);
	return s;
}

const char *parse_rwordbreak(const char *s,
	const char *f,
	int16 limit)
{
	if (!s)
		return 0;
	rskipwsp(s,f);
	if (s==f)
		return s;
	do
		if (_interpUCType[(uint8)*--s]>limit)
			{--s; break;}
	while(s>=f);

	return s;
}

const char *parse_wordhash(const char *s,
	const char *e,
	uint64 *h)
{
	uint64		v = 0;
	unsigned	i, c;

	skipwsp(s,e);
	if (s>=e)
		return s;
	do
		if ((i=_interpUCType[c=(uint8)*s++]-UCFLAT_AT)>UCFUNDSC-UCFLAT_AT)
			{--s; break;}
	while(v=(v<<5)|i,s<e);
	*h = v;
	return s;
}
#endif

const char *parse_entrybreak(const char *str,
	const char *e,
	int32 t)
{
	const char	*s = str, *w = NULL;

	while(s<e)
		switch(*s)
		{
		case '/': /* "//" but not "://" */
			if (++s<e && (*s=='/' || *s=='*') && (s<str+2 || s[-2]!=':'))
				{--s; goto end;}
			w = NULL;
			break;
		case '\t':
		case ' ':
			if (!w)
				w = s;
			++s;
			break;
		case '\\':
			if (++s>=e)
				{--s; goto end;}
			w = NULL;
			break;
		case '[':
		case ']':
			if (t>1u) /* valid for field (after first char) */
				goto end;
		case '!':
		case '#':
		case '%':
		case '+':
		case '-':
		case '.':
		case '_':
		case '|':
		case '`':
		case ':':
		case '@':
		case '&':
		case '?':
		case '*':
		case '(':
		case ')':
		CASE_NUMERIC
		CASE_ALPHABETIC
			w = NULL, ++s;
			break;
		case '<':
		case '>':
			if (t<5)
				goto end;
			w = NULL, ++s;
			break;
		case '=':
			if (t)
				{w = NULL, ++s; break;}

		default:
#ifdef CHR_NB
			if (*s==CHR_NB)
				{w = NULL, ++s; break;}
#endif
			goto end;
		}
end:	if (w)
		s = w;
	return MIN(s,e);
}

/* extended standard */

long sreadline(const char *p,
	long z,
	uint32 *lecount,
	char escape)
{
	const char	*s = p;
	uint32		lec;
	uint8		c, e = 0;

	if (z<=0)
		return 0;
	lec = 0;
	if (lecount)
		lec = *lecount;
	do
	{
		c = *s++;
		if (escape && c==escape)
			e = e!=1;
		else
		{
			if (c-10u<4u && c!='\f') /* LF/VT/FF/CR */
			{
				if (e>1)
				{
					e = 0;
					if (c==10) /* (DOS line-ending) */
						continue;
				} else if (c==13)
					e+= e;
				++lec;
				if (!e) /* end of escaped line */
					{--s; break;} /* last return not recorded */
			}
			e&=~1;
		}
	} while(--z);

	if (lecount)
		*lecount = lec;
	return s-p;
}

char *freadline(FILE *f,
	uint32 *length,
	char *d,
	void *(__cdecl reallocfunc)(void*,
		size_t),
	uint32 *lecount,
	int escape)
{
	void		*p;
	uint32		r, z, lec;
	uint8		e = 0;

	if (!reallocfunc)
		reallocfunc = realloc;
	lec = z = 0;
	if (lecount)
		lec = *lecount;
	if (!length)
		r =~0u;
	else if ((r=*length)==0)
		goto end;
	do
	{
		int		c;

		if ((c=fgetc(f))<0) //==EOF)
		{ /* empty on full EOF */
			if (d && !z)
				d = reallocfunc(d,0);
			break;
		}
		if (escape && c==escape)
			e = e!=1;
		else
		{
		#if FREADLINETAKEFF
			if (c-10u<4u) /* LF/VT/FF/CR */
		#else
			if (c-10u<4u && c!=12) /* LF/VT/CR */
		#endif
			{
				if (e>1)
				{
					e = 0;
					if (c==10) /* (DOS line-ending) */
						continue;
				} else if (c==13)
					e+= e;
				c = '\n', ++lec;
				if (!e) /* end of escaped line */
					break; /* last return not recorded */
			}
			e&=~1;
		}
		if (!(z&7)) /* add some extra bytes for fewer calls */
		{
			if ((p=reallocfunc(d,8+z))==0)
				goto nix;
			d = p;
		}
		d[z] = (char)c;
		if (++z>0x7FFFFFEF)
			goto end; /* can't terminate */
	} while(--r);

	if (d)
	{ /* terminate */
		if (!(z&7))
		{
			if ((p=reallocfunc(d,1+z))==0)
			{
			nix:	d = reallocfunc(d,0);
				goto end;
			}
			d = p;
		}
		d[z] = 0;
	}
end:	if (length)
		*length = z;
	if (lecount)
		*lecount = lec;
	return d;
}

char *freadfile(FILE *f,
	int32 *z,
	int (__cdecl delim)(const char*,
		const char*))
{
	char		b[65], *d = NULL;
	uint32		t = 0, o, lo = 0;
	/* because of stdin, only fgetc or fgets can be used */
	while(fgets(b,sizeof(b),f) && (o=strlen(b))!=0) /* no nulls allowed here */
	{
		char	*p, nl = b[o-1]==13 || b[o-1]==10;

		if ((t+=o)>0x7FFFFFF0) /* clip */
			o = 0x7FFFFFF0-t-o;
		if ((p=realloc(d,t))==NULL)
			{d = realloc(d,0); break;}
		d = p;
		memcpy(d+t-o,b,o);
		if (feof(f))
			break;
		if (nl)
		{
			if (delim && !delim(d+lo,d+t))
				break;
			lo = t;
		}
	}
	if (z)
		*z = t;
	return d;
}

/* ---- number parsing/printing */

const char *readanyinteger(const char *s,
	long z,
	long int *v)
{
	int64		_v;

	s = readanyint64(s,s+z,&_v);
	if (v)
		*v = _v!=(long int)_v ? IMIN(long int):(long int)_v;
	return s;
}

const char *readanyint64(const char *s,
	const char *e,
	int64 *v)
{
	int64		r, m = 0, mxv;
	uint8		c, mxd;

	if (!s || s>e)
		return 0;
	r = 0;
	if (s==e)
		return s;
	c = *s++;
	if ((c+=-'-')==0)
	{
		if (s>=e)
			return 0;
		m =-1, c = *s++;
	} else if ((c+='-'-'+')==0)
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if ((uint8)((c+='+')-'0')>(uint8)9)
	{
#ifdef CHR_HX
		if (c==CHR_HX)
			goto hex;
#endif
		return 0;
	}
	if (c!='0')
	{
		goto dec;
		do
		{
			c = *s++;
		dec:	c-= '0';
			if (c>(uint8)9)
				if ((c+='0'-'A')<=(uint8)('Z'-'A')
						|| (c+='A'-'a')<=(uint8)('z'-'a')
						|| (c+='a'-'_')==0)
					{s = 0; break;}
				else
					{--s; break;}
			mxv = r, r+= r, r+= r<<2;
			r+= (uint32)c;
			if ((mxv|(r&1ull<<63))>922337203685477580ll)
				goto ovfl;
		} while(s<e);
		goto end;
	}

	if (s>=e)
		goto end;
	c = *s++;
	if (c!='x' && c!='X')
	{
		if (c=='o' || c=='O') /* octal (starting-zero-only not supported here) */
			mxd = 7, mxv = 0x1000000000000000ll;
		else if (c=='b' || c=='B') /* binary */
			mxd = 1, mxv = 0x4000000000000000ll;
		else
			goto dec;
		do
		{
			c = *s++, c-= '0';
			if (c>mxd)
				if (c<=(uint8)('9'-'0')
						|| (c+='0'-'A')<=(uint8)('Z'-'A')
						|| (c+='A'-'a')<=(uint8)('z'-'a')
						|| (c+='a'-'_')==0)
					{s = 0; break;}
				else
					{--s; break;}
			if (r>=mxv)
				goto ovfl;
			r+= r;
			if (mxd>1)
				r<<= 2;
			r+= (uint32)c;
		} while(s<e);
		goto end;
	}
#ifdef CHR_HX
hex:
#endif
	if (s>=e) /* only starting character? */
		return 0;
	do
	{
		c = *s++, c-= '0';
		if (c<=(uint8)('9'-'0'))
			goto hd;
		else if ((c+='0'-'A')<=(uint8)('Z'-'A')
				|| (c+='A'-'a')<=(uint8)('z'-'a'))
			if (c>(uint8)('F'-'A'))
				{s = 0; break;}
			else
			{
				c+= 10;
			hd:	if (r>=0x800000000000000ll)
					{ovfl: r = (int64)0x8000000000000000ull; break;}
				r<<= 4, r+= (uint32)c;
			}
		else if ((c+='a'-'_')==0)
			{s = 0; break;}
		else
			{--s; break;}
	} while(s<e);
end:	r+= m, r^= m;
	if (v)
		*v = r;
	return s;
}

int makebasestring(char *outstr,
	const char *basechars,
	long value,
	unsigned base,
	int minlength,
	char signchar)
{
	uint8		store[33]; /* enough space to do binary and sign */
	uint8		*s = sizeof(store)+store;
	unsigned	v = value;
	int		x = 0;

	if (!base)
		return 0;
	/* SC generated erroneous code if `value` wasn't used in the main loop, */
	/* optimizing it out entirely as `v`, nixing this conditional as a result */
	if ((signed)v<0 && signchar)
		v = (unsigned)-(signed)v;
	if (!basechars)
		basechars = _interp_defbasechars;
	if (minlength<=0)
		minlength = 1;
	else if (minlength>sizeof(store))
		minlength = sizeof(store);
	do
	{
		unsigned	i = v;

		*--s = basechars[i-=(v/=base)*base];
	} while(++x<minlength || v);
	minlength = x;

	if (outstr)
	{
		uint8	*d = (uint8*)outstr;

		if (value<0)
		{
			if (signchar=='+')
				signchar = '-';
		} else if (signchar!='+')
			signchar = 0;
		if (signchar)
			*d++ = signchar, ++minlength;
		/* copy to output--no func. calls, unaligned compatible */
		_amcpy(d,s,x);
	}
	return minlength;
}

const char *parse_float(const char *s,
	const char *e,
	s_float80 *_v,
	unsigned char r)
{
	double_t	fb, v;
	uint32		hc, b = 10;
	int		u = 0, a, m, hm, x, xb;
	uint8		c, n;

	if (s>=e)
		return NULL;
	v = 0, hc = 0, hm = x = m = a = 0, n = 0;
	skipwsp(s,e);
#if CHR_MN!='-'
	if ((c=*s++)=='-' || c==(uint8)CHR_MN)
#else
	if ((c=*s++)=='-')
#endif
	{
		n = 0x80;
		if (s>=e)
			return 0;
		c = *s++;
	} else if (c=='+')
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if (c!=mLocaleNPtChar && (uint8)(c-'0')>(uint8)('9'-'0'))
	{
#ifndef PARSE_ONLY_REAL
		if (c!='i')
#endif
		return 0; /* first must be a digit */
	}
#ifdef PARSE_ANY_REAL
	if (c=='i' && s+2<=e && *s=='n' && s[1]=='f')
	{
		v = INFINITY;
		if ((s+=2)==e)
			goto end;
		fb = 0, xb = 0, b = 0, a = 1; /* don't allow extra characters */
		u = 0x8000;
		c = *s++;
		goto start;
	}
#endif
	if (c=='0' && s<e)
	{
		if ((c=*s-'X')==0 || (c+='X'-'x')==0)
			u = 4, b = 16, ++s; /* hexadecimal, can have 'p' */
		else if ((c+='x'-'h')==0)
			u = 3, b = 16, ++s; /* hexadecimal */
		else if ((c+='h'-'O')==0 || (c+='O'-'o')==0)
			b = 8, ++s; /* octal (starting-zero-only not supported) */
		else if ((c+='o'-'b')==0)
			b = 2, ++s; /* binary */
		else if ((c+='b'-'d')==0)
			b = 10, r = 0, ++s; /* fractionless decimal */
		else if ((c+='d'-'z')==0)
			b = 36, u = 3, ++s; /* base 36 (all alphanumeric) */
		else
			u|= 0x8000; /* decimal, zero parsed */
		if (s>=e)
			return 0; /* must follow digits */
		c = *s++;
	}
	fb = (double)b, xb = b;
	goto start;
	do
	{
		c = *s++;
	start:	if ((c+=-'0')<=(uint8)('9'-'0')) /* 0-9 */
		{
			if (c>=(uint8)b)
				return 0; /* digit invalid for this base number */
			u|= 0x8000;
			if (a>0) /* this could easily overflow without clipping */
				x = x>32767 ? 32768:xb*x+c;
			else
				goto adddig;
		} else if (((c+='0'-'A')<='Z'-'A' || (c+='A'-'a')<='z'-'a') && (uint8)u>2)
		{
			if ((c+=10)>=(uint8)b)
			{
				if (b==16 && c=='p'-'a'+10)
				{
					if ((uint8)u==3)
						return 0; /* must be in 0xM.M form for p */
					u = 0, xb = 10; /* now parsing exponent in decimal */
					goto de;
				}
			#ifdef PARSE_ANY_REAL
				goto back;
			#else
				return 0; /* digit invalid for this base number */
			#endif
			}
			u|= 0x8000;
		adddig:	if (hc<=32767)
				hc = b*hc+c, hm = m+a;
			else
				v = fb*v+c, hm+= a+1;
			m+= a;
		} else if ((c+='A'-'E')==0) /* Ee */
		{ /* can't have more than one `E' */
		de:	if (a==1)
				return 0;
			a = 1, u&=~0x8000;
		} else if ((c=s[-1])==mLocaleNPtChar)
		{ /* can't have multiple points or for integer or notation */
			if (a || !r)
		#ifdef PARSE_ANY_REAL
				goto back;
		#else
				return 0;
		#endif
			a =-1;
		} else if (c==mLocaleNSepChar)
		{
			if (b!=10)
				return 0;
#if CHR_MN!='-'
		} else if ((c-='-')==0 || c==(uint8)(CHR_MN-'-'))
#else
		} else if ((c-='-')==0)
#endif
		{ /* only one '-' valid here for notation; cannot come after `E+' */
			if (a<=0 || u&0x8001)
				goto back;
			u = 1;
		} else if ((c+='-'-'+')==0) /* + */
		{ /* only one '+' valid here for notation; cannot come after `E-' */
			if (a<=0 || u&0x80FF)
				goto back;
			u = 2;
		} else if ((c+='+'-'_')==0)
		{ /* strict: can't directly follow an underscore */
		#if INTERP_VERS>0x197
			continue;
		#else
			return 0;
		#endif
		} else /* everything else ends */
		{
		back:	--s;
			break;
		}
	} while(s<e);

	if (!(u&0x8000))
	{
	#ifdef HAVE_COMPLEX_EVAL
		if (s[-1]=='E')
			--s;
		else
	#endif
		return 0; /* no digit chars after start or notation */
	}
#ifdef PARSE_ANY_REAL
end:
#endif
	if (_v)
	{
		if ((uint8)u==1)
			x =-x;
		*_v = i64to80(hc);
		switch(b)
		{
		default: m+= x, hm+= x; *_v = f80_scal10(f80_add(f80_scal10(*_v,hm-m),f64to80(v)),m);
			break;
		case 8: m+= m*2, hm+= hm*2; goto sclb;
		case 16: m+= m, hm+= hm;
		case 4:	m+= m, hm+= hm;
		case 2:
		sclb:	m+= x, hm+= x; *_v = f80_scalb(f80_add(f80_scalb(*_v,hm-m),f64to80(v)),m);
			break;
		case 36:
		//printf("%d %d %d\n",m,hm,hm-m);
			m+= x, hm+= x; *_v = f80_scal36(f80_add(f80_scal36(*_v,hm-m),f64to80(v)),m);
		}
		((int8*)&_v->sgnexp)[!BIG_ENDIAN]^= n;
	}
	return s;
}

const char *parse_double(const char *s,
	const char *e,
	double *_v,
	unsigned char r)
{
	s_float80	v;

	s = parse_float(s,e,&v,r);
	if (_v)
		*_v = f80to64(v);
	return s;
}

int sprint_evalue(char *b, /* needs at least 63 chars */
	evalvalue v,
	int p)
{
#if EVAL_FLOATING
	int		r = 0;

	if (p<0)
	{
	#ifdef HAVE_COMPLEX_EVAL
		if (creal(v) || !cimag(v))
		{
			r = sprint_fphex(b,creal(v));
			if (cimag(v) && !signbit(cimag(v)))
				b[r++] = '+';
		}
		if (cimag(v))
		{
			r+= sprint_fphex(r+b,cimag(v));
			b[r++] = 'I';
			b[r] = 0;
		}
	#else
		r = sprint_fphex(b,v);
	#endif
		return r;
	}
 #ifdef HAVE_COMPLEX_EVAL
	if (isnan(creal(v)) || isnan(cimag(v)))
	{ /* undefined */
		strcpy(b,"nan");
		return 3;
	}
	if (creal(v) || !cimag(v))
	{
		r = sprint_double(b,creal(v),p,mLocaleNPtChar,'-');
		if (cimag(v) && !signbit(cimag(v)))
			b[r++] = '+';
	}
	if (cimag(v))
	{
		r+= sprint_double(r+b,cimag(v),p,mLocaleNPtChar,'-');
		b[r++] = 'I';
		b[r] = 0;
	}
	return r;
 #else
	return sprint_double(b,v,p,mLocaleNPtChar,'-');
 #endif
#else
	p = makebasestring(b,0,v,10,0,'-');
	b[p] = 0;
	return p;
#endif
}

/* this requires that the characters have already been processed */
/* (use parse_textloop) */
int parse_literal(const char *s,
	int64 *r,
	long z)
{
	const char	*e;
	int64		v;

	if (z+0u>8u) /* handle negatives too */
		return _e_general_error;
	e = s+z;
	v = 0;
	while(s<e)
		v = (v<<8)|(uint8)*s++;
	if (r)
		*r = v;
	return z<3 ? z:z<5 ? 4:8;
}

static void __cdecl getl(void *u,
	char v)
{
	**(char**)u = v, ++*(char**)u;
}

const char *parse_eval_literal(const char *str,
	const char *e,
	char q,
	evalvalue *v)
{
	char		b[10], *p = b;
	int64		i;
	const char	*s;
	int32		z = 8;

	s = parse_textloop(str,e,q,&z,NULL,NULL,&p,getl);
	if (parse_literal(b,&i,z)<0)
		z = 0;
#if EVAL_FLOATING
	evrset(*v,int64tox(i));
#else
	*v = (int)int64tol(i);
#endif
	return z ? s:NULL;
}

int smake_utf8(uint8 *d,
	uint32 v)
{
	int		z = 0, i = 0;

	if (v<0x80)
		z = 1;
	else if (v<0x800)
		z = 2;
	else if (v<0x10000)
	{
		if (v-0xD800u<0x800u)
			goto inv; /* surrogates are invalid points */
		z = 3;
	} else if (v<0x200000)
		z = 4;
	else /* 'invalid' replacement char */
	inv:	z = 3, v = 0xFFFD;
	i = z;
	while(--i)
		d[i] = (uint8)((v&0xBF)|0x80), v>>= 6;
	*d = (uint8)(v|(int8)(-0x200>>z)>>1);
	return z;
}

const char *parse_textloop(const char *t,
	const char *e,
	char q,
	int32 *_z,
	uint8 *_t,
	char **r,
	void *u,
	void (__cdecl b)(void *u,
		char v))
{
	static uint8	ub[6];
	uint8		*us;
	const char	*s;
	char		*p = 0;
	uint32		m;
	long		z = 0, w, vc;
	uint8		c, l;

	if ((s=t)==NULL)
		return s;
	us = NULL, z = m = w = vc = 0, l = 0;
	if (r)
		p = *r;
	while(s<e)
	{
		c = *s++;
		if (w)
			if (!--vc)
				goto form;
			else if ((uint8)(c-'0')>(uint8)9)
			{
				if (w>3)
				{
					if ((c+=-'A')<=(uint8)('F'-'A')
							|| (c+='A'-'a')<=(uint8)('f'-'a'))
						{c+= 10, m = m<<4|c, w&=63; continue;}
					if (w&0x80) /* no digits */
						{--s; goto bad;}
				}
				goto form;
			} else if ((c-'0')>>w)
			{
			form:	if (us)
				{
					if (vc)
						goto bad;
					us[smake_utf8(us,m)] = 0;
					c = *us++, m = 0;
				} else
					c = (uint8)m;
				if (m>255) /* no ind. chars out of range */
					goto bad;
				m = w = vc = 0, --s;
			} else
				{m = m<<w|(c-'0'), w&=63; continue;}
		else if (c==q)
			break;
		else if (c=='\\')
		{
			if (s>=e-1)
			{
			bad:	z|=-1<<31;
				goto end;
			}
		/*esc:*/switch(c=*s++)
			{
#if STRICT_ESCAPES /* note that \l is not available with this on */
			case 'n': c = CHR_NL; break; /* new line */
			case 'r': c = CHR_NL^7; break; /* carriage return */
#else
			case 'Z': c = 0; break; /* null */
			case 'h': c = 1; break; /* heading start */
			case 'S': c = 2; break; /* text start */
			case 'e': c = 3; break; /* text end/enter */
			case 'E': c = 4; break; /* end transmission */
			case 'q': c = 5; break; /* enquiry */
			case 'k': c = 6; break; /* acknowledge */
			case 'n':
			case 'l': c = 10; break; /* linefeed */
			case 'r': c = 13; break; /* carriage return */
			case 'o': c = 14; break; /* shift out */
			case 'i': c = 15; break; /* shift in */
			case 'd': c = 16; break; /* data link esc. */
			case 'N': c = 17; break; /* dev.ctl.1-XN */
			case 'A': c = 18; break; /* device control 2 */
			case 'O': c = 19; break; /* dev.ctl.3-XO */
			case 'D': c = 20; break; /* dev.ctl.4-stop */
			case 'z': c = 21; break; /* negative ack./zero */
			case 'j': c = 22; break; /* sync. idle/join */
			case 'w': c = 23; break; /* text block end/word */
			case 'c': c = 24; break; /* cancel */
			case 'm': c = 25; break; /* medium end */
			case 's': c = 26; break; /* substitute */
			case 'y': c = 27; break; /* escape/yield */
			case 'F': c = 28; break; /* file separator */
			case 'G': c = 29; break; /* group separator */
			case 'R': c = 30; break; /* record separator */
			case 'T': c = 31; break; /* unit separator - was U */
#endif /* GCC supports all of 7-13 */
			case 'a': c = 7; break; /* bell/alert */
			case 'b': c = 8; break; /* backspace */
			case 't': c = 9; break; /* horiz. tab */
			case 'v': c = 11; break; /* vertical tab */
			case 'f': c = 12; break; /* formfeed */
#if !STRICT_ESCAPES
			case 'p': /* Pascal str */
				if (!z) /* only on starting byte */
					l = 1;
				break;
#endif
			case 'U': vc = 6; goto unis; /* Unicode character - 32-bit */
			case 'u': vc = 2; /* Unicode character - 16-bit */
			unis:	us = ub;
			case 'x': /* hexadecimal */
				w = 0x84;
				goto bc;

			default:
				if ((uint8)(c-'0')<=(uint8)7u) /* octal */
				{
					c&= 7, m = c, w = 0x03;
				bc:	vc+= 3;
					continue;
				} else if ((uint8)(c-10)<=(uint8)3) /* line-ending */
					continue; /* escaped here = ignored */
			}
		}
		do
		{
			if (b)
				b(u,c);
			if (r)
			{
				if (!(z&7))
					if ((p=realloc(p,8+z))==0)
					{
						*r = realloc(*r,0);
						z =-1<<31;
						goto end;
					} else
						*r = p;
				p[z] = c;
			}
			if (++z>0x7FFFFFF0)
				goto done; /* overflow! */
		} while(us && (c=*us++)!=0);
	}
done:
#if !STRICT_ESCAPES
	if (l)
	{
		if (z>255+1)
			z = 256; /* clip overflow */
		if (p)
			*p = (char)(z-1); /* length byte */
		if (_t)
			*_t = 1;
	} else
#endif
	 if (_t)
		*_t = 0;
end:	if (_z)
		*_z = z;
	return s;
}

#ifdef HAVE_PRINT_TCHARS
/* note that pre-v2 used swr_fputc for these */
uint32 print_tchars(FILE *out,
	const char *s,
	long z,
	char q,
	uint8 bounded)
{
	return z<0 ? 0:tprint_tchars(swr_fwrite,out,s,z,q,bounded,1);
}
#endif

uint32 tprint_tchars(sio_func_t wf,
	void *out,
	const char *s,
	long z,
	char q,
	uint8 bounded,
	uint8 strict)
{
	char		b[40], *d;
	const char	*ta = "ZhSeEqkabtlvfroidNAODzjwcmsyFGRT";
#if CHR_NL==10
	const char	*ts = "\0\0\0\0\0\0\0abtnvfr\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
#elif CHR_NL==13
	const char	*ts = "\0\0\0\0\0\0\0abtrvfn\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
#else
	const char	*ts = "\0\0\0\0\0\0\0abt\0vf\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
#endif
	const char	*t;
	long		r;
	uint8		c, lim, i;

	if (!out || !s || (z<=0 && (!z || (z=strlen(s))==0)))
		return 0;
	r = 0;
	if (strict)
		lim = 126-32u, t = ts; /* PORC doesn't support high chars */
	else
		lim = 255-32u, t = ta;
	d = b;
	do
	{
		--z, c = *s++;
		if ((uint8)(c-32u)>lim)
		{
			uint8	o = !z ? !bounded:(uint8)*s-'0'+0u<8u;

			*d++ = '\\';
			if (c<7 && !o)
				c+= '0';
			/* PORC doesn't support \a, \l is not portable */
			else if (c<32 && (i=t[c])!=0)
				c = i;
			else
			{
				if (c>077 || o)
				{
					*d++ = (c>>6)+'0', c&= 077;
					goto mid; /* fix - middle 0 was missing */
				}
				if (c>7 || o)
				mid:	*d++ = (c>>3)+'0', c&= 7;
				c+= '0';
			}
			goto pc;
		} else if (c==q || c=='\\' || (q==',' && (c=='"' || c=='\'')) ||
				(strict>1 && c=='?'))
			{*d++ = '\\'; goto pc;}
		else
		pc:	*d++ = c;
		if (d>=b+32)
			r+= wf(b,d-b,out), d = b;
	} while(z);
	if (d!=b)
		r+= wf(b,d-b,out);
	return r;
}

const char *parse_skipwsp(const char *s,
	const char *e)
{
	if (!s)
		return 0;
	skipwsp(s,e);

	return s;
}

const char *parse_rskipwsp(const char *s,
	const char *f)
{
	if (!s)
		return 0;
	rskipwsp(s,f);

	return s;
}

/* ---- evaluation */

#if EVAL_FLOATING
 #ifdef HAVE_COMPLEX_EVAL
  #define eabs cabs
 #else
  #define eabs fabs
 #endif
#else
 #define eabs ABS
#endif

#if defined(HAVE_COMPLEX_EVAL) && __POCC__>1100
/* basic op with no checking */
tcomplex _cdivz(tcomplex a,
	tcomplex b)
{
	double		denom = creal(b)*creal(b)+cimag(b)*cimag(b);

	cretri((creal(a)*creal(b)+cimag(a)*cimag(b))/denom,
		(cimag(a)*creal(b)-creal(a)*cimag(b))/denom);
}
#endif

evalvalue esqrt(evalvalue a)
{
#if EVAL_FLOATING
 #ifdef HAVE_COMPLEX_EVAL
  #if __POCC__>1100
   #define THRESH  0x1.A827999FCEF32p+1022
	tcomplex	z;
	double		r = creal(a), i = cimag(a), t;
	char		scale;
	/* special cases */
	if (!r && !i)
		{csetri(z,0,i); goto end;}
	if (isinf(i))
		{csetri(z,INFINITY,i); goto end;}
	if (isnan(r))
	{
		i = (i-i)/(i-i); /* raise invalid if imag is not nan */
		csetri(z,r,i); goto end; /* return nan+nani */
	}
	scale = 0;
	if (isinf(r))
	{	/*
			+inf+nani -> inf+nani
			+inf+xI   -> inf
			-inf+nani -> nan+-infi
			-inf+xI   -> infi
		*/
		if (signbit(r))
			{csetri(z,fabs(i-i),copysign(r,i)); goto end;}
		else
			{csetri(z,r,copysign(i-i,i)); goto end;}
	}
	/* scale to avoid overflow */
	if (fabs(r)>=THRESH || fabs(i)>=THRESH)
		r*= 0.25, i*= 0.25, scale = 1;

	/* here comes the use of hypot */
	if (r<0)
	{
		t = sqrt((-r+hypot(r,i))*0.5);
		csetri(z,fabs(i)/(t+t),copysign(t,i));
	} else
	{
		t = sqrt((r+hypot(r,i))*0.5), t+= t;
		csetri(z,t,i/t);
	}
	if (scale)
		evvadd(z,z);
end:	return z;
  #else
	return csqrt(a);
  #endif
 #else
	return sqrt(a);
 #endif
#else
	return usqrt(a);
#endif
}

#ifdef HAVE_COMPLEX_EVAL
evalvalue emod(evalvalue a,
	evalvalue b)
{
	evalvalue	q, f;

	if (!cimag(b))
		cretri(fmod(creal(a),creal(b)),fmod(creal(b),creal(b)));
	q = edivi(a,b);
	csetri(f,floor(creal(q)),floor(cimag(q)));
	q = emult(f,b);
	evvsub(a,q);
	return a;
}

tcomplex epol2com(double_t r,
	double_t th)
{
	double		costh;
	int		n;

	costh = th = remquo(th,1+.570796326794896619,&n);
	if (n&1)
		th = 1+.570796326794896619-th;
	if (n&2)
		th =-th;
	--n;
	if (n&1)
		costh = 1+.570796326794896619-costh;
	if (!(n&2))
		costh =-costh;
	cretri(costh ? r*sin(costh):0,th ? r*sin(th):0);
}
#endif

evalvalue epow(evalvalue a,
	evalvalue b)
{
#if EVAL_FLOATING
	double_t	r, u;
 #ifdef HAVE_COMPLEX_EVAL
	double_t	th;
 #else
  #define creal(n) (n)
  #define cpow pow
 #endif
	/* some conflict with 1**inf or 1**nan */
	/* 1**INF was undefined here because the 1.0 could be resulting from
	  lost precision from a non-1 value, BUT IEEE 754-2008 calls for =1.0
	  for (+/1)**inf AND **NaN, and <anything>**0, except (-1)**NaN */
	if (!evisnil(b))
	{
	#ifdef HAVE_COMPLEX_EVAL
		if (!cimag(b))
		{
	#endif
			int		t = (int8)creal(b);

			/* official implementation for cpow is inaccurate with whole exponents */
			if (!(creal(b)-t))
			{
				evrset(b,1);
				if (t<0)
				{
					t =-t;
					do
					{
						if (t&1)
							evrdiv(b,a);
						evrmul(a,a);
					} while(t>>=1);
				} else
					do
					{
						if (t&1)
						{
							if (fabs(creal(a))==INFINITY)
							{
								evradd(b,INFINITY);
								return b;
							}
							evrmul(b,a);
						}
						evrmul(a,a);
					} while(t>>=1);
				return b;
			}
		#ifdef HAVE_COMPLEX_EVAL
			if (!cimag(a))
			{
		#endif
				if (creal(a)<0)
				{ /* negative real base */
					if (creal(a)==-1 && creal(b)==INFINITY)
						{evrset(a,1); return a;}
					u = 1/creal(b);
					t = (int)u;
					if (u==t)
					{
				#ifdef HAVE_COMPLEX_EVAL
						if (t==2) /* principle square root */
							cretri(0,sqrt(-creal(a)));
				#endif
						r = pow(-creal(a),creal(b));
						if (t&1) /* real for odd root */
							{evrset(a,-r); return a;}
					}
				#ifdef HAVE_COMPLEX_EVAL
					else
						r = pow(-creal(a),creal(b));
					/* cpow is broken */
					th = M_PI*creal(b);
					goto po2ca;
				#endif
				} else if (creal(a)==1)
					{evrset(a,1); return a;}
			  	r = exp(log(creal(a))*creal(b));
				evrset(a,r); return a;
	#ifdef HAVE_COMPLEX_EVAL
			}
		}
		/* seriously, cpow doesn't work the way it should */
		u = hypot(creal(a),cimag(a)); /* abs */
		if (!u)
			if (!creal(b))
				cretri(NAN,0); /* indeterminate angle; 2020.04.17 */
			else
				cretri(0,0);
		th = atan2(cimag(a),creal(a));
		r = pow(u,creal(b));
		r*= exp(-cimag(b)*th);
		th*= creal(b);
		th+= cimag(b)*log(u);
	po2ca:	return epol2com(r,th);
	#endif
	}
	evrset(a,1); return a;
#else
	if (b<=0)
		return !a ? EVAL_HUGEVAL:!b ? 1:++b ? 0:1/a;
	else if (a==1 || !--b)
		return a;
	else if (b>31)
		return !a ? 0:++a ? EVAL_HUGEVAL:(~(long)b&1)*2-1;
	else
	{
		int64		v;
		evalvalue	m = a;

		do
		{
			(void)muls64s32(v,a,m);
			if ((long)((a=(evalvalue)int64lo(v))>>31)!=int64hi(v))
				return EVAL_HUGEVAL; /* overflow */
		} while(--b);
	}
	return a;
#endif
}

evalvalue emult(evalvalue a,
	evalvalue b)
{
#if EVAL_FLOATING && defined(HAVE_COMPLEX_EVAL)
	double_t	t, u;
	/* ac+adi+bci-bd */
	t = creal(a)*creal(b)-cimag(a)*cimag(b);
	u = creal(a)*cimag(b)+cimag(a)*creal(b);
	if (isnan(t) || isnan(u))
	{
		if (!isnan(cimag(a)))
		{
			u = cabs(a)*cabs(b);
			if (!isnan(u))
			{
				t = atan2(cimag(a),creal(a));
				t+= atan2(cimag(b),creal(b));
				return epol2com(u,t);
			}
		}
		if (!creal(b)&&!cimag(b)) /* fix for 1.9.6r2; 2020.04.16 */
			cretri(NAN,0); /* even complex-inf*0 = undefined */
		/* complex-inf * inf is still complex-inf */
		/* nan will override if passed */
		caddri(a,0,NAN);
		return a;

	}
	cretri(t,u);
#else
	return a*b;
#endif
}

evalvalue edivi(evalvalue a,
	evalvalue b)
{
#if EVAL_FLOATING
 #ifdef HAVE_COMPLEX_EVAL
	double_t	t, u, v, denr;
	int		dene, j;
	/* extract the binary scale to avoid any overflow with denom */
	dene = ilogb(creal(b));
	j = ilogb(cimag(b));
	if (dene>65536 || j>65536)
	{
		if (fabs(creal(a))==INFINITY || fabs(cimag(a))==INFINITY
				|| isnan(cimag(a))) /* 1/0/inf is undefined; 2020.04.17 */
			cretri(NAN,0); /* any cross division of inf -> undefined */
		else /* any division of inf -> *0 */
			cretri(creal(a)*0,cimag(a)*0);
	}
	/* max binary scale */
	/* ABS(FP_ILOGB0) will actually result in the more pos. exp since -IMIN==IMIN */
	if (ABS(j)>ABS(dene))
		dene = j;
	if (dene<-65536) /* zero */
	{ /* this previously made complex infinity, BUT fp zeroes are infinitesimals */
		if (creal(a))
			cretri(creal(a)*INFINITY,0);
		cretri(0,cimag(a)*INFINITY);
	}
	/* will rescale in the other direction at the end, so negate */
	/* to increase accuracy in extreme cases (e.g., 1.6e308/1.7e308), --exp here */
	dene =~dene;
	t = scalb(creal(b),dene);
	u = scalb(cimag(b),dene);
	denr = t*t+u*u;
	v = t;
	t = creal(a)*v+cimag(a)*u;
	u = cimag(a)*v-creal(a)*u;
	denr+= denr, ++dene;
	if (isnan(t) || isnan(u))
	{
		if (!isnan(cimag(a))) /* fix for 1.9.6r2; 2020.04.16 */
		{
			//printf("polar division %g/%g/%g\n",
			//	creal(a),creal(b),denr);
			u = scalb(cabs(a)/denr,dene);
			if (!isnan(u))
			{
				t = atan2(cimag(a),creal(a));
				t-= atan2(cimag(b),creal(b));
				return epol2com(u,t);
			}
		}
		caddri(a,0,NAN);
		return a;
	}
	//printf("t=%g,u=%g,denr=%g,denom exp %d\n",
	//	t,u,denr,dene);
	cretri(scalb(t/denr,dene),scalb(u/denr,dene));
 #else
	return a/b;
 #endif
#else
	return !b ? ~0u:(unsigned)a/(unsigned)b;
#endif
}

#ifdef HAVE_COMPLEX_EVAL
evinttype evintt(evalvalue v)
{
	if (cimag(v))
		return evintmin;
	return creal(v);
}

evalvalue evdbl32(evalvalue v)
{
	int64		r, i;

	*(double*)&r = creal(v);
	if (~((int32*)&r)[!BIG_ENDIAN]&0x7FF00000)
		((int32*)&r)[!BIG_ENDIAN]+= ((uint32*)&r)[BIG_ENDIAN]>>31;
	((int32*)&r)[BIG_ENDIAN] = 0;
	*(double*)&i = cimag(v);
	if (~((int32*)&i)[!BIG_ENDIAN]&0x7FF00000)
		((int32*)&i)[!BIG_ENDIAN]+= ((uint32*)&i)[BIG_ENDIAN]>>31;
	((int32*)&i)[BIG_ENDIAN] = 0;
	cretri(*(double*)&r,*(double*)&i);
}
#else
evalvalue evdbl32(evalvalue v)
{
	int64		r;

	*(double*)&r = evreal(v);
	if (~((int32*)&r)[!BIG_ENDIAN]&0x7FF00000)
		((int32*)&r)[!BIG_ENDIAN]+= ((uint32*)&r)[BIG_ENDIAN]>>31;
	((int32*)&r)[BIG_ENDIAN] = 0;
	return *(double*)&r;
}
#endif

/*
 entry bits, encoded after every data:
  #8-15: type encoded or top bits of a small number
	(int8 when compiled against EVAL_FLOATING)
  #7: interrogation (`?`), or already packed (doesn't apply here)
  #6: with interrogation, else (`:`)
  #0-5: opcode
*/

/* these functions assume the real portion is in the first 8 bytes */

#if sizeof_evalvalue>7
 #define SKIP_UNPACK(a,o)	do a-=((o&0xFF00u)!=(EXT_EVAL<<8) ? 1:sizeof_evalvalue/2); while(0)
#else
 #define SKIP_UNPACK(a,o)	do a-=(o>(EXT_EVAL<<8)+0xFF) ? 0:2; while(0)
#endif

/* packing and unpacking functions for evalvalues */
static evalent *unpack_eval(evalent *a,
	t_ceval o,
	evalvalue *v)
{
#if sizeof_evalvalue>7
	if ((o&=0xFF00u)!=(EXT_EVAL<<8))
	{
		((int32*)v)[!!BIG_ENDIAN] = 0;
		((int32*)v)[!BIG_ENDIAN] = (((uint32)o<<8)|*--a)<<8;
	#if sizeof_evalvalue==16
	 {
		double		u = *(double*)v;

		evrset(*v,u);
	 }
	#endif
		return a;
	}
	a-= sizeof_evalvalue/2;
	*v = *(const evalvalue*)a;
	((evalent*)v)[!!BIG_ENDIAN*3] = a[0]; /* ents first in LSB */
	((evalent*)v)[!!BIG_ENDIAN+1] = a[1];
#else
	if (o>(EXT_EVAL<<8)+0xFF)
	{
		*v = o>>8;
		return a;
	}
	((int16*)v)[!!BIG_ENDIAN] = *(a-=2);
	((int16*)v)[!BIG_ENDIAN] = a[1];
#endif
	return a;
}

static evalent *pack_eval(evalent *a,
	evalent *h,
	const evalvalue *v)
{
	uint32		b = ((uint32*)v)[BIG_ENDIAN ? 0:sizeof(evalvalue)>7];

	*h = (uint8)EXT_EVAL<<8;
#if sizeof_evalvalue>7
	if ((b>>24)!=EXT_EVAL && !(((int32*)v)[!!BIG_ENDIAN*3]|(b&0xFF))
 #ifdef HAVE_COMPLEX_EVAL
			&& !cimag(*v)
 #endif
			)
	{ /* short-encoding: high 24 bits of double, zero low-fraction */
		*h = ((int16*)v)[!BIG_ENDIAN*3]&0xFF00;
		a[0] = (evalent)(b>>8);
		return a+1;
	}
	((double*)a)[0] = *(double*)v;
	a[0] = ((evalent*)v)[!!BIG_ENDIAN*3]; /* ents first in LSB */
	a[1] = ((evalent*)v)[!!BIG_ENDIAN+1];
 #ifdef HAVE_COMPLEX_EVAL
	((double*)a)[1] = cimag(*v);
 #endif
#else
	if (b-(EXT_EVAL+1)<0xFF)
		{*h = (evalent)(b<<8); return a;}
	a[0] = (evalent)b;
	a[1] = (evalent)(b>>16);
#endif
	return a+ENTS_PER_EVAL;
}

/*
 evaluate_chunk
	Evaluates all operations until after an l-to-r dyad.
	(_op_rml is an r-to-l dyad.)
 NOTE:
	Remember to pass the correct ptr if popping from the stack!
 Result:
	Next stack ptr.
*/
static evalent *evaluate_chunk(evalent o,
	const evalent *x,
	evalent *s,
	evalvalue *v,
	evalvalue (__cdecl geti)(evalvalue))
{
	evalop		c;

	while((c=o&_op_mask)!=_op_ret)
	{
		evalvalue	a;
		native_eval	b;
		evinttype	t;
		int		i;

		if ((evalent)(c-_op_una)<=_op_unb-_op_una)
			evrset(a,0);
		else
			s = unpack_eval(s,o,&a);
		if (v)
		{
			b = *v;
			switch(c)
			{
			#include "evaluate.h"

			default: evrset(a,0);
			}
			*v = a;
		}
		if (s<=x || (c>_op_unb && c!=_op_rml))
			break;
		o = *--s;
	}
end:	return s;
}

/* ---- general line processing */

#ifndef EVAL_STACKSIZE
 #define EVAL_STACKSIZE		(ENTS_PER_EVAL*65+1)
#endif

#define SET_OP(c)		(void)(o=c,h=(evalent)(EXT_EVAL<<8))

#if defined(NO_MIDOPWSP) || INTERP_VERS>0x198
 #define midskipwsp(s,e)
#else
 #define midskipwsp(s,e)	skipwsp(s,e)
#endif

static int opStrPrecedesAS(int ch,
	const char *s,
	const char *e)
{
	switch(ch)
	{
	case '*': case '/': case '%':
		return s>=e ? 0:*s!='=';
	case '(': /* an implied multiply */
#ifdef CHR_DV
	case CHR_DV:
#endif
		return 1;
	case '=':
		return s+2>=e ? 0:*s!='=' || s[1]!='>';

	default:
		return 0;
	}
}

static int opStrPrecedesCSHL(int ch,
	const char *s,
	const char *e,
	evalop op)
{
	switch(ch)
	{
	case '+': case '-':
	case '*': case '/': case '%': return 1;
	case '<': return op<_op_shr || (op>_op_shl && (op>_op_cle || (s+1<e && *s=='<')));
	case '>': return op<_op_shr || (op>_op_shl && (op>_op_cle || (s+1<e && *s=='>')));
	case '^': return op<_op_eor;
	case '&': return (s+1<e && *s=='&') ? op<_op_ban:op<_op_and;
	case '|': return (s+1<e && *s=='|') ? op<_op_bor:op<_op_ior;

	default:
		return 0;
	}
}

const uint8 interp_op_precedence[] =
{
	[_op_and] = _pr_and,
	[_op_ior] = _pr_ior,
	[_op_eor] = _pr_eor,
	[_op_ban] = _pr_ban,
	[_op_bor] = _pr_bor,
	[_op_asr] = _pr_shx,
	[_op_shr] = _pr_shx,
	[_op_shl] = _pr_shx,

	[_op_add] = _pr_ads,
	[_op_sub] = _pr_ads,
	[_op_mul] = _pr_mdv,
	[_op_pow] = _pr_pow,
	[_op_div] = _pr_mdv,
	[_op_dvs] = _pr_mdv,
	[_op_mod] = _pr_mdv,
	[_op_mds] = _pr_mdv,

	[_op_clt] = _pr_ciq,
	[_op_cge] = _pr_ciq,
	[_op_cgt] = _pr_ciq,
	[_op_cle] = _pr_ciq,
	[_op_cne] = _pr_ceq,
	[_op_ceq] = _pr_ceq,
	[_op_cni] = _pr_ceq,
	[_op_cid] = _pr_ceq,

	[_op_cae] = _pr_ceq
};

#if INTERP_VERS>0x199
static const char *get_keyword(const char *s,
	const char *e,
	const uint64 *d,
	evalop *c)
{
	int32		x;
	uint64		h, v;

	if (!d)
		return NULL;
	e = parse_wordhash(s,e,&h), ++s;
	if ((x=e-s)<12u) /* good length? */
		for(; (v=*d)!=0; ++d)
		{
			if ((v>>8)!=h)
				continue;
			*c = (uint8)v;
			return e;
		}
	return NULL;
}

 #define lCon state.conQ
int handleEOLimplied(evalstate *state,
	evalop parsedObj,
	int err)
{
	switch(state->lCur.key)
	{ /* handle line break keyword paths */
	case _lexk_if:
	case _lexk_if_t:
	case _lexk_if_p:
		err = _e_no_expr;
		break;
	case _lexk_then:
	case _lexk_then_b:
	case _lexk_else:
		if (state->lCur.key==parsedObj)
			state->lStack[state->lCount++] = state->lCur;

	default:
		;
	}
	return err;
}
#else
 #define lCon 0
#endif

long evaluate_line(const char *str,
	const char *e,
	int *_err,
	evalstate *_state)
{
	evalent		xstack[EVAL_STACKSIZE];
	evalstate	state;
	evalent		*a = xstack;
	const char	*p, *s = str;
	evalparse	i;
	int		err = 0, ch;
	long		rep = 0;
	evalent		h; /* operator to be set, high opcode bits */
	evalop		c, o = _op_ret, so; /* previous op, super-op (when applicable) */
	uint8		q = 0, qc = 0; /* interrogation/conditional state; 0x80=then, 0xC0=else */
#if ALLOW_RUNON_WSP
 #define r			1
#else
	uint8		r = 0; /* when run-on mult. is allowed */
#endif
	state.flags = 0;
	evrset(state.result,0);
	if (_state)
		state = *_state;
#if INTERP_VERS>0x199
	qc = (uint8)state.conShift<<7;
#endif
	c = h = 0, state.qlv = 0;
	ch = *s;
	so = _op_rml;
	while(s<e)
	{
		if (state.flags&_within_longcomment)
		{
			if (ch==']' && ++s<e && *s=='#')
				state.flags&=~_within_longcomment;
			goto next;
		}
		switch(ch) /* try not to get stuck */
		{
		case '#':
			++s; /* ignore first '#' for numeric */
			if (s>=e || *s=='#') /* end of line/comment */
				{--s; goto end;}
			if (*s=='[')
				{state.flags|= _within_longcomment; goto next;}
#ifndef ALLOW_NAME_DOT
		case '.':
			if (s>=e)
				{--s; goto bad;} /* period by itself */
			p = s;
#endif
#ifdef CHR_HX
		case CHR_HX:
#endif
		CASE_NUMERIC
			/* value w/o parentheses */
			if (o==_op_obj)
				{err = _e_param_separation; goto end;}
			if (o-_opd_add+0u<=_op_stv-_opd_add)
				{err = _e_lvalue; goto end;}
			if ((p=parse_eval(s,e,&i))==0)
				goto bad;
			s = p;
			if (!(qc&0x80))
				state.result = i;
#if !EVAL_FLOATING
			if (*s==mLocaleNPtChar)
			{ /* skip unused fraction */
				++s;
				skipdig(s,e);
			}
#endif
			goto runon;
		case '!': so = _pr_ceq; /* if dyadic, an equality comparator precedence */
			if (a>xstack && a[-1]-_pr_fn+0u<=_pr_ceq-_pr_fn)
				goto prdyad;
			c = _op_bno;
			if (o==_op_bno)
				{o = _op_dno; goto next;}
			goto unary;
		case '+': so = _pr_ads; /* if dyadic, add or subtract precedence */
			if (o==_op_obj)
			{
				c = _op_add;
				if (a>xstack && a[-1]-_pr_fn+0u<=_pr_ads-_pr_fn)
					goto prdyad;
				goto dyadic;
			}
			c = _op_abs;
			if (o==_op_abs)
				goto next; /* 2013.02.24 */
			goto unary;
		case '-': so = _pr_ads;
			if (o==_op_obj)
			{
				c = _op_sub;
				if (a>xstack && a[-1]-_pr_fn+0u<=_pr_ads-_pr_fn)
					goto prdyad;
				goto dyadic;
			}
			if (o==_op_neg)
				{o = _op_ldv; goto next;}
			c = _op_neg;
			goto unary;
		case '%': so = _pr_mdv; /* if dyadic, mul/div/mod precedence */
			if (o==_op_obj)
			{
				c = _op_mod;
				if (a>xstack && a[-1]-_pr_fn+0u<=_pr_mdv-_pr_fn)
					goto prdyad;
			}
			if (o==_op_mod) /* %% */
				c = _op_mds|0x80, o = _op_obj;
			else if (o!=_op_obj && state.getv)
				goto gid;
			goto dyadic;
		case '*': so = _pr_mdv;
			if (o==_op_obj)
			{
				c = _op_mul;
				if (s+1<e)
					if (s[1]=='*') /* ** */
						c = _op_pow, so = _pr_pow, ++s;
				#ifdef HAVE_COMPLEX_EVAL
					else if (s[1]==')')
					{ /* note that the op is done here at the parser level */
					  /* for vm, this will have to be done differently */
						while(a>xstack && a[-1]-_pr_fn+0u<=_pr_bor-_pr_fn)
						{
							so = *(a-=2);
							if ((a=evaluate_chunk((evalent)so,xstack,a,
									qc&0x80 ? NULL
									:&state.result,state.geti))<xstack)
								goto serr;
						}
						c = _op_cnj;
						state.result = conj(state.result);
						o = _op_obj;
						goto next;
					}
				#endif
				if (a>xstack && a[-1]-_pr_fn+0u<=so-_pr_fn+0u)
					goto prdyad;
			} else if (state.getv)
				goto gid;
			goto dyadic;
		case '/': so = _pr_mdv;
			if (o==_op_obj)
			{
				c = _op_div;
				if (a>xstack && a[-1]-_pr_fn+0u<=_pr_mdv-_pr_fn)
					goto prdyad;
			}
			if (o==_op_div) /* // */
				c = _op_dvs|0x80, o = _op_obj;
			goto dyadic;
		case '<': so = _pr_ciq; /* if dyadic, inequality comparator precedence */
			p = s; /* s saved because of increment and `<<` has different precedence */
			if (o==_op_obj)
			{
				c = _op_clt;
				if (s+1<e && s[1]=='<') /* << */
					c = _op_shl, so = _pr_shx, ++s;
				if (a>xstack && a[-1]-_pr_fn+0u<=so-_pr_fn+0u)
					goto prdyad;
			}
#ifdef ALLOW_BIN_HI_LO
			else
				{c = _op_blo; goto unary;}
#endif
			goto dyadic;
		case '=': so = _pr_ceq;
			p = s;
			if (o==_op_obj)
			{
				c = _op_ceq;
				if (a>xstack && a[-1]-_pr_fn+0u<=_pr_ceq-_pr_fn)
					goto prdyad;
			}
			if (o==_op_ceq) /* == */
				c = _op_cid|0x80, o = _op_obj;
			else if (o==_op_clt) /* <= */
				c = _op_cle|0x80, o = _op_obj;
			else if (o==_op_cgt) /* >= */
				c = _op_cge|0x80, o = _op_obj;
#if 1//def ALLOW_DSTMOD
			else if ((unsigned)o-_op_ior<=_op_mds-_op_ior)
			{ /* _=> */
				o|= 0x20, ++s;
				midskipwsp(s,e);
				if (s>=e || *s!='>')
					goto bad;
				goto next;
			}
#endif
			goto dyadic;
		case '>': so = _pr_ciq;
			p = s;
			if (o==_op_obj)
			{
				c = _op_cgt;
				if (s+1<e)
					if (s[1]=='>')
						c = _op_shr, so = _pr_shx, ++s;
					else if (s[1]==';') /* >; */
					{
						/* group state change; handle precedent-delayed operators */
						while(a>xstack && a[-1]-_pr_fn+0u<=_pr_bor-_pr_fn)
						{
							so = *(a-=2);
							if ((a=evaluate_chunk((evalent)so,xstack,a,
									qc&0x80 ? NULL
									:&state.result,state.geti))<xstack)
								goto serr;
						}
						if (qc&0x80) /* this part not being executed */
							{++s; goto next;}
						if (_err)
							*_err = 0;
						if (_state)
							*_state = state;
						return _e_return;
					}
				if (a>xstack && a[-1]-_pr_fn+0u<=so-_pr_fn+0u)
					goto prdyad;
			}
			if (o==_op_clt) /* <> */
				c = _op_cni|0x80, o = _op_obj;
			else if (o==_op_shr) /* >>> */
				c = _op_asr|0x80, o = _op_obj;
			else if (o==_op_cid) /* ==> */
			{
				o = _op_stv;
				goto next;
			}
#ifdef ALLOW_BIN_HI_LO
			else if (o!=_op_obj)
				{c = _op_bhi; goto unary;}
#endif
			goto dyadic;
		case '&': so = _pr_and; /* if dyadic, bit-and comparator precedence */
			p = s;
			if (o==_op_obj)
			{
				c = _op_and;
				if (s+1<e && s[1]=='&') /* && */
					c = _op_ban, so = _pr_ban, ++s;
				if (a>xstack && a[-1]-_pr_fn+0u<=so-_pr_fn+0u)
					goto prdyad;
			}
			goto dyadic;
		case '|': so = _pr_ior; /* if dyadic, bit-or comparator precedence */
			p = s;
			if (o==_op_obj)
			{
				c = _op_ior;
				if (s+1<e && s[1]=='|') /* || */
					c = _op_bor, so = _pr_bor, ++s;
				if (a>xstack && a[-1]-_pr_fn+0u<=so-_pr_fn+0u)
					goto prdyad;
			}
			goto dyadic;
		case '^': so = _pr_eor; /* if dyadic, bit-exor comparator precedence */
			c = _op_eor;
			if (o==_op_obj && a>xstack && a[-1]-_pr_fn+0u<=_pr_eor-_pr_fn)
				goto prdyad;
			goto dyadic;

		case ';': /* break */
			if (state.flags&_disallow_breaks)
				goto bad;
			state.qlv = 0;
			/* group state change; handle precedent-delayed operators */
			while(a>xstack && a[-1]-_pr_fn+0u<=_pr_bor-_pr_fn)
			{
				so = *(a-=2);
				if ((a=evaluate_chunk((evalent)so,xstack,a,
						qc&0x80 ? NULL
						:&state.result,state.geti))<xstack)
					goto serr;
			}
			++s; /* incomplete expression tests */
			if (o==_op_ret)
			{
				if (q&0xC0)
					goto end;
			} else if ((int8)o>0 || (o!=_op_obj
					/* NOTE: the cast is to rid the assignment warning */
					/* i will not be used */
					&& (a=unpack_eval(a,h,(evalvalue*)&i))!=xstack))
				goto end;
		#if INTERP_VERS>0x198
			else
				q&=~0x80, qc = q|lCon;
		#endif
			o = _op_ret;
			ch = *s;
			break;
		case ',': /* parameter/list separator */
#ifdef ALLOW_LISTS
			/* group state change; handle precedent-delayed operators */
			while(a>xstack && a[-1]-_pr_fn+0u<=_pr_bor-_pr_fn)
			{
				so = *(a-=2);
				if ((a=evaluate_chunk((evalent)so,xstack,a,
						qc&0x80 ? NULL
						:&state.result,state.geti))<xstack)
					goto serr;
			}
			c = _op_lph;
			goto dyadic;
#else
			goto end;
#endif
		case ')':
			if (o!=_op_obj)
				goto end;
			if (a<=xstack)
				{err = _e_stack_error; goto end;} /* after op., empty or underflow */
			/* execute precedent-delayed operators */
			while(a[-1]-_pr_fn+0u<=_pr_bor-_pr_fn) //_pr_bor-
			{
				o = *(a-=2);
				if (a<=xstack)
					{err = _e_stack_error; goto end;} /* after op., empty or underflow */
				if ((a=evaluate_chunk((evalent)o,xstack,a,
						qc&0x80 ? NULL
						:&state.result,state.geti))<=xstack)
					goto serr;
			}
			o = *--a;
			if (o==_pr_cnd) /* 2024.05.20 */
				o = *--a;
			q = (uint8)(o&~_op_mask); /* restore interrogation state */
			qc = q|lCon;
			++s;
			if ((a=evaluate_chunk((evalent)o,xstack,a,
					qc&0x80 ? NULL
					:&state.result,state.geti))<xstack)
				goto bad;
			SET_OP(_op_obj);
#if !ALLOW_RUNON_WSP
			r = 1;
#endif
			ch = *s;
			break;
		case '@': /* indirection */
			if (state.geti)
				{c = _op_ldi; goto unary;}
#ifdef ALLOW_NAME_DOT
		case '.':
#endif
		case '_':
		CASE_ALPHABETIC
			if (o-_opd_add+0u<=_op_stv-_opd_add)
			{ /* set value for identifier(s); 2011.06.11 */
				if (!state.setv) /* a little weird getting setv but not getv */
					{err = _e_handler_error; goto end;}
				if (o<_op_stv && !(qc&0x80))
				{ /* p set for disabled case */
					if (!state.getv) /* no routine, two in a row or no value */
						{err = _e_handler_error; goto end;}
					if ((p=state.getv(s,e,qc&0x80 ? NULL
							:&state.result,state.vbuf))==0)
						goto nosymb;
					if ((a=evaluate_chunk((evalent)(o|h),xstack,a,
							qc&0x80 ? NULL
							:&state.result,state.geti))<xstack)
						goto bad;
				} else if ((a=unpack_eval(a,h,&state.result))<xstack)
					{err = _e_stack_error; goto end;}
				if (qc&0x80) /* skip term - won't be created */
					s = p = parse_wordbreak(s,e,NULL,UCFDIG_0,UCFUNDSC);
				else if ((p=state.setv(s,e,state.result,&err,state.vbuf))==0)
				{
					if (!err)
						err = _e_no_identifier;
					goto end;
				}
			} else
			{ /* get value for identifier(s) */
			gid:	if (o==_op_obj)
					{err = _e_param_separation; goto end;}
				if (!state.getv) /* no routine, two in a row or no value */
					{err = _e_handler_error; goto end;}
			#if INTERP_VERS>0x199
				p = s;
				if (state.lKeys && (s=get_keyword(s,e,state.lKeys,&c))!=0)
				{
					if (o==_op_rml)
					{
						o = _op_obj;
						//a = unpack_eval(a,h,&state.result);
						SKIP_UNPACK(a,h);
						if (a<xstack)
							{err = _e_stack_error; goto end;}
					}
					switch(c)
					{
					case _op_dno:
					case _op_bno:
					case _op_srt:
					case _op_abs:
					case _op_neg:
					case _op_not:
						if (o==_op_obj)
							{err = _e_param_separation, s = p; goto end;}
						goto xunary;
					case _op_add:
					case _op_sub:
					case _op_mul:
					case _op_pow:
					case _op_div:
					case _op_dvs:
					case _op_mod:
					case _op_mds:
					case _op_clt:
					case _op_cge:
					case _op_cgt:
					case _op_cle:
					case _op_cne:
					case _op_ceq:
					case _op_cni:
					case _op_cid:
						so = interp_op_precedence[c];
						if (o==_op_obj && a>xstack && a[-1]-_pr_fn+0u<=so-_pr_fn+0u)
							goto prdyad;
						goto dyadic;
					case _lexk_then:
					case _lexk_then_b:
						if (o!=_op_obj)
							{err = _e_no_expr; goto bad;}
						while(a>xstack && a[-1]-_pr_fn+0u<=_pr_cnd-_pr_fn)
						{
							so = *(a-=2);
							if ((a=evaluate_chunk((evalent)so,xstack,a,
									qc&0x80 ? NULL
									:&state.result,state.geti))<xstack)
								goto serr;
						}
						if (state.lCur.key-_lexk_if+0u>_lexk_if_p-_lexk_if+0u)
							{err = _e_superfluous_keyword; goto bad;}
						if (c==_lexk_then)
						{
							if (evisnil(state.result))
								state.conQ = 0x80;
						} else if (evveq(state.result,state.trueVal))
							state.conQ = 0;
						else if (evveq(state.result,state.falseVal))
							state.conQ = 0x80;
						else
							{err = _e_expected_boolean; goto bad;}
						qc = q|lCon;
						goto setkey;
					case _lexk_else:
						if (state.lCur.key!=_lexk_then)
							{err = _e_superfluous_keyword; goto bad;}
						state.conQ^= 0x80;
						state.conShift>>= 1; /* technically an end */
						if (state.lCount<=0)
							goto sterr;
						--state.lCount;
					setkey:	state.lCur.key = c;
						evrset(state.result,0);
						break;

					case _lexk_if_t:
					case _lexk_if_p:
						if (state.lCur.key-_lexk_if+0u<=_lexk_if_p-_lexk_if+0u)
							{err = _e_general_error; goto bad;}
						if (state.conShift&1ull<<63)
							{err = _e_stack_error; goto end;}
						skipwsp(s,e);
						if (c==_lexk_if_p && *s!='(')
							{err = _e_general_error; goto bad;}

					default:
						if (a!=xstack)
							goto bad;
						if (state.lCount>EVALLANG_STACKSIZE-2)
							goto sterr;
						state.lCur.key = c;
						state.lCur.offset = s-str+state.lineOffs;
						break;
					}
					ch = *s;
					continue;
				}
				s = p;
			#endif
				p = state.getv(s,e,qc&0x80 ? NULL
					:&state.result,state.vbuf);
				goto evo;
			}
			o = _op_obj, s = p;
			ch = *s;
			break;
		case '?': /* interrogative conditional */
			/* group state change; handle precedent-delayed operators */
			while(a>xstack && a[-1]-_pr_fn+0u<_pr_cnd-_pr_fn)
			{
				so = *(a-=2);
				if ((a=evaluate_chunk((evalent)so,xstack,a,
						qc&0x80 ? NULL
						:&state.result,state.geti))<xstack)
					goto serr;
			}
			if (state.flags&_allow_qrepeat && s+1<e && s[1]=='>')
			{
				s+= 2;
				if (evisnil(state.result))
					{rep = _e_repeat_end; goto end;}
				rep = o==_op_ret ? _e_repeat_last:_e_repeat;
				o = _op_ret;
				ch = *s;
				break;
			}
			so = _pr_cnd;
			if (o!=_op_obj || q&0x40) /* oper after oper or interrogation open */
			{
				err = o==_op_obj ? _e_open_interrogator:_e_general_error;
				goto bad;
			}
			/* eval. 'else' range only if 0; otherwise 'then' only */
			q|= 0xC0; /* | and & operations used due to POCC bugs SIGH */
			qc = q|lCon;
			++state.qlv;
			if (!evisnil(state.result))
			{
				if (state.flags&_CPEMDAS_ordering)
				{
					if (a>xstack && (so=a[-1])==_pr_cnd)
						goto cskp;
					*a++ = (evalent)_pr_cnd;
				}
				q&=~0x80;
				state.result = 0; /* value parsed; clear */
			}
			qc = q|lCon;
		cskp:	o = _op_ret;
			goto next;
		case ':':
			if (o!=_op_obj /* oper after oper, or unimplemented */
					|| !(q&0x40) || !state.qlv)
				{err = _e_incomplete_error; goto bad;}
			/* group state change; handle precedent-delayed operators */
			while(a>xstack && a[-1]-_pr_fn+0u<=_pr_bor-_pr_fn)
			{
				so = *(a-=2);
				if ((a=evaluate_chunk((evalent)so,xstack,a,
						qc&0x80 ? NULL
						:&state.result,state.geti))<xstack)
					goto serr;
			}
			--state.qlv;
			q^= 0xC0; /* clears bit #6 */
			if (a>xstack && (so=a[-1])==_pr_cnd)
				q|= 0x80;
			qc = q|lCon;
			o = _op_ret;
			goto next;
		case 042:
			if (!(state.flags&_dquote_literals))
				goto bad;
		case 047:
			if (o==_op_obj)
				goto bad;
			p = s;
			if ((s=parse_eval_literal(s+1,e,*s,&state.result))==NULL)
				{s = p; goto bad;}
			goto runon;
#if 1//def ALLOW_ESCNL
		case '\\':
			if (++s>=e || *s-10u>3u || *s=='\f')
				goto bad;
			goto next;
#endif
		case '~': p = s;
			if (o==_op_obj && s+1<e && *s=='='
					&& a>xstack && a[-1]-_pr_fn+0u<=_pr_ceq-_pr_fn)
				goto prdyad;
			c = _op_not;
		unary:	p = s, ++s;
			midskipwsp(s,e);
			if (c==_op_bno && s<e && *s=='=')
				{c = _op_cne; goto dyadic;}
			if (state.flags&_allow_approxeq && s<e && *s=='=')
				{c = _op_cae; goto dyadic;}
			if (o==_op_obj)
			{
				err = _e_param_separation;
				s = p;
				goto end;
			}
			if (c==_op_not && s<e && *s=='/') /* ~/ */
				{c = _op_srt, ++s;}
	#if INTERP_VERS>0x199
		xunary:
	#endif
			if (a>=xstack+EVAL_STACKSIZE-1)
				{err = _e_stack_error; goto end;} /* stack overflow */
			*a++ = (evalent)(o|h);
			o = c;
			ch = *s;
			break;
		case ' ':
		whitespace:
			skipwsp(s,e);
			if (state.flags&_whitespace_breaks)
				r = 0;
			ch = *s;
			break;
		default:
			if (ch-10u<4u && ch!='\f')
				goto end; /* 1.9.9: end of line just terminates */
#ifdef CHR_NE
			if (ch==CHR_NE)
				{c = _op_cne; goto dyadic;}
#endif
#ifdef CHR_LE
			if (ch==CHR_LE)
				{c = _op_cle; goto dyadic;}
#endif
#ifdef CHR_GE
			if (ch==CHR_GE)
				{c = _op_cge; goto dyadic;}
#endif
#ifdef CHR_AE
			if (ch==CHR_AE)
				{c = _op_cae; goto dyadic;}
#endif
#ifdef CHR_RA
			if (ch==CHR_RA)
				{c = _op_srt; goto unary;}
#endif
#ifdef CHR_SL
			if (ch==CHR_SL)
				{c = _op_shl; goto dyadic;}
#endif
#ifdef CHR_SR
			if (ch==CHR_SR)
				{c = _op_shr; goto dyadic;}
#endif
#if CHR_MN!='-'
			if (ch==CHR_MN)
				{c = _op_neg; goto unary;}
#endif
#ifdef CHR_ML
			if (ch==CHR_ML)
				{c = _op_mul; goto dyadic;}
#endif
#ifdef CHR_DV
			if (ch==CHR_DV)
				{c = _op_dve; goto dyadic;}
#endif
#if defined(CHR_PI) && defined(HAVE_COMPLEX_EVAL) && INTERP_VERS>0x198
			if (ch==CHR_PI)
			{
				if (o==_op_obj)
					{err = _e_param_separation; goto end;}
				if (!state.getv) /* no routine, two in a row or no value */
					{err = _e_handler_error; goto end;}
				p = "Pi";
				p = state.getv(p,p+2,qc&0x80 ? NULL
					:&state.result,state.vbuf);
				if (p)
					p = s+1;
				goto evo;
			}
#endif
#ifdef CHR_NB
			if (ch==CHR_NB)
				goto whitespace;
#endif
			if (ch=='\t')
				goto whitespace;
			err = _e_unknown_symbol;
			goto bad;

		evo:	if (!p)
			{
				o = _op_obj;
			nosymb:	err = *s=='*' || *s=='%'
					? _e_general_error:_e_no_identifier;
				goto end;
			}
			s = p;
		runon:	if (!(state.flags&_whitespace_breaks))
				skipwsp(s,e);
			ch = 0;
			if (s<e)
			{
				ch = *s; /* test: 152==>x; 0|=>x <=1 ? 1>; ;2 */
				if (state.flags&_CPEMDAS_ordering)
					switch(c)
					{
					case _op_and: case _op_eor: case _op_ior:
					case _op_ban: case _op_bor:
					case _op_asr: case _op_shr: case _op_shl:
					case _op_clt: case _op_cge: case _op_cgt: case _op_cle:
					case _op_cne: case _op_ceq: case _op_cni: case _op_cid:
					case _op_cae:
						if (opStrPrecedesCSHL(ch,s+1,e,c))
						{
							*a++ = (evalent)(c|h);
							*a++ = interp_op_precedence[c];
							o = _op_obj;
							s = p; /* saved due to potential incr. */
							ch = *s;
							continue;
						}
						break;
					case _op_add:
					case _op_sub:
						if (opStrPrecedesAS(ch,s+1,e))
						{
							*a++ = (evalent)(c|h);
							*a++ = _pr_ads;
							o = _op_obj;
							ch = *s;
							continue;
						}
					default: break;
					case _op_mul:
					case _op_div:
					case _op_dvs:
					case _op_mod:
					case _op_mds:
						if (ch=='*' && s+1<e && s[1]=='*')
						{ /* only function, arrays and exponentiation */
							*a++ = (evalent)(c|h);
							*a++ = _pr_mdv;
							o = _op_obj;
							ch = *s;
							continue;
						}
					}
				if (ch=='(')
				{ /* only allow known immutables here */
				  /* needed to soundly remove confusion in func calls */
					if (state.flags&_CPEMDAS_ordering)
						ch = '*', --s;
					else
						goto mparen;
				} else if (((ch&~32)-'A')+0u<='Z'-'A'
 #if defined(CHR_PI) && defined(HAVE_COMPLEX_EVAL) && INTERP_VERS>0x198
						|| ch==CHR_PI
 #endif
 						)
					goto evu;
				else if (ch=='@' || (ch=='~' && s+1<e && s[1]=='/')
#if defined(HAVE_COMPLEX_EVAL) || INTERP_VERS>0x197
 #ifdef CHR_RA
						|| ch==CHR_RA
 #endif
#endif
						)
				{ /* superceded multiply/other */
					if (state.flags&_CPEMDAS_ordering)
						ch = '*', --s;
					else
					{
					evu:	if (a>=xstack+EVAL_STACKSIZE-ENTS_PER_EVAL-1)
							{err = _e_stack_error; goto end;} /* overflow */
						*a++ = (evalent)(o|h), o = _op_rml;
						a = pack_eval(a,&h,&state.result);
						ch = *s;
						break;
					}
				}
			}
			if ((a=evaluate_chunk((evalent)(o|h),xstack,a,qc&0x80 ? NULL
					:&state.result,state.geti))<xstack)
				{err = _e_stack_error; goto end;}
			o = _op_obj;
			if (state.flags&_whitespace_breaks)
				r = 1;
			break;
		case '(':
			so = _op_rml;
			if (o==_op_obj)
			{ /* check for implied multiply */
				if (!r) /* concat not implemented */
					{err = _e_param_separation; goto end;}
			mparen:	if (a>=xstack+EVAL_STACKSIZE-ENTS_PER_EVAL-2)
					{err = _e_stack_error; goto end;} /* overflow */
				if (o==_op_obj)
					o = _op_ret;
				*a++ = (evalent)(o|h), o = so;
				a = pack_eval(a,&h,&state.result);
			} else if (a>=xstack+EVAL_STACKSIZE-1)
				goto sterr;
			o&= _op_mask, o|= q; /* save interrogation state */
			*a++ = (evalent)(o|h);
			SET_OP(_op_ret); c=_op_ret;
			//qsc = 0;
			if ((q&=0x80)==0)
				evrset(state.result,0); /* init. interrogation result */
			qc = q|lCon;
			goto next;
		prdyad:	so = *(a-=2);
			if ((a=evaluate_chunk((evalent)so,xstack,a,qc&0x80 ? NULL
					:&state.result,state.geti))<xstack)
				{sterr: err = _e_stack_error; goto end;}
		dyadic:	if (o!=_op_obj)
			{ /* dyadic after dyadic or unary */
			bad:	if (!err)
					err = _e_general_error;
				goto end;
			}
			if (a>=xstack+EVAL_STACKSIZE-ENTS_PER_EVAL)
				{err = _e_stack_error; goto end;} /* stack overflow */
			if (!(c&0x80)) /* if not already packed */
				a = pack_eval(a,&h,&state.result);
			o = c&_op_mask;
		next:	ch = *++s;
		}
	}
end:	if (((int8)o>=0 && o!=_op_ret || a>xstack) && err>=0)
	{
		if (o!=_op_obj)
			goto serr;
		/* execute precedent-delayed operators */
		while(a>xstack && (o=*--a)-_pr_fn+0u<=_pr_cnd-_pr_fn)
		{
			if (o==_pr_cnd)
				continue;
			o = *--a;
			if ((a=evaluate_chunk((evalent)o,xstack,a,qc&0x80 ? NULL
					:&state.result,state.geti))<xstack)
				goto serr;
		}
		if (a>xstack)
		serr:	err = o-_opd_add+0u<=_op_stv-_opd_add ? _e_no_identifier:_e_incomplete_error;
	}
#if INTERP_VERS>0x199
	if (err>=0 && (state.flags&(_implied_block_breaks|_within_longcomment))==_implied_block_breaks)
		err = handleEOLimplied(&state,c,err);
#endif
	if (_err)
		*_err = err;
	if (_state)
		*_state = state;
	if (state.flags&_allow_qrepeat)
	{
		if (!err && rep)
			return rep;
		if (o==_op_ret)
			return 0; /* clarify, empty line */
	}
	return s-str;
}

#undef r

static const char *quotedtext_scan(const char *s,
	const char *e,
	uint8 v)
{
	if (e<=s)
		return 0;
	do
	{
		uint8	c = *s++;

		if (c=='\\') /* escape; skip next char */
			++s;
		else if (c==v)
			return s;
	} while(s<e);

	return 0; /* no closing quote */
}

const char *parse_parameter(const char *text,
	long length,
	char *outstr,
	int *outerr,
	const char *(__cdecl getv)(const char *s,
		const char *e,
		evalvalue*,
		uint8 **vbuf),
	const char *(__cdecl setv)(const char *s,
		const char *e,
		evalvalue,
		int *err,
		uint8 **vbuf),
	uint8 **vbuf,
	evalvalue (__cdecl geti)(evalvalue),
	uint8 allowSQ)
{
	const char	*s, *e;
	evalstate	state;
	long		r;
	int		err;

	if (outstr)
		*outstr = '\0';

	if ((s=text)==0)
		return 0;
	if (length<=0)
		goto blank;
	e = length+s;
	skipwsp(s,e);

	if (s==e)
	{
	blank:	if (outstr)
			*outstr = ' ';
		goto end;
	}

	memset(&state,0,sizeof(state));
	state.flags = _whitespace_breaks|_disallow_breaks;
	state.getv = getv;
	state.setv = setv;
	state.geti = geti;
	state.vbuf = vbuf;
	if (*s=='\042') /* text - double quote */
		{s = quotedtext_scan(++s,e,'\042'); goto end;}
	else if (*s=='\047' && allowSQ) /* text - single quote */
		{s = quotedtext_scan(++s,e,'\047'); goto end;}
	r = evaluate_line(s,e,&err,&state);
	if (outerr)
		*outerr = err;
	if (r==_e_return)
		s = e;
	else
		s+= r;
	if (err<0)
		goto end;
	if (outstr)
		(void)sprint_evalue(outstr,state.result,16);
end:	return s;
}
