/*
 var.c
  interpreter 1.9.9: variable storage and sprint

  Created: 2013.11.23
  Revised: 2024.05.20

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#if TARGET_OS_MAC
 #include <fp.h>	/* for nan */
 #define NAN nan("255")
#else
 #include <math.h>
#endif

#define INTERP_BASIC_SCRIPT_VARS
#include "../include/interpreter.h"
#include "../include/ftrig.h"
#include "../include/my_cseries.h"

/* ---- globals */

int8		_typSzArray[32] =
{
/*-32*/  2, 2, 1, 1, /* _itck16, _ifix16, _uflt8, _styl8 */
	 4, 2, 1, 1, /* _ufrc32, _ufrc16, _ufrc8, _alpha8 */
	 4, 4, 4, 4, /* _ufract, _ifract, _ufixed, _ifixed */
	10, 4, 2, 2, /* _ext80, _dbl32, _hpf16, _flt16 */
/*-16*/	 8, 4, 2, 1, /* _double, _float, _ifrc16, _bool */
	 8, 4, 2, 1, /* _hint64, _hint32, _hint16, _hint8 */
	 8, 4, 2, 1, /* _uint64, _uint32, _uint16, _uint8 */
	 8, 4, 2, 1  /* _int64, _int32, _int16, _int8 */
};

/* ---- general */

int sprint_fphex(char *b,
	double_t v)
{
	uint64		m = *(uint64*)&v;
	int32		h = (int32)(m>>32);
	int		r = 0, p = (int)(h>>20&2047);

	if (h<0)
		b[r++] = '-';
	m<<= 12;
	if (p==2047)
		{memcpy(b,m ? "nan":"inf",4); return r+3;}
	h = !!p;
	if (((p-=1023)!=0 && h) || m)
	{
		b[r++] = '0', b[r++] = 'x', b[r++] = '1';
		if (!h) /* optimize subnormal */
			h = cntlz64(m), p-= h, m<<= h, m+= m;
		if (m)
		{
			b[r++] = mLocaleNPtChar;
			do
				b[r++] = _interp_defbasechars[m>>60];
			while(m<<=4);
		}
		if (p)
		{
			b[r++] = 'p';
			r+= makebasestring(b+r,NULL,p,10,0,'-');
		}
	} else
		b[r++] = (char)(h+'0');
	b[r] = 0;
	return r;
}

static const uint64	ipow10tab[21] =
{
	1,10,100,1000,
	10000,100000,1000000,10000000,
	100000000,1000000000,10000000000,100000000000,
	1000000000000,10000000000000,100000000000000,1000000000000000,
	10000000000000000,100000000000000000,1000000000000000000,
	1250000000000000000ull /* downshifted 3 to fit */
};

int sprint_float(char *out, /* needs at least 25-31 */
	s_float80 v,
	int p,
	int res,
	char ptchar,
	char sepchar,
	char signchar)
{
	s_float80	u;
	char		*d = out, *l;
	uint64		n;
	int32		z = 0;
	int32		flag = p>>31, frrem, fgpl;
	int		x, r;//, is1;

	if (!signchar)
		signchar = '-';
	if (v.sgnexp<0)
		*d++ = signchar!='+' ? signchar:'-';
	else if (signchar=='+')
		*d++ = '+';
	if ((r=res)-1u>19-1)
		r = 20;
	if ((x=v.sgnexp&=32767)==0)
	{
		x =-cntlz64(v.man);
		if (x==-64)
			z = MK4CC('0',ptchar,'0',0);
	} else if (x==32767 || !(v.man&1ull<<63))
		z = (uint64)(v.man*2) ? MK4CC('n','a','n',0):MK4CC('i','n','f',0);
	if ((*(int32*)d=z)!=0)
	{
		if (z&MK4CC(0x40,0,0,0) || p-1u<r-1u)
			{d+= 3; goto end;}
		if (!p || p>r)
			{*++d = 0; goto end;}
	}
	/* get log10 */
	x = (x-0x3FFF)*(int)(M_LOG10_2*(1<<16)+1)>>16;
	u = f80_scal10(v,-x);
	//is1 = u.man==(1ull<<63);
	z = *(const int32*)((const int8*)&u+!BIG_ENDIAN*6);
	if (z+0u>=0x4002A000u)
		++x;//, is1 = u.man==(10ull<<60);
	else if (z+0u<0x3FFF8000u)
		--x;//, is1 = u.man==(10ull<<60);
	if (!(p^=flag))
		p = 6;
	else if (p>r)
	{ /* in low resolution case, compact to most compact number;
	    it's simplest to go by log10(v) plus room for partial top digits;
	    a top digit of 1 means lower digits occupy most of the resolution
	  */
		if (x<-4932)
			p = x+4932+20;
		else if (r<18)
			if (r>7)
			{
				p = x+308+17;
				// precision rounding factors below are very different
				// doubles can represent 10000000000000005e-324
				//r = MAX(p,2)-1;
			} else /* untested */
				p = x+39+7;
		p = PIN(p,0,r);
	}
	p-= !!p;
	flag|= x+5u>5+9u; /* notation if outside [-5,9] */
	if (!flag && p<x)
		p = x;
	frrem = p-x;
	/* scale to uint64 (with possible compression) */
	//if (frrem>0)
	//	v = f80_div(v,f80_scal10(f80_one,-frrem));
	//else
	 v = f80_scal10(v,frrem);
	n = v.man;
	flag&=~2;
	res = 0x3FFF+62-v.sgnexp;
	if (res<0) /* in 20-digit mode, res will be in the range of [0,3] */
		res =~res, v.sgnexp = 0;
	else
	{
		flag|= (int)(v.man<<3>>res&6); /* to avoid too much rounding, two more bits used */
		n = v.man>>res, res = 0;
	}
	/* to eliminate the need to copy from buffer, accurately predict digit counts */
	z = 0;
	if (flag&1)
		frrem = p;
	else if (x<0)
		p-= x, z = x; /* need zero on top in non-notation under 1.0 */
	if (!p && n>18)
		++p, ++frrem, ++x;
	d+= frrem && frrem<=p; /* anticipate when decimal point will show up */
	l = d;
	z+= p;
	fgpl = 0;
	if (p)
	{
		uint64		onestr = ipow10tab[v.sgnexp ? MIN(z,r):res ? 19:18];
		int		rs = !v.sgnexp && x+13u<20+13u; //15u

		if (v.sgnexp)
			onestr+= onestr;
		if (p>r-2)
		{ /* max precision: round to compact number that matches via parse_number */
			uint64	testn = n;
			int	resn = res;

			if (v.sgnexp)
				z = (int)(n%200u); /* low two digits with half bit */
			else
			{
				++resn;
				z = (int)(n%2000u)*2;
				z = (int)(((uint32)z%(200u>>res))<<res);
				if (res)
					testn>>= 3-res;
			}
			r = z%20u; /* low digit with half bit */
			//printf("\nr%d z%d n %llu/%llu res%d\n",r,z,testn,onestr,res);
			if ((testn>>3)<onestr) /* top digit <8 */
			{ /* at least three levels of precision required - using top digit */
			  /* doubles can represent 1.90000000000000013 and 1.88888888888888884 */
			  /* and rounding max to 1.797693134862316e+308 as double makes inf */
			  /* numbers like 2.4444444444444446 also can't be rounded to 16 digits */
			  /* others: 1e-93, 1e-102, 1e-104, 8.2222, 6.40293, 6.402917, 20/9, 22/9 */
				uint32		rr = r*8+(flag&6);

				testn>>= 1; /* onestr now twostr in comparison */
				if (rr<(((testn>>1)<onestr ? testn<onestr
						? 20u:46u:67u)>>rs)) /* low values for top<{2,4,8} */
				{
					if (res==3 && z>183 && !rs)
						++n, n&=~1;
					else
						n-= r>>resn;
				} else if (rr>(((testn>>1)<onestr ? testn<onestr
						? res ? 127u:134u /* high values for top */
						:res ? 95u:102u:89u)<<rs))
					n+= (20-r)>>resn;
				//else
				//	n+= 5-r; -- fails transcoding on `4/3`
			} /*else if (is1)
				n+= (n&3)==3 ? 1:-(n&3);*/
			else if (!rs)
				if (z*8+(flag&6)>1437) //1449 //739*2 //93*8
					n+= (200-z)>>resn;
				else if (r<4*2 || z<18)
					n-= r>>resn;
				else if (r>7*2)
					n+= (20-r)>>resn;
		}
		if (res)
		{ /* 20-digit case */
			if (res==3 && n>=onestr*10)
				++p, ++x, ++frrem; /* top digit carried */
			if (sepchar && !(flag&1) && x>2)
				r = x/3u, d+= r, l+= r;
			r = (int)n, z = 200>>res;
			n/= (unsigned)z, r-= (int)((uint32)n*z);
			r<<= res, z = r/10u;
			r-= z*10u;
			d[p] = _interp_defbasechars[r-(r>4)];
			n*= 20, n+= z;
			++l, --p, --frrem;
			if (r && rs)
				flag|= (int)(1u<<31);
			fgpl = 2;
		} else
		{
			if (v.sgnexp)
			{
				n = (n+1)>>1;
				if (n>=onestr*5) /* just downshifted n */
					++p, ++x, ++frrem; /* top digit carried */
			}
			if (sepchar && !(flag&1) && x>2)
				r = x/3u, d+= r, l+= r;
		}
		l+= p;
		do
		{
			r = (int)n;
			n/= 10, r-= (int)(n*10);
			d[p] = _interp_defbasechars[r];
			if (frrem>0)
			{
				if (!--frrem)
					--d, d[p] = ptchar;
				if (r || flag<0)
					flag|= (int)(1u<<31);
				else
					l = d+p-1;
			} else if (sepchar && --frrem==-3)
				(--d)[p] = sepchar, frrem = 0;
		} while(--p);
	} else
		n = (n+1)>>1;
	*d = _interp_defbasechars[(int)n];
	d = l;
	d+= *d!=ptchar;
	if (flag&1)
	{ /* add scientific exponent */
		*d++ = 'e';
		d+= makebasestring(d,NULL,x,10,0,signchar);
	}
	*d = 0;
end:	return d-out;
}

/* ---- basic string storage */

int deltobj(tobj *o)
{
	if (o->z<=0 || !o->s)
		{o->s = NULL, o->z = 0; return 0;}
	free(o->s);
	o->s = NULL, o->z = 0;
	return 1;
}

int unrtobj(tobj *o)
{
	char		*d;
	int32		z;

	if (!o->s)
		{o->z = 0; return 0;}
	if ((z=o->z)>=0)
		return 0;
	if ((d=malloc(z&=0x7FFFFFFF))==NULL)
		return _e_memory_error;
	memcpy(d,o->s,z);
	o->s = d;
	return 1;
}

/* ---- scanning */

int32 get_hbuf_entry(const char *n,
	int z,
	uint8 h,
	const uint8 *u,
	uint32 dz)
{
	const uint8	*l;

	if ((l=u)==NULL)
		return ~0;
	for(++dz; *l++; l+=*l,l+=dz)
	{
#ifndef PARSE_NO_CONSTANTS
		int		k = *l;

		if (l[k]=='~')
			--k;
		if (k==z && h==l[-1] && !memcmp(l+1,n,k))
			return 1+*l+l-u;
#else
		if (*l==z && h==l[-1] && !memcmp(l+1,n,z))
			return 1+z+l-u;
#endif
	}
	return u-l; /* becomes ~total */
}

int32 get_hname_index(const char *n,
	int z,
	uint8 h,
	const uint8 *u,
	uint32 tz)
{
	const uint8	*l, *e;
	int32		c;

	if ((l=u)==NULL)
		return ~0;
	for(e=l+tz,c=0; l<e; l+=*l,++c)
		if (*++l==z && h==l[-1] && !memcmp(l+1,n,z))
			return c;
	return u-l; /* becomes ~total */
}

/* ---- calculation */

int32 elem_tsize(int32 typ)
{
	int32		t = 0, z;
	int		h = 9;

	if (typ>=0)
		return typ;
	z = typ|~511;
	goto start;
	do
	{
		h = 5, z = (typ|(~287|~511))&~256;
	start:	if (--z>-257)
			t-= z;
		else if ((z+=288)<0)
			return _e_op_error;
		else
			t+= _typSzArray[z];
	} while((typ>>=h)<~30u);
	return t;
}

double cvt_tdata(int32 z,
	const void *p)
{
	char		b[8];

	if (z<-255)
		switch(z)
		{
		case _hint8:
		case _uint8:
		case _styl8: return (double)*(const uint8*)p;
		case _hint16:
		case _uint16: return (double)*(const uint16*)p;
		case _hint32:
		case _uint32: return (double)*(const uint32*)p;
		case _hint64:
		case _uint64: return (double)int64tox(*(const uint64*)p);
		case _int8: return (double)*(const int8*)p;
		case _int16: return (double)*(const int16*)p;
		case _int32: return (double)*(const int32*)p;
		case _int64: *(int64*)b = *(const int64*)p;
			return (double)int64tox(*(int64*)b);
		case _uflt8: return uflt2dbl(*(const uint8*)p);
		case _bool: return *(const int8*)p!=0;
		case _ufrc8: return fix2double(*(const uint8*)p,8,3);
		case _ufrc16: return *(const uint16*)p/65536.;
		case _ufrc32: return *(const uint32*)p/4294967296.;
		case _ifix16: return fix2double(*(const int16*)p,-8,3);
		case _ifrc16: return fix2double(*(const int16*)p,-10,4);
		case _itck16: z = *(const int16*)p; goto ftck;
		//case _itck32: z = *(const int32*)p;
		ftck:	return z/30.;
		case _ifixed: return *(const int32*)p/65536.;
		case _ufixed: return *(const uint32*)p/65536.;
		case _ifract: return *(const int32*)p/1073741824.;
		case _ufract: return *(const uint32*)p/1073741824.;
		case _flt16:
			((int16*)b)[BIG_ENDIAN] = 0;
			((int16*)b)[!BIG_ENDIAN] = *(const int16*)p;
			return *(float*)b;
		case _float: return *(const float*)p;
		case _dbl32:
			((int32*)b)[BIG_ENDIAN] = 0;
			((int32*)b)[!BIG_ENDIAN] = *(const int32*)p;
			return *(double*)b;
		case _double: return *(const double*)p;
		case _ext80: return f80to64(*(const s_float80*)p);
		case _hpf16: return half2float(*(const u_fhalf16*)p);

		default: goto end;
		}
	if (z<0)
		z =-z, p = 1+(const int8*)p;
	if (parse_number(p,z+(const char*)p,(double*)b,1))
		return *(double*)b;
end:	return NAN;
}

/* ---- parsing */

static int parsehex(const char *s,
	const char *e,
	int8 *d,
	int z)
{
	int		t = 0;
	uint16		v = 0x80, c;

	if (s<e && *s=='#')
		++s;
	while(z>0)
	{
		if (s>=e)
			return _e_op_error;
		if ((c=*s++)=='\t' || c==' ')
			continue;
		if ((c-='0')<=9u)
			v = v<<4|c;
		else if ((c-='A'-'0')<=25u || (c-='a'-'A')<=25u)
			c+= 10, v = v<<4|c;
		else
			return _e_op_error;
		if (v&0x8000)
			d[t++] = (int8)v, v = 0x80, --z;
	}
	return parse_skipwsp(s,e)!=e ? _e_op_error:t;
}

static int equalstr(const char *d,
	const char *s)
{
	while(*s++)
		if (s[-1]!=tolower((uint8)*d++))
			return 0;
	return 1;
}

int set_tdata(int32 *type,
	const tobj *s,
	int8 *p,
	int8 lit,
	int8 first)
{
	const char	*u, *e, *n;
	s_float80	lv;
	double		v;
	int8		*d;
	int32		z, t, typ;

	if (!type)
		return _e_read_error;
	if ((d=p)==NULL)
		return _e_write_error;
	if (s)
		u = s->s, z = u ? s->z&0x7FFFFFFF:0;
	else
		u = 0, z = 0;
	if ((typ=*type)<0)
	{
		typ|=~511;
		if (!first)
			typ|=~287, typ&=~256;
	}
	t = ABS(typ), e = u+z;
	if (typ<-255)
	{
		int8		uns;

		if (typ>=_hint64)
			if (lit)
			{
				if (z>8)
					goto bdat;
				if (parse_literal(u,(int64*)&v,z)<0)
					{t = _e_range_error; goto end;}
				goto t5;
			} else if ((typ<=_uint64 || typ==_int64)
					&& (n=readanyint64(u,e,(int64*)&v))!=NULL
					&& (n>=e || *n!=mLocaleNPtChar))
			{
			t5:	switch(typ&3)
				{
				default: *d = (int8)int64lo(*(int64*)&v), ++d; break;
				case 3: *(int16*)d = (int16)int64lo(*(int64*)&v), d+= 2; break;
				case 2: *(int32*)d = (int32)int64lo(*(int64*)&v), d+= 4; break;
				case 1: *(int64*)d = *(int64*)&v, d+= 8;
				}
				goto end;
			}
		uns = 0;
		if (*type==_rgb24 && (*u=='#' || memchr(u,' ',z)))
		{
			if ((t=parsehex(u,e,d,3))>=0)
				d+= t;
			*type =~0;
			goto end;
		}
		if ((n=parse_float(u,e,&lv,1))!=NULL)
		{
			v = f80to64(lv);
			if (*(u=n)=='u' || *u=='U')
				++u, uns = 1, ++n;
		} else if (z==4)
			if (equalstr(u,"true"))
				v = 1;
			else if (equalstr(u,"-inf"))
				v =-INFINITY;
			else if (typ==_styl8 && equalstr(u,"bold")) //(typ|1)==_styl16
				v = 1;
			else if (typ==_ifrc16 && equalstr(u,"none"))
				v =-.001;
			else
				goto tbool;
		else if (z==5 && equalstr(u,"false"))
			v = 0;
		else if (z==3 && equalstr(u,"inf"))
			v = INFINITY;
		else if (z==6 && typ==_styl8 && equalstr(u,"italic")) //(typ|1)==_styl16
			v = 2; /* italic */
		else
		tbool:	if (typ==_bool)
				v = (double)(s && s->z ? *u:0);
			else
				goto bdat;
		switch(typ)
		{
		case _ifract: v*= 16384;
		case _ifixed: v*= 65536, v+= v<0 ? -.5:.5;
		case _int32:
			if (v>IMAX(int32))
				*(int32*)d = IMAX(int32);
			else
				*(int32*)d = (int32)v;
			goto nchk;
		case _ufrc16: v*= 65536, v+= .5;
		case _uint16:
			if (v<0 && uns)
				v+= 65536;
			if (v>65535)
				*(uint16*)d = 65535;
			else
				*(uint16*)d = (uint16)v;
			goto nchk;
		case _int8:
			if (v>IMAX(int8))
				*(int8*)d = IMAX(int8);
			else
				*(int8*)d = (int8)v;
			goto nchk;
		case _itck16:
			if (uns)
				goto r16;
			v*= 30./1024;
		case _ifrc16:
			if (uns)
				goto r16;
			v*= 4;
		case _ifix16: v*= 256;
			v+= v<0 ? -.5:.5;
		case _int16:
		r16:	if (v>IMAX(int16))
				*(int16*)d = IMAX(int16);
			else
				*(int16*)d = (int16)v;
			goto nchk;
		case _int64:
			if (v>=-IMIN64)
				int64sethl(*(int64*)d,IMAX(int32),~0ul);
			else
				int64xset(*(int64*)d,v);
			goto nchk;
		case _uflt8:
			if (v<0 && uns)
				v+= 4294967296.;
			//v = dbl2uflt(v);
			//goto u8;
			*(uint8*)d = dbl2uflt(v);
			goto nchk;
		case _alpha8:
			v*= 255./256;
		case _ufrc8:
			v*= 256, v+= .5;
		case _uint8:
			if (v<0 && uns)
				v+= 256;
			*(uint8*)d = v>255 ? 255:(uint8)v;
			goto nchk;
		case _hint8: case _styl8:
			*(uint8*)d = (int32)v;
			goto nchk;
		case _hint16: *(uint16*)d = (int32)v; goto nchk;
		case _ufrc32: v*= 4;
		case _ufract: v*= 16384;
		case _ufixed: v*= 65536, v+= .5;
		case _uint32:
			if (v<0 && uns)
				v+= 4294967296.;
			if (v>~0ul)
				*(uint32*)d =~0ul;
			else
				*(uint32*)d = (uint32)v;
			goto nchk;
		case _hint32: /* fmod method is too expensive */
			int64xset(*(int64*)&v,v);
			*(uint32*)d = int64lo(*(int64*)&v);
			goto nchk;
		case _uint64:
			if (v<0 && uns)
				v+=-2*IMIN64;
			if (v>=-2*IMIN64)
				int64sethl(*(int64*)d,~0ul,~0ul);
			else
		case _hint64: int64xset(*(uint64*)d,v); goto nchk;
		case _bool: *(int8*)d = v!=0; goto nchk;
		case _float: *(float*)d = (float)v; goto nchk;
		case _double: *(double*)d = v; goto nchk;
		case _flt16: *(float*)&v = (float)v;
			if (~*(int32*)&v&0x7F800000)
				*(int32*)&v+= 0x8000;
			*(int16*)d = ((int16*)&v)[!BIG_ENDIAN];
			goto nchk;
		case _dbl32: if (~((int32*)&v)[!BIG_ENDIAN]&0x7FF00000)
				((int32*)&v)[!BIG_ENDIAN]+= ((uint32*)&v)[BIG_ENDIAN]>>31;
			*(int32*)d = ((int32*)&v)[!BIG_ENDIAN];
			goto nchk;
		case _ext80: *(s_float80*)d = lv;
			goto nchk;
		case _hpf16: *(u_fhalf16*)d = float2half((float)v);
		nchk:	if (n)
			{
				while(*n==' ' || *n==9)
					++n;
				if (memchr("!@$%^&*()<>?{}[]+~|",*n,19))
					t = _e_general_error;
				else if (*n>126)
					t = _e_unknown_symbol;
			}
			break;
		bdat:
		default: t = _e_op_error;
			goto end;
		}
		d+= _typSzArray[typ+287];
	} else
	{
		if (typ<0)
			++d;
		memset(d,0,ABS(typ)); /* zero out undefined chars - C term. auto */
		if (z>t)
			z = t;
		if (typ<0)
			d[-1] = (char)z;
		memcpy(d,u,z);
		d+= t;
	}
end:	if ((*type>>=first ? 9:5)>~31u)
		*type = 0;
	return t<0 ? t:d-p;
}

const int8 *read_typ_be(void *p,
	const int8 *s,
	int32 typ)
{
	s_float80	n;
	int32		z;
	int8		*d = p;
	int		h = 9;

	if (typ>=0)
		{_amcpy(d,s,typ); goto end;}
	z = typ|~511;
	goto start;
	do
	{
		h = 5, z = (typ|(~287|~511))&~256;
	start:	switch(z)
		{
		case _int8:
		case _uint8:
		case _hint8:
		case _bool:
		case _ufrc8:
		case _uflt8:
		case _alpha8:
		case _styl8:
			*d++ = *s++;
			break;
		case _int16:
		case _uint16:
		case _hint16:
		case _ufrc16:
		case _flt16:
		case _hpf16:
		case _ifix16:
		case _ifrc16:
		case _itck16:
			*(uint16*)d = (uint16)va_be16(s), d+= 2;
			break;
		case _int32:
		case _uint32:
		case _hint32:
		case _ifixed:
		case _ufixed:
		case _ifract:
		case _ufract:
		case _ufrc32:
		case _dbl32:
		case _float:
			*(uint32*)d = va_be32(s), d+= 4;
			break;
		case _int64:
		case _uint64:
		case _hint64:
		case _double:
			((uint32*)&n.man)[!BIG_ENDIAN] = va_be32(s);
			((uint32*)&n.man)[BIG_ENDIAN] = va_be32(s);
			*(int64*)d = n.man, d+= 8;
			break;
		case _ext80:
			n.sgnexp = (int16)va_be16(s);
			((uint32*)&n.man)[!BIG_ENDIAN] = va_be32(s);
			((uint32*)&n.man)[BIG_ENDIAN] = va_be32(s);
			*(s_float80*)d = n, d+= 10;
			break;

		default: if (--z>-257)
			{
				z =-z;
				_amcpy(d,s,z);
			}
		}
	} while((typ>>=h)<~30u);
end:	return s;
}

#define tcase(t)		case _##t:return#t;

const char *elem_tname(int16 type)
{
	switch(type)
	{
	tcase(int8)
	tcase(int16)
	tcase(int32)
	tcase(int64)

	tcase(uint8)
	tcase(uint16)
	tcase(uint32)
	tcase(uint64)

	tcase(hint8)
	tcase(hint16)
	tcase(hint32)
	tcase(hint64)

	tcase(bool)
	tcase(ifrc16)
	tcase(float)
	tcase(double)

	tcase(flt16)
	tcase(hpf16)
	tcase(dbl32)
	tcase(ext80)

	tcase(ifixed)
	tcase(ufixed)
	tcase(ifract)
	tcase(ufract)

	tcase(alpha8)
	tcase(ufrc8)
	tcase(ufrc16)
	tcase(ufrc32)

	tcase(styl8)
	tcase(uflt8)
	tcase(ifix16)
	tcase(itck16)

	case _tilist: return "_list";
	case _tifunc: return "_func";

	default:
		if (type>_int8)
			return "_pstr";
		if (type>=_tiref)
			return "_ref";
		return "!invalid!";
	}
}

/* ---- basic t-obj printing */

int32 tprint_string(sio_func_t f,
	void *out,
	const char *s,
	int32 z,
	int t,
	int w)
{
	uint8		b;

	if (z<0)
	{
		z =-z;
		if (*(const uint8*)s<z)
			z = *(const uint8*)s;
		++s;
	} else /* don't print trailing nulls (C-string default) */
		while(z)
			if (s[--z])
				{++z; break;}
	if (z && w && parse_entrybreak(s,s+z,t)==s+z)
		return f(s,z,out);
	b = '"';
	t = f(&b,1,out);
	t+= tprint_tchars(f,out,s,z,'"',1,1);
	return t+f(&b,1,out);
}

static int sprint_tnum(char *b,
	const void *p,
	int type)
{
	double		v;
	int		m, r;

	switch(type)
	{
	case _hint8: r = *(const uint8*)p; goto h10;
	case _hint16: r = *(const uint16*)p; goto h10;
	case _hint32: r = *(const uint32*)p;
	h10:	*b = '0', b[1] = 'x';
		return 2+makebasestring(b+2,NULL,r,16,1,0);
	case _hint64:
		return sprintf(b,"0x%llX",
			*(const uint64*)p);
	case _int8: r = *(const int8*)p; goto i10;
	case _int16: r = *(const int16*)p; goto i10;
	case _int32: r = *(const int32*)p;
	i10:	return makebasestring(b,NULL,r,10,1,'-');
	case _int64:
		*(int64*)b = *(const int64*)p;
		return sprintf(b,"%lld",
			*(int64*)b);
	case _uint8:
	case _styl8: r = *(const uint8*)p; goto u10;
	case _uflt8:
		return sprintf(b,"%.3g",
			uflt2dbl(*(const uint8*)p));
	case _uint16: r = *(const uint16*)p; goto u10;
	case _uint32: r = *(const uint32*)p;
	u10:	return makebasestring(b,NULL,r,10,1,0);
	case _uint64:
		return sprintf(b,"%llu",
			*(const uint64*)p);
	case _bool: r = *(const int8*)p!=0; goto u10;
	case _ufrc8: v = fix2double(*(const uint8*)p,8,3);
		goto xfs;
	case _ufrc16: r = *(const uint16*)p, m = 16; goto frc;
	case _ufrc32: m = 32; goto x30;
	case _ifix16: v = fix2double(*(const int16*)p,-8,3);
		goto xfs;
	case _ifrc16:
		if (*(const int16*)p==-1)
		{
			*b = '-', b[1] = '1', b[2] = 'u', b[3] = 0;
			return 3;
		}
		v = fix2double(*(const int16*)p,-10,4);
		goto xfs;
	case _itck16: r = *(const int16*)p;
		m = 0;
		if (r<0)
			r =-r, m = 1, *b = '-';
		m+= makebasestring(b+m,NULL,r/30,10,0,0);
		if (r%=30)
		{
			b[m++] = mLocaleNPtChar;
			m+= makebasestring(b+m,NULL,(r*20+3)/6,10,2,0);
			if (b[m-1]=='0')
				--m;
		}
		b[m] = 0;
		break;
	case _ifixed: m =-16; goto ifrc;
	case _ufixed: m = 16;
	ifrc:	r = *(const int32*)p;
	frc:	v = fix2double(r,m,5);
		goto xfs;
	case _ifract: m =-30; goto x30;
	case _ufract: m = 30;
	x30:	v = fix2double(*(const int32*)p,m,10);
	xfs:	return sprintf(b,"%.10g",
			v);
	case _hpf16: *(float*)b = half2float(*(const u_fhalf16*)p); goto flt;
	case _flt16:
		((int16*)b)[BIG_ENDIAN] = 0;
		((int16*)b)[!BIG_ENDIAN] = *(const int16*)p;
	flt:	return sprintf(b,"%.4g",
			*(float*)b);
	case _float:
		return sprintf(b,"%.8g",
			*(const float*)p);
	case _dbl32:
		((int32*)b)[BIG_ENDIAN] = 0;
		((int32*)b)[!BIG_ENDIAN] = *(const int32*)p;
		return sprintf(b,"%.8g",
			*(double*)b);
	case _ext80:
		return sprint_float(b,*(const s_float80*)p,21,20,mLocaleNPtChar,0,0);
	case _double: v = *(const double*)p;
		return sprint_double(b,v,18,mLocaleNPtChar,0);
	case _alpha8:
		return sprintf(b,"%.3g",
			fix2double(*(const uint8*)p*8421505u>>16,15,3));

	default: return 0;
	}
	return m;
}

int32 tprint_tdata(sio_func_t f,
	void *out,
	int32 typ,
	const void *p,
	int w,
	int32 maxlen)
{
	char		b[32], *s;
	int32		z, m, t = 0;
	int		h = 9;

	goto start;
	do
	{
		h = 5, *b = ',';
		t+= f(b,1,out);
	start:	if ((z=typ)<0)
		{
			z|=~511;
			if (h&4)
				z|=~287, z&=~256;
		}
		if (z>-256)
			{t+= tprint_string(f,out,p,MIN(z,maxlen),1,w); continue;}
		m = sprint_tnum(b,p,z);
		/* contains self-modifying expression to remove unnecessary zero */
		s = b;
		if ((*b=='-' && b[1]=='0' && b[2]==mLocaleNPtChar && (b[1]='-',1))
				|| (*b=='0' && b[1]==mLocaleNPtChar))
			++s, --m;
		t+= f(s,m,out);
	} while(p=elem_tsize(z)+(int8*)p,(typ>>=h)-1<~31u);
	return t;
}

/* ---- t-item printing */

#if INTERP_VERS>0x199
int32 tprint_titem(sio_func_t f,
	void *out,
	const titem *ti,
	int idbreak,
	int32 maxlen)
{
	char		b[64];
	const char	*s;
	const void	*p;
	int32		u, t;

	if (!f || !ti)
		return -1;
	if (!TI_IS_ITEM(ti))
	{
		u = sprint_evalue(b,ti->num,17);
		return f(b,MIN(u,maxlen),out);
	}
	p = TI_GETPTR(ti);
	u = TI_GETTYPE(ti);
	if (u>_tilist)
		return tprint_tdata(f,out,TI_IS_ALLOC(ti) ? ti->pos:u,p,idbreak,maxlen);
	if (TI_UTYPE(u)<13)
		return tprint_tdata(f,out,TI_UTYPE(u),p,idbreak,maxlen);
	s = elem_tname(u);
	*b = '(', t = f(b,1,out);
	t+= f(s,strlen(s),out);
	*b = ')', t+= f(b,1,out);
	return t;
}
#endif
